#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "helpers.h"
#include <stdint.h>
#include "lib/include/zmq.h"
#include "lib/include/czmq.h" 

uint32_t state[] = {4,1,2,9,99};

uint32_t id_count = 0;

static void *zocket = NULL;



int receive_number(zmsg_t *response) {
    char *nr_s = zmsg_popstr(response);
    int nr;
    int rc = sscanf(nr_s,"%d",&nr);
    assert(rc==1);
    free(nr_s);
    return nr;
}

void print_chunk_array(const ChunkArray arr) {
    int i;
    printf("size: %ld\n", arr.size);
    for (i = 0; i < arr.size; i++) {
        puts(arr.chunks[i].data);
    }
}

void print_row(const MatrixRow row) {
    printf("Transition group: %s\n", row.transition_group.data);
    print_chunk_array(row.variables);
}

void print_matrix(const Matrix m) {
    int i;
    for (i = 0; i < m.nr_rows; i++) {
        print_row(m.rows[i]);
    }
}


InitialResponse send_init(void) {
    zmsg_t *request = zmsg_new();
    zmsg_addstr(request,"init");
    zmsg_addstr(request,"%d",id_count);
    zmsg_send(&request, zocket);
    zmsg_destroy(&request);
    zmsg_t *response = zmsg_recv(zocket);
    assert(response);

    InitialResponse resp = get_init_response(response);

    assert(zmsg_size(response) == 0);

    puts("transition groups:");
    print_chunk_array(resp.transition_groups);
    puts("variables");
    int i;
    for (i = 0; i < resp.variables.size; i++) {
        printf("%s (%s)\n", resp.variables.chunks[i].data, resp.variable_types.chunks[i].data);
    }
    print_chunk_array(resp.variables);
    puts("state labels");
    //print_chunk_array(resp.state_labels);
    print_matrix(resp.state_labels);
    

    puts("May Write Matrix:");
    print_matrix(resp.may_write);
    puts("Must Write Matrix:");
    print_matrix(resp.must_write);
    puts("Reads Action Matrix:");
    print_matrix(resp.reads_action);
    puts("Reads Guard Matrix:");
    print_matrix(resp.reads_guard);

    zmsg_destroy(&response);
    return resp;
}


// Note: size is part of the result, it will contain the number of states.
State *send_next_state(State s, char *transitiongroup, int *size) {
    zmsg_t *request = zmsg_new();
    zmsg_addstr(request,"get-next-state");
    zmsg_addstr(request,"%d",id_count);
    zmsg_addstr(request,transitiongroup);

    put_state(request,s);
    zmsg_send(&request, zocket);
    zmsg_destroy(&request);
    zmsg_t *response = zmsg_recv(zocket);
    drop_frame(response);
    drop_frame(response);

    char *nr_of_states_s = zmsg_popstr(response);
    sscanf(nr_of_states_s,"%d",size);
    free(nr_of_states_s);

    State *successors = malloc(sizeof(State) * (*size));
    int i;
    for(i=0;i<(*size);i++) {
        successors[i] = get_state(response);
    }
    zmsg_destroy(&response);
    return successors;
}


int send_get_state_label(State s, char *label) {
    zmsg_t *request = zmsg_new();
    zmsg_addstr(request,"get-state-label");
    zmsg_addstr(request,"%d",id_count);
    zmsg_addstr(request,label);
    put_state(request,s);
    zmsg_send(&request, zocket);
    zmsg_destroy(&request);
    zmsg_t *response = zmsg_recv(zocket);
    drop_frame(response);
    drop_frame(response);


    char *result_s = zmsg_popstr(response);
    int res;
    sscanf(result_s,"%d",&res);

    free(result_s);
    zmsg_destroy(&response);
    return res;
}

void terminate(void) {
    zmsg_t *request = zmsg_new();
    zmsg_addstr(request,"terminate");
    zmsg_addstr(request,"%d",id_count);
    zmsg_send(&request, zocket);
    zmsg_t *response = zmsg_recv(zocket);
    zmsg_destroy(&response);
    zmsg_destroy(&request);
}

void start_ltsmin(void) {

    InitialResponse initial_resp = send_init();


    // Some Demo code

    /* Get the successors of the initial state. 
     DA$init_state is the transition group 
     representing the initialization. The 
     prototype does initializing and constant setup in one step.
 
     I don't know how LTSmin handles this, but DA$init_state is 
     always the only transition group that is possible in the initial state
     and it is never possible to execute DA$init_state in any other state.

     ProB automatically adds it as the first transition group.
    */

    int nr_successors;
    State *successors = send_next_state(initial_resp.initial_state, "DA$init_state", &nr_successors);


    // Send a get label request using the last state of the previous request
    State foo = successors[nr_successors-1];

    // DAinvariant is hard wired because I am too lazy to store it in the init code.
    // However, ProB does send the state label. 
    int val = send_get_state_label(foo,"DAinvariant");
    printf("invariant Result: %d\n",val);

    Matrix state_labels = initial_resp.state_labels;
    int i;
    for (i = 0; i < state_labels.nr_rows; i++) {
        // even though the field is called transition group, it is the state label here
        Chunk invariant = state_labels.rows[i].transition_group;
        char *invariant_data = invariant.data;
        val = send_get_state_label(foo, invariant_data);
        printf("%s Result: %d\n",invariant_data, val);
    }

    terminate();

    destroy_initial_response(&initial_resp);
    // Cleanup
    for (i=0; i<nr_successors; i++) {
        destroy_state(&(successors[i]));
    }
    free(successors);

}

int main(int argc, char *args[]) {

    zctx_t *ctx = zctx_new();
    zocket = zsocket_new (ctx, ZMQ_REQ);
    int rc = zsocket_connect(zocket, "ipc:///tmp/ltsmin.probz");
    assert(rc==0);

    start_ltsmin();
    zsocket_disconnect(zocket, "ipc:///tmp/ltsmin.probz");
    zsocket_destroy(ctx, zocket);
    zctx_destroy(&ctx);

    return 0;
}
