:- module(ltsmin,[ltsmin_init/2,ltsmin_init/4,ltsmin_loop/1, ltsmin_teardown/2]).

:- use_module(library(fastrw),[ fast_buf_read/2,fast_buf_write/3 ]).
:- use_module(library(codesio), [write_to_codes/2]).
:- use_module(library(lists)).


:- use_module('../../src/module_information.pl').

:- use_module(probsrc(specfile)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(b_read_write_info), [tcltk_read_write_matrix/1]).

:- use_module(probsrc(typing_tools)).

:- module_info(group,experimental).
:- module_info(description,'This is the interface to C code for LTSmin integration.').
:- use_module(probsrc(b_state_model_check), [get_guard/2]).
:- use_module(probsrc(bsyntaxtree), [conjunction_to_list/2, same_texpr/2,
                                     conjunct_predicates/2, is_truth/1, create_negation/2]).

foreign(put_chunk, put_chunk(+address(zmsg_t),+address(char),+integer)).
foreign(ltsmin_initialize, ltsmin_initialize(+string, [-address(void)])).
%foreign(ltsmin_loop, ltsmin_loop(+address(void))).
foreign(h_zocket2, h_zocket2(+address(void), [-integer])).
foreign(ltsmin_teardown, ltsmin_teardown(+address(void), +string)).
foreign(put_number, put_number(+address(zmsg_t),+integer)).
foreign(create_message, create_message([-address(zmsg_t)])).
foreign(append_message, append_message(+address(zmsg_t),+address(zmsg_t))).

foreign_resource(ltsmin,
                 [put_chunk,ltsmin_initialize,h_zocket2,/*ltsmin_loop,*/ltsmin_teardown,put_number,create_message,append_message]).

ltsmin_loop(Zocket) :-
    h_zocket2(Zocket, Ret),
    Ret == 0, !, ltsmin_loop(Zocket).
ltsmin_loop(_).



:- dynamic loaded/0.
:- dynamic varname_list/1.

:- use_module(probltlsrc(ltl_tools), [temporal_parser/3]).

write_to_ltl_file(Formula, Path) :-
    open(Path, write, Stream),
    tell(Stream),
    print(Formula),
    told.

setup_ltl_formula(Formula, _) :-
    var(Formula), !.
setup_ltl_formula(Formula, OutputFile) :-
    temporal_parser(Formula, ltl, ParsedFormula),
    ltl_safety:pp_formula_to_ltl2ba_syntax(ParsedFormula, ResAtom),
    write_to_ltl_file(ResAtom, OutputFile).

ltsmin_init(Endpoint,Zocket, LtlFormula, LtlFileOutput) :-
    ltsmin_init(Endpoint, Zocket),
    setup_ltl_formula(LtlFormula, LtlFileOutput).
ltsmin_init(Endpoint,Zocket) :-
    loadfr,
    ltsmin_initialize(Endpoint, Zocket).

loadfr :-
    (loaded -> true
     ; (assert(loaded),
        %assert_dir,
        load_foreign_resource(library(ltsmin)))).
%assert_dir :- (user:library_directory('.') -> true ;
%assert(user:library_directory('.'))).


write_and_put_chunk(Msg, C) :-
    fast_buf_write(C,L,A),
    put_chunk(Msg, A, L).


put_termlist_elements([], _Msg).
put_termlist_elements([H|T], Msg) :-
    write_and_put_chunk(Msg, H),
    put_termlist_elements(T, Msg).

put_termlist(Msg, TermList) :-
    length(TermList,ListSize),
    put_number(Msg, ListSize),
    put_termlist_elements(TermList, Msg).



put_successors([], _Msg).
put_successors([H|T], Msg) :-
    get_state_chunks(H, Vals),
    put_termlist(Msg, Vals),
    put_successors(T, Msg).

get_state_chunks(csp_and_b(CSP,BState),[CSP|Vals]) :- !,
   % TO DO: partition CSP into chunks
   maplist(arg(2), BState, Vals).
get_state_chunks(BState,Vals) :-
   maplist(arg(2), BState, Vals).  % extract values from environment of the form [bind(x,Vx),...]

get_next_states(Msg, State,Op) :- % State is a list of bindings
    compute_successors(Op, State, Succs),
    length(Succs,NrStates),
    %format(" ~w successor states for ~w~n  ~w~n",[NrStates,Op,Succs]),
    put_number(Msg, NrStates),
    put_successors(Succs, Msg).


read_state([],[]).
read_state([Addr|T], [Arg|R]) :- read_state(T,R), fast_buf_read(Arg,Addr).

:- meta_predicate mnf1(0).
mnf1(Call) :- %format('Calling ~w~n',[Call]), flush_output,
  if( on_exception(E,Call,(format('Exception: ~w~n',[E]),flush_output,fail)),
      true, (format('Failed ~w~n~n',[Call]),flush_output,fail)).

:- use_module(probsrc(specfile),[csp_with_bz_mode/0]).
% convert a list of values into a ProB environment [bind(x,Vx),...]
generate_bound_state([CSP|BState],Res) :- csp_with_bz_mode,!,
    Res=csp_and_b(CSP,BoundState),
    varname_list(L),
    %print(gen_bind(BState,L)),nl,
    generate_bindings(BState, L, BoundState).
generate_bound_state(BState,BoundState) :-
    varname_list(L),
    generate_bindings(BState, L, BoundState).

generate_bindings([], [], []) :- !.
generate_bindings([H|T], [Name|TN], [bind(Name, H)|R]) :-
    generate_bindings(T, TN, R).

get_successors(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(get_successors_aux(Msg, OperationA, StateA)).
get_successors_aux(Msg, OperationA, StateA) :-
    read_state(StateA,State),
    %print(generate(State)),nl,
    generate_bound_state(State,BoundState),
    fast_buf_read(Operation,OperationA),
    (debug:debug_mode(on)
     -> format('~n ** get_successors for ~w~n',[Operation]),
        translate:print_state(BoundState),nl
     ; true),
    get_next_states(Msg, BoundState,Operation).

get_state_label(Msg, LabelA, StateA) :- mnf1(get_state_label_aux(Msg, LabelA, StateA)).
get_state_label_aux(Msg, LabelA, StateA) :- 
    read_state(StateA, State),
    generate_bound_state(State,BoundState),
    %print(get_state_label_state(BoundState)),nl,
    fast_buf_read(Label,LabelA), %print(get_state_label(Label)),nl,
    eval_label(Msg, BoundState,Label).

% NOTE: LTSMin assumes the first transition group is the initialisation action, it is called only once
% all other events are not called in the initial state
get_transition_groups(['$init_state'|L]) :-
    (specfile:b_or_z_mode
     -> findall(OperationName, bmachine:b_top_level_operation(OperationName), BL),
        (csp_with_bz_mode -> L = ['$CSP'|BL] ; L=BL)
     ;  L=[]).  % FIXME: provide list for CSP-M

extract_name_and_type([],[], []).
extract_name_and_type([b(identifier(HT),TypeTerm,_)|T], [HT|RT], [TypeTerm|RType]) :-
    extract_name_and_type(T,RT,RType).

/* is that already defined somewhere? */
get_ltsmin_variable_names(Names, Types) :-
    bmachine:b_get_machine_variables(V),
    % bmachine:b_get_machine_all_constants(C), % this includes constants which have been moved to deferred sets; causes issue with hanoi model
    bmachine:b_get_machine_constants(C), % this does not include global set constants
    append(C,V,CV),
    extract_name_and_type(CV,Names,Types).




extract_tg_and_nth_entry([], _, R, R).
extract_tg_and_nth_entry([list(H)|T], Arg, Acc, R) :-
    nth1(1, H, TransitionGroup),
    nth1(Arg, H, X),
    extract_tg_and_nth_entry(T, Arg, [[TransitionGroup, X]|Acc], R).


put_rows([], _Msg).
put_rows([[TransitionGroup, VarList]|R], Msg) :-
    write_and_put_chunk(Msg, TransitionGroup),
    put_termlist(Msg, VarList),
    put_rows(R, Msg).

% add $CSP as read/write dependency to matrixes
% [ [Odd,[]], [Even,[]], [Inc,[count]], [Reset,[count]]]
patch_matrix(M,Res) :- csp_with_bz_mode, !,
   Res = [ ['$CSP',['$CSP_CHUNK']] | PM], % add virtual CSP transition for CSP only events; TO DO: also allow/treat other pure CSP events; the CSP transition can only modify the $CSP chunk
   maplist(patch_row,M,PM).
patch_matrix(M,M).

patch_row([Op,Vars], [Op,['$CSP_CHUNK'|Vars]]).

% patch matrix for CSP||B
put_patched_matrix(Msg, Type,M) :- 
    patch_matrix(M,PM),
    (debug:debug_mode(on) -> print(put_matrix(Type,PM)),nl ; true),
    write_matrix(Msg, PM, Type).

put_may_write_matrix(Msg, M) :- csp_with_bz_mode,!,
    write_matrix(Msg, [ ['$CSP',[]] | M], may_write_matrix).
put_may_write_matrix(Msg, M) :- write_matrix(Msg, M, may_write_matrix).

put_matrix(Msg, M) :-
    length(M, MLength),
    put_number(Msg, MLength),
    put_rows(M, Msg).

build_initial_state(_, VType, VVal) :-
    any_value_for_type(VType, VVal).


term_to_atom(T, A) :-
    write_to_codes(T, Codes),
    atom_codes(A, Codes).


:- use_module(probsrc(predicate_debugger), [get_operation_propositional_guards/3]).


guard_seen(G, [H|_]) :- 
    same_texpr(G, H), !.
guard_seen(G, [_|LofG]) :-
    guard_seen(G, LofG).
    

setify_guards(L, R) :-
    setify_guards(L, [], R).
setify_guards([], Seen, Seen).
setify_guards([G|RGuards], Seen, Res) :-
    (guard_seen(G, Seen)
      -> setify_guards(RGuards, Seen, Res)
      ;  setify_guards(RGuards, [G|Seen], Res)).


guard_vector(ListOfListOfGuards, Res) :-
    append(ListOfListOfGuards, ListOfGuards),
    setify_guards(ListOfGuards, Res).


get_name([H|_], [Name|_], G, Name) :-
    same_texpr(G, H), !.
get_name([_|GV], [_|Names], G, R) :-
    get_name(GV, Names, G, R).


guards_to_indices(GuardVector, Names, ListOfGuards, ListOfIndices) :-
    maplist(get_name(GuardVector, Names), ListOfGuards, ListOfIndices).

split_guards(['$init_state'|TransitionGroups], GuardVector, Names, [[]|IndexVectors]) :-
    %% TODO: normalise guards
    maplist(get_operation_propositional_guards, TransitionGroups, RGuards, _Bodies),
    guard_vector(RGuards, GuardVector),
    generate_names(GuardVector, 'guard_', Names), % get list [guardN, ..., guard1]
    maplist(guards_to_indices(GuardVector, Names), RGuards, IndexVectors).


create_matrix([], [], Res, Res) :- !.
create_matrix([H1|T1], [H2|T2], Acc, Res) :-
    create_matrix(T1, T2, [[H1,H2]|Acc], Res).
create_matrix(Headings, Entries, Res) :-
    create_matrix(Headings, Entries, [], Res).


generate_names([], _, _, Acc, Acc) :- !.
generate_names([_|T], Prefix, N, Acc, Res) :-
    number_codes(N, NC),
    append(Prefix, NC, NamedC),
    atom_codes(Named, NamedC),
    N1 is N+1,
    generate_names(T, Prefix, N1, [Named|Acc], Res).
generate_names(Terms, PrefixAtom, Names) :-
    atom_codes(PrefixAtom, PrefixChars),
    generate_names(Terms, PrefixChars, 1, [], Names).

wrap_named_inv(Type, Name, Invariant, state_label(Name, Type, Invariant)).
:- dynamic state_label/3.


assert_all([]).
assert_all([H|T]) :-
    assert(H), assert_all(T).

/* need to extract this from the atomic props of the formula,
   we can also use the prolog term of a formula as its state label */
get_state_labels(ListOfConjuncts, Names, Type, StateLabelMatrix) :-
    %% format of StateLabelMatrix is
    % [[label1, [id1, id2, ...]]
    %  [label2, [id2, id4, ...]]
    %  ...]
    maplist(wrap_named_inv(Type), Names, ListOfConjuncts, Facts),
    assert_all(Facts),
    maplist(bsyntaxtree:predicate_identifiers, ListOfConjuncts, SLSVars),
    create_matrix(Names, SLSVars, StateLabelMatrix).


split_invariant(ListOfInvariantConjuncts, InvNames) :-
    bmachine:b_get_invariant_from_machine(Invariant),!,
    conjunction_to_list(Invariant, ListOfInvariantConjuncts),
    generate_names(ListOfInvariantConjuncts, 'inv', InvNames).





submsg_type_to_int(initial_state, 1).
submsg_type_to_int(transition_group_list, 2).
submsg_type_to_int(variables_name_list, 3).
submsg_type_to_int(variables_type_list, 4).
submsg_type_to_int(state_label_matrix, 5).
submsg_type_to_int(may_write_matrix, 6).
submsg_type_to_int(must_write_matrix, 7).
submsg_type_to_int(reads_action_matrix, 8).
submsg_type_to_int(reads_guard_matrix, 9).
submsg_type_to_int(guard_info_matrix, 10).
submsg_type_to_int(guard_label_matrix, 11). % split invariant and guard matrices
submsg_type_to_int(necessary_enabling_set, 12).
submsg_type_to_int(necessary_disabling_set, 13).
submsg_type_to_int(do_not_accord_matrix, 14).
submsg_type_to_int(may_be_coenabled_matrix, 15).
submsg_type_to_int(ltl_label_matrix, 16). 




write_chunklist(Msg, List, Type) :-
    create_message(FreshMsg),
    submsg_type_to_int(Type, HeaderNum),
    put_number(FreshMsg, HeaderNum),
    put_termlist(FreshMsg, List),
    append_message(Msg, FreshMsg).

write_matrix(Msg, Matrix, Type) :-
    create_message(FreshMsg),
    submsg_type_to_int(Type, HeaderNum),
    put_number(FreshMsg, HeaderNum),
    put_matrix(FreshMsg, Matrix),
    append_message(Msg, FreshMsg).


:- use_module(probporsrc(static_analysis), [test_path/6, dependent_actions/5]).

does_not_accord_with_aux([], _, []).
does_not_accord_with_aux([H|T], TransitionGroup, [H|R]) :-
    dependent_actions(H, TransitionGroup, 0, 300, _), !, % TODO: handle interrupts?
    does_not_accord_with_aux(T, TransitionGroup, R).
does_not_accord_with_aux([_|T], TransitionGroup, R) :-
    does_not_accord_with_aux(T, TransitionGroup, R).

do_not_accord(['$init_state'|TransitionGroups], [[]|DoNotAccordMatrix]) :-
    maplist(does_not_accord_with_aux(TransitionGroups), TransitionGroups, DoNotAccordMatrix).

can_enable(TransitionGroup, Guard, NGuard) :-
    test_path(NGuard, [TransitionGroup], Guard, 0, 300, Result), % TODO: what if Result is something else, like "unknown" or a variable?
    (Result == ok ; Result == timeout).

calculate_enabling_set_aux([], _Guard, _NGuard, []) :- !.
calculate_enabling_set_aux([Transition|Rest], Guard, NGuard, [Transition|ResList]) :-
    can_enable(Transition, Guard, NGuard), !,
    calculate_enabling_set_aux(Rest, Guard, NGuard, ResList).
calculate_enabling_set_aux([_|Rest], Guard, NGuard, ResList) :-
    calculate_enabling_set_aux(Rest, Guard, NGuard, ResList).

calculate_enabling_set(TransitionGroupList, Guard, Res) :-
    create_negation(Guard, NGuard),
    calculate_enabling_set_aux(TransitionGroupList, Guard, NGuard, Res).
    
necessary_enabling_set(GuardVector, ['$init_state'|TransitionGroups], R) :-
    maplist(calculate_enabling_set(TransitionGroups), GuardVector, R).

necessary_disabling_set(GuardVector, TransitionGroups, R) :-
    maplist(create_negation, GuardVector, NGuardVector),
    necessary_enabling_set(NGuardVector, TransitionGroups, R).


may_be_coenabled_guards(G1, _G2Name, G2) :-
    conjunct_predicates([G1, G2], Conjunction),
    is_truth(Truth),
    test_path(Conjunction, [], Truth, 0, 300, CheckResult), % TODO: what if Result is something else, like "unknown" or a variable?
     (CheckResult == ok ; CheckResult == timeout).
         % fail on interrupt
         
may_be_coenabled2([], [], []).
may_be_coenabled2([_], [_], []).
may_be_coenabled2([H|T], [Name|NameRest], [[Name,R]|RT]) :-
    include(may_be_coenabled_guards(H), NameRest, T, R),
    may_be_coenabled2(T, NameRest, RT).
may_be_coenabled(GuardVector, Names, MayBeCoEnabledMatrix) :-
    % calculates an upper triangle matrix without the elements on the diagonal
    may_be_coenabled2(GuardVector, Names, MayBeCoEnabledMatrix).

:- dynamic(guard_labels/1).


get_label_group_guards(Msg, State) :-
    guard_labels(Labels),
    maplist(eval_label(Msg, State), Labels).


get_state_label_group(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(get_state_label_group_aux(Msg, OperationA, StateA)).
get_state_label_group_aux(Msg, TypeA, StateA) :-
    read_state(StateA,State),
    %print(generate(State)),nl,
    generate_bound_state(State,BoundState),
    fast_buf_read(Type,TypeA),
    (debug:debug_mode(on)
     -> format('~n ** get_state_label_group for ~w~n',[Type]),
        translate:print_state(BoundState),nl
     ; true),
    (Type == '1' -> get_label_group_guards(Msg, BoundState)
                 ;  print(unhandled_group_type(Type)),nl,fail).

get_ltl_labels(ASTs, Names) :-
    findall(ap_representation(Hash, AST), ltl_safety:ap_representation(Hash, AST), APReprs),
    maplist(arg(2), APReprs, ASTs),
    maplist(arg(1), APReprs, Names).


:- use_module(probsrc(error_manager), [add_error_fail/3]).

op_to_reads_guard(Op, ReadSet) :-
    op_guards(Op, GuardList),
    maplist(guard_reads, GuardList, ReadSets),
    append(ReadSets, ReadSet).

oplist_to_reads_guards(OpList, ReadSet) :-
    maplist(op_to_reads_guard, OpList, ReadSets),
    append(ReadSets, ReadSet).
% identifiers might be contained multiple times due to append
get_ltl_readsets(bpred(X), ReadSet) :-
    bsyntaxtree:predicate_identifiers(X, ReadSet).
get_ltl_readsets(enabled(Op), ReadSet) :-
    op_to_reads_guard(Op, ReadSet). % get readset of guards of op
get_ltl_readsets(det(OpList), ReadSet) :-
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(dlk(OpList), ReadSet) :- 
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(ctrl(OpList), ReadSet) :- !,
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(deadlock, ReadSet) :-
    get_transition_groups(OpList),
    oplist_to_reads_guards(OpList, ReadSet).
get_ltl_readsets(AP,_) :- !,
    add_error_fail(ltsmin, 'operation is not supported for LTSmin', AP).

:- dynamic(guard_reads/2).

assert_guard_readsets([]).
assert_guard_readsets([[GuardName, ReadSet]|R]) :-
    assert(guard_reads(GuardName, ReadSet)),
    assert_guard_readsets(R).

:- dynamic(op_guards/2).

assert_op_guards([], []).
assert_op_guards([Op|T], [GuardList|T2]) :-
    assert(op_guards(Op, GuardList)),
    assert_op_guards(T, T2).

ltl_to_state_label_matrix(ListOfAPs, LtlApNames, LtlStateLabelMatrix) :-
    %% format of StateLabelMatrix is
    % [[label1, [id1, id2, ...]]
    %  [label2, [id2, id4, ...]]
    %  ...]
    maplist(wrap_named_inv(ltl_ap), LtlApNames, ListOfAPs, Facts),
    assert_all(Facts),
    maplist(get_ltl_readsets, ListOfAPs, SLSVars),
    create_matrix(LtlApNames, SLSVars, LtlStateLabelMatrix).


calc_and_append_por_info(0, _Msg, _, _, _) :- !.
calc_and_append_por_info(_Por, Msg, GuardVector, GuardNames, TransitionGroups) :- 
    may_be_coenabled(GuardVector, GuardNames, MayBeCoEnabledMatrix),
    write_matrix(Msg, MayBeCoEnabledMatrix, may_be_coenabled_matrix),

    necessary_enabling_set(GuardVector, TransitionGroups, NES),
    create_matrix(GuardNames, NES, NesMatrix),
    write_matrix(Msg, NesMatrix, necessary_enabling_set),

    necessary_disabling_set(GuardVector, TransitionGroups, NDS),
    create_matrix(GuardNames, NDS, NdsMatrix),
    write_matrix(Msg, NdsMatrix, necessary_disabling_set),
    
    do_not_accord(TransitionGroups, DNA),
    create_matrix(TransitionGroups, DNA, DnaMatrix),
    write_matrix(Msg, DnaMatrix, do_not_accord_matrix).



get_init(Msg, Por) :-
            print(pl_init_for_lts_min),nl,flush_output,
            get_ltsmin_variable_names(VS, VTypes),
            retractall(varname_list(_)),
            assert(varname_list(VS)),
            maplist(build_initial_state, VS, VTypes, BInitialState),
            %maplist(any_value_for_type,VTypes,InitialState),
            maplist(term_to_atom, VTypes, VTypeAtoms),
            get_transition_groups(TransitionGroups),
            (csp_with_bz_mode
              -> InitialState = [stop(no_loc_info_available) | BInitialState],  % DUMMY Values
                 Vars = ['$CSP_CHUNK'|VS], Types = ['csp_process'|VTypeAtoms]
               ; InitialState = BInitialState,
                 Vars = VS, Types = VTypeAtoms
            ),
            print(InitialState),nl,
            write_chunklist(Msg, InitialState, initial_state),
            write_chunklist(Msg, TransitionGroups, transition_group_list),
            write_chunklist(Msg, Vars, variables_name_list), % we use $CSP_CHUNK as name for the CSP chunk
            write_chunklist(Msg, Types, variables_type_list),


            print('working on matrix'),nl,
            %% read write matrices
            tcltk_read_write_matrix(M),
            (debug:debug_mode(on) -> print(M),nl,flush_output ; true),
            M = list([_Headings|MR]),
            (debug:debug_mode(on) -> print(MR),nl ; true),
            extract_tg_and_nth_entry(MR, 5, [],  MayWriteMatrix),
            extract_tg_and_nth_entry(MR, 4, [],  MustWriteMatrix),
            extract_tg_and_nth_entry(MR, 3, [],  ReadsActionMatrix),
            extract_tg_and_nth_entry(MR, 2, [],  ReadsGuardMatrix),
            put_may_write_matrix(Msg, MayWriteMatrix),
            put_patched_matrix(Msg, must_write_matrix,MustWriteMatrix),
            put_patched_matrix(Msg, reads_action_matrix,ReadsActionMatrix),
            put_patched_matrix(Msg, reads_guard_matrix,ReadsGuardMatrix),

            %% split invariant to state labels
            split_invariant(ListOfInvariantConjuncts, InvNames),
            get_state_labels(ListOfInvariantConjuncts, InvNames, invariant, InvStateLabelMatrix),

            %% split guards to state labels
            split_guards(TransitionGroups, GuardVector, GuardNames, IndexVectors),
            get_state_labels(GuardVector, GuardNames, guard, GuardStateLabelMatrix),

            % memoize mapping OpName -> GuardNames
            assert_op_guards(TransitionGroups, IndexVectors),
            % memoize mapping GuardName -> ReadSet(Guard)
            assert_guard_readsets(GuardStateLabelMatrix),
            
            reverse(GuardNames, GuardNamesInOrder),
            assert(guard_labels(GuardNamesInOrder)),

            get_ltl_labels(ListOfLabelASTs, LtlApNames),
            ltl_to_state_label_matrix(ListOfLabelASTs, LtlApNames, LtlStateLabelMatrix),

            %% write all state labels (invariant conjuncts and guard conjuncts)
            write_matrix(Msg, InvStateLabelMatrix, state_label_matrix),
            write_matrix(Msg, GuardStateLabelMatrix, guard_label_matrix),
            write_matrix(Msg, LtlStateLabelMatrix, ltl_label_matrix),

            %% create guard info matrix, i.e. which transition depends on which guards
            create_matrix(TransitionGroups, IndexVectors, GuardInfoMatrix),
            write_matrix(Msg, GuardInfoMatrix, guard_info_matrix),
            calc_and_append_por_info(Por, Msg, GuardVector, GuardNames, TransitionGroups).



compute_successors('$init_state',_,Result) :- !,
    findall(S,init_state(S),Result), 
    (debug:debug_mode(on) -> print(res_init_state(Result)), nl, flush_output ; true).
compute_successors(OpName,L,Result) :- 
    (debug:debug_mode(on) -> nl,print(compute_successors(OpName)),nl,flush_output ; true),
    findall(S,specfile:specfile_trans(L,OpName,_,S,_,_),Result).
    % this in turn will call b_execute_top_level_operation_update for normal operations


next_update(Msg, OperationA, StateA) :- %print(prolog_get_succs),nl,flush_output,
    mnf1(next_update_aux(Msg, OperationA, StateA)).
next_update_aux(Msg, OperationA, StateA) :-
    read_state(StateA,State),
    %print(generate(State)),nl,
    generate_bound_state(State,BoundState),
    fast_buf_read(Operation,OperationA),
    (debug:debug_mode(on)
     -> format('~n ** get_successors for ~w~n',[Operation]),
        translate:print_state(BoundState),nl
     ; true),
    get_next_update(Msg, BoundState,Operation).

:- use_module(probsrc(preferences),[temporary_set_preference/3,reset_temporary_preference/2]).
get_next_update(Msg, State,Op) :- % State is a list of bindings
    temporary_set_preference(ltsmin_do_not_evaluate_top_level_guards,true,CHNG),
    get_next_states(Msg, State,Op),
    reset_temporary_preference(ltsmin_do_not_evaluate_top_level_guards,CHNG).

init_state(Res) :- specfile:specfile_trans(root,_,_,S,_,_),
                   init_state2(S,Res). % , nl,print(init(Res)),nl.

init_state2(csp_and_b_root,Res) :- !,
    specfile:specfile_trans(csp_and_b_root,_,_,S,_,_),
    init_state2(S,Res).
init_state2(concrete_constants(L),FullRes) :- !,
    specfile:specfile_trans(concrete_constants(L),_,_,Res,_,_),
    expand_const_and_vars_to_full_store(Res,FullRes).
init_state2(R,R).

eval_label(Msg, State,invariant) :-
    bmachine:b_get_invariant_from_machine(Invariant),!,
    (b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],State) ->
        put_number(Msg, 1); put_number(Msg, 0)).


eval_label(Msg, State,Label) :-
    (debug:debug_mode(on) -> print(eval(State,Label)),nl,flush_output ; true),
    state_label(Label, Type, AST),
    eval_label2(Type, Msg, State, AST).


%:- use_module(probltlsrc(ltl_prepositions), [check_ap/2]).
eval_label2(ltl_ap, Msg, State, AST) :-
    (check_ap_bound_state(AST,State) ->
         put_number(Msg, 1); put_number(Msg, 0)).
eval_label2(_, Msg, State, AST) :-
    (b_interpreter:b_test_boolean_expression_for_ground_state(AST,[],State) ->
         put_number(Msg, 1); put_number(Msg, 0)).
%  foreign predicate  put_number(Msg, X) :: 0 = false, 1=true


op_enabled(Op, State) :-
    get_guard(Op, Guard), % TODO: guard WITH existentially quantified parameter
    b_interpreter:b_test_boolean_expression_for_ground_state(Guard, [], State).

% check an atomic proposition in a given node (i.e. state)
check_ap_bound_state(bpred(Property),Node) :-
    b_interpreter:b_test_boolean_expression_for_ground_state(Property, [], Node).
check_ap_bound_state(enabled(Op),Node) :- !,
    op_enabled(Op, Node).
check_ap_bound_state(det(OpList),Node) :- !,
    max_enabled(OpList, Node, 1, 0, R),
    R < 2.
check_ap_bound_state(dlk(OpList),Node) :- !,
    max_enabled(OpList, Node, 0, 0, 0).
check_ap_bound_state(ctrl(OpList),Node) :- !,
    max_enabled(OpList, Node, 1, 0, 1).
check_ap_bound_state(deadlock,Node) :- !,
    get_transition_groups(OpList),
    check_ap_bound_state(dlk(OpList, Node)).
    
%max_enabled(OpList, Max, EnabledSoFar, ResEnabled)
max_enabled([], _State, _Max, R, R).
max_enabled([Op|T], State, Max, EnabledSoFar, R) :-
    (op_enabled(Op, State)
     -> Enabled is EnabledSoFar + 1, 
        Max >= Enabled,
        max_enabled(T, State, Max, Enabled, R)
     ;  max_enabled(T, State, Max, EnabledSoFar, R)).


