#include "z3++.h"
#include <sicstus/sicstus.h>
#include <unordered_map>
#include <vector>

using namespace z3;
using namespace std;

static context c;
static solver smtt(c);

vector<expr> prolog_list_to_expr_vector(SP_term_ref list);
string prolog_term_to_string(SP_term_ref term);

z3::sort mk_type(SP_term_ref type);
SP_integer insert_id(expr x);
SP_integer insert_id(func_decl x);

extern "C" {
    static unordered_map<string, z3::sort> type_table;

    static unordered_map<SP_integer, expr> id_table;
    static unordered_map<SP_integer, const func_decl> func_table;
    static SP_integer id_counter = 2;

    Z3_symbol mk_couple_name = Z3_mk_string_symbol(c, "couple");
    Z3_symbol couple_proj_names[2];

    void init(void) {
        id_table.insert(make_pair<SP_integer,expr>(0,c.bool_val(false)));
        id_table.insert(make_pair<SP_integer,expr>(1,c.bool_val(true)));

        // types
        type_table.insert(make_pair<string,z3::sort>("integer",c.int_sort()));
        type_table.insert(make_pair<string,z3::sort>("boolean",c.bool_sort()));

        // couples
        couple_proj_names[0] = Z3_mk_string_symbol(c, "get_x");
        couple_proj_names[1] = Z3_mk_string_symbol(c, "get_y");
    }

    SP_integer mk_var(SP_term_ref type, char *varname) {
        z3::sort t = mk_type(type);
        expr x = c.constant(varname, t);
        return insert_id(x);
    }

    SP_integer mk_bounded_var(SP_term_ref type, char *varname) {
        return mk_var(type,varname);
    }

    SP_integer mk_empty_set(SP_term_ref type) {
        // z3 api: mk_empty_set gets the domains / subtypes sort!
        SP_term_ref arg = SP_new_term_ref();
        int ret = SP_get_arg(1,type,arg);
        assert(ret);
        z3::sort t = mk_type(arg);
        expr emptyset = to_expr(c,Z3_mk_empty_set(c,t));
        return insert_id(emptyset);
    }

    SP_integer mk_full_set(SP_term_ref type) {
        try {
            // z3 api: mk_full_set gets the domains / subtypes sort!
            SP_term_ref arg = SP_new_term_ref();
            int ret = SP_get_arg(1,type,arg);
            assert(ret);
            z3::sort t = mk_type(arg);
            expr fullset = to_expr(c,Z3_mk_full_set(c,t));
            return insert_id(fullset);
        } catch(z3::exception& e) {
            cout << "z3exception in mk_full_set: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    SP_integer mk_int_const(int val) {
        try {
            expr x = c.int_val(val);
            return insert_id(x);
        }
        catch(z3::exception& e) {
            cout << "z3exception in mk_int_const: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    SP_integer mk_bool_const(char *val) {
        if (0 == strcmp(val, "boolean_true")) {
            return 1;
        } else if (0 == strcmp(val, "boolean_false")) {
            return 0;
        } else {
            // explode
            assert(0);
        }
    }

    SP_integer mk_record_const(SP_term_ref type, SP_term_ref fields) {
        z3::sort record_type = mk_type(type);
        Z3_func_decl constructor = Z3_get_tuple_sort_mk_decl(c,record_type);

        vector<expr> fields_exprs = prolog_list_to_expr_vector(fields);

        z3::array<Z3_ast> arr(fields_exprs.size());
        for(unsigned int i = 0; i < fields_exprs.size(); i++) {
            arr[i] = fields_exprs[i];
        }

        expr x = to_expr(c,Z3_mk_app(c,constructor,arr.size(),arr.ptr()));
        return insert_id(x);
    }

    SP_integer mk_record_field(SP_integer record, SP_integer field_nr, SP_atom field_name) {
        expr record_expr = id_table.at(record);

        func_decl record_select = to_func_decl(c,Z3_get_tuple_sort_field_decl(c,record_expr.get_sort(),field_nr));
        expr x = record_select(record_expr);
        return insert_id(x);
    }

    SP_integer mk_couple(SP_term_ref type, SP_integer arg1, SP_integer arg2) {
        try {
            z3::sort sort = mk_type(type);
            // cout << "sort " << sort << endl;
            expr a = id_table.at(arg1);
            expr b = id_table.at(arg2);

            func_decl constructor = to_func_decl(c,Z3_get_tuple_sort_mk_decl(c,sort));
            return insert_id(constructor(a,b));
        } catch(z3::exception& e) {
            cout << "z3exception in mk_couple " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    SP_integer mk_op(char *opname, SP_integer arg1, SP_integer arg2) {
        try {
            if(id_table.count(arg1) == 0) {
                func_decl a = func_table.at(arg1);
                expr b = id_table.at(arg2);
                if(strcmp(opname, "function") == 0) {
                    return insert_id(a(b));
                }
            } else {
                expr a = id_table.at(arg1);
                expr b = id_table.at(arg2);

                if(strcmp(opname, "modulo") == 0) {
                    return insert_id(to_expr(c,Z3_mk_rem(c,a,b)));
                }
                if(strcmp(opname, "add") == 0) {
                    return insert_id(a + b);
                }
                if(strcmp(opname, "minus") == 0) {
                    return insert_id(a - b);
                }
                if(strcmp(opname, "power_of") == 0) {
                    return insert_id(pw(a,b));
                }
                if(strcmp(opname, "div") == 0) {
                    // smtt.add(b > 0);
                    return insert_id(ite(a > 0,a / b,ite(b > 0,(a / b)+1,(a / b)-1)));
                }
                if(strcmp(opname, "multiplication") == 0) {
                    return insert_id(a * b);
                }
                if(strcmp(opname, "equal") == 0) {
                    return insert_id(a == b);
                }
                if(strcmp(opname, "implication") == 0) {
                    return insert_id(implies(a,b));
                }
                if(strcmp(opname, "equivalence") == 0) {
                    return insert_id(a == b);
                }
                if(strcmp(opname, "greater") == 0) {
                    return insert_id(a > b);
                }
                if(strcmp(opname, "greater_equal") == 0) {
                    return insert_id(a >= b);
                }
                if(strcmp(opname, "less") == 0) {
                    return insert_id(a < b);
                }
                if(strcmp(opname, "less_equal") == 0) {
                    return insert_id(a <= b);
                }
                if(strcmp(opname, "member") == 0) {
                    return insert_id(to_expr(c,Z3_mk_set_member(c,a,b)));
                }
                if(strcmp(opname, "subset") == 0) {
                    return insert_id(to_expr(c,Z3_mk_set_subset(c,a,b)));
                }
                if(strcmp(opname, "union") == 0) {
                    Z3_ast uargs[2] = {a,b};
                    return insert_id(to_expr(c,Z3_mk_set_union(c,2,uargs)));
                }
                if(strcmp(opname, "intersection") == 0) {
                    Z3_ast uargs[2] = {a,b};
                    return insert_id(to_expr(c,Z3_mk_set_intersect(c,2,uargs)));
                }
                if(strcmp(opname, "set_subtraction") == 0) {
                    return insert_id(to_expr(c,Z3_mk_set_difference(c,a,b)));
                }
            }
        } catch(z3::exception& e) {
            cout << "z3exception in mk_op: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
        }

        cout << "Operator missing in mk_op: " << opname << endl;
        assert(false); // operator missing
    }

    SP_integer mk_quantifier(char *opname, SP_term_ref list, SP_integer ex) {
        vector<expr> args = prolog_list_to_expr_vector(list);
        expr inner = id_table.at(ex);
        try {
            expr_vector vars(c);
            for(unsigned int i = 0; i< args.size(); i++) {
                vars.push_back(args[i]);
            }

            if(strcmp(opname, "forall") == 0) {
                expr r = forall(vars, inner);
                return insert_id(r);
            }
            if(strcmp(opname, "exists") == 0) {
                expr r = exists(vars, inner);
                return insert_id(r);
            }
        } catch(z3::exception& e) {
            cout << "z3exception in mk_quantifier: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
        }
        assert(false);
    }

    SP_integer mk_op_arglist(char *opname, SP_term_ref list) {
        vector<expr> args = prolog_list_to_expr_vector(list);
        try {
            if(strcmp(opname, "negation") == 0) {
                return insert_id(!args.at(0));
            }
            if(strcmp(opname, "unary_minus") == 0) {
                return insert_id(-args.at(0));
            }
            if(strcmp(opname, "if_then_else") == 0) {
                return insert_id(ite(args.at(0),args.at(1),args.at(2)));
            }
            if(strcmp(opname, "conjunct") == 0) {
                expr x = expr(args.at(0));
                for (unsigned int i = 1; i < args.size(); i++) {
                    x = x && args.at(i);
                }
                return insert_id(x);
            }
            if(strcmp(opname, "disjunct") == 0) {
                expr x = expr(args.at(0));
                for (unsigned int i = 1; i < args.size(); i++) {
                    x = x || args.at(i);
                }
                return insert_id(x);
            }
            if(strcmp(opname, "distinct") == 0) {
                z3::array<Z3_ast> arr(args.size());
                for(unsigned int i = 0; i< args.size(); i++) {
                    arr[i] = args[i];
                }

                return insert_id(to_expr(c,Z3_mk_distinct(c, args.size(), arr.ptr())));
            }
        }
        catch(z3::exception& e) {
            cout << "z3exception in mk_op_arglist" << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
        }
        cout << "Operator missing in mk_op_arglist: " << opname << endl;
        assert(false); // operator missing
    }

    SP_term_ref add_interpreter_constraint(SP_integer spe, SP_atom name, SP_integer timeout) {
        SP_term_ref return_value = SP_new_term_ref();
        try{
            // receives a new constraint with a given name,
            // adds it to the smt solver,
            // checks for a contradiction
            // returns unsat(ListOfNamesInUnsatCore), sat or unknown

            expr x = id_table.at(spe);
            // cout << "expr x is " << x << endl;
            smtt.add(x,SP_string_from_atom(name));

            params p(c);
            p.set("timeout", (unsigned int)timeout);
            p.set("unsat-core", true);
            smtt.set(p);

            check_result r = smtt.check();
            if(r == check_result::sat) {
                SP_put_atom(return_value, SP_atom_from_string("sat"));
                return return_value;
            } else if(r == check_result::unsat) {
                SP_term_ref list = SP_new_term_ref();
                expr_vector core = smtt.unsat_core();
                for (unsigned i = 0; i < core.size(); i++) {
                    SP_term_ref head = SP_new_term_ref();
                    std::ostringstream out;
                    out << core[i];
                    int ret = SP_put_atom(head, SP_atom_from_string(out.str().c_str()));
                    assert(ret);
                    ret = SP_cons_list(list,head,list);
                    assert(ret);
                }
                int ret = SP_cons_functor(return_value, SP_atom_from_string("unsat"), 1, list);
                assert(ret);
                return return_value;
            }
        }
        catch(z3::exception& e) {
            cout << "z3exception in add_interpreter_constraint: " << e.msg() << endl;

            if(strcmp(e.msg(),"canceled") != 0) { // occurs on timeout
                SP_term_ref eterm = SP_new_term_ref();
                SP_term_ref arg = SP_new_term_ref();

                int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
                assert(ret);
                ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                                      1, arg);
                assert(ret);

                SP_raise_exception(eterm);
            }
        }
        SP_put_atom(return_value, SP_atom_from_string("unknown"));
        return return_value;
    }

    SP_atom query(SP_integer spe, SP_integer timeout) {
        try {
            expr x = id_table.at(spe);
            // cout << "expr x is " << x << endl;

            params p(c);
            p.set("timeout", (unsigned int)timeout);
            smtt.set(p);

            smtt.push();
            smtt.add(x);
            // cout << "calling smtt.check()" << x << endl;
            check_result r = smtt.check();
            // cout << "smtt.check() done" << x << endl;
            smtt.pop();
            if(r == check_result::sat) {
                // cout << "Z3 found model: " << smtt.get_model() << endl;
                return SP_atom_from_string("sat");
            } else if(r == check_result::unsat) {
                return SP_atom_from_string("unsat");
            }
            return SP_atom_from_string("unknown");
        } catch(z3::exception& e) {
            cout << "z3exception in query: " << e.msg() << endl;

            if(strcmp(e.msg(),"canceled") != 0) { // occurs on timeout
                SP_term_ref eterm = SP_new_term_ref();
                SP_term_ref arg = SP_new_term_ref();

                int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
                assert(ret);
                ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                                      1, arg);
                assert(ret);
                SP_raise_exception(eterm);
            }
            return SP_atom_from_string("unknown");
        }
    }

    void print_functs(model m) {
        for(unsigned int i=0;i<m.num_funcs();i++) {
            func_decl decl = m.get_func_decl(i);
            func_interp interp = m.get_func_interp(decl);

            cout << "found interp for: " << decl.name() << endl;
            for(unsigned int j=0;j<interp.num_entries();j++) {
                func_entry entry = interp.entry(j);
                cout << "entry: " << entry.arg(0) << " -> " << entry.value() << endl;
            }
            cout << "default: " << interp.else_value() << endl;
        }
    }

    SP_atom get_string(SP_integer spexpr) {
        try {
            expr x = id_table.at(spexpr);
            // cout << "for expr:" << x << endl;
            expr model = smtt.get_model().eval(x);

            //print_functs(smtt.get_model());

            std::ostringstream out;
            out << model;
            return SP_atom_from_string(out.str().c_str());
        } catch(z3::exception& e) {
            cout << "z3exception in get_string: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            return SP_atom_from_string("no_atom_here"); // to make the compiler happy
        }
    }

    SP_atom get_full_model_string(void) {
        try {
            std::ostringstream out;
            out << smtt.get_model() << endl;
            return SP_atom_from_string(out.str().c_str());
        } catch(z3::exception& e) {
            cout << "z3exception in get_full_model_string: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();
            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
                                  1, arg);
            assert(ret);
            SP_raise_exception(eterm);
            return SP_atom_from_string("no_atom_here"); // to make the compiler happy
        }
    }

    SP_integer mk_set(SP_term_ref list) {
        try {
            vector<expr> exprs = prolog_list_to_expr_vector(list);

            z3::sort dom_type = exprs[0].get_sort();
            expr theset = to_expr(c,Z3_mk_empty_set(c,dom_type));

            for(unsigned int i=0; i < exprs.size(); i++) {
                theset = to_expr(c,Z3_mk_set_add(c,theset,exprs[i]));
            }
            return insert_id(theset);
        }
        catch(z3::exception& e) {
            cout << "z3exception in mk_set: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    void mk_sort(SP_atom name) {
        const char* setname = SP_string_from_atom(name);
        // cout << "mk_sort with name " << setname << endl;
        if(type_table.count(setname) == 0) {
            z3::sort type = to_sort(c,Z3_mk_uninterpreted_sort(c,Z3_mk_string_symbol(c, setname)));
            type_table.insert(pair<string,z3::sort>(string(setname),type));
        }
    }

    void mk_sort_with_cardinality(SP_atom name, SP_integer cardinality) {
        const char* setname = SP_string_from_atom(name);
        // cout << "new sort with cardinality " << cardinality << " named " << setname << endl;
        if(type_table.count(setname) == 0) {
            // cout << "new sort added to table" << endl;
            z3::sort type = to_sort(c,Z3_mk_finite_domain_sort(c,Z3_mk_string_symbol(c, setname),cardinality));

            type_table.insert(pair<string,z3::sort>(string(setname),type));
        }
    }

    void push_frame(void) {
        try {
            smtt.push();
        } catch(z3::exception& e) {
            cout << "z3exception in push_frame: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    void pop_frame(void) {
        try {
            smtt.pop();
        } catch(z3::exception& e) {
            cout << "z3exception in pop_frame: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }

    void reset(void) {
        smtt.reset();
        id_table.clear();
        func_table.clear();
        id_counter = 2;
        init();
    }

    void setup_enumerated_set(SP_term_ref list) {
        try {
            // the set identifiers are distinct. the cardinality is already defined
            vector<expr> ids = prolog_list_to_expr_vector(list);
            z3::array<Z3_ast> arr(ids.size());
            for(unsigned int i = 0; i < ids.size(); i++) {
                arr[i] = ids[i];
            }

            // can only be distinct if more than one given
            if(ids.size() > 1) {
                expr distinct = to_expr(c,Z3_mk_distinct(c, ids.size(), arr.ptr()));
                smtt.add(distinct);
            }

            // each is enforced to be equal to a corresponding uninterpreted constant
            // if this is a finite domain sort (i.e. was finite all along, not shrunken by prob)
            z3::sort the_sort = ids.begin()->get_sort();
            if(the_sort.is_finite_domain()) {
                int num = 0;
                for (vector<expr>::iterator it = ids.begin(); it != ids.end(); ++it) {
                    expr uc = c.num_val(num,the_sort);
                    expr equal = *it == uc;
                    smtt.add(equal);
                    num++;
                }
            }
        } catch(z3::exception& e) {
            cout << "z3exception in setup_enumerated_set: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),1, arg);
            assert(ret);

            SP_raise_exception(eterm);
        }
    }

    SP_integer conjoin_negated_state(SP_term_ref list, SP_integer old_expr) {
        try {
            vector<expr> exprs = prolog_list_to_expr_vector(list);
            vector<expr> negated_exprs;

            for (vector<expr>::iterator it = exprs.begin(); it != exprs.end(); ++it) {
                expr not_equal = *it != smtt.get_model().eval(*it);
                negated_exprs.push_back(not_equal);
            }

            if(negated_exprs.size() == 1) {
                expr new_expr = negated_exprs.front() && id_table.at(old_expr);
                return insert_id(new_expr);
            } else {
                expr negstate = expr(negated_exprs.at(0));
                for (unsigned int i = 1; i < negated_exprs.size(); i++) {
                    negstate = negstate && negated_exprs.at(i);
                }
                expr new_expr = negstate && id_table.at(old_expr);
                return insert_id(new_expr);
            }
        } catch(z3::exception& e) {
            cout << "z3exception in conjoin_negated_state: " << e.msg() << endl;
            SP_term_ref eterm = SP_new_term_ref();
            SP_term_ref arg = SP_new_term_ref();

            int ret = SP_put_atom(arg,SP_atom_from_string(e.msg()));
            assert(ret);
            ret = SP_cons_functor(eterm, SP_atom_from_string("z3_exception"),
            1, arg);
            assert(ret);

            SP_raise_exception(eterm);
            assert(false);
        }
    }
}

z3::sort mk_type(SP_term_ref type) {
    int ret;
    // simple atom types 'integer' and 'boolean'
    if(SP_is_atom(type)) {
        SP_atom as_atom;
        ret = SP_get_atom(type,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        z3::sort to_return = type_table.at(string(typestr));
        if(type_table.count(typestr) == 0) {
            SP_printf("Type missing in table: %s\n",typestr);
        } else {
            return to_return;
        }
    }

    // types that have been constructed before
    string type_as_string = prolog_term_to_string(type);
    if(type_table.count(type_as_string) > 0) {
        // cout << "found type for " << type_as_string << endl;
        return type_table.at(string(type_as_string));
    }
    // otherwise deconstruct type
    SP_atom name;
    int arity;
    ret = SP_get_functor(type,&name,&arity);
    assert(ret);

    // sets
    const char *atom = SP_string_from_atom(name);
    if(strcmp(atom, "set") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        z3::sort subtype = mk_type(arg);
        type_table.insert(make_pair<string,z3::sort>(string(type_as_string),to_sort(c,Z3_mk_set_sort(c,subtype))));
        // cout << "created type for " << type_as_string << endl;
        return to_sort(c,Z3_mk_set_sort(c,subtype));
    }
    if(strcmp(atom, "global") == 0) {
        SP_term_ref arg = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg);
        assert(ret);
        SP_atom as_atom;
        ret = SP_get_atom(arg,&as_atom);
        assert(ret);
        const char *typestr = SP_string_from_atom(as_atom);
        return type_table.at(string(typestr));
    }
    if(strcmp(atom, "couple") == 0) {
        Z3_func_decl mk_tuple_decl;
        Z3_func_decl proj_decls[2];
        Z3_sort proj_sorts[2];

        SP_term_ref arg1 = SP_new_term_ref();
        ret = SP_get_arg(1,type,arg1);
        assert(ret);
        SP_term_ref arg2 = SP_new_term_ref();
        ret = SP_get_arg(2,type,arg2);
        assert(ret);
        proj_sorts[0] = mk_type(arg1);
        proj_sorts[1] = mk_type(arg2);
        Z3_sort pair_sort = Z3_mk_tuple_sort(c, mk_couple_name, 2, couple_proj_names, proj_sorts, &mk_tuple_decl, proj_decls);
        type_table.insert(make_pair<string,z3::sort>(string(type_as_string),to_sort(c,pair_sort)));
        // cout << "created type for " << type_as_string << endl;
        return to_sort(c,pair_sort);
    }
    if(strcmp(atom, "record") == 0) {
        // first argument is a list of field(Name,Type)
        SP_term_ref list_of_fields = SP_new_term_ref();
        ret = SP_get_arg(1,type,list_of_fields);
        assert(ret);

        // iterate over the list and place fields in vector
        vector<pair<Z3_symbol,z3::sort>> fields;
        SP_term_ref head = SP_new_term_ref();
        SP_term_ref tail = SP_new_term_ref();
        while (SP_get_list(list_of_fields, head, tail)) {
            SP_term_ref name = SP_new_term_ref();
            ret = SP_get_arg(1,head,name);
            assert(ret);
            ret = SP_get_arg(2,head,type);
            assert(ret);

            z3::sort field_type = mk_type(type);

            SP_atom as_atom;
            ret = SP_get_atom(name,&as_atom);
            assert(ret);
            const char *namestr = SP_string_from_atom(as_atom);
            Z3_symbol field_name = Z3_mk_string_symbol(c,namestr);

            fields.push_back(make_pair(field_name,field_type));
            list_of_fields = tail;
        }

        Z3_symbol const_name = Z3_mk_string_symbol(c, "record");
        Z3_func_decl record_constructor;

        z3::array<Z3_func_decl> projections(fields.size());

        z3::array<Z3_sort> proj_sorts(fields.size());
        z3::array<Z3_symbol> proj_names(fields.size());
        for(unsigned int i = 0; i < fields.size(); i++) {
            proj_names[i] = fields[i].first;
            proj_sorts[i] = fields[i].second;
        }

        Z3_sort record_sort = Z3_mk_tuple_sort(c, const_name, fields.size(), proj_names.ptr(), proj_sorts.ptr(), &record_constructor, projections.ptr());

        return to_sort(c,record_sort);
    }

    // TODO: further types?
    printf("Unsupported type: %s\n",atom);
    assert(false);
}

SP_integer insert_id(expr x) {
    SP_integer id = id_counter;
    id_table.insert(pair<SP_integer, expr>( id,  x));
    id_counter++;

    //cout << id << ": " << x << endl;
    return id;
}

SP_integer insert_id(func_decl x) {
    SP_integer id = id_counter;
    func_table.insert(pair<SP_integer, func_decl>( id,  x));
    id_counter++;

    //cout << id << ": " << x << endl;
    return id;
}

vector<expr> prolog_list_to_expr_vector(SP_term_ref list) {
    assert(SP_is_list(list));

    vector<expr> args;
    SP_term_ref head = SP_new_term_ref();
    SP_term_ref tail = SP_new_term_ref();
    while (SP_get_list(list, head, tail)) {
        assert(SP_is_integer(head));
        SP_integer id;
        SP_get_integer(head, &id);

        expr x = id_table.at(id);
        args.push_back(x);
        list = tail;
    }
    return args;
}

string prolog_term_to_string(SP_term_ref term) {
    SP_atom functor;
    int arity;
    SP_term_ref arg = SP_new_term_ref();

    int ret = SP_get_functor(term, &functor, &arity);
    assert(ret);
    if(arity == 0) {
        return string(SP_string_from_atom(functor));
    } else {
        string s = SP_string_from_atom(functor);
        s.append("(");
        for(int i=1;i<arity;i++) {
            ret = SP_get_arg(i,term,arg);
            assert(ret);
            s.append(prolog_term_to_string(arg));
            s.append(",");
        }
        ret = SP_get_arg(arity,term,arg);
        assert(ret);
        s.append(prolog_term_to_string(arg));
        s.append(")");
        return s;
    }
}
