#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>

#include "../../../lib/include/zmq.h"
#include "../../../lib/include/czmq.h"

#include "../../../common/hashtrie.h"
#include "../../../common/shm_common.h"
#include "../../../common/shm_cln.h"

#include "mctrl.h"
#include "mhash.h"
#include "mprint.h"

static int shutdown_check(void);
static int handle_results(zmsg_t *msg);

static void *hash_recv = NULL;
static void *hash_send = NULL;

static int max_states = 0;
size_t checked_states = 0;

extern volatile int nmbr_workers;
static char *filename_hashes = NULL;
static int fd_hashes = -1;
static int fd_transitions = -1; // TODO: for now, just dump it in a seperate file

static void *h_mhash (void *arg);

static void *local_printer;

char *init_hash_module(zctx_t *ctx, zmq_pollitem_t *p_hash_recv, int maximum_states, int port_start) {
#ifndef TMPFOLDER
#define TMPFOLDER "/var/tmp/"
#endif
    local_printer = zsocket_new(ctx, ZMQ_PUSH);
    zsocket_connect(local_printer, "inproc://printer");

    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :init_hash_module :args {:maximum_states %d :port_start %d}", maximum_states, port_start);
    hash_recv = zsocket_new (ctx, ZMQ_PULL);
    zsocket_bind (hash_recv, "tcp://*:%d", port_start + 1);
    hash_send = zsocket_new (ctx, ZMQ_PUB);
    zsocket_bind (hash_send, "tcp://*:%d", port_start);
    filename_hashes = tempnam(TMPFOLDER, "zmqh-");
    fd_hashes = open(filename_hashes, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    fd_transitions = open("transitions", O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH); 
    

    max_states = maximum_states;

    init_shm(ctx);

    *p_hash_recv = (zmq_pollitem_t) { hash_recv, 0, ZMQ_POLLIN };
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :init_hash_module");
    return filename_hashes;
}

void start_h_mhash(pthread_t *hashthread, int *strict_mode) {
    pthread_create(hashthread, NULL, h_mhash, strict_mode);
}

/* this should eventually terminate the master */
static void *h_mhash (void *arg) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :h_mhash");
    int strict_mode = *((int*) arg);

    // not enough states checked, wait for a first message
    while (shutdown_check() == 0 && zsocket_poll(hash_recv, -1) == true) {
        zmsg_t *out = zmsg_new();
        const int max = nmbr_workers;
        int count = 0;
        while (count < max && zsocket_poll(hash_recv, 0) == true) {
            zmsg_t *in = zmsg_recv(hash_recv);

            int has_results = handle_results(in);
            
            if (has_results && strict_mode) {
                zmsg_destroy(&in);
                zmsg_destroy(&out);
                goto exit;
            }
            


            zframe_t *checked = zmsg_last(in);
            char *from = zframe_strdup(checked);
            char readable_from[TRIE_HASHLEN * 2 + 1];
            print_readable_digest((unsigned char*) from, readable_from);

            zframe_t *frame;
            while ((frame = zmsg_pop(in)) != NULL) {
                char *digest = zframe_strdup(frame);
                if (is_checked((unsigned char*) digest + TRIE_HASHLEN)) {
                    int rc = write(fd_hashes, digest, TRIE_HASHLEN); // save the hashes of checked states
                    assert(rc == TRIE_HASHLEN);
                    checked_states++;
                    //if (checked_states % 100000 == 0) {
                    //    ctrl_memusage();
                    //}
                }
                if (debug_level <= DEBUG_VERBOSE && !is_checked((unsigned char*) digest + TRIE_HASHLEN)) {
                    char readable_to[TRIE_HASHLEN * 2 + 1];
                    print_readable_digest((unsigned char*) digest, readable_to);
                    zstr_sendf(local_printer, "{:type :trans :from :%s :to :%s}\n", readable_from, readable_to);
                }

                if (!is_checked((unsigned char*) digest + TRIE_HASHLEN)) {
                    write(fd_transitions, from, TRIE_HASHLEN);
                    write(fd_transitions, digest, TRIE_HASHLEN);
                }

                trie_put(digest, (unsigned char*) digest + TRIE_HASHLEN);
                free(digest);
                zmsg_append(out, &frame);
            }
            free(from);

            zmsg_destroy(&in);
            count++;
            //zstr_send(hash_recv, "ok");
        }

        zmsg_send(&out, hash_send);
        zmsg_destroy(&out);
    }
exit:
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :h_mhash");
    pthread_exit(NULL);
}

static int handle_results(zmsg_t *msg) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :handle_results");
    char *label = zmsg_popstr(msg);
    assert(strcmp(label, "Results:") == 0);
    free(label);

    int has_result = 0;

    zframe_t *frame;
    while ((frame = zmsg_pop(msg)) != NULL) {
        char *res = zframe_strdup(frame);
        if (strcmp(res, "Hashes:") != 0) {
            has_result = 1;
            zstr_sendf(local_printer, "{:type :result :term \"%s\"}\n", res);
            free(res);
            zframe_destroy(&frame);
        } else {
            free(res);
            zframe_destroy(&frame);
            break;
        } 
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :handle_results");
    return has_result;
}

static int shutdown_check(void) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :shutdown_check");
    int checked = trie_get_checked_states_count();
    int total = trie_count_elements();

    if (checked == total || (max_states > 0 && checked >= max_states)) {
        DEBUG_PRINT(DEBUG_NORMAL,,,":exit :shutdown_check :shutting_down true :checked %d", checked);
        return -1;
    }
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :shutdown_check :shutting_down false");
    return 0;
}


void teardown_hash_module(zctx_t *ctx) {
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":enter :teardown_hash_module");
    zsocket_destroy(ctx, hash_recv);
    zsocket_destroy(ctx, hash_send);
    int checked = trie_get_checked_states_count();
    int depth = get_max_depth();
    int states = teardown_hashtrie();

    zstr_sendf(local_printer, "{:type :done :multichecked %ld :checked %d :known %d :trie-depth %d}\n", checked_states - checked, checked, states, depth);
    close(fd_hashes);
    remove(filename_hashes);
    free(filename_hashes);
    DEBUG_PRINT(DEBUG_VERY_VERBOSE,,,":exit :teardown_hash_module");
    zsocket_disconnect(local_printer, "inproc://printer");
    zsocket_destroy(ctx, local_printer);
}

