#include "../include/zmq.h"
#include "../include/czmq.h"

#include "CuTest.h"
#include "../master/mwork.h"
#include "../common/hashtrie.h"
#include "../common/queue.h"
#include "../common/sha.h"

extern zctx_t *ctx;
static void *tmp_recv;
static void *tmp_send;

extern wQueue *assertions;
extern int port;

static void init(void) {
    init_hashtrie();
    tmp_recv = zsocket_new (ctx, ZMQ_DEALER);
    tmp_send = zsocket_new (ctx, ZMQ_DEALER);

    zsocket_set_identity(tmp_recv, "1");
    zsocket_set_identity(tmp_send, "2");

    zsocket_connect(tmp_recv, "tcp://localhost:%d", port);
    zsocket_connect(tmp_send, "tcp://localhost:%d", port+1);
    zmq_pollitem_t ignored;
    init_work_module(ctx, &ignored, port-2);
    port += 2;
    sleep(1);
}

static void teardown(void) {
    zsocket_destroy(ctx, tmp_send);
    zsocket_destroy(ctx, tmp_recv);
    teardown_hashtrie();
    teardown_work_module(ctx);
}

void recv_assertions(CuTest *tc, int n) {
    int rc = zsocket_poll(tmp_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *msg = zmsg_recv(tmp_recv);
    CuAssertIntEquals(tc, 2, zmsg_size(msg));

    char *str = zmsg_popstr(msg);
    CuAssertStrEquals(tc, "0", str);
    free(str);

    str = zmsg_popstr(msg);

    int n_len = snprintf(NULL, 0, "%d", n);
    char *term = build_term(n, n_len+11);
    CuAssertMemEquals(tc, "DScheck\0\1I", term, 10);
    CuAssertMemEquals(tc, term, str+20, 11+n_len);
    char digest[20];
    sha1(term, n_len+11, digest);
    CuAssertMemEquals(tc, digest, str, 20);
    free(term);
    free(str);
    zmsg_destroy(&msg);
}

void Test_assertions(CuTest *tc) {
    init();
    start_work_assertion(2, 1);
    recv_assertions(tc, 0);
    CuAssertIntEquals(tc, 1, q_size(assertions));
    next_assertion(1);
    recv_assertions(tc, 1);
    teardown();
}

void Test_start_checking(CuTest *tc) {
    init();
    char *fwp = "DAroot";
    start_work(fwp, strlen(fwp));
    int rc = zsocket_poll(tmp_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *msg = zmsg_recv(tmp_recv);

    char *str = zmsg_popstr(msg);
    CuAssertStrEquals(tc, "0", str);
    free(str);

    char *wp = zmsg_popstr(msg);
    char digest[20];
    sha1(fwp, strlen(fwp), digest);
    CuAssertMemEquals(tc, digest, wp, 20);
    CuAssertStrEquals(tc, fwp, wp+20);
    free(wp);

    zmsg_destroy(&msg);
    teardown();
}

void Test_build_term(CuTest *tc) {
    char *term = build_term(0, 12); // 12 = DScheck\0\1I0 + null
    CuAssertMemEquals(tc, "DScheck\0\1I0\0", term, 12);
    free(term);
    term = build_term(1, 12);
    CuAssertMemEquals(tc, "DScheck\0\1I1\0", term, 12);
    free(term);
    term = build_term(55, 13);
    CuAssertMemEquals(tc, "DScheck\0\1I55\0", term, 13);
    free(term);
}


void Test_work_balance(CuTest *tc) {
    init();
    
    threaded_work_balancing(2, 1);
    int rc = zsocket_poll(tmp_send, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *req = zmsg_recv(tmp_send);
    char *str = zmsg_popstr(req);
    CuAssertStrEquals(tc, "give work", str);
    free(str);
    zmsg_destroy(&req);

    zmsg_t *rep = zmsg_new();
    zmsg_addstr(rep, "fakedigest0123456789foo");
    zmsg_addstr(rep, "fakedigest1111111111bar");
    zmsg_addstr(rep, "fakedigest2222222222baz");
    zmsg_send(&rep, tmp_send);
    zmsg_destroy(&rep);

    rc = zsocket_poll(tmp_recv, 5000);
    CuAssertTrue(tc, rc);
    zmsg_t *wps = zmsg_recv(tmp_recv);
    CuAssertIntEquals(tc, 4, zmsg_size(wps));
    
    str = zmsg_popstr(wps);
    CuAssertStrEquals(tc, "2", str);
    free(str);
    str = zmsg_popstr(wps);
    CuAssertStrEquals(tc, "fakedigest0123456789foo", str);
    free(str);
    str = zmsg_popstr(wps);
    CuAssertStrEquals(tc, "fakedigest1111111111bar", str);
    free(str);
    str = zmsg_popstr(wps);
    CuAssertStrEquals(tc, "fakedigest2222222222baz", str);
    free(str);

    zmsg_destroy(&wps);


    teardown();
}
