% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_abstract_mappings,[
                               call_unary_test/4,
                               call_binary_test/6,

                               call_alpha/4,

                               call_unary_expression/3,
                               call_binary_expression/4,
                               call_nary_expression/3,

                               glb/3,
                               lub/3
]).

:- use_module(probsrc(error_manager)).
:- use_module(probsrc(tools)).

:- use_module(library(file_systems)).
:- use_module(probsrc(module_information)).
:- module_info(group,plugin_absint).
:- module_info(description,'Experimental abstract interpretation plugin: this module maps regular operators to the ones in the selected abstract domain.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- dynamic domain/2.

% ---------
% loading of the abstract domain modules at compiletime, lookup in directory
% ---------

load_compiletime_domains :-
    findall(Module,
           (directory_member_of_directory('domains',Dir,Fullname),
            atom_concat(Dir,'.pl',PrologFilename),
            file_member_of_directory(Fullname,PrologFilename,_),
            ajoin([Dir,'/',PrologFilename], Modulename),
            Module = abstract_domains(Modulename)),
            Domains),
    load_domains2(Domains).

load_domains2([]).
load_domains2([Module|Prest]) :-
    load_domain(Module,Id),
    (nonvar(Id) -> assert(domain(Id,Module)); true),
    load_domains2(Prest).

load_domain(Module,Id) :-
    catch( ( load_domain2(Module,Id) ->
               ajoin(['Loaded domain ', Id], Msg),
               print_message(informational, Msg)
           ; otherwise ->
               ajoin(['Loading of module ', Id, ' failed: '], Msg),
               print_message(error, Msg)),
           E,
           ( ajoin(['Loading of module ', Id, ' raised an exception: ', E], Msg),
               print_message(error, Msg))).

load_domain2(Module,Id) :-
    use_module(IdVar,Module,[]),
    IdVar = Id,
    ( domain(Id,_) ->
        ajoin(['Domain ID ',Id,'used twice'],Msg),
        print_message(error,Msg),
        fail
    ; otherwise ->
        true).


:- load_compiletime_domains.

% ----------------
% Call GLB and LUB
% ----------------
lub(X, Y, Z) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    current_predicate(DomainModule:lub/1),
    DomainModule:lub(LUB),
    call(DomainModule:LUB, X, Y, Z),
    !.
lub(_, _, _) :-
    add_error(lub, 'Least upper bound is not defined or failed. Define a lub/3 predicate in the abstract domain file'), fail.

glb(X, Y, Z) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    current_predicate(DomainModule:glb/1),
    DomainModule:glb(GLB),
    call(DomainModule:GLB, X, Y, Z),
    !.
glb(_, _, _) :-
    add_error(glb, 'Greatest lower bound is not defined or failed. Define a glb/3 predicate in the abstract domain file'), fail.

% -----------------
% Call Alpha
% -----------------
call_alpha(Concrete,Type,Infos,Abstract) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    DomainModule:map_alpha(alpha, CallAlpha),
    call(DomainModule:CallAlpha,Concrete,Type,Infos,Abstract), !.
call_alpha(C,T,I,top) :-
    add_error(call_alpha, 'no mapping for alpha or no result defined (using top instead) for ', [C,T,I]).

% ----------------
% Call predicates
% ----------------
call_unary_test(UOP, Arg1, NewArg1, Test) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    (DomainModule:map_unary_test_operator(UOP, CallUOP) -> true
        ; Expr =.. [UOP, Arg1],add_error(call_unary_test, 'no mapping for operator or no result defined (using true instead) for ', Expr)),
    !,
    (call(DomainModule:CallUOP, Arg1, NewArg1, Test1) -> Test = Test1
        ; Test = false, Arg1 = NewArg1).

call_binary_test(BOP, Arg1, Arg2, NewArg1, NewArg2, Test) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    (DomainModule:map_binary_test_operator(BOP, CallBOP) -> true
        ; Expr =.. [BOP, Arg1, Arg2],add_error(call_binary_test, 'no mapping for operator or no result defined (using true instead) for ', Expr)),
    !,
    (call(DomainModule:CallBOP, Arg1, Arg2, NewArg1, NewArg2, Test1) -> Test = Test1
        ; Test = false, Arg1 = NewArg1, Arg2 = NewArg2).

% -----------------
% Call expressions
% -----------------
call_unary_expression(UOP, Arg1, Result) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    DomainModule:map_unary_operator(UOP, CallUOP),
    !,
    call(DomainModule:CallUOP, Arg1, Result).
call_unary_expression(UOP, Arg1, top) :-
    Expr =.. [UOP, Arg1],
    add_error_fail(call_unary_expression, 'no mapping for operator or no result defined for ', Expr).

call_binary_expression(BOP, Arg1, Arg2, Result) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    DomainModule:map_binary_operator(BOP, CallBOP),
    !,
    call(DomainModule:CallBOP, Arg1, Arg2, Result).
call_binary_expression(BOP, Arg1, Arg2, top_) :-
    Expr =.. [BOP, Arg1, Arg2],
    add_error_fail(call_binary_expression, 'no mapping for operator or no result defined for ', Expr).

call_nary_expression(OP, Arg, Result) :-
    preferences:preference(plugin(absint,abstract_domain_module), DomainModule),
    DomainModule:map_nary_operator(OP, CallOP),
    !,
    call(DomainModule:CallOP, Arg, Result).
call_nary_expression(OP, Arg, top) :-
    Expr =.. [OP, Arg],
    add_error_fail(call_unary_expression, 'no mapping for operator or no result defined for ', Expr).
