% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(b_global_sets,
          [b_get_global_constants/1, b_get_global_enumerated_sets/1, b_get_global_sets/1,
           lookup_global_constant/2,
           is_b_global_constant/3,
           is_used_b_global_constant/3, is_unused_b_global_constant/3,

           b_global_set/1, b_non_empty_global_set/1, b_empty_global_set/1,
           b_global_deferred_set/1,
           b_global_set_with_potential_symmetry/1, b_global_deferred_set_with_card_gt1/1,
           b_partially_deferred_set/1,

           unfixed_deferred_set/1,
           contains_unfixed_deferred_set/1,
           b_supplementary_global_set/1, % introduced when untyped ids are allowed

           inferred_minimum_global_set_cardinality/2,
           inferred_maximum_global_set_cardinality/2,
           b_exists_global_set_with_potential_symmetry/0,
           b_global_set_cardinality/2,
           
           b_clear_global_set_type_definitions/0,
           b_check_and_precompile_global_set_type_definitions/0,
           %b_precompile_used_global_set_information/0,
           
           b_get_prob_deferred_set_elements/2,add_prob_deferred_set_elements_to_store/3,
           prob_deferred_set_element/4,

           find_inequal_global_set_identifiers/4,
           static_symmetry_reduction_for_global_sets/1,

           b_integer_set/1, b_integer_or_string_set/1,
           b_type2_set/2, try_b_type2global_set/2,

           b_fd_type/3, b_get_fd_type_bounds/3,
           b_fd_card/2,
           is_global_set_of_cardinality_one/2, is_global_set_of_cardinality_two/3,
           global_type/2, global_type_wf/3,
           enum_global_type/2, enumerate_global_type_with_enum_warning/3,

           all_elements_of_type/2, all_elements_of_type_rand/2,

           set_user_defined_scope/2,

           fixed_deferred_set_size/2,

           add_new_global_set/1,
           generate_fresh_supplementary_global_set/1
           %,add_global_set_with_named_constants/2  /* for use by Z,... */
           ]).

:- use_module(debug).
:- use_module(tools).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(error_manager).
:- use_module(bsyntaxtree).
:- use_module(library(between),[between/3]).
:- use_module(gensym,[gensym/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides support for deferred/enumerated set elements in B.').

:- use_module(bmachine).

/* what we call global sets here usually called "given sets", which can be
   either enumerated or deferred */


:- dynamic b_global_constant/3. % named element of a deferred or enumerated set
  % for enumerated sets S = {a,b,c}  we would have entries a,b,c as b_global_constant
  % for deferred set S, we may add CONSTANTS which are element of S as b_global_constant
:- dynamic unused_b_global_constant/3.  % TO DO: compute properly again !!

:- dynamic inferred_minimum_global_set_cardinality/2.
:- dynamic inferred_maximum_global_set_cardinality/2.

b_get_global_enumerated_sets(GSets) :-
  findall(GS, (b_global_set(GS),  is_b_global_constant(GS,_Nr,_Cst)), GSets).
b_get_global_sets(GSets) :-
  findall(GS, b_global_set(GS), GSets).

b_get_global_constants(Csts) :- findall(Cst,is_b_global_constant(_,_,Cst),Csts).

is_b_global_constant(Set,Nr,Cst) :- b_global_constant(Set,Nr,Cst).
is_unused_b_global_constant(Set,Nr,Cst) :- unused_b_global_constant(Set,Nr,Cst).
is_used_b_global_constant(Set,Nr,Cst) :-
   b_global_constant(Set,Nr,Cst), \+ unused_b_global_constant(Set,Nr,_).

:- dynamic lookup_global_constant/2.
%lookup_global_constant(Id,Val) :-
%  b_global_constant(Set,Nr,Id), % not an indexed lookup ; now we assert lookup_global_constant below
%  Val = fd(Nr,Set).


:- dynamic b_precompiled_global_set/1.
b_precompiled_global_set(_) :- print_error('*** b_global_set not precompiled'),fail.

b_global_set(GS) :- b_precompiled_global_set(GS) ; b_supplementary_global_set(GS).

% a version of b_global_set which does not leave a trailing choice point for b_supplementary_global_set around
% has to be called with GS ground
b_global_set_ground(GS) :- if(b_precompiled_global_set(GS),true,b_supplementary_global_set(GS)).


b_global_deferred_set(GS) :-
  b_global_set(GS),
  \+ is_b_global_constant(GS,_Nr,_Cst). % note : some deferred sets are translated into partially_deferred_set

% either an enumerated set with unused constants or a deferred sets where some constants were lifted into the deferred set
b_partially_deferred_set(GS) :-
   b_global_set(GS),
   (is_b_global_constant(GS,_Nr,_Cst) -> true), % GS is in principle enumerated
   (unused_b_global_constant(GS,_X,_) -> true).


b_global_set_with_potential_symmetry(GS) :-
   b_global_set(GS),
   (b_global_deferred_set_with_card_gt1(GS) -> true
    ; ((unused_b_global_constant(GS,X,_),
        unused_b_global_constant(GS,Y,_), X\=Y) -> true % At least two unused constants exist
        ; fail)).

b_global_deferred_set_with_card_gt1(GS) :-
   b_global_deferred_set(GS),
   extract_setsize_from_machine_cache(GS,Low,Up),
   Up>Low.

:- volatile b_exists_global_set_with_potential_symmetry/0.
:- dynamic b_exists_global_set_with_potential_symmetry/0.

b_check_and_precompile_global_set_type_definitions :-
                      /* must be called before the animator can run !! */
  b_reset_global_set_type_definitions,
  retractall(b_exists_global_set_with_potential_symmetry),
  debug_println(9,'Checking existence of global type definitions: '),
  debug_print(9,'%  '),
  b_get_machine_set(Set),
  add_new_global_set(Set),
  (b_fd_type(Set,LowBnd,UpBnd)  /* also computes & caches the size of the SET */
    -> debug_print(9,' '),debug_print(9,Set),
       debug_print(9,'=='),debug_print(9,LowBnd),debug_print(9,'..'),debug_print(9,UpBnd)
    ;  add_internal_error('No b_fd_type/2 definition for global set: ',b_fd_type(Set,_,_))
   ),debug_print(9,'  '),
   fail.
b_check_and_precompile_global_set_type_definitions :-
   find_all_used_enumerated_elements(UsedEnumeratedElements),
   b_fd_type(GS,Low,Up),
   (is_b_global_constant(GS,_Nr,_Cst) -> true), % GS is in principle enumerated
   (user_forced_symmetry(GS)
     -> ForceSymm=true, print('FORCING SYMMETRY: '), print(GS), print(' ')
     ;  ForceSymm=false
   ),
   enum_fd(X,Low,Up), /* enumerate all elements of GS */
   can_symmetry_reduction_be_applied_to_constant(ForceSymm,GS,X,UsedEnumeratedElements,Display),
   assert(unused_b_global_constant(GS,X,deferred)), % by storing that it is not used, we enable symmetry reduction
   debug_println(9,unused_b_global_constant(Display)),
   fail.
b_check_and_precompile_global_set_type_definitions :-
  (b_global_set_with_potential_symmetry(_)
   -> assert(b_exists_global_set_with_potential_symmetry),
      debug_println(9,'% Symmetry is potentially useful for this machine')
   ; true),
  debug_nl(9),
  /* clean up: */
  reset_global_set_user_defined_scope.

% --------------------

% peform MACE style static symmetry reduction for those global constants
% that have not already been fixed
% e.g., for constants aa,bb,cc,dd of type ID and bb/=cc  --> nrs of bb,cc will be fixed as 1 and 2; we will ensure that numbers of aa:1..3 and dd:1..4 and that dd=4 only if aa=3
% TO DO: extend this kind of symmetry reduction to other constants (e.g., total fun over fixed domain f: ... --> ID)
% TO DO: ensure this is also used in enabling analysis, symbolic_model checker, ...
% TO DO: support detection of injections
static_symmetry_reduction_for_global_sets(_ConstantsState) :-
  get_preference(try_kodkod_on_load,true),!. % no idea which numbering Kodkod might return/expect ?!
static_symmetry_reduction_for_global_sets(_ConstantsState) :-
  get_preference(use_static_symmetry_detection,false),!.
static_symmetry_reduction_for_global_sets(ConstantsState) :-
  findall((GS,FirstAvailableNewIdx,Low,Up,Other),
         static_symmetry_reduction_possible(GS,FirstAvailableNewIdx,Low,Up,Other),L),
  maplist(perform_static_symmetry_reduction(ConstantsState),L).

perform_static_symmetry_reduction(ConstantsState,(GS,First,Low,Up,Other)) :-
   if(perform_ssr(Other,[],First,First,GS,Low,Up,ConstantsState),true,
      add_internal_error('Call failed: ',perform_ssr(Other,[],First,First,GS,Low,Up,ConstantsState))).

:- use_module(store,[lookup_value_for_existing_id/3]).
perform_ssr([],PrevNrs,InitialF,FirstNew,_,_,_,_) :- debug_println(10,done_ssr(PrevNrs,InitialF,FirstNew)).
perform_ssr([OtherID|T],PrevNrs,InitialF,FirstNew,GS,Low,Up,ConstantsState) :-
   % adding static symmetry breaking constraints for OtherID of type GS
   lookup_value_for_existing_id(OtherID,ConstantsState,fd(Nr,GS)),
   debug_println(10,ssr(OtherID,Nr,FirstNew,Up)),
   fd_utils_clpfd:in_fd(Nr,Low,FirstNew),
   clpfd_interface:force_post_constraint( (Nr#>InitialF) #<=> Trigger),
   check_contiguous(Trigger,Nr,PrevNrs),
   (FirstNew<Up -> F1 is FirstNew+1 ; F1=Up),
   perform_ssr(T,[Nr|PrevNrs],InitialF,F1,GS,Low,Up,ConstantsState).

:- block check_contiguous(-,?,?),check_contiguous(?,-,?). % to do: do not delay on Nr
check_contiguous(0,_Nr,_PrevNrs). % we have not used a new deferred set number
check_contiguous(1,Nr,PrevNrs) :- % we have used a new number, check that previous number also used
   %print(check_used(Nr,PrevNrs)),nl,
   N1 is Nr-1,
   %clpfd_lists:in_fd_value_list_wf(N1,PrevNrs,_WF). % TO DO: get rid of WF
   (tools:exact_member(N1,PrevNrs) -> true
     ; when(ground(PrevNrs),clpfd_interface:force_clpfd_inlist(N1,PrevNrs))).

% check if we can apply static symmetry reduction for a global deferred set GS
static_symmetry_reduction_possible(GS,FirstAvailableNewIdx,Low,Up,Other) :-
  b_fd_type(GS,Low,Up),
  b_get_disjoint_constants_of_type(GS, DisjointConstants, AllConstantsIDs),
  findall(Nr,is_b_global_constant(GS,Nr,_Cst),Ns),
  max_member(Max,Ns),
  FirstAvailableNewIdx is Max+1,
  sort(AllConstantsIDs,SAll),
  sort(DisjointConstants,SDis),
  ord_subtract(SAll,SDis,Other),
  Other \= [],
  debug_println(10,static_symmetry_possible(GS,FirstAvailableNewIdx,Up,Other,SDis)).

% --------------------

can_symmetry_reduction_be_applied_to_constant(true,GS,X,_Enums,GS/X). % if symmetry enforced, all elements are
                                                               % subject to symmetry reduction
can_symmetry_reduction_be_applied_to_constant(false,GS,X,UsedEnums,Display) :-
    ( is_b_global_constant(GS,X,Name) ->  % sym-reduction can be applied to an enumerated set member
        nonmember(Name,UsedEnums),        % if it's not used anywhere
        Display=Name
    ; otherwise -> Display=GS/X).         % sym-reduction can be applied to deffered set elements


find_all_used_enumerated_elements(Elements) :-
    b_get_all_used_identifiers(Identifiers),
    b_get_machine_variables(TVariables),get_texpr_ids(TVariables,Variables),
    % We assume that b_get_all_used_identifiers/1 is a proper ordset. (sorted with sort/2)
    % As an optimisation, we remove the variables, they cannot be enumerated elements.
    list_to_ord_set(Variables,OVariables),
    ord_subtract(Identifiers,OVariables,Elements),!.
find_all_used_enumerated_elements(_Elements) :-
    add_error_and_fail(b_global_sets, 'find_all_used_enumerated_elements failed').


reset_global_set_user_defined_scope :-
   retract(global_set_user_defined_scope(GS,_Scope)),
   (b_fd_type(GS,_Low,_Up)
     -> true
     ;  add_error(b_global_sets,'Trying to set scope of unknown SET: ',GS)
   ),fail.
reset_global_set_user_defined_scope.

% Return those constants which are implicitly introduced by ProB, but not really
% part of the model. These are the elements of a deferred set.
b_get_prob_deferred_set_elements(TIds,AllOrVisible) :-
    findall( TId,
             ( prob_deferred_set_element(GS,_Nr,Id,AllOrVisible),
                 create_texpr(identifier(Id),global(GS),[],TId)),
             TIds).

add_prob_deferred_set_elements_to_store(OldStore,NewStore,AllOrVisible) :-
    %print(add_prob_deferred_set_elements_to_store(OldStore)),nl,
    findall(bind(Id,fd(Nr,GS)),(prob_deferred_set_element(GS,Nr,Id,AllOrVisible),
                                \+ member(bind(Id,_),OldStore)),
                                NewStore, OldStore).

prob_deferred_set_element(GlobalSet,Elem,Id,AllOrVisible) :-
    b_global_set(GlobalSet),
    atom_codes(GlobalSet,GlobalSetCodes),append(GlobalSetCodes,ECodes,NameCodes),
    b_fd_type(GlobalSet,Low,Up),
    between(Low,Up,Elem),
    (AllOrVisible = all -> true
      ; % only visible ones
        \+ is_b_global_constant(GlobalSet,Elem,_) % the identifier is not used somewhere else
    ),
    number_codes(Elem,ECodes),
    atom_codes(Id,NameCodes),
    \+ b_global_set(Id), % not used as another SET name
    \+ is_b_global_constant(_,_,Id). % not used as another SET element


user_forced_symmetry(GS) :- % check if the user defined FORCE_SYMMETRY_GS == TRUE
     string_concatenate('FORCE_SYMMETRY_',GS,DefName),
     b_get_typed_definition(DefName,[],F), get_texpr_expr(F,boolean_true).

% TO DO: re-add this feature, so that unused enumerated set elements can be detected
% :- use_module(bmachine,[used_identifier/1]).
%b_precompile_used_global_set_information :-
%  b_precompute_used_identifiers,
%  b_global_constant(GlobalSetName,Nr,CstName),
%  (used_identifier(CstName) -> true
%   ; (assert(unused_b_global_constant(GlobalSetName,Nr,CstName)),
%      debug_print(9,unused_b_global_constant(GlobalSetName,Nr,CstName)), debug_nl(9))
%  ),
%  fail.
%b_precompile_used_global_set_information :-
%   retractall(b_exists_deferred_set_or_unused_csts_with_card_gt_1),
%   ((b_global_deferred_set_with_card_gt1(_GS) ;
%     (is_unused_b_global_constant(GS,Nr,_),
%      is_unused_b_global_constant(GS,Nr2,_), Nr2 \= Nr))
%       -> (assert(b_exists_deferred_set_or_unused_csts_with_card_gt_1),
%           print_message('symmetry can be used for this machine'))
%     ; true
%   ).

is_global_set_of_cardinality_two(Type,LowBnd,UpBnd) :-
  b_get_fd_type_bounds(Type,LowBnd,UpBnd),
  LowBnd+1 =:= UpBnd.
  
:- volatile is_global_set_of_cardinality_one/2.
:- dynamic is_global_set_of_cardinality_one/2.
% should be called when unloading a machine, before type-checking etc...
% TO DO: use :- use_module(eventhandling,[register_event_listener/3]).
b_clear_global_set_type_definitions :- % nl,print(clearing_gs),nl,
  retractall(b_supplementary_global_set(_)),
  b_reset_global_set_type_definitions.

b_reset_global_set_type_definitions :- %nl,print(reset_gs),nl,
  % print_message('resetting global sets'),
  retractall(b_precompiled_global_set(_)),
  retractall(enumerated_set(_)),
  retractall(fixed_deferred_set_size(_,_)),
  retractall(extract_setsize_from_machine_cache(_,_,_)),
  retractall(find_minimum_cardinality_cache(_,_)),
  retractall(is_global_set_of_cardinality_one(_,_)),
  retractall(b_global_constant(_,_,_)),
  retractall(lookup_global_constant(_,_)),
  retractall(unused_b_global_constant(_,_,_)),
  retractall(inferred_minimum_global_set_cardinality(_,_)),
  retractall(inferred_maximum_global_set_cardinality(_,_)).

add_new_global_set(Set) :-
  (b_precompiled_global_set(Set)
   -> add_internal_error('Global set already exists: ',add_new_global_set(Set))
   ;  assert(b_precompiled_global_set(Set))).

:- dynamic b_supplementary_global_set/1.

% these should be called before we precompile the global set definitions
generate_fresh_supplementary_global_set(FRESHID) :-
   gensym:gensym('__DEFERREDSET__',FRESHID),
   debug_println(10,generate_fresh_supplementary_global_set(FRESHID)),
   assert(b_supplementary_global_set(FRESHID)).

/* --------------------------------------------------------- */
/* Extracting Finite Domain type information from the B machine */
/* --------------------------------------------------------- */



/* below treats annotations in the form:
DEFINITIONS
    scope_Name == 1..3;
    scope_Code == 4..8
 which inform us about which finidte domain ranges we should
  give to global sets defined in SETS
*/

:- use_module(tools_printing,[print_error/1]).
:- dynamic extract_setsize_from_machine_cache/3.
extract_setsize_from_machine_cache(_,_,_) :-
   print_error('*** extract_setsize_from_machine not precompiled'),fail.

try_extract_setsize_from_machine(GlobalSetName,_LowBound,_UpBound) :-
   start_extracting_setsize(GlobalSetName),!,fail. % cyclic dependency; abort computation
try_extract_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
   extract_setsize_from_machine(GlobalSetName,LowBound,UpBound).

extract_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
  extract_setsize_from_machine_cache(GlobalSetName,L,U),!,
  LowBound=L,UpBound=U.
extract_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
  start_extracting_setsize_from_machine(GlobalSetName),
  b_get_named_machine_set(GlobalSetName,ListOfConstants),!,
  assert_enumerated_set(GlobalSetName,ListOfConstants,LowBound,UpBound), % will also assert_extract_setsize_from_machine_cache
  ((get_user_defined_scope(GlobalSetName,DL,DU), DU-DL =\= UpBound-LowBound)
    -> add_error(extract_setsize_from_machine,'Conflict between scope_ Definition and explicit enumeration of SET:',GlobalSetName)
    ; true
  ).
extract_setsize_from_machine(GlobalSetName,L,U) :-
   b_supplementary_global_set(GlobalSetName),!, % deferred set that was added for untyped ID,...
   default_deferred_set_bounds(LowBound,UpBound),
   assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound),
   (LowBound,UpBound)=(L,U).
extract_setsize_from_machine(GlobalSetName,L,U) :-
   /* we have a DEFERRED SET */
   extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound),
   (LowBound=UpBound -> assert(is_global_set_of_cardinality_one(GlobalSetName,LowBound)) ; true),
   assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound),
   (b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants,_)
    -> /* add those constants to the deferred set as though they were enumerated set elements: improves performance by avoiding redundant permutations; is a kind of symmetry breaking */
     print(adding_disjoint_constants(GlobalSetName,DisjointConstants)),nl,
     ~~mnf(add_named_constants_to_global_set(GlobalSetName,DisjointConstants))
    ; true
   ),
   debug_println(9,setsize(GlobalSetName,LowBound,UpBound)),
   (LowBound,UpBound)=(L,U).

/* for Deferred SETS : */
extract_DEFERRED_setsize_from_machine(GlobalSetName,L,U) :-
   b_get_properties_from_machine(Properties), member_in_conjunction(C,Properties),
  % check if the PROPERTIES contain an expression of the form card(GS) = Nr
   is_equality_card_global_set(C,GlobalSetName,Properties,Card),!,
   assert(fixed_deferred_set_size(GlobalSetName,Card)),
   LowBound=1, UpBound = Card,
   ((get_user_defined_scope(GlobalSetName,DL,DU),
      DU-DL =\= UpBound-LowBound)
    -> add_error(extract_DEFERRED_setsize_from_machine,'Conflict between scope_ Definition and PROPERTIES:',GlobalSetName)
    ; true
   ),
   LowBound=L,UpBound=U.
extract_DEFERRED_setsize_from_machine(GlobalSetName,L,U) :-  /* check if there is a scope_ DEFINITION */
  get_user_defined_scope(GlobalSetName,LowBound,UpBound),!,
  LowBound=L,UpBound=U.
extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
   get_preference(globalsets_fdrange,DefaultUpperBound),
   find_minimum_cardinality(GlobalSetName,MinCard),
   assert(inferred_minimum_global_set_cardinality(GlobalSetName,MinCard)),
   !,
   LowBound = 1,
   (find_maximum_cardinality(GlobalSetName,MaxCard)
    -> (MaxCard=MinCard -> assert(fixed_deferred_set_size(GlobalSetName,MinCard)),
        debug_println(9,fixed_deferred_set_size(GlobalSetName,MinCard)),
        UpBound = MinCard
      ; DefaultUpperBound>MaxCard ->
        debug_println(9,reducing_deferred_set_size(GlobalSetName,MaxCard)),
        UpBound=MaxCard
      ; DefaultUpperBound<MinCard ->
        debug_println(9,inferred_minimum_global_set_cardinality(GlobalSetName,MinCard)),
        UpBound=MinCard
      ; otherwise -> UpBound is MinCard
        )
    ; MinCard>DefaultUpperBound -> UpBound=MinCard
    ; UpBound=DefaultUpperBound),
    debug_println(4,inferred(GlobalSetName,UpBound)).
extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
   % No Minimum cardinality was inferred
   LowBound=1, get_preference(globalsets_fdrange,DefaultUpperBound),
   find_maximum_cardinality(GlobalSetName,MaxCard),
   !,
   (MaxCard=1 -> UpBound=1, assert(fixed_deferred_set_size(GlobalSetName,1))
    ; DefaultUpperBound>MaxCard ->
       debug_println(9,reducing_deferred_set_size(GlobalSetName,MaxCard)),
       UpBound=MaxCard
    ; UpBound=DefaultUpperBound).
extract_DEFERRED_setsize_from_machine(_GlobalSetName,LowBound,UpBound) :-
   % No Minimum cardinality was inferred
   default_deferred_set_bounds(LowBound,UpBound).

default_deferred_set_bounds(1,UpBound) :- get_preference(globalsets_fdrange,UpBound).

find_maximum_cardinality(GlobalSetName,MaxCard) :-
   find_maximum_cardinality1(GlobalSetName,MaxCard),
   assert(inferred_maximum_global_set_cardinality(GlobalSetName,MaxCard)).

find_maximum_cardinality1(GlobalSetName,MaxCard) :-
   b_get_machine_constants(Csts),
   b_get_properties_from_machine(Properties),
   findall(MC2,find_maximum_cardinality2(Csts,Properties,GlobalSetName,MC2),List),
   debug_println(9,maxcard_list(GlobalSetName,List)),
   min_member(MaxCard,List).
find_maximum_cardinality2(_Csts,Properties,GlobalSetName,MaxCard) :-
    member_in_conjunction(C,Properties),
    get_texpr_expr(C,equal(LHS,RHS)),
    (get_texpr_expr(LHS,identifier(GlobalSetName)), R=RHS
     ;
     get_texpr_expr(RHS,identifier(GlobalSetName)), R=LHS
    ),
    get_texpr_expr(R,set_extension(SetExtElements)),
    /* We have an axiom of the form  GlobalSetName = { El1, ..., Eln} */
    length(SetExtElements,MaxCard).

:- dynamic find_minimum_cardinality_cache/2.
find_minimum_cardinality(GS,MC) :-
   find_minimum_cardinality_cache(GS,Nr),!, %print(min_card_cache(GS,Nr)),nl,
   % avoid loops; e.g., if we have N1 >->> N2  and N2 >->> N1
   number(Nr),
   MC=Nr.
find_minimum_cardinality(GlobalSetName,Res) :- %print(find_min_card(GlobalSetName)),nl,
 /* try to find out from the Properties whether there is an implied minimum cardinality */
   b_get_machine_constants(Csts),
   b_get_properties_from_machine(Properties),
   assert(find_minimum_cardinality_cache(GlobalSetName,unknown)),
   findall(MinCard2,find_minimum_cardinality2(Csts,Properties,GlobalSetName,MinCard2),List),
   debug_println(9,mincard_list(GlobalSetName,List)),
   max_member(MinCard,List),
   retract(find_minimum_cardinality_cache(GlobalSetName,_)),
   assert(find_minimum_cardinality_cache(GlobalSetName,MinCard)),
   Res=MinCard.

find_minimum_cardinality2(_,Properties,GlobalSetName,MinCard) :-
   member_in_conjunction(C,Properties),
   find_minimum_cardinality3(C,Properties,GlobalSetName,MinCard).
find_minimum_cardinality2(_Constants,_Properties,GlobalSetName,MinCard) :-
   %print(trying_find_min_card(GlobalSetName,Constants)),nl,
   b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants,_),
   length(DisjointConstants,MinCard), MinCard>1.

find_minimum_cardinality3(C,Properties,GlobalSetName,MinCard) :-
  % check if the PROPERTIES contain an expression of the form card(GS) >= Nr or similar
   is_greaterequalthan_card_global_set(C,GlobalSetName,Properties,MinCard).
find_minimum_cardinality3(C,Properties,GlobalSetName,MinCard) :-
  % find things like partition(GlobalSetName,S1,...,SN)
   get_texpr_expr(C,partition(Set,DisjSets)),
   global_set_identifier(Set,GlobalSetName),
   %  print(partition(Set,DisjSets,GlobalSetName)),nl,
   find_min_card_of_disjoint_sets(DisjSets,Properties,MinCard). % ,print(min(MinCard)),nl.
find_minimum_cardinality3(C,_Properties,GlobalSetName,MinCard) :-
   get_texpr_expr(C,member(_Something,SURJ)),
   is_surjection(SURJ,Set,OTHERSET),
   global_set_identifier(Set,GlobalSetName),
   \+ global_set_identifier(OTHERSET,GlobalSetName), % Surjection on itself; no use in narrowing down cardinality
   (minimum_cardinality(OTHERSET,MinCard) -> true ; print('*** could not compute mincard: '), print(OTHERSET),nl,fail),
   (MinCard=inf -> add_error(b_global_sets,'Set has to be of infinite cardinality: ',GlobalSetName),fail ; true).

is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,total_surjection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,partial_surjection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,total_bijection(FromSet,ToSet)).  % here we know the exact cardinality - TO DO: use this fact
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,partial_bijection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,total_injection(ToSet,FromSet)). % inverse is surjective



% compute the minmum cardinality of an expression
minimum_cardinality(ID,MinCard) :- global_set_identifier(ID,GlobalSet),!,
  (b_fd_type(GlobalSet,LowBnd,UpBnd) -> MinCard is 1+UpBnd-LowBnd
   ; print('*** Do not know card of : '), print(GlobalSet),nl,
   % TO DO: do full blown cardinality analysis of deferred sets
    fail).
minimum_cardinality(b(EXPR,TYPE,_INFO),MinCard) :-
  minimum_cardinality2(EXPR,TYPE,MinCard).

%minimum_cardinality2(A,T,R) :- print(mincard(A,T,R)),nl,fail.
minimum_cardinality2(bool_set,_,2).
minimum_cardinality2(cartesian_product(A,B),_,Res) :- minimum_cardinality(A,CA),
   minimum_cardinality(B,CB), kernel_objects:safe_mul(CA,CB,Res).
minimum_cardinality2(interval(A,B),_,Res) :-
   b_get_properties_from_machine(Properties),
   evaluable_integer_expression(A,Properties,EA),
   evaluable_integer_expression(B,Properties,EB),
   (EB >= EA -> Res is EB+1-EA ; Res=0).
% TO DO: enumerated sets, ... seq, relations, ... ?


 find_min_card_of_disjoint_sets([],_Properties,0).
 find_min_card_of_disjoint_sets([H|T],Properties,R) :-
    find_min_card_of_disjoint_sets(T,Properties,TN),
    (minimum_cardinality_of_set(H,Properties,MC)
      -> true   %,print(min_card(MC,H)),nl
       ; add_internal_error('Call failed: ',minimum_cardinality_of_set(H,Properties,MC)), MC=1),
    R is TN+MC.


minimum_cardinality_of_set(b(Expr,set(Type),_),Properties,MC) :-
  minimum_cardinality_of_set_aux(Expr,Type,Properties,MC).

%minimum_cardinality_of_set_aux(P,Type,_,_MC) :- print(min_card(P,Type)),nl,fail.
minimum_cardinality_of_set_aux(set_extension(List),GlobalSetName,Properties,MinCard) :- !,
   def_get_texpr_ids(List,AtomicIDList),
   find_inequal_global_set_identifiers(AtomicIDList,GlobalSetName,Properties,DisjointConstants),
   length(DisjointConstants,MinCard). % ,   print(disj_set_ext(MinCard,DisjointConstants)),nl,nl.
minimum_cardinality_of_set_aux(identifier(CST),_Type,_Properties,MinCard) :-
   find_minimum_cardinality_cache(CST,Nr),!,
   number(Nr),
   MinCard=Nr.
minimum_cardinality_of_set_aux(identifier(CST),Type,Properties,MinCard) :-
   CST_Template = b(identifier(CST),set(Type),_),
   min_card_of_identifier(CST_Template,Type,Properties,MC),
   !,
   MinCard=MC,
   assert(find_minimum_cardinality_cache(CST,MC)). % global sets and constants cannot have the same name
% TO DO: add more cases ?  sequence extension cannot appear
minimum_cardinality_of_set_aux(_,_,_,0).

% determine minimum cardinality of an identifier; usually a constant (cannot be another global set due to B typing).
min_card_of_identifier(CST_Template,_Type,Properties,MC) :-
   get_texpr_expr(PT,partition(CST_Template,DisjSets)),
   member_in_conjunction(PT,Properties),
   find_min_card_of_disjoint_sets(DisjSets,Properties,MC),!.
min_card_of_identifier(CST_Template,Type,Properties,MinCard) :- % x:ID => card(ID) >= 1
   get_texpr_id(TID,ID),
   get_texpr_expr(EL,member(TID,CST_Template)),
   findall(ID,member_in_conjunction(EL,Properties),AtomicIDList),
   % find all identifiers such that we have id : CST
   AtomicIDList \= [],
   !,
   find_inequal_global_set_identifiers(AtomicIDList,Type,Properties,DisjointConstants),
   length(DisjointConstants,MinCard).
min_card_of_identifier(CST_Template,_Type,Properties,MC) :- % x:ID => card(ID) >= 1
   get_texpr_expr(EL,member(_,CST_Template)),
   member_in_conjunction(EL,Properties),
   %translate:print_bexpr(EL),nl,
   !,
   MC=1.
min_card_of_identifier(CST_Template,Type,Properties,MC) :- % ID /= {}  => card(ID) >= 1
   EMPTYSET = b(empty_set,set(Type),_),
   get_texpr_expr(NEQ,not_equal(LHS,RHS)),
   member_in_conjunction(NEQ,Properties),
   (LHS=CST_Template, RHS=EMPTYSET ;
    RHS=CST_Template, LHS=EMPTYSET ),
   %print(neq(NEQ,CST,RHS)),nl,
   !,
   MC=1.


% the following could be further improved
% try and find out how many different identifiers are in a list in the context of a predicate Prop
% it is also called by b_get_disjoint_constants_of_type_calc in bmachine
%find_inequal_global_set_identifiers(L,GS,_P,_R) :- print(find_inequal_global_set_identifiers(L,GS)),nl,fail.

find_inequal_global_set_identifiers([],_,_,[]).
find_inequal_global_set_identifiers([ID|T],GS,Prop,Res) :-
    (atomic(ID) -> true
     ; add_internal_error('Not atomic id: ',find_inequal_global_set_identifiers([ID|T],GS,Prop,Res)),fail),
    findall(ID2, (member_in_conjunction(C,Prop), %print(chck(C)),nl,
                  inequality(C,GS,ID,ID2),
                  member(ID2,T)), DiffIDs),
    sort(DiffIDs,SortedDiffIDs), % also remove duplicates
    ((SortedDiffIDs=[],T=[_,_|_]) % then try to proceed with T, ignoring ID
      -> find_inequal_global_set_identifiers(T,GS,Prop,Res)
      ;  find_inequal_global_set_identifiers(SortedDiffIDs,GS,Prop,RT), Res = [ID|RT]
     ).

inequality(b(P,pred,_),GS,ID,OtherID) :-
  inequality_aux(P,GS,ID,OtherID).
inequality_aux(not_equal(I1,I2),GS,ID,OtherID) :-
  gs_identifier(I1,GS,ID1), gs_identifier(I2,GS,ID2),
  (ID=ID1 -> OtherID=ID2 ; ID=ID2,OtherID=ID1).
inequality_aux(not_equal(I1,I2),GS,ID,OtherID) :-
  % two identifiers of global sets need not to be equal
  % in order for two set extensions to be inequal
  % first we check if I1, I2 are in fact subsets of global sets
  get_texpr_type(I1,set(global(_))),
  get_texpr_expr(I1,set_extension(IDs1)), % TODO: is there something more general than a set_extension?
  get_texpr_expr(I2,set_extension(IDs2)), % should full sets or set identifiers added?
  % remove the common elements and see if we find a pair of identifiers
  maplist(remove_all_infos_and_ground,IDs1,IDs1NoInfos), maplist(remove_all_infos_and_ground,IDs2,IDs2NoInfos),
  % TODO: maybe it is same to assume the order?
  list_to_ord_set(IDs1NoInfos,IDs1Set), list_to_ord_set(IDs2NoInfos,IDs2Set),
  ord_subtract(IDs1Set,IDs2Set,[E1]), % each set has one identifier that is not in the other one
  ord_subtract(IDs2Set,IDs1Set,[E2]),
  E1 \= E2, % and they are not the same
  gs_identifier(E1,GS,ID1), gs_identifier(E2,GS,ID2),
  (ID=ID1 -> OtherID=ID2 ; ID=ID2,OtherID=ID1).
inequality_aux(negation( b(equal(I1,I2),pred,_)),GS,ID,OtherID) :-
  gs_identifier(I1,GS,ID1), gs_identifier(I2,GS,ID2),
  (ID=ID1 -> OtherID=ID2 ; ID=ID2,OtherID=ID1).
inequality_aux(partition(_Set,DisjSets),GS,ID,OtherID) :- %print(part(Set,DisjSets,GS,ID,OtherID)),nl,
  % detect such things as partition(_,{...ID...},...{...OtherID...}...)
  select(Set1,DisjSets,Rest), set_extension_list(Set1,List1),
  member(O1,List1), gs_identifier(O1,GS,ID),
  select(Set2,Rest,_), set_extension_list(Set2,List2),
  member(O2,List2), gs_identifier(O2,GS,OtherID).
inequality_aux(equal(A,B),GS,ID,OtherID) :-
  all_different(equal(A,B),List1,GS),
  select(O1,List1,RestList), gs_identifier(O1,GS,ID),
  member(O2,RestList), gs_identifier(O2,GS,OtherID).

% detect card({x1,....,xn}) = n  as an all-different predicate
all_different(equal(A,B),List,Type) :-
   all_diff_aux(A,B,List,Type) ; all_diff_aux(B,A,List,Type).
all_diff_aux(A,B,List1,Type) :-
   A=b(card(Set1),_,_),
   get_texpr_type(Set1,set(Type)), % TO DO: also match seq type ? (in case we use predicate in other context)
   set_extension_list(Set1,List1),
   evaluable_integer_expression(B,b(truth,pred,[]),Card),
   length(List1,Card).

set_extension_list(b(set_extension(List),_,_),List).
gs_identifier(b(identifier(ID),GS,_),GS,ID).

:- assert_must_succeed(( b_global_sets:is_equality_card_global_set(b(equal(b(card(b(identifier('FACES'),set(global('FACES')),[])),integer,[]),b(integer(6),integer,[])),pred,[]),GS,[],Card),GS='FACES',Card=6)
).
:- assert_must_succeed(( b_global_sets:is_equality_card_global_set(b(equal(b(card(b(value(global_set('TITLE')),set(global('TITLE')),[])),integer,[]),b(integer(4),integer,[])),pred,[]),GS,[],Card),
 GS=='TITLE', Card=4)
).

is_equality_card_global_set(TE,GlobalSet,Properties,Card) :-
   get_texpr_expr(TE,equal(LHS,RHS)),
   get_texpr_expr(LHS,L), get_texpr_expr(RHS,R),
   (  L=card(C),evaluable_integer_expression(R,Properties,Card)
    ; R=card(C),evaluable_integer_expression(L,Properties,Card)), %print(card(Card)),nl,
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).

%evaluable_integer_expression(X,Y) :- print(ev(X,Y)),nl,fail.
% TO DO: maybe we should do some kind of constant folding or call b_compile ??
evaluable_integer_expression(b(E,integer,_),Properties,R) :-
   evaluable_integer_expression(E,Properties,R).
evaluable_integer_expression(max_int,_,R) :- preferences:preference(maxint,R).
evaluable_integer_expression(min_int,_,R) :- preferences:preference(minint,R).
evaluable_integer_expression(identifier(ID),Properties,R) :-
   get_texpr_expr(Eq,equal(LHS,RHS)), get_texpr_id(LHS,ID),
   select_member_in_conjunction(Eq,Properties,RestProp), % avoids loops; but an identifier cannot be used multiple times N+N
   evaluable_integer_expression(RHS,RestProp,R).
evaluable_integer_expression(card(Set),_Properties,Card) :- % only works if we have already treated GlobalSetName ! TO DO: infer CLPFD constraints or treat sets in some order or look for definition in Properties (beware of cycles)
      Set = b(identifier(GlobalSetName),set(global(GlobalSetName)),_),
      extract_setsize_from_machine_cache(GlobalSetName,L,U),
      Card is U+1-L.
evaluable_integer_expression(integer(Card),_,Card).
evaluable_integer_expression(div(A,B),Properties,Res) :-
      evaluable_integer_expression(A,Properties,RA),
      evaluable_integer_expression(B,Properties,RB), Res is RA // RB.
evaluable_integer_expression(multiplication(A,B),Properties,Res) :-
      evaluable_integer_expression(A,Properties,RA),
      evaluable_integer_expression(B,Properties,RB), Res is RA*RB.
evaluable_integer_expression(add(A,B),Properties,Res) :-
      evaluable_integer_expression(A,Properties,RA),
      evaluable_integer_expression(B,Properties,RB), Res is RA+RB.
evaluable_integer_expression(minus(A,B),Properties,Res) :-
      evaluable_integer_expression(A,Properties,RA),
      evaluable_integer_expression(B,Properties,RB), Res is RA-RB.
evaluable_integer_expression(unary_minus(A),Properties,Res) :-
      evaluable_integer_expression(A,Properties,RA), Res is 0-RA.

global_set_identifier(C,GS) :- get_texpr_expr(C,BE), global_set_identifier2(BE,GS).
global_set_identifier2(identifier(GlobalSet),GlobalSet).
global_set_identifier2(value(global_set(GlobalSet)),GlobalSet). % generated by Z

% check for card(GlobalSet) >= Card or similar
is_greaterequalthan_card_global_set(TE,GlobalSet,Properties,Card) :-
   (get_texpr_expr(TE,greater_equal(LHS,RHS)) ;
    get_texpr_expr(TE,less_equal(RHS,LHS)) ),
   get_texpr_expr(LHS,card(C)),
   get_minimum_or_exact_value(RHS,Properties,Card),
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).
is_greaterequalthan_card_global_set(TE,GlobalSet,Properties,Card1) :-
   (get_texpr_expr(TE,greater(LHS,RHS)) ;
    get_texpr_expr(TE,less(RHS,LHS)) ),
   get_texpr_expr(LHS,card(C)),
   get_minimum_or_exact_value(RHS,Properties,Card),
   number(Card), Card1 is Card+1,
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).
is_greaterequalthan_card_global_set(TE,GlobalSet,_Properties,2) :-
  %% preferences:get_preference(allow_empty_global_sets,false), preference no longer exists
  /* GlobalSet /= { ANY } -> as GlobalSet cannot be empty & as ANY must be well-typed:
     we need at least one more element of GlobalSet */
   get_texpr_expr(TE,not_equal(LHS,RHS)),
   ( get_texpr_expr(LHS,identifier(GlobalSet))
     -> get_texpr_expr(RHS,set_extension([b(_X,global(GlobalSet),_)]))
     ;  get_texpr_expr(RHS,identifier(GlobalSet)),
        get_texpr_expr(LHS,set_extension([b(_X,global(GlobalSet),_)]))
   ).
   %%print(not_equal(GlobalSet,_X)),nl.
%  not_equal(S_CHEMIN_ID,b(set_extension([b(identifier(C_NULL_CHEMIN_ID),global(S_CHEMIN_ID),[nodeid(pos(35,1,80,27,80,42)),loc(local,emi,concrete_constants)])]),set(global(S_CHEMIN_ID)),[nodeid(pos(34,1,80,26,80,43))]))

% get mininimum or exact value for an expression
get_minimum_or_exact_value(TE,Properties,N) :-  %print(get_min(TE,Properties)),nl,nl,
   get_texpr_expr(TE,E),
   get_minimum_or_exact_value_aux(E,Properties,N).
get_minimum_or_exact_value_aux(integer(N),_,N).
get_minimum_or_exact_value_aux(identifier(ID),Properties,N) :-
   (extract_min_or_exact_value_for_id(ID,Properties,N) -> true). % avoid backtracking; TO DO: in future all this should be part of constraint solving
get_minimum_or_exact_value_aux(add(A,B),Properties,Res) :- 
      get_minimum_or_exact_value(A,Properties,RA),
      get_minimum_or_exact_value(B,Properties,RB), Res is RA+RB.
% TO DO: other operators
get_minimum_or_exact_value_aux(card(Set),_Properties,Card) :-
      Set = b(identifier(GlobalSetName),set(global(GlobalSetName)),_),
      try_extract_setsize_from_machine(GlobalSetName,L,U), % will try extraction; unless a cycle is hit
      Card is U+1-L.

extract_min_or_exact_value_for_id(ID,Properties,N) :-
   member_in_conjunction(TE,Properties),
   get_texpr_expr(TE,E),
   get_value_bound(E,ID,N).

% equalities already inlined
get_value_bound(equal(LHS,RHS),ID,N) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N).
get_value_bound(equal(RHS,LHS),ID,N) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N).
get_value_bound(greater(LHS,RHS),ID,N1) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N), N1 is N+1.
get_value_bound(less(RHS,LHS),ID,N1) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N), N1 is N+1.
get_value_bound(greater_equal(LHS,RHS),ID,N) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N).
get_value_bound(less_equal(RHS,LHS),ID,N) :- get_texpr_id(LHS,ID), get_texpr_integer(RHS,N).

get_texpr_integer(b(integer(N),_,_),N).


:- volatile global_set_user_defined_scope/2.
:- dynamic global_set_user_defined_scope/2.

/* allow to set scope for specific sets, e.g., via command-line */
set_user_defined_scope(GS,X) :- \+ number(X),!,
  add_error(set_user_defined_scope,'Scope must be number: ',GS:X).
set_user_defined_scope(GS,X) :-
  print('% Setting user defined scope: '), print(GS), print(' == '), print(X),nl,
  assert(global_set_user_defined_scope(GS,X)).

get_user_defined_scope(GlobalSetName,LowBound,UpBound) :-
   (var(GlobalSetName)
    -> add_error(get_user_defined_scope,'Arg is variable: ',get_user_defined_scope) ; true),
   (global_set_user_defined_scope(GlobalSetName,UpBound)
    -> LowBound=1
    ;  extract_setsize_from_defs(GlobalSetName,LowBound,UpBound)).


extract_setsize_from_defs(GlobalSetName,LowBound,UpBound) :-
  b_get_machine_setscope(GlobalSetName,DefTerm),
  get_texpr_expr(DefTerm,DefExpr),
  extract_integer_range(DefExpr,LowBound,UpBound).

extract_integer_range(interval(LB,UB), LowBound, UpBound) :-
     get_texpr_expr(LB,integer(LowBound)),  get_texpr_expr(UB,integer(UpBound)).
extract_integer_range(set_extension(CstList),LowBound,UpBound) :-
     LowBound = 1, length(CstList,UpBound). /* TO DO: extract names of constants in list for pretty printing */
extract_integer_range(integer(UpBound),LowBound,UpBound) :-
     LowBound = 1, UpBound>0.


:- dynamic start_extracting_setsize/1.

start_extracting_setsize_from_machine(GlobalSetName) :- %print(start(GlobalSetName)),nl,
  (start_extracting_setsize(GlobalSetName)
    -> add_internal_error('Cyclic computation: ',start_extracting_setsize_from_machine(GlobalSetName))
    ;  assert(start_extracting_setsize(GlobalSetName))).

assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound) :-
  retractall(start_extracting_setsize(GlobalSetName)),
  %print(finished(GlobalSetName,LowBound,UpBound)),nl,
  (retract(extract_setsize_from_machine_cache(GlobalSetName,L,U))
    -> print(overriding_set_size(GlobalSetName,LowBound-UpBound,L-U)),nl
    ; true),
  assert(extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound)),
  (UpBound<LowBound
   -> print_error('*** illegal empty global set :'),print_error(GlobalSetName)
    ; true
  ).


  %% b_try_update_fd_cardinality has been removed

:- volatile enumerated_set/1, fixed_deferred_set_size/2.
:- dynamic enumerated_set/1, fixed_deferred_set_size/2.
% true if GS is a deferred set whose size was not fixed; i.e., ProB has not inspected all possible instantiations
unfixed_deferred_set(GS) :-
   b_global_set(GS), \+ enumerated_set(GS), \+ fixed_deferred_set_size(GS,_).

contains_unfixed_deferred_set(Type) :- unfixed_deferred_set(Type).
contains_unfixed_deferred_set(global(Type)) :- contains_unfixed_deferred_set(Type).
contains_unfixed_deferred_set(set(Type)) :- contains_unfixed_deferred_set(Type).
contains_unfixed_deferred_set(seq(Type)) :- contains_unfixed_deferred_set(Type).
contains_unfixed_deferred_set(couple(Type1,Type2)) :-
    contains_unfixed_deferred_set(Type1) ; contains_unfixed_deferred_set(Type2).
contains_unfixed_deferred_set(record(Fields)) :-
   ((member(field(_,Type),Fields), contains_unfixed_deferred_set(Type)) -> true).

:- assert_pre(b_global_sets:assert_enumerated_set(_GS,_L,_,_),true).
% TODO(DP, 7.8.2008)
%                (atomic(GS),type_check(L,list(xml_identifier)))).
:- assert_post(b_global_sets:assert_enumerated_set(_,_,L,U),(number(L),number(U))).

assert_enumerated_set(GlobalSetName,ListOfConstants,LowBound,UpBound) :-
  assert(enumerated_set(GlobalSetName)),
 % print_message(enum_set(GlobalSetName,ListOfConstants)),
  LowBound = 1,
  length(ListOfConstants,UpBound),
  assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound),
  add_named_constants_to_global_set(GlobalSetName,ListOfConstants).

add_named_constants_to_global_set(GlobalSetName,ListOfConstants) :-
  (b_global_constant(GlobalSetName,CurNr,CurCst)
    -> add_error_fail(add_named_constants_to_global_set,'Globalset already has constants: ',
                      b_global_constant(GlobalSetName,CurNr,CurCst))
    ; true),
  nth1(Nr,ListOfConstants,Cst),
  (b_global_constant(GS2,_,Cst)
    -> (GS2=GlobalSetName
         -> add_error(b_global_sets,'Duplicate element in global set:',(Cst,GS2))
         ;  add_error(b_global_sets,'Element appears in multiple sets:',(Cst,GS2,GlobalSetName))
       )
    ; true
  ),
  assert(b_global_constant(GlobalSetName,Nr,Cst)),
  assert(lookup_global_constant(Cst,fd(Nr,GlobalSetName))),
  fail.
add_named_constants_to_global_set(_,_).


%add_global_set_with_named_constants(GlobalSetName,ListOfConstants) :-
%  ~~pp_mnf(b_global_sets:assert_enumerated_set(GlobalSetName,ListOfConstants,_,_)),
%  add_new_global_set(GlobalSetName).



:- assert_pre(b_global_sets:b_fd_type(_G,_L,_U),true).
:- assert_post(b_global_sets:b_fd_type(G,L,U),(atomic(G),(integer(L),integer(U)))).

b_fd_type(GlobalSetName,LowBound,UpBound) :-
  b_global_set(GlobalSetName), % was b_get_machine_set(GlobalSetName),
  extract_setsize_from_machine(GlobalSetName,LowBound,UpBound).

% like b_fd_type above, but with GS known (avoids backtracking b_supplementary_global_set)
b_get_fd_type_bounds(NonvarGlobalSetname,LowBound,UpBound) :- nonvar(NonvarGlobalSetname),
   extract_setsize_from_machine_cache(NonvarGlobalSetname,L,U),
   !,
   (LowBound,UpBound)=(L,U).
b_get_fd_type_bounds(GS,L,U) :- add_internal_error('Illegal call: ',b_get_fd_type_bounds(GS,L,U)),fail.

b_global_set_cardinality('STRING',Card) :- !, Card=inf.
b_global_set_cardinality('NAT',Card) :- !, get_preference(maxint,MaxInt), Card is MaxInt+1.
b_global_set_cardinality('NATURAL',Card) :- !,  Card=inf. %get_preference(maxint,MaxInt), Card is MaxInt+1.
b_global_set_cardinality('NAT1',Card) :- !, get_preference(maxint,Card).
b_global_set_cardinality('NATURAL1',Card) :- !, Card=inf. % get_preference(maxint,Card).
b_global_set_cardinality('INT',Card) :- !, get_preference(maxint,MaxInt),
   get_preference(minint,MinInt),
   (MaxInt >= MinInt -> Card is MaxInt-MinInt+1
      ; add_error(b_global_set_cardinality,'MININT larger than MAXINT',(MinInt:MaxInt)),
        Card = 0
   ).
b_global_set_cardinality('INTEGER',Card) :- !, Card=inf. % b_global_set_cardinality('INT',Card).
b_global_set_cardinality(GlobalSet,Card) :- nonvar(GlobalSet),!,
   (b_get_fd_type_bounds(GlobalSet,Low,Up) -> Card is 1+ Up - Low
     ; add_internal_error('Unknown global set: ', b_global_set_cardinality(GlobalSet,Card)),
       fail).
b_global_set_cardinality(GlobalSet,Card) :- %b_global_set(GlobalSet),
  b_fd_type(GlobalSet,Low,Up), Card is 1+ Up - Low.
  % enum_warning_for_global_set(GlobalSet,'assuming deferred set to be of given cardinality',Card,trigger_true(card)).
  %% Note: The Disprover now checks whether unfixed_deferred sets were involved; keeping track of implicit enumerations of deferred sets is just too difficult

b_fd_card(GlobalSet,Card) :- b_get_fd_type_bounds(GlobalSet,Low,Up), Card is 1+ Up - Low.

b_non_empty_global_set(_). % allow_empty_global_sets no longer exists; all global_sets are non-empty
%b_non_empty_global_set(IntSet) :- b_integer_or_string_set(IntSet).
%b_non_empty_global_set(GlobalSet) :- b_global_set(GlobalSet),
%   b_fd_type(GlobalSet,Low,Up), Up >= Low.

b_empty_global_set(_) :- fail.  % allow_empty_global_sets no longer exists; all global_sets are non-empty
%b_empty_global_set(GlobalSet) :- %b_global_set(GlobalSet),
%   b_fd_type(GlobalSet,Low,Up), Up < Low.


b_integer_or_string_set('STRING').
b_integer_or_string_set(X) :- b_integer_set(X).

b_integer_set('NAT').  b_integer_set('NATURAL').
b_integer_set('NAT1').  b_integer_set('NATURAL1').
b_integer_set('INT').  b_integer_set('INTEGER').

% convert a type term into a global_set term
try_b_type2global_set(string,R) :- !, R= global_set('STRING').
try_b_type2global_set(integer,R) :- !, R= global_set('INTEGER').
try_b_type2global_set(global(GS),R) :- b_global_set(GS),!, R=global_set(GS).
try_b_type2global_set(freetype(ID),R) :- !, R=freetype(ID).
try_b_type2global_set(boolean,R) :- !,
   %custom_explicit_sets:construct_avl_from_lists([pred_false  /* bool_false */,pred_true  /* bool_true */],R).
    R = avl_set(node(pred_false,true,1,empty,node(pred_true,true,0,empty,empty))).

b_type2_set(GS,Res) :- try_b_type2global_set(GS,GR),!, Res=GR.
b_type2_set(Type,Res) :- Res = closure(['_zzzz_unary'],[Type],b(truth,pred,[])).
%b_type2global_set(GS,_) :- add_error_and_fail(b_type2_set,'Type cannot be converted to global_set: ',GS).

%:- use_module(fd_utils).
:- use_module(fd_utils_clpfd).

global_type(fd(X,GlobalSet),GlobalSet) :- %b_global_set(GlobalSet),
   if(b_get_fd_type_bounds(GlobalSet,Low,Up), %print(b_get_fd_type_bounds(X,GlobalSet,Low,Up)),nl, when(nonvar(X),trace),
   % print(in_fd(X,Low,Up)),nl,
       in_fd(X,Low,Up),
       add_internal_error('Illegal global set: ',global_type(fd(X,GlobalSet),GlobalSet))).

global_type_wf(fd(X,GlobalSet),GlobalSet,WF) :- %b_global_set(GlobalSet),
   b_get_fd_type_bounds(GlobalSet,Low,Up), %print(b_fd_type_wf(X,GlobalSet,Low,Up)),nl,
   in_fd_wf(X,Low,Up,WF).



% like enum_global_type but with a nonvar GlobalSet + generate enum_warning
enumerate_global_type_with_enum_warning(R,_GS,_EnumWarning) :- ground(R),!.
enumerate_global_type_with_enum_warning(fd(X,GlobalSet),GlobalSet,_EnumWarning) :-
    b_fd_type(GlobalSet,Low,Up),
    %enum_warning_for_global_set(GlobalSet,'assuming deferred set to be of finite cardinality','?',EnumWarning),
    enum_fd(X,Low,Up).

/* No longer used: Disprover now checks whether unfixed deferred sets involved globally
enum_warning_for_global_set(GlobalSet,Message,Card,EnumWarning) :- % print(enum(GlobalSet,Message,Card)),nl,%trace,
   ((preference(disprover_mode,true),
     % we do not know in Disprover mode whether the axioms used to make GlobalSet finite
     % are also part of the hypotheses ---> better be safe and always generate enumeration warning
     unfixed_deferred_set(GlobalSet) % the disprover now only loads type information as a machine. should be good again.
    )
      -> (Card = '?' -> b_fd_card(GlobalSet,RCard),
                        kernel_objects:enum_warning(GlobalSet,Message,RCard,EnumWarning)
                      ; kernel_objects:enum_warning(GlobalSet,Message,Card,EnumWarning))
      ; true).
*/

enum_global_type(fd(X,GlobalSet),GlobalSet) :- %b_global_set(GlobalSet),
   (nonvar(GlobalSet) ->
      (b_get_fd_type_bounds(GlobalSet,Low,Up) -> enum_global_type_aux(X,Low,Up,GlobalSet)
        ; add_internal_error('Illegal value: ',enum_global_type(fd(X,GlobalSet),GlobalSet)),fail)
    ; b_fd_type(GlobalSet,Low,Up),
      enum_global_type_aux(X,Low,Up,GlobalSet)
   ).

enum_global_type_aux(X,Low,Up,GlobalSet) :- atomic(X),!,
  ((number(X),X >= Low, X =< Up) -> true
    ; add_internal_error('Illegal value: ', enum_global_type_aux(X,Low,Up,GlobalSet))).
enum_global_type_aux(X,Low,Up,_GlobalSet) :- enum_fd(X,Low,Up).

% get range information for global set and generate warning if necessary
%b_fd_type_with_enum_warning(GlobalSet,Low,Up) :-
%    b_fd_type(GlobalSet,Low,Up),
%    enum_warning_for_global_set(GlobalSet,'assuming deferred set to be of finite cardinality',Up,trigger_true(card)).

/* ------------------------------------- */


:- block all_elements_of_type(-,?).
all_elements_of_type(Type,Set) :- % print(all_elements(Type,Set)),nl,
    all_elements_of_type2(Type,Set).

:- use_module(kernel_objects,[all_strings/1, enum_warning/4]).
all_elements_of_type2('STRING',Res) :- !, all_strings(Res).
all_elements_of_type2('NAT',Res) :- !, get_preference(maxint,MaxInt),
      gen_int_set(0,MaxInt,Res).
all_elements_of_type2('NATURAL',Res) :- !, get_preference(maxint,MaxInt),
      enum_warning('NATURAL',0:sup,0:MaxInt,trigger_throw('NATURAL')), %print('### Warning: expanding NATURAL'),nl,
      gen_int_set(0,MaxInt,Res).
all_elements_of_type2('NAT1',Res) :- !, get_preference(maxint,MaxInt),
      gen_int_set(1,MaxInt,Res).
all_elements_of_type2('NATURAL1',Res) :- !, get_preference(maxint,MaxInt),
      enum_warning('NATURAL1',1:sup,1:MaxInt,trigger_throw('NATURAL1')), %print('### Warning: expanding NATURAL1'),nl,
      gen_int_set(1,MaxInt,Res).
all_elements_of_type2('INT',Res) :- !, get_preference(maxint,MaxInt),
      get_preference(minint,MinInt),
      gen_int_set(MinInt,MaxInt,Res).
all_elements_of_type2('INTEGER',Res) :- !, get_preference(maxint,MaxInt),
      get_preference(minint,MinInt),
      enum_warning('INTEGER',inf:sup,MinInt:MaxInt,trigger_throw('INTEGER')), %print('### Warning: expanding INTEGER'),nl,
      gen_int_set(MinInt,MaxInt,Res).
all_elements_of_type2(Type,Set) :-
       b_get_fd_type_bounds(Type,Low,Up),
       % ensure we do not use Random enumeration
       findall(fd(El,Type),enum_fd_linear(El,Low,Up),Set).

% a version of all_elements_of_type which may randomise element order if preference set
:- block all_elements_of_type_rand(-,?).
all_elements_of_type_rand(Type,Set) :- % print(all_elements(Type,Set)),nl,
    all_elements_of_type_rand2(Type,Set).

all_elements_of_type_rand2(Type,Set) :- b_global_set(Type),!,
       b_get_fd_type_bounds(Type,Low,Up),
       % may use Random enumeration if preference RANDOMISE_ENUMERATION_ORDER set
       findall(fd(El,Type),enum_fd(El,Low,Up),Set).
all_elements_of_type_rand2(Type,Set) :- % TO DO: we could permute the solution here; but currently all_elements_of_type_rand2 is only called for enumerated/deferred sets
       all_elements_of_type2(Type,Set).


gen_int_set(Min,Max,Set) :-
   ((Min>Max) -> (Set = [])
              ;  (Set = [int(Min)|RSet],
                  M1 is Min+1,
                  gen_int_set(M1,Max,RSet))).
