% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(b_interpreter_eventb,
          [ b_event/6
          , b_event_with_change_set/7
          , b_event_constraints/7
           ]).

:- use_module(tools).
:- use_module(bsyntaxtree).
:- use_module(store).
:- use_module(bmachine).
:- use_module(kernel_objects).
:- use_module(error_manager).
:- use_module(kernel_waitflags).
:- use_module(b_enumerate).
:- use_module(state_space).

:- use_module(b_interpreter).

:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This is an Event-B specific part of the interpreter. It contains the interpreter rules for events (as opposed to operations in classical B.').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Event-B multilevel execution
%
% used shortcuts: IS=InState, OS=OutState, LS=LocalState

% execute an Event:
% difference with b_event_with_change_set : OutState already setup
b_event(TBody,Parameters,NormalisedParaValues,InState,OutState,Path) :-
    get_texpr_expr(TBody,Event),
    b_event_with_change_set_aux(Event,Parameters,NormalisedParaValues,InState,no_change_set_available,OutState,Path).


:- use_module(store,[copy_variable_store/4]).
% a version of b_event which gets the operation ChangeSet passed
b_event_with_change_set(TBody,Parameters,NormalisedParaValues,InState,ChangeSet,Updates,Path) :-
    get_texpr_expr(TBody,Event),
    b_event_with_change_set_aux(Event,Parameters,NormalisedParaValues,InState,ChangeSet,Updates,Path).

/* TO DO: 
b_event_with_change_set_aux(lazy_let_subst(Id,IdExpr,TBody),Parameters,NormalisedParaValues,InState,ChangeSet,Updates,Path) :- !,
  add_lazy_let_id_and_expression(Id,IdExpr,LocalState,InState,NewLocalState,WF),
  b_event_with_change_set(TBody,Parameters,NormalisedParaValues,InState,ChangeSet,Updates,Path). */
b_event_with_change_set_aux(Event,Parameters,NormalisedParaValues,InState,ChangeSet,Updates,Path) :-
    extract_event(Event,Name,Section,Guard),
    setup_event_state(Parameters,ParaValues,NormalisedParaValues,ParamTypedVals,LS),
    % evaluate the event's guard
    evaluate_predicate_if_necessary(Guard,ParamTypedVals,LS,InState),
    % execute the event
    init_trace_of_event(Name,Section,Parameters,ParaValues,Guard,Path,RestPath,Desc),
    (ChangeSet==no_change_set_available -> OutState=Updates
      ; copy_variable_store(InState,ChangeSet,Updates,OutState)), 
    execute_event_body(Event,InState,LS,OutState,RestPath,Desc).

evaluate_predicate_if_necessary(Guard,ParamTypedVals,LS,InState) :-
    ((pge_algo: do_not_evaluate_guard, ParamTypedVals==[]) ->
        % if we already know that the event is enabled
        % in the current state, we don't need to evaluate the guard here
        debug_println(9,guard_will_not_be_evaluated_b_event)
    ; otherwise -> 
        evaluate_predicate(Guard,ParamTypedVals,LS,InState)
    ).

% TO DO: add support for lazy_let_subst by calling add_lazy_let_id_and_expression
:- use_module(error_manager,[add_error/2]).
extract_event(Event,Name,Section,FullGuard) :-
    Event = rlevent(Name,Section,_Status,_Params,Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),
    !,
    % check for user-defined additional guard
    full_guard(Guard,FullGuard),
    ((Name='INITIALISATION',FullGuard \== Guard)
        % The guard is evaluated on the state before ! It should be executed on the state after ! %% TO DO
      -> add_error(b_interpreter_eventb,'Additional guards not yet supported for Event-B INITIALISATION')
      ; true).
extract_event(Event,Name,Section,FullGuard) :-
     add_internal_error('Illegal Event-B Event: ',extract_event(Event,Name,Section,FullGuard)),fail.

setup_event_state(Parameters,ParaValues,NormalisedParaValues,ParamTypedVals,LS) :-
    empty_state(EmptyState),
    set_up_typed_localstate(Parameters,ParaValues,ParamTypedVals,EmptyState,LS,positive), 
    store:l_expand_and_normalise_values(ParaValues,NormalisedParaValues,Parameters).

init_trace_of_event(Name,Section,Parameters,ParaValues,FullGuard,[event(Path),eventtrace(Trace)],RestPath,D2) :-
    Path = [event(Name,ParaValues)|RestPath],
    eventtrace_init(Name,Desc,Trace),
    eventtrace_add(event(Name,Section),Desc,D1),
    eventtrace_add(true_guard(Parameters,ParaValues,FullGuard),D1,D2).

full_guard(Guard,FullGuard) :-
    bmachine:b_machine_temp_predicate(TempPred),!,
    conjunct_predicates([TempPred,Guard],FullGuard).
full_guard(Guard,Guard).

execute_event_body(Event,InState,LS,OutState,AbstractPath,D) :-
    Event = rlevent(_Name,_Section,Status,_Params,_Guard,Theorems,Actions,VWitnesses,PWitnesses,Unmodifiables,AbstractEvents),
    % for convergent/anticipated events: check if the variant is non-negative
    check_event_entry_status(Status,InState,LS,VariantValue,D,D2),
    % check if the theorems hold
    check_event_theorems(Theorems,InState,LS,D2,D3),
    % execute the actions
    b_execute_eventb_actions(Actions,InState,LS,OutState,D3,D4),
    % for convergent/anticipated events: check if the variant has decreased/not increased
    check_event_exit_status(VariantValue,Status,OutState,LS,D4,D5),
    check_if_invalid_vars_are_modified(Unmodifiables,InState,OutState,D5),
    %print(go_for_abstract(AbstractEvents,VWitnesses,PWitnesses)),nl,
    go_for_abstract_events(AbstractEvents,VWitnesses,PWitnesses,InState,LS,OutState,AbstractPath,D5).


go_for_abstract_events([],_VWit,_PWit,_IS,_LS,_OS,[],D) :-
    % no more refined level exists, stop here
    !, eventtrace_finalise(D,_).
go_for_abstract_events(AbstractEvents,[],[],IS,LS,OS,AbstractPath,Din) :-
    !, % optimisation: If no witness has to be evaluated, do not copy primed variables
    execute_abstract_events(AbstractEvents,[],[],IS,LS,OS,AbstractPath,Din).
go_for_abstract_events(AbstractEvents,VWitnesses,PWitnesses,IS,LS,OS,AbstractPath,Din) :-
    % copy a primed version of all values in the outstate into the local state
    copy_primed(OS,LS,NewLS),
    % evaluate witnesses for dropped abstract vars
    evaluate_witnesses_for_abstract_vars(VWitnesses,IS,NewLS,OS,Din,D1),
    % evaluate witnesses for dropped parameters
    evaluate_witnesses_for_abstract_parameters(PWitnesses,IS,NewLS,AbsLS,Params,PValues,D1,D2),
    % continue with the next refinement level
    execute_abstract_events(AbstractEvents,Params,PValues,IS,AbsLS,OS,AbstractPath,D2).

check_event_entry_status(TStatus,IS,LS,VariantValue,Din,Dout) :-
    get_texpr_expr(TStatus,Status),
    check_event_entry_status2(Status,IS,LS,VariantValue,Din,Dout).
check_event_entry_status2(ordinary,_IS,_LS,-,D,D).
check_event_entry_status2(anticipated(Variant),IS,LS,Val,Din,Dout) :-
    variant_is_not_negative(Variant,anticipated,IS,LS,Val,Din,Dout).
check_event_entry_status2(convergent(Variant),IS,LS,Val,Din,Dout) :-
    variant_is_not_negative(Variant,convergent,IS,LS,Val,Din,Dout).
variant_is_not_negative(Variant,_CType,IS,LS,Value,D,D) :-
    get_texpr_type(Variant, set(_)),!,
    % if variant is of type set, it cannot be negative
    % nevertheless, we compute its value here to check later if the set has been decreased
    b_compute_expression(Variant,LS,IS,Value).  % Daniel: why are LS & IS swapped ?
variant_is_not_negative(Variant,CType,IS,LS,Value,Din,Dout) :-
   % print_bt_message(computing_variant), print_bexpr(Variant),nl,
   % print(local(LS)), nl, print(global(IS)),nl,
    b_compute_expression(Variant,LS,IS,Value), % Daniel: why are LS & IS swapped ?
    %print_bt_message(variant(Value)),
    ( less_than(Value, int(0)) ->
        eventtrace_store_event_error(variant_negative(CType,Variant,Value),Din,Dout)
    ; otherwise -> eventtrace_add(variant_checked_pre(CType,Variant,Value),Din,Dout)).

check_event_exit_status(-,_TStatus,_OS,_LS,D,D) :- !.
check_event_exit_status(EntryValue,TStatus,OS,LS,Din,Dout) :-
    get_texpr_expr(TStatus,Status),
    % Status is convergent(Expr) or anticipated(Expr)
    functor(Status,CType,1),arg(1,Status,Expr),
    b_compute_expression(Expr,LS,OS,ExitValue),
    get_texpr_type(Expr,Type),
    ( variant_condition(CType,Type,EntryValue,ExitValue) ->
        eventtrace_add(variant_checked_post(CType,Expr,EntryValue,ExitValue),Din,Dout)
    ; otherwise ->
        eventtrace_store_event_error(invalid_variant(CType,Expr,EntryValue,ExitValue),Din,Dout)).

variant_condition(anticipated, integer, Pre, Post) :- !, less_than_equal(Post,Pre).
variant_condition(anticipated, set(_),  Pre, Post) :- !, check_subset_of(Post,Pre).
variant_condition(convergent,  integer, Pre, Post) :- !, less_than(Post,Pre).
variant_condition(convergent,  set(_),  Pre, Post) :- !, strict_subset_of(Post,Pre).

check_event_theorems([],_State,_LState,D,D).
check_event_theorems([Theorem|Trest],State,LState,Din,Dout) :-
    check_event_theorem(Theorem,State,LState,Din,Dinter),
    check_event_theorems(Trest,State,LState,Dinter,Dout).
check_event_theorem(Theorem,State,LState,Din,Dout) :-
    ( evaluate_predicate(Theorem,[],LState,State) ->
        Din = Dout
        %eventtrace_add(theorem_checked(Theorem),Din,Dout)
    ; otherwise ->
        eventtrace_store_event_error(invalid_theorem_in_guard(Theorem),Din,Dout)
    ).

% check if variables that are not modified by an abstract event or newly introduced into
% the current refinement level are modified.
check_if_invalid_vars_are_modified([],_IS,_OS,_D).
check_if_invalid_vars_are_modified([Unmod|UnmodRest],IS,OS,D) :-
    check_unmodifiable(Unmod,IS,OS,D),
    check_if_invalid_vars_are_modified(UnmodRest,IS,OS,D).
check_unmodifiable(Var,IS,OS,D) :-
    def_get_texpr_id(Var,Id),
    get_id_value(Id,IS,Pre),!,
    get_id_value(Id,OS,Post),!,
    ( equal_object(Pre,Post) -> true
    ; otherwise ->
        eventtrace_failure(invalid_modification(Var,Pre,Post),D)).

evaluate_witnesses_for_abstract_parameters(Witnesses,IS,LS,AbsLS,Params,PValues,Din,Dout) :-
    evaluate_witnesses_for_abstract_parameters2(Witnesses,IS,LS,LS,AbsLS,Params,PValues,Din,Dout).
evaluate_witnesses_for_abstract_parameters2([],_State,_OrigLS,LS,LS,[],[],D,D).
evaluate_witnesses_for_abstract_parameters2([Witness|WRest],State,OrigLS,LS,AbsLS,[TId|Params],[Value|PValues],Din,Dout) :-
    witness_substate(Witness,OrigLS,LS,TId,Value,Predicate,TypedVals,WitnessLS,ExtLS),
    if( evaluate_predicate(Predicate,TypedVals,WitnessLS,State),
        ( eventtrace_add(eval_witness(parameter,TId,Value,Predicate),Din,D1),
            evaluate_witnesses_for_abstract_parameters2(WRest,State,OrigLS,ExtLS,AbsLS,Params,PValues,D1,Dout)),
        eventtrace_failure(no_witness_found(parameter,TId,Predicate), Din) ).

witness_substate(Witness,OrigLS,LS,TId,Value,Predicate,TypedVals,WitnessLS,ExtLS) :-
    extract_witness(Witness,TId,Id,_Type,Predicate),
    set_up_typed_localstate([TId],[Value],TypedVals,LS,ExtLS,positive),
    add_var_to_localstate(Id,Value,OrigLS,WitnessLS).

extract_witness(Witness,TId,Id,Type,Predicate) :-
    get_texpr_expr(Witness,witness(TId,Predicate)),
    def_get_texpr_id(TId,Id),
    get_texpr_type(TId,Type).

eventtrace_init(EventName, desc(EventName,D,D), D).
eventtrace_add( Entry, desc(EN,D,[Entry|L]), desc(EN,D,L) ).
eventtrace_name( desc(EN,_,_), EN).
eventtrace_finalise(desc(_EN,D,[]),D).
eventtrace_store_event_error(Error,Din,Dout) :-
    eventtrace_name(Din, EventName),
    copy_term(Din,Dcopy),
    eventtrace_add(Error,Dcopy,Dc1),
    eventtrace_finalise(Dc1,D),
    store_error_for_context_state(eventerror(EventName,Error,D),Id),
    eventtrace_add(error(Error,Id),Din,Dout).
eventtrace_failure(Error,Desc) :-
    eventtrace_name(Desc,EventName),
    eventtrace_add(Error,Desc,D1),
    eventtrace_finalise(D1,D),
    store_error_for_context_state(eventerror(EventName,Error,D),_Id),
    fail.

evaluate_predicate(Predicate,TypedVals,LS,State) :-
    init_wait_flags(WF),
    enumerate_predicate(Predicate,TypedVals,LS,State,WF),
    ground_wait_flags(WF).

enumerate_predicate(Predicate,TypedVals,LS,State,WF) :-
    b_tighter_enumerate_values(TypedVals,WF),
    b_test_boolean_expression(Predicate,LS,State,WF).

execute_abstract_events(Events,Params,PValues,IS,LS,OS,[event(Name,PValues)|Path],Din) :-
    if( select_event(Events,Params,PValues,IS,LS,Event,Name,Din,D1),
        execute_event_body(Event,IS,LS,OS,Path,D1),
        ( generate_simulation_error(Events,Error), eventtrace_failure(Error,Din) )).

generate_simulation_error(Events,simulation_error(FailedEvents)) :-
    generate_simulation_error2(Events,FailedEvents).
generate_simulation_error2([],[]).
generate_simulation_error2([TEvent|Erest],[event(Name,Section,Guard)|Srest]) :-
    get_texpr_expr(TEvent,Event),
    Event = rlevent(Name,Section,_Status,_Params,Guard,_Thms,_Actions,_VWitnesses,_PWitnesses,_Unmodifiables,_AbstractEvents),
    generate_simulation_error2(Erest,Srest).

select_event(Events,Params,PValues,IS,LS,Event,Name,Din,Dout) :-
    member(TEvent,Events),
    get_texpr_expr(TEvent,Event),
    Event = rlevent(Name,Section,_Status,_Params,Guard,_Thms,_Act,_VWit,_PWit,_Unmod,_Evt),
    b_test_boolean_expression_wf(Guard,LS,IS),
    eventtrace_add(event(Name,Section),Din,D1),
    eventtrace_add(true_guard(Params,PValues,Guard),D1,Dout).

evaluate_witnesses_for_abstract_vars([],_IS,_LS,_OS,D,D).
evaluate_witnesses_for_abstract_vars([Witness|WRest],IS,LS,OS,Din,Dout) :-
    extract_witness(Witness,TId,Id,Type,Predicate),
    if( (init_wait_flags(WF),
            b_enumerate_values_in_store([Id],[Type],[Value],LS,WF),
            b_test_boolean_expression(Predicate,LS,IS,WF),
            ground_wait_flags(WF)),
        ( eventtrace_add(eval_witness(variable,TId,Value,Predicate),Din,D1),
            evaluate_witnesses_for_abstract_vars(WRest,IS,LS,OS,D1,Dout)),
        eventtrace_failure(no_witness_found(variable,TId,Predicate),Din) ).

b_execute_eventb_actions([],_IS,_LS,_OS,D,D).
b_execute_eventb_actions([TAction|ARest],IS,LS,OS,Din,Dout) :-
    get_texpr_expr(TAction,Action),
    if( ( init_wait_flags(WF),
            b_execute_eventb_action(Action,IS,LS,OS,WF,Step),
            ground_wait_flags(WF)),
        ( eventtrace_add(Step,Din,D1),
          b_execute_eventb_actions(ARest,IS,LS,OS,D1,Dout)),
        eventtrace_failure(action_not_executable(TAction),Din)).
%get_assigned_vars(TAction,Assigned) :-
%    get_texpr_info(TAction,Info),
%    member(assigned_after(Assigned),Info),!.
b_execute_eventb_action(assign(Lhs,Exprs),IS,LS,OS,WF,Step) :-
    b_interpreter:b_compute_expressions(Exprs,LS,IS,VALs,WF), % VALs - a list of computed values 
    b_interpreter:b_assign_values_or_functions(Lhs,VALs,LS,IS,NewValues,WF,output_required),
    eventb_store_new_values(NewValues,OS),
    Step = action(Lhs,Exprs,VALs).
b_execute_eventb_action(becomes_element_of(Lhs,Rhs),IS,LS,OS,WF,Step) :-
    convert_pairs_into_list(Lhs,Val,ValList),
    get_texpr_types(Lhs,Types),
    b_compute_expression(Rhs,LS,IS,ValueSet,WF),
    check_element_of_wf(Val,ValueSet,WF),
    b_enumerate_typed_values(ValList,Types,WF),
    % ground_det_wait_flag(WF),
    b_interpreter:b_assign_values_or_functions(Lhs,ValList,LS,IS,NewValues,WF,output_required),
    eventb_store_new_values(NewValues,OS),
    Step = action_set(Lhs,Rhs,ValueSet,ValList).
b_execute_eventb_action(becomes_such(Ids,Pred),IS,LS,OS,WF,Step) :-
    add_primed_vars(Ids,OS,LS,NewLS,TypedVals,Values),
    b_enumerate_values(TypedVals,WF),
    b_test_boolean_expression(Pred,NewLS,IS,WF),
    Step = action_pred(Ids,Pred,Values).

% primed vars always trigger an enum warning
% this might be split up depending on where they are introduced
add_primed_vars([],_State,LS,LS,[],[]).
add_primed_vars([TId|IRest],State,In,Out,[typedval(Value,Type,Id,trigger_true(Primed))|TypeValRest],[Value|Vrest]) :-
    def_get_texpr_id(TId,Id), get_texpr_type(TId,Type),
    lookup_value(Id,State,Value),
    atom_concat(Id,'\'',Primed),
    add_var_to_localstate(Primed,Value,In,Inter),
    add_primed_vars(IRest,State,Inter,Out,TypeValRest,Vrest).

eventb_store_new_values([],_Store).
eventb_store_new_values([bind(Id,Value)|IRest],Store) :-
    lookup_value(Id,Store,StoreValue),
    equal_object(StoreValue,Value),
    eventb_store_new_values(IRest,Store).

% copy_primed(+Store, +InStore, -OutStore)
% copies and primes all variables from Store
% TO DO: maybe only copy those that are really used as primed variables in witnesses,...
copy_primed([], InStore, InStore).
copy_primed([bind(OldVar,Value)|Srest], InStore, OutStore) :-
    atom_concat(OldVar,'\'',NewVar), % is faster than caching this using a dynamic fact
    copy_primed(Srest,[bind(NewVar,Value)|InStore], OutStore).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Multilevel-animation: Version that is intended to set up constraints
% between two states and not to find errors in one event

b_event_constraints(TBody,Parameters,NormalisedParaValues,IS,OS,Path,WF) :-
    get_texpr_expr(TBody,Event),
    b_event_constraints_aux(Event,Parameters,NormalisedParaValues,IS,OS,Path,WF).

b_event_constraints_aux(Event,Parameters,NormalisedParaValues,IS,OS,Path,WF) :-
    extract_event(Event,Name,Section,Guard),
    setup_event_state(Parameters,ParaValues,NormalisedParaValues,ParamTypedVals,LS),
    % evaluate guard
    enumerate_predicate(Guard,ParamTypedVals,LS,IS,WF),
    % execute the event
    init_trace_of_event(Name,Section,Parameters,ParaValues,Guard,Path,RestPath,Desc),
    event_body_constraints(Event,IS,LS,OS,RestPath,Desc,WF).

event_body_constraints(Event,IS,LS,OS,AbstractPath,D,WF) :-
    Event = rlevent(_Name,_Section,_Status,_Params,_Guard,_Thms,Actions,VWitnesses,PWitnesses,_Unmodifiables,AbstractEvents),
    % for convergent/anticipated events: check if the variant is non-negative
    % check_event_entry_status(Status,IS,LS,VariantValue,D,D2),
    eventb_actions_constraints(Actions,IS,LS,OS,D,D2,WF),
    % for convergent/anticipated events: check if the variant has decreased/not increased
    % check_event_exit_status(VariantValue,Status,OS,LS,D3,D4),
    % check_if_invalid_vars_are_modified(Unmodifiables,IS,OS,D4),
    abstract_event_constraints(AbstractEvents,VWitnesses,PWitnesses,IS,LS,OS,AbstractPath,D2,WF).

eventb_actions_constraints([],_IS,_LS,_OS,D,D,_WF).
eventb_actions_constraints([TAction|ARest],IS,LS,OS,Din,Dout,WF) :-
    get_texpr_expr(TAction,Action),
    b_execute_eventb_action(Action,IS,LS,OS,WF,Step),
    eventtrace_add(Step,Din,D1),
    eventb_actions_constraints(ARest,IS,LS,OS,D1,Dout,WF).

abstract_event_constraints([],_VWit,_PWit,_IS,_LS,_OS,[],D,_WF) :-
    % no more refined level exists, stop here
    !, eventtrace_finalise(D,_).
abstract_event_constraints(AbstractEvents,VWitnesses,PWitnesses,IS,LS,OS,AbstractPath,Din,WF) :-
    % copy a primed version of all values in the outstate
    % into the local state
    copy_primed(OS,LS,NewLS),
    % evaluate witnesses for dropped abstract vars
    abstract_var_witnesses_constraints(VWitnesses,IS,NewLS,OS,Din,D1,WF),
    % evaluate witnesses for dropped parameters
    abstract_parameter_witnesses_constraints(PWitnesses,IS,NewLS,AbsLS,_Params,_PValues,D1,D2,WF),
    % continue with the next refinement level
    execute_abstract_events(AbstractEvents,IS,AbsLS,OS,AbstractPath,D2,WF).

abstract_var_witnesses_constraints([],_IS,_LS,_OS,D,D,_WF).
abstract_var_witnesses_constraints([Witness|WRest],IS,LS,OS,Din,Dout,WF) :-
    extract_witness(Witness,TId,Id,Type,Predicate),
    b_enumerate_values_in_store([Id],[Type],[Value],LS,WF),
    b_test_boolean_expression(Predicate,LS,IS,WF),
    eventtrace_add(eval_witness(variable,TId,Value,Predicate),Din,D1),
    abstract_var_witnesses_constraints(WRest,IS,LS,OS,D1,Dout,WF).

abstract_parameter_witnesses_constraints(Witnesses,IS,LS,AbsLS,Params,PValues,Din,Dout,WF) :-
    abstract_parameter_witnesses_constraints2(Witnesses,IS,LS,LS,AbsLS,Params,PValues,Din,Dout,WF).
abstract_parameter_witnesses_constraints2([],_State,_OrigLS,LS,LS,[],[],D,D,_WF).
abstract_parameter_witnesses_constraints2([Witness|WRest],State,OrigLS,LS,AbsLS,[TId|Params],[Value|PValues],Din,Dout,WF) :-
    witness_substate(Witness,OrigLS,LS,TId,Value,Predicate,TypedVals,WitnessLS,ExtLS),
    enumerate_predicate(Predicate,TypedVals,WitnessLS,State,WF),
    eventtrace_add(eval_witness(parameter,TId,Value,Predicate),Din,D1),
    abstract_parameter_witnesses_constraints2(WRest,State,OrigLS,ExtLS,AbsLS,Params,PValues,D1,Dout,WF).

execute_abstract_events(AbstractEvents,IS,LS,OS,Path,D,WF) :-
    % This introduces a choicepoint when there are more than one abstract event.
    % But the usual case is just one abstract event
    member(TEvent,AbstractEvents),
    get_texpr_expr(TEvent,Event),
    event_body_constraints(Event,IS,LS,OS,Path,D,WF).
