% (c) 2009-2014 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bmachine_eventb,[check_event_b_project/4
                          ,prime_variable/2
                          ,prime_variables/2
                          ,deferred_set_equality_without_enumeration_axioms/2
                          ,is_eventb_additional_info/1
                          ,raw_event/10
                          ,is_primed_id/1
                          ]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(codesio),[read_from_codes/2]).

:- use_module(error_manager).
:- use_module(tools,[remove_all/3,foldl/4,foldl/5,foldl/6,ajoin_with_sep/3,unique_id/2]).
:- use_module(self_check).
:- use_module(btypechecker).
:- use_module(b_ast_cleanup).
:- use_module(bsyntaxtree).
:- use_module(debug).
:- use_module(bmachine_construction).
:- use_module(preferences,[get_preference/2]).
:- use_module(input_syntax_tree,[extract_raw_identifiers/2,create_fresh_identifier/3]).
:- use_module(translate,[translate_bexpression/2]).
:- use_module(pragmas,[extract_pragmas_from_event_b_extensions/2,apply_pragmas_to_event_b_machine/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module contains the rules for loading a Event-B machine or context. ').

:- use_module(specfile,[eventb_mode/0]).
deferred_set_equality_without_enumeration_axioms(Set,ExtSet) :-
   eventb_mode,
   deferred_set_eq_wo_enumeration_axioms(Set,ExtSet),
   % there is a single equality Set = {id1,...,idn} and no partition or not enough disequalities
   % quite often the user forgot to add those disequalities
   \+ (deferred_set_eq_wo_enumeration_axioms(Set,ExtS2), ExtS2 \= ExtSet).

:- dynamic deferred_set_eq_wo_enumeration_axioms/2.

check_event_b_project(RawModels,RawContextes,Extensions,MachineWithPragmas) :-
    retractall(deferred_set_eq_wo_enumeration_axioms(_,_)),
    catch( check_event_b_project1(RawModels,RawContextes,Extensions,MachineWithPragmas),
           typecheck_errors(Errors),
           ( add_all_perrors(Errors),fail )).
check_event_b_project1([],RawContextes,Extensions,MachineWithPragmas) :-
    !,
    check_extensions(Extensions,Freetypes,Ops,InitEnv),   % Theories
    check_contextes(InitEnv,RawContextes,Contextes),
    convert_event_b(context,Freetypes,Ops,Contextes,[],[],Machine),
    extract_pragmas_from_event_b_extensions(Extensions,Pragmas),
    apply_pragmas_to_event_b_machine(Pragmas,Machine,MachineWithPragmas).
check_event_b_project1(RawModels,RawContextes,Extensions,MachineWithPragmas) :-
    compute_animation_chain(RawModels,AnimationChain),
    check_event_b_project2(RawModels,RawContextes,Extensions,
                           AnimationChain,Machine),
    extract_pragmas_from_event_b_extensions(Extensions,Pragmas),
    apply_pragmas_to_event_b_machine(Pragmas,Machine,MachineWithPragmas).

check_event_b_project2(RawModels,RawContextes,Extensions,AnimationChain,Machine) :-
    RawModels=[event_b_model(_,Main,_)|_],
    check_extensions(Extensions,Freetypes,Ops,InitEnv),
    % generate type-checked contextes
    check_contextes(InitEnv,RawContextes,Contextes),
    % generate type-checked models
    check_models(InitEnv,RawModels,Contextes,Models),
    % merge all the constants/vars/predicates and events into one machine
    convert_event_b(Main,Freetypes,Ops,Contextes,Models,AnimationChain,Machine).

% the filter chain contains the name of all models that
% should be contained in the animation
compute_animation_chain(Models,AnimationChain) :-
    get_preference(number_animated_abstractions,N),
    debug_println(9,compute_animation_chain(N)),
    length(Models,L),
    N1 is min(N+1,L),
    prefix_length(Models, AnimatedModels, N1),
    extract_model_names(AnimatedModels,AnimationChain).
extract_model_names([],[]).
extract_model_names([event_b_model(_,Name,_)|MRest],[Name|NRest]) :-
    extract_model_names(MRest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% context

check_contextes(InitEnv,RawContextes,Contextes) :-
    debug_println(9,'Checking Contextes'),
    % make sure that a extended context is treated before
    % the extending context
    topological_sort(RawContextes,SortedContextes),
    check_contextes2(SortedContextes,InitEnv,[],Contextes).
check_contextes2([],_,_,[]).
check_contextes2([RawContext|Rrest],InitEnv,PrevContextes,[Context|Crest]) :-
    check_context(RawContext,InitEnv,PrevContextes,Context),
    check_contextes2(Rrest,InitEnv,[Context|PrevContextes],Crest).
    
check_context(RawContext,InitEnv,PrevContextes,Context) :-
    extract_ctx_sections(RawContext,Name,Extends,
                         RawSets,RawConstants,RawAbstractConstants,RawAxioms,RawTheorems),
    Context = context(Name,Extends,Sets,Constants,AbstractConstants,Axioms,Theorems),

    % create deferred sets
    create_sets(RawSets,Name,Sets),
    % and constants
    IdInfos = [loc(Name,constants),section(Name)],
    create_identifiers(RawConstants,IdInfos,Constants),
    create_identifiers(RawAbstractConstants,IdInfos,AbstractConstants),

    % with the constants and set of seen contextes,
    create_ctx_env(Extends,PrevContextes,InitEnv,PrevEnv),
    % and the sets and constants
    add_unique_variables(Sets,PrevEnv,SetEnv),
    add_unique_variables(Constants,SetEnv,EnvT),
    add_unique_variables(AbstractConstants,EnvT,Env),

    % now typecheck the axioms and theorems,
    % this should determine the constant's types
    typecheck_together([RawAxioms,RawTheorems],Env,[Axioms,Theorems],axioms).

extract_ctx_sections(RawContext,Name,Extends,RawSets,RawConstants,RawAbstractConstants,RawAxioms,RawTheorems) :-
    RawContext = event_b_context(_Pos,Name,Sections),
    optional_rawmachine_section(extends,Sections,[],Extends),
    optional_rawmachine_section(sets,Sections,[],RawSets),
    optional_rawmachine_section(constants,Sections,[],RawConstants),
    optional_rawmachine_section(abstract_constants,Sections,[],RawAbstractConstants),
    optional_rawmachine_section(axioms,Sections,[],RawAxioms),
    optional_rawmachine_section(theorems,Sections,[],RawTheorems).


create_sets([],_,[]).
create_sets([RawSet|RSrest],CtxName,[Set|Srest]) :-
    ext2int(RawSet,deferred_set(SetName),_,set(global(SetName)),
            identifier(SetName),[section(CtxName),given_set],Set),
    create_sets(RSrest,CtxName,Srest).

create_identifiers([],_,[]).
create_identifiers([RawIdentifier|RIrest],Infos,[Identifier|Irest]) :-
    ext2int(RawIdentifier,I,_,_,I,Infos,Identifier),
    I=identifier(_),
    create_identifiers(RIrest,Infos,Irest).

create_ctx_env(Extends, Contextes, Old, New) :-
    % find all seen contextes (and their extended contextes,
    % and again their extended contextes, ...)
    transitive_contextes(Extends,Contextes,TransExt),
    % and all variables of extended machines
    add_env_vars_for_extended2(TransExt,Contextes,Old,New).
transitive_contextes(Extends,Contextes,All) :-
    findall(E, transext(Extends,Contextes,E), All1), sort(All1, All).
transext(Extends,_,E) :- member(E,Extends).
transext(Extends,Contextes,E) :-
    member(Name,Extends),
    internal_context_extends(Context,Name,NewExtends),
    member(Context,Contextes),
    transext(NewExtends,Contextes,E).
% add_env_vars_for_extended2(Names,Contextes,OldEnv,NewEnv):
%  Names: a list of the names of the contextes whose constants and sets
%     should be added to the type environment
%  Contextes: a list of all existing contextes
%  OldEnv: the input environment
%  NewEnv: the input environment plus the constants and sets
add_env_vars_for_extended2([],_,Env,Env).
add_env_vars_for_extended2([Name|Erest],Contextes,Old,New) :-
    internal_context_sets(Context,Name,Sets),
    internal_context_constants(Context,Name,Constants),
    internal_context_abstract_constants(Context,Name,AbstractConstants),
    member(Context,Contextes),!,
    add_unique_variables(Sets,Old,Env1),
    add_unique_variables(Constants,Env1,Env2),
    add_unique_variables(AbstractConstants,Env2,Env),
    add_env_vars_for_extended2(Erest,Contextes,Env,New).

% copy&paste from bmachine_construction
add_unique_variables([],Env,Env).
add_unique_variables([Var|Rest],Old,New) :-
    def_get_texpr_id(Var,Name),
    get_texpr_type(Var,Type),
    get_texpr_info(Var,Infos),
    ( env_lookup_type(Name,Old,_) ->
        add_error(bmachine_eventb,'Identifier declared twice',Name),
        fail
    ; otherwise -> true),
    env_store(Name,Type,Infos,Old,Env),
    add_unique_variables(Rest,Env,New).

% add the primed (') versions of the variables to the
% type environment
add_primed_variables(Variables,OrigEnv,NewEnv) :-
    prime_variables(Variables,Primed),
    add_unique_variables(Primed,OrigEnv,NewEnv).
% add a prime (') to each typed identifier
prime_variables([],[]).
prime_variables([V|Vrest],[P|Prest]) :-
    prime_variable(V,P),
    prime_variables(Vrest,Prest).
% add a prime (') to a typed identifier
prime_variable(V,P) :-
    def_get_texpr_id(V,Id),
    prime_id(Id,PId),
    rename_bt(V,[rename(Id,PId)],P).
% add (or remove) a prime (') to a plain identifier (without syntax structure)
prime_id(UnPrimed,Primed) :-
    atom_concat(UnPrimed,'\'',Primed).

check_for_errors(Errors) :-
% TO DO: extract and add warnings in list
    Errors = [_|_],!,throw(typecheck_errors(Errors)).
check_for_errors([]).

typecheck_l(Untyped,Env,Type,Typed,Context) :-
    all_same(Untyped,Type,Types),
    btype_ground_dl(Untyped,Env,[],Types,Typed1,Errors,[]),
    check_for_errors(Errors),
    clean_up_l_with_optimizations(Typed1,[],Typed,Context),!.
typecheck_l(Untyped,Env,Type,Typed,Context) :- 
     add_failed_call_error(typecheck_l(Untyped,Env,Type,Typed,Context)),fail.
all_same([],_,[]).
all_same([_|R],E,[E|T]) :- all_same(R,E,T).

typecheck(Untyped,Env,Type,Typed) :-
    typecheck_with_freevars(Untyped,Env,[],Type,Typed).
typecheck_with_freevars(Untyped,Env,Freevars,Type,Typed) :-
    btype_ground_dl([Untyped],Env,Freevars,[Type],[Typed1],Errors,[]),
    clean_up_pred_or_expr(Freevars,Typed1,Typed),
    check_for_errors(Errors).





%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% models

check_models(InitEnv,RawModels,Contextes,Models) :- 
    debug_println(9,'CHECKING MODELS'),
    topological_sort(RawModels,SortedModels),
    check_models2(SortedModels,InitEnv,[],Contextes,Models).

check_models2([],_,_,_,[]).
check_models2([RawModel|RMrest],InitEnv,PrevModels,Contextes,[Model|Mrest]) :-
    (check_model(RawModel,InitEnv,PrevModels,Contextes,Model) -> true
      ; add_failed_call_error(check_model(RawModel,InitEnv,PrevModels,Contextes,Model)),
        fail),
    check_models2(RMrest,InitEnv,[Model|PrevModels],Contextes,Mrest).

check_model(RawModel,InitEnv,PrevModels,Contextes,Model) :- 
    extract_model_sections(RawModel,Name,Refines,Sees,
                           RawVariables,RawInvariant,
                           RawTheorems,RawVariant,RawEvents),
    Model = model(Name,Refines,Sees,ConcreteVariables,NewVariables,Invariant,Theorems,Events),
    get_refined_model(Refines,PrevModels,RefinedModel),

    get_all_abstract_variables(Refines,PrevModels,AbstractionVariables),
    length(PrevModels,Level),
    IdInfos = [loc(Name,variables),section(Name),level(Level)],
    create_identifiers(RawVariables,IdInfos,ConcreteVariables),
    split_abstract_vars(AbstractionVariables,ConcreteVariables,
                        NewVariables,DroppedAbstractVars),

    create_ctx_env(Sees,Contextes,InitEnv,CtxEnv),
    add_unique_variables(ConcreteVariables,CtxEnv,ModelEnv),
    add_unique_variables(DroppedAbstractVars,ModelEnv,InvEnv),
    % add primed versions of the concrete variables to
    % permit before-after predicates in the witnesses
    add_primed_variables(ConcreteVariables,InvEnv,BeforeAfterEnv),

    typecheck_together([RawInvariant,RawTheorems],InvEnv,[Invariant,Theorems],invariant),

    typecheck_variant(RawVariant,InvEnv,Variant),

    get_abstract_events(RefinedModel,AbstractEvents),

    check_events(RawEvents,Name,ModelEnv,BeforeAfterEnv,InvEnv,
                 DroppedAbstractVars,ConcreteVariables,AbstractEvents,
                 NewVariables,Variant,Events).

% Context is just an information term about the context (axioms, ...)
typecheck_together(Raws,Env,Typed,Context) :-
    maplist(same_length, Raws, Typed),
    append(Raws,RawL), append(Typed,TypedL),
    typecheck_l(RawL,Env,pred,TypedL,Context).

typecheck_variant(-, _Env, -) :- !.
typecheck_variant(RawVariant, Env, Variant) :-
    typecheck(RawVariant, Env, Type, Variant),
    ( Type==integer       -> true
    ; functor(Type,set,1) -> true
    ; otherwise ->
        add_error(bmachine_eventb,'Invalid type of variant ',Type),
        fail).

remove_linking_parts([],_AnimationChain,[]).
remove_linking_parts([Inv|Irest],AnimationChain,Invariants) :-
    ( abstraction_var_occures(Inv,AnimationChain) ->
        Invariants = RestInvariants
    ; otherwise ->
        Invariants = [Inv|RestInvariants]),
    remove_linking_parts(Irest,AnimationChain,RestInvariants).
    
abstraction_var_occures(TExpr,AnimationChain) :-
    get_texpr_expr(TExpr,identifier(_)),!,
    get_texpr_section(TExpr,Section),
    \+ member(Section,AnimationChain).
abstraction_var_occures(TExpr,AnimationChain) :-
    syntaxtraversion(TExpr,_,_,_,Subs,_),
    abstraction_var_occures_l(Subs,AnimationChain).
abstraction_var_occures_l([Sub|_],AnimationChain) :-
    abstraction_var_occures(Sub,AnimationChain),!.
abstraction_var_occures_l([_|Srest],AnimationChain) :-
    abstraction_var_occures_l(Srest,AnimationChain).

get_texpr_section(TExpr,Section) :-
    get_texpr_info(TExpr,Info),member(section(Section),Info),!.

extract_model_sections(RawModel,Name,Refines,Sees,Variables,Invariant,Theorems,Variant,Events) :-
    RawModel = event_b_model(_Pos,Name,Sections),
    optional_rawmachine_section(refines,Sections,-,Refines),
    optional_rawmachine_section(sees,Sections,[],Sees),
    optional_rawmachine_section(variables,Sections,[],Variables),
    optional_rawmachine_section(invariant,Sections,[],Invariant),
    optional_rawmachine_section(theorems,Sections,[],Theorems),
    optional_rawmachine_section(variant,Sections,-,Variant),
    optional_rawmachine_section(events,Sections,[],Events).

%get_abstraction_variables(-,[]).
%get_abstraction_variables(Model,Variables) :-
%    internal_model_vars(Model,_Name,Variables).

get_abstract_events(-,[]).
get_abstract_events(Model,Events) :- internal_model_events(Model,_Name,Events).

get_refined_model(-,_Models,-) :- !.
get_refined_model(Name,Models,Model) :-
    internal_model(M,Name),
    ( memberchk(M,Models) ->
        M=Model
    ; otherwise ->
        add_error(bmachine_eventb,'Unable to find refined model ',Name),
        fail).

split_abstract_vars(AbstractVars,ConcreteVars,NewVars,DroppedVars) :-
    split_abstract_vars2(AbstractVars,ConcreteVars,DroppedVars),
    find_new_variables(ConcreteVars,AbstractVars,NewVars).
split_abstract_vars2([],_,[]).
split_abstract_vars2([AbstractVar|AVrest],ConcreteVars,Dropped) :-
    ( find_identifier_in_list(AbstractVar,ConcreteVars,ConcreteVar) ->
        % Both variables have the same type
        get_texpr_type(AbstractVar,Type),
        get_texpr_type(ConcreteVar,Type),
        Dropped = DroppedRest
    ; otherwise ->
        Dropped = [AbstractVar|DroppedRest]),
    split_abstract_vars2(AVrest,ConcreteVars,DroppedRest).
find_new_variables([],_,[]).
find_new_variables([ConcreteVar|CVrest],AbstractVars,NewVars) :-
    ( find_identifier_in_list(ConcreteVar,AbstractVars,_AbstractVar) -> NewVars = Nrest
    ; otherwise -> NewVars = [ConcreteVar|Nrest]),
    find_new_variables(CVrest,AbstractVars,Nrest).

% find_identifier_in_list(TypedIdToFind,List,Elem):
%  search for a typed identifier Elem in List wich
%  has the same id as TypedIdToFind, but type and informations
%  may be different
find_identifier_in_list(TypedIdToFind,List,Found) :-
    def_get_texpr_id(TypedIdToFind,Id),
    get_texpr_id(Found,Id),
    member(Found,List).

get_all_abstract_variables(Refines,AbstractModels,AbstractVariables) :-
    get_all_abstract_variables2(Refines,AbstractModels,_EncounteredIds,AbstractVariables,[]).
get_all_abstract_variables2(-,_AbstractModels,EncounteredIds) --> 
    {!,empty_avl(EncounteredIds)}.
get_all_abstract_variables2(Name,AbstractModels,EncounteredIds) -->
    {get_refined_model(Name,AbstractModels,Model),
     internal_model_refines(Model,_,Refines)},
    get_all_abstract_variables2(Refines,AbstractModels,EncounteredIdsAbs),
    get_all_abstract_variables3(Model,EncounteredIdsAbs,EncounteredIds).
get_all_abstract_variables3(Model,EncounteredIdsIn,EncounteredIdsOut,
                            FoundVarsIn,FoundVarsOut) :-
    internal_model_vars(Model,_,Vars),
    extract_new_variables(Vars,EncounteredIdsIn,NewVars),
    get_texpr_ids(NewVars,NewIds),
    add_all_to_avl(NewIds,EncounteredIdsIn,EncounteredIdsOut),
    add_all(NewVars,FoundVarsIn,FoundVarsOut).
add_all_to_avl([],Avl,Avl).
add_all_to_avl([H|T],In,Out) :-
    avl_store(H,In,1,Inter),
    add_all_to_avl(T,Inter,Out).
add_all([],L,L).
add_all([H|T],[H|L1],L2) :- add_all(T,L1,L2).
extract_new_variables([],_EncounteredIdsIn,[]).
extract_new_variables([Var|Vrest],EncounteredIds,NewVars) :-
    get_texpr_id(Var,Id),
    ( avl_fetch(Id,EncounteredIds) ->
        NewVars = NewRest
    ; otherwise ->
        NewVars = [Var|NewRest]),
    extract_new_variables(Vrest,EncounteredIds,NewRest).

    
    
    

check_events([],_ModelName,_Env,_GEnv,_InvEnv,_Dropped,_ConcreteVars,_AbstractEvents,_NewVars,_Variant,[]).
check_events([RawEvent|RawRest],ModelName,Env,GEnv,InvEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,[Event|Erest]) :-
    (check_event(RawEvent,ModelName,Env,GEnv,InvEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Event)
      -> true
      ;  add_failed_call_error(check_event(RawEvent,ModelName,Env,GEnv,InvEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Event))),
    check_events(RawRest,ModelName,Env,GEnv,InvEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Erest).


check_event(RawEvent,ModelName,Env,GEnv,InvEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVariables,Variant,
            event(EvName,ModelName,Status,TParameters,TGuard,TTheorems,MActions,TVarWitnesses,TParamWitnesses,Unmodifiables,Refines)) :-
    
    raw_event(RawEvent,_,EvName,RawStatus,Refines,Params,Guards,Theorems,Actions,Witnesses),
    check_status(RawStatus, Variant, Status, Refines, AbstractEvents, eventlocation(ModelName,EvName)),
    get_abstract_parameters(Refines,AbstractEvents,AbstractParameters),
    create_identifiers(Params,[loc(ModelName,event(EvName))],TParameters),
    split_abstract_vars(AbstractParameters,TParameters,
                        _NewParameters,DroppedParameters1),
    annotate_dropped_parameters(DroppedParameters1,Refines,AbstractEvents,DroppedParameters),
    conjunct_guards(Guards,Guard),
    maplist(adapt_becomes_such,Actions,Actions1),
    % add new parameters to the typing environments:
    add_identifiers_to_environment(TParameters,Env,Subenv),
    add_identifiers_to_environment(TParameters,GEnv,SubGenv),
    add_identifiers_to_environment(TParameters,InvEnv,SubTheoremsEnv), % theorems can access dropped abstract variables; just like the invariant
    typecheck(Guard,Subenv,pred,TGuard),
    typecheck_l(Theorems,SubTheoremsEnv,pred,TTheorems,theorems),
    typecheck_l(Actions1,Subenv,subst,TActions,actions),
    check_witnesses(Witnesses,SubGenv,DroppedParameters,DroppedVars,TVarWitnesses,TParamWitnesses),
    % add checks that some variables may not be modified, e.g
    % if skip is refined
    get_abstract_actions(Refines,AbstractEvents,AbstractActions),
    ~~mnf( add_modified_vars_to_actions(TActions,MActions1) ),
    add_unmodified_assignments(AbstractActions,MActions1,ConcreteVars,MActions),
    add_equality_check(MActions,AbstractActions,NewVariables,ConcreteVars,Unmodifiables).

% add assignments of the form x:=x if a not dropped abstract variable
% is modified by the abstract event, but not the dropped event
add_unmodified_assignments(AbstractActions,ConcreteActions,ConcreteVars,NewActions) :-
    get_modified_vars_of_actions(AbstractActions,AbstractModified),
    get_modified_vars_of_actions(ConcreteActions,ConcreteModified),
    remove_all(AbstractModified,ConcreteModified,OnlyAbstractModified),
    findall( Var,
             ( member(Id,OnlyAbstractModified), get_texpr_id(Var,Id), memberchk(Var,ConcreteVars)),
             ToKeep),
    ( ToKeep = []    -> NewActions = ConcreteActions
    ; ToKeep = [_|_] ->
        create_texpr(assign(ToKeep,ToKeep),subst,[],Action1),
        add_modified_vars_to_action(Action1,Action),
        NewActions = [Action|ConcreteActions]).

check_status(RawStatus, Variant, TConv, Refines, AbstractEvents, Location) :-
    remove_pos(RawStatus,Status),
    check_status2(Status, Variant, Refines, AbstractEvents, Location, Conv, Infos),
    create_texpr(Conv, status, Infos, TConv).
check_status2(ordinary, _Variant, _Refines, _AbstractEvents, _Location, ordinary, []).
check_status2(convergent, _Variant, Refines, AbstractEvents, _Location, ordinary, [convergence_proved]) :-
     % if the refined event is also convergent, we can (or must because the
     % variant may be missing in the refinement) omit the check for decreasing the
     % variant. We do this by treating the event as ordinary.
     convergence_ensured_by_abstract_event(Refines, AbstractEvents),!.
check_status2(convergent, Variant, _Refines, _AbstractEvents, Location, convergent(RVariant), [convergence_proved]) :-
     variant_must_be_set(Variant,Location,RVariant).
check_status2(anticipated, Variant, _Refines, _AbstractEvents, _Location, anticipated(RVariant), []) :- 
     get_variant(Variant,RVariant).

:- use_module(tools_strings,[ajoin/2]).
variant_must_be_set(-,Location,RVariant) :- !,
    Location = eventlocation(Model,Name),
    ajoin(['No variant defined for ', Name, ' in model ', Model],Msg),
    add_error(bmachine_eventb,Msg),
    get_variant(-,RVariant).
    % removed fail to allow animation
variant_must_be_set(V,_,V).

% get_variant(ModelVariant,Variant) checks profides a default variant (0) if
% the model does not have a variant
get_variant(-,V) :- !, create_texpr(integer(0),integer,[default_variant],V).
get_variant(V,V).

convergence_ensured_by_abstract_event(Refines, AbstractEvents) :-
    % find a refined event which status is convergent
    internal_event_status(Event,Ref,Status),
    get_texpr_info(Status,StatusInfo),
    member(Event, AbstractEvents),
    member(Ref, Refines),
    member(convergence_proved,StatusInfo),!.

annotate_dropped_parameters([],_,_,[]).
annotate_dropped_parameters([P|PRest],Refines,AbstractEvents,[A|ARest]) :-
    def_get_texpr_id(P,Id),
    add_texpr_infos(P,[droppedby(DropList)],A),
    findall(Event, event_drops_param(Event,Refines,AbstractEvents,Id), DropList),
    annotate_dropped_parameters(PRest,Refines,AbstractEvents,ARest).
event_drops_param(Event,Refines,AbstractEvents,Id) :-
    get_texpr_id(Pattern,Id),
    member(Event,Refines),
    get_event_parameters(Event,AbstractEvents,AbParams),
    member(Pattern,AbParams).

get_abstract_parameters([],_,R) :- !,R=[].
get_abstract_parameters([Refines|Rest],Events,Params) :- !,
    get_event_parameters(Refines,Events,LocalParams),
    append(LocalParams,RestParams,Params),
    get_abstract_parameters(Rest,Events,RestParams).
get_abstract_parameters(Refines,_,Params) :- add_error(get_abstract_parameters,'Illegal Refines List: ',Refines),
    Params=[].
get_event_parameters(Refines,Events,Params) :-
    internal_event_params(Event,Refines,Params),
    member(Event,Events),!.

get_abstract_actions([],_,[]).
get_abstract_actions([Refines|Rest],Events,Actions) :-
    get_event_action(Refines,Events,LocalActions),
    append(LocalActions,RestActions,Actions),
    get_abstract_actions(Rest,Events,RestActions).
get_event_action(Refines,Events,Actions) :-
    internal_event_actions(Event,Refines,Actions),
    member(Event,Events),!.

check_witnesses([],_Env,_DroppedParams,_DroppedVars,[],[]).
check_witnesses([Witness|Rest],Env,DroppedParams,DroppedVars,TVarWitnesses,TParamWitnesses) :-
    check_witness(Witness,Env,DroppedParams,DroppedVars,TWitness,WType),
    ( WType = droppedvar ->
        TVarWitnesses = [TWitness|TWVRest],
        TParamWitnesses = TWPRest
    ; WType = droppedparam(_DropList) ->
        TVarWitnesses = TWVRest,
        TParamWitnesses = [TWitness|TWPRest]),
    check_witnesses(Rest,Env,DroppedParams,DroppedVars,TWVRest,TWPRest).
check_witness(witness(_Pos,RawId,Pred),Env,DroppedParams,DroppedVars,TWitness,WType) :-
    RawId = identifier(_,WitnessId),
    get_witness_identifier(WitnessId,DroppedParams,DroppedVars,TId,Info,WType),
    get_texpr_type(TId,Type),
    env_store(WitnessId,Type,Info,Env,Subenv),
    typecheck(Pred,Subenv,pred,TPred),
    create_texpr(witness(TId,TPred),witness,[],TWitness).
get_witness_identifier(PrimedId,_DroppedParams,DroppedVars,TPrimeId,[],droppedvar) :-
    prime_id(Id,PrimedId),
    get_texpr_id(TId,Id),
    member(TId,DroppedVars),!,
    prime_variables([TId],[TPrimeId]).
get_witness_identifier(Id,DroppedParams,_DroppedVars,TId,Info,droppedparam(DropList)) :-
    get_texpr_id(TId,Id),
    member(TId,DroppedParams),!,
    get_texpr_info(TId,Info),
    member(droppedby(DropList),Info),!.

conjunct_guards([],truth(none)).
conjunct_guards([H|T],Result) :- conjunct_guards2(T,H,Result).
conjunct_guards2([],G,G).
conjunct_guards2([H|T],G,Result) :- conjunct_guards2(T,conjunct(none,G,H),Result).

/*
parallel_actions(Actions,Pos,Action) :-
    ( Actions = [] -> Action = skip(Pos)
    ; Actions = [S] -> Action = S
    ; otherwise -> Action = parallel(Pos,Actions)).
    */

adapt_becomes_such(becomes_such(Pos,Ids,Pred),evb2_becomes_such(Pos,Ids,Pred)) :- !.
adapt_becomes_such(Subst,Subst).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convert B project to classical B machine

:- use_module(bmachine_structure,[create_machine/2]).

convert_event_b(Main,Freetypes,Ops,Contextes,Models,AnimationChain,Machine) :-
    debug_println(9,'Converting to classical B'),
    select_models(AnimationChain,Models,AnimatedModels),
    % todo: access to private elements of bmachine_construction
    bmachine_structure:create_machine(Main,Empty),
    convert_event_b2(Freetypes,Ops,AnimatedModels,Contextes,AnimationChain,
                     Empty,Machine),!.
convert_event_b(Main,Freetypes,Ops,Contextes,Models,AnimationChain,Machine) :-
    add_failed_call_error(convert_event_b(Main,Freetypes,Ops,Contextes,Models,AnimationChain,Machine)),
    fail.
convert_event_b2(Freetypes,Ops,AnimatedModels,Contextes,AnimationChain) -->
    append_to_section(freetypes,Freetypes),
    append_to_section(operators,Ops),
    create_constants(AnimatedModels,Contextes,AnimatedContextNames),
    {append(AnimatedContextNames,AnimationChain,AnimatedSectionNames)},
    create_variables(AnimatedModels,AnimatedSectionNames),
    create_events(AnimatedModels,AnimatedSectionNames),
    create_animated_section_info(AnimatedContextNames,AnimationChain).

create_animated_section_info(Contextes,Models) -->
    {reverse(Models,FromAbstractToConcrete),
     maplist(mark_as_model,FromAbstractToConcrete,MarkedModels),
     maplist(mark_as_context,Contextes,MarkedContextes),
     append(MarkedContextes,MarkedModels,Sections)},
    append_to_section(meta,[animated_sections(Sections)]).

mark_as_model(Name,model(Name)).
mark_as_context(Name,context(Name)).

/*
get_operation_identifiers([],[]).
get_operation_identifiers([TExpr|Rest],[Id|IdRest]) :-
    get_texpr_expr(TExpr,operation(Id,_,_,_)),
    get_operation_identifiers(Rest,IdRest).
*/

create_variables(Models,AnimatedSectionNames) -->
    {collect_model_infos(Models,AnimatedSectionNames,Variables,Invariants,Theorems)},
    append_to_section(abstract_variables,Variables),
    conjunct_to_section(invariant,Invariants),
    append_to_section(assertions,Theorems).

select_contexts([],_Contexts,[]).
select_contexts([Name|NRest],Contexts,[Context|CRest]) :-
    internal_context(Context,Name),
    memberchk(Context,Contexts),
    select_contexts(NRest,Contexts,CRest).

select_models([],_Models,[]).
select_models([Name|NRest],Models,[Model|MRest]) :-
    internal_model(Model,Name), memberchk(Model,Models),
    select_models(NRest,Models,MRest).

% collect_model_infos(Models,Variables,Invariants,Theorems):
%  collect all variables, invariants and theorems of the given models
collect_model_infos([],_,[],[],[]) :- !.
collect_model_infos(Models,AnimatedSectionNames,Vars,Invs,Thms) :-
    collect_model_infos2(Models,AnimatedSectionNames,AllVars,Vars1,Invs,Thms),
    % add to every variable the information in which models
    % it is present
    add_var_model_occurrences(Vars1,AllVars,Vars).
collect_model_infos2([],_,[],[],[],[]).
collect_model_infos2([Model|MRest],AnimatedSectionNames,Vars,NewVars,Invs,Thms) :-
    internal_model_vars(Model,_,LVars), append(LVars,RVars,Vars),
    % use only newly introduced variables of a model, to prevend
    % double declarations
    % in case of a limited animation chain, it is important to introduce
    % all variables for the most abstract model
    ( MRest == [] -> LNewVars = LVars
    ; otherwise   -> internal_model_newvars(Model,_,LNewVars)),
    append(LNewVars,RNewVars,NewVars),
    internal_model_invs(Model,_,LInvs1), append_stripped_predicates(LInvs1,AnimatedSectionNames,RInvs,Invs),
    internal_model_thms(Model,_,LThms1), append_stripped_predicates(LThms1,AnimatedSectionNames,RThms,Thms),
    collect_model_infos2(MRest,AnimatedSectionNames,RVars,RNewVars,RInvs,RThms).
add_var_model_occurrences([],_,[]).
add_var_model_occurrences([Var|Vrest],AllVars,[MVar|Mrest]) :-
    add_var_model_occurrence(Var,AllVars,MVar),
    add_var_model_occurrences(Vrest,AllVars,Mrest).
add_var_model_occurrence(Var,AllVars,MVar) :-
    findall(Section,
            ( find_identifier_in_list(Var,AllVars,M),
                get_texpr_info(M,Infos),
                member(section(Section),Infos)),
            Sections),
    reverse(Sections,FromAbstractToConcrete),
    add_texpr_infos(Var,[occurrences(FromAbstractToConcrete)],MVar).
append_stripped_predicates(PredsIn,AnimatedSectionNames,RestPreds,Preds) :-
    remove_linking_parts(PredsIn,AnimatedSectionNames,LPreds),
    append(LPreds,RestPreds,Preds).

all_used_contextes(AnimatedModels,Contextes,Sees) :-
    ( AnimatedModels = [] ->
        all_context_names(Contextes,Sees)
    ; otherwise ->
        collect_sees(AnimatedModels,Sees)).
all_context_names([],[]).
all_context_names([Context|Crest],[Name|Nrest]) :-
    internal_context(Context,Name),
    all_context_names(Crest,Nrest).
collect_sees([],[]).
collect_sees([Model|MRest],Sees) :-
    internal_model_sees(Model,_,LocalSees), append(LocalSees,RestSees,Sees),
    collect_sees(MRest,RestSees).

create_constants(AnimatedModels,Contextes,AnimatedContextNames) -->
    {all_used_contextes(AnimatedModels,Contextes,Sees),
     get_constants(Sees,Contextes,DefSets,EnumSets,EnumElems,Constants,AbstractConstants,Axioms,Theorems,AnimatedContextNames)},
    append_to_section(deferred_sets,DefSets),
    append_to_section(enumerated_sets,EnumSets),
    append_to_section(enumerated_elements,EnumElems),
    append_to_section(concrete_constants,Constants),
    append_to_section(abstract_constants,AbstractConstants),
    conjunct_to_section(properties,Axioms),
    append_to_section(assertions,Theorems).

get_constants(Sees,Contextes,DefSets,EnumSets,EnumElems,Constants,AbstractConstants,Axioms,Theorems,AllContextNames) :-
    transitive_contextes(Sees,Contextes,AllContextNames),
    select_contexts(AllContextNames,Contextes,AllContextes),
    collect_context_infos(AllContextes,Sets,Constants1,AbstractConstants,Axioms1,Theorems),
    enumerated_deferred_sets(Sets,Constants1,Axioms1,DefSets,Constants,Axioms,
                             EnumSets,EnumElems).

collect_context_infos([],[],[],[],[],[]).
collect_context_infos([context(Name,_Ex,LSets1,LCsts1,LAbsCsts1,LAxms,LThms)|Rest],Sets,Constants,AbstractConstants,Axioms,Theorems) :-
    add_context_occurrences(LSets1,Name,LSets),
    add_context_occurrences(LCsts1,Name,LCsts),
    add_context_occurrences(LAbsCsts1,Name,LAbsCsts),
    append(LSets,RSets,Sets), 
    append(LCsts,RCsts,Constants),
    append(LAbsCsts,RAbsCsts,AbstractConstants),
    append(LAxms,RAxms,Axioms),
    append(LThms,RThms,Theorems),
    collect_context_infos(Rest,RSets,RCsts,RAbsCsts,RAxms,RThms).
add_context_occurrences([],_,[]).
add_context_occurrences([Constant|Crest],Name,[CC|CCrest]) :-
    add_texpr_infos(Constant,[occurrences([Name])],CC),
    add_context_occurrences(Crest,Name,CCrest).


% merge the events and write them to a machine
create_events(AnimatedModels,AnimatedSectionNames) -->
    {get_events(AnimatedModels,AnimatedSectionNames,Events,Initialisation,Promoted)},
    append_to_section(operation_bodies,Events),
    bmachine_construction:write_section(initialisation,Initialisation),
    append_to_section(promoted,Promoted).

% merge the events of multiple refinement levels
get_events([],_,[],Initialisation,[]) :- !,
    create_texpr(skip,subst,[],Initialisation).
get_events(AnimatedModels,AnimatedSectionNames,Events,Initialisation,Promoted) :-
    get_events_to_animate(AnimatedModels,AnimEvents),
    merge_refinement_levels(AnimEvents,AnimatedModels,AnimatedSectionNames,MergedEvents,Promoted1),
    remove_initialisation(Promoted1,Promoted2),
    add_op_parenthesis(Promoted2,Promoted),
    select_initialisation(MergedEvents,PreOps,Initialisation),
    create_operations(PreOps,Events).
remove_initialisation(In,Out) :-
    get_texpr_id(Expr,'INITIALISATION'),
    select(Expr,In,Out),!.
remove_initialisation(In,Out) :-
    % add_error(check_event_b_project,'Model has no INITIALISATION'), %err msg already generated below
    Out=In.
add_op_parenthesis([],[]).
add_op_parenthesis([Old|ORest],[New|NRest]) :-
    def_get_texpr_id(Old,Id),
    rename_bt(Old,[rename(Id,op(Id))],New),
    add_op_parenthesis(ORest,NRest).

select_initialisation(In,Ops,Init) :-
    select(eventop('INITIALISATION',Section,[],Init1,op([],[]),_),In,Ops),!,
    add_texpr_infos(Init1,[eventb(Section)],Init).
select_initialisation(In,Ops,Init) :-
    add_error(select_initialisation,'Model has no INITIALISATION'),
    Init = b(skip,subst,[]), % replace skip by non-det assign
    Ops=In.
create_operations([],[]).
create_operations([E|ERest],[O|ORest]) :-
    create_operation(E,O),
    create_operations(ERest,ORest).
create_operation(eventop(Name,Section,Parameters,Subst,Type,Refines),Operation) :-
    create_texpr(identifier(op(Name)),Type,[],Id),
    get_texpr_info(Subst,SubstInfo),
    Mod=modifies(_),memberchk(Mod,SubstInfo),
    Reads=reads(_),memberchk(Reads,SubstInfo),
    ( Refines == [] -> RefInfo = []
    ; otherwise -> RefInfo = [refines(Refines)]),
    create_texpr(operation(Id,[],Parameters,Subst),Type,[eventb(Section),Mod,Reads|RefInfo],Operation).

get_events_to_animate([Model|_],Events) :- internal_model_events(Model,_,Events).

merge_refinement_levels([],_,_,[],[]).
merge_refinement_levels([Event|ERest],AnimModels,AnimatedSectionNames,[Merged|MRest],[Prom|PRest]) :-
    merge_refinement_levels2(Event,AnimModels,AnimatedSectionNames,Merged,Prom),
    merge_refinement_levels(ERest,AnimModels,AnimatedSectionNames,MRest,PRest).

merge_refinement_levels2(Event,AnimatedLevels,AnimatedSectionNames,
                         eventop(Name,Section,Parameters,Subst,OpType,Refines),Prom) :-
    AnimatedLevels = [CurrentLevel|_Abstractions],
    internal_model(CurrentLevel,Section),
    internal_event_params(Event,Name,Parameters),
    internal_event_refined(Event,Name,Refines),
    %merge_parameters(Refines,Parameters,_Abstractions,MergedParameters),
    merge_events_to_subst(Name,AnimatedLevels,AnimatedSectionNames,[],Subst1),
    optimise_events(Subst1,Subst),
    get_texpr_types(Parameters,Types),OpType = op(Types,[]),
    create_texpr(identifier(Name),OpType,[],Prom).
merge_events_to_subst(Name,[CurrentLevel|Abstractions],AnimatedSectionNames,RefinedParameters,TEvent) :-
    create_texpr(RefLevelEvent,subst,[modifies(Vars),reads(Reads)],TEvent),
    RefLevelEvent = rlevent(Name,Section,Status,Params,Guard,Theorems,Actions,VWitnesses,PWitnesses,Unmodifiables,AbstractEvents),
    InternalEvent = event(Name,Section,Status,AllParams,Guard,Theorems,Actions,VWitnesses1,PWitnesses1,Unmodifiables,Refines),
    get_event_from_model(Name,CurrentLevel,InternalEvent),
    strip_witnesses(VWitnesses1,AnimatedSectionNames,VWitnesses),
    strip_witnesses(PWitnesses1,AnimatedSectionNames,PWitnesses),
    remove_refined_parameters(AllParams,RefinedParameters,Params),
    ( Abstractions = [] ->
        AbstractEvents = []
    ; otherwise ->
        append(RefinedParameters,Params,NewRefinedParameters),
        merge_events_to_subst_l(Refines,Abstractions,AnimatedSectionNames,NewRefinedParameters,AbstractEvents1),
        normalise_merged_events(AbstractEvents1,AbstractEvents)),
    collect_modified_vars(Actions,AbstractEvents,Vars),
    collect_read_vars(AllParams,Guard,Theorems,Actions,VWitnesses,PWitnesses,AbstractEvents,Reads).
merge_events_to_subst_l([],_,_,_,[]).
merge_events_to_subst_l([Name|NRest],Abstractions,AnimatedSectionNames,RefinedParameters,[TEvent|TRest]) :-
    merge_events_to_subst(Name,Abstractions,AnimatedSectionNames,RefinedParameters,TEvent),
    merge_events_to_subst_l(NRest,Abstractions,AnimatedSectionNames,RefinedParameters,TRest).

% removes a witness completely if the witnessed variable belongs
% to a refinement level that is not animated or removes part of the
% witness-predicate if references to not animated levels are made.
strip_witnesses([],_AnimatedSectionNames,[]).
strip_witnesses([Witness|WRest],AnimatedSectionNames,Result) :-
    strip_witness(Witness,AnimatedSectionNames,StrippedWitness),
    append(StrippedWitness,SWRest,Result),
    strip_witnesses(WRest,AnimatedSectionNames,SWRest).
strip_witness(Witness,AnimatedSectionNames,StrippedWitness) :-
    get_texpr_expr(Witness,witness(W,P)),
    ( abstraction_var_occures(W,AnimatedSectionNames) ->
        % The abstract variable is not animated, skip the complete witness
        StrippedWitness = []
    ; otherwise ->
        conjunction_to_list(P,Preds),
        remove_linking_parts(Preds,AnimatedSectionNames,SPreds),
        conjunct_predicates(SPreds,SP),
        get_texpr_type(Witness,Type),
        get_texpr_info(Witness,Info),
        create_texpr(witness(W,SP),Type,Info,TWitness),
        StrippedWitness = [TWitness]).

remove_refined_parameters([],_RefinedParameters,[]).
remove_refined_parameters([Param|APRest],RefinedParameters,Parameters) :-
    ( find_identifier_in_list(Param,RefinedParameters,_RefParam) ->
        Parameters = PRest
    ; otherwise ->
        Parameters = [Param|PRest]),
    remove_refined_parameters(APRest,RefinedParameters,PRest).

/* currently commented out :
merge_parameters([],Parameters,_Abstractions,Parameters).
merge_parameters([_|_],Parameters,[],Parameters) :- !.
merge_parameters([Refines|Rest],OldParameters,Abstractions,Parameters) :-
    Abstractions = [Abstract|FurtherAbstractions],
    get_event_from_model(Refines,Abstract,Event),
    internal_event_params(Event,_,NewParameters),
    internal_event_refined(Event,_,NewRefines),
    add_unique_parameters(NewParameters,OldParameters,Parameters2),
    merge_parameters(NewRefines,Parameters2,FurtherAbstractions,Parameters3),
    merge_parameters(Rest,Parameters3,Abstractions,Parameters).
add_unique_parameters([],Parameters,Parameters).
add_unique_parameters([P|PRest],InParameters,OutParameters) :-
    ( find_identifier_in_list(P,InParameters,_) ->
        !,Parameters = InParameters
    ; otherwise ->
        Parameters = [P|InParameters]),
    add_unique_parameters(PRest,Parameters,OutParameters).
*/

get_event_from_model(Name,Model,Event) :-
    internal_model_events(Model,_,Events),
    internal_event(Event,Name),
    member(Event,Events),!.
    
collect_modified_vars(Actions,Events,Vars) :-
    ~~mnf( get_modified_vars_of_actions(Actions,M1) ),
    append_modified_vars_of_events(Events,M2),
    append(M1,M2,M3),sort(M3,Vars).
get_modified_vars_of_actions([],[]).
get_modified_vars_of_actions([Action|Arest],Vars) :-
    get_modified_vars_of_action(Action,Vlocal),
    append(Vlocal,Vrest,Vars),
    get_modified_vars_of_actions(Arest,Vrest).
get_modified_vars_of_action(Action,Vars) :-
    get_texpr_info(Action,Infos),
    member(modifies(Vars), Infos),!.

add_modified_vars_to_actions([],[]).
%add_modified_vars_to_actions([TAction|ARest],MRest) :- 
%    get_texpr_expr(TAction,skip),!, % remove skip actions % for Daniel: is this better than keeping them ??
%    add_modified_vars_to_actions(ARest,MRest).
add_modified_vars_to_actions([TAction|ARest],[MAction|MRest]) :-
    (add_modified_vars_to_action(TAction,MAction) ->true
      ; add_internal_error('Call failed: ',add_modified_vars_to_action(TAction,MAction)),
        MAction = TAction),
    add_modified_vars_to_actions(ARest,MRest).
add_modified_vars_to_action(TAction,MAction) :-
    get_lhs_ids(TAction,Vars),
    maplist(prime_id,Vars,Primed),
    add_texpr_infos(TAction,[assigned_after(Primed),modifies(Vars)],MAction).

get_lhs_ids(TAction,Ids) :-
    get_lhs_tids(TAction,Vars),
    get_texpr_ids(Vars,Ids).
get_lhs_tids(TAction,Vars) :-
    get_texpr_expr(TAction,Action),
    split_lhs_rhs(Action,Lhs,_),
    extract_lhs_ids(Lhs,Vars).

split_lhs_rhs(skip,[],[]). % not needed as skip removed above
split_lhs_rhs(assign(Ids,Exprs),Ids,Exprs).
split_lhs_rhs(assign_single_id(Id,Expr),[Id],[Expr]).
split_lhs_rhs(becomes_element_of(Ids,Expr),Ids,[Expr]).
split_lhs_rhs(becomes_such(Ids,Pred),Ids,[Pred]).
extract_lhs_ids([],[]).
extract_lhs_ids([TId|TRest],[Var|VRest]) :-
    extract_lhs_id(TId,Var),
    extract_lhs_ids(TRest,VRest).
extract_lhs_id(Id,Var) :-
    get_texpr_id(Id,_),!,Var=Id.
extract_lhs_id(Func,Var) :-
    get_texpr_expr(Func,function(Id,_)),
    get_texpr_id(Id,Var).

append_modified_vars_of_events(Events,Vars) :-
    append_modified_vars_of_events2(Events,Unsorted),
    sort(Unsorted,Vars).
append_modified_vars_of_events2([],[]).
append_modified_vars_of_events2([TEvent|ERest],Vars) :-
    get_modified_vars_of_event(TEvent,Local),
    append(Local,VRest,Vars),
    append_modified_vars_of_events2(ERest,VRest).
get_modified_vars_of_event(TEvent,Local) :-
    get_texpr_info(TEvent,Info),
    memberchk(modifies(Local),Info).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get all the variables that are accessed in an event
collect_read_vars(AllParams,Guard,Theorems,Actions,VWitnesses,PWitnesses,AbstractEvents,Reads) :-
    get_texpr_ids(AllParams,Params), % access to parameters is not reading a variable
    find_identifier_uses(Guard,Params,GrdVars),
    find_identifier_uses_l(Theorems,Params,ThmVars),
    maplist(find_all_action_reads(Params),Actions,ActVars0),      ord_union(ActVars0, ActVars),
    maplist(find_all_witness_reads(Params),VWitnesses,VWitVars0), ord_union(VWitVars0,VWitVars),
    maplist(find_all_witness_reads(Params),PWitnesses,PWitVars0), ord_union(PWitVars0,PWitVars),
    join_abstract_reads(AbstractEvents,AbstrVars),
    ord_union([GrdVars,ThmVars,ActVars,VWitVars,PWitVars,AbstrVars],Reads).
%find_all_identifiers(Ignore,Expr,Vars) :-
%    find_identifier_uses(Expr,Ignore,Vars). % Just switch arguments to allow the use of maplist
find_all_action_reads(Ignore,TAction,Vars) :-
    get_texpr_expr(TAction,Action),
    find_action_reads2(Action,Ignore,Vars).
find_action_reads2(skip,_,Vars) :- Vars=[].
find_action_reads2(assign(_,Exprs),Ignore,Vars) :- find_identifier_uses_l(Exprs,Ignore,Vars).
find_action_reads2(assign_single_id(_,Expr),Ignore,Vars) :- find_identifier_uses(Expr,Ignore,Vars).
find_action_reads2(becomes_element_of(_,Expr),Ignore,Vars) :- find_identifier_uses(Expr,Ignore,Vars).
find_action_reads2(becomes_such(Ids,Pred),Ignore,Vars) :-
    prime_variables(Ids,PIds),get_texpr_ids(PIds,Primes),
    append(Primes,Ignore,Ignore0),sort(Ignore0,Ignore1),
    find_identifier_uses(Pred,Ignore1,Vars).
find_all_witness_reads(Ignore,Witness,Vars) :-
    get_texpr_expr(Witness,witness(I,P)),
    get_texpr_id(I,Id),
    sort([Id|Ignore],Ignore2),
    find_identifier_uses(P,Ignore2,Vars1),
    % a small hack to remove the primed variables that refer to the after-state
    % without the need to put all variables in the Ingore list:
    exclude(is_primed_id,Vars1,Vars).
is_primed_id(Id) :- atom_codes(Id,Codes),[Prime]="'",last(Codes,Prime).
join_abstract_reads([],[]) :- !.
join_abstract_reads(AbstractEvents,Reads) :-
    maplist(get_read_from_info,AbstractEvents,AbsReads),
    ord_union(AbsReads,ReadBySome),
    maplist(get_modifies_from_info,AbstractEvents,AbsModifies),
    ord_union(AbsModifies,ModifiedBySome),
    ord_intersection(AbsModifies,ModifiedByAll),
    ord_subtract(ModifiedBySome,ModifiedByAll,ModifiedByNotAll),
    ord_union(ReadBySome,ModifiedByNotAll,Reads).
get_read_from_info(AbstrEvents,Reads) :-
    get_texpr_info(AbstrEvents,Info),memberchk(reads(Reads),Info).
get_modifies_from_info(AbstrEvents,Modifies) :-
    get_texpr_info(AbstrEvents,Info),memberchk(modifies(Modifies),Info).

add_equality_check(Actions,AbstractActions,NewVars,Variables,Unmodifiables) :-
    get_modified_vars_of_actions(Actions,Modified),
    get_modified_vars_of_actions(AbstractActions,AbstractModified),
    unmodifiables(Modified,AbstractModified,NewVars,Variables,Unmodifiables).
unmodifiables([],_AbstractModified,_NewVars,_Variables,[]).
unmodifiables([UM|Modified],AbstractModified,NewVars,Variables,Unmodifiables) :-
    ( is_modifiable(UM,AbstractModified,NewVars) ->
        Unmodifiables = UMrest  
    ; otherwise ->
        get_texpr_id(TUM,UM),
        member(TUM,Variables),!,
        Unmodifiables = [TUM|UMrest]),
    unmodifiables(Modified,AbstractModified,NewVars,Variables,UMrest).
% the abstract event modifies it, so we can do it, too
is_modifiable(UM,AbstractModified,_NewVars) :- member(UM,AbstractModified),!.
% the variables is newly introduced, so we can modifiy it
is_modifiable(UM,_AbstractModified,NewVars) :-
    get_texpr_id(TUM,UM), member(TUM,NewVars).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% normalising merged events means that if two refined events A and B
% change a different set of variables then the event A gets an additional
% assignment of the form x:=x with x as a variable that is changed in B
% but not in A.
% This can happen even though merged events have the same assignments because
% they again can refine events that have different assignments.

% None or one refined event do not have to be normalised
normalise_merged_events([],[]) :- !.
normalise_merged_events([E],[E]) :- !.
normalise_merged_events(Events,Normalised) :-
    append_modified_vars_of_events(Events,AllModifiedVars),
    find_typed_version_in_events(AllModifiedVars,Events,TypedVars),
    normalise_merged_events2(Events,AllModifiedVars,TypedVars,Normalised).
normalise_merged_events2([],_AllModifiedVars,_TypedVars,[]).
normalise_merged_events2([Event|Erest],AllModifiedVars,TypedVars,[Normalised|Nrest]) :-
    normalise_merged_event(Event,AllModifiedVars,TypedVars,Normalised),
    normalise_merged_events2(Erest,AllModifiedVars,TypedVars,Nrest).
normalise_merged_event(Event,AllModifiedVars,TypedVars,Normalised) :-
    get_modified_vars_of_event(Event,Modified),
    remove_all(AllModifiedVars,Modified,MissingVars),
    add_missing_assignments(MissingVars,Event,AllModifiedVars,TypedVars,Normalised).
add_missing_assignments([],Event,_,_,Event) :- !.
add_missing_assignments(MissingVars,TEvent,AllModifiedVars,TypedVars,Normalised) :-
    get_texpr_expr(TEvent,Event),
    get_texpr_type(TEvent,Type),
    get_texpr_info(TEvent,Infos),
    selectchk(modifies(_),Infos,RestInfos), % TO DO: also add reads Information
    create_texpr(NewEvent,Type,[modifies(AllModifiedVars)|RestInfos],Normalised),
    Event = rlevent(Name,Sec,St,Prm,Grd,Thm,Actions,VWit,PWit,Unmod,AbsEvents),
    NewEvent = rlevent(Name,Sec,St,Prm,Grd,Thm,NewActions,VWit,PWit,Unmod,AbsEvents),
    append(Actions,NewAssignments,NewActions),
    create_missing_assignments(MissingVars,TypedVars,NewAssignments).
create_missing_assignments([],_TypedVars,[]).
create_missing_assignments([MissingVar|Mrest],TypedVars,[Assign|Arest]) :-
    create_missing_assignment(MissingVar,TypedVars,Assign),
    create_missing_assignments(Mrest,TypedVars,Arest).
create_missing_assignment(Var,TypedVars,Assignment) :-
    get_texpr_id(TId,Var),
    memberchk(TId,TypedVars),
    prime_id(Var,Primed),
    create_texpr(assign([TId],[TId]),subst,[assigned_after(Primed),modifies(Var)],Assignment).

find_typed_version_in_events([],_Events,[]).
find_typed_version_in_events([Name|Nrest],Events,[Typed|Trest]) :-
    find_typed_version_in_events2(Name,Events,Typed),
    find_typed_version_in_events(Nrest,Events,Trest).
find_typed_version_in_events2(Name,Events,Typed) :-
    get_texpr_id(Typed,Name),
    member(Event,Events),
    get_texpr_expr(Event,rlevent(_Name,_Sec,_St,_Prm,_Grd,_Thm,Actions,_VWit,_PWit,_Unmod,AbsEvents)),
    find_typed_version_in_events3(Name,Actions,AbsEvents,Typed),!.
find_typed_version_in_events3(_Name,Actions,_AbsEvents,Typed) :-
    member(Action,Actions),
    get_lhs_tids(Action,Vars),
    member(Typed,Vars),!.
find_typed_version_in_events3(Name,_Actions,AbsEvents,Typed) :-
    find_typed_version_in_events2(Name,AbsEvents,Typed).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% optimisation of guards:
% A guard in a refinement does not need to be evaluated in in an abstract event
% again. This happens quiet frequently, especially if "extends" is used.
% The same applies for assignments: If in a refinement an action is performed,
% we do not have to check if the same action can be performed in a more abstract
% level.
optimise_events(TEvent,NTEvent) :-
    empty_avl(Empty),
    optimise_events2(Empty,TEvent,NTEvent).
optimise_events2(KnownElements,TEvent,NTEvent) :-
    create_texpr(Event,Type,Infos,TEvent),
    create_texpr(NewEvent,Type,Infos,NTEvent),
    Event = rlevent(Name,Section,Status,Params,OldGuard,OldTheorems,OldActions,
                    VWitnesses,PWitnesses,Unmodifiables,OldAbstractEvents),
    NewEvent = rlevent(Name,Section,Status,Params,NewGuard,NewTheorems,NewActions,
                    VWitnesses,PWitnesses,Unmodifiables,NewAbstractEvents),
    optimise_guards(OldGuard,OldTheorems,KnownElements,NewGuard,NewTheorems,KnownElements1),
    optimise_formulas(OldActions,KnownElements1,NewActions,NewKnownElements),
    maplist(optimise_events2(NewKnownElements),OldAbstractEvents,NewAbstractEvents).
optimise_guards(OldGuard,OldTheorems,KnownElements,NewGuard,NewTheorems,NewKnownElements) :-
    conjunction_to_list(OldGuard,OldGuards),
    optimise_formulas(OldGuards,KnownElements,NewGuards,KnownElements1),
    % If no elements were removed, we just use the old guard
    ( OldGuards=NewGuards -> OldGuard=NewGuard
    ; otherwise           -> conjunct_predicates(NewGuards,NewGuard)
    ),
    optimise_formulas(OldTheorems,KnownElements1,NewTheorems,NewKnownElements).
% optimise_formulas(+AllFormulas,+KnownElements,-NewFormulas,-NewNormed)
%   AllFormulas:      A list of formulas
%   KnownElementes:   A list of normed versions of already encountered elements
%   NewFormulas:      A sub-list of AllFormulas which are new (not in KnownElements)
%   NewKnownElements: The updated list of already encountered elements
% If no formulas are removed, it is guaranteed that NewFormulas=AllFormulas
optimise_formulas(AllFormulas,KnownElements,NewFormulas,NewKnownElements) :-
    % convert formulas to a list: Formula/Normed where Normed is a normalised form
    maplist(add_stripped_ast,AllFormulas,AllFormNormed),
    % remove the already known formulas
    exclude(is_duplicate_formula(KnownElements),AllFormNormed,NewFormNormed),
    % split the list [Form/Normed] into [Form] and [Normed]
    maplist(unzipslash,NewFormNormed,NewFormulas1,NewNormed),
    ( AllFormNormed = NewFormNormed -> % No formulas removed, keep everything as it is
        NewFormulas = AllFormulas
    ; otherwise ->
        NewFormulas = NewFormulas1
    ), 
    foldl(add_to_known_elements,NewNormed,KnownElements,NewKnownElements).
add_stripped_ast(Formula,Formula/Stripped) :-
    strip_and_norm_ast(Formula,Stripped).
is_duplicate_formula(KnownElements,_Formula/Stripped) :-
    avl_fetch(Stripped,KnownElements).  % remark: if some ast_cleanup operations introduce fresh identifiers (e.g., comp_result) then we may not detect formulas as identical here !
unzipslash(A/B,A,B).
add_to_known_elements(New,In,Out) :-
    avl_store(New,In,true,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% internal event handling (just data structure)
internal_event(event(Name,_Section,_Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name).
internal_event_status(event(Name,_Section,Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name,Status).
internal_event_params(event(Name,_Section,_Status,Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name,Params).
internal_event_actions(event(Name,_Section,_Status,_Params,_Grd,_Thm,Act,_VWit,_PWit,_Unmod,_Ref),Name,Act).
internal_event_refined(event(Name,_Section,_Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,Ref),Name,Ref).

internal_context(context(Name,_Extends,_Sets,_Constants,_AbstractConstants,_Axioms,_Theorems),Name).
internal_context_extends(context(Name,Extends,_Sets,_Constants,_AbstractConstants,_Axioms,_Theorems),Name,Extends).
internal_context_sets(context(Name,_Extends,Sets,_Constants,_AbstractConstants,_Axioms,_Theorems),Name,Sets).
internal_context_constants(context(Name,_Extends,_Sets,Constants,_AbstractConstants,_Axioms,_Theorems),Name,Constants).
internal_context_abstract_constants(context(Name,_Extends,_Sets,_Constants,AbstractConstants,_Axioms,_Theorems),Name,AbstractConstants).

internal_model(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name).
internal_model_refines(model(Name,Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Refines).
internal_model_sees(model(Name,_Refines,Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Sees).
internal_model_vars(model(Name,_Refines,_Sees,Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Variables).
internal_model_newvars(model(Name,_Refines,_Sees,_Variables,NewVariables,_Invariants,_Theorems,_Events),Name,NewVariables).
internal_model_invs(model(Name,_Refines,_Sees,_Variables,_NewVariables,Invariants,_Theorems,_Events),Name,Invariants).
internal_model_thms(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,Theorems,_Events),Name,Theorems).
internal_model_events(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,Events),Name,Events).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% analyse deferred sets and generate enumerated sets if possible




%% :- include(bmachine_eventb_common).
%% code below used to be in separate file, as there was the single-level and multi-level mode:



extract_name(event_b_model(_,Name,_Sections),Res) :- !, Res=Name.
extract_name(event_b_context(_,Name,_Sections),Res) :- !, Res=Name.
extract_name(MODEL,_) :- add_error(extract_name,'Could not extract model name: ',MODEL),fail.

extract_sections(event_b_model(_,_Name,Sections),Sections).
extract_sections(event_b_context(_,_Name,Sections),Sections).

% topological sort

topological_sort(List,Sorted) :-
    get_dependencies(List,Dependencies),
    topolsort(Dependencies,[],SortedNames),
    get_sorted(SortedNames,List,Sorted),!.
topological_sort(List,Sorted) :- add_failed_call_error(topological_sort(List,Sorted)),fail.

get_dependencies([],[]).
get_dependencies([E|Rest],[dep(Name,Deps)|Drest]) :-
    extract_name(E,Name),
    findall(D,dependency(E,D),Deps1),
    sort(Deps1,Deps),
    get_dependencies(Rest,Drest).

dependency(Elem,Dependend) :-
    extract_sections(Elem,Sections),
    ( rawmachine_section(extends,Sections,Extends),
        member(Dependend,Extends)
    ; rawmachine_section(refines,Sections,Dependend)).

topolsort([],_,[]) :- !.
topolsort(Deps,Fulfilled,[First|Rest]) :-
    select(dep(First,Before),Deps,Drest),
    is_sublist(Before,Fulfilled),!,
    sort([First|Fulfilled],NewFulfilled),
    topolsort(Drest,NewFulfilled,Rest).
is_sublist([],_).
is_sublist([H|T],[F|R]) :-
    ( H=F -> is_sublist(T,R)
    ; otherwise -> is_sublist([H|T],R)).

get_sorted([],_,[]).
get_sorted([Name|Nrest],Elems,[First|Srest]) :-
    member(First,Elems),
    extract_name(First,Name),!,
    get_sorted(Nrest,Elems,Srest).
    
% copy&paste from bmachine_construction
rawmachine_section(Elem,List,Result) :-
    functor(Pattern,Elem,2),
    arg(2,Pattern,Result),
    member(Pattern,List),!.

optional_rawmachine_section(Elem,List,Default,Result) :-
    ( rawmachine_section(Elem,List,Result1) -> true
    ; Result1=Default),
    Result1=Result.

% ----------------------------------------


% older versions of the .eventb files do not have a status included
raw_event(event(Pos,Name,Refines,Params,Guards,Actions,Witnesses),
      Pos,Name,ordinary(none),Refines,Params,Guards,[],Actions,Witnesses) :-
    (\+((Refines==[] ; Refines=[_|_]))
     -> add_error(b_machine_eventb,'Warning: outdated .eventb format') ; true).
raw_event(event(Pos,Name,Status,Refines,Params,Guards,Actions,Witnesses),
      Pos,Name,Status,Refines,Params,Guards,[],Actions,Witnesses).
raw_event(event(Pos,Name,Status,Refines,Params,Guards,Theorems,Actions,Witnesses),
      Pos,Name,Status,Refines,Params,Guards,Theorems,Actions,Witnesses).

% code common to bmachine_eventb.pl and bmachine_eventb_ml.pl

:- use_module(bmachine_structure,[select_section/5]).
append_to_section(Section,List,Old,New) :-
    % todo: access to private elements of bmachine_construction
    bmachine_structure:select_section(Section,OldList,NewList,Old,New),
    append(OldList,List,NewList).
conjunct_to_section(Section,Preds,Old,New) :-
    % todo: access to private elements of bmachine_construction
    bmachine_structure:select_section(Section,OldPred,NewPred,Old,New),
    conjunct_predicates([OldPred|Preds],NewPred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% analyse Event-B deferred sets and generate enumerated sets if possible

enumerated_deferred_sets(Sets,Constants,Axioms,NewSets,NewConstants,NewAxioms,TSet,TElements) :-
    conjunct_predicates(Axioms,Axiom),
    conjunction_to_list(Axiom,Axiom2),
    enumerated_deferred_sets2(Sets,Constants,Axiom2,NewSets,NewConstants,NewAxioms,TSet,TElements).

enumerated_deferred_sets2([],Csts,Axioms,[],Csts,Axioms,[],[]).
enumerated_deferred_sets2([Set|SetRest],Csts,Axioms,NewSets,NewCsts,NewAxioms,[TSet|TSetRest],TElementsAll) :-
    % case 1: the set is an enumerated set
    enumerated_deferred_set(Set,Csts,Axioms,ICsts,IAxioms,TSet,TElements),!,
    append(TElements,TElementsRest,TElementsAll),
    enumerated_deferred_sets2(SetRest,ICsts,IAxioms,NewSets,NewCsts,NewAxioms,TSetRest,TElementsRest).
enumerated_deferred_sets2([Set|SetRest],Csts,Axioms,[Set|NewSets],NewCsts,NewAxioms,TSets,TElements) :-
    % case 2: the set is a normal deferred set
    enumerated_deferred_sets2(SetRest,Csts,Axioms,NewSets,NewCsts,NewAxioms,TSets,TElements).

enumerated_deferred_set(Set,Constants,Axioms,NewConstants,NewAxioms,Set,TElements) :-
   %% print(checking_for_enumerated_deferred_sets(Axioms)),nl,
    has_enumeration_axioms(Set,Elements,Axioms,NewAxioms),
 %% print(found_enumerated_set(Set)),nl, %%
    % remove Elem1,...,ElemN from the constants
    remove_constants(Elements,Constants,NewConstants),
    % add enumerated set information to the elements
    get_texpr_type(Set,SType),
    (SType=set(global(Type)) -> true
     ; add_error_and_fail(enumerated_deferred_set,'Unexpected Set: ',(Set:SType))),
    (typeset_enumerated(Elements,Type,TElements) -> true
     ; add_error_and_fail(enumerated_deferred_set,'Could not typeset: ',(Type,Elements))).


has_enumeration_axioms(Set,Elements,AxiomsIn,AxiomsOut) :-
    has_partition_axioms(Set,Elements,AxiomsIn,AxiomsOut),!.
has_enumeration_axioms(Set,Elements,AxiomsIn,AxiomsOut) :-
    % there in an axiom Set = {Elem1,...,ElemN}
    select_equal_extension_set_of_ids(Set,AxiomsIn,Axioms1,ExtSet,Elements),
    % check if all axioms needed for an enumeration are
    % present, and remove them
    % second case: All elements of the set are mutually unequal
    
    % generate the unequalities we have to check for
    generate_unequals(Elements,Unequals),
    % check if they are all in the axioms, and remove them from the axioms
    (all_unequals_present(Unequals,Axioms1,AxiomsOut) -> true
      ; assert(deferred_set_eq_wo_enumeration_axioms(Set,ExtSet)),
        print('Deferred set with equality but *not* recognized as enumerated set: '),
        translate:print_bexpr(Set), print(' = '), translate:print_bexpr(ExtSet),nl,
        fail
     ).

has_partition_axioms(Set,PartitionElements,AxiomsIn,AxiomsOut) :-
    % print(checking_for_enumeration_axioms(Set)),nl,
    % first case: A partion predicate is present with
    % all elements of the set as its arguments
    %  ( e.g. partition(S,{a},{b},...) )

    % create a pattern for the partition:
    get_texpr_id(Set,SetId),get_texpr_id(SetRef,SetId),
    create_texpr(partition(SetRef,Singletons),pred,_,Partition),
    % remove the partition from the axioms
    % print(looking_for_partition(SetId,Partition)),nl,nl,
    select(Partition,AxiomsIn,AxiomsOut1),
    %%print(found_partition(SetID,Partition)),nl,
    % check if all arguments are singletons and get the
    % singleton elements
    all_singletons(Singletons,PartitionElements,AxiomsOut1,AxiomsOut2,Recursive),
    (Recursive==true
     -> AxiomsOut=AxiomsIn  /* then we still need the partition axioms */
     ; AxiomsOut=AxiomsOut2).
%%    ,print(all_singletons(Singletons)),nl.
    %% print(singletons(PartitionElements)),nl,
    % check if all arguments and the identifiers
    % of the set are the same
    % same_elements(Elements,PartitionElements).


% all elements of a list are singletons, the second argument
% returns the list of elements without the set around it
all_singletons([],[],Ax,Ax,_).
all_singletons([Set|SRest],[Expr|ERest],AxiomsIn,AxiomsOut,Recursive) :-
    get_texpr_expr(Set,set_extension([Expr])),!,
    all_singletons(SRest,ERest,AxiomsIn,AxiomsOut,Recursive).
all_singletons([ID1|SRest],Result,AxiomsIn,AxiomsOut,Recursive) :- 
    ID1 = b(identifier(_ID),set(global(_)),_Info),
    Recursive=true,
    % print(trying_to_look_for_recursive_partition(_ID)),nl, %%
    has_enumeration_axioms(ID1,PartitionElements,AxiomsIn,AxiomsOut1),
   % print(recursive_partition(ID,PartitionElements)),nl,
    append(PartitionElements,ERest,Result),
    all_singletons(SRest,ERest,AxiomsOut1,AxiomsOut,_).
    
% the two lists of typed identifiers have the same identifiers
%same_elements(ElemsA,ElemsB) :-
%    get_texpr_ids(ElemsA,IdsA),sort(IdsA,A),
%    get_texpr_ids(ElemsB,IdsB),sort(IdsB,B),
%    A=B.

typeset_enumerated([],_Type,[]).
typeset_enumerated([Elem|Erest],Type,[TElem|Trest]) :-
    create_texpr(identifier(Id),global(Type),Infos,Elem),
    create_texpr(identifier(Id),global(Type),[enumerated_set_element|Infos],TElem),
    typeset_enumerated(Erest,Type,Trest).

remove_constants([],Cst,Cst).
remove_constants([TId|Erest],In,Out) :-
    get_texpr_id(TId,Id),
    get_texpr_id(Pattern,Id),
    select(Pattern,In,Env),
    remove_constants(Erest,Env,Out).

% find an equality Set = {id1,...,idk}
select_equal_extension_set_of_ids(Set,In,Out,Ext,Elements) :-
    get_texpr_id(Set,SetId),
    get_texpr_id(Expr,SetId),
    create_texpr(set_extension(Elements),_,_,Ext),
    create_texpr(Equal,_,_,TEqual),
    ( Equal = equal(Expr,Ext); Equal = equal(Ext,Expr) ),
    select(TEqual,In,Out),!.

generate_unequals(Elements,Unequals) :-
    get_texpr_ids(Elements,Ids),
    findall(unequal(A,B),two_ids(Ids,A,B),Unequals).
two_ids([A|Rest],A,B) :- member(B,Rest).
two_ids([_|Rest],A,B) :- two_ids(Rest,A,B).

all_unequals_present([],Axioms,Axioms).
all_unequals_present([unequal(A,B)|Rest],InAxioms,OutAxioms) :-
    remove_unequalities(InAxioms,A,B,IntAxioms,0,N), N>0,!, % at least one occurrence
    all_unequals_present(Rest,IntAxioms,OutAxioms).
    
remove_unequalities([],_,_,[],N,N).
remove_unequalities([TAxiom|T],A,B,Rest,NrRemoved,R) :-
    get_texpr_id(AExpr,A), get_texpr_id(BExpr,B),
    get_texpr_expr(TAxiom,Axiom),
    is_unequality(Axiom,AExpr,BExpr),!, N1 is NrRemoved+1,
    remove_unequalities(T,A,B,Rest,N1,R).
remove_unequalities([Axiom|T],A,B,[Axiom|TR],N,R) :- % not an unequality between A and B
    remove_unequalities(T,A,B,TR,N,R).    

is_unequality(not_equal(A,B),A,B) :- !.
is_unequality(not_equal(B,A),A,B) :- !.
is_unequality(negation(TEqual),A,B) :-
    get_texpr_expr(TEqual,Equal),
    is_equality(Equal,A,B).
is_equality(equal(A,B),A,B) :- !.
is_equality(equal(B,A),A,B) :- !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% creating a type environment with information about theories inside

check_extensions(Infos,Freetypes,Ops,TheoryEnv) :-
    findall( T, (member(T,Infos),is_eventb_additional_info(T)), Theories),
    default_theory_ops(InitialEnv),
    foldl(check_theory,Theories,FreetypesL,OpsL,InitialEnv,TheoryEnv),
    append(FreetypesL,Freetypes),append(OpsL,Ops).
is_eventb_additional_info(Term) :-
    nonvar(Term),functor(Term,theory,Arity),
    memberchk(Arity,[3,4,5,7]).
default_theory_ops(Env) :-
    env_empty(Empty),
    cond_operator(Id,Op),
    env_store_operator(Id,Op,Empty,Env).
cond_operator('COND',bmachine_eventb:direct_definition(Params,DirectDef,[PT])) :-
    PT = '__IFTHENELSETYPE',
    TypeExpr = identifier(none,PT),
    Params = [argument('THEN',TypeExpr),
              argument('ELSE',TypeExpr),
              argument('IF',set_extension(none,[truth(none)]))],
    DirectDef = if_then_else(none,identifier(none,'IF'),
                             identifier(none,'THEN'),identifier(none,'ELSE')).
check_theory(Theory,Freetypes,Ops,Env,OutEnv) :-
    check_theory2(Theory,Env,Freetypes,Ops),
    store_operator_list(Ops,Env,OutEnv).
store_operator_list(Ops,Env,OutEnv) :-
    keys_and_values(Ops,Ids,Operators),
    foldl(env_store_operator,Ids,Operators,Env,OutEnv).
check_theory2(theory(PT,Datatypes,OP),Env,Freetypes,Ops) :-
    % just to be backward-compatible to a version without axiomatic definitions
    check_theory2(theory(PT,Datatypes,OP,[]),Env,Freetypes,Ops).
check_theory2(theory(PT,Datatypes,OP,AxDefs),Env,Freetypes,Ops) :-
    % just to be backward-compatible to a version without operator mapping
    % definitions
    check_theory2(theory(PT,Datatypes,OP,AxDefs,[]),Env,Freetypes,Ops).
check_theory2(theory(PT,Datatypes,OP,AxDefs,Mappings),Env,Freetypes,Ops) :-
    check_theory2(theory(unknown,unknown,PT,Datatypes,OP,AxDefs,Mappings),Env,Freetypes,Ops).
check_theory2(theory(_Id,_Imported,PT,Datatypes,OP,AxDefs,Mappings),Env,Freetypes,AllOps) :-
    extract_freetypes(Env,Datatypes,Freetypes,FreetypeOps),
    foldl(handle_mappings(PT),Mappings,MappedOpsL,OP/AxDefs,OP1/AxDefs1),
    append(MappedOpsL,MappedOps),
    maplist(store_operator(PT),OP1,DefinedOps),
    maplist(store_axiomatic_operator(PT),AxDefs1,AxOpsL),
    append(AxOpsL,AxOps),
    append([FreetypeOps,MappedOps,DefinedOps,AxOps],AllOps1),
    % The operators are used for call-backs, a module prefix is mandatory
    maplist(prefix_with_module,AllOps1,AllOps).
prefix_with_module(Id-Op,Id-(bmachine_eventb:Op)).

store_operator(PT,operator(Id,Parameters,_WD,DirectDef,RecursiveDef),Id-Op) :-
    store_operator2(DirectDef,RecursiveDef,Parameters,PT,Op).
store_operator2(_,RecursiveDef,Parameters,PT,Op) :-
    RecursiveDef=[_|_],!,Op=recursive_operator(Parameters,RecursiveDef,PT).
store_operator2([Def],_,Parameters,PT,Op) :- !,
    Op=direct_definition(Parameters,Def,PT).
store_operator2(_,_,_,_,Op) :-
    Op = unsupported_operator('Operators without direct definition are currently unsupported').

extract_freetypes(ParamEnv,DataTypes,Freetypes,FreetypeOps) :-
    maplist(extract_freetype(DataTypes,ParamEnv),DataTypes,Freetypes,FreetypeOpsL),
    append(FreetypeOpsL,FreetypeOps).
extract_freetype(AllDataTypes,Env,
                 datatype(Id,TypeParams,Constructors),
                 freetype(FreetypeId,Cases),
                 [Id-freetype_operator(FreetypeId,Types,TypeParams,Usage)|CaseOps]) :-
    foldl(add_datatype_to_env,AllDataTypes,Env,Env1),
    foldl(add_type_parameter,TypeParams,Types,Env1,ParamEnv),
    create_freetype_typeid(Id,Types,FreetypeId),
    Type = freetype(FreetypeId),
    extract_used_type_parameters(Constructors,TypeParams,Usage),
    store_eventb_operator(Id,freetype_operator_as_identifier(Id),ParamEnv,ParamEnv2),
    maplist(extract_case(ParamEnv2,Type,Types),Constructors,Cases,CaseOpsL),
    append(CaseOpsL,CaseOps).
add_datatype_to_env(datatype(Id,Params,_),In,Out) :-
    length(Params,NumberOfParams),
    functor(FreetypeId,Id,NumberOfParams),
    env_store(Id,set(freetype(FreetypeId)),[],In,Out).
add_type_parameter(TypeParam,Type,In,Out) :-
    extract_id(TypeParam,ParameterId),
    env_store(ParameterId,set(Type),[],In,Out).
extract_id(identifier(_,Id),Id).
% constructor without parameters
extract_case(_ParamEnv,freetype(TypeId),_Types, constructor(Id,[]),        
             case(Id,constant([Id])),Ops) :- !,
    Ops=[Id-constructor_operator(TypeId,[])].
% constructor with parameters
extract_case(ParamEnv,Freetype,ParamTypes,constructor(Id,Destructors),
             case(Id,Type),
             [Id-constructor_operator(FreetypeId,Types)|DestructOps]) :-
    maplist(extract_destructor(ParamEnv,Freetype,ParamTypes,Id,Type),
            Destructors,Projections,TypesAndOps),
    maplist(split_type_and_op,TypesAndOps,Types,DestructOps),
    Freetype=freetype(FreetypeId),
    combine_typelist_with_prj(Types,Type,Projections).
extract_destructor(Env,freetype(FreetypeId),ParamTypes,Case,CType,
                   destructor(Name,TypeExpr),Projection,Type-Name-Op) :-
    typecheck_with_freevars(TypeExpr,Env,ParamTypes,set(Type),_),
    Op = destructor_operator(FreetypeId,Projection,Case,CType).
split_type_and_op(Type-Name-Op,Type,Name-Op).

create_freetype_typeid(Id,Types,FreetypeId) :-
    FreetypeId =.. [Id|Types].

:- assert_must_succeed(( combine_typelist_with_prj([a,b],T,P),
                         T == couple(a,b), P == [[prj1],[prj2]] )).
:- assert_must_succeed(( combine_typelist_with_prj([a,b,c],T,P),
                         T == couple(a,couple(b,c)),
                         P == [[prj1],[prj2,prj1],[prj2,prj2]] )).
:- assert_must_succeed(( combine_typelist_with_prj([a,b,c,d],T,P),
                         T == couple(couple(a,b),couple(c,d)),
                         P == [[prj1,prj1],[prj1,prj2],[prj2,prj1],[prj2,prj2]] )).
combine_typelist_with_prj([T],R,Projections) :- !,R=T,Projections=[[]].
combine_typelist_with_prj(Types,couple(TypeL,TypeR),Projections) :-
    split_list_in_middle(Types,TypesLeft,TypesRight),
    same_length(TypesLeft,PrjLeft),same_length(TypesRight,PrjRight),
    append(PrjLeft,PrjRight,Projections),
    maplist(cons(prj1),PrjLeft1,PrjLeft),
    maplist(cons(prj2),PrjRight1,PrjRight),
    combine_typelist_with_prj(TypesLeft,TypeL,PrjLeft1),
    combine_typelist_with_prj(TypesRight,TypeR,PrjRight1).
split_list_in_middle(List,Left,Right) :-
    length(List,N),
    Half is floor(N / 2),
    append_length(Left,Right,List,Half).

combine_exprlist_to_couple([T],R) :- !,T=R.
combine_exprlist_to_couple(Exprs,R) :-
    split_list_in_middle(Exprs,ExprsLeft,ExprsRight),
    combine_exprlist_to_couple(ExprsLeft,TL),
    combine_exprlist_to_couple(ExprsRight,TR),
    get_texpr_types([TL,TR],[TypeL,TypeR]),
    create_texpr(couple(TL,TR),couple(TypeL,TypeR),[],R).

combine_exprlist_to_cprod([T],R) :- !,T=R.
combine_exprlist_to_cprod(Exprs,R) :-
    split_list_in_middle(Exprs,ExprsLeft,ExprsRight),
    combine_exprlist_to_cprod(ExprsLeft,TL),
    combine_exprlist_to_cprod(ExprsRight,TR),
    get_texpr_types([TL,TR],[set(TypeL),set(TypeR)]),
    create_texpr(cartesian_product(TL,TR),set(couple(TypeL,TypeR)),[],R).

extract_used_type_parameters(Constructors,TypeParams,Usages) :-
    maplist(extract_id,TypeParams,TypeParamIds),
    convlist(used_type_parameters_of_constructor(TypeParamIds),Constructors,Usages).
used_type_parameters_of_constructor(TypeParamIds,constructor(Id,Destructors),
                                    used_param_types(Id,UsedParamTypes,DestructorExprs)) :-
    maplist(destructor_expr,Destructors,DestructorExprs),
    maplist(used_type_parameters_of_destructor(TypeParamIds),
            DestructorExprs,UsedParamTypeL),
    UsedParamTypeL = [_|_],   % store the information only for constructors that actually use
                              % type parameters
    ord_union(UsedParamTypeL,UsedParamTypes).
used_type_parameters_of_destructor(TypeParamIds,TypeExpr,UsedTypes) :-
    extract_raw_identifiers(TypeExpr,UsedIds),
    findall( N, (nth1(N,TypeParamIds,T),memberchk(T,UsedIds)), UsedTypes1),
    sort(UsedTypes1,UsedTypes).
destructor_expr(destructor(_,Expr),Expr).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% callback predicates for operators

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unsupported_operator(Msg,Id,_Args,Pos,_Env,ExPr,Result,TRin,TRout) :-
    store_typecheck_error(Msg,Pos,TRin,TRout),
    failure_syntax_element(ExPr,Id,F,Type), % just to make sure that the result
    create_texpr(F,Type,[],Result).         % is nonvar, to prevent later errors
failure_syntax_element(expr,Id,identifier(Id),_).
failure_syntax_element(pred,_Id,falsity,pred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: direct definitions
direct_definition(Params,DirectDef,PT,
                  _Id,TArgs,Pos,Env,ExPr,Result,TRin,TRout) :-
    apply_direct_definition(TArgs,Params,DirectDef,PT,Env,Pos,TLetIds,TLetValues,TDef,TRin,TRout),
    create_let(ExPr,TLetIds,TLetValues,TDef,[],Result).

pp_direct_definition(_Params,DirectDef,_PT,EventBId,I,O) :- 
    %print(ddef(_Params,DirectDef,_PT,Id,I,O)),nl,
    %functor(DirectDef,F,_), atom_codes(F,FC),  % TO DO: pp raw definition
    raw_bexpr_to_codes(DirectDef,DDCodes),
    functor(EventBId,FId,_), atom_codes(FId,FIdC),
    append(["By direct definition ",FIdC," -> ",DDCodes],Codes),
    append(Codes,O,I). % TODO: improve pretty printed message

:- use_module(library(codesio),[write_to_codes/2]).
raw_bexpr_to_codes(Raw,Codes) :- strip_raw(Raw,Stripped), write_to_codes(Stripped,Codes).
strip_raw(F,Res) :- atomic(F),!,Res=F.
strip_raw(identifier(_,I),Res) :- !, Res=I.
strip_raw(F,Res) :- F=..[Op,_|A],!, maplist(strip_raw,A,SA), Res =..[Op|SA].
strip_raw(F,F).

% Example: ddef([argument(R,pow_subset(none,identifier(none,T)))],seq(none,identifier(none,R)),[T],seq)



% Args are the arguments of the operator where it is used
% Params are the parameters in the operator definition
:- public apply_direct_definition/11.
apply_direct_definition(TArgs,Params,DirectDef,PT,Env,_Pos,
                        TLetIds,TLetValues,TDef,TRin,TRout) :-
    same_length(TArgs,Params),!,
    maplist(create_typed_id,PT,ParametricTypes,TPT),
    add_identifiers_to_environment(TPT,Env,SubEnv),
    % type check the arguments given in the operator usage
    get_texpr_types(TArgs,ArgTypes),
    typecheck_arguments(Params,SubEnv,ArgTypes,TParams,TRin,TR2),
    % type check the direct definition of the operator. To do this we need the parameters of the operators
    % in the type environment. Their type is determined by the arguments of the operator call.
    add_identifiers_to_environment(TParams,SubEnv,DefEnv),
    btype(apply_direct_definition,DirectDef,DefEnv,_Type,TDef,TR2,TRout),
    append(TParams,TPT,TLetIds),
    maplist(create_typeset,ParametricTypes,TypeValues),
    append(TArgs,TypeValues,TLetValues).
apply_direct_definition(_TArgs,_Arguments,_DirectDef,_PT,_Env,Pos,
                        _TParams,_TDef,TRin,TRout) :-
    store_typecheck_error('Wrong number of operator arguments',Pos,TRin,TRout).
split_arg_and_typedef(argument(Id,TypeExpr),Id,TypeExpr).
map_to_set_type(Type,set(Type)).
typecheck_arguments(Params,Env,ArgTypes,TParams,TRin,TRout) :-
    % type check the parameters' type expressions of the operator definition, they should have the same type
    % as the arguments (except being a set)
    % E.g. an operator "o" has one argument "arg" and type expression "T**BOOL" (with T being a parameter type of the theory),
    %   then for "o(5|->TRUE)" we have "couple(integer,boolean)" as the argument's type,
    %   we add "T" to the type environment (with type set(_)),
    %   then we type check the parameter expression "T**BOOL", expecting the type set(coule(integer,boolean)),
    %   the type check results in T being of type set(integer).
    maplist(split_arg_and_typedef,Params,ParamIds,ParamTypeExprs),
    same_length(Params,ArgTypes),
    maplist(map_to_set_type,ArgTypes,SetArgTypes),
    btype_l(ParamTypeExprs,Env,SetArgTypes,_TParamTypeExprs,TRin,TRout),
    maplist(create_typed_id,ParamIds,ArgTypes,TParams).
create_typeset(Type,TExpr) :- create_texpr(typeset,Type,[],TExpr).
create_let(ExPr,Ids,Values,Inner,Infos,Result) :-
    remove_unused_identifiers(Ids,Values,Inner,RIds,RValues),
    ( RIds=[] -> Result = Inner
    ; ExPr=expr ->
        get_texpr_type(Inner,Type),
        create_texpr(let_expression(RIds,RValues,Inner),Type,Infos,Result)
    ; ExPr=pred ->
        create_texpr(let_predicate( RIds,RValues,Inner),pred,Infos,Result)).
remove_unused_identifiers(TLetIds1,TLetValues1,TDef,TLetIds,TLetValues) :-
    find_identifier_uses(TDef,[],UsedIds),
    maplist(combine_with_minus,TLetIds1,TLetValues1,IdValuePairs),
    include(id_is_used(UsedIds),IdValuePairs,NewIdValuePairs),
    maplist(combine_with_minus,TLetIds,TLetValues,NewIdValuePairs).
combine_with_minus(A,B,A-B).
id_is_used(UsedIds,TId-_Value) :-
    get_texpr_id(TId,Id),memberchk(Id,UsedIds).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: CHOOSE operator
% Maybe the use of the direct_definition operator would be possible but it turned out
% that constructing the necessary parameters was quite tricky.
choose_operator(Id,[TArg],Pos,_Env,ExPr,Result,TRin,TRout) :- !,
    ( get_texpr_type(TArg,set(Type)) ->
        create_texpr(external_function_call('CHOOSE',[TArg]),Type,[],Result),
        TRin = TRout
    ; otherwise ->
        failure_syntax_element(ExPr,Id,Expr,Type),
        create_texpr(Expr,Type,[],Result), % just to avoid free variables in the AST
        store_typecheck_error('Expected set for choose operator argument',Pos,TRin,TRout)
    ).
choose_operator(_Id,_TArgs,Pos,_Env,_ExPr,_Result,TRin,TRout) :-
    store_typecheck_error('Wrong number of operator arguments',Pos,TRin,TRout).
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: Recursive operators
recursive_operator(Arguments,Cases,PT,
                   Id,TArgs,_Pos,Env,ExPr,Result,TRin,TRout) :-
    % parameter types
    maplist(create_typed_id,PT,ParametricTypes,TPT),
    add_identifiers_to_environment(TPT,Env,Env1),
    maplist(create_typeset,ParametricTypes,LetValues), % needed for the LET
    typecheck_arguments(Arguments,Env1,ArgTypes,TParamIds,TRin,TR1),
    % TArgs: A,B,Cf
    % TArg: A|->B|->C
    get_texpr_types(TArgs,ArgTypes),
    create_couple(TArgs,TArg),
    length(TArgs,NumberOfArgs),OpInfo = [theory_operator(Id,NumberOfArgs)],
    unique_id("opresult.",ResultId),create_typed_id(ResultId,Type,TResultId),
    ( ExPr=expr ->              % Result: Comp(x|->y|->z)
        % we need to append a result parameter in the comprehension set
        create_texpr(function(TCompSet,TArg),Type,OpInfo,InnerResult),
        append(TParamIds,[TResultId],CompSetIds),append(ArgTypes,[Type],CompSetTypes),
        % make sure that we do not run into an infinite loop
        store_eventb_operator(Id,freetype_operator_as_function(RecId,Type),Env1,Env2)
    ; ExPr=pred -> % Result: x|->y|->z : Comp
        create_texpr(member(TArg,TCompSet),pred,OpInfo,InnerResult),
        CompSetIds = TParamIds,ArgTypes=CompSetTypes,
        % make sure that we do not run into an infinite loop
        store_eventb_operator(Id,freetype_operator_as_set(RecId),Env1,Env2)
    ),
    couplise_list(CompSetTypes,CompSetType),
    % Comp: {i,r| Cond}
    create_recursive_compset(CompSetIds,Cond,set(CompSetType),[],RecId,TCompSet),
    % Params: a,b,c
    % i = a|->b|->c
    add_identifiers_to_environment(TParamIds,Env2,Subenv),
    foldl(recursive_operator_case(Subenv,TResultId),Cases,TCases,TR1,TRout),
    conjunct_predicates(TCases,Cond),
    create_let(ExPr,TPT,LetValues,InnerResult,OpInfo,Result).


recursive_operator_case(Env,TResultId,case(ParamId,_,Expression,Formula),Cond,TRin,TRout) :-
    % Add free identifiers of case to environment
    extract_free_identifiers(Expression,TFreeIdentifiers,Env,Subenv),
    % IsCase => CaseCheck
    create_texpr(implication(IsCase,TCaseCheck),pred,[],Cond),
    % IsCase: ParamId = constructor(x,y,z)
    typecheck_parameter(ParamId,Env,TParamId),
    recop_create_is_case(Expression,Subenv,IsCase,TParamId,FT,Case,TRin,TR1),
    % CaseCheck: #x,y,z | Destruction & Predicate
    recop_create_in_case(Formula,TParamId,TResultId,Subenv,TFreeIdentifiers,FT,Case,TCaseCheck,TR1,TRout).
extract_free_identifiers(typeof(_Pos,ExtExpr,_TypeExpr),TIds,In,Out) :-
    extract_free_identifiers(ExtExpr,TIds,In,Out).
extract_free_identifiers(extended_expr(_Pos,_Case,Exprs,Preds),TIds,In,Out) :-
    append(Exprs,Preds,Ids),
    maplist(extract_free_identifier,Ids,TIds),
    add_identifiers_to_environment(TIds,In,Out).
extract_free_identifier(identifier(_Pos,Id),TId) :-
    create_typed_id(Id,_,TId).
typecheck_parameter(ParamId,Env,TParamId) :-
    env_lookup_type(ParamId,Env,ParamType),
    create_typed_id(ParamId,ParamType,TParamId).
env_store_type(Id,Type,In,Out) :-
    % Version without Info field
    env_store(Id,Type,[],In,Out).
recop_create_is_case(typeof(_,Expr,_TypeExpr),Env,IsCase,TParamId,FT,Case,TRin,TRout) :-
    recop_create_is_case(Expr,Env,IsCase,TParamId,FT,Case,TRin,TRout).
recop_create_is_case(Expression,Env,IsCase,TParamId,FT,Case,TRin,TRout) :-
    Expression = extended_expr(_,Case,_,_),
    get_texpr_type(TParamId,freetype(FT)),
    btype(recop_create_is_case,Expression,Env,freetype(FT),_TCons,TRin,TRout),
    create_texpr(freetype_case(FT,Case,TParamId),pred,[],IsCase).
recop_create_in_case(Formula,TParamId,TResultId,Subenv,TFreeIdentifiers,FT,Case,Cond,TRin,TRout) :-
    % Predicate
    btype(recop_create_in_case,Formula,Subenv,Type,TFormula,TRin,TRout),
    ( Type=pred -> Predicate = TFormula
    ; otherwise ->
        get_texpr_types([TResultId,TFormula],[TypeA,TypeB]),
        unify_types_strict(TypeA,TypeB),
        create_texpr(equal(TResultId,TFormula),pred,[],Predicate)),
    % Destruction: x->y|->z = destructor(ParamId)
    ( TFreeIdentifiers = [] ->
        Cond = Predicate
    ; TFreeIdentifiers = [_|_] ->
        get_texpr_types(TFreeIdentifiers,FTypes),
        combine_typelist_with_prj(FTypes,FType,Projections),
        unique_id("destructed",DId),create_typed_id(DId,FType,TDId),
        create_texpr(freetype_destructor(FT,Case,TParamId),FType,[],Destructor),
        maplist(create_inner_let(TDId),Projections,TLetExpressions),
        create_let(pred,TFreeIdentifiers,TLetExpressions,Predicate,[],InnerLets),
        create_let(pred,[TDId],          [Destructor],   InnerLets,[],Cond)
    ).

create_inner_let(TOrig,Projections,TExpr) :-
    apply_projections(Projections,TOrig,TExpr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype sets
:- public freetype_operator/12.
freetype_operator(FreetypeId1,Types1,TypeParams1,Usage1,
                  Id,TArgs,_Pos,Env,_ExPr,Result,TR,TR) :-
    % make a fresh copy to prevent that several instances of the freetype propagate
    % their types to each other (resulting in spurious typing errors)
    copy_term(freetype(FreetypeId1,Types1,TypeParams1,Usage1),
              freetype(FreetypeId, Types, TypeParams, Usage)),
    maplist(map_to_set_type,Types,SetTypes),
    get_texpr_types(TArgs,SetTypes),
    create_freetype_compset(TArgs,Id,FreetypeId,TypeParams,Usage,Env,Result).
create_freetype_compset(TArgs,SetId,FreetypeId,TypeParams,Usage,Env,Result) :-
    find_constructors_with_complex_types(TArgs,Usage,Constructors),
    ( Constructors = [] ->
        create_texpr(freetype_set(FreetypeId),set(freetype(FreetypeId)),[],Result)
    ; otherwise ->
        create_freetype_compset2(TArgs,SetId,Constructors,FreetypeId,TypeParams,Env,Result)).
create_freetype_compset2(TArgs,SetId,Constructors,FreetypeId,TypeParams,Env,Result) :-
    btypechecker:add_ext_variables(TypeParams,Env,TParams,Subenv,TRin,TRout),
    get_texpr_types(TParams,Types),get_texpr_types(TArgs,Types),
    Type=freetype(FreetypeId),
    create_texpr(identifier(freetype_element),Type,[],TId),
    do_prettyprint_freetype(FreetypeId,TArgs,PP),
    create_recursive_compset([TId],Let,set(Type),[freetype(PP)],RecId,Result),
    store_eventb_operator(SetId,freetype_operator_as_identifier(RecId),Subenv,Subenv2),
    foldl(create_cons_check(FreetypeId,Subenv2,TId),Constructors,Checks,TRin,TRout),
    conjunct_predicates(Checks,Cond),
    create_let(pred,TParams,TArgs,Cond,[],Let).
create_cons_check(FreetypeId,Env,TId,cons(Case,Exprs),Check,TRin,TRout) :-
    btype_l(Exprs,Env,_Types,TExprs,TRin,TRout),
    combine_exprlist_to_cprod(TExprs,Set),
    get_texpr_type(Set,set(Type)),
    create_texpr(freetype_case(FreetypeId,Case,TId),pred,[],CheckCase),
    create_texpr(freetype_destructor(FreetypeId,Case,TId),Type,[],Value),
    create_texpr(member(Value,Set),pred,[],Membercheck),
    create_texpr(implication(CheckCase,Membercheck),pred,[],Check).
% this is used to pretty-print the resulting comprehension set
do_prettyprint_freetype(Id,Params,R) :-
    maplist(translate_bexpression,Params,Strs),
    ajoin_with_sep(Strs,',',S),
    functor(Id,I,_),
    ajoin([I,'(',S,')'],R).

find_constructors_with_complex_types(TArgs,Usage,Constructors) :-
    find_complex_type_params(TArgs,ComplexTP),
    findall(cons(Id,DestructorExprs),
            ( member(used_param_types(Id,UsedParamTypes,DestructorExprs),Usage),
                ord_intersect(UsedParamTypes,ComplexTP)),
            Constructors).
% returns the indeces of the type arguments that are not only types
find_complex_type_params(TArgs,ComplexTP) :-
    findall(N,(nth1(N,TArgs,TArg),\+ is_just_type(TArg)),ComplexTP1),
    sort(ComplexTP1,ComplexTP).

freetype_operator_as_identifier(RecId,Id,TArgs,_Pos,_Env,_ExPr,Result,TR,TR) :-
    get_texpr_types(TArgs,SetTypes),
    maplist(map_to_set_type,Types,SetTypes),
    create_freetype_typeid(Id,Types,FreetypeId),
    create_recursion_ref(RecId,set(freetype(FreetypeId)),Result).

freetype_operator_as_function(RecId,RType,_Id,TArgs,_Pos,_Env,_ExPr,Result,TR,TR) :-
    create_couple(TArgs,TArg), get_texpr_type(TArg,Type),
    create_recursion_ref(RecId,set(couple(Type,RType)),TFunction),
    create_texpr(function(TFunction,TArg),RType,[],Result).

freetype_operator_as_set(RecId,_Id,TArgs,_Pos,_Env,_ExPr,Result,TR,TR) :-
    create_couple(TArgs,TArg), get_texpr_type(TArg,Type),
    create_recursion_ref(RecId,set(Type),TSet),
    create_texpr(member(TArg,TSet),pred,[],Result).

create_recursion_ref(RecId,Type,TId) :-
    create_texpr(identifier(RecId),Type,[for_recursion],TId).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype constructors
:- public constructor_operator/10. % TO DO: ensure by another mechanism that Spider detects that this is called
constructor_operator(FreetypeId1,[],Id,[],_Pos,_Env,_ExPr,Result,TRin,TRout) :-
    !, copy_term(FreetypeId1,FreetypeId),
    create_texpr(value(freeval(FreetypeId,Id,term(Id))),
                 freetype(FreetypeId),[],Result),
    TRin = TRout.
constructor_operator(FreetypeId1,ParamTypes1,
                     Id,TArgs,_Pos,_Env,_ExPr,Result,TR,TR) :-
    copy_term(constructor(FreetypeId1,ParamTypes1),constructor(FreetypeId,ParamTypes)),
    get_texpr_types(TArgs,ParamTypes),
    combine_exprlist_to_couple(TArgs,TArg),
    create_texpr(freetype_constructor(FreetypeId,Id,TArg),freetype(FreetypeId),[],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype destructors
:- public destructor_operator/12. % TO DO: ensure by another mechanism that Spider detects that this is called
destructor_operator(FreetypeId1,Projections,Case,CType1,
                    _Id,[TArg],_Pos,_Env,_ExPr,Result,TR,TR) :-
    copy_term(destructor(FreetypeId1,CType1),destructor(FreetypeId,CType)),
    get_texpr_type(TArg,freetype(FreetypeId)),
    create_texpr(freetype_destructor(FreetypeId,Case,TArg),CType,[],Destructed),
    apply_projections(Projections,Destructed,Result).

% create an projection expression (or a combination of arbitrary many) to
% access a part of a couple. The input is a list of prj1/prj2 atoms,
% e.g. [prj1,prj1,prj2]: Take the first of the first of the second element.
apply_projections(Projections,Expr,Result) :-
    rev(Projections,Prjs),
    apply_projections1(Prjs,Expr,Result).
apply_projections1([],Expr,Expr).
apply_projections1([P|Rest],Expr,Result) :-
    get_texpr_type(Inner,couple(A,B)),
    apply_projections2(P,A,B,Result,Inner),
    apply_projections1(Rest,Expr,Inner).
apply_projections2(prj1,A,_B,Result,Inner) :-
    create_texpr(first_of_pair(Inner),A,[],Result).
apply_projections2(prj2,_A,B,Result,Inner) :-
    create_texpr(second_of_pair(Inner),B,[],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: axiomatic definitions of operators
store_axiomatic_operator(_PT,axiomatic_def_block(_Id,_Types,OpDefs,_Axioms),Ops) :-
    % in case we do not recognize the operator, we store an error message
    % that is shown as soon the operator is used somewhere
    maplist(store_axdef_error,OpDefs,Ops).
store_axdef_error(opdef(Id,_Arguments,_WD),Id-unsupported_operator(Msg)) :-
    ajoin(['Axiomatic defined operator \"',Id,'\" not recognized. Be sure to add a .ptm file to your Theory project.'],Msg).

    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: Tagged operators that can be replaced by internal constructs
handle_mappings(PT,tag(Opname,Tag),Ops,OPin/AxDefsin,OPout/AxDefsOut) :- !,
    ( extract_operator(Opname,Operator,OPin/AxDefsin,OPout/AxDefsOut) ->
        (debug:debug_mode(on) -> format("Mapping Theory Operator ~s -> ~w~n",[Tag,Opname]) ; true),
        (store_operator_by_tag(Tag,Opname,Operator,PT,Ops) -> true
          ; add_internal_error('Error mapping Theory Operator (the *.ptm file maybe incorrect or you need a newer version of ProB): ',Tag:Opname))
    ; otherwise ->
        ajoin(['ProB theory mapping: operator ',Opname,' not found, mapping entry ignored.'],Msg),
        add_error(bmachine_eventb,Msg),
        OPin/AxDefsin=OPout/AxDefsOut,Ops=[]
    ).
handle_mappings(_PT,Mapping,[],X,X) :-
    functor(Mapping,Functor,Arity),
    ajoin(['ProB theory mapping: unexpected entry of type ',Functor,'/',Arity,', entry ignored.'],Msg),
    add_error(bmachine_eventb,Msg).

extract_operator(Opname,Operator,OPin/AxDefs,OPout/AxDefs) :-
    Operator = operator(Opname,_Parameters,_WD,_DirectDef,_RecursiveDef),
    selectchk(Operator,OPin,OPout),!.
extract_operator(Opname,Operator,OP/AxDefsIn,OP/AxDefsOut) :-
    Operator = axdef(Opname,Parameters,WD,block(Id,Types,Axioms)),
    select(axiomatic_def_block(Id,Types,OpDefsIn,Axioms),AxDefsIn,axiomatic_def_block(Id,Types,OpDefsOut,Axioms),AxDefsOut),
    select(opdef(Opname,Parameters,WD),OpDefsIn,OpDefsOut),!.

% two versions are supported: The older one expects an atom Tag, the new one a string
% The string version is then parsed as a Prolog term.
store_operator_by_tag(Tag,Opname,Operator,PT,Ops) :-
    is_list(Tag),!,       % A string
    append(Tag,".",Tag1), % read_from_codes/2 needs a full-stop at the end
    read_from_codes(Tag1,Term),
    store_operator_by_tag(Term,Opname,Operator,PT,Ops).
store_operator_by_tag(Tag,Opname,Operator,PT,Ops) :-
    store_operator_by_tag1(Tag,Opname,Operator,PT,Ops).
store_operator_by_tag1(Tag,Opname,Operator,PT,Ops) :-
    % SIGMA or PI
    aggregation_operator(Tag,Functor),!,
    get_operator_parameters(Operator,Parameters),
    ( Parameters = [Arg] ->
        create_raw_aggregation_op(Functor,Arg,ClassicalBOp),
        Ops=[Opname-direct_definition([Arg],ClassicalBOp,PT)]
    ; otherwise ->
        parameter_error(Parameters,1,Opname),Ops=[]
    ).
store_operator_by_tag1(mkinat,OpName,Operator,PT,Ops) :- !,
    store_operator_by_tag1(mkinat(iZero,iSucc),OpName,Operator,PT,Ops).
store_operator_by_tag1(mkinat(ZeroOp,SuccOp),OpName,Operator,_PT,Ops) :- !,
    get_operator_parameters(Operator,Parameters),
    ( Parameters = [argument(Arg,_ArgType)] ->
        Ops=[OpName-make_inductive_natural(Arg,ZeroOp,SuccOp)]
    ; otherwise ->
        parameter_error(Parameters,1,OpName),Ops=[]
    ).
store_operator_by_tag1(choose,OpName,Operator,_PT,Ops) :- !,
    get_operator_parameters(Operator,Parameters),
    ( Parameters = [argument(_Arg,_ArgType)] ->
        Ops=[OpName-choose_operator]
    ; otherwise ->
        parameter_error(Parameters,1,OpName),Ops=[]
    ).
store_operator_by_tag1(BOperator,Opname,Operator,PT,Ops) :- unary_operator_raw_term(BOperator,Arg,Expr), !,
    get_operator_parameters(Operator,Parameters),
    ( Parameters = [argument(Arg,_ArgType)] ->
        Ops=[Opname-direct_definition(Parameters,Expr,PT)]
    ; otherwise ->
        parameter_error(Parameters,1,Opname),Ops=[]
    ).
store_operator_by_tag1(BOperator,Opname,Operator,PT,Ops) :- %print(try(BOperator,Opname,Operator,PT,Ops)),nl,
    binary_operator_raw_term(BOperator,Arg1,Arg2,Expr), !,
    get_operator_parameters(Operator,Parameters),
    ( Parameters = [argument(Arg1,_ArgType1), argument(Arg2,_ArgType2)] ->
        Ops=[Opname-direct_definition(Parameters,Expr,PT)]
    ; otherwise ->
        parameter_error(Parameters,1,Opname),Ops=[]
    ).

% TO DO ternary operator ?? son/3

% construct a raw term with arguments for replacement in AST
% first argument is functor name as it appears in the theory mapping file *.ptm
binary_operator_raw_term(Functor,Arg1,Arg2,Term) :- binary_operator_to_ast(Functor,ASTFunctor),
    Term =.. [ASTFunctor,none,identifier(none,Arg1),identifier(none,Arg2)].
binary_operator_to_ast('^',concat).
binary_operator_to_ast(concat,concat).
binary_operator_to_ast(div,floored_div).
binary_operator_to_ast(restrict_front,restrict_front).
binary_operator_to_ast(restrict_tail,restrict_tail).
% binary tree operators
binary_operator_to_ast(arity,arity).
binary_operator_to_ast(bin,bin).
binary_operator_to_ast(const,const).
binary_operator_to_ast(father,father).
binary_operator_to_ast(rank,rank).
binary_operator_to_ast(subtree,subtree).

% construct a raw term with argument for replacement in AST
% first argument is functor name as it appears in the theory mapping file *.ptm
unary_operator_raw_term(Functor,Arg,Term) :- unary_operator_to_ast(Functor,ASTFunctor),
    Term =.. [ASTFunctor,none,identifier(none,Arg)].
unary_operator_to_ast(closure,reflexive_closure). % reflexive_closure corresponds to closure B operator !
unary_operator_to_ast(closure1,closure).
unary_operator_to_ast(conc,general_concat).
unary_operator_to_ast(fnc,trans_function).
unary_operator_to_ast(rel,trans_relation).
unary_operator_to_ast(succ,successor). % also exist in Event-B
unary_operator_to_ast(pred,predecessor). % also exist in Event-B
unary_operator_to_ast(bool,convert_bool). % also exist in Event-B
unary_operator_to_ast(id,identity). % exists in similar way in Event-B
unary_operator_to_ast(union,general_union).
unary_operator_to_ast(inter,general_intersection).
unary_operator_to_ast(F,F) :- unary_operator(F). % operators that have same name in B and AST

% sequence operators
unary_operator(seq).
unary_operator(seq1).
unary_operator(iseq).
unary_operator(iseq1).
unary_operator(perm).
unary_operator(first).
unary_operator(size).
unary_operator(front).
unary_operator(tail).
unary_operator(rev).
unary_operator(last).
unary_operator(mu).
% tree operators
unary_operator(tree).
unary_operator(btree).
unary_operator(top).
unary_operator(sons).
unary_operator(bin).
unary_operator(left).
unary_operator(right).
unary_operator(sizet).
unary_operator(prefix).
unary_operator(postfix).
%unary_operator(infix). % not yet supported
%unary_operator(mirror). % not yet supported
% general operators
unary_operator(min). % also exist in Event-B
unary_operator(max). % also exist in Event-B
unary_operator(domain).
unary_operator(range).
% z operators
unary_operator(compaction).
unary_operator(bag_items).
    
aggregation_operator('SIGMA',general_sum).
aggregation_operator('PI',general_product).

create_raw_aggregation_op(Op,argument(Arg,_ArgType),Result) :-
    % this construction will immediately wrap _agg_op1/2 inside Sigma or Pi, and thus _agg_op1/2 will not be visibile outside
    A = identifier(none,Arg),                 % User-given argument
    create_fresh_identifier('_agg_op1',[A],I1), % starting with _ ensures user cannot create such an id
    create_fresh_identifier('_agg_op2',[A],I2),
    I = identifier(none,I1),                  % I is the (not used) index variable
    N = identifier(none,I2),                  % N is the integer to sum or multiply
    Pred = member(none,couple(none,I,N),A),   % I|->N : Arg
    Result =.. [Op,none,[I,N],Pred,N].        % SIGMA(I,N . I|->N:Arg | N)  (or PI)

parameter_error(Parameters,Expected,Opname) :-
        length(Parameters,N),
        ajoin(['ProB theory mapping: Expected ',Expected,' parameter for operator ',
               Opname,' but found ',N],Msg),
        add_error(bmachine_eventb,Msg).

get_operator_parameters(operator(_Opname,Parameters,_WD,_DirectDef,_RecursiveDef),Parameters).
get_operator_parameters(axdef(_Opname,Parameters,_WD,_Block),Parameters).

store_eventb_operator(Id,Op,Ein,Eout) :-
    env_store_operator(Id,bmachine_eventb:Op,Ein,Eout).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Operator for constructing inductive natural numbers
:- public make_inductive_natural/11.
make_inductive_natural(ArgId,ZeroOp,SuccOp,
                       Id,[TParam],_Pos,Env,expr,Result,TRin,TRout) :-
    unique_id("opresult.",ResultId),
    create_typed_id(ResultId,NatType,TResult),
    create_typed_id(ArgId,integer,TArgId),
    % prepare environment:
    add_identifiers_to_environment([TArgId,TResult],Env,Env1),
    % we create a recursive function by using a (as recursive annotated) comprehension
    % set. We can refer to the compset with RecId.
    create_recursive_compset([TArgId,TResult],Cond,set(couple(integer,NatType)),[],RecId,TCompSet),
    % used for recursive call:
    env_store(RecId,set(couple(integer,NatType)),[],Env1,SubEnv),
    % just type iZero to retrieve its type:
    IZero=extended_expr(none,ZeroOp,[],[]),
    btype(make_inductive_natural,IZero,SubEnv,NatType,_,TRin,TR1),
    % Arg=0 => Result=iZero
    RawCase1=implication(none,
                         equal(none,identifier(none,ArgId),integer(none,0)),
                         equal(none,identifier(none,ResultId),IZero)),
    btype(make_inductive_natural,RawCase1,SubEnv,pred,Case1,TR1,TR2),
    % Arg>0 => Result=iSucc( recursive_call(Arg-1) )
    RawCase2=implication(none,
                         greater(none,identifier(none,ArgId),integer(none,0)),
                         equal(none,identifier(none,ResultId),extended_expr(none,SuccOp,[Minus1],[]))),
    Minus1=function(none,identifier(none,RecId),minus(none,identifier(none,ArgId),integer(none,1))),
    btype(make_inductive_natural,RawCase2,SubEnv,pred,Case2,TR2,TRout),
    % {Arg,Result | Arg=0=>Result=iZero & Arg>0=>Result=iSucc( RC(Arg-1) )}
    conjunct_predicates([Case1,Case2],Cond),
    % {Arg,Result | Arg=0=>Result=iZero & Arg>0=>Result=iSucc( RC(Arg-1) )}(TParam)
    create_texpr(function(TCompSet,TParam),NatType,[theory_operator(Id,1)],Result).
