% (c) 2013-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bmachine_static_checks, [static_check_machine/1, toplevel_raw_predicate_sanity_check/2]).

:- use_module(probsrc(module_information)).
:- module_info(group,typechecker).
:- module_info(description,'Static checking of B machines upon construction. Detection of name clashes and uninitalised variables.').

:- use_module(bmachine_structure, [get_section/3]).
:- use_module(b_read_write_info, [get_accessed_vars/4]).
:- use_module(bsyntaxtree, [get_texpr_expr/2,get_texpr_id/2,get_texpr_ids/2]).
:- use_module(error_manager, [add_warning/2, add_warning/3, add_warning/4]).
:- use_module(tools, [ajoin/2]).
:- use_module(library(lists)).

static_check_machine(Machine) :-
    initialises_all_variables(Machine),
    check_name_clashes(Machine).


% ---------------------------------------------
% check for unnatural invariants and properties
% ---------------------------------------------
% the priority of the implication is sometimes surprising
%  a & (x=1) => (y=2) & c   --> top-level symbol is the implication !
% the test is done on the raw predicate before ast_cleanup removes potential typing predicates; see tests 106, 107
toplevel_raw_predicate_sanity_check(invariant,RawPredicate) :-
    check_top_level(RawPredicate,'INVARIANT'),
    fail.
toplevel_raw_predicate_sanity_check(properties,RawPredicate) :-
    check_top_level(RawPredicate,'PROPERTIES'),
    fail.
toplevel_raw_predicate_sanity_check(_,_).

check_top_level(implication(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level implication (=>) in clause: ',SECT).
check_top_level(disjunct(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level disjunction (or) in clause: ',SECT).
check_top_level(equivalence(_,_,_),SECT) :-
    add_warning(static_check_machine,'Top-level equivalence (<=>) in clause: ',SECT).

    
% ---------------------
% Checks is all variables are initialised by the machine
% ---------------------
initialises_all_variables(Machine) :-
    get_section(initialisation,Machine,Initialisation),
    % check wich variables are read / modified by the initialisation
    get_accessed_vars(Initialisation,[],Modifies,_Reads),
    get_machine_variables(Machine,SortedAllVars),
    % check for each variable, if the initialisation modifies it
    exclude(is_initialised(Modifies),SortedAllVars,Uninitialised),
    % generate a warning if unitialised is not empty
    generate_uninitialised_warning(Uninitialised),
    % now check order of initialisation sequences
    check_initialisation_order(Initialisation,SortedAllVars,[],_).

get_machine_variables(Machine,SortedAllVars) :-
    % get all variables that should be initialised
    get_section(abstract_variables,Machine,AbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    append(AbsVars,ConcVars,TAllVars),
    get_texpr_ids(TAllVars,AllVars),
    sort(AllVars,SortedAllVars).

is_initialised(Modifies,Id) :-
    memberchk(Id,Modifies).

generate_uninitialised_warning([]) :- !.
generate_uninitialised_warning(Vars) :-
    ajoin(['Machine may not initialise some of its variables: ', Vars], Msg),
    add_warning(bmachine_static_checks,Msg).


:- use_module(library(ordsets)).
% Check if order of sequential compositions in INITIALISATION is ok
% TO DO: add support for if(LIST) and while
check_initialisation_order(b(choice([First|T]),subst,Info),AllVars,AlreadyInit,OutInit) :- !,
    check_initialisation_order(First,AllVars,AlreadyInit,OutInit), % we pick the output of the first choice 
    (T=[] -> true ; check_initialisation_order(b(choice(T),subst,Info),AllVars,AlreadyInit,_)).
check_initialisation_order(b(parallel([First|T]),subst,Info),AllVars,AlreadyInit,OutInit) :- !,
    check_initialisation_order(First,AllVars,AlreadyInit,OutInit1), % we pick the output of the first choice 
    (T=[] -> OutInit=OutInit1
      ; check_initialisation_order(b(parallel(T),subst,Info),AllVars,AlreadyInit,OutInitRest),
        ord_union(OutInit1,OutInitRest,OutInit)
    ).
check_initialisation_order(b(init_parallel(S),subst,Info),AllVars,AlreadyInit,OutInit) :- !,
    check_initialisation_order(b(parallel(S),subst,Info),AllVars,AlreadyInit,OutInit).
check_initialisation_order(b(sequence([First|T]),subst,Info),AllVars) --> !,
    check_initialisation_order(First,AllVars),
    ({T=[]} -> [] ; check_initialisation_order(b(sequence(T),subst,Info),AllVars)).
% TO DO: also check if-then-else, while, ...
check_initialisation_order(Subst,AllVars,AlreadyInit,OutInit) :-
    % print('CHCK '), translate:print_subst(Subst),nl,
    get_accessed_vars(Subst,[],Modifies,Reads), 
    ord_union(Modifies,AlreadyInit,OutInit), % after Subst we have initialised OutInit
    % below is already checked by type checker:
    %ord_subtract(Modifies,AllVars,IllegalAssignments),
    %(IllegalAssignments=[] -> true
    % ; add_warning(bmachine_static_checks,'INITIALISATION writes illegal variables: ',IllegalAssignments,Subst)),
    ord_intersection(Reads,AllVars,ReadFromSameMachine),
    ord_subtract(ReadFromSameMachine,OutInit,IllegalReads), % we use OutInit: there could be an if with sequence inside
    (IllegalReads=[] -> true
     ; add_warning(bmachine_static_checks,'INITIALISATION reads variables which are not yet initialised: ',IllegalReads,Subst)).

% ---------------------
% Checks if an operations parameter or local variable clashes with a variable
% ---------------------
check_name_clashes(Machine) :-
    debug:debug_println(10,'Checking for name clashes'),
    % get all variables and constants that might clash
    get_section_ids(abstract_variables,Machine,'variable',AbsVars),
    get_section_ids(concrete_variables,Machine,'variable',ConcVars),
    get_section_ids(abstract_constants,Machine,'constant',AbsCons),
    get_section_ids(concrete_constants,Machine,'constant',ConcCons),
    get_section_ids(deferred_sets,Machine,'set',DefSets),
    get_section_ids(enumerated_sets,Machine,'set',EnumSets),
    get_section_ids(enumerated_elements,Machine,'set element',EnumElems),
    % enumerated  b_get_named_machine_set(GlobalSetName,ListOfConstants) + b_get_machine_set(S)
    append([AbsVars,ConcVars,AbsCons,ConcCons,DefSets,EnumSets,EnumElems],AllIds),
    sort(AllIds,SortedAllIds),
    %print(SortedAllIds),nl,nl,
    % for each operation, compare the parameter names with existing vars/constants
    get_section(operation_bodies,Machine,Operations),
    maplist(op_name_clashes(SortedAllIds),Operations),
    get_section(definitions,Machine,Defs),
    maplist(def_name_clashes(SortedAllIds),Defs).

get_section_ids(Section,Machine,Class,ResultList) :-
    get_section(Section,Machine,Vars),
    findall(machine_id(ID,Class),(member(TID,Vars),get_texpr_id(TID,ID)),ResultList).
   

:- use_module(bsyntaxtree,[map_over_typed_bexpr/2]).
:- use_module(external_functions,[is_external_function_name/1]).

def_name_clashes(AllIds,definition_decl(Name,DefType,Pos,Args,_RawExpr,_Deps)) :- !,
    findall(ID,member(identifier(Pos,ID),Args),ArgIds),  % args can sometimes not be identifiers; see test 1711
    debug:debug_println(4,checking_def(Name,DefType,ArgIds)),
    % exclude external functions/predicates/...; 
    % EXTERNAL_ declarations do not seem to be included; TO DO: check whether external function actually declared
    (is_external_function_name(Name) -> true
     ; include(clash_warnings('DEFINITION parameter',AllIds,Name),ArgIds,_ArgsCausingWarning)
       % TO DO: check Body; for this we need a map_over_raw_expression to detect local variables introduced !
     ).
def_name_clashes(_,D) :- print(unknown_def(D)),nl.



op_name_clashes(AllIds,Operation) :-
    get_texpr_expr(Operation,operation(IdFull,Results,Params,Subst)),
    get_texpr_id(IdFull,op(Id)),
    get_texpr_ids(Results,ResultIds), get_texpr_ids(Params,ParamIds),
    include(clash_warnings('Operation parameter',AllIds,Id),ParamIds,_ParamsCausingWarning),
    include(clash_warnings('Operation result variable',AllIds,Id),ResultIds,_ResultsCausingWarning),
    (map_over_typed_bexpr(bmachine_static_checks:local_variable_clash(AllIds,Id),Subst),fail ; true).

:- public local_variable_clash/3.
local_variable_clash(AllIds,Operation,TSubst) :-
   get_texpr_expr(TSubst,Subst),
   introduces_local_variable(Subst,ID),
   %print(check(ID,Operation)),nl,
   clash_local_warnings(AllIds,Operation,ID,TSubst).

% check for constructs which introduced local variables
introduces_local_variable(var(Parameters,_),ID) :-   % currently B-interpreter cannot correctly deal with this in the context of operation_call 
   member(TID,Parameters), get_texpr_id(TID,ID).

% ord_member does not work below because of free variable Class
my_ord_member(Name,Class,[machine_id(Name1,_)|T]) :-
    Name @> Name1, !, 
    my_ord_member(Name,Class,T).
my_ord_member(Name,Class,[machine_id(Name,Class)|_]).   

clash_warnings(Context,AllIds,OperationId,Name) :-
    my_ord_member(Name,Class,AllIds), !,
    ajoin([Context, ' "', Name, '" has the same name as a ', Class, ' in operation "', OperationId,'".'], Msg),
    add_warning(bmachine_static_checks,Msg).

clash_local_warnings(AllIds,OperationId,Name,Pos) :-
    my_ord_member(Name,Class,AllIds), !,
    ajoin(['Local variable has the same name as a ', Class, ' in operation ', OperationId,': '], Msg),
    add_warning(bmachine_static_checks,Msg,Name,Pos).

% TODO: this does not and can not work here:
% - Some preconditions are removed (typing only....) during machine simplification
% - Needs to be verified during typechecking
% ---------------------
% Checks if an operations parameter is not typed by a pre condition
% ---------------------
%parameters_without_pre_condition(Machine) :-
%    get_section(operation_bodies,Machine,Operations),
%    maplist(parameters_without_pre_condition_aux,Operations).
%
%parameters_without_pre_condition_aux(Operation) :-trace,
%    get_texpr_expr(Operation,operation(IdFull,_Results,Params,Subst)),
%    get_texpr_id(IdFull,op(Id)),
%    (Params == []
%     -> true % no parameters
%     ;  (get_texpr_expr(Subst,precondition(_,_))
%         -> true % parameters and precondition
%         ;  ajoin(['Operation ', Id, ' has parameters but no pre-condition.'], Msg),
%            add_warning(bmachine_static_checks,Msg)
%          )).