% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(bvisual2,[ reset_bvisual/0
                   , bv_get_top_level/1
                   , bv_expand_formula/3
                   , bv_get_values/3
                   , bv_get_values/2
                   , bv_insert_formula/3
                   , bv_include_variables/0
                   , bv_get_value_unlimited/3
                   , bv_is_explanation_node/1
                   , bv_print_to_file/3
                   , bv_write_all_variables_and_constants/2
                   , set_translation_mode/1
                   , bv_get_top_level_formula/4
                   , bv_get_top_level_formula/5
                   , bv_value_to_atom/2
                   ]).

:- use_module(library(lists)).
:- use_module(library(codesio)).
%:- use_module(library(timeout)).

:- use_module(bmachine,      [ b_machine_has_constants_or_properties/0
                             , b_get_machine_variables/1
                             , b_get_machine_constants/1
                             , b_get_machine_operation/4
                             , b_get_properties_from_machine/1
                             , b_get_invariant_from_machine/1
                             , b_get_static_assertions_from_machine/1
                             , b_get_assertions_from_main_machine/2
                             , b_get_dynamic_assertions_from_machine/1
                             , b_get_assertion_count/3
                             ]). 
:- use_module(specfile,      [ state_corresponds_to_initialised_b_machine/2
                             , state_corresponds_to_set_up_constants/2
                             ]).
:- use_module(state_space,   [ visited_expression/2
                             , current_state_id/1
                             , invariant_not_yet_checked/1
                             , invariant_violated/1
                             , transition/3
                             ]).
:- use_module(store,         [ empty_state/1
                             , normalise_store/2
                             ]).
:- use_module(b_interpreter, [ b_test_boolean_expression_wf/3
                             , b_not_test_boolean_expression_wf/3
                             , b_test_boolean_expression/4
                             , b_compute_expression/4
                             , set_up_typed_localstate/6
                             ]).
:- use_module(translate,     [ translate_subst_or_bexpr_in_mode/3
                             , translate_bexpression_to_codes/2
                             %, translate_bvalue/2
                             , translate_bvalue_for_expression/3
                             , translate_bvalue/2
                             ]).
:- use_module(bsyntaxtree,   [ get_texpr_expr/2
                             , get_texpr_type/2
                             , get_texpr_info/2
                             %, get_texpr_ids/2
                             , is_texpr/1
                             , create_texpr/4
                             , safe_create_texpr/3
                             , syntaxtraversion/6
                             , conjunction_to_list/2
                             , conjunct_predicates/2
                             , create_exists/3, create_forall/3
                             %, find_identifier_uses/3
                             ]).
:- use_module(b_ast_cleanup, [ clean_up_pred/3
                             ]).
:- use_module(error_manager, [ get_all_errors_and_reset/1
                             , time_out_with_enum_warning_one_solution/3
                             ]).
:- use_module(preferences,   [ get_preference/2
                             , set_preference/2
                             ]).
:- use_module(kernel_waitflags, [ init_wait_flags/1
                                , ground_wait_flags/1]).
:- use_module(b_enumerate,   [ b_tighter_enumerate_values/2 ]).
:- use_module(tools,         [ put_codes/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,analysis).
:- module_info(description,'This module contains functionality to analyse B expressions and predicates by evaluating and decomposing them into substructures.').

:- dynamic top_level_node/1, subnode/2, supernode/2.
:- dynamic stored_formula/2, expanded/1, explanation_parent/1.
:- dynamic id_counter/1.
:- dynamic formula_cache/3.
% We store the information whether a node can introduce new values as an (counter-)example
% in example_node/1; the value is computed in explore_node/2.
% For such nodes, the found examples are stored in local_state/3.
:- dynamic example_node/1,local_state/3.


bv_value_to_atom(p(P),R) :- !, % a predicate value
   (P=true -> R='TRUE' ; P=false -> R='FALSE' ; R=P).
bv_value_to_atom(e(_),R) :- !, R='ERROR'.
bv_value_to_atom(v(V),R) :- !, R=V. % a value, already pretty-printed !??
bv_value_to_atom(i,R) :- !, R='INACTIVE'. 
bv_value_to_atom(V,V).


:- dynamic variables_should_be_included/0.
bv_include_variables :-
    ( variables_should_be_included -> true
    ; otherwise -> assert(variables_should_be_included)).

% This predicate should be called after loading a new specification
% and before using the other predicates below
reset_bvisual :-
    retractall(top_level_node(_)),
    retractall(subnode(_,_)),
    retractall(supernode(_,_)),
    retractall(stored_formula(_,_)),
    retractall(expanded(_)),
    retractall(explanation_parent(_)),
    retractall(formula_cache(_,_,_)),
    retractall(example_node(_)),
    retractall(local_state(_,_,_)),
    register_top_level.

% returns the top-level nodes as a list of IDs
bv_get_top_level(Tops) :-
    findall( Id, top_level_node(Id), Tops ).

% bv_expand_formula(+FormulaID,-LabelAtom,-)
% Input: Id
% Output: Label of Formula, list of children
bv_expand_formula(Id,Label,Children) :-
    expanded(Id),!,
    get_node_label(Id,Label),
    findall(C, subnode(Id,C), Children).
bv_expand_formula(Id,Label,Children) :-
    get_node_label(Id,Label),
    explore_node(Id,Children).

% bv_get_values in current state
bv_get_values(Ids,Values) :-
   current_state_id(StateId),
   bv_get_values(Ids,StateId,Values).

% bv_get_values(+IdsOfFormulas,+CurrentStateID,-ValuesOfFormulas)
bv_get_values(Ids,StateId,Values) :-
    visited_expression(StateId,State),!,
    get_preference(time_out,Timeout),
    bv_get_values2(Ids,State,StateId,Timeout,Values).
bv_get_values(Ids,_StateId,Values) :-
    % in case that the state ID cannot be resolved, we
    % return an error for each formula
    same_length(Ids,Values),
    same_value(Values,e('unkown state')).
same_value([],_).
same_value([V|Rest],V) :- same_value(Rest,V).


% Insert a new formula
% bv_insert_formula(+TypeCheckedExpression,+ParentID,-IDofNewFormula)
bv_insert_formula(TExpr,ParentId,Id) :-
    get_new_id(Id),
    assert( stored_formula(Id,TExpr) ),
    assert( subnode(ParentId,Id) ),
    assert( supernode(Id,ParentId) ),
    ( ParentId == top ->
        assert( top_level_node(Id) )
    ; otherwise ->
        true).


:- meta_predicate bv_time_out_call(0,-,-,-).
bv_time_out_call(Call,ValueFromCall,Timeout,Value) :- 
    (time_out_with_enum_warning_one_solution( Call,
              Timeout,
              Result)
     ->
        ( Result == success -> Value = ValueFromCall
        ; Result = virtual_time_out(failure_enumeration_warning(_Info,_,_,_,critical)) -> Value = e('?(\x221E\)') % Value = e('\x22A5\?(\x221E\)') % 8734 in decimal % Feedback for user: increasing MAXINT,... could mean ProB can find a solution; However, increasing TIMEOUT value will not help
        ; Result =  virtual_time_out(_) -> Value = e('?(\x221E\)')  % infinity symbol: Feedback to the user: probably no way to solve the issue apart from ensuring that set comprehensions are finite ...
        ; otherwise         -> Value = e(timeout))
    ).

bv_get_values2([],_,_,_,[]).
bv_get_values2([Id|Irest],State,StateId,Timeout,[Value|Vrest]) :-
    bv_get_value(Id,State,StateId,Timeout,Value),
    bv_get_values2(Irest,State,StateId,Timeout,Vrest).
bv_get_value(Id,State1,StateId,Timeout,FinalValue) :-
    bv_get_value_unprocessed(Id,State1,StateId,Timeout,Value),
    value_post_processing(Value,FinalValue).
bv_get_value_unprocessed(Id,State1,StateId,Timeout,Value) :-
    is_active(Id,StateId,State1,State,LocalState),!,
    ( is_cached(Id,StateId,State1,Value1) ->
        FromCache = true
    ; bv_get_value1(Id,State,LocalState,Timeout,Value1) ->
        FromCache = false
    ; otherwise ->
        Value1 = e('evaluation failed'),
        FromCache = false),
    handle_errors(Value1,Value),
    ( FromCache==false, should_be_cached(Id) ->
        write_to_cache(Id,StateId,State1,Value)
    ; otherwise -> true).
bv_get_value_unprocessed(_Id,_State1,_StateId,_Timeout,i). % INACTIVE 
bv_get_value1(Id,State,LocalState,Timeout,Value) :-
    stored_formula(Id,Formula),
    bv_get_value2(Formula,State,LocalState,Timeout,Value).
bv_get_value2(Formula,State,LocalState,Timeout,Value) :-
    is_texpr(Formula),!,
    get_texpr_type(Formula,Type),
    bv_get_value3(Type,Formula,LocalState,State,Timeout,Value).
bv_get_value2(bind(_ID,Value),_,_,_,btvalue(Value)).
bv_get_value2(textnode(Value,_),_State,_LS,_,v(Value)).
%bv_get_value2(cbc_path(_LastStateID,Path,_Last),_State,_LS,_,v(NrOfPaths)) :-
%   append(Path,_,Prefix),
%   findall(Last,sap:cb_path(_,Prefix,Last),AllPathsWithPrefix),
%   length(AllPathsWithPrefix,NrOfPaths).
bv_get_value2(variant(_Name,_ConvOrAnt,Variant),State,LocalState,Timeout,Value) :-
    bv_get_value2(Variant,State,LocalState,Timeout,Value).
bv_get_value2(guard(_Name,Parameters,Guard),State,LocalState,Timeout,Value) :- !,
    ( Parameters = [] -> Expr = Guard
    ; otherwise       -> create_exists(Parameters,Guard,Expr)),
    bv_get_value2(Expr,State,LocalState,Timeout,Value).
bv_get_value2(guard_theorems(_Name,Parameters,Guard,Theorems),State,LocalState,Timeout,Value) :- !,
    ( Parameters = [] -> Expr = Theorems
    ; otherwise       -> create_forall(Parameters,b(implication(Guard,Theorems),pred,[]),Expr)),
    bv_get_value2(Expr,State,LocalState,Timeout,Value).
bv_get_value3(pred,Formula,LocalState,State,Timeout,Value) :-
    !,
    % print('CHECKING PREDICATE: '), translate:print_bexpr(Formula),nl,
    ( bv_time_out_call(b_test_boolean_expression_wf(Formula,LocalState,State),true,Timeout,ValuePos) -> true
    ; otherwise -> ValuePos=false),
    ( get_preference(double_evaluation_when_analysing,true) ->
        ( bv_time_out_call(b_not_test_boolean_expression_wf(Formula,LocalState,State),true,Timeout,ValueNeg) -> true
        ; otherwise -> ValueNeg=false),
        combine_predicate_values(ValuePos,ValueNeg,Value)
    ; otherwise -> encode_predicate_value(ValuePos,Value)).
bv_get_value3(_,Formula,LocalState,State,Timeout,Res) :-
    bv_time_out_call(b_compute_expression(Formula,LocalState,State,FValue),FValue,Timeout,Value),!,
    ( Value = e(E) -> Res = e(E)
    ; otherwise    -> Res = btvalue(Value,Formula)
    ).


encode_predicate_value(e(Error),e(Error)).
encode_predicate_value(true,p(true)).
encode_predicate_value(false,p(false)).
combine_predicate_values(e(_Error),true,p(false)) :- print('### Ignoring Timeout in positive case'),nl.
combine_predicate_values(e(_Error),false,p(false)) :- print('### Ignoring Timeout in positive case'),nl.
combine_predicate_values(e(Error),e(_),e(Error)).
combine_predicate_values(true,false,p(true)).
combine_predicate_values(false,true,p(false)).
combine_predicate_values(true,true,e('both true and false')).
combine_predicate_values(false,false,e(undefined)).
combine_predicate_values(true,e(_Error),p(true)) :- print('### Ignoring Timeout in negative case'),nl.
combine_predicate_values(false,e(_Error),p(false)) :- print('### Ignoring Timeout in negative case'),nl.

value_post_processing(In,Out) :-
    ( value_post_processing2(In,O) -> O=Out
    ; otherwise -> Out=e('internal error: failed to post-process value')).
value_post_processing2(v(Value),v(Value)).
value_post_processing2(p(Value),p(Value)).
value_post_processing2(e(Error),e(Error)).
value_post_processing2(i,i).

value_post_processing2(btvalue(BValue,Expr),v(Value)) :-
    translate_bvalue_for_expression(BValue,Expr,Value).
value_post_processing2(btvalue(BValue),v(Value)) :-
    translate_bvalue(BValue,Value).
    
    


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% caching of some values

% we do not compute the theorems node, only its subnodes
is_cached(theoremsc,_,_,v('')) :- !.
is_cached(theoremscm,_,_,v('')) :- !.
is_cached(theoremsv,_,_,v('')) :- !.
is_cached(theoremsvm,_,_,v('')) :- !.
is_cached(goal,_,_,v('')) :- !.
% we use the invariant_violated flag of the state space as a kind of cache
is_cached(inv,StateId,_State,Value) :-
    \+ invariant_not_yet_checked(StateId),
    ( invariant_violated(StateId) -> Value=p(false)
    ; otherwise                   -> Value=p(true)).
% axioms are usually true in the constants' state
is_cached(axioms,StateId,concrete_constants(State),R) :- !,
    % only if a partial_setup_constants event led to that
    % state, then something did not work
    ( transition(root,'$partial_setup_constants',StateId) ->
        (member(bind(_,abort(_)),State)
          -> R = e('some constants have no values')
          ; animation_minor_mode(eventb) -> R = e('some axioms may be false')
          ; otherwise -> R = e('some properties may be false')
        )
    ; otherwise ->
        R = p(true)).
% this formula was already computed and cached for the given state
is_cached(FormulaId,StateId,_State,Value) :-
    formula_cache(FormulaId,StateId,Value),!.
% in case of axiom, we can look at the reverences constants' state
is_cached(FormulaId,_,const_and_vars(ConstId,_),Value) :-
    is_axiom_or_static_theorem(FormulaId),
    visited_expression(ConstId,ConstState),
    is_cached(FormulaId,ConstId,ConstState,Value),!.

should_be_cached(FormulaId) :-
    is_axiom_or_static_theorem(FormulaId).

write_to_cache(FormulaId,_StateId,const_and_vars(ConstId,_),Value) :-
    is_axiom_or_static_theorem(FormulaId),!,
    visited_expression(ConstId,ConstState),    
    write_to_cache(FormulaId,ConstId,ConstState,Value).
write_to_cache(FormulaId,StateId,_State,Value) :-
    assert(formula_cache(FormulaId,StateId,Value)).

is_axiom_or_static_theorem(axioms) :- !.
is_axiom_or_static_theorem(theoremsc) :- !.
is_axiom_or_static_theorem(theoremscm) :- !.
is_axiom_or_static_theorem(FormulaId) :-
    supernode(FormulaId,Parent),is_axiom_or_static_theorem(Parent).

handle_errors(_,e(ErrorMsg)) :-
    get_all_errors_and_reset([Error|_]),!,
    write_to_codes(Error,ErrorCodes),
    limit_error(ErrorCodes,ErrorCodes2),
    atom_codes(ErrorMsg,ErrorCodes2).
handle_errors(Value,Value).

limit_error(ErrorCodes,ErrorCodes) :-
    length(ErrorCodes,L),L < 200,!.
limit_error(In,Out) :-
    prefix_length(In,Start,200),
    append(Start,"...",Out).

% is_active(+FormulaID,+InState,-PossiblyExpandedState)
% says if a FormulaID makes sense for the current state, if not: it will be grayed out
is_active(user,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(inv,_StateId,InState,OutState,LocalState) :- !,
    empty_state(LocalState),
    state_corresponds_to_initialised_b_machine(InState,OutState).
is_active(axioms,_StateId,InState,OutState,LocalState) :- !,
    InState \= root,
    empty_state(LocalState),
    state_corresponds_to_set_up_constants(InState,OutState),
    !.
is_active(variants,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(theoremsc,StateId,InState,OutState,LocalState) :- !,
    is_active(axioms,StateId,InState,OutState,LocalState).
is_active(theoremscm,StateId,InState,OutState,LocalState) :- !,
    is_active(axioms,StateId,InState,OutState,LocalState).
is_active(theoremsv,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(theoremsvm,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(guards,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(guard_theorems,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(variables,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
is_active(constants,StateId,InState,OutState,LocalState) :- !,
    is_active(axioms,StateId,InState,OutState,LocalState).
is_active(sets,_StateId,InState,OutState,LocalState) :- !,
    empty_state(LocalState), InState=OutState.
is_active(goal,StateId,InState,OutState,LocalState) :- !,
    is_active(inv,StateId,InState,OutState,LocalState).
%is_active(cbc_tests,_StateId,_InState,_OutState,_LocalState) :- !,
%    sap:cb_path(_,[_],_). % show cbc_tests if we ran CBC test-case generator
is_active(NodeId,StateId,InState,OutState,LocalState) :-
    supernode(NodeId,Super),
    is_active(Super,StateId,InState,OutState,InLocalState),
    % check if the parent node introduces example values (e.g. in a exists-clause)
    % If yes, the local state is extended by the values
    ( example_node(Super) ->
        find_example(Super,StateId,OutState,InLocalState,LocalState)
    ; otherwise ->
        InLocalState = LocalState).

find_example(NodeId,StateId,_State,_InLocalState,LocalState) :-
    % check if the value was computed and cached before
    local_state(NodeId,StateId,Result),!,
    % if Result is not of the form example/1, no example
    % was found. Just fail to show inactive state
    Result = example(LocalState).
find_example(NodeId,StateId,State,InLocalState,LocalState) :-
    % how the example is computed depends on the expression (e.g. exists)
    % and the outcome of the parent node (true or false)
    bv_get_values([NodeId],StateId,[V]),!,
    stored_formula(NodeId,Formula),
    ( find_example1(Formula,V,State,InLocalState,LocalState) ->
        Result = example(LocalState)
    ; otherwise ->
        Result = not_found),
    % currently we just ignore errors
    handle_errors(_,_),
    assert( local_state(NodeId,StateId,Result) ),
    Result = example(_).
find_example1(TExpr,V,State,InLocalState,LocalState) :-
    is_texpr(TExpr),!,
    get_texpr_expr(TExpr,Expr),
    find_example2(Expr,V,State,InLocalState,LocalState).
find_example1(guard(_Name,Parameters,Guard),V,State,InLocalState,LocalState) :-
    create_exists(Parameters,Guard,Exists),
    find_example1(Exists,V,State,InLocalState,LocalState).
find_example1(guard_theorems(_Name,Parameters,Guard,Theorems),V,State,InLocalState,LocalState) :-
    create_forall(Parameters,b(implication(Guard,Theorems),pred,[]),ForAll),
    find_example1(ForAll,V,State,InLocalState,LocalState).
find_example2(exists(Ids,Predicate),p(true),State,InLocalState,LocalState) :-
    find_solution_for_predicate(Ids,Predicate,State,InLocalState,LocalState).
find_example2(exists(Ids,Predicate),p(false),State,InLocalState,LocalState) :-
    safe_create_texpr(negation(Predicate),pred,Negation),
    find_solution_for_predicate(Ids,Negation,State,InLocalState,LocalState).
find_example2(forall(Ids,Pre,Condition),p(false),State,InLocalState,LocalState) :-
    safe_create_texpr(negation(Condition),pred,Negation),
    conjunct_predicates([Pre,Negation],Predicate),
    find_solution_for_predicate(Ids,Predicate,State,InLocalState,LocalState).
find_example2(forall(Ids,Pre,Condition),p(true),State,InLocalState,LocalState) :-
    conjunct_predicates([Pre,Condition],Predicate),
    find_solution_for_predicate(Ids,Predicate,State,InLocalState,LocalState).
find_solution_for_predicate(Ids,Predicate,State,InLocalState,NormedLocalState) :-
    get_preference(time_out,Timeout),
    set_up_typed_localstate(Ids,_FreshVars,TypedVals,InLocalState,LocalState,positive),
    (time_out_with_enum_warning_one_solution( (init_wait_flags(WF),
                  b_tighter_enumerate_values(TypedVals,WF),
                  b_test_boolean_expression(Predicate,LocalState,State,WF),
                  ground_wait_flags(WF)),
              Timeout,Result)
     -> Result == success,
        normalise_store(LocalState,NormedLocalState)
    ).


explore_node(Id,Children) :-
    stored_formula(Id,Formula),
    check_if_example_node(Id,Formula),
    get_subformulas(Formula,Subs,Kind),
    assert( expanded(Id) ),
    (Kind = explanation
     -> assert_as_explanation(Id,ExplId), Children = [ExplId|Children1],
        register_formulas(Subs,Id,Children1)
     ; register_formulas(Subs,Id,Children)).

assert_as_explanation(Parent,Id) :- 
    assert(explanation_parent(Parent)),
    % add a text line; showing that we did an equivalence rewrite
    get_new_id(Id),
    assert( stored_formula(Id,textnode('\x21D4\',[])) ), % unicode translation
    assert( subnode(Parent,Id) ),
    assert( supernode(Id,Parent) ).
    
bv_is_explanation_node(NodeId) :-
   supernode(NodeId,ParentId),
   explanation_parent(ParentId).

% the following predicates determine whether the node introduces
% example values and stores that information in example_node/1.
check_if_example_node(Id,Formula) :-
    is_example_node(Formula),!,
    assert( example_node(Id) ).
check_if_example_node(_Id,_Formula).
is_example_node(TFormula) :-
    is_texpr(TFormula),!,
    get_texpr_expr(TFormula,Formula),
    is_example_node2(Formula).
is_example_node(guard(_Name,[_|_],_Guard)).
is_example_node(guard_theorems(_Name,[_|_],_Guard,_Theorems)).
is_example_node2(exists(_Ids,_Cond)).
is_example_node2(forall(_Ids,_Pre,_Cond)).

% subformula_rule/4 defines wich subformulas should be
%  shown in an evaluation tree for a formula. If no
%  rule matches, the standard subformulas are used
subformula_rule(equal(A,B),TF,[Sub1,Sub2],explanation) :-
    get_texpr_type(A,T),T=set(_),
    is_not_bvexpr(TF),
    no_empty_set(A),no_empty_set(B),!,
    create_bvexpr(set_subtraction(A,B),T,AminusB),
    create_bvexpr(set_subtraction(B,A),T,BminusA),
    create_bvexpr(empty_set,T,Empty),
    create_bvexpr(equal(AminusB,Empty),pred,Sub1),
    create_bvexpr(equal(BminusA,Empty),pred,Sub2).
subformula_rule(subset(A,B),_,[Equals],explanation) :-
    subset_rule(A,B,Equals).
subformula_rule(subset_strict(A,B),_,[AmBEmpty,NotEquals],explanation) :-
    subset_rule(A,B,AmBEmpty),
    get_texpr_type(A,T),
    create_bvexpr(set_subtraction(B,A),T,BminusA),
    create_bvexpr(empty_set,T,Empty),
    create_bvexpr(not_equal(BminusA,Empty),pred,NotEquals).
subformula_rule(not_subset(A,B),_,Subs,Kind) :-
    subformula_rule(subset(A,B),_,Subs,Kind). % it is an equivalent rewrite to the negation
subformula_rule(not_subset_strict(A,B),_,Subs,Kind) :-
    subformula_rule(subset_strict(A,B),_,Subs,Kind). % it is an equivalent rewrite to the negation
subformula_rule(forall(Ids,P,R),_,Subs,quantifier) :-
    append(Ids,[P,R],Subs).
subformula_rule(exists(Ids,P),_,Subs,quantifier) :-
    append(Ids,[P],Subs).
subformula_rule(conjunct(A,B),_,Subs,normal) :-
    safe_create_texpr(conjunct(A,B),pred,E),
    conjunction_to_list(E,Subs).
subformula_rule(disjunct(A,B),_,Subs,normal) :-
    disjunction_to_list(A,B,Subs).
subformula_rule(member(TM,TS),_,Children,explanation) :-
    get_texpr_expr(TS,S),
    subformula_member_rule(S,TM,TS,Children),!.

subset_rule(A,B,Equals) :-
    get_texpr_type(A,T),
    create_bvexpr(set_subtraction(A,B),T,AminusB),
    create_bvexpr(empty_set,T,Empty),
    create_bvexpr(equal(AminusB,Empty),pred,Equals).
    
:- use_module(bsyntaxtree,[is_just_type/1]).

subformula_member_rule(partial_function(TDom,TRan),TM,TS,[TDoubles]) :-
    bsyntaxtree:is_just_type(TDom),
    bsyntaxtree:is_just_type(TRan),
    !,
    % this rule describes the plain function check, without
    % checking domains
    texpr_function_type(TS,FunctionType,DomType,RanType),
    create_texpr(identifier(d_1),DomType,[],TDomId),
    create_texpr(identifier(r_1),RanType,[],TRanId1),
    create_texpr(identifier(r_2),RanType,[],TRanId2),
    create_texpr(comprehension_set([TDomId,TRanId1],TExists),FunctionType,[],TDoubles),
    % create_exists([TRanId2],TPred,TExists1), % moved below so that used ids can be computed
    create_texpr(couple(TDomId,TRanId1),couple(DomType,RanType),[],TCouple1),
    create_texpr(couple(TDomId,TRanId2),couple(DomType,RanType),[],TCouple2),
    safe_create_texpr(not_equal(TRanId1,TRanId2),pred,Unequal),
    create_texpr(member(TCouple1,TM),pred,[],TMember1),
    create_texpr(member(TCouple2,TM),pred,[],TMember2),
    conjunct_predicates([Unequal,TMember1,TMember2],TPred),
    create_exists([TRanId2],TPred,TExists1), % now also computes used identifiers; cleanup no longer required ?
    % This is necessary to register the used identifiers
    clean_up_pred(TExists1,[],TExists).
subformula_member_rule(partial_function(TA,TB),TM,TS,[TFunCheck|SetChecks]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_domain_range_checks(TDomain,TRange,TA,TB,SetChecks).
subformula_member_rule(total_function(TA,TB),TM,TS,[TFunCheck|SetChecks]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_total_domain_range_checks(TDomain,TRange,TA,TB,SetChecks).
subformula_member_rule(partial_injection(TA,TB),TM,TS,[TFunCheck,TInjCheck|SetChecks]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_injection_check(TM,TS,TInjCheck),
    create_domain_range_checks(TDomain,TRange,TA,TB,SetChecks).
subformula_member_rule(total_injection(TA,TB),TM,TS,[TFunCheck,TInjCheck|SetChecks]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_injection_check(TM,TS,TInjCheck),
    create_total_domain_range_checks(TDomain,TRange,TA,TB,SetChecks).
subformula_member_rule(partial_surjection(TA,TB),TM,TS,Children) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_texpr(equal(TRange,TB),pred,[],TRanCheck),
    create_optional_subset_check(TDomain,TA,DomChecks),
    append([TFunCheck|DomChecks],[TRanCheck],Children).
subformula_member_rule(total_surjection(TA,TB),TM,TS,[TFunCheck,TDomCheck,TRanCheck]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_texpr(equal(TDomain,TA),pred,[],TDomCheck),
    create_texpr(equal(TRange,TB),pred,[],TRanCheck).
subformula_member_rule(total_bijection(TA,TB),TM,TS,[TFunCheck,TInjCheck,TDomCheck,TRanCheck]) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_injection_check(TM,TS,TInjCheck),
    create_texpr(equal(TDomain,TA),pred,[],TDomCheck),
    create_texpr(equal(TRange,TB),pred,[],TRanCheck).
subformula_member_rule(partial_bijection(TA,TB),TM,TS,Children) :-
    create_function_check(TM,TS,TFunCheck,TDomain,TRange),
    create_injection_check(TM,TS,TInjCheck),
    create_optional_subset_check(TDomain,TA,DomChecks),
    create_texpr(equal(TRange,TB),pred,[],TRanCheck),
    append([TFunCheck,TInjCheck|DomChecks],[TRanCheck],Children).
subformula_member_rule(pow_subset(TA),TM,_TS,[NewCheck]) :-
   % TM : POW(TA)  <=>  TM \ TA = {}
    get_texpr_type(TM,Type),
    create_bvexpr(set_subtraction(TM,TA),Type,Diff),
    create_bvexpr(empty_set,Type,Empty),
    create_bvexpr(equal(Diff,Empty),pred,NewCheck).
% TO DO: add rules for <:, <<:, FIN, POW1, FIN1

create_function_check(TM,TS,TFunCheck,TDomain,TRange) :-
    texpr_function_type(TS,FunctionType,DomType,RanType),
    translate:set_type_to_maximal_texpr(DomType,TDom),
    translate:set_type_to_maximal_texpr(RanType,TRan),
    create_bvexpr(partial_function(TDom,TRan),set(FunctionType),TFun),
    create_bvexpr(member(TM,TFun),pred,TFunCheck),
    create_bvexpr(domain(TM),set(DomType),TDomain),
    create_bvexpr(range(TM),set(RanType),TRange).

create_domain_range_checks(TDomain,TRange,TA,TB,SetChecks) :-
    create_optional_subset_check(TDomain,TA,DomCheck),
    create_optional_subset_check(TRange,TB,RanCheck),
    append(DomCheck,RanCheck,SetChecks).
create_total_domain_range_checks(TDomain,TRange,TA,TB,[TIsTotal|RanCheck]) :-
    create_texpr(equal(TDomain,TA),pred,[],TIsTotal),
    create_optional_subset_check(TRange,TB,RanCheck).
create_optional_subset_check(TSubset,TSuperset,Check) :-
    ( bsyntaxtree:is_just_type(TSuperset) ->
        Check = []
    ; otherwise ->
        create_texpr(subset(TSubset,TSuperset),pred,[],TPred),
        Check = [TPred]).    

create_injection_check(TM,TS,TCheckExpr) :-
    texpr_function_type(TS,FunctionType,DomType,RanType),
    create_texpr(identifier(d_1),DomType,[],TDomId1),
    create_texpr(identifier(d_2),DomType,[],TDomId2),
    create_texpr(identifier(r_1),RanType,[],TRanId),
    create_texpr(comprehension_set([TDomId1,TRanId],TExists),FunctionType,[bv_function_check(TM)],TCheckExpr),
    %create_exists([TDomId2],TPred,TExists1),
    create_texpr(couple(TDomId1,TRanId),couple(DomType,RanType),[],TCouple1),
    create_texpr(couple(TDomId2,TRanId),couple(DomType,RanType),[],TCouple2),
    create_texpr(not_equal(TDomId1,TDomId2),pred,[],Unequal),
    create_texpr(member(TCouple1,TM),pred,[],TMember1),
    create_texpr(member(TCouple2,TM),pred,[],TMember2),
    conjunct_predicates([Unequal,TMember1,TMember2],TPred),
    create_exists([TDomId2],TPred,TExists1), % now also computes used identifiers; cleanup no longer required ?
    % This is necessary to register the used identifiers
    clean_up_pred(TExists1,[],TExists).

texpr_function_type(TFun,FunType,DomType,RanType) :-
    get_texpr_type(TFun,set(FunType)),
    FunType = set(couple(DomType,RanType)).
    
        
create_bvexpr(Expr,Type,TExpr) :-
    create_texpr(Expr,Type,[bvisual],TExpr).
is_not_bvexpr(TExpr) :-
    get_texpr_info(TExpr,Infos),nonmember(bvisual,Infos).
%is_bvexpr(TExpr) :-   get_texpr_info(TExpr,Infos),memberchk(bvisual,Infos).

no_empty_set(E) :- \+ get_texpr_expr(E,empty_set).

get_subformulas(TFormula,Subs,Kind) :-
    is_texpr(TFormula),
    get_texpr_expr(TFormula,Formula),
    subformula_rule(Formula,TFormula,Subs,Kind),!.
get_subformulas(TFormula,Subs,syntaxtraversion) :-
    is_texpr(TFormula),!,
    syntaxtraversion(TFormula,_,_,_,Subs1,Names),
    ( Names = [] ->
        filter_trivial_expressions(Subs1,Subs)
    ; otherwise ->  Subs = []).
get_subformulas(textnode(_,Subs),Subs,textnode) :- !.
get_subformulas(guard(_Name,Parameters,Guard),Subs,guard) :- !,
    conjunction_to_list(Guard,GuardSubs),
    append(Parameters,GuardSubs,Subs).
get_subformulas(variant(_Name,_ConvOrAnt,Variant),Subs,variant) :- !,
    Subs = [Variant].
get_subformulas(guard_theorems(_Name,[],_Guard,Theorems),Subs,guard_theorems) :- !,
    conjunction_to_list(Theorems,Subs).
get_subformulas(guard_theorems(_Name,Parameters,Guard,Theorems),Subs,guard_theorems) :- !,
    append(Parameters,[b(implication(Guard,Theorems),pred,[])],Subs).
%get_subformulas(cbc_path(_,Path,_),Subs) :- !,
%    append(Path,[X],XPath),
%    findall(cbc_path(LastStateID,XPath,X), sap:cb_path(LastStateID,XPath,_), Subs).
get_subformulas(_,[],none).

filter_trivial_expressions([],[]).
filter_trivial_expressions([TI|Irest],Out) :-
    get_texpr_expr(TI,I),get_texpr_info(TI,Info),
    ( is_trivial(I,Info) -> Out = Orest
    ; otherwise          -> Out = [TI|Orest]),
    filter_trivial_expressions(Irest,Orest).

% is_trivial/2 is true for expressions that should not appear as single nodes
%  in the evaluation tree, because they are too simple.
%  The first argument is the expression, the second its information list
is_trivial(integer(_),_).
is_trivial(integer_set(_),_).
is_trivial(empty_set,_).
is_trivial(boolean_true,_).
is_trivial(boolean_false,_).
is_trivial(bool_set,_).
is_trivial(truth,_).
is_trivial(falsity,_).
is_trivial(identifier(_),Info) :-
    memberchk(given_set,Info).
is_trivial(identifier(_),Info) :-
    memberchk(enumerated_set_element,Info).


:- use_module(tools_strings,[ajoin/2]).
:- use_module(specfile,[animation_minor_mode/1, get_specification_description/2]).
get_node_label(inv,invariants) :- animation_minor_mode(eventb),!.
get_node_label(inv,I):- !, get_specification_description(invariant,I).
get_node_label(axioms,axioms) :- animation_minor_mode(eventb),!.
get_node_label(axioms,S):- !, get_specification_description(properties,S).
get_node_label(variants,variants) :- animation_minor_mode(eventb),!.
get_node_label(variants,'VARIANT') :- !. % not used there
get_node_label(theoremsc,'theorems (on constants)') :- animation_minor_mode(eventb),!.
get_node_label(theoremsc,'ASSUME (on constants)') :- animation_minor_mode(tla),!.
get_node_label(theoremsc,'ALL ASSERTIONS (on CONSTANTS)') :- b_get_assertion_count(static,AllNr,MainNr), MainNr < AllNr, !.
get_node_label(theoremsc,'ASSERTIONS (on CONSTANTS)') :- !.
get_node_label(theoremscm,'MAIN ASSERTIONS (on CONSTANTS)') :- !. % should only trigger in B mode at the moment
get_node_label(theoremsv,'theorems (on variables)') :- animation_minor_mode(eventb),!.
get_node_label(theoremsv,'ALL ASSERTIONS (on VARIABLES)') :- b_get_assertion_count(dynamic,AllNr,MainNr), MainNr < AllNr, !.
get_node_label(theoremsv,'ASSERTIONS (on VARIABLES)') :- !. 
get_node_label(theoremsvm,'MAIN ASSERTIONS (on VARIABLES)') :- !. % should only trigger in B mode at the moment
get_node_label(sets,'sets') :- animation_minor_mode(eventb),!.
get_node_label(sets,'SETS').
get_node_label(goal,'GOAL').
get_node_label(variables,'variables') :- animation_minor_mode(eventb),!.
get_node_label(variables,'VARIABLES').
get_node_label(constants,'constants') :- animation_minor_mode(eventb),!.
get_node_label(constants,'CONSTANTS').
get_node_label(guards,'event guards') :- animation_minor_mode(eventb),!.
get_node_label(guards,'OPERATIONS (guards/preconditions)') :- !. % GUARDS/PRE
get_node_label(guard_theorems,'theorems (in guards)') :- !.
%get_node_label(cbc_tests,'CBC_TESTS') :- !.
get_node_label(Id,Label) :-
    stored_formula(Id,Formula),
    get_node_label2(Formula,Label).

get_node_label2(textnode(Text,_),Text) :- !.
get_node_label2(bind(ID,_),ID) :- !.
get_node_label2(guard(Name,_Parameters,_Guard),Name) :- !.
get_node_label2(guard_theorems(Name,_Parameters,_Guard,_Theorems),Name) :- !.
get_node_label2(variant(OpName,ConvOrAnt,_Variant),Name) :- !,
    ajoin([OpName,' |-> ',ConvOrAnt],Name).
get_node_label2(TExpr,Label) :-
    is_texpr(TExpr),!,
    bvisual2_translation_mode(Mode),
    translate_subst_or_bexpr_in_mode(Mode,TExpr,Label).  % use unicode, latex or ascii; but unicode_mode makes Viewer very slow
get_node_label2(Term,Label) :-
    functor(Term,F,Arity),
    ajoin(['unknown node type: ',F,'/',Arity],Label).

:- dynamic bvisual2_translation_mode/1.
bvisual2_translation_mode(unicode). % or ascii or latex
set_translation_mode(X) :- retractall(bvisual2_translation_mode(_)),
   assert(bvisual2_translation_mode(X)).

register_top_level :-
    top_level(Id),
    assert( top_level_node(Id) ),
    assert( expanded(Id) ),
    fail.
register_top_level.

top_level(Id) :-
    top_level2(Id,Formula,Subs),
    register_top_level_formula(Id,Formula,Subs).

:- use_module(specfile,[b_mode/0]).
:- use_module(b_global_sets,[b_global_set/1,b_integer_or_string_set/1,all_elements_of_type/2]).
:- use_module(bmachine,[b_get_machine_goal/1]).
top_level2(variables,textnode('',Variables),Variables) :-
    variables_should_be_included,
    b_get_machine_variables(Variables).
top_level2(constants,textnode('',Constants),Constants) :-
    variables_should_be_included,
    b_get_machine_constants(Constants).
top_level2(sets,textnode('',Sets),Sets) :-
    b_mode,
    findall(bind(S,All),%b(identifier(S),set(global(S)),[]), %b(value(All),set(global(S)),[]),
        top_level_set(S,All),Sets), Sets \= [].
top_level2(goal,Goal,Subs) :- b_mode,
     b_get_machine_goal(Goal),
     conjunction_to_list(Goal,Subs).
top_level2(inv,Invariant,Subs) :- b_mode,
    b_get_invariant_from_machine(Invariant),
    conjunction_to_list(Invariant,Subs).
top_level2(variants,textnode('',Variants),Variants) :- animation_minor_mode(eventb),
        findall(variant(Name,ConvOrAnt,Variant),
            ( b_get_machine_operation(Name,_,_,TBody), get_texpr_expr(TBody,Body),
                Body = rlevent(Name,_Section,Status,
                       _Params,_Guard,_Theorems,_Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),
                get_texpr_expr(Status,S),
                functor(S,ConvOrAnt,1),
                (ConvOrAnt = convergent ; ConvOrAnt = anticipated),
                arg(1,S,Variant) ),
            Variants),
    Variants = [_|_]. 
top_level2(axioms,Props,Subs) :- b_mode,
    b_machine_has_constants_or_properties,
    b_get_properties_from_machine(Props),
    conjunction_to_list(Props,Subs).
top_level2(theoremsc,Pred,Assertions) :- b_mode,
    b_get_static_assertions_from_machine(Assertions),Assertions\=[],
    conjunct_predicates(Assertions,Pred).
top_level2(theoremscm,Pred,Assertions) :- b_mode,
    b_get_assertions_from_main_machine(static,Assertions),
    %Assertions\=[], % in case we have less: show that there are none
    b_get_assertion_count(static,AllNr,MainNr), MainNr < AllNr,
    conjunct_predicates(Assertions,Pred).
top_level2(theoremsv,Pred,Assertions) :- b_mode,
    b_get_dynamic_assertions_from_machine(Assertions),Assertions\=[],
    conjunct_predicates(Assertions,Pred).
top_level2(theoremsvm,Pred,Assertions) :- b_mode,
    b_get_assertions_from_main_machine(dynamic,Assertions),
    %Assertions\=[], % in case we have less: show that there are none
    b_get_assertion_count(dynamic,AllNr,MainNr), MainNr < AllNr,
    conjunct_predicates(Assertions,Pred).
top_level2(guards,textnode('',EventGuards),EventGuards) :- b_mode,
    findall(guard(Name,Params,Guard),
            get_guard(Name,Params,Guard),EventGuards),
    EventGuards = [_|_].    
top_level2(guard_theorems,textnode('',EventGuards),EventGuards) :- b_mode,
    findall(guard_theorems(Name,Params,Guard,GuardTheorems),
            ( b_get_machine_operation(Name,_,_,TBody),
                get_texpr_expr(TBody,Body),
                Body = rlevent(_Name,_Section,_Status,Params,Guard,Theorems,
                              _Actions,_VWitnesses,_PWitnesses,_Unmod,_AbstractEvents),
                Theorems \= [],
                conjunct_predicates(Theorems,GuardTheorems)
            ),
            EventGuards),
    EventGuards = [_|_].
% we are now using a custom tree_inspector for the CBC Tests
/* top_level2(cbc_tests,textnode('',Paths),Paths) :- b_mode,
    get_preference(user_is_an_expert_with_accessto_source_distribution,true),
    %sap:cb_path(_,[_],_), % we have generated at least one test-case
    % TO DO: we need a way to refresh this information after test-cases have been generated
    %  currently the only way seems to close the evaluation view, reload, generate the tests and then open the view
    Paths = [cbc_path(root,[],'INITIALISATION')].
*/

:- use_module(predicate_debugger,[get_unsimplified_operation_enabling_condition/4]).
get_guard(OpName,Params,Guard) :-
    get_unsimplified_operation_enabling_condition(OpName,Params,Guard,_BecomesSuchVars),
    \+ is_initialisation_op(OpName).

is_initialisation_op('$setup_constants').
is_initialisation_op('$initialise_machine').

top_level_set(Set,AllEls) :- b_mode, b_global_set(Set),
   \+ b_integer_or_string_set(Set),
   all_elements_of_type(Set,AllEls).
top_level_set('INT',INTVAL) :- b_mode,
   \+ animation_minor_mode(_), % INT only exists in pure B mode
   get_preference(maxint,MAXINT), get_preference(minint,MININT),
   custom_explicit_sets:construct_interval_closure(MININT,MAXINT,INTVAL).

register_top_level_formula(Id,Formula,Subs) :-
    assert( stored_formula(Id,Formula) ),
    register_formulas(Subs,Id,_).
    
register_formulas([],_Parent,[]).
register_formulas([Formula|Frest],Parent,[SubId|Srest]) :-
    register_formula(Formula,Parent,SubId),
    register_formulas(Frest,Parent,Srest).
register_formula(Formula,Parent,Id) :-
    get_new_id(Id),
    assert( stored_formula(Id,Formula) ),
    assert( subnode(Parent,Id) ),
    assert( supernode(Id,Parent) ).

get_new_id(Id) :-
    ( retract( id_counter(Old) ) -> true
    ; otherwise -> Old = 0),
    Id is Old+1,
    assert( id_counter(Id) ).


disjunction_to_list(A,B,Out) :-
    disjunction_to_list2(A,[B],Out,[]).
disjunction_to_list2(Expr,Rest) -->
    {get_texpr_expr(Expr,disjunct(A,B)),!},
    disjunction_to_list2(A,[B|Rest]).
disjunction_to_list2(Expr,Rest) -->
    [Expr],disjunction_to_list3(Rest).
disjunction_to_list3([]) --> !.
disjunction_to_list3([H|T]) --> disjunction_to_list2(H,T).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some predicates for the support of the Tcl/Tk UI

:- meta_predicate call_with_temp_preference(*,*,0).
call_with_temp_preference(Prefs,Values,Call) :-
    maplist(get_preference,Prefs,OldValues),
    maplist(set_preference,Prefs,Values),
    call_cleanup(Call,maplist(set_preference,Prefs,OldValues)),!.

bv_get_value_unlimited(Id,StateId,Value) :-
    visited_expression(StateId,State),!,
    get_preference(time_out,Timeout),
    call_with_temp_preference([expand_avl_upto],[-1],bv_get_value(Id,State,StateId,Timeout,Value)).
bv_get_value_unlimited(_Id,_StateId,e('unknown state')).

bv_get_value_unlimited_and_unprocessed(Id,StateId,Value) :-
    visited_expression(StateId,State),!,
    get_preference(time_out,Timeout),
    call_with_temp_preference([expand_avl_upto],[-1],bv_get_value_unprocessed(Id,State,StateId,Timeout,Value)).
bv_get_value_unlimited_and_unprocessed(_Id,_StateId,e('unknown state')).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
:- use_module(tools_io).
bv_print_to_file(IdList,StateId,Filename) :-
    call_with_temp_preference([expand_avl_upto],[-1],bv_print_to_file_aux(IdList,StateId,Filename)).
bv_print_to_file_aux(IdList,StateId,Filename) :-
    safe_open_file(Filename,write,S,[]),
    bv_print_to_stream_l(IdList,S,StateId,false),
    close(S).

bv_print_to_stream_l([],_S,_StateId,_WriteAnd).
bv_print_to_stream_l([Id|Rest],S,StateId,WriteAnd) :-
    bv_print_to_stream(Id,S,StateId,OutType,WriteAnd),
    ( OutType=predicate -> WriteAnd2=true
    ; otherwise         -> WriteAnd2=WriteAnd),
    bv_print_to_stream_l(Rest,S,StateId,WriteAnd2).

bv_print_to_stream(Id,S,StateId,predicate,WriteAnd) :-
    stored_formula(Id,Formula),
    bv_print_formula_to_stream(Formula,Id,S,StateId,WriteAnd),!,nl(S).
bv_print_to_stream(Id,S,StateId,other,_WriteAnd) :-
    get_node_label(Id,Label),
    bv_get_values([Id],StateId,[Value]),
    write(S,'/* '),write(S,Label),write(S,': '),
    bv_print_std_value(Value,S),write(S,' */\n').
bv_print_std_value(e(Msg),S) :- write(S,'ERROR( '), write(S,Msg), write(S,' )').
bv_print_std_value(p(Value),S) :- write(S,Value).
bv_print_std_value(v(Value),S) :- write(S,Value).
bv_print_std_value(i,S) :- write(S,'INACTIVE').

bv_print_formula_to_stream(Formula,Id,S,StateId,WriteAnd) :-
    is_texpr(Formula),
    get_texpr_type(Formula,Type),
    bv_get_value_unlimited_and_unprocessed(Id,StateId,Value),
    bv_print_formula_to_stream2(Type,Formula,Value,S,WriteAnd).

bv_print_formula_to_stream2(pred,Formula,p(Value),S,WriteAnd) :- !,
    bv_print_predicate_to_stream(Value,Formula,S,WriteAnd).
bv_print_formula_to_stream2(Type,Formula,btvalue(Value,_),S,WriteAnd) :- !,
    create_texpr(value(Value),Type,[],TValue),
    create_texpr(equal(Formula,TValue),pred,[],Equal),
    write_bexpression(S,Equal,WriteAnd).
    
bv_print_predicate_to_stream(true,Formula,S,WriteAnd) :-
    write_bexpression(S,Formula,WriteAnd).
bv_print_predicate_to_stream(false,Formula,S,WriteAnd) :-
    create_texpr(negation(Formula),pred,[],NegatedFormula),
    write_bexpression(S,NegatedFormula,WriteAnd).

write_bexpression(Stream,Expression,WriteAnd) :-
    ( WriteAnd = true -> write(Stream,' & \n')
    ; otherwise       -> true),
    translate_bexpression_to_codes(Expression,Codes),
    put_codes(Codes,Stream).

bv_write_all_variables_and_constants(StateId,Filename) :-
    call_with_temp_preference([expand_avl_upto],[-1],bv_write_all_variables_and_constants_aux(StateId,Filename)).
bv_write_all_variables_and_constants_aux(StateId,Filename) :-
    bv_expand_formula(constants,_CLabel,ConstNodes),
    bv_expand_formula(variables,_VLabel,VarNodes),
    append(ConstNodes,VarNodes,Subnodes),expand_all(Subnodes),
    append([constants|ConstNodes],[variables|VarNodes],Nodes),
    bv_print_to_file(Nodes,StateId,Filename).
expand_all(Subnodes) :-
    member(Node,Subnodes),
    bv_expand_formula(Node,_,_),
    fail.
expand_all(_Subnodes).


:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(specification_initialised,reset_bvisual,
                           'Initialise module bvisual2.').
:- register_event_listener(change_of_animation_mode,reset_bvisual,
                           'Initialise module bvisual2.').

:- use_module(self_check).
% a very small use-case:
:- assert_must_succeed(( 
       reset_bvisual,
       bv_get_top_level(T), %print(top(T)),nl,
       %T == [variables,constants,inv],
       member(inv,T),
       bv_expand_formula(inv,Text,Sub),
       %print(expand(Text,Sub)),nl,
       Text == 'INVARIANT',
       Sub = [H|_],
       bv_expand_formula(H,_Text1,_Sub1),
       %print(expand1(Text1,Sub1)),nl,
       bv_get_values(Sub,root,Values),
       %print(values(Values)),nl, % should be [i,...] or e('unknown state')
       Values = [VI|_], (VI=i ; VI=e(_)),
       reset_bvisual
       )).

% example bv_get_top_level_formula(theoremsc,Txt,Labels,Vals)

bv_get_top_level_formula(Category,CatText,Labels,Values) :-
       current_state_id(StateId),
       bv_get_top_level_formula(Category,CatText,StateId,Labels,Values).
bv_get_top_level_formula(Category,CatText,StateId,Labels,Values) :-
      bv_get_top_level(TL),
      member(Category,TL),
      bv_expand_formula(Category,CatText,Subs),
      maplist(get_node_label,Subs,Labels),
      bv_get_values(Subs,StateId,Values).
