% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(coverage_tools,[
                          print_coverage_info/0,
                          print_coverage_info/1,
			  get_file_coverage_info/1, get_file_coverage_info/2,
                          gen_pltables_coverage_rows/4,
                          gen_pltables_descriptions_table/3,
                          export_tables/3,
                          export_tables_html/1,
                          export_tables_latex/1,
                          export_tables_html_stream/1,
                          export_tables_latex_stream/1,
                          colorize_source_files/2,
                          export_emma/1,
                          '$NOT_COVERED'/1,
                          uncovered/1
                          ]).

:- use_module(library(lists)).
:- use_module(library(between)).
:- use_module(library(sets)).

:- use_module(tools).
:- use_module(extension('pltables/pltables')).
:- use_module(self_check).
:- use_module(junit_tests).


:- use_module(module_information).
:- module_info(group,coverage_analysis).
:- module_info(description,'This module automatically generates the SICStus Prolog coverage information reports.').

:- volatile summary_row/9, module_row/10.
:- dynamic summary_row/9, module_row/10.

:- use_module(coverage_tools_annotations).
% defines: '$NOT_COVERED'(_X) :- true.

print_coverage_info :-
    coverage_data(D),
    merge_coverage_pred_data(D,MD),
    print_uncovered(MD), nl,
    merge_coverage_file_data(MD,FD),
    print_file_stats(FD).

print_coverage_info(Module) :-
    coverage_data(D),
    merge_coverage_pred_data(D,MD),
    print_uncovered(Module,MD), nl,
    merge_coverage_file_data(MD,FD),
    print_file_stats(Module,FD).

get_file_coverage_info(FDSorted) :-
    coverage_data(Data),
    get_file_coverage_info(Data,FDSorted).

get_file_coverage_info(Data,FDSorted) :-
    merge_coverage_pred_data(Data,MD),
    merge_coverage_file_data(MD,FD),
    sort(FD, FDSorted).

merge_coverage_pred_data([],[]).
merge_coverage_pred_data([CD | Tail],Res) :-
   create_fresh_pred_data(CD,NewData),
   merge_coverage_pred_data(Tail,NewData,Res).

merge_coverage_pred_data([],LastData,[LastData]).
merge_coverage_pred_data([CD | Tail],LastData,Res) :-
  (merge_pred_data(CD,LastData,NewData) -> Res=NR
    ;  Res=[LastData|NR], create_fresh_pred_data(CD,NewData)
  ),merge_coverage_pred_data(Tail,NewData,NR).

% translate a list of sorted counter data into summary of data for every predicate
% pred_data(FileName,PredicateSpecification,NrOfClauses,NrOfCoveredClauses,WasTheLastClauseAlreadyCovered,ListOfCoveredLines,ListOfUncoveredLines)
merge_pred_data(counter(File,PredSpec,ClauseNo,LineNr)-Tagged,
                pred_data(File,PredSpec,MaxClause,Covered,LastCov,OldLineNrCovered,OldLineNrUncovered,OldLineNrNonDet),
                pred_data(File,PredSpec,NewMax,NewCovered,NewLastCov,NewLineNrCovered,NewLineNrUncovered,NewLineNrNonDet)) :-
    (Tagged = nondet(_) -> NewLineNrNonDet = [LineNr|OldLineNrNonDet] ; NewLineNrNonDet = OldLineNrNonDet),
    (uncovered(Tagged)
       -> NewLineNrCovered = OldLineNrCovered, NewLineNrUncovered = [LineNr|OldLineNrUncovered]
       ;  NewLineNrCovered = [LineNr|OldLineNrCovered], NewLineNrUncovered = OldLineNrUncovered),
    (ClauseNo>MaxClause
      -> NewMax = ClauseNo,
         (uncovered(Tagged)
         -> NewCovered=Covered, NewLastCov=0
          ; NewCovered is Covered+1, NewLastCov=1)
       ; NewMax=MaxClause,
         (ClauseNo<MaxClause -> print('*** WARNING: ClauseNr decreasing: '),nl,
            print(pred_data(File,PredSpec,MaxClause,Covered)),nl,
            print(new_clause_no(ClauseNo)), nl, trace,
            NewCovered=Covered, NewLastCov=LastCov
            ; (uncovered(Tagged);LastCov=1) -> NewCovered=Covered, NewLastCov=LastCov
            ; NewLastCov=1, NewCovered is Covered+1 % we cover a clause that was thus far uncovered
         )
    ).

create_fresh_pred_data(counter(File,PredSpec,ClauseNo,LineNr)-Tagged,
                       pred_data(File,PredSpec,ClauseNo,Covered,LastCov,CovLines,UnCovLines,NonDetLines)) :-
	(Tagged = nondet(_) -> NonDetLines = [LineNr] ; NonDetLines = []),
        (uncovered(Tagged) -> Covered=0,CovLines=[],UnCovLines=[LineNr] ; Covered=1,CovLines=[LineNr],UnCovLines=[]),
        LastCov=Covered,
        (ClauseNo<1 -> print('### Warning: clause number <1 : '), print(ClauseNo),nl,
                       print('### '),print_pred(File,PredSpec),nl,trace
                    ; true).

uncovered(det(0)).
uncovered(nondet(0)).

print_uncovered(T) :- print('UNCOVERED PREDICATES:'),nl,print_uncovered2(T).
print_uncovered2([]).
print_uncovered2([pred_data(File,PredSpec,_M,0,_,_,_,_)|T]) :-  !,
    print('   '),print_pred(File,PredSpec), nl, print_uncovered2(T).
print_uncovered2([_|T]) :- print_uncovered2(T).

print_uncovered(Module,T) :- print('UNCOVERED PREDICATES:'),nl,print_uncovered2(Module,T).
print_uncovered2(_Module,[]).
print_uncovered2(Module,[pred_data(File,PredSpec,_M,0,_,_,_,_)|T]) :-  !,
    (
        PredSpec = Module:_/_
        ->  print('   '),print_pred(File,PredSpec), nl
        ;   true
    ),
    print_uncovered2(Module,T).
print_uncovered2(Module,[_|T]) :- print_uncovered2(Module,T).


print_pred(File,Pred) :- print(Pred), print('  in  '), print(File).

print_file_stats(L) :- print('COVERAGE PER FILE:'),nl,
   init_program_data(PD),print_file_stats2(L,PD).
print_file_stats2([],prog_data(NrFiles,Preds,CovPreds,Clauses,CovClauses)) :- % now print total summary:
   PPerc is CovPreds*100/Preds, CPerc is CovClauses*100/Clauses,
   format("TOTAL #FILES: ~w~n  ~w/~w (~2F %) Predicates~n  ~w/~w (~2F %) Clauses~n",
          [NrFiles,CovPreds,Preds,PPerc,CovClauses,Clauses,CPerc]).
print_file_stats2([FileData|T],PD) :-
   FileData = file_data(File,Preds,CovPreds,Clauses,CovClauses,_,_,_),
   PPerc is CovPreds*100/Preds, CPerc is CovClauses*100/Clauses,
   format("~w~n  ~w/~w (~2F %) Predicates~n  ~w/~w (~2F %) Clauses~n",
          [File,CovPreds,Preds,PPerc,CovClauses,Clauses,CPerc]),
   merge_program_data(FileData,PD,NPD),
   print_file_stats2(T,NPD).

print_file_stats(_Module,[]).
print_file_stats(Module,[FileData|T]) :-
   FileData = file_data(File,Preds,CovPreds,Clauses,CovClauses,_,_,_),
   get_modulename_filename(File,Module)
   ->  PPerc is CovPreds*100/Preds, CPerc is CovClauses*100/Clauses,
       format("~w~n  ~w/~w (~2F %) Predicates~n  ~w/~w (~2F %) Clauses~n",[File,CovPreds,Preds,PPerc,CovClauses,Clauses,CPerc])
    ;  print_file_stats(Module,T).


% translate a list of sorted predicate data into summary of data for every file
% file_data(FileName,NrOfPredicates,NrOfPredicatesCovered,NrOfClauses,NrOfClausesCovered,CoveredLines,UncoveredLines,NonDetLines)
merge_coverage_file_data([],[]).
merge_coverage_file_data([PD | Tail],Res) :-
   create_fresh_file_data(PD,NewData),
   merge_coverage_file_data(Tail,NewData,Res).

merge_coverage_file_data([],LastData,[LastData]).
merge_coverage_file_data([PD | Tail],LastData,Res) :-
  (merge_file_data(PD,LastData,NewData) -> Res=NR
    ;  Res=[LastData|NR], create_fresh_file_data(PD,NewData)
  ),merge_coverage_file_data(Tail,NewData,NR).

:-assert_must_succeed(coverage_tools:merge_file_data(pred_data(file,spec,1,2,_,[5],[5],[5]),file_data(file,5,2,6,3,[6],[6],[6]),file_data(file,6,3,7,5,[5,6],[5,6],[5,6]))).
:-assert_must_fail(coverage_tools:merge_file_data(pred_data(file,spec,1,2,_,[4],[4],[5]),file_data(file,5,2,6,3,[6],[7],[5]),file_data(file,6,3,7,5,[5,6],[5,6],[7]))).
:-assert_must_fail(coverage_tools:merge_file_data(pred_data(file,spec,1,2,_,_,_,_),file_data(another_file,5,2,6,3,_,_,_),file_data(file,6,3,7,5,_,_,_))).
merge_file_data(pred_data(File,_PredSpec,MaxClause,Covered,_,CoveredLines,UncoveredLines,NonDetLines),
                file_data(File,Preds,PCov,Clauses,CCov,AlreadyCoveredLines,AlreadyUncoveredLines,AlreadyNonDetLines),
                file_data(File,P1,PC1,C1,CC1,NowCoveredLines,NowUncoveredLines,NowNonDetLines)) :-
    append(NonDetLines,AlreadyNonDetLines,NowNonDetLines),
    append(CoveredLines,AlreadyCoveredLines,NowCoveredLines),
    append(UncoveredLines,AlreadyUncoveredLines,NowUncoveredLines),
    P1 is Preds+1,
    (Covered>0 -> PC1 is PCov+1 ; PC1=PCov),
    CC1 is CCov+Covered,
    C1 is Clauses+MaxClause.

:-assert_must_succeed(coverage_tools:create_fresh_file_data(pred_data(file,spec,15,2,_,[5],[7],[7]),file_data(file,1,1,15,2,[5],[7],[7]))).
:-assert_must_fail(coverage_tools:create_fresh_file_data(pred_data(file,spec,15,2,_,[5],[6],[6]),file_data(file,1,1,15,2,[7],[8],[6]))).
:-assert_must_fail(coverage_tools:create_fresh_file_data(pred_data(file,spec,15,0,_,[5],[6],[6]),file_data(file,1,1,15,2,[5],[6],[6]))).
create_fresh_file_data(pred_data(File,_PredSpec,MaxClause,Covered,_,CoveredLines,UncoveredLines,NonDetLines),
                       file_data(File,1,PCov,MaxClause,Covered,CoveredLines,UncoveredLines,NonDetLines)) :-
    (Covered>0 -> PCov=1 ; PCov=0).

init_program_data(prog_data(0,0,0,0,0)).
merge_program_data(file_data(_,P,CP,C,CC,_,_,_),prog_data(Files1,P1,CP1,C1,CC1),
                   prog_data(Files2,P2,CP2,C2,CC2)) :-
            Files2 is Files1+1,
            P2 is P1+P, CP2 is CP1+CP,
            C2 is C1+C, CC2 is CC1+CC.

:- assert_must_succeed(coverage_tools:coverage_attributes(50, 50,[total=[53,48],covered=[53,48],covered_percentage=[49,48,48,46,48]])).
:- assert_must_fail(coverage_tools:coverage_attributes(50,2,[total=[53,48],covered=[52],covered_percentage=[52,46,48]])).
:- assert_must_succeed(coverage_tools:coverage_attributes(100,100,[total=[49,48,48],covered=[49,48,48],covered_percentage=[49,48,48,46,48]])).
coverage_attributes(Total,Covered,Res) :-
  Perc is Covered*100/Total,
  number_codes(Total,TC), number_codes(Covered,CC), number_codes(Perc,PC),
  Res = ['='(total,TC), '='(covered,CC), '='(covered_percentage,PC)].

gen_pltables_descriptions_table(TableName, TableAttributes, ImportantGroups) :-
  new_table(TableName, TableAttributes),
  add_column(TableName, 'name', []),
  add_column(TableName, 'description', []), !,
  findall(Module, (get_module_info(Module,group,MGroup),member(MGroup,ImportantGroups)), ListImportant),
  gen_pltables_descriptions_rows(TableName, ListImportant),
  findall(Module, (get_module_info(Module,group,MGroup),\+(member(MGroup,ImportantGroups))), ListNotImportant),
  gen_pltables_descriptions_rows(TableName, ListNotImportant).

gen_pltables_descriptions_rows(_, []).
gen_pltables_descriptions_rows(TableName, [Module | T]) :-
  (get_module_info(Module, description, Desc)
   -> add_row(TableName, [Module, Desc], [])
   ;  true),
  gen_pltables_descriptions_rows(TableName,T).

gen_pltables_coverage_table(TableName, TableAttributes, FileData) :-
  new_table(TableName, TableAttributes),
  add_column(TableName, 'module', [align(left)]),
  add_column(TableName, 'clauses', [align(right)]),
  add_column(TableName, 'covered', [align(right)]),
  add_column(TableName, 'uncoverable', [align(right)]),
  add_column(TableName, '%', [align(right)]),
  add_column(TableName, 'preds', [align(right)]),
  add_column(TableName, 'covered', [align(right)]),
  add_column(TableName, 'uncoverable', [align(right)]),
  add_column(TableName, '%', [align(right)]),
  summary_row(TableName, FileData, [0,0,0,0,0,0]).

summary_row(ModuleGroup, [], [Preds,CovPreds,IntPreds,Clauses,CovClauses,IntClauses]) :-
  PercentageClauses1 is CovClauses*100/(Clauses-IntClauses),
  RoundedClauses is round(PercentageClauses1*10),
  PercentageClauses is RoundedClauses/10,
  (PercentageClauses > 100 -> junit_coverage_report_error(coverage_tables,'Covered more than 100% of all Clauses') ; true),
  PercentagePreds1 is CovPreds*100/(Preds-IntPreds),
  RoundedPreds is round(PercentagePreds1*10),
  PercentagePreds is RoundedPreds/10,
  (PercentagePreds > 100 -> junit_coverage_report_error(coverage_tables,'Covered more than 100% of all Clauses') ; true),
  add_row(ModuleGroup, [summary,Clauses,CovClauses,IntClauses,PercentageClauses,Preds,CovPreds,IntPreds,PercentagePreds], [bold]),
  assert(summary_row(ModuleGroup,Clauses,CovClauses,IntClauses,PercentageClauses,Preds,CovPreds,IntPreds,PercentagePreds)).
summary_row(ModuleGroup, [file_data(File,Preds,CovPreds,Clauses,CovClauses,CoveredLines,_,_)|T],[CPreds,CCovPreds,CFailPreds,CClauses,CCovClauses,CFailClauses]) :-
  allowed_file(File), !,
  get_modulename_filename(File, Module),
  (get_module_info(Module, group, ModuleGroupFile) ; ModuleGroupFile = other),
  (  ModuleGroup = ModuleGroupFile ->
       (  NCPreds is CPreds + Preds,
          NCCovPreds is CCovPreds + CovPreds,
          NCClauses is CClauses + Clauses,
          NCCovClauses is CCovClauses + CovClauses,
          uncoverable_preds(File,IntPreds),
          uncoverable_clauses(File,IntClauses,CoveredLines),
          NCFailPreds is IntPreds + CFailPreds,
          NCFailClauses is IntClauses + CFailClauses,
          summary_row(ModuleGroup,T,[NCPreds,NCCovPreds,NCFailPreds,NCClauses,NCCovClauses,NCFailClauses])
       )
    ;  summary_row(ModuleGroup,T,[CPreds,CCovPreds,CFailPreds,CClauses,CCovClauses,CFailClauses])
  ).
% file was not allowed / wrong ending
summary_row(ModuleGroup, [_H|T],[CPreds,CCovPreds,CFailPreds,CClauses,CCovClauses,CFailClauses]) :-
  summary_row(ModuleGroup,T,[CPreds,CCovPreds,CFailPreds,CClauses,CCovClauses,CFailClauses]).

gen_pltables_coverage_rows([], _, _, _).
gen_pltables_coverage_rows([file_data(File,Preds,CovPreds,Clauses,CovClauses,CoveredLines,_,_)|T],
                           TableAttributes, ImportantGroups, ColorizedURL) :-
  allowed_file(File), !,
  get_tail_filename(File,TF),
  get_modulename_filename(File, Module),
  uncoverable_preds(File,IntPreds),
  uncoverable_clauses(File,IntClauses,CoveredLines),
  (Clauses-IntClauses =:= 0
   -> PercentageClauses1 = 0
   ;  PercentageClauses1 is CovClauses*100/(Clauses-IntClauses)),
  RoundedClauses is round(PercentageClauses1*10),
  PercentageClauses is RoundedClauses/10,
  (PercentageClauses > 100 -> junit_coverage_report_error(coverage_tables,'Covered more than 100% of all Clauses') ; true),
  (Preds - IntPreds =:= 0
   -> PercentagePreds1 = 0
   ;  PercentagePreds1 is CovPreds*100/(Preds-IntPreds)),
  RoundedPreds is round(PercentagePreds1*10),
  PercentagePreds is RoundedPreds/10,
  (PercentagePreds > 100 -> junit_coverage_report_error(coverage_tables,'Covered more than 100% of all Clauses') ; true),
  (get_module_info(Module, group, ModuleGroup) ; ModuleGroup = other),
  (
  	PercentageClauses < 10 ->
  		RowAttrs1 = [color('FF0000')] ;
  		RowAttrs1 = []
  ),
  (
  	PercentagePreds < 10 ->
  		RowAttrs2 = [color('FF0000')|RowAttrs1] ;
  		RowAttrs2 = RowAttrs1
  ),
  (pltables:pltable(ModuleGroup,_) -> true
    ;
    ((
      member(ModuleGroup, ImportantGroups) ->
  	  ToAppend = [description(ModuleGroup, [bold])] ;
  	  ToAppend = [description(ModuleGroup, [])]
  	),
    append(ToAppend,TableAttributes, TableAttributes2),
    gen_pltables_coverage_table(ModuleGroup, TableAttributes2, [file_data(File,Preds,CovPreds,Clauses,CovClauses,_,_,_)|T])
  )),
  ajoin([ColorizedURL,Module,'.html'],URL),
  append([url(URL)],RowAttrs2,RowAttrs3),
  add_row(ModuleGroup, [TF,Clauses,CovClauses,IntClauses,PercentageClauses,Preds,CovPreds,IntPreds,PercentagePreds],RowAttrs3),
  assert(module_row(ModuleGroup,TF,Clauses,CovClauses,IntClauses,PercentageClauses,Preds,CovPreds,IntPreds,PercentagePreds)),
  gen_pltables_coverage_rows(T, TableAttributes, ImportantGroups,ColorizedURL).
% file was not allowed / wrong ending
gen_pltables_coverage_rows([_H|T], TableAttributes, ImportantGroups, ColorizedURL) :-
  gen_pltables_coverage_rows(T, TableAttributes, ImportantGroups,ColorizedURL).

:- assert_must_succeed(coverage_tools:allowed_file('/a/b/c/d.pl')).
:- assert_must_succeed(coverage_tools:allowed_file('d.pl')).
:- assert_must_fail(coverage_tools:allowed_file('/a/b/c/d.eventb')).
:- assert_must_fail(coverage_tools:allowed_file('/a/b/c/d.P')).
:- assert_must_fail(coverage_tools:allowed_file('/a/b/c/d.P')).
:- assert_must_fail(coverage_tools:allowed_file('d.eventb')).
:- assert_must_fail(coverage_tools:allowed_file('d.P')).
:- assert_must_fail(coverage_tools:allowed_file('/a/b/c/d.pml.pl')).
:- assert_must_fail(coverage_tools:allowed_file('TestingUnicodeFeatures_unicode_saved.csp.pl')).
allowed_file(File) :-
  get_tail_filename(File,TF),
  atom_codes(TF,TFCodes),
  suffix(TFCodes,".pl"),
  \+ suffix(TFCodes, ".pml.pl"),
  \+ suffix(TFCodes, ".csp.pl").

export_tables(FilenameHTML,FilenameTex,FilenameXML) :-
  findall(TableName, pltables:pltable(TableName, _), UnsortedListOfTables),
  sort(UnsortedListOfTables,ListOfTables),
  tables_to_html(ListOfTables, FilenameHTML),
  tables_to_latex(ListOfTables, FilenameTex),
  tables_to_xml(ListOfTables, FilenameXML).

export_tables_html(FilenameHTML) :-
  findall(TableName, pltables:pltable(TableName, _), UnsortedListOfTables),
  sort(UnsortedListOfTables,ListOfTables),
  tables_to_html(ListOfTables, FilenameHTML).

export_tables_latex(FilenameTex) :-
  findall(TableName, pltables:pltable(TableName, _), UnsortedListOfTables),
  sort(UnsortedListOfTables,ListOfTables),
  tables_to_latex(ListOfTables, FilenameTex).

export_tables_html_stream(Stream) :-
  findall(TableName, pltables:pltable(TableName, _), UnsortedListOfTables),
  sort(UnsortedListOfTables,ListOfTables),
  tables_to_html_stream(ListOfTables, Stream).

export_tables_latex_stream(Stream) :-
  findall(TableName, pltables:pltable(TableName, _), UnsortedListOfTables),
  sort(UnsortedListOfTables,ListOfTables),
  tables_to_latex_stream(ListOfTables, Stream).

colorize_source_files([],_ColorizedPath).
colorize_source_files([file_data(File,_Preds,_CovPreds,_Clauses,_CovClauses,CoveredLines,UncoveredLines,NonDetLines)|T],ColorizedPath) :-
  colorize_source_file(File,CoveredLines,UncoveredLines,NonDetLines,ColorizedPath),
  colorize_source_files(T,ColorizedPath).

colorize_source_file(FileIn,CoveredLines,UncoveredLines,NonDetLines,ColorizedPath) :-
  colored_html_path(ColorizedPath,FileIn,FileOut),
  open(FileIn,read,StreamIn),
  open(FileOut,write,StreamOut),
  write(StreamOut, '<html>\n<head>\n'),
  write(StreamOut, '  <meta http-equiv="content-type" content="text/html; charset=ISO-8859-1"/>\n'),
  write(StreamOut, '  <style type="text/css">\n'),
  write(StreamOut, '    .uncovered    {color:#D02020}\n'),
  write(StreamOut, '    .covered      {color:#109010}\n'),
  write(StreamOut, '    .mixed        {color:#FFA500}\n'),
  write(StreamOut, '    .unexecutable {color:#000000}\n'),
  write(StreamOut, '    .uncoverable  {color:#0000FF}\n'),
  write(StreamOut, '    .red          {color:#FF0000}\n'),
  write(StreamOut, '  </style>\n</head>\n<body>\n'),
  write(StreamOut, '<table style="white-space:pre;font-family:monospace;" cellspacing="0" cellpadding="0">\n'),
  colorize_source_lines(StreamIn,StreamOut,FileIn,CoveredLines,UncoveredLines,NonDetLines,1),
  write(StreamOut, '</table>\n</body>\n</html>\n'),
  close(StreamIn), close(StreamOut).

colorize_source_lines(StreamIn,StreamOut,FileIn,CoveredLines,UncoveredLines,NonDetLines,CurrentLine) :-
  read_line(StreamIn, LineIn),
  (
      LineIn = end_of_file -> true
    ; output_source_line(StreamOut,FileIn,LineIn,CoveredLines,UncoveredLines,NonDetLines,CurrentLine),
      CL2 is CurrentLine + 1,
      colorize_source_lines(StreamIn,StreamOut,FileIn,CoveredLines,UncoveredLines,NonDetLines,CL2)
  ).

output_source_line(StreamOut,FileIn,Line,CoveredLines,UncoveredLines,NonDetLines,CurrentLine) :-
  source_line_class(FileIn,CurrentLine,CoveredLines,UncoveredLines,Class),
  line_det(CurrentLine,NonDetLines,LineDet,LineDetClass),
  escape_line(Line,EscapedLine),
  format(StreamOut,'<tr><td>~w </td><td class="~w">~w</td><td class="~w">~s</td></tr>\n',[CurrentLine,LineDetClass,LineDet,Class,EscapedLine]).

line_det(CurrentLine,NonDetLines,LineDet,red) :-
  member(CurrentLine,NonDetLines) -> LineDet = '?' ; LineDet = ''.

:- assert_must_succeed(coverage_tools:source_line_class('Filename',15,[1,2,3,15],[4,5,6],covered)).
:- assert_must_succeed(coverage_tools:source_line_class('Filename',15,[1,2,3],[4,5,6,15],uncovered)).
:- assert_must_succeed(coverage_tools:source_line_class('Filename',15,[1,2,3,15],[4,5,6,15],mixed)).
:- assert_must_fail((coverage_tools:source_line_class('Filename',15,[1,2,3,15],[4,5,6,15],X),X=covered)).
:- assert_must_succeed(coverage_tools:source_line_class('Filename',15,[1,2,3,15],[4,5,6,15],uncovered)).
source_line_class(_FileIn,CurrentLine,CoveredLines,UncoveredLines,mixed) :-
    memberchk(CurrentLine,CoveredLines),
    memberchk(CurrentLine,UncoveredLines), !.
source_line_class(_FileIn,CurrentLine,CoveredLines,_UncoveredLines,covered) :-
    memberchk(CurrentLine,CoveredLines),!.
source_line_class(FileIn,CurrentLine,_Cov,_Uncov,uncoverable) :-
    (user:clause_counter_collected(FileIn,_,_,CurrentLine,_,_);user:call_counter_collected(FileIn,_,_,CurrentLine)),!.
source_line_class(_FileIn,CurrentLine,_CoveredLines,UncoveredLines,uncovered) :-
    memberchk(CurrentLine,UncoveredLines),!.
source_line_class(_FileIn,_CurrentLine,_CoveredLines,_UncoveredLines,unexecutable).

:- assert_must_succeed(coverage_tools:escape_line("Test mit < Zeichen", "Test mit &lt; Zeichen")).
:- assert_must_succeed(coverage_tools:escape_line("Test mit > Zeichen", "Test mit &gt; Zeichen")).
escape_line([],[]).
% less than symbol
escape_line([60|T],[38,108,116,59|T2]) :-
    escape_line(T,T2).
% greater than symbol
escape_line([62|T],[38,103,116,59|T2]) :-
    escape_line(T,T2).
escape_line([X|T],[X|T2]) :-
    escape_line(T,T2).

:- assert_must_succeed(coverage_tools:colored_html_path('/path/','test.pl','/path/test.html')).
colored_html_path(ColorizedPath,FileIn,FileOut) :-
    get_modulename_filename(FileIn,ModuleName),
    ajoin([ColorizedPath, ModuleName, '.html'], FileOut).

% a predicate is only counted as failing if all definitions end in a failing predicate
uncoverable_preds(File,Count) :-
    findall(clause(Module:Name/Arity,Counter), user:clause_counter_collected(File,Module:Name/Arity,Counter,_LayoutHead,_FirstLine,_LastLine),ListTemp),
    remove_dups(ListTemp,List),
    count_uncoverable_preds(List,Count).

count_uncoverable_preds([],0).
count_uncoverable_preds([clause(Module:Name/Arity,_Counter)|T],Count) :-
    delete_all(T,Module:Name/Arity,Res),
    count_uncoverable_preds(Res,Count2),
    length(T,LengthBeforeDelete), length(Res,LengthAfterDelete),
    Occurences is LengthBeforeDelete - LengthAfterDelete + 1,
    user:definitionnr(Module,Name/Arity,NoOfDefinitions),
    (Occurences = NoOfDefinitions -> Count is Count2 + 1 ; Count = Count2).

delete_all([],_PredSpec,[]).
delete_all([clause(PredSpec,_)|T], PredSpec, NT) :-
    !, delete_all(T,PredSpec,NT).
delete_all([clause(PredSpec2,Counter)|T], PredSpec, [clause(PredSpec2,Counter)|NT]) :-
    PredSpec2 \= PredSpec,
    delete_all(T,PredSpec,NT).

% a clause is counted as uncoverable, if at least one of the definitions ends in a failing predicate
% this might need to be changed for clauses that might or might not fail
uncoverable_clauses(File,Count,CoveredLines) :-
  findall(clause(Module:Name/Arity,Counter),
	  (user:clause_counter_collected(File,Module:Name/Arity,Counter,_LayoutHead,FirstLine,LastLine),
	  numlist(FirstLine,LastLine,AllLines),
	  intersection(AllLines,CoveredLines,[])),
	  ListTemp),
  remove_dups(ListTemp,List),
  length(List,Count).

sum_all(List,[A,B,C,D,E,F]) :- sum_all(List,[0,0,0,0,0,0],[A,B,C,D,E,F]).
sum_all([],[A,B,C,D,E,F],[A,B,C,D,E,F]).
sum_all([summary_row(_ModuleGroup,Clauses,CovClauses,IntClauses,_PercentageClauses,Preds,CovPreds,IntPreds,_PercentagePreds)|T],
        [ClausesIn,CovClausesIn,IntClausesIn,PredsIn,CovPredsIn,IntPredsIn],[A,B,C,D,E,F]) :-
        NClauses is Clauses + ClausesIn,
        NCovClauses is CovClauses + CovClausesIn,
        NIntClauses is IntClauses + IntClausesIn,
        NPreds is Preds + PredsIn,
        NCovPreds is CovPreds + CovPredsIn,
        NIntPreds is IntPreds + IntPredsIn,
        sum_all(T,[NClauses,NCovClauses,NIntClauses,NPreds,NCovPreds,NIntPreds],[A,B,C,D,E,F]).

emma_header(Stream) :-
    findall(summary_row(A,B,C,D,E,F,G,H,I), summary_row(A,B,C,D,E,F,G,H,I), List),
    sum_all(List,[Clauses,CovClauses,IntClauses,Preds,CovPreds,IntPreds]),
    module_information:number_of_groups(Groups),
    module_information:number_of_modules(Modules),
    TCovClauses is CovClauses + IntClauses,
    PClauses is TCovClauses/Clauses*100,
    (PClauses > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Clauses') ; true),
    TCovPreds is CovPreds + IntPreds,
    PPreds is TCovPreds/Preds*100,
    (PPreds > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Predicates') ; true),
    format(Stream, '  <stats>~n', []),
    format(Stream, '    <packages value="~w"/>~n', [Groups]),
    format(Stream, '    <clases value="~w"/>~n', [Modules]),
    format(Stream, '    <methods value="~w"/>~n', [Clauses]),
    format(Stream, '    <srcfiles value="~w"/>~n', [Modules]),
    format(Stream, '    <srclines value="0"/>~n', []), % TODO: Lines?
    format(Stream, '  </stats>~n', []),
    format(Stream, '  <data>~n', []),
    format(Stream, '    <all name="all classes">~n', []),
    format(Stream, '        <coverage type="class, %" value="100% (~w/~w)"/>~n', [Modules,Modules]),
    format(Stream, '        <coverage type="method, %" value="~w% (~w/~w)"/>~n', [PClauses,TCovClauses,Clauses]),
    format(Stream, '        <coverage type="block, %" value="~w% (~w/~w)"/>~n', [PPreds,TCovPreds,Preds]),
    format(Stream, '        <coverage type="line, %" value="0% (0/0)"/>~n~n', []). % TODO?

emma_data(Stream) :-
    summary_row(ModuleGroup,Clauses,CovClauses,IntClauses,_PercentageClauses,Preds,CovPreds,IntPreds,_PercentagePreds),
    TCovClauses is CovClauses + IntClauses,
    PClauses is TCovClauses/Clauses*100,
    (PClauses > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Clauses') ; true),
    TCovPreds is CovPreds + IntPreds,
    PPreds is TCovPreds/Preds*100,
    (PPreds > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Predicates') ; true),
    module_information:number_of_modules(ModuleGroup,Modules),
    format(Stream, '        <package name="~w">~n', [ModuleGroup]),
    format(Stream, '            <coverage type="class, %" value="100% (~w/~w)"/>~n', [Modules,Modules]),
    format(Stream, '            <coverage type="method, %" value="~w% (~w/~w)"/>~n', [PClauses,TCovClauses,Clauses]),
    format(Stream, '                <coverage type="block, %" value="~w% (~w/~w)"/>~n', [PPreds,TCovPreds,Preds]),
    format(Stream, '            <coverage type="line, %" value="0% (0/0)"/>~n~n', []), % TODO?
    emma_data(Stream,ModuleGroup),
    format(Stream, '        </package>~n~n', []),
    fail.
emma_data(_S).

emma_data(Stream,ModuleGroup) :-
    module_row(ModuleGroup,TF,Clauses,CovClauses,IntClauses,_PercentageClauses,Preds,CovPreds,IntPreds,_PercentagePreds),
    TCovClauses is CovClauses + IntClauses,
    PClauses is TCovClauses/Clauses*100,
    (PClauses > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Clauses') ; true),
    TCovPreds is CovPreds + IntPreds,
    PPreds is TCovPreds/Preds*100,
    (PPreds > 100 -> junit_coverage_report_error(emma,'Covered more than 100% of all Predicates') ; true),
    format(Stream, '            <srcfile name="~w">~n', [TF]),
    format(Stream, '                <coverage type="class, %" value="100% (1/1)"/>~n', []),
    format(Stream, '                <coverage type="method, %" value="~w% (~w/~w)"/>~n', [PClauses,TCovClauses,Clauses]),
    format(Stream, '                <coverage type="block, %" value="~w% (~w/~w)"/>~n', [PPreds,TCovPreds,Preds]),
    format(Stream, '                <coverage type="line, %" value="0% (0/0)"/>~n~n', []), % TODO?
    format(Stream, '                <class name="~w">~n', [TF]),
    format(Stream, '                    <coverage type="class, %" value="100% (1/1)"/>~n', []),
    format(Stream, '                    <coverage type="method, %" value="~w% (~w/~w)"/>~n', [PClauses,TCovClauses,Clauses]),
    format(Stream, '                    <coverage type="block, %" value="~w% (~w/~w)"/>~n', [PPreds,TCovPreds,Preds]),
    format(Stream, '                    <coverage type="line, %" value="0% (0/0)"/>~n~n', []), % TODO?
    format(Stream, '                </class>~n', []),
    format(Stream, '            </srcfile>~n', []),
    fail.
emma_data(_S,_M).

export_emma(Filename) :-
    open(Filename,write,Stream),
    write(Stream,'<?xml version="1.0" encoding="UTF-8"?>\n'),
    write(Stream,'<report>\n'),
    emma_header(Stream),
    emma_data(Stream),
    write(Stream, '    </all>\n'),
    write(Stream, '  </data>\n'),
    write(Stream, '</report>\n'),
    close(Stream).

junit_coverage_report_error(Id,ErrorText) :-
    create_junit_result(Id,'Coverage Reports','Coverage_Reports',0,error([ErrorText]),Result),
    print_junit([Result],'CoverageReports').
