% (c) 2012-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(external_functions,[call_external_function/6, call_external_predicate/8,
                              call_external_substitution/6,
                              reset_external_functions/0,
                              external_function_argument_list/2,
                              external_pred_always_true/1,
                              check_if_predicates_were_evaluated/1,
                              reset_side_effect_occurred/0,
                              side_effect_occurred/1,
                              performs_io/1, not_declarative/1,
                              is_external_function_name/1,
                              external_subst_enabling_condition/3,
                              external_fun_has_wd_condition/1,
                              external_fun_can_be_inverted/1,
                              do_not_evaluate_args/1,
                              reset_argv/0, set_argv_from_list/1, set_argv_from_atom/1,

                              observe_parameters/2, observe_state/1, observe_value/2
                              ]).

% Note: there is also the predicate external_procedure_used/1 in bmachine_construction

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(avl)).
:- use_module(library(lists)).
:- use_module(translate).

% TO DO: add external Prolog files
% Should we do auto-conversion of B to Prolog types ,... ?

:- use_module(kernel_waitflags).
:- use_module(debug).
:- use_module(bsyntaxtree).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides a way to call external functions and predicates.').

:- dynamic side_effect_occurred/1.
% TRUE if a side effect has occured
% possible values (user_output, file)
reset_side_effect_occurred :- retractall(side_effect_occurred(_)).
assert_side_effect_occurred(Location) :-
  (side_effect_occurred(Location) -> true ; assert(side_effect_occurred(Location))).

:- discontiguous expects_waitflag/1.
:- discontiguous expects_spaninfo/1.
:- discontiguous expects_type/1. % only used for external functions, not predicates or subst
:- discontiguous performs_io/1.
:- discontiguous is_not_declarative/1. % does not perform io but is still not declarative
:- discontiguous does_not_modify_state/1.
:- discontiguous external_function_argument_list/2.
:- discontiguous expects_unevaluated_args/1. % currently only used for external predicates
:- discontiguous do_not_evaluate_args/1. % currently only used for external predicates
:- discontiguous expects_state/1. % currently only used for external predicates
:- discontiguous external_pred_always_true/1. % external predicates which are always true
:- discontiguous external_subst_enabling_condition/3. % external predicates which are always true
:- discontiguous external_fun_has_wd_condition/1.
:- discontiguous external_fun_can_be_inverted/1.
:- discontiguous external_fun_type/3. % external_fun_type(FUN,ListOfTypeParameters,ListOfArgTypes); also applicable to external predicates

:- meta_predicate safe_call(0,-).

not_declarative(FunName) :- (performs_io(FunName) -> true ; is_not_declarative(FunName)).

is_external_function_name(Name) :- external_fun_type(Name,_,_).
is_external_function_name(Name) :- expects_waitflag(Name), \+ external_fun_type(Name,_,_).
is_external_function_name(Name) :- expects_spaninfo(Name), \+ expects_waitflag(Name), \+ external_fun_type(Name,_,_).


:- assert_must_succeed(( external_functions:call_external_function('COS',[int(0)],int(100),integer,unknown,_WF) )).
:- assert_must_succeed(( external_functions:call_external_function('SIN',[int(0)],int(0),integer,unknown,_WF) )).
% first a few hard-wired external functions for efficiency:
call_external_function('COS',[X],Value,_,_,_WF) :- !, 'COS'(X,Value). % on_exception ?
call_external_function('SIN',[X],Value,_,_,_WF) :- !, 'SIN'(X,Value).
call_external_function('TAN',[X],Value,_,_,_WF) :- !, 'TAN'(X,Value).
% now the generic case:
call_external_function(F,Args,Value,Type,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   (expects_type(F) -> E3=[Type|E2] ; E3=E2),
   append(Args,[Value|E3],AV),
    % Check if there is better way to do this
   Call =.. [F|AV],
   % print('EXTERNAL FUNCTION CALL: '), print(F), print(' '), print(Call),nl,
   safe_call(Call,SpanInfo)
   %, print(result(Value)),nl.
   .

call_external_predicate(F,Args,EvaluatedArgs,LocalState,State,PREDRES,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   (expects_unevaluated_args(F) -> E3=[Args|E2] ; E3=E2),
   (expects_state(F) -> E4=[LocalState,State|E3] ; E4=E3),
   append(EvaluatedArgs,[PREDRES|E4],AV),
   Call =.. [F|AV],
   % print('EXTERNAL PREDICATE CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).

call_external_substitution(FunName,Args,InState,OutState,Info,WF) :-
 % TO DO: signal if substitution can be backtracked
   get_wait_flag_for_subst(FunName,WF,Flag),
   (does_not_modify_state(FunName) -> OutState=[] ; OutState = OutState2),
   call_external_substitution2(Flag,FunName,Args,InState,OutState2,Info,WF).

get_wait_flag_for_subst(FunName,WF,Flag) :- perform_directly(FunName),!,
   get_wait_flag1(WF,Flag). % ensure that precondition, ... have been evaluated
get_wait_flag_for_subst(FunName,WF,Flag) :- performs_io(FunName),!,
   get_enumeration_finished_wait_flag(WF,Flag). % ensure that precondition, ... have been evaluated
%   get_last_wait_flag(external(FunName),WF,Flag).
get_wait_flag_for_subst(FunName,WF,Flag) :- get_binary_choice_wait_flag(external(FunName),WF,Flag).

:- block call_external_substitution2(-,?,?,?,?,?,?).
call_external_substitution2(_,F,Args,InState,OutState,_Info,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[InState,OutState|E2],AV),
   Call =.. [F|AV],
   % print('EXTERNAL SUBSTITUTION CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).

safe_call(X,SpanInfo) :-
         on_exception(Exception, call(X),
                (add_exception_error(Exception,X,SpanInfo),fail)).

%safe_is(Context,X,Y) :- safe_is(Context,X,Y,unknown).
safe_is(Context,X,Y,Span) :-   % print(safe_is(Context,X,Y)),nl,
         on_exception(Exception, X is Y,
                (add_exception_error(Exception,Context,Span),fail)). %, print(res(X)),nl.

add_exception_error(error(existence_error(procedure,external_functions:FUN),_),X,SpanInfo) :- !,
   add_error(external_functions,'### EXTERNAL function does not exist: ',FUN:X,SpanInfo).
add_exception_error(error(undefined,ERR),X,SpanInfo) :-
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(undefined),ERR),X,SpanInfo) :-
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(zero_divisor),ERR),X,SpanInfo) :-
   is_error(ERR),!,
   % TO DO: use add_wd_error_set_result:
   add_error(external_functions,'### Division by 0 in arithmetic operation: ',X,SpanInfo).
add_exception_error(error(existence_error(source_sink,File),_),X,SpanInfo) :- !,
   add_error(external_functions,'### File does not exist (calling external function): ',File:X,SpanInfo).
add_exception_error(error(consistency_error(Msg,_,format_arguments),_),_X,SpanInfo) :- !,
   add_error(external_functions,'### Illegal format string (see SICStus Prolog reference manual for syntax): ',Msg,SpanInfo).
add_exception_error(enumeration_warning(A,B,C,D,E),X,SpanInfo) :-
   Exception = enumeration_warning(A,B,C,D,E),!,
   add_warning(external_functions,'### Enumeration warning occured while calling external function: ',X:Exception,SpanInfo),
   % or should we just print a message on the console ?
   throw(Exception).
add_exception_error(user_interrupt_signal,_X,_SpanInfo) :- !,
    print('User Interrupt during external function evaluation'),nl,
    throw(user_interrupt_signal).
add_exception_error(Exception,X,SpanInfo) :-
   add_error(external_functions,'### Exception occurred while calling external function: ',X:Exception,SpanInfo).

is_error(evaluation_error('is'(_,_),2,_,_)).
is_error(domain_error('is'(_,_),2,_,_)).

% STANDARD EXTERNAL FUNCTIONS


% -------------------------------

% MATHEMATICAL FUNCTIONS

external_fun_type('COS',[],[integer,integer]).
:- public 'COS'/2.
:- block 'COS'(-,?). % redundant; we could add CLPFD constraints for R
'COS'(int(X),int(R)) :- cos2(100,X,R,unknown).
external_fun_type('COSx',[],[integer,integer,integer]).
expects_spaninfo('COSx').
:- public 'COSx'/4.
:- block 'COSx'(-,?,?,?),'COSx'(?,-,?,?).
'COSx'(int(PrecisionMultiplier),int(X),int(R),Span) :- cos2(PrecisionMultiplier,X,R,Span).
:- block cos2(-,?,?,?),cos2(?,-,?,?).
cos2(Multiplier,X,R,Span) :- safe_is('COSx',R,round(Multiplier*cos(X/Multiplier)),Span).

external_fun_type('SIN',[],[integer,integer]).
:- public 'SIN'/2.
:- block 'SIN'(-,?). % redundant; we could add CLPFD constraints for R
'SIN'(int(X),int(R)) :- sin2(100,X,R,unknown).
expects_spaninfo('SINx').
:- public 'SINx'/4.
:- block 'SINx'(-,?,?,?),'SINx'(?,-,?,?).
'SINx'(int(PrecisionMultiplier),int(X),int(R),Span) :- %print('SINx'(PrecisionMultiplier,X,R)),nl,
  sin2(PrecisionMultiplier,X,R,Span).
:- block sin2(-,?,?,?),sin2(?,-,?,?).
sin2(Multiplier,X,R,Span) :- safe_is('SINx',R,round(Multiplier*sin(X/Multiplier)),Span).

external_fun_type('TAN',[],[integer,integer]).
:- public 'TAN'/2.
:- block 'TAN'(-,?). % redundant; we could add CLPFD constraints for R
'TAN'(int(X),int(R)) :- tan2(100,X,R,unknown).
external_fun_type('TANx',[],[integer,integer,integer]).
expects_spaninfo('TANx').
:- public 'TANx'/4.
:- block 'TANx'(-,?,?,?),'TANx'(?,-,?,?).
'TANx'(int(PrecisionMultiplier),int(X),int(R),Span) :- tan2(PrecisionMultiplier,X,R,Span).
:- block tan2(?,-,?,?),tan2(-,?,?,?).
tan2(Multiplier,X,R,Span) :- safe_is('TANx',R,round(Multiplier*tan(X/Multiplier)),Span).


external_fun_type('PROLOG_FUN',[],[string,integer,integer,integer]).
external_fun_has_wd_condition('PROLOG_FUN').
expects_spaninfo('PROLOG_FUN').
:- public 'PROLOG_FUN'/5.
'PROLOG_FUN'(string(Function),int(PrecisionMultiplier),int(Argument),int(Result),Span) :-
  prolog_fun2(Function,PrecisionMultiplier,Argument,Result,Span).
% Function can be:
% sin, cos, tan, cot, sinh, cosh, tanh, asin, acos, atan, acot, asinh, acosh, atanh, acoth, sqrt, log, exp
:- block prolog_fun2(?,-,?,?,?),prolog_fun2(-,?,?,?,?), prolog_fun2(?,?,-,?,?).
prolog_fun2(Function,Multiplier,X,R,Span) :-
   E =.. [Function,X/Multiplier],
   safe_is(Function,R,round(Multiplier*E),Span).

external_fun_has_wd_condition('LOGx').
expects_spaninfo('LOGx').
:- public 'LOGx'/5.
:- block 'LOGx'(-,?,?,?,?),'LOGx'(?,-,?,?,?), 'LOGx'(?,?,-,?,?).
'LOGx'(int(PrecisionMultiplier),int(Base),int(X),int(R),Span) :-
  log2(PrecisionMultiplier,Base,X,R,Span).
:- block log2(?,-,?,?,?),log2(-,?,?,?,?),log2(?,?,-,?,?).
log2(Multiplier,Base,X,R,Span) :-
  safe_is('LOGx',R,round(Multiplier*log(Base/Multiplier,X/Multiplier)),Span).

external_fun_type('GCD',[],[integer,integer,integer]).
:- public 'GCD'/3.
:- block 'GCD'(-,?,?),'GCD'(?,-,?).
'GCD'(int(X),int(Y),int(R)) :- gcd2(X,Y,R).
:- block gcd2(?,-,?),gcd2(-,?,?).
% we could assert abs(X) #>= R, abs(Y) #>= R
gcd2(X,Y,R) :- R is gcd(X,Y).

external_fun_type('ABS',[],[integer,integer]).
:- public 'ABS'/2.
:- use_module(clpfd_interface,[clpfd_eq_expr/2]).
:- block 'ABS'(-,-). % do we need this ??
'ABS'(int(X),int(R)) :-
   clpfd_eq_expr(R,abs(X)).


external_fun_type('MSB',[],[integer,integer]).
expects_spaninfo('MSB').
:- public 'MSB'/3.
:- block 'MSB'(-,?,?).
'MSB'(int(X),int(R),Span) :- msb2(X,R,Span). % most significant bit
:- block msb2(-,?,?).
msb2(X,R,Span) :- safe_is('MSB',R,msb(X),Span). % TO DO: could be made reversible by propagating constraint if R known

% -------------------------------

% THE CHOOSE operator (for TLA, recursive functions over sets,...)
% or choice operator from Abrial Book, page 60ff

:- use_module(custom_explicit_sets).
:- use_module(kernel_waitflags,[add_wd_error_span/4]).
external_fun_type('CHOOSE',[X],[set(X),X]).
external_fun_has_wd_condition('CHOOSE').
expects_waitflag('CHOOSE').
expects_spaninfo('CHOOSE').
:- public 'CHOOSE'/4.
:- block 'CHOOSE'(-,?,?,?).
'CHOOSE'([],_,Span,WF) :-
   add_wd_error_span('CHOOSE applied to empty set: ',[],Span,WF).
'CHOOSE'(CS,H,_,_) :- is_interval_closure_or_integerset(CS,Low,Up),
       !,
       choose_interval(Low,Up,H).
'CHOOSE'([H|T],C,Span,WF) :- !,
       when(ground([H|T]),
           (convert_to_avl([H|T],AVL) -> 'CHOOSE'(AVL,C,Span,WF)
             ; add_wd_error_span('Cannot determine minimum element for CHOOSE: ',[H|T],Span,WF),
               kernel_objects:equal_object(C,H))
            ).
'CHOOSE'(avl_set(A),H,_,_) :- !, avl_min(A,Min),kernel_objects:equal_object(H,Min).
'CHOOSE'(global_set(GS),H,_,_) :- !, H = fd(1,GS). % assume GS not empty, integer sets covered above
'CHOOSE'(freetype(ID),H,Span,WF) :- !,
    (kernel_freetypes:enumerate_freetype(basic,FV,freetype(ID))
     -> kernel_objects:equal_object(FV,H)
     ; add_internal_error('Could not generate freetype element: ','CHOOSE'(freetype(ID),H,Span,WF))).
'CHOOSE'(CS,H,Span,WF) :- custom_explicit_sets:is_cartesian_product_closure(CS,Set1,Set2),!,
     'CHOOSE'(Set1,H1,Span,WF),
     'CHOOSE'(Set2,H2,Span,WF),
     kernel_objects:equal_object(H,(H1,H2)).
'CHOOSE'(closure(P,T,B),H,Span,WF) :-
     custom_explicit_sets:expand_custom_set(closure(P,T,B),ER),
    'CHOOSE'(ER,H,Span,WF).

:- block choose_interval(-,?,?).
choose_interval(Low,_Up,H) :- number(Low),!,H=int(Low).
choose_interval(_Low,Up,H) :- choose_interval2(Up,H). % lower-bound is minus infinity
:- block choose_interval2(-,?).
choose_interval2(Up,H) :- number(Up),!,H=int(Up). % for those intervals we choose Up
choose_interval2(_,int(0)). % we have INTEGER or something equivalent; simply choose 0


% sort a set and generate a sequence
external_fun_has_wd_condition('SORT').
expects_waitflag('SORT').
expects_spaninfo('SORT').
external_fun_type('SORT',[X],[set(X),seq(X)]).
:- public 'SORT'/4.
:- block 'SORT'(-,?,?,?).
'SORT'([],Res,_Span,_WF) :- !, Res=[].
'SORT'(CS,Res,_,WF) :- custom_explicit_sets:is_custom_explicit_set(CS),!,
    custom_explicit_sets:expand_custom_set_to_list(CS,ER,Done,'SORT'),
    sort_aux(Done,ER,Res,WF).
'SORT'([H|T],Res,_,WF) :- when(ground([H|T]),(sort([H|T],ER),sort_aux(done,ER,Res,WF))).

:- block sort_aux(-,?,?,?).
sort_aux(_,ER,Res,_WF) :- convert_set_to_seq(ER,Res).

convert_set_to_seq(ER,Res) :- convert_set_to_seq(ER,1,Seq), equal_object_optimized(Seq,Res).

convert_set_to_seq([],_,[]).
convert_set_to_seq([H|T],N,[(int(N),H)|CT]) :- N1 is N+1, convert_set_to_seq(T,N1,CT).

% Prolog term order on encoding
external_fun_type('LESS',[X],[X,X,boolean]).
:- public 'LESS'/3.
:- block 'LESS'(-,?,?), 'LESS'(?,-,?).
'LESS'(string(A),string(B),Res) :- !, less2(A,B,Res).
'LESS'(int(A),int(B),Res) :- !, less_clpfd(A,B,Res). % use CLPFD less ?
'LESS'(fd(A,G),fd(B,G),Res) :- !, less_clpfd(A,B,Res).
'LESS'(pred_false,B,Res) :- !, B=Res. % if B=pred_true then < , otherwise not
'LESS'(pred_true,_,Res) :- !, Res=pred_false.
% TO DO: expand finite closures, couples, avl_sets, ...
'LESS'(X,Y,Res) :- when((ground(X),ground(Y)), less2(X,Y,Res)).
% TO DO: use kernel_ordering,[ordered_value/2]). 

:- use_module(clpfd_interface,[clpfd_leq_expr/2,clpfd_lt_expr/2]).
:- block less_clpfd(-,?,-), less_clpfd(?,-,-).
less_clpfd(A,B,Res) :- Res==pred_true, !, clpfd_lt_expr(A,B).
less_clpfd(A,B,Res) :- Res==pred_false, !, clpfd_leq_expr(B,A).
less_clpfd(A,B,Res) :- (A < B -> Res=pred_true ; Res=pred_false).


:- block less2(-,?,?), less2(?,-,?).
less2(A,B,Res) :- (A @< B -> Res=pred_true ; Res=pred_false).

:- use_module(kernel_ordering,[leq_ordered_value/3]).

external_fun_type('LEQ_SYM_BREAK',[X],[X,X,boolean]).
:- public 'LEQ_SYM_BREAK'/3.
:- block 'LEQ_SYM_BREAK'(-,?,?), 'LEQ_SYM_BREAK'(?,-,?).
% true if arg1 <= arg2 according to Prolog term order; may return true for some symbolic representations
%'LEQ_SYM_BREAK'(X,Y,Res) :- print('LEQ_SYM_BREAK'(X,Y,Res)),nl,fail.
'LEQ_SYM_BREAK'(string(A),string(B),Res) :- !, less_eq2(A,B,Res).
'LEQ_SYM_BREAK'(int(A),int(B),Res) :- !, less_eq2(A,B,Res). % use CLPFD less ?
'LEQ_SYM_BREAK'(fd(A,G),fd(B,G),Res) :- !, less_eq2(A,B,Res). % use CLPFD less
'LEQ_SYM_BREAK'(pred_false,B,Res) :- !, less_eq2(pred_false,B,Res).
'LEQ_SYM_BREAK'(pred_true,B,Res) :- !, less_eq2(pred_true,B,Res).
% TO DO: expand finite closures, couples, avl_sets, ...
'LEQ_SYM_BREAK'(closure(_,_,_),_,Res) :- !, Res=pred_true. % otherwise we would have to expand it
'LEQ_SYM_BREAK'(_,closure(_,_,_),Res) :- !, Res=pred_true.
'LEQ_SYM_BREAK'((A,B),(X,Y),Res) :-  Res==pred_true,!, leq_ordered_value((A,B),(X,Y),_Equal).
'LEQ_SYM_BREAK'(X,Y,Res) :- when((ground(X),ground(Y)), less_eq2(X,Y,Res)). % unsafe to use; we should normalize AVLs ...
% TO DO: use kernel_ordering:leq_ordered_value/2 more 
% Note: it is reified in b_interpreter_check

:- block less_eq2(-,?,?), less_eq2(?,-,?).
less_eq2(A,B,Res) :- (A @=< B -> Res=pred_true ; Res=pred_false).

% -------------------------------

% a first prototype for implementing recursive functions
% still a bit clumsy to define recursive functions (using STRING parameter for names)
% a restriction is that we can only have one version per machine, i.e.
%  one can only define recursive function in the CONSTANTS section and only
%  in a deterministic fashion not depending on enumerated constants,...
% also: we still need to ensure that the recurisve closures are not expanded
%  this should probably go hand in hand with the symbolic pragma

reset_external_functions :-
    retractall(recursive_function(_,_)),
    retractall(evaluated_predicate(_)),
    retractall(last_walltime(_)),
    reset_profiler, % maybe we should do this only if profiler was used ?
    reset_ident,
    reset_kodkod,
    close_all_files.

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_external_functions,
                           'Reset B External Functions.').

:- volatile recursive_function/2.
:- dynamic recursive_function/2.

expects_waitflag('REC'). % DEPRECATED
expects_spaninfo('REC').
:- public 'REC'/5.
:- block 'REC'(?,-,?,?,?).
'REC'(string(FunName),ArgValue,Value,Span,WF) :-
  % print('REC'(FunName,ArgValue)),nl,
   (recursive_function(FunName,Fun)
    -> bsets_clp:apply_to(Fun,ArgValue,Value,Span,WF)
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

expects_spaninfo('RECx'). % DEPRECATED
:- public 'RECx'/3.
:- block 'RECx'(-,?,?).
'RECx'(string(FunName),Result,Span) :-
       print('DEPRECATED RECx'(FunName)),nl, %%
   (recursive_function(FunName,FunBody)
    -> kernel_objects:equal_object(Result,FunBody,'RECx')
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

:- use_module(custom_explicit_sets,[mark_closure_as_recursive/2]).
%expects_waitflag('REC_LET').
expects_spaninfo('REC_LET').  % DEPRECATED
external_function_argument_list('REC_LET',[normal,symbolic]).
:- public 'REC_LET'/4.
'REC_LET'(string(FunName),FunBody,Result,Span) :-
   when(ground(FunBody),
       ( mark_closure_as_recursive(FunBody,SFunBody),  % what we actually need is that FunBody is evaluated already with a symbolic annotation !
       % TO DO: also marks as possibly being non-wd (infinite loop; avoid over-eager computation by b_interpreter_check)
         store_recursive_function(FunName,SFunBody,Span),
         kernel_objects:equal_object(Result,FunBody,'REC_LET') % pass the argument through
         )).

store_recursive_function(FunName,FunBody,Span) :-
  print('STORING RECURSIVE FUNCTION : '), print(FunName),nl,  % print(FunBody),nl, %
  (recursive_function(FunName,OtherBody)
    -> (OtherBody=FunBody -> true
         ; add_error(external_functions,'### Multiple definitions of recursive function not yet allowed: ',FunName:FunBody:OtherBody,Span)
        )
    ;  assert(recursive_function(FunName,FunBody))
  ).

% -------------------------------

expects_waitflag('FORCE').
:- public 'FORCE'/3.
:- block 'FORCE'(-,?,?).
'FORCE'(avl_set(A),Result,_) :- !, equal_object(avl_set(A),Result).
'FORCE'([],Result,_) :- !, equal_object([],Result).
'FORCE'(closure(P,T,B),Result,_) :- !,
   print('FORCING : '),translate:print_bvalue(closure(P,T,B)),nl,
   custom_explicit_sets:expand_closure_to_avl_or_list(P,T,B,ExpandedValue,no_check),
   print('RESULT : '),translate:print_bvalue(ExpandedValue),nl,
   print('  ASSIGNING TO : '), translate:print_bvalue(Result),nl,
   equal_object(ExpandedValue,Result).
'FORCE'(A,B,_) :- equal_object_optimized(A,B,'FORCE').


% force enumeration of certain basic datatypes; at the moment this is quite limited : TO DO: extend
% maybe we also want to add priorities or several variables in one go (several variables can be sent as a pair though)
expects_waitflag('ENUMERATE').
:- public 'ENUMERATE'/3.
:- block 'ENUMERATE'(-,?,?).
'ENUMERATE'(Obj,pred_true,WF) :- get_wait_flag1(WF,WF1),
    format('*** ENUMERATE: ~w~n',Obj),
    enumerate_aux(Obj,WF1,WF).

:- block enumerate_aux(?,-,?).
enumerate_aux(fd(X,GS),_,WF) :-  !, kernel_objects:enumerate_basic_type_wf(fd(X,GS),global(GS),WF).
enumerate_aux(int(X),_,WF) :-  !, kernel_objects:enumerate_basic_type_wf(int(X),integer,WF).
enumerate_aux((A,B),WF1,WF) :- !, enumerate_aux2(A,WF1,WF), enumerate_aux2(B,WF1,WF).
enumerate_aux(Obj,_,_WF) :- add_error(external_functions,'Cannot enumerate value yet: ',Obj).

:- block enumerate_aux2(-,?,?).
enumerate_aux2(Obj,WF1,WF) :- enumerate_aux(Obj,WF1,WF).

% a predicate to ensure that certain variables are enumerated together:
% as soon as first value is known, the second value gets enumerated
expects_waitflag('LINKED_ENUMERATION').
:- public 'LINKED_ENUMERATION'/4.
:- block 'LINKED_ENUMERATION'(-,?,?,?).
'LINKED_ENUMERATION'(Obj1,Obj2,pred_true,WF) :- link_enum_aux(Obj1,Obj2,WF).

link_enum_aux(fd(X,_),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(int(X),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(string(X),Obj2,WF) :- !, link_enum_block(X,Obj2,WF).
link_enum_aux(Obj1,Obj2,WF) :- when(ground(Obj1),link_enum_block(1,Obj2,WF)).

:- block link_enum_block(-,?,?).
link_enum_block(_X,Obj2,WF) :- %print(linked_enumeration(_X,Obj2)), tools_printing:print_arg(Obj2),nl,
   get_wait_flag1(WF,WF1), % give chance to run deterministic propagations
   enumerate_aux(Obj2,WF1,WF).

expects_unevaluated_args('DO_NOT_ENUMERATE').
:- public 'DO_NOT_ENUMERATE'/3.
:- block 'DO_NOT_ENUMERATE'(-,?,?).
'DO_NOT_ENUMERATE'((A,B),PredRes,UEA) :- !, 'DO_NOT_ENUMERATE'(A,PredRes,UEA),'DO_NOT_ENUMERATE'(B,PredRes,UEA).
'DO_NOT_ENUMERATE'(fd(V,_),PredRes,UEA) :- !, 'DO_NOT_ENUMERATE'(V,PredRes,UEA).
'DO_NOT_ENUMERATE'(int(V),PredRes,UEA) :- !, 'DO_NOT_ENUMERATE'(V,PredRes,UEA).
'DO_NOT_ENUMERATE'(string(V),PredRes,UEA) :- !, 'DO_NOT_ENUMERATE'(V,PredRes,UEA).
'DO_NOT_ENUMERATE'(_Obj,PredRes,_UEA) :- 
   %print('INSTANTIATED: '),nl, _UEA=[UnEvalArg],translate:print_bexpr(UnEvalArg),nl, print(_Obj),nl,
   PredRes=pred_true.
   % add_message('DO_NOT_ENUMERATE','Enumerating: ',UnEvalArgs).


% -------------------------------

:- assert_must_succeed(( external_functions:call_external_function('STRING_APPEND',[string(a),string(b)],string(ab),string,unknown,_WF) )).
% STRING MANIPULATION FUNCTIONS
:- use_module(kernel_waitflags).
% function to append two strings, reversible
external_fun_can_be_inverted('STRING_APPEND').
expects_waitflag('STRING_APPEND').
external_fun_type('STRING_APPEND',[],[string,string,string]).
:- public 'STRING_APPEND'/4.
:- block 'STRING_APPEND'(-,?,-,?), 'STRING_APPEND'(?,-,-,?).
'STRING_APPEND'(string(A),string(B),string(C),WF) :- %print(app(A,B,C)),nl,
  app2(A,B,C,WF).
:- block app2(-,?,-,?), app2(?,-,-,?).
app2(A,B,C,_WF) :- %print(app2(A,B,C)),nl,
   nonvar(A),nonvar(B),!,atom_concat(A,B,CC), /* overcome bug in SICStus; delay unification */
   CC=C. %, print(app2(A,B,C)),nl.
app2(A,B,C,WF) :- atom_codes(C,CC), app3(A,B,CC,WF).
app3(A,B,CC,WF) :-
   (   nonvar(B) -> atom_codes(B,BB), append(AA,BB,CC), atom_codes(A,AA)
     ; nonvar(A) -> atom_codes(A,AA), append(AA,BB,CC), atom_codes(B,BB)
     ; length(CC,CLen), %print(enum(CC,CLen)),nl,
       get_wait_flag(CLen,'STRING_APPEND',WF,LWF),
       app4(A,B,CC,WF,LWF)
    ).
:- block app4(-,?,-,?,-), app4(?,-,-,?,-).
app4(A,B,CC,WF,_) :- (nonvar(A) ; nonvar(B)), !, % no need to enumerate
   app3(A,B,CC,WF).
app4(A,B,CC,_WF,_) :- %print(enumerating(CC)),nl,
   append(AA,BB,CC), % will be non-deterministic
   atom_codes(A,AA), atom_codes(B,BB).

external_fun_can_be_inverted('STRING_CHARS').
external_fun_type('STRING_CHARS',[],[string,seq(string)]).
:- public 'STRING_CHARS'/2.
:- block 'STRING_CHARS'(-,-).
% to do: make it reversible
'STRING_CHARS'(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
'STRING_CHARS'(Str,SeqRes) :- when((ground(Str);ground(SeqRes)),string_chars1(Str,SeqRes)).

string_chars1(Str,SeqRes) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars1(Str,SeqRes) :-
   custom_explicit_sets:expand_custom_set_to_list(SeqRes,ExpandedAtomList,Done,string_chars1),
   string_chars3(Str,SeqRes,ExpandedAtomList,Done).

string_chars2(A,SeqRes) :- atom_codes(A,AA), generate_char_seq(AA,1,CharSeq),
   equal_object_optimized(CharSeq,SeqRes,string_chars2).

:- use_module(probsrc(tools_strings),[ajoin/2]).

:- block string_chars3(-,?,?,-).
string_chars3(Str,SeqRes,_ExpandedAtomList,_Done) :- nonvar(Str), Str=string(A), ground(A),!,
   string_chars2(A,SeqRes).
string_chars3(Str,_SeqRes,ExpandedAtomList,Done) :- nonvar(Done),!,
   sort(ExpandedAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
      (ajoin(PL,Atom),
       equal_object(Str,string(Atom)))).
string_chars3(Str,SeqRes,ExpandedAtomList,Done) :- % Str is only partially instantiated
   when((ground(Str);nonvar(Done)),string_chars3(Str,SeqRes,ExpandedAtomList,Done)).
generate_char_seq([],_,[]).
generate_char_seq([Code|T],Nr,[(int(Nr),string(CS))|TSeq]) :-
   atom_codes(CS,[Code]),
   N1 is Nr+1, generate_char_seq(T,N1,TSeq).

external_fun_can_be_inverted('STRING_CODES').
external_fun_type('STRING_CODES',[],[string,seq(integer)]).
:- public 'STRING_CODES'/2.
:- block 'STRING_CODES'(-,-).
'STRING_CODES'(string(A),SeqRes) :- string_codes2(A,SeqRes).
:- block string_codes2(-,-).
string_codes2(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes2(A,SeqRes) :- SeqRes==[],!, empty_string_atom(A).
string_codes2(A,SeqRes) :- expand_custom_set_to_list(SeqRes,SeqList,_,string_codes2),
   when((nonvar(A);ground(SeqList)), string_codes3(A,SeqList)).

string_codes3(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes3(A,SeqRes) :- sort(SeqRes,SSeqRes),extract_codes(SSeqRes,1,Codes),
   atom_codes(A,Codes).
string_codes4(A,SeqRes) :- atom_codes(A,AA), generate_code_seq(AA,1,CodeSeq),
   equal_object_optimized(CodeSeq,SeqRes,string_codes4).

generate_code_seq([],_,[]).
generate_code_seq([Code|T],Nr,[(int(Nr),int(Code))|TSeq]) :-
   N1 is Nr+1, generate_code_seq(T,N1,TSeq).

extract_codes([],_,[]).
extract_codes([(int(Nr),int(Code))|T],N,[Code|CT]) :-
  (Nr==N -> true ; add_error(extract_codes,'Unexpected index: ',(Nr,N))),
  N1 is N+1, extract_codes(T,N1,CT).


empty_string(string('')).
empty_string_atom('').

:- assert_must_succeed(( external_functions:call_external_function('STRING_LENGTH',[string(a)],int(1),integer,unknown,_WF) )).
% function to get length of a string
external_fun_type('STRING_LENGTH',[],[string,integer]).
:- public 'STRING_LENGTH'/2.
'STRING_LENGTH'(SA,int(L)) :- greater_than_equal(int(L),int(0)),
   string_len2(SA,L).
:- block string_len2(-,-).
string_len2(SA,L)  :- L==0,!, empty_string(SA).
string_len2(string(A),L) :- string_len3(A,L).
:- block string_len3(-,-).
string_len3(A,L)  :- L==0,!, empty_string_atom(A).
% in case A is not known and L=1 we could enumerate chars ??
string_len3(A,L) :- string_len4(A,L).
:- block string_len4(-,?).
string_len4(A,L) :- atom_length(A,LL), LL=L. % delay unification due to bug in SICStus atom_length
% bug in SICStus: dif(X,1), atom_length(a,X) succeeds in 4.2.0 and 4.2.1

% function to split a string into a list of strings which were delimited by a separator
% WARNING: if the seperator is of length more than one, then first match-strategy will be used
external_fun_can_be_inverted('STRING_SPLIT').
:- public 'STRING_SPLIT'/3.
:- block 'STRING_SPLIT'(?,-,?), 'STRING_SPLIT'(-,?,-).
'STRING_SPLIT'(string(A),string(B),R) :- string_split2(A,B,R).
:- block string_split2(?,-,?),string_split2(-,?,-).
string_split2(Atom,Seperator,SplitAtomList) :-
  (var(Atom) ; var(Seperator)),!, % currently : separator always known
  custom_explicit_sets:expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_split3(Atom,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).
string_split2(Atom,Separator,SplitAtomList) :- % normal forward mode: atom and separator known:
  string_split_forward(Atom,Separator,SplitAtomList).

string_split_forward(Atom,Separator,SplitAtomList) :-
  split_atom_string(Atom,Separator,List), % safe_call ?
  convert_prolog_to_b_list(List,SplitAtomList).

:- block string_split3(?,-,?,?,?),string_split3(-,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_split3(Atom,Seperator,SplitAtomList,Done,_ExpandedSplitAtomList) :- var(Done),!,
   string_split_forward(Atom,Seperator,SplitAtomList).
string_split3(Atom,Seperator,SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done),
   atom_codes(Seperator,SepCodes),
   append(SepCodes,_,SepCodes2),
   split4(Done,SepCodes2,PL,Seperator,Atom,SplitAtomList).

:- block split4(-,?,?,?,-,?), split4(-,?,?,-,?,?).
% unblock either when Done or when both Atom and Sperator known
split4(Done,_SepCodes2,_PL,Seperator,Atom,SplitAtomList) :- var(Done),!,
   % we can now compute forwards anyhow; ignore backwards direction
   string_split_forward(Atom,Seperator,SplitAtomList).
split4(_,SepCodes2,PL,Seperator,Atom,_) :-
      maplist(not_suffix_atom(SepCodes2),PL), % check that seperator occurs in no split atom: e.g. STRING_SPLIT(r,"_") = ["a","_","c"] should fail
       insert_sep(PL,Seperator,PL2),
       ajoin(PL2,Atom).

insert_sep([],_,[]).
insert_sep([H],_,R) :- !, R=[H].
insert_sep([H|T],Sep,[H,Sep|IT]) :- insert_sep(T,Sep,IT).

not_suffix_atom(SepCodes,Atom) :- \+ suffix_atom(SepCodes,Atom).
suffix_atom(SepCodes,Atom) :-
   atom_codes(Atom,AL),
   append(_,SepCodes,AL).

:- public 'STRING_JOIN'/3.
:- block 'STRING_JOIN'(?,-,?), 'STRING_JOIN'(-,?,?).
'STRING_JOIN'(SplitAtoms,string(Sep),Res) :- string_join2(SplitAtoms,Sep,Res).

% this is not reversible  ["a","b"],"_"  ->  "a_b"  but ["a_b"],"_" -> "a_b"
:- block string_join2(?,-,?),string_join2(-,?,?).
string_join2(SplitAtomList,Seperator,Result) :-
  custom_explicit_sets:expand_custom_set_to_list(SplitAtomList,ExpandedSplitAtomList,Done,string_split2),
  string_join3(Result,Seperator,SplitAtomList,Done,ExpandedSplitAtomList).

:- block string_join3(?,-,?,?,?),string_join3(?,?,?,-,?). % we need to know the seperator: TO DO : improve this
string_join3(Result,Seperator,_SplitAtomList,_Done,ExpandedSplitAtomList) :-
   ExpandedSplitAtomList \= [],  % split("",sep) --> [""] not the empty list
   sort(ExpandedSplitAtomList,SL),
   maplist(drop_index,SL,IL),
   convert_b_to_prolog_atoms(IL,PL,Done2),
   when(nonvar(Done2),
     (insert_sep(PL,Seperator,PL2),
      ajoin(PL2,Atom),
      equal_object(Result,string(Atom)))).


:- assert_must_succeed((external_functions:split_atom_string('ef,g',',',R), R==[ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab,cd,ef,g',',',R), R==['ab','cd',ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab','a',R), R==['','b'])).
:- assert_must_succeed((external_functions:split_atom_string('STRING1','',R), R==['STRING1'])).
:- assert_must_succeed((external_functions:split_atom_string('mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286',',',R), R==[mod274,mod276,mod277,mod282,mod283,mod284,mod285,mod286])).

split_atom_string(Atom,Sep,SplitList) :- %print(split_atom_string(Atom,Sep)),nl,
  atom_chars(Sep,SepAscii),
  (SepAscii=[] -> SplitList = [Atom]
   ; SepAscii = [H|T], atom_chars(Atom,ListAscii),
     split3(ListAscii,H,T,Match,Match,SplitList)).

% MatchSoFar is passed in two variables: one to instantiate and one with the Result of the match
% this avoids calling reverse
split3([],_,_,MatchSoFarIn,MatchSoFarRes,R) :- !,
   MatchSoFarIn=[],  % match complete, ground tail of match
   atom_chars(Atom,MatchSoFarRes),R=[Atom].
split3([H|List],H,Sep,MatchSoFarIn,MatchSoFarRes,Res) :-
   append(Sep,Tail,List),
   !, % we have a match with a separator
   MatchSoFarIn=[], % match complete
   atom_chars(Atom,MatchSoFarRes),
   Res=[Atom|R2], split3(Tail,H,Sep,NewMatch,NewMatch,R2).
split3([H|T],HS,Sep,[H|MatchSoFarIn],MatchSoFarRes,Res) :- % no match
   split3(T,HS,Sep,MatchSoFarIn,MatchSoFarRes,Res).


convert_prolog_to_b_list(PL,BL) :- convert_prolog_to_b_list_aux(PL,1,CPL),
  equal_object(CPL,BL,convert_prolog_to_b_list).

convert_prolog_to_b_list_aux([],_,[]).
convert_prolog_to_b_list_aux([H|T],Index,[(int(Index),CH)|CT]) :-
  convert_prolog_to_b_term(H,CH),
  I1 is Index+1, convert_prolog_to_b_list_aux(T,I1,CT).

convert_prolog_to_b_term(N,R) :- number(N),!, R=int(N).
convert_prolog_to_b_term(A,R) :- atomic(A),!, R=string(A).
convert_prolog_to_b_term(A,R) :-
   add_internal_error('Illegal Prolog term: ',convert_prolog_to_b_term(A,R)), R=A.


drop_index((int(_),R),R).

% a version that delays converting and sets Done to done when all B Atoms have been grounded
convert_b_to_prolog_atoms([],[],done).
convert_b_to_prolog_atoms([BAtom|T],[PrologAtom|PT],Done) :-
   convert_b_to_prolog_atoms_aux(BAtom,PrologAtom,DoneAtom),
   convert_b_to_prolog_atoms(T,PT,DoneT),
   both_done(DoneAtom,DoneT,Done).

:- block both_done(-,?,?), both_done(?,-,?).
both_done(_,_,done).

:- block convert_b_to_prolog_atoms_aux(-,?,?).
convert_b_to_prolog_atoms_aux(pred_true,'TRUE',done).
convert_b_to_prolog_atoms_aux(pred_false,'FALSE',done).
convert_b_to_prolog_atoms_aux(string(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).
convert_b_to_prolog_atoms_aux(int(S),PrologAtom,Done) :-
   convert_b_to_prolog_atoms_aux2(S,PrologAtom,Done).

:- block convert_b_to_prolog_atoms_aux2(-,?,?).
convert_b_to_prolog_atoms_aux2(Atom,Atom,done).


external_fun_can_be_inverted('STRING_TO_ENUM').
external_fun_has_wd_condition('STRING_TO_ENUM').
:- public 'STRING_TO_ENUM'/2.
:- block 'STRING_TO_ENUM'(-,-).
'STRING_TO_ENUM'(string(S),FD) :- string_to_enum2(S,FD).

:- block string_to_enum2(-,-).
string_to_enum2(S,FD) :- nonvar(S),!,
   (b_global_sets:is_b_global_constant(GS,Nr,S) ->
       FD = fd(Nr,GS)
     ; % TO DO: use add_wd_error_set_result
      add_error('Could not convert string to enumerated set element: ',S),fail).
string_to_enum2(S,fd(Nr,GS)) :-
   string_to_enum3(S,Nr,GS).

:- block string_to_enum3(-,-,?),string_to_enum3(-,?,-).
string_to_enum3(S,Nr,GS) :- nonvar(S),!,
   (b_global_sets:is_b_global_constant(SGS,SNr,S) -> (SGS,SNr) = (GS,Nr)
     ; add_error('Could not convert string to enumerated set element: ',S),fail).
string_to_enum3(S,Nr,GS) :-
   (b_global_sets:is_b_global_constant(GS,Nr,GNS) -> GNS=S
     ; add_error('Could not convert deferred set element to string: ',GS:Nr),fail).

:- public 'STRING_IS_INT'/2.
:- block 'STRING_IS_INT'(-,?).
'STRING_IS_INT'(string(S),I) :- string_is_int2(S,I).
:- block string_is_int2(-,?).
string_is_int2(S,Res) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
           (number_codes(Num,C),
            (integer(Num)->Res=pred_true;Res=pred_false)),
         Res=pred_false).

% A Utility to convert Decimal Strings to Integers
% TO DO: make a more precise version using just integers:
external_fun_has_wd_condition('DEC_STRING_TO_INT').
expects_spaninfo('DEC_STRING_TO_INT').
:- public 'DEC_STRING_TO_INT'/4.
:- block 'DEC_STRING_TO_INT'(-,?,?,?),'DEC_STRING_TO_INT'(?,-,?,?).
'DEC_STRING_TO_INT'(string(S),int(Precision),int(Res),Span) :- 
  dec_str_to_int(S,Precision,Res,Span).
   
:- block dec_str_to_int(-,?,?,?),dec_str_to_int(?,-,?,?).
dec_str_to_int(S,Precision,Res,Span) :- 
   atom_codes(S,Codes),
   my_codes_to_numbers(Codes,Number),
   (integer(Number) -> Res = Number
    ; safe_is('DEC_STRING_TO_INT',Res,round(Number*(10**Precision)),Span)).

% a version which accepts numbers like 11. or .5  (but not . on its own).
my_codes_to_numbers(Codes,Number) :- safe_number_codes(Number,Codes),!.
my_codes_to_numbers([46|Codes],Number) :- !, safe_number_codes(Number,[48,46|Codes]). % if number starts with . add leading 0
my_codes_to_numbers(Codes,Number) :- append(_,".",Codes),!, % number ends with ".", add 0
   append(Codes,"0",C0), safe_number_codes(Number,C0).

external_fun_has_wd_condition('STRING_TO_INT').
expects_waitflag('STRING_TO_INT').
expects_spaninfo('STRING_TO_INT').
:- public 'STRING_TO_INT'/4.
:- block 'STRING_TO_INT'(-,-,?,?).
'STRING_TO_INT'(string(S),int(I),Span,WF) :- string_to_int2(S,I,Span,WF).

:- block string_to_int2(-,-,?,?).
string_to_int2(S,Res,Span,WF) :- var(S), % we know the number; we cannot construct the string as leading spaces/0s are ok
   % with an injective string_to_int conversion we could invert the function
   !,
   frozen(S,Goal),  %print(s2int(S,Res,Goal)),nl,
   (incompatible_goal(Goal,S,Res) ->  %print(prune(S,Res)),nl,
     fail ; true),
   strint_to_int3(S,Res,Span,WF).
string_to_int2(S,Res,Span,WF) :- strint_to_int3(S,Res,Span,WF).

:- use_module(kernel_waitflags,[add_wd_error_set_result/6]).
:- block strint_to_int3(-,?,?,?).
strint_to_int3(S,Res,Span,WF) :-
    atom_codes(S,C),
    on_exception(error(syntax_error(_),_),
         integer_number_codes(C,S,Res,Span,WF),
         add_wd_error_set_result('### Could not convert string to integer: ',S,Res,0,Span,WF)).
         %add_error_and_fail(external_functions,'### Could not convert string to integer: ',S)),

integer_number_codes(C,S,Res,Span,WF) :- number_codes(Num,C),
  (integer(Num) -> Res=Num
   ; %add_error_and_fail(external_functions,'### String represents a floating point number (expected integer): ',S)).
     add_wd_error_set_result('### String represents a floating point number (expected integer): ',S,Res,0,Span,WF)).

% check if another pending co-routine transforms the same string into another number
incompatible_goal((A,B),S,Res) :- (incompatible_goal(A,S,Res) -> true ; incompatible_goal(B,S,Res)).
incompatible_goal(external_functions:strint_to_int3(S2,Res2,_,_),S,Res) :- number(Res2),
    S==S2, Res2 \= Res.


:- public 'INT_TO_STRING'/2.
:- block 'INT_TO_STRING'(-,?).
'INT_TO_STRING'(int(I),S) :- int_to_string2(I,S).

:- block int_to_string2(-,?).
int_to_string2(Num,Res) :- number_codes(Num,C), atom_codes(S,C), Res=string(S).

:- public 'TO_STRING'/2.
:- block 'TO_STRING'(-,?).
'TO_STRING'(int(I),S) :- !,int_to_string2(I,S).
'TO_STRING'(Value,S) :- ground_value_check(Value,GrValue), to_string_aux(GrValue,Value,S).

:- block to_string_aux(-,?,?).
to_string_aux(_,Value,Str) :- to_string_aux(Value,Str).

to_string_aux(Value,Str) :-
  translate_bvalue(Value,Atom), Str=string(Atom).
  

:- public 'FORMAT_TO_STRING'/3.
:- block 'FORMAT_TO_STRING'(-,?,?).
'FORMAT_TO_STRING'(string(FormatString),ListOfStrings,Res) :- 
    custom_explicit_sets:expand_custom_set_to_list(ListOfStrings,ESet,_,'FORMAT_TO_STRING'),
    ground_value_check(ESet,GrESet),
    format_to_string_aux(GrESet,FormatString,ESet,Res).

:- use_module(library(codesio),[format_to_codes/3]).
:- block format_to_string_aux(-,?,?,?).
format_to_string_aux(_,FormatString,ESet,Res) :-
   sort(ESet,SortedESet),
   maplist(get_string,SortedESet,Atoms),
   %print(format_to_codes(FormatString,Atoms,Codes)),nl,
   format_to_codes(FormatString,Atoms,Codes),
   atom_codes(Atom,Codes),
   Res = string(Atom).

get_string((_,string(S)),R) :- !,R=S.
get_string((_,X),R) :- to_string_aux(X,string(R)).


:- public 'HASH'/2.
:- block 'HASH'(-,?).
'HASH'(Value,Int) :-
    % TO DO: use ground_value_check
    when(ground(Value),(term_hash(Value,H),Int=int(H))).

% -------------------------------

% STRING FILE SYSTEM FUNCTIONS


:- use_module(kernel_objects).
:- use_module(library(file_systems)).
:- public 'FILES'/2.
:- block 'FILES'(-,?).
'FILES'(string(A),L) :- files2(A,L).
:- block files2(-,?).
files2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(F),file_member_of_directory(ADir,F,_FLong),L),
   equal_object_optimized(L,List,files2).
:- public 'FULL_FILES'/2.
:- block 'FULL_FILES'(-,?).
'FULL_FILES'(string(A),L) :- full_files2(A,L).
:- block full_files2(-,?).
full_files2(Dir,List) :-
   b_absolute_file_name(Dir,ADir),
   findall(string(FLong),file_member_of_directory(ADir,_F,FLong),L),
   equal_object_optimized(L,List,full_files2).


:- public 'DIRECTORIES'/2.
:- block 'DIRECTORIES'(-,?).
'DIRECTORIES'(string(A),L) :- dir2(A,L).
:- block dir2(-,?).
dir2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(F),directory_member_of_directory(ADir,F,_FLong),L),
   equal_object_optimized(L,List,dir2).
:- public 'FULL_DIRECTORIES'/2.
:- block 'FULL_DIRECTORIES'(-,?).
'FULL_DIRECTORIES'(string(A),L) :- full_dir2(A,L).
:- block full_dir2(-,?).
full_dir2(Dir,List) :- b_absolute_file_name(Dir,ADir),
   findall(string(FLong),directory_member_of_directory(ADir,_,FLong),L),
   equal_object_optimized(L,List,full_dir2).

:- public 'FILE_EXISTS'/2.
:- block 'FILE_EXISTS'(-,?).
'FILE_EXISTS'(string(A),PREDRES) :- file_exists2(A,PREDRES).
:- block file_exists2(-,?).
file_exists2(A,PREDRES) :- b_absolute_file_name(A,AA),
   (file_exists(AA) -> PREDRES=pred_true ; PREDRES=pred_false).
:- public 'DIRECTORY_EXISTS'/2.
:- block 'DIRECTORY_EXISTS'(-,?).
'DIRECTORY_EXISTS'(string(A),PREDRES) :- dir_exists2(A,PREDRES).
:- block dir_exists2(-,?).
dir_exists2(A,PREDRES) :-  b_absolute_file_name(A,AA),
   (directory_exists(AA) -> PREDRES=pred_true ; PREDRES=pred_false).


b_absolute_file_name(File,AbsFileName) :-
   bmachine:b_get_main_filename(MainFileName),
   absolute_file_name(File,AbsFileName,[relative_to(MainFileName)]).

:- public 'FILE_PROPERTY'/3.
:- block 'FILE_PROPERTY'(-,?,?), 'FILE_PROPERTY'(?,-,?).
'FILE_PROPERTY'(string(A),string(P),BoolRes) :-
   file_property2(A,P,BoolRes).
:- block file_property2(-,?,?), file_property2(?,-,?).
file_property2(File,Prop,BoolRes) :-
    b_absolute_file_name(File,AFile),
    file_property(AFile,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.
% possible properties: readable, writable, executable

:- public 'FILE_PROPERTY_VALUE'/3.
:- block 'FILE_PROPERTY_VALUE'(-,?,?), 'FILE_PROPERTY_VALUE'(?,-,?).
'FILE_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   file_property_val2(A,P,R).
:- block file_property_val2(-,?,?), file_property_val2(?,-,?).
file_property_val2(File,Prop,Res) :-
    b_absolute_file_name(File,AFile),
    file_property(AFile,Prop,Val) -> Res=Val
         ; add_error(external_functions,'Illegal FILE_PROPERTY',Prop),Res=int(-1).
% possible properties:
%  size_in_bytes, create_timestamp, modify_timestamp, access_timestamp
% + on unix the follwing is also an integer: owner_user_id, owner_group_id

:- public 'DIRECTORY_PROPERTY'/3.
:- block 'DIRECTORY_PROPERTY'(-,?,?), 'DIRECTORY_PROPERTY'(?,-,?).
'DIRECTORY_PROPERTY'(string(A),string(P),BoolRes) :-
   dir_property2(A,P,BoolRes).
:- block dir_property2(-,?,?), dir_property2(?,-,?).
dir_property2(File,Prop,BoolRes) :-
    b_absolute_file_name(File,AFile),
    directory_property(AFile,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.

:- public 'DIRECTORY_PROPERTY_VALUE'/3.
:- block 'DIRECTORY_PROPERTY_VALUE'(-,?,?), 'DIRECTORY_PROPERTY_VALUE'(?,-,?).
'DIRECTORY_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   dir_property_val2(A,P,R).
:- block dir_property_val2(-,?,?), dir_property_val2(?,-,?).
dir_property_val2(File,Prop,Res) :-
    b_absolute_file_name(File,AFile),
    directory_property(AFile,Prop,Val) -> Res=Val
           ; add_error(external_functions,'Illegal DIRECTORY_PROPERTY',Prop), Res=int(-1).
% --------------------------------

% I/O


% printf as predicate; can also be used as a function to BOOL
expects_spaninfo('fprintf').
external_pred_always_true('fprintf').
:- public fprintf/5.
:- block fprintf(-,?,?,?,?), fprintf(?,-,?,?,?), fprintf(?,?,-,?,?).
fprintf(string(File),string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
   fprintf2(File,FormatString,Value,Span)).
fprintf(File,F,V,P,S) :- add_internal_error('Illegal call: ',fprintf(File,F,V,P,S)), P=pred_true.

expects_spaninfo('printf_opt_trace').
external_pred_always_true('printf_opt_trace').
:- public printf_opt_trace/5.
:- assert_must_succeed(external_functions:printf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printf_opt_trace(-,?,?,?,?), printf_opt_trace(?,-,?,?,?).
printf_opt_trace(string(FormatString),Value,Trace,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),
          (Trace==pred_true -> trace ; true)
         )
         ).
printf_opt_trace(F,V,T,P,S) :- add_internal_error('Illegal call: ',printf_opt_trace(F,V,T,P,S)), P=pred_true.


expects_spaninfo('printf').
external_pred_always_true('printf').
:- assert_must_succeed(external_functions:printf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printf(-,?,?,?), printf(?,-,?,?).
printf(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         fprintf2(user_output,FormatString,Value,Span)).
printf(F,V,P,S) :- add_internal_error('Illegal call: ',printf(F,V,P,S)), P=pred_true.

expects_spaninfo('printf_two'). % print two values, triggers as soon as first value is known
external_pred_always_true('printf_two').
:- assert_must_succeed(external_functions:printf_two(string('test ok = ~w~n'),pred_true,string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- public printf_two/6.
:- block printf_two(-,?,?,?,?,?), printf_two(?,-,?,?,?,?).
printf_two(string(FormatString),Value,string(FormatString2),Value2,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),
          fprintf2(user_output,FormatString2,Value2,Span))).
printf_two(F,V,F2,V2,P,S) :- add_internal_error('Illegal call: ',printf_two(F,V,F2,V2,P,S)), P=pred_true.


expects_spaninfo('printf_nonvar'). % print as soon as value becomes nonvar (printf waits until the value is ground)
external_pred_always_true('printf_nonvar').
:- public printf_nonvar/4.
:- block printf_nonvar(-,?,?,?), printf_nonvar(?,-,?,?).
printf_nonvar(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),nonvar(Value)),
         fprintf2(user_output,FormatString,Value,Span)).
printf_nonvar(F,V,P,S) :- add_internal_error('Illegal call: ',printf(F,V,P,S)), P=pred_true.


expects_spaninfo('dprintf'). % a version of printf which delays setting itself to TRUE until printing is performed
external_pred_always_true('dprintf').
:- public dprintf/4.
:- block dprintf(-,?,?,?), dprintf(?,-,?,?).
dprintf(string(FormatString),Value,PredRes,Span) :- !,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),PredRes = pred_true)).
dprintf(F,V,P,S) :- add_internal_error('Illegal call: ',dprintf(F,V,P,S)), P=pred_true.

fprintf2(File,FormatString,Value,Span) :-
   translate:translate_bvalue(Value,TV),
   file_format(File,FormatString,[TV],Span),
   print_hash((FormatString,Value)),
   print_stats,
   fprint_log_info(File,Span).
%fprintf2(File,FormatString,Value,Span) :- print('BACKTRACKING ') , file_format(File,FormatString,['?'],Span),nl,fail.
% warning: format can call an argument as goal with ~@; this is dangerous
% and we should prohibit this
% Luckily PRINTF("@: <>~@<>~n","print(danger)") where print(danger)
% is arbitrary code is not a security exploit:
% our PRINTF translates the argument to an atom, i.e., one could
% only call Prolog predicates without arguments, anything else leads to an exception
% (even PRINTF("@: <>~@<>~n","nl") leads to an exception)

:- use_module(hashing,[my_term_hash/2]).
% optionally print hash and allow breakpoints
print_hash(V) :-
  external_fun_pref(printf_hash,'TRUE'),   % use SET_PREF("printf_hash","TRUE")
  !,
  my_term_hash(V,H), format('% HASH = ~w~n',H),
  %((H = 158201315 ; H=29759067) -> trace ; true),
  true.
print_hash(_).


:- dynamic nr_of_printfs/1.
nr_of_printfs(0).
print_stats :-
  external_fun_pref(printf_stats,PFS),   % use SET_PREF("printf_stats","TRUE")
  PFS='TRUE',
  retract(nr_of_printfs(N)),
  !,
  N1 is N+1,
  assert(nr_of_printfs(N1)),
  statistics(walltime,[X,SL]),
  statistics(memory_used,M), MB is M  / 1048576,
  statistics(gc_count,GCs),
  format('% Call ~w at walltime ~w ms (since last ~w ms), memory ~3f MB (~w GCs)~n',[N1,X,SL,MB,GCs]).
print_stats.

:- dynamic external_fun_pref/2.
% valid keys: log_info
valid_external_fun_pref(log_info).
valid_external_fun_pref(printf_stats).
valid_external_fun_pref(printf_hash).

fprint_log_info(File,Span) :- external_fun_pref(log_info,X),
   'GET_INFO'(string(X),string(Res)),
   file_format(File,'%%% ~w~n',[Res],Span).
fprint_log_info(_,_).

% example usage SET_PREF("log_info","time")
external_fun_type('SET_PREF',[],[string,string,boolean]).
:- public 'SET_PREF'/3.
:- block 'SET_PREF'(-,?,?), 'SET_PREF'(?,-,?).
'SET_PREF'(string(S),string(Val),R) :-
   preferences:eclipse_preference(S,Pref),!, R=pred_true,
   format('Setting ProB preference ~w := ~w~n',[S,Val]),
   preferences:set_preference(Pref,Val).
'SET_PREF'(string(S),string(Val),R) :- valid_external_fun_pref(S),!,
   R=pred_true,
   retractall(external_fun_pref(S,_)),
   format('Setting external function preference ~w := ~w~n',[S,Val]),
   assert(external_fun_pref(S,Val)).
'SET_PREF'(P,V,R) :-
   add_internal_error('Illegal call: ','SET_PREF'(P,V,R)), R=pred_false.

external_fun_type('GET_PREF',[],[string,string]).
expects_waitflag('GET_PREF').
expects_spaninfo('GET_PREF').
:- public 'GET_PREF'/4.
'GET_PREF'(Str,Res,Span,WF) :-
    kernel_waitflags:get_wait_flag(10,'GET_PREF',WF,WF10), % will enumerate possible input strings
    get_pref2(Str,Res,_,WF10,Span).
external_fun_type('GET_PREF_DEFAULT',[],[string,string]).
expects_waitflag('GET_PREF_DEFAULT').
expects_spaninfo('GET_PREF_DEFAULT').
:- public 'GET_PREF_DEFAULT'/4.
'GET_PREF_DEFAULT'(Str,Res,Span,WF) :-
    kernel_waitflags:get_wait_flag(10,'GET_PREF',WF,WF10), % will enumerate possible input strings
    get_pref2(Str,_,Res,WF10,Span).

:- block get_pref2(-,?,?,-,?).
get_pref2(string(S),Res,Res2,_,Span) :-
     if((preferences:eclipse_preference(S,Pref),
         preferences:get_preference(Pref,Val),
         preferences:preference_default_value(Pref,DefVal)),
         (make_string(Val,Res),make_string(DefVal,Res2)),
        (add_error(external_functions,'Illegal argument for GET_PREF: ',S,Span),
         fail)).


% printf followed by fail; useful for print fail loops in the REPL
expects_spaninfo('printfail').
:- assert_must_fail(external_functions:printfail(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printfail(-,?,?,?), printfail(?,-,?,?).
printfail(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),fail)).
printfail(F,V,P,S) :-
   add_internal_error('Illegal call: ',printfail(F,V,P,S)), P=pred_true.

% a version of printf with automatic indenting and backtracking feedback
expects_spaninfo('iprintf').
external_pred_always_true('iprintf').
:- assert_must_succeed(external_functions:iprintf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block iprintf(-,?,?,?), iprintf(?,-,?,?).
iprintf(string(FormatString),Value,PredRes,Span) :- !, PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         ifprintf2(user_output,FormatString,Value,no_trace,Span)).
iprintf(F,V,P,S) :- add_internal_error('Illegal call: ',iprintf(F,V,P,S)), P=pred_true.
ifprintf2(File,FormatString,Value,Trace,Span) :-
   my_term_hash((FormatString,Value),Hash),
   format(user_output,'[hash ~w] ',[Hash]), % (Hash=1087266573122589201 -> (trace ; trace,fail) ; true),
   translate:translate_bvalue(Value,TV),
   indent_format(File,FormatString,[TV],Trace),
   fprint_log_info(File,Span).
expects_spaninfo('itprintf').
external_pred_always_true('itprintf').
:- public itprintf/4.
:- assert_must_succeed(external_functions:iprintf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block itprintf(-,?,?,?), itprintf(?,-,?,?).
itprintf(string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         ifprintf2(user_output,FormatString,Value,trace,Span)).


:- public observe/5.
expects_spaninfo('observe').
expects_unevaluated_args('observe').
expects_waitflag('observe').
external_pred_always_true('observe').
% observe pairs of values
observe(Value,PredRes,UnEvalArgs,Span,WF) :-
  observe2(Value,PredRes,UnEvalArgs,no_trace,Span,WF).

:- public tobserve/5.
expects_spaninfo('tobserve'). % observe but with tracing enabled
expects_unevaluated_args('tobserve').
expects_waitflag('tobserve').
external_pred_always_true('tobserve').
tobserve(Value,PredRes,UnEvalArgs,Span,WF) :-
   observe2(Value,PredRes,UnEvalArgs,trace,Span,WF).

observe2(Value,PredRes,[Arg],Trace,Span,WF) :- PredRes=pred_true,
   %kernel_waitflags:portray_waitflags(WF),
   kernel_waitflags:get_wait_flag0(WF,WF0),
   kernel_waitflags:get_wait_flag1(WF,WF1),
   when(nonvar(WF0),(indent_format(user_output,'WAITFLAG 0 set~n',[],no_trace),kernel_waitflags:portray_waitflags(WF))),
   when(nonvar(WF1),(indent_format(user_output,'WAITFLAG 1 set~n',[],no_trace),kernel_waitflags:portray_waitflags(WF))),
   Hook = print_value_as_table_if_possible(Arg,Value),
   observe_aux(Value,Arg,1,_,Trace,Span,Hook),
   iprintf(string("observe full value complete: ~w~n~n"),Value,_,Span).

%print_value_as_table_if_possible(_,_) :- !.
print_value_as_table_if_possible(Arg,Value) :- 
  translate:print_bexpr(Arg),nl,
  (table_tools:print_value_as_table(Arg,Value) -> true ; true).

observe_aux((V1,V2),b(couple(A1,A2),_,_),Nr,R,Trace,Span,Hook) :- !,
    observe_aux(V1,A1,Nr,N1,Trace,Span,Hook),
    observe_aux(V2,A2,N1,R,Trace,Span,Hook).
observe_aux(V,Arg,Nr,R,Trace,Span,Hook) :- R is Nr+1,
   translate:translate_bexpression(Arg,ArgString), %print(obs(Nr,ArgString)),nl,
    atom_codes(ArgString,ArgCodes),
    observe_value(V,ArgCodes,Nr,Trace,Span,Hook).

% entry for other modules:
observe_value(Value,Info) :- (Info=[_|_] -> PCodes = Info ; atom_codes(Info,PCodes)),
    observe_value(Value,PCodes,1,no_trace,unknown,true).

observe_value(Value,ArgCodes,Nr,Trace,Span,Hook) :-  %print(obs(Nr,ArgString)),nl,
   number_codes(Nr,NrCodes), append(NrCodes,") = ~w~n",Tail1), append(" (",Tail1,Tail2),
   append(ArgCodes,Tail2,FormatString),
   observe_set(Value,ArgCodes,1,Trace,Span,Value,Hook),
   when((ground(FormatString),ground(Value)),ifprintf2(user_output,FormatString,Value,Trace,Span)).
observe_value_with_full_value(Value,ArgCodes,Nr,Trace,Span,FullValue,Hook) :-  %print(obs(Nr,ArgString)),nl,
   number_codes(Nr,NrCodes), append(NrCodes,"):FULL = ~w~n",Tail1), append(" (",Tail1,Tail2),
   append(ArgCodes,Tail2,FormatString),
   observe_set(Value,ArgCodes,1,Trace,Span,Value,Hook),
   when((ground(FormatString),ground(Value)),
           observe_ground(FormatString,Value,FullValue,Trace,Span,Hook)).

observe_ground(FormatString,Value,FullValue,Trace,Span,Hook) :-
    ifprintf2(user_output,FormatString,rec([field(value,Value),field(full,FullValue)]),Trace,Span),
    (call(Hook) -> true ; print(hook_failed(Hook)),nl).



% observe sets and records:
:- block observe_set(-,?,?,?,?,?,?).
observe_set([],_ArgCodes,_,_Trace,_Span,_,_) :- !.
observe_set([Value|T],ArgCodes,Pos,Trace,Span,FullValue,Hook) :- !, %print(obs(Value,Pos)),nl,
   append(ArgCodes," el -> ",ArgN),
   observe_value_with_full_value(Value,ArgN,Pos,Trace,Span,FullValue,Hook),
   %tools:put_codes(ArgN,user_output),print(Pos),print(' = '),tools_printing:print_arg(Value),nl,
   P1 is Pos+1, observe_set(T,ArgCodes,P1,Trace,Span,FullValue,Hook).
observe_set((V1,V2),ArgCodes,Pos,Trace,Span,FullValue,Hook) :-
   preferences:preference(observe_precise_values,true),
   !, %print(obs(Value,Pos)),nl,
   append(ArgCodes," prj1 -> ",ArgN1),
   observe_value_with_full_value(V1,ArgN1,Pos,Trace,Span,FullValue,Hook),
   %tools:put_codes(ArgN,user_output),print(Pos),print(' = '),tools_printing:print_arg(Value),nl,
   Pos2 is Pos+1,
   append(ArgCodes," prj2 -> ",ArgN2),
   observe_value_with_full_value(V2,ArgN2,Pos2,Trace,Span,FullValue,Hook).
observe_set(rec(Fields),ArgCodes,Pos,Trace,Span,_FullValue,Hook) :-
   preferences:preference(observe_precise_values,true),
   !,
   append(ArgCodes," rec -> ",ArgN),
   observe_field(Fields,ArgN,Pos,Trace,Span,Hook).
observe_set(_,_,_,_,_,_,_).

:- block observe_field(-,?,?,?,?,?).
observe_field([],ArgN,Pos,Trace,Span,_Hook) :-
    append(ArgN," complete (~w)~n",FormatString),
    ifprintf2(user_output,FormatString,Pos,Trace,Span).
observe_field([field(Name,Value)|TFields],ArgN,Pos,Trace,Span,Hook) :-
    observe_value((string(Name),Value),ArgN,Pos,Trace,Span,Hook),
    observe_field(TFields,ArgN,Pos,Trace,Span,Hook).

:- use_module(bsyntaxtree,[get_texpr_id/2]).
:- use_module(store,[lookup_value_with_span/5]).
% utility function for other modules to use observe on parameters
observe_parameters(Parameters,LocalState) :-
   skel(LocalState,SetupDone),
   reset_ident, reset_walltime,
   when(nonvar(SetupDone),observe_parameters_aux(Parameters,1,LocalState)).
:- block skel(-,?).
skel([],pred_true).
skel([_|T],Done) :- skel(T,Done).
:- block observe_parameters_aux(-,?,?), observe_parameters_aux(?,?,-).
observe_parameters_aux([],_,_).
observe_parameters_aux([TID|T],Nr,LS) :-
   safe_get_texpr_id(TID,ID), atom_codes(ID,PCodes),
   store:lookup_value_with_span(ID,[],LS,Value,TID),
   debug_println(9,observing(ID,Value,LS)),
   Hook = print_value_as_table_if_possible(TID,Value),
   observe_value(Value,PCodes,Nr,no_trace,unknown,Hook),
   N1 is Nr+1,
   observe_parameters_aux(T,N1,LS).
safe_get_texpr_id(TID,ID) :- (atom(TID) -> ID=TID ; get_texpr_id(TID,ID)).

% utility to observe a state using the observe external function
observe_state(State) :-
   reset_ident, reset_walltime,
   observe_state_aux(State,1).
:- block observe_state_aux(-,?).
observe_state_aux(no_state,_) :- !.
observe_state_aux([],_) :- !.
observe_state_aux(concrete_constants(C),Nr) :- !, observe_state_aux(C,Nr).
observe_state_aux([bind(Id,Value)|T],Nr) :-
   atom_codes(Id,PCodes), !,
   %print(observe(Id,Value,Nr,T)),nl,
   observe_value(Value,PCodes,Nr,no_trace,unknown,true),
   N1 is Nr+1,
   observe_state_aux(T,N1).
observe_state_aux(State,Nr) :- add_internal_error('Illegal state: ',observe_state_aux(State,Nr)).


external_fun_type('observe_fun',[X],[X,boolean]).
expects_unevaluated_args('observe_fun').
external_pred_always_true('observe_fun').
:- public observe_fun/3.
% a specific predicate to observe enumeration of functions,...
:- block observe_fun(-,-,?).
observe_fun(Value,PredRes,[UnEvalArg]) :-
   translate:translate_bexpression(UnEvalArg,Str), format('Observe function : ~w~n',[Str]),
   obs_fun(Value,Str,Value),PredRes=pred_true.

:- block obs_fun(-,?,?).
obs_fun([],_,_) :- !.
obs_fun([(A,B)|T],Str,FullVal) :- !,
    obs_idx(A,B,Str,FullVal),
    obs_fun(T,Str,FullVal).
obs_fun(V,Str,_) :- print(uncovered_obs_fun(V,Str)),nl.

:- block obs_idx(-,?,?,?), obs_idx(?,-,?,?).
obs_idx(A,B,Str,FullVal) :-
  when(ground((A,B)),obs_idx_complete(FullVal,A,B,Str)).

obs_idx_complete(FullVal,A,B,Str) :-
   format('~n ~w = ',[Str]),obs_idx_complete_aux(FullVal,A,B).
obs_idx_complete(FullVal,A,B,Str) :-
   format('~n [BACKTRACK] ~w = ',[Str]),obs_idx_complete_aux(FullVal,A,B),fail.

obs_idx_complete_aux(V,_,_) :- var(V),!,print(' OPEN '),nl.
obs_idx_complete_aux([],_,_) :- !, nl.
obs_idx_complete_aux([(X,Y)|T],A,B) :- !,
  ((X,Y)==(A,B) -> print(' * ') ; print('   ')), print_bvalue((X,Y)),
  obs_idx_complete_aux(T,A,B).
obs_idx_complete_aux(_,_,_) :- print(unknown),nl.


% -----------------------------

:- public tprintf/4.
expects_spaninfo('tprintf').
external_pred_always_true('tprintf').
% a debugging version that switches into trace mode
:- block tprintf(-,?,?,?), tprintf(?,-,?,?).
tprintf(string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),trace)).

:- public 'prolog_printf'/5.
expects_spaninfo('prolog_printf').
expects_waitflag('prolog_printf').
external_pred_always_true('prolog_printf').
prolog_printf(string(FormatString),Value,PredRes,Span,WF) :- PredRes = pred_true,
   file_format(user_output,FormatString,[initial(Value,WF)],Span),
   when(nonvar(Value),file_format(user_output,FormatString,[nonvar(Value,WF)],Span)),
   when((ground(FormatString),ground(Value)),
         fprintf2(user_output,FormatString,Value,Span)).

% vprintf: expression which returns argument as value
expects_spaninfo('vprintf').
:- public vprintf/4.
vprintf(Format,Value,ValueRes,Span) :-
    printf(Format,Value,pred_true,Span), % use itprintf for tracing
    equal_object(Value,ValueRes,vprintf). % we could generate a version which only instantiates when printing done
expects_spaninfo('fvprintf').
:- public fvprintf/5.
fvprintf(File,Format,Value,ValueRes,Span) :-
    fprintf(File,Format,Value,pred_true,Span),
    equal_object(Value,ValueRes,fvprintf). % we could generate a version which only instantiates when printing done


% PRINT a single value as substitution
performs_io('PRINT').
external_subst_enabling_condition('PRINT',_,Truth) :- create_texpr(truth,pred,[],Truth).
does_not_modify_state('PRINT').
:- public 'PRINT'/3.
:- block 'PRINT'(-,?,?), 'PRINT'(?,-,?).
'PRINT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   assert_side_effect_occurred(user_output),
   print_value(Val),nl, OutEnv=[].

% PRINT a single value as substitution; show backtracking
performs_io('PRINT_BT').
perform_directly('PRINT_BT').
external_subst_enabling_condition('PRINT_BT',_,Truth) :- create_texpr(truth,pred,[],Truth).
does_not_modify_state('PRINT_BT').
:- public 'PRINT_BT'/3.
:- block 'PRINT_BT'(-,?,?), 'PRINT_BT'(?,-,?).
'PRINT_BT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   assert_side_effect_occurred(user_output),
   print_value(Val),nl, OutEnv=[].
'PRINT_BT'(Val,_Env,_OutEnv) :-
   assert_side_effect_occurred(user_output),
   print('BACKTRACKING: '),print_value(Val),nl, fail.

print_value(Val) :- translate:translate_bvalue(Val,TV), print(TV).


performs_io('DEBUG_PRINT_STATE').
external_subst_enabling_condition('DEBUG_PRINT_STATE',_,Truth) :- create_texpr(truth,pred,[],Truth).
does_not_modify_state('DEBUG_PRINT_STATE').
:- public 'DEBUG_PRINT_STATE'/3.
:- block 'DEBUG_PRINT_STATE'(-,?,?).
'DEBUG_PRINT_STATE'(Value,Env,OutEnv) :-
    assert_side_effect_occurred(user_output),
    print_value(Value),translate:print_bstate(Env),nl, OutEnv=[].

 % PRINTF as substitution
expects_spaninfo('PRINTF').
performs_io('PRINTF').
external_subst_enabling_condition('PRINTF',_,Truth) :- create_texpr(truth,pred,[],Truth).
does_not_modify_state('PRINTF').
:- public 'PRINTF'/5.
:- block 'PRINTF'(-,?,?,?,?), 'PRINTF'(?,-,?,?,?), 'PRINTF'(?,?,-,?,?).
'PRINTF'(string(FormatString),Value,_Env,OutEnvModifications,Span) :-
  % print('PRINTF'(FormatString,Value,_Env)),nl,
   when((ground(FormatString),ground(Value)),
         (fprintf2(user_output,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).

expects_spaninfo('FPRINTF').
performs_io('FPRINTF').
external_subst_enabling_condition('FPRINTF',_,Truth) :- create_texpr(truth,pred,[],Truth).
does_not_modify_state('FPRINTF').
:- public 'FPRINTF'/6.
:- block 'FPRINTF'(-,?,?,?,?,?), 'FPRINTF'(?,-,?,?,?,?), 'FPRINTF'(?,?,-,?,?,?).
'FPRINTF'(string(File),string(FormatString),Value,_Env,OutEnvModifications,Span) :-
   when((ground(FormatString),ground(Value)),
         (fprintf2(File,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).

file_format(user_output,FormatString,Args,Span) :- !,
    assert_side_effect_occurred(user_output),
    safe_call(myformat(user_output,FormatString,Args),Span).
file_format(File,FormatString,Args,Span) :-
    open(File,append,Stream),
    assert_side_effect_occurred(file),
    safe_call(myformat(Stream,FormatString,Args),Span),
    close(Stream). % TO DO: call_cleanup

:- public myformat/3.
myformat(Stream,FormatString,Args) :-
   on_exception(error(consistency_error(_,_,format_arguments),_),
               format(Stream,FormatString,Args),
               % backup: the user probably has not provided ~w annotations:
               (format(Stream,FormatString,[]), format_args(Stream,Args))),
   flush_output(Stream).

format_args(_Stream,[]).
format_args(Stream,[H|T]) :- format(Stream,'~w~n',H), format_args(Stream,T).

% a format with indentation and backtracking info:
:- dynamic indentation_level/1.
indentation_level(0).
inc_indent :- retract(indentation_level(X)), X1 is X+1, assert(indentation_level(X1)).
dec_indent :- retract(indentation_level(X)), X1 is X-1, assert(indentation_level(X1)).
reset_ident :- retractall(indentation_level(_)), assert(indentation_level(0)).
indent(Stream) :- indentation_level(L), %(L>28 -> trace ; true),
    indent_aux(L,Stream).
indent_aux(X,Stream) :- X < 1, !, write(Stream,' ').
indent_aux(X,Stream) :- write(Stream,'+-'), X1 is X-1, indent_aux(X1,Stream).

indent_format(Stream,Format,Args,Trace) :- inc_indent,
  indent(Stream), format(Stream,Format,Args),
  print_walltime(Stream),
  % my_term_hash(Args,Hash), format(Stream,'[hash ~w]~n',[Hash]), (Hash=101755762 -> trace ; true), %%
  flush_output(Stream),
  (Trace=trace -> trace ; true).  
%indent_format(_Stream,_Format,Args,_Trace) :- my_term_hash(Args,Hash),(Hash=553878463258718990 -> trace).
indent_format(Stream,Format,Args,Trace) :-
  indent(Stream), write(Stream,'BACKTRACKING '),
    % my_term_hash(Args,Hash), format(Stream,'[hash ~w] ',[Hash]), %% (Hash=553878463258718990 -> trace ; true), %%
                  format(Stream,Format,Args),
  print_walltime(Stream), %% comment in to see walltime info
  flush_output(Stream),
  dec_indent,
  (Trace=trace -> trace ; true),
  fail.

:- dynamic ref_walltime/1, prev_walltime/1.
reset_walltime :- retractall(prev_walltime(_)), retractall(ref_walltime(_)),
   statistics(walltime,[WT,_]),
   assert(prev_walltime(WT)),
   assert(ref_walltime(WT)).
print_walltime(File) :- statistics(walltime,[WT,_]),
   (ref_walltime(RefWT) -> RWT is WT-RefWT ; RWT = 0, assert(ref_walltime(WT))),
   (retract(prev_walltime(PWT)) -> T is WT-PWT, indent(File),format(File,' ~w ms (walltime; total ~w ms)~n',[T,RWT])
     ; true),
    assert(prev_walltime(WT)).
:- public 'TRACE'/2.
:- block 'TRACE'(-,?).
'TRACE'(pred_true,R) :- !,trace,R=pred_true.
'TRACE'(pred_false,pred_true).

% --------------------------------------------

% ARGV, ARGC

:- dynamic argv_string/2, argc_number/1.
argc_number(0).
reset_argv :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   assert(argc_number(0)).

% call to set the commandline arguments, e.g., from probcli
set_argv_from_atom(X) :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   split_atom_string(X,' ',List),
   set_argv_from_list(List).
set_argv_from_list(List) :-
   retractall(argv_string(_,_)),
   retractall(argc_number(_)),
   set_argv_from_list_aux(List).
set_argv_from_list_aux(List) :-
   length(List,Len),
   assert(argc_number(Len)),
   debug:debug_println(9,argc(Len)),
   nth1(N,List,String), % we number the arguments in B style
   debug:debug_println(9,argv(N,String)),
   assert(argv_string(N,String)),
   fail.
set_argv_from_list_aux(_).

external_fun_type('ARGV',[],[integer,string]).
:- public 'ARGV'/2.
:- block 'ARGV'(-,?).
'ARGV'(int(N),S) :- argv2(N,S).
:- block argv2(-,?).
argv2(Nr,Res) :- % print(get_argv(Nr,Res)),nl,
  (argv_string(Nr,S) -> Res=string(S)
   ; add_error(external_functions,'Command-line argument does not exist: ',Nr), Res=string('')).

external_fun_type('ARGC',[],[integer]).
:- public 'ARGC'/1.
'ARGC'(int(X)) :- argc_number(X).

% --------------------------------------------

% RANDOM

:- assert_must_succeed(( external_functions:call_external_function('RANDOM',[int(0),int(1)],int(0),integer,unknown,_WF) )).
:- use_module(library(random)).
% Generate a random number >= Low and < Up
% warning: this is not a mathematical function
:- public 'RANDOM'/3.
is_not_declarative('RANDOM').
:- block 'RANDOM'(-,?,?), 'RANDOM'(?,-,?).
'RANDOM'(int(Low),int(Up),int(R)) :- random2(Low,Up,R).
:- block random2(-,?,?), random2(?,-,?).
random2(L,U,R) :- random(L, U, R).

% pick a random element from a set
:- public 'random_element'/4.
is_not_declarative('random_element').
expects_waitflag('random_element').
expects_spaninfo('random_element').
:- block 'random_element'(-,?,?,?).
random_element([],_,Span,WF) :- !,
   add_wd_error_span('random_element applied to empty set: ',[],Span,WF).
random_element([H|T],El,_,_) :- !,
   block_length(T,1,Len),
   random_select_from_list([H|T],Len,El).
random_element(CS,Res,_,_) :- is_interval_closure_or_integerset(CS,Low,Up),!,
   U1 is Up+1,
   random(Low,U1,H), Res = int(H).
random_element(CS,X,Span,WF) :-
   custom_explicit_sets:expand_custom_set(CS,ER),
   random_element(ER,X,Span,WF).

:- block block_length(-,?,?).
block_length([],Acc,Acc).
block_length([_|T],Acc,R) :- A1 is Acc+1, block_length(T,A1,R).

:- block random_select_from_list(?,-,?).
random_select_from_list(List,Len,Res) :-
   random(0,Len,RandNr),
   nth0(RandNr,List,El),
   kernel_objects:equal_object(El,Res).

:- use_module(library(system)).

:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(year)],int(Y),integer,unknown,_WF), number(Y), Y>2011 )).
:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(month)],int(M),integer,unknown,_WF), number(M), M>0, M<13 )).
:- public 'TIME'/2.
is_not_declarative('TIME'). % as time can change with next call
'TIME'(string(S),int(N)) :- time2(S,N).
:- block time2(-,?).
time2(now,N) :- !, now(N). % provide UNIX timesstamp
time2(year,N) :- !,datime(datime(N,_,_,_,_,_)).
time2(month,N) :- !,datime(datime(_,N,_,_,_,_)).
time2(day,N) :- !,datime(datime(_,_,N,_,_,_)).
time2(hour,N) :- !,datime(datime(_,_,_,N,_,_)).
time2(min,N) :- !,datime(datime(_,_,_,_,N,_)).
time2(sec,N) :- !,datime(datime(_,_,_,_,_,N)).
time2(OTHER,_) :- add_error(time2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- assert_must_succeed(( external_functions:call_external_function('GET_INFO',[string(time)],string(M),string,unknown,_WF), atom(M) )).
:- public 'GET_INFO'/2.
'GET_INFO'(string(S),Res) :- get_info2(S,Res).
:- use_module(library(codesio)).
:- block get_info2(-,?).
get_info2(time,Res) :- !, datime(datime(Year,Month,Day,Hour,Min,Sec)),
    (Min<10
      -> format_to_codes('~w/~w/~w - ~wh0~w ~ws',[Day,Month,Year,Hour,Min,Sec],Codes)
       ; format_to_codes('~w/~w/~w - ~wh~w ~ws',[Day,Month,Year,Hour,Min,Sec],Codes)
    ),
    atom_codes(INFO, Codes), Res=string(INFO).
get_info2(OTHER,_) :- add_error(time2,'Illegal GET_INFO field (must be time): ',OTHER),fail.

:- public 'TIMESTAMP_INFO'/3.
is_not_declarative('TIMESTAMP_INFO'). % as time can change with next call
:- block 'TIMESTAMP_INFO'(-,-,-).
'TIMESTAMP_INFO'(string(S),int(Ticks),int(Res)) :- time_info2(S,Ticks,Res).
:- block time_info2(-,?,?), time_info2(?,-,?).
time_info2(year,Ticks,N) :- !,datime(Ticks,datime(N,_,_,_,_,_)).
time_info2(month,Ticks,N) :- !,datime(Ticks,datime(_,N,_,_,_,_)).
time_info2(day,Ticks,N) :- !,datime(Ticks,datime(_,_,N,_,_,_)).
time_info2(hour,Ticks,N) :- !,datime(Ticks,datime(_,_,_,N,_,_)).
time_info2(min,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,N,_)).
time_info2(sec,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,_,N)).
time_info2(OTHER,_,_) :- add_error(time_info2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- public 'TIMESTAMP'/7.
:- block 'TIMESTAMP'(-,-,-,-,-,-,-).
'TIMESTAMP'(int(Y),int(M),int(D),int(H),int(Min),int(S),int(Res)) :-
   timestamp2(Y,M,D,H,Min,S,Res).

:- block timestamp2(-,?,?,?,?,?,-), timestamp2(?,-,?,?,?,?,-), timestamp2(?,?,-,?,?,?,-),
         timestamp2(?,?,?,-,?,?,-), timestamp2(?,?,?,?,?,-,-).
timestamp2(Y,M,D,H,Min,S,ResTimeStamp) :- datime(ResTimeStamp, datime(Y,M,D,H,Min,S)).


:- public 'WALLTIME'/1.
is_not_declarative('WALLTIME'). % as time can change with next call
'WALLTIME'(int(X)) :- statistics(walltime,[X,_]).

:- volatile last_walltime/1.
:- dynamic last_walltime/1.
:- public 'DELTA_WALLTIME'/1.
is_not_declarative('DELTA_WALLTIME'). % as time can change with next call
'DELTA_WALLTIME'(int(D)) :- statistics(walltime,[X,_]),
  (retract(last_walltime(L)) -> D is X-L ; D = 0),
  assert(last_walltime(X)).

:- public 'RUNTIME'/1.
is_not_declarative('RUNTIME'). % as time can change with next call
'RUNTIME'(int(X)) :- statistics(runtime,[X,_]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a given expression is valued in the deterministic phase; warn otherwise

:- use_module(kernel_tools,[ground_value_check/2]).
expects_spaninfo('CHECK_DET').
expects_unevaluated_args('CHECK_DET').
expects_waitflag('CHECK_DET').
expects_waitflag('observe').
:- public 'CHECK_DET'/5.
'CHECK_DET'(Value,PredRes,[UnEvalArg],Span,WF) :-
   ground_value_check(Value,Ground),
   kernel_waitflags:get_wait_flag0(WF,WF0),
   kernel_waitflags:get_wait_flag1(WF,WF1),
   PredRes=pred_true,
   check_det_aux(Ground,Value,WF0,WF1,UnEvalArg,Span).

:- block check_det_aux(-,?,?,-,?,?).
check_det_aux(Ground,Value,_WF0,_WF1,UnEvalArg,Span) :- var(Ground),!,
   translate:translate_bexpression(UnEvalArg,S),
   add_error(external_functions,'Argument has not been computed deterministically: ',S,Span),
   translate:print_bvalue(Value),nl.
check_det_aux(_Ground,Value,WF0,WF1,UnEvalArg,_Span) :- debug:debug_mode(on),!,
   translate:translate_bexpression(UnEvalArg,S),
   format('Value determined (WF0=~w,WF1=~w) : ~w~n',[WF0,WF1,S]),
   translate:print_bvalue(Value),nl.
check_det_aux(_Ground,_Value,_WF0,_WF1,_UnEvalArgs,_Span).

% check if a given expression has already a ground value when
% the predicate is called

% the predicate is true when the value is ground, false otherwise
:- public 'IS_DETERMINED'/2.
'IS_DETERMINED'(Value,PredRes) :- ground(Value),!,PredRes=pred_true.
'IS_DETERMINED'(_Value,pred_false).

% other variant: the predicate is always true, the information if the variable is written
% to a file to check it later.
expects_spaninfo('IS_DETERMINED_INTO_FILE').
:- public 'IS_DETERMINED_INTO_FILE'/5.
'IS_DETERMINED_INTO_FILE'(FileValue,VariableValue,Value,PredRes,SpanInfo) :-
    check_value_is_string(FileValue,'IS_DETERMINED_INTO_FILE','file name',SpanInfo,File),
    check_value_is_string(VariableValue,'IS_DETERMINED_INTO_FILE','variable name',SpanInfo,Varname),
    ( ground(File) ->
        open(File,write,S),
        call_cleanup( check_determined2(Value,S,Varname), close(S))
    ; otherwise ->
        add_error(external_functions,'### EXTERNAL Unspecified file in check_determined','',SpanInfo)),
    PredRes = pred_true.
check_determined2(Value,S,Varname) :-
    check_determined3(Value,Varname,Term),
    writeq(S,Term),write(S,'.\n').
check_determined3(Value,Name,Term) :- ground(Value),!,Term=determined_value(Name,Value).
check_determined3(_Value,Name,not_determined_value(Name)).

check_value_is_string(string(S),_PredName,_Location,_SpanInfo,String) :-
    !, String=S.
check_value_is_string(Value,PredName,Location,SpanInfo,_String) :-
    atom_concat('### EXTERNAL ',PredName,Msg1),
    atom_concat(Msg1,': Type error, expected string as ',Msg2),
    atom_concat(Msg2,Location,Msg),
    add_error(external_functions,Msg,Value,SpanInfo),
    fail.


:- public 'SEE'/3.
is_not_declarative('SEE').
:- dynamic open_stream/2.
'SEE'(string(File),_S,[]) :- b_absolute_file_name(File,AFile),
   open_stream(AFile,_),!,
   print('File already open: '), print(File),nl.
'SEE'(string(File),_S,[]) :-
   b_absolute_file_name(File,AFile),
   open_absolute_file_for_reading(AFile,_Stream).

open_file_for_reading_if_necessary(File,Stream) :-
   b_absolute_file_name(File,AFile),
   (open_stream(AFile,S) -> Stream=S
     ; open_absolute_file_for_reading(AFile,Stream)).
open_absolute_file_for_reading(AFile,Stream) :-
   open(AFile,read,Stream),
   print(opened_file(AFile,Stream)),nl,
   assert_side_effect_occurred(file),
   assert(open_stream(AFile,Stream)).

%correct_file_open_mode(read).
%correct_file_open_mode(write).
%correct_file_open_mode(append).

:- public 'GET_CODE'/2.
performs_io('GET_CODE').
'GET_CODE'(string(File),int(Code)) :-
   b_absolute_file_name(File,AFile),
   open_stream(AFile,S),
   assert_side_effect_occurred(file),
   get_code(S,Code). %,print(got_code(S,Code)),nl.

:- public 'GET_CODE_STDIN'/1.
'GET_CODE_STDIN'(int(Code)) :- get_code(user_input,Code).

:- public 'SEEN'/3.
is_not_declarative('SEEN').
'SEEN'(string(File),_S,[]) :- b_absolute_file_name(File,AFile),
   retract(open_stream(AFile,S)),
   assert_side_effect_occurred(file),
   close(S).

close_all_files :- retract(open_stream(_File,S)),
   close(S),fail.
close_all_files.


:- public 'READ_FILE_AS_STRINGS'/3.
performs_io('READ_FILE_AS_STRINGS').
external_subst_enabling_condition('READ_FILE_AS_STRINGS',_,Truth) :- create_texpr(truth,pred,[],Truth).
expects_spaninfo('READ_FILE_AS_STRINGS').
:- block 'READ_FILE_AS_STRINGS'(-,?,?).
'READ_FILE_AS_STRINGS'(string(File),Res,SpanInfo) :-
    b_absolute_file_name(File,AFile),
    safe_call(open(AFile,read,Stream),SpanInfo),!,
    format('% Opened file: ~w~n',[File]),
    % no side-effect: as everything read in one go
    call_cleanup(read_lines(Stream,Res),
                 close(Stream)).
'READ_FILE_AS_STRINGS'(string(File),_Res,_SpanInfo) :-
   format('% Opening file failed: ~w~n',[File]),fail.
read_lines(File,Res) :- read_lines_loop(1,File,Lines),
   kernel_objects:equal_object(Res,Lines).
read_lines_loop(Nr,Stream,Lines) :- read_line(Stream,Line),
   debug:debug_println(9,read_line(Nr,Line)),
   (Line = end_of_file -> Lines = []
    ; atom_codes(Atom,Line), Lines = [(int(Nr),string(Atom))|T],
      N1 is Nr+1,read_lines_loop(N1,Stream,T)).

:- public 'READ_LINE'/3.
expects_spaninfo('READ_LINE').
external_subst_enabling_condition('READ_LINE',_,Truth) :- create_texpr(truth,pred,[],Truth).
performs_io('READ_LINE').
'READ_LINE'(string(File),Res,_SpanInfo) :-
   open_file_for_reading_if_necessary(File,Stream),
   read_line(Stream,Line),
   debug:debug_println(9,read_line(Line)),
   assert_side_effect_occurred(file),
   (Line = end_of_file
    -> add_error(external_functions,'Attempting to read past EOF: ',File),
       Res = ""
     ; atom_codes(Atom,Line), Res = string(Atom)).

:- public 'EOF'/2.
'EOF'(string(File),Res) :- eof2(File,Res).
:- block eof2(-,?).
eof2(File,Res) :-
   open_file_for_reading_if_necessary(File,Stream),
   ((peek_code(Stream,C), C = -1) -> Res = pred_true ; Res = pred_false).  % for Linux compatibility
%   (at_end_of_stream(Stream) -> Res = pred_true ; Res=pred_false).

:- public 'EOF_STDIN'/2.
'EOF_STDIN'(_,Res) :- % dummy argument to enable valid external predicate type declarations
   ((peek_code(user_input,C), C = -1) -> Res = pred_true ; Res = pred_false).


:- public 'CURRENT_FILE_POSITION'/3.
'CURRENT_FILE_POSITION'(string(File),string(Field),Res) :-
   open_file_for_reading_if_necessary(File,Stream),
   stream_position(Stream,Pos),
   % POS must be one of byte_count [only for binary streams], line_count, character_count, line_position
   stream_position_data(Field,Pos,R), Res=int(R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- public 'ADD_ERROR'/5.
expects_spaninfo('ADD_ERROR').
'ADD_ERROR'(string(Msg),Expr,_S,[],SpanInfo) :-
   add_error(b_user_program,Msg,Expr,SpanInfo).


:- public 'ASSERT_TRUE'/5.
expects_spaninfo('ASSERT_TRUE').
expects_waitflag('ASSERT_TRUE').

'ASSERT_TRUE'(PredVal,string(Msg),PredRes,SpanInfo,WF) :-
    assert_true(PredVal,Msg,PredRes,SpanInfo,WF).

:- block assert_true(-,?,?,?,?).
assert_true(pred_true,_Msg,pred_true,_SpanInfo,_WF).
assert_true(pred_false,Msg,Res,SpanInfo,WF) :- print(Msg),nl,
    add_error_ef('ASSERT_TRUE',Msg,'',SpanInfo,WF,Res).

% we could also use: add_wd_error_set_result
add_error_ef(Source,Msg,Term,Span,WF,Res) :-
    get_enumeration_finished_wait_flag(WF,AWF),
    add_error_ef_aux(AWF,Source,Msg,Term,Span,WF,Res).

:- block add_error_ef_aux(-,?,?,?,?,?,?).
add_error_ef_aux(_AWF,Source,Msg,Term,Span,_WF,Res) :-
    add_error(Source,Msg,Term,Span),
    Res = pred_false.

:- public 'ASSERT_EXPR'/6.
expects_spaninfo('ASSERT_EXPR').
expects_waitflag('ASSERT_EXPR').

'ASSERT_EXPR'(PredVal,string(Msg),ExprVal,Res,SpanInfo,WF) :-
    assert_expr(PredVal,Msg,ExprVal,Res,SpanInfo,WF).

:- block assert_expr(-,?,?,?,?,?).
assert_expr(pred_true,_Msg,ExprVal,Res,_SpanInfo,_WF) :-
    Res = ExprVal.
assert_expr(pred_false,Msg,_ExprVal,Res,SpanInfo,WF) :-
    add_error_ef('ASSERT_EXPR',Msg,'',SpanInfo,WF,Res).

% --------------------------
% Reflection Functions

external_fun_type('STATE_AS_STRING',[],[integer,string]).
:- public 'STATE_AS_STRING'/3.
expects_waitflag('STATE_AS_STRING').
'STATE_AS_STRING'(int(X),StateStr,WF) :-
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_AS_STRING',WF,LWF),
    get_state2(X,StateStr,LWF).
:- block get_state2(-,?,-).
get_state2(X,StateStr,_) :-
    convert_id_to_nr(IDX,X),
    state_space:visited_expression(IDX,S),
    translate:translate_bstate(S,Str),
    StateStr = string(Str).

:- public 'STATE_SUCC'/4.
external_fun_type('STATE_SUCC',[],[integer,integer,boolean]).
expects_waitflag('STATE_SUCC').
'STATE_SUCC'(X,Y,Res,WF) :- 'STATE_TRANS'(X,_,Y,Res,WF).

external_fun_type('STATE_TRANS',[],[integer,string,integer,boolean]).
expects_waitflag('STATE_TRANS').
'STATE_TRANS'(int(X),OperationName,int(Y),Res,WF) :-
    'STATE_TRANS_ARGS'(int(X),OperationName,_,int(Y),Res,WF).

external_fun_type('STATE_TRANS_ARGS',[X],[integer,string,X,integer,boolean]).
expects_waitflag('STATE_TRANS_ARGS').
'STATE_TRANS_ARGS'(int(X),OperationName,OperationArgs,int(Y),Res,WF) :-
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_TRANS',WF,LWF),
    get_wait_flag1(WF,WF1),
    get_succ2(X,OperationName,OperationArgs,Y,Res,LWF,WF1).
:- block get_succ2(-,?,?,-,?,-,?),
         get_succ2(?,?,?,?,?,?,-). % wait at least until WF1 (or at least wf0); as we do non-deterministic enumeration ! (e.g., otherwise b_interprter_components gets confused !!)
get_succ2(X,OperationName,OperationArgs,Y,Res,_,_) :- convert_id_to_nr(IDX,X), convert_id_to_nr(IDY,Y),
    if(state_space_transition(IDX,IDY,OperationName,OperationArgs),Res=pred_true,Res=pred_false).
    
state_space_transition(IDX,IDY,OperationName,OperationArgs) :-
   state_space:transition(IDX,Transition,IDY),
   op_functor(Transition,OperationName),
   op_args(Transition,OperationArgs).

% root is represented as -1
:- block convert_id_to_nr(-,-).
convert_id_to_nr(ID,X) :- ID==root, !, X is -1.
convert_id_to_nr(ID,X) :- X == -1, !, ID=root.
convert_id_to_nr(X,X).

:- use_module(specfile,[get_operation_name/2,get_operation_arguments/2]).
op_functor(OpTerm,Res) :- get_operation_name(OpTerm,Name),
   Res = string(Name).

:- use_module(tools,[convert_list_into_pairs/2]).
op_args(OpTerm,OperationArgs) :- get_operation_arguments(OpTerm,As),
   convert_list_into_pairs(As,ArgTerm),
   kernel_objects:equal_object(OperationArgs,ArgTerm).


:- public 'STATE_SAT'/4.
expects_waitflag('STATE_SAT').
external_fun_type('STATE_SAT',[],[integer,string,boolean]).
% check whether formula as string holds in a state
:- block 'STATE_SAT'(?,-,?,?).
'STATE_SAT'(int(X),string(Formula),Res,WF) :- %print(parsing(Formula)),nl, trace,
    parse_pred(Formula,TFormula),
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_SAT',WF,LWF),
    state_sat2(X,TFormula,Res,LWF,WF).
:- block state_sat2(?,-,?,?,?),state_sat2(-,?,?,-,?).
state_sat2(X,TFormula,PredRes,_,WF) :-
    (number(X) ->  test_formula_in_state(X,TFormula,PredRes,WF)
     ; test_formula_in_state(ID,TFormula,PredRes,WF), ID \= root, X=ID % avoid CLPFD errors on X
    ).

:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/3]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2,state_corresponds_to_fully_setup_b_machine/2]).
%:- use_module(b_interpreter,[b_try_check_boolean_expression_wf/5]).
test_formula_in_state(ID,Pred,PredRes,WF) :-
    state_space:visited_expression(ID,State),
    state_corresponds_to_fully_setup_b_machine(State,FullState),
    add_prob_deferred_set_elements_to_store(FullState,FullState1,visible),
    debug_println(9,check(Pred,FullState1,PredRes)),
    b_interpreter:b_try_check_boolean_expression_wf(Pred,[],FullState1,WF,PredRes).


:- public 'EVAL'/4.
expects_waitflag('EVAL'). % external function with unknown return type
'EVAL'(int(X),string(Expr),Value,WF) :-
     'STATE_EVAL'(int(X),string(Expr),Value,pred_true,WF).
:- public 'STATE_EVAL'/5.
expects_waitflag('STATE_EVAL'). % external predicate
% evaluate expression and check whether value corresponds to provided value
% (needed for type inference of the external predicate)
:- block 'STATE_EVAL'(?,-,?,?,?).
'STATE_EVAL'(int(X),string(Expr),Value,PredRes,WF) :-
    % string must be provided ; we assume it will be ground
    parse_expr(Expr,TExpr),
    state_space:get_state_space_stats(NrNodes,_NrTransitions,_ProcessedNodes),
    get_wait_flag(NrNodes,'STATE_SAT',WF,LWF),
    state_eval2(X,TExpr,Value,LWF,PredRes).
:- block state_eval2(?,-,?,?,?),state_eval2(-,?,?,-,?).
state_eval2(X,TExpr,Value,_,PredRes) :-
    (number(X) ->  eval_formula_in_state(X,TExpr,Value,PredRes)
     ; eval_formula_in_state(ID,TExpr,Value,PredRes), ID \= root, X=ID % avoid CLPFD errors on X
    ).
eval_formula_in_state(ID,TExpr,Value,PredRes) :-
    state_space:visited_expression(ID,State),
    state_corresponds_to_fully_setup_b_machine(State,FullState),
    add_prob_deferred_set_elements_to_store(FullState,FullState1,visible),
    debug_println(9,eval(TExpr,FullState1,Value)),
    kernel_equality:equality_objects(Value,TExprValue,PredRes), % TO DO: Stronger Checks that types compatible
    b_interpreter:b_compute_expression(TExpr,[],FullState1,TExprValue).

:- use_module(library(terms),[term_hash/2]).
:- use_module(bmachine,[b_parse_machine_predicate/3]).
:- dynamic parse_pred_cache/3.
parse_pred(PredStr,TPred) :- term_hash(PredStr,Hash), parse_pred(PredStr,Hash,TPred).
parse_pred(PredStr,Hash,TPred) :- parse_pred_cache(Hash,PredStr,TC),!, TPred=TC.
parse_pred(PredStr,Hash,TPred) :-
    bmachine:b_parse_machine_predicate(PredStr,[prob_ids(visible),variables],TC),
    assert(parse_pred_cache(Hash,PredStr,TC)),
    TPred=TC.

:- dynamic parse_expr_cache/3.
parse_expr(ExprStr,TExpr) :- term_hash(ExprStr,Hash), parse_expr(ExprStr,Hash,TExpr).
parse_expr(ExprStr,Hash,TExpr) :- parse_expr_cache(Hash,ExprStr,TC),!, TExpr=TC.
parse_expr(ExprStr,Hash,TExpr) :-
    atom_codes(ExprStr,Codes),  %print(parsing(Expr,Codes)),nl,
    bmachine:b_parse_machine_expression_from_codes_with_prob_ids(Codes,TC),
    assert(parse_expr_cache(Hash,ExprStr,TC)),
    TExpr=TC.

external_fun_type('HISTORY',[],[seq(integer)]).
:- public 'HISTORY'/1.
'HISTORY'(Value) :-
   state_space:history(RH), reverse(RH,[root|Rest]),
   generate_code_seq(Rest,1,Hist),
   kernel_objects:equal_object(Value,Hist).

:- use_module(state_space,[is_initial_state_id/1]).

:- public 'INITIAL_STATE'/3.
expects_waitflag('INITIAL_STATE'). % external function with unknown return type
'INITIAL_STATE'(int(InitialState),PredRes,WF) :-
    get_wait_flag(2,'INITIAL_STATE',WF,LWF),
    is_init2(InitialState,PredRes,LWF).
:- block is_init2(-,-,?), is_init2(-,?,-).
is_init2(InitialState,PredRes,_) :-
   PredRes == pred_true,!,
   is_initial_state_id(InitialState).
is_init2(InitialState,PredRes,_) :- number(InitialState),!,
   (is_initial_state_id(InitialState) -> PredRes=pred_true ; PredRes=pred_false).
is_init2(InitialState,PredRes,_LWF) :-
   PredRes == pred_false,!,
   when(nonvar(InitialState), \+is_initial_state_id(InitialState)).
is_init2(InitialState,PredRes,LWF) :-
   add_internal_error('Uncovered case: ', is_init2(InitialState,PredRes,LWF)),fail.


% provide information about a particular B machine in the project
:- public 'MACHINE_INFO'/4.
expects_waitflag('MACHINE_INFO').
:- block 'MACHINE_INFO'(?,-,?,?).
'MACHINE_INFO'(M,string(I),Value,WF) :-
    get_wait_flag(2,'INITIAL_STATE',WF,LWF), % TO DO: use number of machines
    print(minfo(I,M,Value)),nl,
    machine_info2(I,M,Value,LWF).

:- use_module(b_machine_hierarchy,[machine_type/2]).
:- block machine_info2(-,?,?,?), machine_info2(?,-,?,-).
machine_info2('TYPE',M,Value,_LWF) :-
   (ground(M) -> M = string(MachineName) ; true),
   b_machine_hierarchy:machine_type(MachineName,Type),
   (M,Value) = (string(MachineName),string(Type)).
   

% returns list of strings for info string such as files, variables,...
external_fun_type('PROJECT_INFO',[],[string,set(string)]).
expects_waitflag('PROJECT_INFO').
expects_spaninfo('PROJECT_INFO').
:- public 'PROJECT_INFO'/4.
'PROJECT_INFO'(Str,Res,Span,WF) :-
    kernel_waitflags:get_wait_flag(10,'PROJECT_INFO',WF,WF10), % will enumerate possible input strings
    project_info1(Str,Res,WF10,Span).

:- block project_info1(-,?,-,?).
project_info1(string(INFO),Res,_,_Span) :- INFO==help,!,
     L = [files, absolute-files, main-file, variables, constants, sets, operations, assertion_labels, invariant_labels],
     maplist(make_string,L,SL),
     kernel_objects:equal_object(SL,Res).
project_info1(string(INFO),Res,_,Span) :-
     if(list_info(INFO,L),
         (maplist(make_string,L,SL),
         kernel_objects:equal_object(SL,Res)),
        (add_error(external_functions,'Illegal argument for PROJECT_INFO/:list (use help to obtain valid arguments) : ',INFO,Span),
         fail)).

:- use_module(bmachine).
:- use_module(bsyntaxtree, [get_texpr_ids/2, get_texpr_label/2]).
:- use_module(tools,[get_tail_filename/2]).
list_info(files,TL) :- (b_get_all_used_filenames(L) -> true ; L=[]), maplist(get_tail_filename,L,TL).
list_info('absolute-files',L) :- (b_get_all_used_filenames(L) -> true ; L=[]).
list_info('main-file',L) :- (b_get_main_filename(F) -> get_tail_filename(F,TF),L=[TF] ; L=[]).
list_info(variables,L) :- b_get_machine_variables(LT), get_texpr_ids(LT,L).
list_info(constants,L) :- b_get_machine_constants(LT), get_texpr_ids(LT,L).
list_info(sets,L) :- findall(S,b_get_machine_set(S),L).
list_info(operations,L) :- findall(S,b_top_level_operation(S),L).
list_info(assertion_labels,L) :- 
      findall(ID,(b_get_assertions(all,_,Ass), member(Assertion,Ass),get_texpr_label(Assertion,ID)),L).
list_info(invariant_labels,L) :- 
      findall(ID,(b_get_invariant_from_machine(C), conjunction_to_list(C,CL),
                  member(Inv,CL),get_texpr_label(Inv,ID)),L).
      

make_string(X,Res) :- atom(X),!,Res=string(X).
make_string(X,Res) :- format_to_codes('~w',X,C), atom_codes(Atom,C), Res=string(Atom).

% obtain the source code of an identifier
external_fun_type('SOURCE',[],[string,string]).
:- public 'SOURCE'/2.
'SOURCE'(string(ID),SourceCode) :- source2(ID,SourceCode).
:- block source2(-,?).
source2(ID,SourceCode) :-
    source_code_for_identifier(ID,_,_,_OriginStr,_,Source),
    translate:translate_bexpression(Source,SC),
    SourceCode = string(SC).


% obtain various information about ProB (STRING result)
external_fun_type('PROB_INFO_STR',[],[string,string]).
expects_waitflag('PROB_INFO_STR').
:- public 'PROB_INFO_STR'/3.
% :- block 'PROB_INFO_STR'(-,?). % no block: this way we can via backtracking generate solutions
'PROB_INFO_STR'(Str,Res,WF) :-
    kernel_waitflags:get_wait_flag(10,'PROB_INFO_STR',WF,WF10), % will enumerate possible input strings
    prob_info1(Str,Res,WF10).

:- block prob_info1(-,?,-).
prob_info1(string(INFO),Res,_) :-
    prob_info2(INFO,Str), make_string(Str,Res).

:- use_module(parsercall,[get_parser_version/1, get_java_command_path/1, get_java_fullversion/3]).
:- use_module(version, [version_str/1, revision/1, lastchangeddate/1]).
prob_info2('prob-version',V) :- version_str(V).
prob_info2('parser-version',V) :- get_parser_version(V).
prob_info2('prolog-version',V) :- prolog_flag(version,V).
prob_info2('prob-revision',V) :- revision(V).
prob_info2('prob-last-changed-date',V) :- lastchangeddate(V).
prob_info2('java-version',V) :- get_java_fullversion(Maj,Min,Sfx), ajoin([Maj,'.',Min,'.',Sfx],V).
prob_info2('java-command-path',V) :- get_java_command_path(V).
prob_info2('current-time',V) :- get_info2(time,string(V)).  % Maybe we can get rid of GET_INFO ?


% obtain various information about ProB (INTEGER result)
external_fun_type('PROB_INFO_INT',[],[string,integer]).
expects_waitflag('PROB_INFO_INT').
:- public 'PROB_INFO_INT'/3.
'PROB_INFO_INT'(Str,Res,WF) :-
    kernel_waitflags:get_wait_flag(10,'PROB_INFO_INT',WF,WF10), % will enumerate possible input strings
    prob_info_int1(Str,Res,WF10).

:- block prob_info_int1(-,?,-).
prob_info_int1(string(INFO),Res,_) :-
    prob_info_int1(INFO,I), Res=int(I).

:- use_module(state_space,[current_state_id/1, get_state_space_stats/3]).
prob_info_int1('current-state-id',StateIDNr) :- current_state_id(ID),convert_id_to_nr(ID,StateIDNr).
prob_info_int1('states',NrNodes) :- get_state_space_stats(NrNodes,_,_).
prob_info_int1('transitions',NrTransitions) :- get_state_space_stats(_,NrTransitions,_).
prob_info_int1('processed-states',ProcessedNodes) :- get_state_space_stats(_,_,ProcessedNodes).
prob_info_int1('now-timestamp',N) :- now(N). % provide UNIX timesstamp
prob_info_int1('prolog-runtime',V) :- statistics(runtime,[V,_]).
prob_info_int1('prolog-walltime',V) :- statistics(walltime,[V,_]).
prob_info_int1('prolog-memory-bytes-used',V) :- statistics(memory_used,V).
prob_info_int1('prolog-memory-bytes-free',V) :- statistics(memory_free,V).
prob_info_int1('prolog-global-stack-bytes-used',V) :- statistics(global_stack_used,V).
prob_info_int1('prolog-local-stack-bytes-used',V) :- statistics(local_stack_used,V).
%prob_info_int1('prolog-global-stack-bytes-free',V) :- statistics(global_stack_free,V).
%prob_info_int1('prolog-local-stack-bytes-free',V) :- statistics(local_stack_free,V).
prob_info_int1('prolog-trail-bytes-used',V) :- statistics(trail_used,V).
prob_info_int1('prolog-choice-bytes-used',V) :- statistics(choice_used,V).
prob_info_int1('prolog-atoms-bytes-used',V) :- statistics(atoms_used,V).
prob_info_int1('prolog-atoms-nb-used',V) :- statistics(atoms_nbused,V).
prob_info_int1('prolog-gc-count',V) :- statistics(gc_count,V).
prob_info_int1('prolog-gc-time',V) :- statistics(gc_time,V).


% obtain various information about bvisual2 Formulas
external_fun_type('FORMULA_INFOS',[],[string,seq(string)]).
expects_waitflag('FORMULA_INFOS').
:- public 'FORMULA_INFOS'/3.
'FORMULA_INFOS'(Str,Res,WF) :-
    kernel_waitflags:get_wait_flag(10,'FORMULA_INFOS',WF,WF10), % will enumerate possible input strings
    formula_info(Str,_,Res,_,WF10).
external_fun_type('FORMULA_VALUES',[],[string,seq(string)]).
expects_waitflag('FORMULA_VALUES').
:- public 'FORMULA_VALUES'/3.
'FORMULA_VALUES'(Str,Res,WF) :-
    kernel_waitflags:get_wait_flag(10,'FORMULA_VALUES',WF,WF10), % will enumerate possible input strings
    formula_info(Str,_,_,Res,WF10).

:- use_module(bvisual2,[bv_get_top_level_formula/4, bv_value_to_atom/2]).

:- block formula_info(-,?,?,?,-).
formula_info(string(TopLevelCategory),CatTextRes,LabelRes,ValuesRes,_) :-
    bv_get_top_level_formula(TopLevelCategory,CatText,Labels,Values),
    CatTextRes = string(CatText),
    maplist(make_string,Labels,LS),
    convert_set_to_seq(LS,LabelRes),
    maplist(make_value,Values,LV),
    convert_set_to_seq(LV,ValuesRes).
    
make_value(V,string(A)) :- bv_value_to_atom(V,A).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% store the information that a predicate has actually been used
% We can check this later with check_if_predicates_were_evaluated/1
% below. The CLI offers a way to use the predicate.

:- use_module(tools, [assert_once/1]).
:- dynamic evaluated_predicate/1.
external_fun_type('IS_EVALUATED',[],[string,boolean]).
expects_spaninfo('IS_EVALUATED').
:- public 'IS_EVALUATED'/3.
'IS_EVALUATED'(Value,PredRes,SpanInfo) :-
    check_value_is_string(Value,'IS_EVALUATED',name,SpanInfo,Name),
    assert_once( evaluated_predicate(Name) ),
    PredRes = pred_true.

check_if_predicates_were_evaluated(List) :-
    maplist(check_if_predicate_was_evaluated,List).
check_if_predicate_was_evaluated(evaluated(P)) :-
    format('Checking if "~w" has been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> true
    ; otherwise -> add_error(external_functions,'Predicate has not been evaluated', P)).
check_if_predicate_was_evaluated(not_evaluated(P)) :-
    format('Checking if "~w" has not been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> add_error(external_functions,'Predicate has been evaluated', P)
    ; otherwise -> true).

% --------------------------

:- use_module(hit_profiler,[add_hit/2, add_hit_if_not_covered/2, profile_statistics/0, reset_profiler/0,
                        get_profile_statistics/2, profile_functor/1]). %profiler:profile_statistics.

:- public 'GET_PROFILE_STATISTICS'/3.
is_not_declarative('GET_PROFILE_STATISTICS').
:- block 'GET_PROFILE_STATISTICS'(-,?,?).
'GET_PROFILE_STATISTICS'(string(PP),_DummyType,Res) :- % DummyType is just there for typing purposes
  if(profile_functor(PP),
      (get_profile_statistics(PP,Hits),
       maplist(translate_hit_to_b,Hits,List)),
      List = []),
  equal_object_optimized(List,Res, 'GET_PROFILE_STATISTICS').
  %equal_object(List,Res,'GET_PROFILE_STATISTICS'), print(res(Res)),nl.

translate_hit_to_b(hit(Arg,Nr),(Arg,int(Nr))).

:- public 'GET_PROFILE_POINTS'/1.
'GET_PROFILE_POINTS'(Res) :- findall(string(S),profile_functor(S),List),
  equal_object_optimized(List,Res,'PROFILE_POINT').

expects_spaninfo('PROFILE').
:- public 'PROFILE'/5.
:- block 'PROFILE'(-,?,?,?,?), 'PROFILE'(?,-,?,?,?). %, 'PROFILE'(?,-,?,?,?).
'PROFILE'(string(PP),Count,Value,ValueRes,SpanInfo) :-
    add_profile_exf_hit(Value,PP,SpanInfo,Count),
    equal_object(Value,ValueRes,'PROFILE').

:- block add_profile_exf_hit(-,?,?,?), add_profile_exf_hit(?,-,?,?).
add_profile_exf_hit(pred_true,PP,SpanInfo,Count) :- !, add_value_hit(Count,pred_true,PP,SpanInfo).
add_profile_exf_hit(pred_false,PP,SpanInfo,Count) :- !, add_value_hit(Count,pred_false,PP,SpanInfo).
add_profile_exf_hit(int(X),PP,SpanInfo,Count) :- !, add_int_hit(X,PP,SpanInfo,Count).
add_profile_exf_hit(string(X),PP,SpanInfo,Count) :- !, add_string_hit(X,PP,SpanInfo,Count).
add_profile_exf_hit(Value,PP,SpanInfo,Count) :- when(ground(Value), add_complex_hit(Value,PP,SpanInfo,Count)).

add_complex_hit(Value,PP,SpanInfo,Count) :- print(PP), print(' : '), translate:print_bvalue(Value),nl,
   add_value_hit(Count,Value,PP,SpanInfo).

:- block add_int_hit(-,?,?,?).
add_int_hit(X,PP,SpanInfo,Count) :- add_value_hit(Count,int(X),PP,SpanInfo).
:- block add_string_hit(-,?,?,?).
add_string_hit(X,PP,SpanInfo,Count) :- add_value_hit(Count,string(X),PP,SpanInfo).

add_value_hit(pred_true,Val,PP,_) :- add_hit(PP,Val).
add_value_hit(pred_false,Val,PP,_) :- add_hit_if_not_covered(PP,Val).

% -------------------

reset_kodkod :- 
   true. %retractall(kodkod_pred(_,_)).

%:- use_module(probsrc(bsyntaxtree), [b_compute_expression/5]).

%  EXTERNAL_PREDICATE_KODKOD == INT*BOOL;
%:- dynamic kodkod_pred/2. % now stored as attribute in WF store
do_not_evaluate_args('KODKOD').
expects_unevaluated_args('KODKOD').
expects_waitflag('KODKOD').
expects_state('KODKOD').

:- use_module(bsyntaxtree,[create_negation/2]).
:- public 'KODKOD'/5.
:- block 'KODKOD'(-,?,?,?,?).
'KODKOD'(PredRes,LocalState,State,UnevaluatedArgs,WF) :- 
    UnevaluatedArgs = [TID,IDEXPR,WAITIDEXPR,UEA],
    get_integer(TID,ID),
    construct_predicate(UEA,KPredicate),
    (PredRes == pred_false -> create_negation(KPredicate,Predicate) ; Predicate=KPredicate),
    must_get_identifiers(IDEXPR,TIdentifiers),
    get_texpr_ids(TIdentifiers,Ids),
    b_interpreter:b_compute_expression(WAITIDEXPR,LocalState,State,WaitVal,WF),
    when(ground(WaitVal),
      (  %print(compile_for(ID,Ids)),nl,
         b_compiler:b_optimize(Predicate,Ids,LocalState,State,CPredicate,WF),
         format('STORING FOR KODKOD ~w (~w): ',[ID,PredRes]), translate:print_bexpr(CPredicate),nl,
         kernel_waitflags:my_add_predicate_to_kodkod(WF,ID,CPredicate),
         %assert(kodkod_pred(ID,CPredicate)), % TO DO: attach to WF Store !; assert will break Prolog variable links !
         % TO DO: retract upon backtracking + store negation if PredRes = pred_false
         PredRes=pred_true
      )).

get_integer(b(V,_,_),I) :- get_integer2(V,I).
get_integer2(integer(V),V).
get_integer2(value(IV),V) :- nonvar(IV), IV=int(V), number(V).

:- use_module(probsrc(bsyntaxtree), [safe_create_texpr/3]).
construct_predicate(b(convert_bool(P),_,_),Res) :- !, Res=P.
construct_predicate(BoolExpr,Res) :-
   safe_create_texpr(equal(BoolExpr,b(boolean_true,boolean,[])),pred,Res).

:- use_module('kodkod/kodkod', [replace_by_kodkod/3]).
expects_unevaluated_args('KODKOD_SOLVE').
expects_waitflag('KODKOD_SOLVE').
expects_state('KODKOD_SOLVE').
:- public 'KODKOD_SOLVE'/8.
:- block 'KODKOD_SOLVE'(-,?,?,?,?,?,?,?).
'KODKOD_SOLVE'(int(ID),Ids,WaitIds,PredRes,LocalState,State,UEA,WF) :-
    when(ground(WaitIds),
            kodkod_solve_aux(ID,Ids,WaitIds,PredRes,LocalState,State,UEA,WF)).

:- use_module(library(ordsets),[ord_subtract/3,ord_intersection/3]).
kodkod_solve_aux(ID,Ids,_WaitIds,PredRes,LocalState,State,UEA,WF) :-
    %print(trigger_SOLVE(_WaitIds)),nl,
    get_wait_flag(2,'KODKOD_SOLVE',WF,LWF), % give other co-routines chance to fire
    kodkod_solve_aux2(ID,Ids,PredRes,LocalState,State,UEA,WF,LWF).

:- block kodkod_solve_aux2(?,?,?,?,?,?,?,-).
kodkod_solve_aux2(ID,_,PredRes,LocalState,State,[_,IDEXPR,WAITIDEXPR],WF,_) :-
    print('KODKOD_SOLVE:'),nl,
    %print(ls(LocalState,State)),nl,
    %findall(P,kodkod_pred(ID,P),Ps), % TO DO: obtain from WF Store
    kernel_waitflags:my_get_kodkod_predicates(WF,ID,Ps),
    nl,print(got_predicates(ID)),nl,
    length(Ps,Len), print(len(Len)),nl,
    reverse(Ps,RPs), % we reverse Ps to put predicates in same order they were added by KODKOD / seems to influence analysis time 
    conjunct_predicates(RPs,Predicate),
    translate:print_bexpr(Predicate),nl,
    must_get_identifiers(IDEXPR,TIdentifiers),
    get_texpr_ids(TIdentifiers,Ids),
    print(ids(Ids)),nl,
    ((get_identifiers(WAITIDEXPR,TWaitIds,[]), get_texpr_ids(TWaitIds,WaitIds),
     sort(WaitIds,SWIds), print(wait(SWIds)),nl,
     sort(Ids,SIds),
     ord_subtract(SIds,SWIds,IdsToSolve))
      -> b_compiler:b_optimize(Predicate,IdsToSolve,LocalState,State,CPredicate,WF) % inline known values
           ,print('compiled: '),translate:print_bexpr(CPredicate),nl,nl
      ;  CPredicate=Predicate
    ),
    (replace_by_kodkod(TIdentifiers,CPredicate,NewPredicate)
     ->  print('AFTER KODKOD: '),translate:print_bexpr(NewPredicate),nl,
         b_interpreter:b_test_boolean_expression(NewPredicate,LocalState,State,WF)
      ;  print('KODKOD cannot be applied'),nl,
         b_interpreter:b_test_boolean_expression(CPredicate,LocalState,State,WF)
    ),
    PredRes=pred_true.

must_get_identifiers(TExpr,Res) :-
   (get_identifiers(TExpr,Ids,[]) -> Res=Ids
     ; add_error('KODKOD_SOLVE','Could not get identifiers: ',TExpr),fail).

get_identifiers(b(couple(A1,A2),_,_)) --> !, get_identifiers(A1), get_identifiers(A2).
get_identifiers(b(identifier(ID),T,I)) --> [b(identifier(ID),T,I)].
get_identifiers(b(integer(_),_,_)) --> []. % to do: accept other ground values; in case they are compiled
get_identifiers(b(value(_),_,_)) --> []. % identifier has already been compiled

    
% -------------------

expects_waitflag('UNSAT_CORE'). % WF not used at the moment
expects_spaninfo('UNSAT_CORE').
:- public 'UNSAT_CORE'/5.
%  EXTERNAL_FUNCTION_UNSAT_CORE(T) == (POW(T)-->BOOL)*POW(T) --> POW(T);
%  UNSAT_CORE(Fun,Set) == Set
% example: UNSAT_CORE(f,0..200) = {2}   with   f = %x.(x:POW(INTEGER)|bool(x /\ 1..2={}))
'UNSAT_CORE'(FunToBool,Set,ValueRes,SpanInfo,WF) :-
   'UNSAT_CORE_ACC'(FunToBool,Set,[],ValueRes,SpanInfo,WF).


expects_waitflag('UNSAT_CORE_ACC'). % WF not used at the moment
expects_spaninfo('UNSAT_CORE_ACC').
%  EXTERNAL_FUNCTION_UNSAT_CORE_ACC(T) == (POW(T)-->BOOL)*POW(T)*POW(T) --> POW(T);
%  UNSAT_CORE_ACC(Fun,Set,Acc) == Set\/Acc;
% example: UNSAT_CORE_ACC(f,0..200,{}) = {2}   &   f = %x.(x:POW(INTEGER)|bool(x /\ 1..2={}))
% provides a way to initialise the accumulator of the unsat core computation
'UNSAT_CORE_ACC'(FunToBool,Set,Acc,ValueRes,SpanInfo,WF) :-
    custom_explicit_sets:expand_custom_set_to_list(Set,ESet,_,compute_apply_true),
    custom_explicit_sets:expand_custom_set_to_list(Acc,EAcc,_,compute_apply_true),
    debug_println(9,computing_unsat_core(ESet,EAcc)),
    when(ground((FunToBool,ESet)), compute_unsat_core(ESet,FunToBool,no_time_out,EAcc,ValueRes,SpanInfo,WF)).

%  EXTERNAL_FUNCTION_UNSAT_CORE_ACC_TIMEOUT(T) == (POW(T)-->BOOL)*POW(T)*POW(T)*INTEGER --> POW(T);
%  UNSAT_CORE_ACC_TIMEOUT(Fun,Set,Acc,TO) == Set\/Acc;
% example: UNSAT_CORE_ACC_TIMEOUT(f,0..200,{},100) = {2}   &   f = %x.(x:POW(INTEGER)|bool(x /\ 1..2={}))
expects_waitflag('UNSAT_CORE_ACC_TIMEOUT'). % WF not used at the moment
expects_spaninfo('UNSAT_CORE_ACC_TIMEOUT').
'UNSAT_CORE_ACC_TIMEOUT'(FunToBool,Set,Acc,int(TimeOut),ValueRes,SpanInfo,WF) :- 
    custom_explicit_sets:expand_custom_set_to_list(Set,ESet,_,compute_apply_true),
    custom_explicit_sets:expand_custom_set_to_list(Acc,EAcc,_,compute_apply_true),
    debug_println(9,computing_unsat_core(ESet,EAcc)),
    when(ground((FunToBool,ESet,TimeOut)), compute_unsat_core(ESet,FunToBool,TimeOut,EAcc,ValueRes,SpanInfo,WF)).

compute_unsat_core([],_FunToBool,_TimeOut,Core,Result,_SpanInfo,_WF) :- equal_object(Result,Core).
compute_unsat_core([H|T],FunToBool,TimeOut,Core,Result,SpanInfo,WF) :-
    debug_println(9,trying_to_remove(H,core(Core))),
    append(Core,T,CoreT),
    if(compute_apply_true_timeout(TimeOut,FunToBool,CoreT,SpanInfo),
        compute_unsat_core(T,FunToBool,TimeOut,[H|Core],Result,SpanInfo,WF), % we found a solution, H is needed in core
        compute_unsat_core(T,FunToBool,TimeOut,   Core, Result,SpanInfo,WF)  % contradiction also exists without H
      ).

:- use_module(library(timeout)).
compute_apply_true_timeout(no_time_out,FunToBool,CoreT,SpanInfo) :- !,
   compute_apply_true(FunToBool,CoreT,SpanInfo).
compute_apply_true_timeout(TO,FunToBool,CoreT,SpanInfo) :-
   time_out(compute_apply_true(FunToBool,CoreT,SpanInfo), TO, TORes),
   (TORes == success -> true ,format(user_output,'~nSUCCESS~n~n',[])
    ; format(user_output,'~nTIME-OUT: ~w~n~n',[TORes]),fail). % treat time-out like failure

compute_apply_true(FunToBool,CoreT,SpanInfo) :- init_wait_flags(WF1),
    bsets_clp:apply_to(FunToBool,CoreT,pred_true,SpanInfo,WF1), % TO DO: also pass function type
    ground_wait_flags(WF1).

expects_waitflag('MAX_SAT').
expects_spaninfo('MAX_SAT').
:- public 'MAX_SAT'/5.
%  EXTERNAL_FUNCTION_MAX_SAT(T) == (POW(T)-->BOOL)*POW(T) --> POW(T);
%  MAX_SAT(Fun,Set) == Set;
% example: MAX_SAT(f,0..3) = {0,3}   &   f = %x.(x:POW(INTEGER)|bool(x /\ 1..2={}))

'MAX_SAT'(FunToBool,Set,ValueRes,SpanInfo,WF) :-
    custom_explicit_sets:expand_custom_set_to_list(Set,ESet,_,compute_apply_true),
    debug_println(9,computing_unsat_core(ESet)),
    when(ground((FunToBool,ESet)), compute_maxsat_ground(ESet,FunToBool,ValueRes,SpanInfo,WF)).

compute_maxsat_ground(ESet,FunToBool,ValueRes,SpanInfo,WF) :-
    debug_println(9,computing_maxsat(ESet)), %nl,reset_walltime,
    get_wait_flag(1,'MAX_SAT',WF,LWF), % avoid computing straightaway in case pending co-routines fail; not really sure it is needed
    when(nonvar(LWF),compute_maxsat(ESet,FunToBool,[],ValueRes,SpanInfo,WF)).
compute_maxsat([],_FunToBool,Core,Result,_SpanInfo,_WF) :- equal_object(Result,Core).
compute_maxsat([H|T],FunToBool,Core,Result,SpanInfo,WF) :-
    debug_println(9,trying_to_add(H,core(Core))),
    %nl,nl,nl,print(maxsat(H,Core)),nl, print_walltime(user_output),nl,
    if(compute_apply_true(FunToBool,[H|Core],SpanInfo),
        compute_maxsat(T,FunToBool,[H|Core],Result,SpanInfo,WF),
        compute_maxsat(T,FunToBool,   Core, Result,SpanInfo,WF)  % contradiction also exists without H
      ).

:- use_module(clpfd_interface,[clpfd_domain/3, clpfd_inrange/3, clpfd_size/2, clpfd_eq/2]).
expects_waitflag('MAXIMIZE_EXPR').
expects_spaninfo('MAXIMIZE_EXPR').
:- public 'MAXIMIZE_EXPR'/4.
'MAXIMIZE_EXPR'(int(X),int(X),Span,WF) :-
    get_wait_flag1(WF,WF1),
    maximize_expr_wf(WF1,X,Span).
:- block maximize_expr_wf(-,?,?).
maximize_expr_wf(_WF,X,_Span) :-
    clpfd_size(X,Size), Size \= sup,
    clpfd_domain(X,Min,Max),
    Mid is round((Min + Max)/2),
    maximize_expr_wf(Min,Mid,Max,X).
maximize_expr_wf(_,_,Span) :-
    add_error(external_functions,'MAXIMIZE_EXPR / MINIMIZE_EXPR called on unbounded expression',Span),
    fail.
maximize_expr_wf(X,X,X,X).
maximize_expr_wf(_Min,Mid,Max,X) :-
    clpfd_inrange(X,Mid,Max), !,
    NMid is round((Mid+Max)/2),
    maximize_expr_wf(Mid,NMid,Max,X).
maximize_expr_wf(Min,Mid,_Max,X) :-
    clpfd_inrange(X,Min,Mid), !,
    NMid is round((Min+Mid)/2),
    maximize_expr_wf(Min,NMid,Mid,X).

expects_waitflag('MINIMIZE_EXPR').
expects_spaninfo('MINIMIZE_EXPR').
:- public 'MINIMIZE_EXPR'/4.
'MINIMIZE_EXPR'(int(X),int(X),Span,WF) :-
    clpfd_eq(-X,MaxVar),
    'MAXIMIZE_EXPR'(int(MaxVar),int(MaxVar),Span,WF).

:- use_module(kernel_tools,[bexpr_variables/2]).
% maximize is actually quite similar to CHOOSE; but it will choose maximum and use special algorithm for closures
expects_waitflag('MAXIMIZE').
expects_spaninfo('MAXIMIZE').
:- public 'MAXIMIZE'/4.
:- block 'MAXIMIZE'(-,?,?,?).
'MAXIMIZE'([],_,Span,WF) :- !,
   add_wd_error_span('MAXIMIZE applied to empty set: ',[],Span,WF).
'MAXIMIZE'(avl_set(A),H,_,_) :- !, avl_max(A,Max),kernel_objects:equal_object(H,Max).
'MAXIMIZE'(closure(P,T,Body),FunValue,SpanInfo,WF) :- !,
    kernel_tools:bexpr_variables(Body,ClosureWaitVars),
    print(wait_maximize(P,ClosureWaitVars)),nl,
    when(ground(ClosureWaitVars),maximize_closure(P,T,Body,FunValue,SpanInfo,WF)).
'MAXIMIZE'([H|T],C,Span,WF) :- !,
       when(ground([H|T]),
           (convert_to_avl([H|T],AVL) -> 'MAXIMIZE'(AVL,C,Span,WF)
             ; add_wd_error_span('Cannot determine maximum element for MAXIMIZE: ',[H|T],Span,WF),
               kernel_objects:equal_object(C,H))
            ).
'MAXIMIZE'(_,_,SpanInfo,_WF) :-
    add_error(external_functions,'MAXIMIZE requires set of values',SpanInfo),
    fail.

maximize_closure(Parameters,ParameterTypes,ClosureBody,FunValue,_SpanInfo,WF) :-
    debug_println(maximizing(Parameters)),
    length(Parameters,Len), length(ParValues,Len),
    test_closure(Parameters,ParameterTypes,ClosureBody,ParValues),
    debug_println(found_initial_solution(ParValues)),
    !,
    maximize_closure2(Parameters,ParameterTypes,ClosureBody,ParValues,FunValue,WF).
maximize_closure(_Parameters,_ParameterTypes,_ClosureBody,_FunValue,SpanInfo,_WF) :-
    add_error(external_functions,'MAXIMIZE: could not find solution for set comprehension',SpanInfo),
    fail.

maximize_closure2(Parameters,ParameterTypes,ClosureBody,PrevSol,FunValue,WF) :-
    Parameters=[PMax|_], ParameterTypes=[T1|_],
    PrevSol = [ PrevVal | _],
    print(trying_to_improve_upon(PrevVal,PMax)),nl,
    better_sol_pred(PMax,T1,PrevVal,Better), translate:print_bexpr(Better),nl,
    bsyntaxtree:conjunct_predicates([Better,ClosureBody],NewBody),
    length(Parameters,Len), length(NewParValues,Len),
    test_closure(Parameters,ParameterTypes,NewBody,NewParValues),
    debug_println(found_better_solution(NewParValues)),
    !,
    maximize_closure2(Parameters,ParameterTypes,ClosureBody,NewParValues,FunValue,WF).
maximize_closure2(_Parameters,_ParameterTypes,_ClosureBody,PrevSol,FunValue,_WF) :-
    print(cannot_improve),nl,
    tools:convert_list_into_pairs(PrevSol,ParTuple),
    kernel_objects:equal_object(ParTuple,FunValue).

test_closure(Parameters,ParameterTypes,Body,ParValues) :-
    kodkod:apply_kodkod(Parameters,ParameterTypes,Body,Body2),
    custom_explicit_sets:b_test_closure(Parameters,ParameterTypes,Body2,ParValues,positive). % positive instead of all_solutions

% generate predicate for better solution; % TO DO: support MINIMIZE by inverting args
better_sol_pred(PMax,integer,PrevVal,Better) :- !,
    Better = b(less(b(value(PrevVal),integer,[]),b(identifier(PMax),integer,[])),pred,[]).
better_sol_pred(PMax,set(T1),PrevVal,Better) :- !,
    % WARNING: for sets we will only find one maximum, not all due to the cuts in maximize_closure(2) above
    Better = b(strict_subset(b(value(PrevVal),set(T1),[]),b(identifier(PMax),set(T1),[])),pred,[]).
better_sol_pred(PMax,Type,_Val,_) :-
   add_error(external_functions,'MAXIMIZE requires first argument of set comprehension to be integer or set; cannot optimize: ',PMax:Type),fail.

% ----------------------------------

expects_spaninfo('READ_XML').
:- public 'READ_XML'/3.
:- block 'READ_XML'(-,?,?).
'READ_XML'(string(File),Res,Span) :-
   read_xml(File,auto,Res,Span).

expects_spaninfo('READ_XML').
:- public 'READ_XML'/4.
:- block 'READ_XML'(-,?,?,?),'READ_XML'(?,-,?,?).
'READ_XML'(string(File),string(Encoding),Res,Span) :-
   read_xml(File,Encoding,Res,Span).

% EXTERNAL_FUNCTION_READ_XML == STRING --> seq(XML_ELement_Type);
% with  XML_ELement_Type == struct( recId: NATURAL1, pId:NATURAL, element:STRING, attributes: STRING +-> STRING, meta: STRING +-> STRING );

:- use_module(tools, [read_string_from_file/3]).
:- use_module(xml2b, [convert_xml_to_b/2]).
:- use_module(preferences, [is_of_type/2]).
:- block read_xml(-,?,?,?), read_xml(?,-,?,?).
read_xml(File,EncodingPref,Res,Span) :-
    (is_of_type(EncodingPref,text_encoding) -> Encoding = EncodingPref
        ; add_error(read_xml,'Illegal encoding preference (must be "auto", "UTF-8", "UTF-16", "ISO-8859-1",...): ',EncodingPref),
          Encoding=auto
    ),
    b_absolute_file_name(File,AFile),
    statistics(walltime,_),
    safe_call(read_string_from_file(AFile,Encoding,Codes),Span),
    statistics(walltime,[_,W1]), formatsilent('% Walltime ~w ms to read in ~w~n',[W1,AFile]),
    (convert_xml_to_b(Codes,BDataSequence) -> true
       ; add_error(read_xml,'Converting XML file to B failed: ',AFile,Span),fail),
    statistics(walltime,[_,W2]),
    formatsilent('% Walltime ~w ms to parse and convert XML in ~w~n',[W2,AFile]),
    kernel_objects:equal_object(Res,BDataSequence).
    

expects_spaninfo('READ_XML_FROM_STRING').
:- public 'READ_XML_FROM_STRING'/3.
:- block 'READ_XML_FROM_STRING'(-,?,?).
'READ_XML_FROM_STRING'(string(Contents),Res,Span) :-
    atom_codes(Contents,Codes),
    (convert_xml_to_b(Codes,BDataSequence) -> true
       ; add_error(read_xml,'Converting XML string to B failed: ',Contents,Span),fail),
    statistics(walltime,[_,W2]), 
    formatsilent('% Walltime ~w ms to parse and convert XML~n',[W2]),
    kernel_objects:equal_object(Res,BDataSequence).

% ----------------------------------
% A flexible CVS Reader that always uses the same type
% rec([field(attributes,BAttributes),field(lineNr,LineNr)])

% EXTERNAL_FUNCTION_READ_CSV_STRINGS == STRING --> seq(CSV_ELement_Type);
% with  CSV_ELement_Type == STRING +-> STRING ;

expects_spaninfo('READ_CSV_STRINGS').
:- public 'READ_CSV_STRINGS'/3.
:- block 'READ_CSV_STRINGS'(-,?,?).
'READ_CSV_STRINGS'(string(File),Res,Span) :- read_csv_strings(File,Res,Span).

read_csv_strings(File,ResAVL,Span) :- 
    b_absolute_file_name(File,AFile),
    safe_call(open(AFile,read,Stream),Span),
    statistics(walltime,_),
    read_line(Stream,Line1),
    (Line1 = end_of_file -> ResAVL = []
    ; split_csv_line([],HeadersCodes,Span,Line1,[]),
      maplist(process_header,HeadersCodes,Headers),
      formatsilent('CSV Line Headers = ~w for ~w~n',[Headers,AFile]),
      safe_call(read_csv_strings_lines(Stream,1,Span,Headers,Res),Span),
      statistics(walltime,[_,W1]), formatsilent('% Walltime ~w ms to read in ~w~n',[W1,AFile]),
      kernel_objects:equal_object_optimized(Res,ResAVL),
      statistics(walltime,[_,W2]), formatsilent('% Walltime ~w ms to normalise in ~w~n',[W2,AFile])
    ),
    close(Stream). % TO DO: call_cleanup

process_header(HeaderC,HeaderAtom) :- safe_atom_codes(HeaderAtom,HeaderC).

read_csv_strings_lines(Stream,PrevLineNr,Span,Headers,Res) :-
   read_line(Stream,Line),
   (Line = end_of_file -> Res = []
    ; LineNr is PrevLineNr+1,
      split_csv_line([],AtomsC,Span,Line,[]),
      Res = [(int(PrevLineNr),Attrs)|ResT],
      construct_attributes(AtomsC,Headers,Attrs,Span,LineNr),
      read_csv_strings_lines(Stream,LineNr,Span,Headers,ResT)
    ).

construct_attributes([],_,[],_,_).
construct_attributes([ [] | T],[ _|HT],Res,Span,LineNr) :- !,
  construct_attributes(T,HT,Res,Span,LineNr).
construct_attributes([ AtomC | _T], [], Res,Span,LineNr) :- !,
       safe_atom_codes(A,AtomC),
       ajoin(['Additional field on line ',LineNr,' : '],Msg),
       add_error('READ_CSV_STRINGS',Msg,A,Span),
       Res=[].
construct_attributes([ AtomC | T], [ Header | HT], [ (string(Header),string(A)) | ResT],Span,LineNr) :-
   safe_atom_codes(A,AtomC),
   construct_attributes(T,HT,ResT,Span,LineNr).

% ----------------------------------
% A CSV Reader that uses the B types to adapt how it reads in values

expects_spaninfo('READ_CSV').
expects_type('READ_CSV').

:- public 'READ_CSV'/6.
:- block 'READ_CSV'(-,?,?,?,?,?).
'READ_CSV'(string(File),BoolSkipLine1,AllowExtraValues,Res,Type,Span) :-
   read_csv(File,BoolSkipLine1,AllowExtraValues,Res,Type,Span).

:- public 'READ_CSV'/4.
'READ_CSV'(string(File),Res,Type,Span) :-
   'READ_CSV'(string(File),pred_false,pred_false,Res,Type,Span).

:- block read_csv(-,?,?,?,?,?).
read_csv(File,BoolSkipLine1,AllowExtraValues,Res,Type,Span) :-
    get_column_types(Type,Types,Skel,Vars,Span),
    b_absolute_file_name(File,AFile),
    safe_call(open(AFile,read,Stream),Span),
    (debug:debug_mode(off) -> true
     ; formatsilent('~nReading CSV file ~w with Types = ~w and Skel = ~w for Vars = ~w~n~n',[AFile,Types,Skel,Vars])),
    (read_csv_lines(1,BoolSkipLine1,AllowExtraValues,Stream,Types,Skel,Vars,Res,Span) -> true
       ; add_error('READ_CSV','Reading CSV File failed: ',AFile)),
    close(Stream). % TO DO: call_cleanup

read_csv_lines(Nr,BoolSkipLine1,AllowExtraValues,Stream,Types,Skel,Vars,Res,Span) :-  %print(line(Nr,Types)),nl,
    read_line(Stream,Line),
    debug:debug_println(9,read_line(Nr,Line)),
   (Line = end_of_file -> Res = []
    ; split_csv_line([],AtomsC,Span,Line,[]),
      %print(atoms(Nr,Atoms)),nl,
      %maplist(convert_atom_codes(Span),Types,Atoms,Tuple),  % TO DO: check same length
      ((Nr=1,BoolSkipLine1=pred_true)
       -> maplist(codes_to_atom,AtomsC,Atoms), format('Skipping line 1: ~w~n',[Atoms]),
          Res=ResT
       ;  l_convert_atom_codes(Types,AtomsC,Tuple,Span,Nr,AllowExtraValues),
          copy_term((Vars,Skel),(Tuple,BValue)),
          Res = [BValue|ResT]
      ),
      N1 is Nr+1,
      read_csv_lines(N1,BoolSkipLine1,AllowExtraValues,Stream,Types,Skel,Vars,ResT,Span)).

codes_to_atom(C,A) :- atom_codes(A,C).

l_convert_atom_codes([],[],Vals,_Span,_LineNr,_) :- !, Vals=[].
l_convert_atom_codes([Type|TT],[AtomC|TC],[Val|VT],Span,LineNr,AllowExtraValues) :- !,
      convert_atom_codes(Span,LineNr,Type,AtomC,Val),
      l_convert_atom_codes(TT,TC,VT,Span,LineNr,AllowExtraValues).
l_convert_atom_codes([Type|TT],[],_,Span,LineNr,_) :- !,
    ajoin(['Missing value(s) on CSV line ',LineNr,' for types: '],Msg),
    add_error('READ_CSV',Msg,[Type|TT],Span),fail.
l_convert_atom_codes([],[AtomC|AT],Vals,Span,LineNr,AllowExtraValues) :- !,
    (AllowExtraValues=pred_true -> true
     ; safe_atom_codes(Atom,AtomC),
       ajoin(['Extra value(s) on CSV line ',LineNr,' : ',Atom,' Rest = '],Msg),
       add_error('READ_CSV',Msg,AT,Span)
    ), Vals=[].

convert_atom_codes(Span,LineNr,T,AtomC,Val) :-
  (convert_atom2(T,AtomC,Val,Span,LineNr) -> true ;
    atom_codes(Atom,AtomC),
    ajoin(['Could not convert value to type ',T,' on CSV line ',LineNr,': '],Msg),
    add_error('READ_CSV',Msg,Atom,Span),
    Val = term('READ_CSV_ERROR')).
:- use_module(tools,[safe_number_codes/2, safe_atom_codes/2]).
convert_atom2(string,C,string(A),_,_) :- safe_atom_codes(A,C).
convert_atom2(integer,C,int(Nr),_Span,LineNr) :-
   (C=[] -> (debug:debug_mode(off) -> true ; format('Translating empty string to -1 on line ~w~n',[LineNr])),
            % add_warning('READ_CSV','Translating empty string to -1 on line: ',LineNr,Span), 
            Nr is -1
    ; safe_number_codes(Nr,C)).
convert_atom2(global(G),C,FD,Span,LineNr) :- safe_atom_codes(A,C),convert_to_enum(G,A,FD,Span,LineNr).
convert_atom2(boolean,C,Bool,Span,LineNr) :- safe_atom_codes(A,C),
  (convert_to_bool(A,Bool) -> true
   ; ajoin(['Could not convert value to BOOL on line ',LineNr,'; assuming FALSE: '],Msg),
     add_warning('READ_CSV',Msg,A,Span),
     Bool=pred_false).
% TO DO: support set(_), seq(_)

convert_to_bool('0',pred_false).
convert_to_bool('1',pred_true).
convert_to_bool('FALSE',pred_false).
convert_to_bool('TRUE',pred_true).
convert_to_bool('false',pred_false).
convert_to_bool('true',pred_true).

convert_to_enum(G,A,FD,Span,LineNr) :- string_to_enum2(A,FD),
  (FD=fd(_,G) -> true ; add_error('READ_CSV','Enumerated element of incorrect type: ',A:expected(G):line(LineNr),Span)).

% split a CSV line into individual AtomCode lists
split_csv_line([],T,Span) --> " ", !, split_csv_line([],T,Span).
split_csv_line(Acc,[AtomC|T],Span) --> sep, !,
   {reverse(Acc,AtomC)}, split_csv_line([],T,Span).
split_csv_line([],[String|T],Span) --> [34], % "
   scan_to_end_of_string([],String,Span),
   (sep -> split_csv_line([],T,Span)
    ; [XX] -> {add_error('READ_CSV','String not followed by separator: ',String,Span)}, split_csv_line([XX],T,Span)
    ;  {T=[]}).
split_csv_line(Acc,Atoms,Span) --> [X], !, split_csv_line([X|Acc],Atoms,Span).
split_csv_line(Acc,[AtomC],_Span) --> !, % end of line
   {reverse(Acc,AtomC)}.

sep --> " ",!, sep.
sep --> ("," ; ";" ; " ,").

scan_to_end_of_string(Acc,AtomC,_) --> [34],!,{reverse(Acc,AtomC)}.
scan_to_end_of_string(Acc,String,Span) --> [X],!, scan_to_end_of_string([X|Acc],String,Span).
scan_to_end_of_string(Acc,String,Span) -->
  {reverse(Acc,RAcc),atom_codes(String,RAcc),add_error('READ_CSV','String not terminated: ',String,Span)}.

% get the list of column types along with Value skeleton with variables in the places where CSV values should be inserted
get_column_types(set(X),Types,Skel,Vars,_) :- !,flatten_pairs(X,Types,[]), %print(types(Types)),nl,
   flatten_pairs_value_skeleton(X,Skel,Vars,[]). % print(skel(Skel,Vars)),nl.
get_column_types(seq(X),Types,Skel,Vars,Span) :- !,
    get_column_types(set(couple(integer,X)),Types,Skel,Vars,Span).
get_column_types(Type,_,_,_,Span) :- add_error('READ_CSV','Invalid return type (must be relation): ',Type,Span),fail.

% flatten couple pairs and generate skeleton term
flatten_pairs(couple(A,B)) --> !,flatten_pairs(A), flatten_pairs(B).
flatten_pairs(record(Fields)) --> !, flatten_fields(Fields).
flatten_pairs(X) --> [X].

/* note: currently record fields have to be alphabetically in the same order as in CSV file: the fields are sorted by ProB's kernel ! : TO DO: try and recover the original order of the fields */
flatten_fields([]) --> [].
flatten_fields([field(_Name,Type)|T]) --> [Type], % no nesting allowed; this has to be a basic type for CSV reading to work
   flatten_fields(T).

% generate a value skeleton
flatten_pairs_value_skeleton(couple(A,B),(CA,CB)) --> !,
    flatten_pairs_value_skeleton(A,CA), flatten_pairs_value_skeleton(B,CB).
flatten_pairs_value_skeleton(record(Fields),rec(SF)) --> !, flatten_field_skel(Fields,SF).
flatten_pairs_value_skeleton(_X,Var) --> [Var].

flatten_field_skel([],[]) --> [].
flatten_field_skel([field(Name,_Type)|T],[field(Name,Var)|ST]) --> [Var],
   flatten_field_skel(T,ST).
