% (c) 2012-16 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kodkod_integer_recalc,[integer_recalc/2]).

:- use_module('../module_information',[module_info/2]).
:- module_info(group,kodkod).
:- module_info(description,'This module provides functionality to adapt the outcome of the interval analysis to the needs of the Kodkod translation.').

:- use_module(library(lists)).

:- use_module(probsrc(self_check)).

:- use_module(probsrc(bsyntaxtree),[create_texpr/4,syntaxtransformation/5,get_texpr_info/2,get_texpr_type/2]).
:- use_module(probsrc(interval_calc)).
:- use_module(probsrc(inf_arith)).

:- use_module(probsrc(predicate_analysis),[max_type_cardinality/2]).

% re-calculate integer expressions
% The predicate analysis computes the possible integer intervals of expressions
% for valid solutions. E.g. for the predicate "x:1..100 & y:2..88 & x+y = x*y"
% it finds out that "x*y" lies in the range of "2..200" because the expresssion
% equals "x+y" which lies obviously in "3..188".
% But if we use that information as a basis for Kodkod's bitwidth, we can receive
% invalid solutions because of an integer overflow on the right side.
% E.g. Kodkod may choose a bitwidth that allows 511 as the maximum integer, then
% it can find the solution x=28 & y=20 because x*y = 560 which leads to an
% (unnoticed) overflow such that x*y = x*y mod 512 = 560 mod 512 = 48 = x+y
%
% To prevend this behaviour, we use the interval of basic integer expressions
% (variables, integers, cardinality) to re-calc the possible interval of
% integer expressions

:- assert_must_succeed((
    create_texpr(identifier(x),integer,[analysis([interval:irange(1,100)])],X),
    create_texpr(identifier(y),integer,[analysis([interval:irange(2,88)])],Y),
    create_texpr(multiplication(X,Y),integer,[other,analysis([interval:irange(3,188)])],M),
    integer_recalc(M,N),
    ground(N),                   % make sure no variables are in the result
    get_texpr_info(N,NInfo),
    memberchk(other,NInfo),      % the old information must still be there
    memberchk(analysis([A]),NInfo),
    A = interval:irange(2,8800)  % the updated interval information should reflect the multiplication
                       )).

integer_recalc(TExpr,NTExpr) :-
    create_texpr(Expr,Type,Info,TExpr),
    create_texpr(NExpr,Type,NInfo,NTExpr),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    maplist(integer_recalc,Subs,NSubs),
    ( integer_recalc2(NExpr,Interval) ->
        safe_add_interval_info(Interval,Info,NInfo)
    ; otherwise ->
        Info=NInfo).

:- use_module(probsrc(bsyntaxtree),[get_texpr_set_type/2]).
:- use_module(probsrc(debug),[debug_println/2]).

integer_recalc2(value(IV),irange(I,I)) :- nonvar(IV), IV=int(I), number(I).
integer_recalc2(integer(I),irange(I,I)).
integer_recalc2(add(A,B),C) :-
    get_interval_info(A,AI), get_interval_info(B,BI),
    interval_addition(AI,BI,C).
integer_recalc2(minus(A,B),C) :-
    get_interval_info(A,AI), get_interval_info(B,BI),
    interval_subtraction(AI,BI,C).
integer_recalc2(unary_minus(A),C) :-
    get_interval_info(A,AI),
    interval_negation(AI,C).
integer_recalc2(multiplication(A,B),C) :-
    get_interval_info(A,AI), get_interval_info(B,BI),
    interval_multiplication(AI,BI,C).
integer_recalc2(division(A,B),C) :-
    get_interval_info(A,AI), get_interval_info(B,BI),
    AI=irange(Am,AM),BI=irange(Bm,_BM), 
    ( number(Bm), Bm>=0 -> C = AI
    ; number(Am),number(AM) ->
        M is max(abs(Am),abs(AM)),
        C = irange(-M,M)
    ; otherwise ->
        C = irange(-inf,inf)).
integer_recalc2(modulo(_A,B),irange(0,M)) :-
    get_interval_info(B,irange(_Bm,BM)),
    M infis BM-1.
integer_recalc2(card(S),irange(MinCard,MaxCard)) :-
    % TODO[DP,2016-04-20]: This is a very pessimistic approach:
    % The cardinality is maximally as high as the type allows
    % Maybe an alternative approach would be to add a second infered information to each node,
    % the "guaranteed" integer interval which is independent of overflows. E.g. "card({a,b,c}) : 1..3"
    get_texpr_set_type(S,Type),
    % print(recalc(Type,S)),nl,
    ( extract_min_max_card(S,Type,C1,C2)        -> (MinCard,MaxCard)=(C1,C2),
                                                    debug_println(9,extract_max_card_for_kodkod(C1,C2))
    ; max_type_cardinality(Type,MaxCard) -> MinCard=0
    %; otherwise                       -> Card=64 % TO DO allow user to provide a custom upper bound ???
    ; otherwise                       -> MinCard=0, MaxCard = inf).

% TO DO: use definitely_not_empty_and_finite from b_ast_cleanup for MinCard ?

:- use_module(library(lists),[maplist/3]).
extract_min_max_card(b(E,_,_),Type,MinCard,MaxCard) :- extract_min_max_card2(E,MinCard,MaxCardE),
   ((max_type_cardinality(Type,CType), number(CType), CType < MaxCardE)
     -> MaxCard = CType
      ; MaxCard = MaxCardE).
extract_min_max_card2(comprehension_set(IDs,_), 0, Card) :- 
     maplist(extract_max_card_id,IDs,Cards),
     simple_mul_list(Cards,1,Card).
extract_min_max_card2(set_extension(L), MinCard, MaxCard) :- length(L,MaxCard),
     (MaxCard>0 -> MinCard=1 ; MinCard=0). % should always be 1
extract_min_max_card2(sequence_extension(L), MinCard, MaxCard) :- length(L,MaxCard),
     (MaxCard>0 -> MinCard=1 ; MinCard=0). % should always be 1
extract_min_max_card2(empty_set,0,0).
extract_min_max_card2(empty_sequence,0,0).
% integer_set(_) ?
% TO DO: check with Daniel if ok
% I think this is ok as the comprehension set will be computed separately and will lead a value between 0 and Card, if  extract_max_card_id succeeds; no overflow should ever occur here

extract_max_card_id(b(identifier(_),_,Info),MaxCard) :-
   member(analysis(I),Info),
   member(interval:irange(A,B),I), % should we support other analysis infos ??
   number(A),number(B),
   !,
   MaxCard is B+1-A.
extract_max_card_id(b(_,Type,_),MaxCard) :- % use max card of type; maybe other identifiers have restricted cardinality
   max_type_cardinality(Type,MaxCard), number(MaxCard).


simple_mul_list([],A,A).
simple_mul_list([H|T],Acc,R) :- NA is Acc*H, simple_mul_list(T,NA,R).

% b(identifier(p),integer,[analysis([possible_values:irange(0,16),interval:irange(1,16)]),nodeid(pos(4360,1,383,8,383,8))])

get_interval_info(TExpr,I) :-
    get_texpr_info(TExpr,Info),
    memberchk(analysis(An),Info),
    memberchk(interval:I,An).

safe_add_interval_info(Interval,Info,[analysis([interval:Interval|AnRest])|InfoRest]) :-
    selectchk(analysis(A),Info,InfoRest),!,
    ( selectchk(interval:_,A,AnRest) -> true
    ; otherwise -> AnRest=A).
safe_add_interval_info(Interval,Info,[analysis([interval:Interval])|Info]).
