% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(logger, [set_log_file/1, get_log_file/1, set_logging_mode/1, reset_logger/0,
                              writeln_log/1,
                              writeln_log_time/1,
                              write_xml_element_to_log/2,
                              write_prolog_term_as_xml_to_log/1,
                              start_xml_group_in_log/1, start_xml_group_in_log/3,
                              stop_xml_group_in_log/1, stop_xml_group_in_log_no_statistics/1,
                              close_all_xml_groups_in_log/0,
                              logging_is_enabled/0]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'This module is responsible for (xml and prolog) logging.').


:- use_module(self_check).
:- use_module(error_manager).
:- use_module(library(xml)).
:- use_module(library(lists)).

:- dynamic logfile/1.
set_log_file(F) :- retractall(logfile(_)), assert(logfile(F)).
get_log_file(F) :- logfile(F).

:- dynamic logging_mode/1.
logging_mode(prolog).
% valid modes: prolog and xml
set_logging_mode(Mode) :- retractall(logging_mode(_)), assert(logging_mode(Mode)),
   (Mode=xml -> format_log_header(reset,'<?xml version="1.0" encoding="ASCII"?>~n',[]) ; true).
   
   % TO DO: use:
    %get_preference(xml_encoding,EncodingPref),

logging_is_enabled :- logfile(_),!.

prolog_log_file(F) :- logfile(F), logging_mode(prolog).

reset_logger :- retractall(open_xml_group(_,_)),
   retractall(logfile(_)), set_logging_mode(prolog).

% TO DO: try and move most writeln_log calls to write_xml_element_to_log format
writeln_log(Term) :-
    (prolog_log_file(F)
      -> (open(F,append,S),
          write_term(S,Term,[quoted(true)]), write(S,'.'),nl(S),
          close(S))
     ; true
    ).

format_log(FormatString,Args) :- %format(FormatString,Args),nl,
    (logfile(F)
      -> (open(F,append,S),
          format(S,FormatString,Args),
          close(S))
     ; true
    ).

writeln_log_time(Term) :-
    (prolog_log_file(_) ->
       statistics(runtime,[Time,_]),
       statistics(walltime,[WTime,_]),
       Term=..[H|Args],
       append(Args,[Time,WTime],NArgs),
       NT =.. [H|NArgs],
       writeln_log(NT)
    ;  true).

:- use_module(library(file_systems),[file_exists/1]).
format_log_header(reset,FormatString,Args) :- !, % reset means we want to start with a fresh log file
    (logfile(F)
      -> (open(F,write,S),
          format(S,FormatString,Args),
          close(S))
     ; true
    ).
format_log_header(_,FormatString,Args) :-
% like format_log, but only writes if the file does not exist yet
    ((logfile(F), \+ file_exists(F))
      -> (open(F,append,S),
          format(S,FormatString,Args),
          close(S))
     ; true
    ).

:- assert_must_succeed((logger:xml_encode_text("b<>c",R),R=="b&lt;&gt;c")).
xml_encode_text(Codes,Encoded) :- XML = xml([],[pcdata(Codes)]), xml_parse(Encoded,XML).
% we could also use xml:pcdata_generation(Codes, Encoded, []).

:- assert_must_succeed((logger:xml_encode_element(check_goal,[true/1],R),R=="<check_goal true=\"1\" />")).
xml_encode_element(Tag,Attributes,Encoded2) :- 
   maplist(prepare_attribute,Attributes,XMLAttr),
   XML = xml([],[element(Tag,XMLAttr,[])]),
   xml_parse(Encoded,XML),
   peel_off_leading_newline(Encoded,Encoded2).
   
peel_off_leading_newline([10|T],R) :- !, peel_off_leading_newline(T,R).
peel_off_leading_newline([13|T],R) :- !, peel_off_leading_newline(T,R).
peel_off_leading_newline(R,R).

% write a tag with attributes to the log file
write_xml_element_to_log(_,_) :- \+ logfile(_), !.
write_xml_element_to_log(Tag,Attributes) :- logging_mode(xml),!,
   xml_encode_element(Tag,Attributes,Encoded),
   indent_log(WS),
   format_log("~s~s~n",[WS,Encoded]).
write_xml_element_to_log(Tag,Attributes) :-
   Term =.. [Tag,Attributes],
   format_log("~w.~n",[Term]).

% write a Prolog Term either as Prolog Term in Prolog mode or in nested XML form
write_term_to_log(Term) :- logging_mode(xml),!,
   write_prolog_term_as_xml_to_log(Term).
write_term_to_log(Term) :- writeln_log(Term).

write_prolog_term_as_xml_to_log(A) :- number(A),!,
   indent_log(WS),
   format_log("~s<number>~w</number>~n",[WS,A]).
write_prolog_term_as_xml_to_log(A) :- atomic(A),!, convert_to_codes(A,Codes),
   xml_encode_text(Codes,Encoded),
   indent_log(WS),
   (is_a_file_path(Encoded) -> format_log("~s<path>~s</path>~n",[WS,Encoded])
    ; format_log("~s<~s />~n",[WS,Encoded])).
write_prolog_term_as_xml_to_log([H|T]) :- !, % Note: we assume we have a proper list !
   start_xml_group_in_log(list),
   maplist(write_prolog_term_as_xml_to_log,[H|T]),
   stop_xml_group_in_log_no_statistics(list).
write_prolog_term_as_xml_to_log(T) :- T =.. [Functor|Args],
   start_xml_group_in_log(Functor),
   maplist(write_prolog_term_as_xml_to_log,Args),
   stop_xml_group_in_log_no_statistics(Functor).

:- use_module(tools,[host_platform/1]).
is_a_file_path(Codes) :- member(47,Codes).
is_a_file_path(Codes) :- host_platform(windows), member(92,Codes). % windows

prepare_attribute('='(Tag,Atom),'='(Tag,Codes)) :- convert_to_codes(Atom,Codes).
prepare_attribute('/'(Tag,Atom),'='(Tag,Codes)) :- convert_to_codes(Atom,Codes).

:- use_module(library(codesio),[write_to_codes/2]).
convert_to_codes(V,Codes) :- var(V),!,Codes="_".
convert_to_codes([H|T],Codes) :- number(H),!, Codes=[H|T].
convert_to_codes(N,Codes) :- number(N),!, number_codes(N,Codes).
convert_to_codes(A,Codes) :- atom(A),!,atom_codes(A,Codes).
convert_to_codes(A,Codes) :- write_to_codes(A,Codes).

:- dynamic open_xml_group/2, nesting_level/1.
nesting_level(0).
update_nesting_level(X) :- retract(nesting_level(Y)),
  New is Y+X, assert(nesting_level(New)).

space(32).
indent_log(WS) :- nesting_level(Lvl), length(WS,Lvl),
    maplist(space,WS).

check_and_generate_group_stats(Group,Stats) :- open_xml_group(A,_),!,
   (A=Group
    -> retract(open_xml_group(Group,WTimeStart)),
       (Stats=no_statistics -> true
        ; statistics(walltime,[WTimeEnd,_]),
          Delta is WTimeEnd - WTimeStart,
          statistics(memory_used,M),
          write_xml_element_to_log(statistics,[walltime/Delta,walltime_since_start/WTimeEnd,memory_used/M])
        ),
       update_nesting_level(-1)
    ; add_internal_error('XML closing tag mismatch: ', Group/A)).
check_and_generate_group_stats(Group,_) :- 
    add_internal_error('XML closing tag error, no tag open: ', Group).

start_xml_group_in_log(Group) :- logging_mode(xml),!,
   statistics(walltime,[WTime,_]),
   indent_log(WS),
   asserta(open_xml_group(Group,WTime)),
   update_nesting_level(1),
   format_log("~s<~w>~n",[WS,Group]).
start_xml_group_in_log(_).

% we currently only support a single attribute and value
start_xml_group_in_log(Group,Attr,Value) :- logging_mode(xml),!,
   statistics(walltime,[WTime,_]),
   indent_log(WS),
   asserta(open_xml_group(Group,WTime)),
   update_nesting_level(1),
   format_log("~s<~w ~w=\"~w\">~n",[WS,Group,Attr,Value]).
start_xml_group_in_log(_,_,_).

stop_xml_group_in_log(Group,Stats) :- logging_mode(xml),!,
    check_and_generate_group_stats(Group,Stats),
    indent_log(WS),
    format_log("~s</~w>~n",[WS,Group]).
stop_xml_group_in_log(_,_).


stop_xml_group_in_log(Group) :- stop_xml_group_in_log(Group,statistics).
stop_xml_group_in_log_no_statistics(Group) :- stop_xml_group_in_log(Group,no_statistics).

% call if you need to prematurely exit probcli
close_all_xml_groups_in_log :- open_xml_group(Group,_), !,
   stop_xml_group_in_log(Group),
   close_all_xml_groups_in_log.
close_all_xml_groups_in_log.

 /* 
 <?xml version="1.0" encoding="ASCII"?>
 
  | ?- xml_parse("<PT ID=\"2\" stID=\"3\"/>",R).
R = xml([],[element('PT',['ID'=[50],stID=[51]],[])]) ? 
  
<PointsTelegram elementID="W90" stationID="FR" interlockingID="FR" interlockingElementID="W90"/>


*/