% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% LTL model checking

:- module(ltl,[ltl_model_check/4,ltl_model_check2/4,
               ltl_check_assertions/2, parse_ltlfile/2,
               ltl_model_check_with_ce/4,
               parse_and_pp_ltl_formula/2, % imported from ltl_tools.pl
               ltl_formula_structure/3,
               pp_ltl_formula/2,
               counterexample_node/1, counterexample_op/1,
               preprocess_formula/2,
               extract_formulas/4,
               %% reset_ltl/0,
               find_atomic_property_formula/2,
               init_states/1,
               executetrace/1,set_trace_to_init/1,
               parse_and_preprocess_formula/3,
               is_fairness_implication/1
              ]).

/* SICSTUS libraries */
:- use_module(library(lists)).
:- use_module(library(ordsets)).

/* ProB standard libraries */
% B 
:- use_module(probsrc(bsyntaxtree), [get_texpr_expr/2,get_texpr_pos/2,predicate_identifiers/2]).
:- use_module(probsrc(bmachine),[b_get_machine_operation/4,
                        b_definition_prefixed/4, b_get_typed_definition/3, b_top_level_operation/1]).
% CSP 
:- use_module(probcspsrc(haskell_csp),[filter_formulas_from_pragmas/3]).
% utility modules
:- use_module(probsrc(tools)).
:- use_module(probsrc(translate)).
:- use_module(probsrc(debug)).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(tools_printing), [print_error/1]).
:- use_module(probsrc(preferences),[get_preference/2,preference/2]).
% state space relevant modules
:- use_module(probsrc(specfile),[animation_mode/1]).
:- use_module(probsrc(state_space),[set_max_nr_of_new_impl_trans_nodes/1,get_max_nr_of_new_impl_trans_nodes/1,impl_trans_term/3,
                           current_state_id/1,visited_expression/2, visited_expression_id/1,
                           transition/3,transition/4,
                           state_space_add/2,retract_visited_expression/2,  % used for test cases
                           impl_trans_term_all/2,state_space_reset/0,
                           execute_id_trace_from_current/3, set_context_state/1,
                           state_space_initialise/0, compute_transitions_if_necessary/1, not_all_transitions_added/1]).
:- use_module(probsrc(state_space_exploration_modes),[depth_breadth_first_mode/1,set_depth_breadth_first_mode/1]).
% Optimisations (POR and PGE)
:- use_module(probporsrc(ample_sets),[stutter_action/1, visible_action/1,check_cycle_proviso/0,reset_runtime_dynamic_predicates_POR/0]).
:- use_module(probporsrc(static_analysis),[compute_dependendency_relation_of_all_events_in_the_model/3]).
:- use_module(probpgesrc(pge_algo), [is_pge_opt_on/0]).
% static analyses
:- use_module(probsrc(b_read_write_info),[b_get_read_write/3,b_get_operation_read_guard_vars/3]).
% other 
:- use_module(probsrc(plugins)).

/* ProB LTL modules */
:- use_module(probltlsrc(ltl_fairness)).
:- use_module(probltlsrc(ltl_verification)).
:- use_module(probltlsrc(ltl_safety)).
:- use_module(probltlsrc(ltl_tools)).
:- use_module(probltlsrc(ltl_translate)).
:- use_module(probltlsrc(ltl_propositions)).
:- use_module(probltlsrc(state_space_explorer)).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides LTL model checking on models.').

/* LTL relevant PL modules */
% import the low-level C-library wrapper
:- use_module(extension('ltlc/ltlc')).

parse_and_preprocess_formula(FormulaAsAtom,Type,Result) :-
  temporal_parser(FormulaAsAtom,Type,ParseResult),
  preprocess_formula(ParseResult,Result).

% a utility to find a node satisfying a restricted atomic LTL property
find_atomic_property_formula(AtomFormula,NodeID) :-
    parse_and_preprocess_formula(AtomFormula,ltl,PF),
    visited_expression_id(NodeID),
    check_atomic_property_formula(PF,NodeID).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get LTL formulas from 'DEFINITIONS' for B machines and
% from pragmas ({-# ... #-}) for CSP-M specifications
:- dynamic ltl_formula_cache/2.

get_ltl_formulas(Names,Strings,Formulas) :-
    (animation_mode(b)
        -> get_ltl_formula_strings(Names,Strings)
    ; animation_mode(cspm)
        -> filter_formulas_from_pragmas(ltl,Names,Strings)
    ; animation_mode(csp_and_b)
        -> get_ltl_formula_strings(BNames,BStrings),
           filter_formulas_from_pragmas(ltl,CSPNames,CSPStrings),
           append(BNames,CSPNames,Names),
           append(BStrings,CSPStrings,Strings)
    ),
    (ltl_formula_cache(Strings,Formulas) -> true
    ; otherwise ->
        retractall(ltl_formula_cache(_,_)),
        temporal_parser_l(Strings,ltl,Formulas),
        assert(ltl_formula_cache(Strings,Formulas))).

get_ltl_formula_strings(Names,Formulas) :-
    findall(ltl_assertion(Name,F),get_ltl_formula_string(Name,F),Assertions),
    findall(N,member(ltl_assertion(N,_),Assertions),Names),
    findall(F,member(ltl_assertion(_,F),Assertions),Formulas).
get_ltl_formula_string(Tail,S) :-
    b_definition_prefixed(_,'ASSERT_LTL',Tail,DefName),
    b_get_typed_definition(DefName,[],F),
    ( get_texpr_expr(F,string(S)) -> true
    ; get_texpr_pos(F,Pos) ->
        add_all_perrors([error('Expected String for LTL formula',Pos)]),
        fail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check all LTL assertions in a machine

ltl_check_assertions(Limit,Outcome) :-
    get_ltl_formulas(Names,Strings,Formulas),
    extract_expectations(Names,Exp),
    ltl_check_assertions2(Names,Strings,Formulas,Exp,Limit,Outcomes),
    merge_outcomes(Outcomes,Outcome).

ltl_check_assertions2([],[],[],[],_,[]).
ltl_check_assertions2([Name|Nrest],[String|Srest],[Formula|Frest],[Exp|Erest],Limit,[Outcome|Orest]) :-
    print('Checking LTL formula '),print(Name),print(': '),print(String),nl,
    flush_output(user_output),
    catch( ltl_model_check2(Formula,Limit,init,Status),
           E,
           Status = exception(E)),
    ( Exp=pass, Status=ok -> Outcome=pass, print('no counter-example found, test passed\n')
    ; Exp=fail, Status=no -> Outcome=pass, print('counter-example found, test passed\n')
    ; Status=exception(E) ->
        add_error(ltl,'Caught exception: ',E),Outcome=error
    ; Exp=pass, Status=no ->
        Outcome=fail, add_error(ltl, 'Expected correct formula, but found counter-example')
    ; Exp=fail, Status=ok ->
        Outcome=fail, add_error(ltl, 'Expected counter-example, but model checker found none')
    ; otherwise ->
        add_error(ltl,'Unexpected result: ',(Exp,Status)),Outcome=fail),
    ltl_check_assertions2(Nrest,Srest,Frest,Erest,Limit,Orest).

extract_expectations([],[]).
extract_expectations([N|Nrest],[E|Erest]) :-
    name(N,Codes),
    ( append(_,"FAIL",Codes) -> E=fail; E=pass),
    extract_expectations(Nrest,Erest).

merge_outcomes(List,Result) :-
    merge_outcomes2(List,no_tests,Result).
merge_outcomes2([],R,R).
merge_outcomes2([O|Rest],P,R) :-
    merge_outcomes3(P,O,I),
    merge_outcomes2(Rest,I,R).
merge_outcomes3(no_tests,X,X) :- !.
merge_outcomes3(_,error,error) :- !.
merge_outcomes3(P,fail,fail) :- P\=error,!.
merge_outcomes3(P,pass,P).

%*******************************************************************************

% ltl_model_check(FormulaAsAtom,Max,Mode,Status):
%  FormulaAsAtom: The LTL formula as an atom
%  Max: The maximum number of states that should be newly calculated
%  Mode: Where to start the search in state space
%     starthere: start the evaluation of the formula in the current state
%     init: start in the (possible several) initialisation states of the model
%     checkhere: start the search as in init, but check the formula F in the
%        current state by constructing a new formula (current -> F)
%  Status:
%     no: the model does not fulfil the formula, a counter-example is shown by
%         modifying the history
%     ok: the model does fulfil the formula
%     incomplete: the search was stopped (usually because the limit Max was reached)
%         before finding a counter-example or guarantee the absent of a counter-example
ltl_model_check(FormulaAsAtom,Max,Mode,Status) :-
    print_message(parsing_ltl_formula),
    retractall(counterexample_node(_)),
    retractall(counterexample_op(_)),
    ( temporal_parser(FormulaAsAtom,ltl,Ltl) ->
        true
    ; otherwise ->
        add_error(ltl,'LTL Parser failed: ',FormulaAsAtom),fail),
    ltl_model_check2(Ltl,Max,Mode,Status).

ltl_model_check2(Ltl,Max,Mode,Status) :-
    print_message(initialising_ltlc(Max,Mode)),
    ( Ltl=syntax_error ->
        Status=syntax_error
    ; otherwise ->
        ltl_model_check3(Ltl,Max,Mode,Result),
        convert_result(Result,Mode,Status)).

ltl_model_check3(Ltl,Max,Mode,Result) :-
    perform_static_analyses_if_necessary(Ltl),
    set_max_nr_of_new_impl_trans_nodes(Max),
    % initialising all given fairness constraints
    initialise_fairness_assumptions(Ltl,WeakFairnessAssumption,StrongFairnessAssumption,LtlFormula),
    debug_println(9,fairness_constraints(weak(WeakFairnessAssumption),strong(StrongFairnessAssumption))),
    select_ltl_mode(Mode,LtlFormula,Startnodes,Ltl2),
    preprocess_formula(Ltl2,Ltl3),
    debug_println(9,preprocess_formula(LtlFormula,Ltl3)),
    ltl_model_check4(Ltl3,Startnodes,WeakFairnessAssumption,StrongFairnessAssumption,Result),
    % sanity check for the result
    call_ltl_verification(Result,Ltl3),
    call_ltl_fairness_verification(Result,WeakFairnessAssumption,StrongFairnessAssumption),
    debug_println(9,ltl_result(Result)).

% deciding whether to run the safety model checker or the ordinary LTL model checker
ltl_model_check4(Ltl,Startnodes,WeakFairnessAssumption,StrongFairnessAssumption,Result) :-
    ((is_safety_property(Ltl,NLtl),is_ltl2ba_tool_installed) % checkig whether the property is a safety LTL formula
      -> debug_println(9,'Formula is a safety property'),
         debug_println(9,normalized(NLtl)),
         ltl_mc_safety_property(Ltl,Startnodes,Result)
    ;    call_c_initialisation,
         ltl_model_check5(Startnodes,Ltl,WeakFairnessAssumption,StrongFairnessAssumption,Result)
    ).

ltl_model_check5([],_Ltl,_WeakFairnessAssumption,_StrongFairnessAssumption,nostart) :- !.
ltl_model_check5(Startnodes,Ltl,WeakFairnessAssumption,StrongFairnessAssumption,Result) :-
    get_preference(por,ample_sets),!,
    % explore state space without performing any checks
    ((get_preference(pge,Pref), Pref\=off) -> PGE=1; PGE=0),
    depth_breadth_first_mode(DFMODE),
    set_depth_breadth_first_mode(depth_first), % set to depth-first mode
    print_message('******** starting a static LTL model check with partial order reduction ********'),
    statistics(runtime,[T1,_]),
    explore_state_space(0,-1,NodesAnalysed,_LimitTime,Res),
    statistics(runtime,[T2,_]),
    D is T2-T1,
    debug_println(9,statspace_stats(NodesAnalysed,Res)),
    format('state space exploration needed ~w ms to explore ~w states\n',[D,NodesAnalysed]),
    print_message('******** state space explored using partial order reduction ********'),
    set_depth_breadth_first_mode(DFMODE), % reset to previous mode
    print_message('******** start LTL model checking on the explored state space ********'),
    ltl_model_check_core(Ltl,Startnodes,WeakFairnessAssumption,StrongFairnessAssumption,Result).
ltl_model_check5(Startnodes,Ltl,WeakFairnessAssumption,StrongFairnessAssumption,Result) :-
     ltl_model_check_core(Ltl,Startnodes,WeakFairnessAssumption,StrongFairnessAssumption,Result).

ltl_model_check_core(Ltl,Startnodes,WeakFairnessAssumption,StrongFairnessAssumption,Result) :-
    % checking first whether some fairness constraints have been set up
    (fairness_check_on(WeakFairnessAssumption,StrongFairnessAssumption) ->
      c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback, none,
               ltl_fairness:cltl_get_transition_ids_callback, ltl_fairness:cltl_get_enabled_actions_callback, WeakFairnessAssumption,StrongFairnessAssumption)
    ; c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback) % no fairness assumptions have been imposed
    ).

call_c_initialisation :- ltlc_init,!.
call_c_initialisation :-
    add_error(ltl,'Initialisation of LTL-C module failed'),
    fail.

select_ltl_mode(starthere,Ltl,[Startnode],not(Ltl)) :-
    !,current_state_id(Startnode).
select_ltl_mode(init,Ltl,Startnodes,not(Ltl)) :-
    !,init_states(Startnodes).
select_ltl_mode(checkhere,Ltl,Startnodes,not(globally(implies(ap(current),Ltl)))) :-
    !,init_states(Startnodes).
select_ltl_mode(specific_node(Startnode),Ltl,[Startnode],not(Ltl)) :- !.
select_ltl_mode(Mode,Ltl,[],Ltl) :-
    add_error(ltl,'Internal error: Unexpected mode: ', Mode),
    fail.

%:- use_module(probsrc(model_checker),[model_checking_is_incomplete/6]).
convert_result(nostart,_,Result) :- !,Result=nostart.
convert_result(none,_,Result) :- !,  Result=ok.
  % TO DO: maybe return different result when model_checking_is_incomplete(0,1,0,0,_Msg,_Term) is true
convert_result(incomplete,_,Result) :- !,Result=incomplete.
convert_result(model(CE,_Entry),Mode,Result) :- !, Result=no,
    translate_ctrace(CE,Op,N,Dest),
    ( (Mode=init; Mode=checkhere) ->
        ce_get_first_state(CE,UsedInitState),
        set_trace_to_init(UsedInitState)
    ; Mode=specific_node(InitState) ->
        set_trace_to_init(InitState)
    ; otherwise ->
        true),
    execute_id_trace_from_current(Dest,Op,N).
convert_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result: ', Result),
    fail.

ce_get_first_state([atom(State,_,_)|_],State).

:- dynamic counterexample_node/1.
:- dynamic counterexample_op/1.

reset_ltl :-
   retractall(counterexample_node(_)),
   retractall(counterexample_op(_)),
   retractall(ltl_formula_cache(_,_)),
   retractall(ltl_counter_examples_cache(_,_,_,_)).

:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- register_event_listener(reset_specification,reset_ltl,
                           'Reset LTL Model Checker.').

translate_ctrace(Atoms,OpIds,StateIds,Dest) :-
    translate_ctrace2(Atoms,beginning,OpIds,StateIds,Dest).
translate_ctrace2([atom(StartNode,_,OpId)|RestAtoms],_PreviousState,Ops,Nodes,Dest) :-
    assert(counterexample_node(StartNode)),
    assert(counterexample_op(OpId)),
    ( OpId = none -> Ops=[],Nodes=[],Dest=StartNode
    ; otherwise ->
        transition(StartNode,_Op,OpId,NextNode),
        Ops=[OpId|OpRest], Nodes=[NextNode|NextRest],
        translate_ctrace2(RestAtoms,NextNode,OpRest,NextRest,Dest)).
translate_ctrace2([],Dest,[],[],Dest).

% replacing unparsed/1 by atomic propositions
% DEAD CODE
%% unparsed_ap(ap(P),ap(P)) --> !.
%% unparsed_ap(action(unparsed_b(OpName/Arity/Types,Exprs)),action(b(OpName/Arity,Es))) --> !,
%%     add_unparsed_exprs(Exprs,Types,Es).
%% unparsed_ap(action(b(OpName/Arity/_Types)),action(b(OpName/Arity))) --> !.
%% unparsed_ap(action(Op),action(Op)) --> !.
%% unparsed_ap(unparsed(Unparsed),ap(P)) --> !,[pred(Unparsed,P)].
%% unparsed_ap(F,N) --> {F =..[Functor|Args]},unparsed_ap_l(Args,NArgs),{N=..[Functor|NArgs]}.
%% unparsed_ap_l([],[]) --> [].
%% unparsed_ap_l([F|Rest],[N|NRest]) --> unparsed_ap(F,N), unparsed_ap_l(Rest,NRest).

%% add_unparsed_exprs([],[],[]) --> [].
%% add_unparsed_exprs([expr(P,U)|UR],[T|TR],[expr(P,E)|ER]) -->
%%     [expr(U,T,E)], add_unparsed_exprs(UR,TR,ER).

%*******************************************************************************
% set current trace

executetrace(Trace) :-
    translate_trace(Trace,OpIds,Ids,Dest),
    debug_println(9,exec(Dest,OpIds,Ids)),
    execute_id_trace_from_current(Dest,OpIds,Ids),!.
executetrace(Trace) :- add_internal_error('Could not execute counter-example: ',executetrace(Trace)).

translate_trace([Dest],[],[],Dest) :-
    assert(counterexample_node(Dest)).
translate_trace([Start,Id|RestNodes],[OpId|OpRest],[Id|IdRest],Dest) :-
    !,transition(Start,_Op,OpId,Id),
    assert(counterexample_node(Start)),
    assert(counterexample_op(OpId)),
    translate_trace([Id|RestNodes],OpRest,IdRest,Dest).

%*******************************************************************************
% find initialised states
% TODO: does NOT work with XTL (CSP?) and constants

init_states(Init) :-
    animation_mode(Mode),
    ( is_registered_plugin(Mode) ->
        findall(I,find_plugin_inits(root,Mode,_,I),Init)
    ; mode_cst_init(Mode) ->
        findall(I,find_init(root,I,_),Init)
    ; mode_one_step(Mode) ->
        next_states(root,Init)
    ; otherwise ->
        fail).

mode_cst_init(b).
mode_cst_init(z).
mode_cst_init(csp_and_b).

mode_one_step(csp).
mode_one_step(cspm).
mode_one_step(xtl).
mode_one_step(promela).

next_states(Start,States) :-
    impl_trans_term_all(Start,Ops),
    findall(S, member(op(_Id,_,S),Ops), States).

find_init(Start,Init,Trace) :- Start==Init,!,Trace=[]. % usually called with Start=Init=root
find_init(Start,Init,[State|Rest]) :-
    impl_trans_term(Start,O,State),
    find_init2(O,State,Init,Rest).
find_init2(O,Init,Init,[]) :-
    has_functor_and_maybe_tau(O,'$initialise_machine').
find_init2(O,State,Init,Path) :-
    has_functor_and_maybe_tau(O,'$setup_constants'),
    find_init(State,Init,Path).
find_init2(start_cspm_MAIN,State,Init,Path) :-
    find_init(State,Init,Path).
find_init2(start_cspm(_Proc),State,Init,Path) :-
    find_init(State,Init,Path).

% has_functor_and_maybe_tau(Term,Functor)
% checks if Term has the form "Functor(...)" or "tau(Functor(...))"
% this is used for CSP||B specification where the initialisation is wrapped with
% in a tau operator
has_functor_and_maybe_tau(tau(Term),Functor) :-
    has_functor_and_maybe_tau(Term,Functor),!.
has_functor_and_maybe_tau(Term,Functor) :-
    functor(Term,Functor,_).


find_plugin_inits(Start,Id,[Start|Rest],StateId) :-
    visited_expression(Start,State),
    ( plugin_is_initialised_state(Id,State) ->
        StateId = Start,
        Rest = []
    ; otherwise ->
        impl_trans_term(Start,_,Next),
        find_plugin_inits(Next,Id,Rest,StateId)).

set_trace_to_init(Initstate) :-
    animation_mode(Mode),
    find_trace_to_init(Mode,Initstate,Trace),!,
    state_space_reset,
    executetrace(Trace).
set_trace_to_init(Initstate) :- add_internal_error('Could not replay counter-example: ',set_trace_to_init(Initstate)).

find_trace_to_init(Mode,Target,[root,Target]) :-
    mode_one_step(Mode).
find_trace_to_init(Mode,Target,[root|Trace]) :-
    mode_cst_init(Mode),
    find_init(root,Target,Trace).
find_trace_to_init(Mode,Target,Trace) :-
    is_registered_plugin(Mode),
    find_plugin_inits(root,Mode,Trace,Target).

perform_static_analyses_if_necessary(Ltl) :-
  % currently only for B and Event-B
  (animation_mode(b) -> perform_static_analyses_if_necessary2(Ltl); true).

perform_static_analyses_if_necessary2(Ltl) :-
  get_preference(por,ample_sets),
  prepare_ltl_model_checker_for_por(Ltl),
  fail.
perform_static_analyses_if_necessary2(_Ltl) :-
  is_pge_opt_on,
  pge_algo: compute_operation_relations_for_pge(0), % 0 for not including the invariant into the analysis
  fail.
perform_static_analyses_if_necessary2(_Ltl).

%%% Predicates for determining which actions are visible and which are stutter in regard to the given LTL formula.
%%% An operation Op in a B machine should be identified as visible for a given LTL formula f, if Op writes variables used
%%% in the predicates in f or writes variables read in guard of operations inside the e(-) atoms in f.
%%% In case the respective LTL formula f contains a X (next) operator then M_{POR} |= f and M |= f are not equivalent.
%%% It is not clear how to handle LTL formulae containing transition propositions ([-]). (investigate e.g. LTL formulae of the form: ([Op] => g))
prepare_ltl_model_checker_for_por(Ltl) :-
    debug_println(9,prepare_ltl_model_checker_for_por(Ltl)),
    % state space should be reseted every time when a new LTL_X formula is checked,
    % because of the stutter condition
    state_space_initialise,
    reset_runtime_dynamic_predicates_POR,
    get_stutter_visible_actions(Ltl,StutterActions,VisibleActions,NextOrTP),
    (NextOrTP -> pp_ltl_formula(Ltl,LTLString),add_warning(ltl_mc_por, 'LTL formula contains X or Y operator, or [-] connstruct: ', LTLString); true),
    debug_println(9,get_stutter_visible_actions(Ltl,stutter(StutterActions),visible(VisibleActions),next_tps(NextOrTP))),
    assert_all_visible_stutter_actions(StutterActions,VisibleActions),
    % if the model is not reloaded _EnableGraph will not be computed every time a new LTL formula is checked
    get_por_preferences(POROptions),
    assert(check_cycle_proviso),
    compute_dependendency_relation_of_all_events_in_the_model(0,POROptions,_EnableGraph).

% get_stutter_visible_actions(+Ltl,-StutterActions,-VisibleActions)
% Ltl: the parsed Ltl formula presented as Prolog term.
% StutterActions: the set of stutter actions in regard to Ltl
% VisibleActions: the set of visible actions in regard to Ltl
get_stutter_visible_actions(Ltl,StutterActions,VisibleActions,NextOrTP) :-
    findall(Op,b_top_level_operation(Op),Ops),
    list_to_ord_set(Ops,Operations),
    look_up_for_visible_actions(Ltl,Operations,VisActions,NextOrTP),
    list_to_ord_set(VisActions,VisibleActions),
    ord_subtract(Operations,VisibleActions,StutterActions).

look_up_for_visible_actions(Ltl,Operations,VisActions,NextOrTP) :-
    look_up_for_visible_actions1(Ltl,Operations,VisActions,NextOrTPs),
    (NextOrTPs = [] -> NextOrTP=fail;NextOrTP=true).

look_up_for_visible_actions1(ap(bpred(Pred)),Ops,Acts,[]) :- !,
    predicate_identifiers(Pred,Ids),
    list_to_ord_set(Ids,Vars),
    get_actions_modifying_variables(Vars,Ops,Acts).
look_up_for_visible_actions1(action(bop(Name,_NArgs,_NRes,_Args,_Outputs)),_Ops,[Name],[true]) :- !.
    %% in case of [evt], 'evt' will be recognised as visible 
% All actions that can enable or disable an action checked for enabledness should be
% asserted as visible. This means that every action that writes the variables of action
% "a" which is checked for enabledness (e(a)) in the LTL-formula.
look_up_for_visible_actions1(enabled(bop(Name,_NArgs,_NRes,_Args,_Outputs)),Ops,Acts,[]) :- !,
    b_get_operation_read_guard_vars(Name,true,GuardReads),
    list_to_ord_set(GuardReads,Vars),
    get_actions_modifying_variables(Vars,Ops,Acts).
look_up_for_visible_actions1(next(F),Ops,Acts,[true|NextOrTPs]) :- !,
    look_up_for_visible_actions1(F,Ops,Acts,NextOrTPs).
look_up_for_visible_actions1(yesterday(F),Ops,Acts,[true|NextOrTPs]) :- !,
    look_up_for_visible_actions1(F,Ops,Acts,NextOrTPs).
look_up_for_visible_actions1(Term,_Ops,[],[]) :- % atomic can be deadlock or sink
    atomic(Term),!. % can Term be a variable???
% the recursive clause
look_up_for_visible_actions1(Term,Ops,Actions,NextOrTPs) :-
    compound(Term),!,
    functor(Term,_Functor,N),
    look_up_for_visible_actions_args(N,Term,Ops,Acts,NOrTPs),
    append(Acts,Actions),append(NOrTPs,NextOrTPs).

look_up_for_visible_actions_args(0,_,_,[],[]) :- !.
look_up_for_visible_actions_args(N,Term,Ops,Actions,NextOrTPs) :-
    N>0,!,
    arg(N,Term,Arg),
    look_up_for_visible_actions1(Arg,Ops,Acts,NTPs),
    Actions = [Acts|ActsRest], NextOrTPs = [NTPs|NTPsRest],
    N1 is N-1,
    look_up_for_visible_actions_args(N1,Term,Ops,ActsRest,NTPsRest).

get_actions_modifying_variables(Vars,Ops,Actions) :-
    maplist(variable_modified_by_action(Vars),Ops,Acts),
    append(Acts,Actions).

variable_modified_by_action(Vars,Op,Res) :-
    b_get_read_write(Op,_Read,Write),
    (ord_intersect(Vars,Write)
        -> Res=[Op]
        ;  Res=[]
    ).

% asserting all stutter and visible actions before starting the LTL model checker
assert_all_visible_stutter_actions(StutterActions,VisibleActions) :-
    maplist(assert_stutter_action,StutterActions),
    maplist(assert_visible_action,VisibleActions).

assert_stutter_action(Act) :-
    assert(stutter_action(Act)).

assert_visible_action(Act) :-
    assert(visible_action(Act)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% callback from LTLc:

% cltl_callback(+Node, -Aps, -Out, +Queries, +Specs):
%  Callback predicate that evaluates for a state "Node" all atomic
%  propositions and returns a list of outgoing transitions.
%  It evaluates if each of the transitions met the propositions on transistions.
%  "Queries" is a list of atomic propositions (ap), "Aps" is a lists where
%  each entry is 0 resp 1 if the corresponding atomic proposition is true resp. false.
%  "Specs" is a list of propositions on transitions (tp).
%  "Out" is a list of outgoing transitions, for details, see below check_tps/4.
:- public cltl_callback/5.
cltl_callback(Node, Aps, Out, Queries, Specs) :-
    check_aps(Queries,Node,Aps),      % generate a list of 0s and 1s, depending on whether the AP is met or not
    impl_trans_term_all1(Node,Trans), % find all outgoing transitions
    check_tps(Trans,Specs,Node,Out),  % evaluate the TP on each transition and generate a list of transitions, each
                                      % annotated with 0s and 1s, depending on whether the TP is met or not
    debug_println(9,cltl_callback(Node, Aps, Out, Queries, Specs)).

impl_trans_term_all1(Node,Trans) :-
  % In case that sink/deadlock is checked or the e(-) predicate is used in the LTL formula,
  % the state will be completely explored by execution of the check_aps/3 predicate (see implementation of trans/3).
  ( (animation_mode(b), is_pge_opt_on, not_all_transitions_added(Node))
    ->  set_context_state(Node),
        visited_expression(Node,CurState),
        pge_algo: compute_transitions_pge(Node,CurState),
        findall(op(Id,ActionAsTerm,To),
                transition(Node,ActionAsTerm,Id,To),
                Trans)
    ;  impl_trans_term_all(Node,Trans)
  ).

:- public cltl_callback_reexplore/5.
% predicate used only in case POR is enabled 
% to satisfy the cycle condition
% at this moment of the call the node has already been explored
% and we need to apply the rest of the enabled events at node
% to force the full exploration of it
cltl_callback_reexplore(Node,Aps,Out,Queries,Specs) :-
    check_aps(Queries,Node,Aps),
    impl_trans_term_rest(Node,Trans),
    check_tps(Trans,Specs,Node,Out),
    debug_println(8,cltl_callback_reexplore(Node,Aps,Out,Queries,Specs)).

impl_trans_term_rest(Node,Trans) :-
   get_all_already_explored_transitions(Node,Ops_so_far),
   state_space: compute_transitions_if_necessary_saved(From),
   findall(op(Id,ActionAsTerm,To),
     (transition(From,ActionAsTerm,Id,To),nonmember(op(Id,ActionAsTerm,To),Ops_so_far)),
     Trans).

get_all_already_explored_transitions(Node,Ops) :-
   findall(op(Id,ActionAsTerm,To),
     transition(Node,ActionAsTerm,Id,To),Ops).

get_por_preferences(por(TYPE,UseEnableGraph,Depth,UsePGE)) :-
  get_preference(por,TYPE),
  get_preference(enable_graph,UseEnableGraph),
  get_preference(enable_graph_depth,Depth),
  get_preference(pge,UsePGE).

get_por_preference(Pref) :-
  (animation_mode(b) -> get_preference(por,Pref); Pref=off).

% check list of atomic propositions (ap)
check_aps([],_,[]).
check_aps([Query|QRest],Node,[Result|RRest]) :-
    (check_ap(Query,Node) -> !,Result=1; Result=0),
    check_aps(QRest,Node,RRest).

% check list of transition properties:
% check_tps(+Transitions, +Specifications, +State, -Result):
%  "Transitions" is a list of terms of the form op(Id,Operation,NextState)
%          each term represents a transition from "State" to "NextState",
%          with the id "Id" and operation "Operation"
%  "Specifications" is a list of propositions on transistions (tp)
%  "Result" is a list where each element corresponds to a transition in "Transitions"
%          each entry has the form "trans(Next,Id,ResVector)" where "Next" and "Id"
%          are the destination state and the id of the transition, respectively, and
%          "ResVector" a list where each element corresponds to a proposition (tp)
%          in "Specifications". It is 0 resp. 1 if the proposition is false resp. true.
check_tps([],_,_,[]).
check_tps([op(Id,Op,Next)|OpRest],Specs,Node,[trans(Next,Id,Tp)|TpRest]) :-
    check_tps2(Specs,Node,Next,Op,Tp),
    check_tps(OpRest,Specs,Node,TpRest).
check_tps2([],_,_,_,[]).
check_tps2([Spec|SRest],SrcNode,DstNode,Op,[Result|RRest]) :-
    (check_transition(Spec,Op,SrcNode,DstNode) -> !,Result=1; Result=0),
    check_tps2(SRest,SrcNode,DstNode,Op,RRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

ltl_formula_structure(Formula,Atomics,Structure) :-
    c_ltl_aptp(Formula,ApTp),
    pp_formulas(ApTp,Atomics),
    ltl_structure(Formula,ApTp,Structure).

ltl_structure(fairnessimplication(_FAssumptions,F),Subformulas,Structure) :- !,
    ltl_structure(F,Subformulas,Structure).
ltl_structure(ap(AP),Subformulas,ap(N)) :-
    !,nth0(N,Subformulas,ap(AP)),!.
ltl_structure(action(TP),Subformulas,tp(N)) :-
    !,nth0(N,Subformulas,action(TP)),!.
ltl_structure(F,ApTp,N) :-
    F =.. [Functor|Args],
    same_length(Args,NewArgs),
    N =.. [Functor|NewArgs],
    ltl_structure_l(Args,ApTp,NewArgs).
ltl_structure_l([],_,[]).
ltl_structure_l([F|Frest],ApTp,[N|Nrest]) :-
    ltl_structure(F,ApTp,N),
    ltl_structure_l(Frest,ApTp,Nrest).

ltl_model_check_with_ce(Formula,Max,Mode,Result) :-
    ltl_model_check3(Formula,Max,Mode,R1),
    ce_result(R1,Formula,Result).

ce_result(nostart,_,nostart) :- !.
ce_result(none,_,ok) :- !.
ce_result(incomplete,_,incomplete) :- !.
ce_result(model(CE,LoopEntry),Formula,counterexample(Counterexample,LoopEntry,PathToCE)) :- !,
    ce_result_initpath(CE,PathToCE),
    c_ltl_aptp(Formula,ApTp),
    length(ApTp,N),
    convert_counterexample(CE,N,Counterexample).
ce_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result: ', Result),
    fail.

ce_result_initpath(CE,Path) :-
    ce_get_first_state(CE,InitState),
    find_init(root,InitState,StatePath),
    find_operation_ids(StatePath,root,Path).
find_operation_ids([],_,[]).
find_operation_ids([Dst|Srest],Src,[(Id,Op,Dst)|Irest]) :-
    transition(Src,Op,Id,Dst),
    find_operation_ids(Srest,Dst,Irest).

/* just restrict the evaluation bits to the atomic formulas */
convert_counterexample([],_N,[]).
convert_counterexample([atom(State,Evals,NextOp) |Arest],N,
                       [atom(State,NEvals,OpTuple)|Erest]) :-
    prefix_length(Evals,NEvals,N),
    ( NextOp = none          -> OpTuple = none
    ; otherwise              -> transition(State,Action,NextOp,Dst),OpTuple=(NextOp,Action,Dst)),
    convert_counterexample(Arest,N,Erest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% dynamic predicate for saving the LTL counterexamples for the Tcl/Tk interface
:- dynamic ltl_counter_examples_cache/4.

ltl_model_check_with_ce1(Proc,FormulaAsAtom,Max,Mode,Result) :-
    print_message(parsing_ltl_formula),
    ( temporal_parser(FormulaAsAtom,ltl,Ltl) ->
        true
    ; otherwise ->
        add_error_fail(ltl,'LTL Parser failed: ',FormulaAsAtom)),
    ltl_model_check_with_ce2(Proc,Ltl,Max,Mode,Result).

ltl_model_check_with_ce2(Proc,Ltl,Max,Mode,ActionTrace) :-
    print_message(initialising_ltlc2(Max,Mode)),
    ( Ltl=syntax_error ->
        Result=syntax_error
    ; otherwise ->
        interruptable_ltl_model_check3(Ltl,Max,Mode,Result),
        convert_counterexample_to_action_trace(Proc,Result,Mode,Ltl,ActionTrace)
    ).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).
interruptable_ltl_model_check3(Ltl,Max,Mode,Result) :-
  user_interruptable_call_det(user:catch_interrupt_assertion_call(ltl: ltl_model_check3(Ltl,Max,Mode,Result)),InterruptResult),
  (InterruptResult=interrupted ->
    Result=interrupted
  ; (real_error_occurred ->
      print_error('% *** Errors occurred while LTL model checking ! ***'),nl,nl,fail
    ; print('LTL assertion check completed.'),nl)).

convert_counterexample_to_action_trace(Proc,Result,Mode,Ltl,ActionTrace) :-
  ( Result = model(CE,LoopEntry) ->
    pp_ltl_counterexample(LoopEntry,CE,ActionTrace),
    (ltl_counter_examples_cache(Proc,Ltl,Mode,_) -> true; assert(ltl_counter_examples_cache(Proc,Ltl,Mode,Result)))
  ; Result = interrupted ->
    ActionTrace = interrupted
  ; otherwise ->
    convert_result(Result,Mode,ActionTrace)).

pp_ltl_counterexample(deadlock,Counterexample,ce(Trace)) :- !,
  length(Counterexample,N),
  (N >= 2 ->
    maplist(get_operation_name,Counterexample,Trace1),
    ajoin_with_sep(Trace1,' -> ',Trace)
    ; otherwise ->
    Trace = deadlock).
pp_ltl_counterexample(no_loop,Counterexample,ce(Trace)) :- !,
  maplist(get_operation_name,Counterexample,Trace1),
  ajoin_with_sep(Trace1,' -> ',Trace).
pp_ltl_counterexample(loop(N),Counterexample,ce(Trace)) :- !,
  prefix_length(Counterexample,PathToCE,N),
  maplist(get_operation_name,PathToCE,PrefixTrace),
  ajoin_with_sep(PrefixTrace,' -> ',PrefixTrace1),
  append(PathToCE,CEPath,Counterexample),
  maplist(get_operation_name,CEPath,SuffixTrace),
  ajoin_with_sep(SuffixTrace,' -> ',SuffixTrace1),
  ajoin([PrefixTrace1,' -> (',SuffixTrace1,')+'],Trace).
pp_ltl_counterexample(LoopEntry,_Counterexample,_CE) :-
  add_error_fail(ltl, 'Internal error: unknown loop entry type: ', LoopEntry).

get_operation_name(atom(State,_NEvasl,OpId),Res) :-
  (OpId = none
    -> Res = '|'
  ; transition(State,Op,OpId,_NextState) ->
       translate_event(Op,Res)
  ; otherwise
    ->  add_error_fail(ltl_assertions_result, 'Internal Error: unknown operation in CE trace ', atom(State,_NEvasl,OpId))).

reset_ce_marked_nodes :-
   retractall(counterexample_node(_)),
   retractall(counterexample_op(_)).

:- register_event_listener(play_counterexample,reset_ce_marked_nodes,
                           'Reset marked nodes from previous counterexamples.').

tcltk_play_ltl_counterexample(Proc,Ltl) :-
   ltl_counter_examples_cache(Proc,Ltl,Mode,model(CE,_Entry)),
   translate_ctrace(CE,Op,N,Dest),
    ( (Mode = init; Mode = checkhere) ->
      ce_get_first_state(CE,UsedInitState),
      set_trace_to_init(UsedInitState)
    ; Mode = specific_node(NodeID) ->
      set_trace_to_init(NodeID)
    ; otherwise ->
      true
    ),execute_id_trace_from_current(Dest,Op,N).
tcltk_play_ltl_counterexample(Proc,Ltl) :-
  add_error_fail(ltl_counter_example, 'Internal error: No counterexample has been asserted for the following configuration: ',(Proc,Ltl)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% verify a counter-example

callback_tp(TP,StateId,TransId) :-
    ( transition(StateId,Op,TransId,DestId) ->
        check_transition(TP,Op,StateId,DestId)
    ; otherwise ->
        add_error(ltl,'Internal error: callback_tp called for unknown transition: ',
                  state_trans(StateId,TransId)),
        fail
    ).

call_ltl_verification(model(Atoms,Status),Formula) :-
    !,call_ltl_verification2(Status,Atoms,Formula).
call_ltl_verification(_,_Formula).
call_ltl_verification2(Status,Atoms,Formula) :-
    maplist(convert_atom_to_statetransition,Atoms,StateTransitions),
    ( evaluate_ltl_formula(Formula,StateTransitions,Status,
                           check_ap,callback_tp,Result) ->
        ( Result = true -> true
        ; otherwise ->
            add_error(ltl,'Evaluation of LTL yielded unexpected result: ',Result))
    ; otherwise ->
        add_error(ltl,'Evaluation of LTL formula failed: ',Formula)).

convert_atom_to_statetransition(atom(NodeId,_,TransId),strans(NodeId,TransId)).


% fairness
is_fairness_implication(fairnessimplication(_FAssumptions,_F)).

call_ltl_fairness_verification(model(Atoms,Status),WeakDNF,StrongDNF) :-
    !,maplist(convert_atom_to_statetransition,Atoms,ST),
    call_ltl_fairness_verification1(WeakDNF,weak,ST,Status),
    call_ltl_fairness_verification1(StrongDNF,strong,ST,Status).
call_ltl_fairness_verification(_,_WeakDNF,_StrongDNF).

call_ltl_fairness_verification1([],_Type,_ST,_Status) :- !.
call_ltl_fairness_verification1(all,Type,ST,Status) :- !,
    get_fairness_specification(Status,ST,Assumption),
    call_ltl_fairness_verification1(Assumption,Type,ST,Status).
call_ltl_fairness_verification1(DNF,Type,ST,Status) :-
    ( call_ltl_fairness_verification2(DNF,Type,ST,Status,Result) ->
        true
    ; otherwise ->
        add_error(ltl,'Evalutation of LTL fairness failed: ',Type/DNF)),
    ( Result = fair -> true
    ; otherwise ->
        print(Type/DNF),nl,
        add_error(ltl,'Evaluation of LTL yields that the result violates fairness: ',Type/Result)).

call_ltl_fairness_verification2([],_Type,_ST,_Status,unfair([])).
call_ltl_fairness_verification2([Clause|CRest],Type,ST,Status,Result) :-
    call_ltl_fairness_verification3(Clause,Type,ST,Status,CResult),
    ( CResult = fair -> Result=fair
    ; CResult = unfair(Cause) ->
        call_ltl_fairness_verification2(CRest,Type,ST,Status,RResult),
        ( RResult = fair           -> Result=fair
        ; RResult = unfair(Causes) -> Result=unfair([Cause|Causes]))).
call_ltl_fairness_verification3([],_Type,_ST,_Status,fair).
call_ltl_fairness_verification3([Spec|SRest],Type,ST,Status,Result) :-
    evaluate_ltl_fairness(Type,ST,Status,ev_fairness_id(Spec),ev_fairness_id(Spec),LResult),
    ( LResult = fair ->
        call_ltl_fairness_verification3(SRest,Type,ST,Status,Result)
    ; LResult = unfair(Id) ->
        Result = unfair(Id)
    ; otherwise ->
        add_error_fail(ltl,'Unknown result after fairness verification check: ', evaluate_ltl_fairness(Type,ST,Status,ev_fairness_id(Spec),LResult))).
