% (c) 2014-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(ltl_safety,[is_safety_property/2,ltl_mc_safety_property/3,
                      is_ltl2ba_tool_installed/0,
                      aut_transition/3, aut_finite_state/1, aut_initial_state/1]).

/* SICSTUS Libraries */
:- use_module(library(lists)).
:- use_module(library(codesio)).
:- use_module(library(terms)).
:- use_module(library(file_systems)).

/* ProB Libraries */
:- use_module(probsrc(tools),[ajoin/2,host_platform/1]).
:- use_module(probsrc(error_manager)).
%% :- use_module(probsrc(debug)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(preferences),[preference/2,set_preference/2,get_preference/2]).
:- use_module(probsrc(system_call),[system_call/5]).
:- use_module(probsrc(debug),[debug_println/2,debug_mode/1]).
:- use_module(probsrc(bsyntaxtree),[create_texpr/4,create_negation/2]).

:- use_module(extension('ltl2ba/ltl2ba')).

/* ProB LTL Model Checker Modules */
:- use_module(probltlsrc(safety_mc),[start_mc_safety_property/3]).
:- use_module(probltlsrc(ltl_translate),[pp_ap/2,pp_tp/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides predicates for checking whether an LTL formula is a safety property.').

:- dynamic ap_representation/2, tp_representation/2, tps_occurring/0.

:- dynamic aut_transition/3, aut_finite_state/1, aut_initial_state/1.

clear_automaton_facts :-
    retractall(aut_transition(_,_,_)),
    retractall(aut_initial_state(_)),
    retractall(aut_finite_state(_)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% Checking whether an LTL property represents a safety property %%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The predicates check whether the parsed LTL property represents a safety property
% by checking the syntactical structure of the formula.
% The idea is borrowed from the following theorem from the paper 
% "Safety, Liveness and Fairness in Temporal Logic" of A. Sistla: 
%% THEOREM: Every propositional formula is a safety formula and if f,g are safety formulae,
%% then so are f & g, f or g, X f, f W g, and G f, where 'W' is the weak-until operator.
%
% Note that the checked formula should be a positive formula. An LTL formula is said to
% be a positive if none of the temporal operators appears in the scope of 'not'.

% Testing various properties for being safety or liveness. Some of 
% the unit tests here represent examples from the paper mentioned above.
% Note that the first argument of is_safety_property/2 should be negated, as 
% this is done after negating the original property that should be model checked (see select_ltl_mode/4 in ltl.pl)
:- assert_must_succeed((F = ap(p),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(deadlock),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = weakuntil(not(ap(alloc)),ap(request)),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = and(weakuntil(not(ap(alloc)),ap(request)),
                                globally(implies(ap(alloc),next(weakuntil(not(ap(alloc)),ap(request)))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(implies(not(ap(request)),or(ap(alloc),or(next(ap(alloc)),next(next(ap(alloc))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(or(not(ap(p)),globally(not(ap(p))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),ap(p3)))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),until(ap(p3),ap(p4))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(until(ap(p1),until(ap(p2),until(ap(p3),until(ap(p4),until(ap(p5),until(ap(p6),ap(p7)))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = not(not(globally(deadlock))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = globally(or(ap(p),and(next(ap(q)),not(next(ap(q)))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_succeed((F = and(globally(or(ap(p),next(globally(ap(q))))),globally(or(ap(p),next(globally(not(ap(q))))))),
                        is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(implies(ap(request),finally(ap(alloc)))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(finally(ap(alloc))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = globally(until(ap(request),finally(ap(alloc)))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = implies(globally(finally(ap(send))),globally(finally(receive))),
                     is_safety_property_test(not(F),_NF))).
:- assert_must_fail((F = not(until(ap(p1),not(until(ap(p2),until(ap(p3),ap(p4)))))),
                        is_safety_property_test(not(F),_NF))).

is_safety_property_test(F,NF) :-
  get_preference(use_safety_ltl_model_checker,Val),
  set_preference(use_safety_ltl_model_checker,true),
  is_safety_property(F,NF),
  set_preference(use_safety_ltl_model_checker,Val).  

is_safety_property(not(F),NF) :-
   preference(use_safety_ltl_model_checker,true),
   normalize_to_release_pnf(F,NF),
   is_syntactically_safe(NF).

is_syntactically_safe(deadlock) :- !.
is_syntactically_safe(false) :- !.
is_syntactically_safe(true) :- !.
is_syntactically_safe(ap(_)) :- !.
is_syntactically_safe(action(_)) :- !.
is_syntactically_safe(and(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(or(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(next(F)) :- !,
   is_syntactically_safe(F).
is_syntactically_safe(weakuntil(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(release(F,G)) :- !,
   is_syntactically_safe(F),
   is_syntactically_safe(G).
is_syntactically_safe(not(F)) :- !,
   is_atomic_proposition(F).

is_atomic_proposition(F) :-
   F =.. [Functor|_Args],
   memberchk(Functor,[true,false,ap,action]).

% Release PNF (the Release PNF does not imply exponential blow-up of the origin LTL formula)
% f = true | false | a | not a | f1 & f2 | f1 or f2 | X f | f1 U f2 | f1 R f2 
normalize_to_release_pnf(deadlock,deadlock) :- !.
normalize_to_release_pnf(false,false) :- !.
normalize_to_release_pnf(true,true) :- !.
normalize_to_release_pnf(ap(P),ap(P)) :- !.
normalize_to_release_pnf(action(P),action(P)) :- !.
normalize_to_release_pnf(and(F,G),and(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(or(F,G),or(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(implies(F,G),NImplies) :- !,
   normalize_to_release_pnf(or(not(F),G),NImplies).
normalize_to_release_pnf(next(F),next(NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(globally(F),release(false,NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(finally(F),until(true,NF)) :- !,
   normalize_to_release_pnf(F,NF).
normalize_to_release_pnf(weakuntil(F,G),weakuntil(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(release(F,G),release(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
normalize_to_release_pnf(until(F,G),until(NF,NG)) :- !,
   normalize_to_release_pnf(F,NF),
   normalize_to_release_pnf(G,NG).
% not rules
normalize_to_release_pnf(not(true),false) :- !.
normalize_to_release_pnf(not(false),true) :- !.
normalize_to_release_pnf(not(and(F,G)),NOr) :- !,
   normalize_to_release_pnf(or(not(F),not(G)),NOr).
normalize_to_release_pnf(not(or(F,G)),NAnd) :- !,
   normalize_to_release_pnf(and(not(F),not(G)),NAnd).
normalize_to_release_pnf(not(next(F)),NNext) :- !,
   normalize_to_release_pnf(next(not(F)),NNext).
normalize_to_release_pnf(not(finally(F)),NGlobally) :- !,
   normalize_to_release_pnf(release(false,not(F)),NGlobally).
normalize_to_release_pnf(not(globally(F)),NUntil) :- !,
   normalize_to_release_pnf(until(true,not(F)),NUntil).
normalize_to_release_pnf(not(until(F,G)),NRelease) :- !,
   normalize_to_release_pnf(release(not(F),not(G)),NRelease).
normalize_to_release_pnf(not(F),NF) :- !,
   (F = not(F1) ->
     normalize_to_release_pnf(F1,NF)
   ; normalize_to_release_pnf(F,NF1),
     NF=not(NF1)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% Predicates for using the LTL2BA Tool %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ltl_mc_safety_property(Ltl,StartNodes,Result) :-
  translate_to_ltl2ba_syntax(Ltl,Ltl1),
  ltl2ba_init,
  clear_automaton_facts,
  (debug_mode(on) -> Verbose = true; Verbose = false),
  c_call_ltl2ba(Ltl1,ltl_safety:assert_aut_initial_state,ltl_safety:assert_aut_finite_state,ltl_safety:assert_aut_transition,Verbose,ExitResult),
  debug_print_ba_automaton,
  (ExitResult = ok ->
    load_and_check_safety_property(StartNodes,Result)
  ; ExitResult = error ->
    add_error_fail(translating_ltl2ba,'An error appeared while translating LTL to BA for the following formula: ',Ltl1)
  ; otherwise ->
    add_error_fail(translating_ltl2ba,'Unexpected error occurred while translating LTL to BA for the following formula: ',Ltl1)
  ).
ltl_mc_safety_property(_Ltl,_StarNodes,_Result) :- 
    add_error_fail(translating_ltl2ba,'Please ensure that ltl2ba is installed (see Help menu for ProB Tcl/Tk)','').

load_and_check_safety_property(StartNodes,Result) :-
  %% start_mc_safety_property(random,StartNodes,Result).
  start_mc_safety_property(breadth_first,StartNodes,Result).

is_ltl2ba_tool_installed :-
  host_platform(Platform),
  get_platform_specifix_ltl2ba_library(Platform,File),
  absolute_file_name(prob_lib(File),LTL2BALibrary),
  (file_exists(LTL2BALibrary) ->
    debug_println(9,'LTL2BA is installed.')
  ; otherwise ->
    add_warning(translating_ltl2ba,
      'ltl2ba-library archive could not be found in lib/ directory, please ensure that ltl2ba is installed'),
    fail % important to fail since we test here for the availability of ltl2ba
    ).

get_platform_specifix_ltl2ba_library(darwin,'ltl2ba.bundle'):- !.
get_platform_specifix_ltl2ba_library(linux,'ltl2ba.so') :- !.
get_platform_specifix_ltl2ba_library(windows,'ltl2ba.dll') :- !.
get_platform_specifix_ltl2ba_library(Unknown,_) :-
  add_error_fail(translating_ltl2ba,'Unknown platform: ', Unknown).

assert_aut_initial_state(StateDescription) :-
  debug_println(9,asserting_initialstate(StateDescription)),
  assert(aut_initial_state(StateDescription)).
assert_aut_finite_state(StateDescription) :-
  debug_println(9,asserting_finitestate(StateDescription)),
  assert(aut_finite_state(StateDescription)).
assert_aut_transition(Src,Atom,Dst) :-
  atom_codes(Atom,Codes),
  read_term_from_codes_exception(Codes,PrepRep,[]),
  redefine_transition_label(PrepRep,Label),
  debug_println(9,asserting_transition(aut_transition(Src,Label,Dst))),
  assert(aut_transition(Src,Label,Dst)).

read_term_from_codes_exception(Codes,PrepRep,Options) :-
  (read_term_from_codes(Codes,PrepRep,Options) 
    -> true
    ;  atom_codes(Atom,Codes),
       add_error(translating_ltl3ba_callback, 'Unexpected atom representation (does atom ends with fullstop?): ', Atom)
  ).

redefine_automaton_facts :-
  aut_transition(Src,P,Dst),
  redefine_transition(Src,P,Dst),
  fail.
redefine_automaton_facts.

redefine_transition(Src,Predicate,Dst) :-
  retract(aut_transition(Src,Predicate,Dst)),
  redefine_transition_label(Predicate,Preposition),
  assert(aut_transition(Src,Preposition,Dst)).

redefine_transition_label(true,Res) :- !,
  Res = true.
redefine_transition_label(false,Res) :- !,
  Res = false.
redefine_transition_label(not(AP),Res) :-
  ap_representation(AP,bpred(Pred)),!,
  create_negation(Pred,NegPred),
  create_atomic_b_predicate(NegPred,Res).
redefine_transition_label(AP,Res) :-
  ap_representation(AP,Rep),!,
  Res = ap(Rep).
redefine_transition_label(TP,Res) :-
  tp_representation(TP,Rep),!,
  Res = tp(Rep).
redefine_transition_label(F,Res) :-
  F =.. [Functor|Args],
  maplist(redefine_transition_label,Args,NewArgs),
  Res =.. [Functor|NewArgs].

create_atomic_b_predicate(BPred,Res) :-
  Res = ap(bpred(BPred)).

debug_print_ba_automaton :-
  (debug_mode(on)
    -> print_transitions_fail_loop
     ; true).

print_transitions_fail_loop :-
  aut_transition(Src,Label,Dst),
  print('Transition: '),print(Src),
  translate_and_print_label(Label),
  print(Dst),nl,fail.
print_transitions_fail_loop.

translate_and_print_label(Label) :-
  pp_ba_label(Label,Atom,[]),
  ajoin(Atom,Text),
  print_label(Text).

%% pretty-printing predicate in Spin syntax (created just to test if ltl2ba produces the correct automaton)
pp_ba_label(true,['true'|S],S) :- !.
pp_ba_label(false,['false'|S],S) :- !.
pp_ba_label(ap(AP),[Atom|S],S) :- !,
  pp_ap(AP,Atom).
pp_ba_label(tp(TP),[Atom|S],S) :- !,
  pp_tp(TP,Atom).
pp_ba_label(not(Pred),['! ('|S],T) :- !,
  pp_ba_label(Pred,S,[')'|T]).
pp_ba_label(and(Pred1,Pred2),['('|S],T) :- !,
  pp_ba_label(Pred1,S,[' && '| S1]),
  pp_ba_label(Pred2,S1,[')'|T]).
pp_ba_label(or(Pred1,Pred2),['('|S],T) :- !,
  pp_ba_label(Pred1,S,[' || '| S1]),
  pp_ba_label(Pred2,S1,[')'|T]).
pp_ba_label(Otherwise,['???'|S],S) :-
  add_error(printing_ba_label_for_ltl2ba,'Non-expected label: ',Otherwise).

print_label(Label) :-
  print(' --('),print(Label),print(')--> ').

translate_to_ltl2ba_syntax(Formula,Text) :-
  retractall(ap_representation(_,_)),
  retractall(tp_representation(_,_)),
  retractall(tps_occurring),
  pp_formula_to_ltl2ba_syntax(Formula,Text).

pp_formula_to_ltl2ba_syntax(Formula,Res) :-
  pp_to_ltl2ba_syntax(Formula,Atoms,[]),
  ajoin(Atoms,Res).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%% Unit Tests %%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed((F = and(or(true,false),true),
                        pp_formula_to_ltl2ba_syntax(F,Text),
                        Text = '((true || false) && true)')).
:- assert_must_succeed((F = implies(deadlock,not(true)),
                        pp_formula_to_ltl2ba_syntax(F,Text),
                        Text = '(deadlock -> ! true)')).
:- assert_must_succeed((F = not(until(true,next(implies(deadlock,not(true))))),
                        pp_formula_to_ltl2ba_syntax(F,Text),
                        Text = '! (true U X ((deadlock -> ! true)))')).
:- assert_must_succeed((F = weakuntil(true,false),
                        pp_formula_to_ltl2ba_syntax(F,Text),
                        Text = '((! true || false) V (true || false))')).
:- assert_must_succeed((F = globally(finally(true)),
                        pp_formula_to_ltl2ba_syntax(F,Text),
                        Text = '[] (<> (true))')).


pp_to_ltl2ba_syntax(true,['true'|S],S) :- !.
pp_to_ltl2ba_syntax(false,['false'|S],S) :- !.
pp_to_ltl2ba_syntax(deadlock,['deadlock'|S],S) :- !.
pp_to_ltl2ba_syntax(ap(X),[AP|S],S) :- !,
   term_hash(X,Hash),
   ajoin(['ap',Hash],AP),
   (\+ap_representation(AP,_) -> assert(ap_representation(AP,X)); true).
pp_to_ltl2ba_syntax(action(X),[TP|S],S) :- !,
   assert(tps_occurring),
   term_hash(X,Hash),
   ajoin(['tp',Hash],TP),
   (\+tp_representation(TP,_) -> assert(tp_representation(TP,X)); true).
pp_to_ltl2ba_syntax(not(X),['! '|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,T).
pp_to_ltl2ba_syntax(and(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' && '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(or(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' || '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(implies(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' -> '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(next(X),['X ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(globally(X),['[] ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(finally(X),['<> ('|S],T) :- !,
   pp_to_ltl2ba_syntax(X,S,[')'|T]).
pp_to_ltl2ba_syntax(until(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' U '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(release(X,Y),['('|S],T) :- !,
    pp_to_ltl2ba_syntax(X,S,[' V '|S1]),
    pp_to_ltl2ba_syntax(Y,S1,[')'|T]).
pp_to_ltl2ba_syntax(weakuntil(X,Y),S,T) :- !,
    pp_to_ltl2ba_syntax(release(or(not(X),Y),or(X,Y)),S,T).


%% Obsolete predicates (to be removed later)
load_and_check_safety_property(File,StartNodes,Result) :-
  safety_consult_without_redefine_warning(File),
  redefine_automaton_facts,
  %% start_mc_safety_property(breadth_first,StartNodes,Result).
  start_mc_safety_property(random,StartNodes,Result).

create_ba_pl_file(Text,Filename) :-
  write_string_to_file('prob_buchi_temp.pl',Filename,Text).

write_string_to_file(BaseName,Filename,Machine) :-
    open(temp(BaseName),write,S,[if_exists(generate_unique_name)]),
    stream_property(S, file_name(Filename)),
    print_string(Machine,S),
    close(S).

print_string([],_).
print_string([C|Rest],S) :- put_code(S,C),print_string(Rest,S).

safety_consult_without_redefine_warning(File) :-
    clear_automaton_facts,
    (file_exists(File) -> true ; add_error_fail(safety_mc,'Automaton PL File does not exist: ',File)),
    print(consult_without_redefine(File)),nl,
    prolog_flag(redefine_warnings, Old, off),
    prolog_flag(single_var_warnings, Old2, off), 
    prolog_flag(discontiguous_warnings, Old3, off), 
    (load_files(File,[compilation_mode(assert_all)]) %consult(File)
      -> OK=true ; OK=false),
    prolog_flag(redefine_warnings, _, Old),
    prolog_flag(single_var_warnings, _, Old2),
    prolog_flag(discontiguous_warnings, _, Old3),
    OK=true.
