% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(parsercall, [load_b_machine_as_term/2,
                       load_b_machine_probfile_as_term/2, % a way to directly load a .prob file
                       load_b_machine_list_of_facts_as_term/2,
                       get_parser_version/1,
                       get_java_command_path/1,
                       check_java_version/2,
                       get_java_fullversion/1, get_java_fullversion/3, get_java_version/1,
                       ensure_console_parser_launched/0, % just make sure the console parser is up and running
                       call_ltl_parser/3,
                       call_tla2b_parser/1, tla2b_filename/2,
                       call_promela_parser/1, promela_prolog_filename/2,
                       tla2prob_filename/2,
                       parse/3,
                       parse_formula/2, parse_predicate/2, parse_expression/2,
                       parse_substitution/2]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module takes care of calling the Java B Parser if necessary.').

:- use_module(library(lists)).
:- use_module(library(process)).
:- use_module(library(file_systems)).
:- use_module(library(codesio)).
:- use_module(library(system)).

:- use_module(error_manager,[add_error/2,add_error/3,add_error_fail/3,add_failed_call_error/1, add_internal_error/2, add_warning/3, real_error_occurred/0]).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(tools,[host_platform/1, split_filename/3, get_PROBPATH/1, safe_atom_chars/3, platform_is_64_bit/0, string_concatenate/3]).
:- use_module(tools_strings,[ajoin_with_sep/3, ajoin/2]).
:- use_module(tools_printing,[print_error/1]).
:- use_module(debug).
:- use_module(bmachine,[b_get_definition/5]).
:- use_module(specfile, [b_mode/0]).

:- dynamic java_parser_version/1.
:- dynamic parser_location/1.

% stores the java process of the parser, consisting of ProcessIdentifier, STDInStream, STDOutStream
:- dynamic java_parser_process/4.
:- volatile java_parser_process/4.

send_definition(ToParser,def(Name,Type,Arity)) :-
    format(ToParser,'definition\n',[]),
    format(ToParser,'~w\n~w\n~w\n',[Name,Type,Arity]).

send_definitions(ToParser) :-
    b_mode, !,
    findall(def(A,B,Ar),(b_get_definition(A,B,C,_,_),length(C,Ar)),L),
    maplist(send_definition(ToParser),L).
send_definitions(_).

cleanup(Codes,CleanCodes) :-
    exclude(crlf,Codes,CleanCodes).

crlf(10).
crlf(13).

send_extended_kind(Extended,Kind,In) :-
    (Extended == true
     -> ajoin([extended,Kind,'\n'],ToParser)
     ;  ajoin([Kind,'\n'], ToParser)),
    format(In,ToParser,[]).

:- mode parse_x(+Kind,+Extended,+Codes,-Tree).
parse_x(Kind,Extended,CodesWithNewlines,Tree) :-
    % Java parser expects only one line of input -> remove newlines
    cleanup(CodesWithNewlines,Codes),
    % statistics(walltime,[Tot,Delta]), format('~s~n~w~n0     ~w (Delta ~w) ms  ~n',[Codes,Kind,Tot,Delta]),
    get_console_parser(In,Out,Err),
    send_definitions(In),
    send_extended_kind(Extended,Kind,In),
    format(In,'~s\n',[Codes]),
    read_line(Out,CodesIn),
    %format('read: ~s~n',[CodesIn]),
    catch(
         (my_read_from_codes(CodesIn,Tree,Out,Err),handle_parser_exception(Tree)),
          error(_,_),
          (append("Illegal parser result exception: ",CodesIn,ErrMsgCodes),
           atom_codes(ErrMsg,ErrMsgCodes),
           throw(parse_errors([internal_error(ErrMsg,none)])))).
    %statistics(walltime,[Tot4,Delta4]), format('4     ~w (Delta ~w) ms ~n ~n',[Tot4,Delta4]).


my_read_from_codes(end_of_file,_,_,Err) :- !,
    read_line_if_ready(Err,Err1Codes),  % we just read one line
    safe_name(ErrMsg,Err1Codes),
    add_error(parsercall,'Abnormal termination of parser: ',ErrMsg),
    missing_parser_diagnostics,
    ajoin(['Parser not available (',ErrMsg,')'],FullErrMsg),
    read_lines_and_add_as_error(Err),  % add other info on error stream as single error
    throw(parse_errors([error(FullErrMsg,none)])).
my_read_from_codes(ErrorCodes,_Tree,Out,_) :-
    append("Error",_,ErrorCodes),!,
    atom_codes(ErrMsg,ErrorCodes),
    add_error(parsercall,'Error running/starting parser: ',ErrMsg),
    read_lines_and_add_as_error(Out),
    fail.
my_read_from_codes(CodesIn,Tree,_,_) :- read_from_codes(CodesIn,Tree).

% a version of my_read_from_codes which has no access to error stream
my_read_from_codes(end_of_file,_) :- !,
    missing_parser_diagnostics,
    throw(parse_errors([error('Parser not available (end_of_file on input stream)',none)])).
my_read_from_codes(ErrorCodes,_Tree) :-
    append("Error",_,ErrorCodes),!,
    atom_codes(ErrMsg,ErrorCodes),
    add_error(parsercall,'Error running/starting parser: ',ErrMsg),
    fail.
my_read_from_codes(CodesIn,Tree) :- read_from_codes(CodesIn,Tree).

missing_parser_diagnostics :-
    parser_location(Classpath), debug_println(9,classpath(Classpath)),fail.
missing_parser_diagnostics :- runtime_application_path(P),
   atom_concat(P,'/lib/probcliparser.jar',ConParser),
   (file_exists(ConParser)
      -> add_error(parsercall,'The Java B parser (probcliparser.jar) cannot be launched: ',ConParser),
         add_error(parsercall,'Please check that Java and B Parser are available, e.g., using "probcli -check_java_version" command.')
      ;  add_internal_error('Java B parser (probcliparser.jar) is missing from lib directory in: ',P)
    ).

% the position reported by the parser is slightly off,
% because the formula has to be submitted as
% "#FORMULA ...", "#EXPRESSION ..." or "#PREDICATE ..."
%handle_parser_exception(E) :- nl,print(E),nl,fail.
handle_parser_exception(parse_exception(Pos,Msg)) :-
    remove_msg_posinfo_known(Msg,SanitizedMsg),
    throw(parse_errors([error(SanitizedMsg,Pos)])).
handle_parser_exception(io_exception(Pos,Msg)) :-
    throw(parse_errors([error(Msg,Pos)])).
handle_parser_exception(io_exception(Msg)) :-
    throw(parse_errors([error(Msg,unknown)])).
handle_parser_exception(exception(Msg)) :-
    throw(parse_errors([error(Msg,unknown)])). % TO DO: get Pos from Java parser !
handle_parser_exception(_).

:- mode parse_formula(+Codes,-Tree).
parse_formula(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(formula,Extended,Codes,Tree).

:- mode parse_expression(+Codes,-Tree).
parse_expression(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(expression,Extended,Codes,Tree).

:- mode parse_predicate(+Codes,-Tree).
parse_predicate(Codes,Tree) :-
    get_preference(bool_expression_as_predicate,Extended),
    parse_x(predicate,Extended,Codes,Tree).

:- mode parse_substitution(+Codes,-Tree).
parse_substitution(Codes,Tree) :-
    parse_x(substitution,false,Codes,Tree).

:- mode parse(+Kind,+Codes,-Tree).
parse(formula,Codes,Tree) :- parse_formula(Codes,Tree).
parse(expression,Codes,Tree) :- parse_expression(Codes,Tree).
parse(predicate,Codes,Tree) :- parse_predicate(Codes,Tree).
parse(substitution,Codes,Tree) :- parse_substitution(Codes,Tree).

:- mode parse_temporal_formula(+Kind,+LanguageExtension,+Formula,-Tree).
parse_temporal_formula(Kind,LanguageExtension,Formula,Tree) :-
    write_to_codes(Formula,CodesWithNewlines),
    % Java parser expects only one line of input -> remove newlines
    cleanup(CodesWithNewlines,Codes),
    get_console_parser(In,Out,Err),
    send_definitions(In),
    format(In,'~s\n~s\n', [Kind,LanguageExtension]),
    format(In,'~s\n',[Codes]),
    read_line(Out,CodesIn), % TODO: Improve feedback from java!
    catch(my_read_from_codes(CodesIn,Tree,Out,Err),_,throw(parse_errors([error(exception,none)]))).

% throws parse_errors(Errors) in case of syntax errors
load_b_machine_as_term(Filename, Machine) :- %print(load_b_machine(Filename)),nl,
    prob_filename(Filename, ProBFile),
    debug_println(9,prob_filename(Filename,ProBFile)),
    ( dont_need_compilation(Filename,ProBFile,Machine,LoadResult) ->
        % no compilation is needed
        (LoadResult=loaded -> debug_println(20,'.prob file up-to-date')
        ; print_error('*** Loading failed *** '),nl,fail)
    ; otherwise ->
        debug_println(20,'Calling Parser'),
        (debug:debug_mode(on) ->
            time(call_console_parser(Filename,ProBFile))
        ;  call_console_parser(Filename,ProBFile)
        ),
        load_b_machine_probfile_as_term(ProBFile, Machine)
    ).

% converts the filename of the machine into a filename for the parsed machine
prob_filename(Filename, ProB) :-
    split_filename(Filename,Basename,_Extension),
    safe_atom_chars(Basename,BasenameC,prob_filename1),
    append(BasenameC,['.','p','r','o','b'],ProBC),
    safe_atom_chars(ProB,ProBC,prob_filename2),!.
prob_filename(Filename, ProB) :-
    add_failed_call_error(prob_filename(Filename,ProB)),fail.

check_version_and_read_machine(ProBFile, Version, Machine) :-
    (Version='$UNKNOWN' -> print('*** Unknown parser version number. Trying to load parsed file.'),nl
                          /* keep CheckVersionNr as variable */
     ; CheckVersionNr=Version),
    read_machine_term_from_file(ProBFile,parser_version(CheckVersionNr),Machine).

read_machine_term_from_file(ProBFile,VersionTerm,CompleteMachine) :-
   (debug:debug_mode(on)
     ->  time(read_machine_term_from_file_aux(ProBFile,VersionTerm,CompleteMachine))
     ;   read_machine_term_from_file_aux(ProBFile,VersionTerm,CompleteMachine)).

read_machine_term_from_file_aux(ProBFile,VersionTerm,complete_machine(Main,Machines,Files)) :-
    open(ProBFile, read, S),
    call_cleanup(( read(S,VersionTerm),
                   %read(S,classical_b(Main,Files)), read_machines(S,Machines)),
                   read_machines(S,Machines,Main,Files),
                   (var(Main) -> add_internal_error('Missing classical_b fact',ProBFile), Main=unknown,Files=[]
                     ; true)
                   ),
                 close(S)).

load_b_machine_list_of_facts_as_term(Facts,Machine) :-
    read_machine_term_from_list_of_facts(Facts,_,Machine).
read_machine_term_from_list_of_facts(Facts,VersionTerm,complete_machine(Main,Machines,Files)) :-
    (select(parser_version(VERS),Facts,F2) -> VersionTerm = parser_version(VERS)
      ; add_internal_error('No parser_version available',read_machine_from_list_of_facts),
        F2=Facts, VersionTerm = unknown),
    (select(classical_b(Main,Files),F2,F3) -> true
      ; add_internal_error('No classical_b file information available',read_machine_from_list_of_facts),
        fail),
     print(loading_classical_b(VersionTerm,Main,Files)),nl,
     include(is_machine_term,F3,F4),
     maplist(get_machine_term,F4,Machines).

is_machine_term(machine(_M)).
get_machine_term(machine(M),M).

load_b_machine_probfile_as_term(ProBFile, Machine) :-
    debug_println(9,consulting(ProBFile)),
    ( read_machine_term_from_file(ProBFile,_VersionTerm,Machine) -> true
    ; otherwise ->
        add_error(parsercall,'Failed to read parser output: wrong format?',ProBFile),fail).

read_machines(S,Machines,Main,Files) :-
    read(S,Term),
    ( Term == end_of_file ->
        Machines = []
    ; Term = classical_b(M,F) ->
        ((M,F)=(Main,Files) -> true ; add_internal_error('Inconsistent classical_b fact: ',classical_b(M,F))),
        read_machines(S,Machines,Main,Files)
    ; Term = machine(M) ->
        Machines = [M|Mrest],
        read_machines(S,Mrest,Main,Files)
    ).

% checks if all files that the prob-file depends on
% are older than the prob-file
dont_need_compilation(Filename,ProB,Machine,LoadResult) :-
    file_exists(Filename),
    catch( (   get_parser_version(Version),
               file_property(Filename, modify_timestamp, BTime),
               debug_print(19,'Parser Version: '), debug_println(19,Version),
               debug_println(9,' Filename Timestamp: '), debug_print(9,Filename), debug_print(9,' : '),debug_println(9,BTime),
               file_exists(ProB), % the .prob file exists
               file_property(ProB, modify_timestamp, ProbTime),
               debug_println(9,' .prob Timestamp: '), debug_print(9,ProB), debug_print(9,' : '),debug_println(9,ProbTime),
               % print(time_stamps(ProB,ProbTime,Filename,BTime)),nl,
               BTime < ProbTime, % .mch file is older than .prob file
               check_version_and_read_machine(ProB, Version, Machine),
               Machine = complete_machine(_,_,Files),
               % print(checking_all_older(Files,ProbTime)),nl,
               all_older(Files, ProbTime),
               LoadResult = loaded),
           error(A,B),
           (debug_println(9,error(A,B)),
            (A=resource_error(memory)
              -> add_error(load_b_machine,'Insufficient memory to load file.\nTry starting up ProB with more memory\n (e.g., setting GLOBALSTKSIZE=500M before starting ProB).'),
                 LoadResult=A
              ; fail))).
all_older([],_).
all_older([File|Rest],PTime) :- % print(checking(File)),nl,trace,
    (file_exists(File) -> true ; print('File does not exist anymore: '), print(File),nl,fail),
    file_property(File, modify_timestamp, BTime),
    debug_print(9,' Subsidiary File Timestamp: '), debug_print(9,File), debug_print(9,' : '),debug_println(9,BTime),
    ( BTime < PTime -> true
    ; print('File has changed : '), print(File), print(' : '), print(BTime),nl,fail ),
    all_older(Rest,PTime).

:- use_module(system_call).

% use the former LARGE_JVM by default
% if we are running on a 64-bit system
%get_jvm_options(['-Xmx1500m','-Xss5m']) :- platform_is_64_bit, !.  % use -Xmx221500000m instead of -Xss5m or Xmx to force exception and test ProB's response
get_jvm_options(['-Xss5m']) :- platform_is_64_bit, !.
get_jvm_options([]).

% ensure that the console Java process is running
ensure_console_parser_launched :-
   on_exception(E,parse_formula("1",_),
      (add_internal_error('Cannot launch console parser (probcliparser.jar)',E),fail)).


get_console_parser(In,Out,Err) :-
    java_parser_process(_PID,In,Out,Err),
    % calling is_process seams to be unreliable (if the process has crashed)
    % hence, we call process_wait with a timeout of 0
    % if this does not return an exit value of 'timeout'
    % the process has either been killed or died on its own
    % -- turns out that using process_wait does not work either
    % -- process_wait(PID,timeout,[timeout(0)]), !.
    % now we just check if in and output streams are still available
    \+ at_end_of_stream(Out), !,
    (at_end_of_stream(Err)
     -> add_error(parsercall,'Java process not running'),
        read_lines_and_add_as_error(Out)
     ;  true).
get_console_parser(In,Out,Err) :-
    clear_old_console_parsers,
    get_java_command(JavaCmd,Classpath),
    % maybe move to a predicate similar to get_jvm_options for Properties
    get_PROBPATH(PROBPATH),
    atom_concat('-Dprob.stdlib=', PROBPATH, Stdlibpref),
    %
    get_jvm_options(XOpts),
    append(XOpts, [Stdlibpref, '-cp',Classpath,'de.prob.cliparser.CliBParser',
                   %% '-v', % comment in for verbose
                   '-prolog','-lineno','-prepl'], FullOpts),
    debug_println(15,launching_java_console_parser(JavaCmd,FullOpts,probpath(PROBPATH))),
    system_call_keep_open(JavaCmd,FullOpts,PID,In,Out,Err,[]), % note: if probcliparser.jar does not exist we will not yet get an error here !
    debug_println(15,java_console_parser_launched(PID)),
    assert(java_parser_process(PID,In,Out,Err)), !.
get_console_parser(_,_,_) :- missing_parser_diagnostics, fail.

clear_old_console_parsers :-
    retract(java_parser_process(PID,_,_,_)),
    process_release(PID),
    fail.
clear_old_console_parsers.

call_console_parser(Filename, ProB) :-
    get_console_parser(In,Out,Err),

    % path to the .prob file
    replace_windows_path_backslashes(ProB,ProBW),
    % the input file
    replace_windows_path_backslashes(Filename,FilenameW),
    format(In,'machine\n',[]),
    format(In,'~w\n',[FilenameW]),
    format(In,'~w\n',[ProBW]),

    safe_read_line(Out,Err,1,[],Term),
    %print(result(Term)),nl,
    handle_console_parser_result(Term,Err).

safe_read_line(Out,Err,LineNr,SoFar,Term) :-
    safe_read_line_aux(read_line,Out,Err,LineNr,SoFar,Term).
safe_read_line_if_ready(Out,Err,LineNr,SoFar,Term) :-
    safe_read_line_aux(read_line_if_ready,Out,Err,LineNr,SoFar,Term).

safe_read_line_aux(Pred,Out,Err,LineNr,SoFar,Term) :-
    call(Pred,Out,Codes),    % read another line
    Codes \= end_of_file, !, % happens if parser crashed and stream is closed
    append(SoFar,Codes,NewCodes),
    catch(my_read_from_codes(NewCodes,Term,Out,Err),Exception,
           safe_read_exception(Out,Err,LineNr,Exception,NewCodes,Term)).
safe_read_line_aux(_Pred,_Out,Err,_LineNr,CodesSoFar,_) :-
    safe_name(T,CodesSoFar),
    add_error(parsercall,'Unexpected error while parsing machine: ',T),
    read_lines_and_add_as_error(Err),fail.


safe_read_exception(Out,_,1,_Exception,CodesSoFar,_Term) :- append("Error",_,CodesSoFar),!,
    % probably some Java error, we do not obtain a correct Prolog term, but probably something like:
    % "Error occurred during initialization of VM"
    safe_name(T,CodesSoFar),
    add_error(parsercall,'Java error while parsing machine: ',T),
    read_lines_and_add_as_error(Out),fail.
safe_read_exception(Out,Err,LineNr,error(syntax_error(M),_),CodesSoFar,Term) :- LineNr<10, % avoid infinite loop in case unexpected errors occur which cannot be solved by reading more input lines
    !,
    %(debug:debug_mode(off) -> true ; format('Syntax error in parser result (line ~w): ~w~nTrying to read another line.~n',[LineNr,M])),
    add_warning(parsercall,'Syntax error in parser result; trying to read another line. Line:Error = ',LineNr:M),
    L1 is LineNr+1,
    safe_read_line_if_ready(Out,Err,L1,CodesSoFar,Term). % try and read another line; we assume the syntax error is because the TERM is not complete yet and transmitted on multiple lines (Windows issue)
safe_read_exception(_,_,_LineNr,parse_errors(Errors),_CodesSoFar,_Term) :- !,
    throw(parse_errors(Errors)).
safe_read_exception(_,_,_LineNr,Exception,_CodesSoFar,_Term) :-
    add_internal_error('Unexpected exception occurred during parsing: ',Exception),fail.
    %safe_name(T,CodesSoFar), add_error_fail(parsercall,'Unexpected error while parsing machine: ',T).

safe_name(Name,Codes) :- var(Name),Codes == end_of_file,!,
   %add_internal_error('Parser does not seem to be available',''),
   Name=end_of_file.
safe_name(Name,Codes) :-
  on_exception(E,name(Name,Codes),
               (print('Exception during name: '),print(E),nl,nl,throw(E))).

% see also handle_parser_exception
%handle_console_parser_result(E,_) :- nl,print(E),nl,fail.
handle_console_parser_result(exit(0),_) :- !.
handle_console_parser_result(io_exception(Msg),_) :- !,
    add_error_fail(parsercall,'Error while opening the B file: ',Msg).
handle_console_parser_result(preparse_exception(Tokens,Msg),_) :- !,
    remove_msg_posinfo(Msg,SanitizedMsg,MsgPos),
    findall(error(SanitizedMsg,Pos),member(Pos,Tokens),Errors1),
    (Errors1 = []
     -> /* if there are no error tokens: use pos. info from text Msg */
        Errors = [error(SanitizedMsg,MsgPos)]
     ;  Errors = Errors1),
    throw(parse_errors(Errors)).
handle_console_parser_result(parse_exception(Pos,Msg),_) :- !,
    remove_msg_posinfo_known(Msg,SanitizedMsg),  %print(sanitized(SanitizedMsg)),nl,
    throw(parse_errors([error(SanitizedMsg,Pos)])).
handle_console_parser_result(exception(Msg),_) :- !,
    %add_error(bmachine,'Error in the classical B parser: ', Msg),
    remove_msg_posinfo(Msg,SanitizedMsg,Pos),
    throw(parse_errors([error(SanitizedMsg,Pos)])).
handle_console_parser_result(end_of_file,Err) :- !,
     % probably NullPointerException or similar in parser; no result on StdOut
     debug_println(9,end_of_file_on_parser_output_stream),
     format(user_error,'Detected abnormal termination of B Parser~n',[]), % print in case we block
     read_line_if_ready(Err,Err1),
     (Err1=end_of_file -> Msg = 'Abnormal termination of B Parser: EOF on streams'
      ; append("Abnormal termination of B Parser: ",Err1,ErrMsgCodes),
        safe_name(Msg,ErrMsgCodes),
        read_lines_and_add_as_error(Err) % also show additional error infos
     ),
     throw(parse_errors([error(Msg,none)])).
handle_console_parser_result(Term,_) :- !,
    write_term_to_codes(Term,Text,[]),
    safe_name(Msg,Text),
    %add_error(bmachine,'Error in the classical B parser: ', Msg),
    remove_msg_posinfo(Msg,SanitizedMsg,Pos),
    throw(parse_errors([error(SanitizedMsg,Pos)])).

% now replaced by remove_msg_posinfo:
%extract_position_info(Text,pos(Row,Col,Filename)) :-
%    atom_codes(Text,Codes),
%    epi(Row,Col,Filename,Codes,_),!.
%extract_position_info(_Text,none).

:- assert_must_succeed((parsercall:epi(R,C,F,"[2,3] xxx in file: /usr/lib.c",[]), R==2,C==3,F=='/usr/lib.c')).
:- assert_must_succeed((parsercall:epi(R,C,F,"[22,33] xxx  yyy ",_), R==22,C==33,F=='unknown')).

% we expect error messages to start with [Line,Col] information
epi(Row,Col,Filename) --> " ",!,epi(Row,Col,Filename).
epi(Row,Col,Filename) --> "[", epi_number(Row),",",epi_number(Col),"]",
   (epi_filename(Filename) -> [] ; {Filename=unknown}).
epi_number(N) --> epi_numbers(Chars),{safe_name(N,Chars)}.
epi_numbers([C|Rest]) --> [C],{is_number(C)},epi_numbers2(Rest).
epi_numbers2([C|Rest]) --> [C],{is_number(C),!},epi_numbers2(Rest).
epi_numbers2("") --> !.
is_number(C) :- member(C,"0123456789"),!.
epi_filename(F) --> " in file: ",!,epi_path2(Chars),{safe_name(F,Chars)}.
epi_filename(F) --> [_], epi_filename(F).
epi_path2([C|Rest]) --> [C],!,epi_path2(Rest). % assume everything until the end is a filename
epi_path2([]) --> [].

:- assert_must_succeed((parsercall:remove_rowcol(R,C,"[22,33] xxx  yyy ",_), R==22,C==33)).
% remove parser location info from message to avoid user clutter
remove_rowcol(Row,Col) --> " ",!, remove_rowcol(Row,Col).
remove_rowcol(Row,Col) --> "[", epi_number(Row),",",epi_number(Col),"]".

:- assert_must_succeed((parsercall:remove_filename(F,C,"xxx  yyy in file: a.out",R), F == 'a.out', C == "xxx  yyy", R==[])).
remove_filename(F,[]) --> " in file: ",!,epi_path2(Chars),{safe_name(F,Chars)}.
remove_filename(F,[C|T]) --> [C], remove_filename(F,T).

:- assert_must_succeed((parsercall:remove_filename_from_codes("xxx  yyy in file: a.out",F,C), F == 'a.out', C == "xxx  yyy")).
:- assert_must_succeed((parsercall:remove_filename_from_codes("xxx  yyy zzz",F,C), F == 'unknown', C == "xxx  yyy zzz")).
remove_filename_from_codes(Codes,F,NewCodes) :-
   (remove_filename(F,C1,Codes,C2) -> append(C1,C2,NewCodes) ; F=unknown, NewCodes=Codes).

% obtain position information from a message atom and remove the parts from the message at the same time
remove_msg_posinfo_known(Msg,NewMsg) :- remove_msg_posinfo(Msg,NewMsg,_,pos_already_known).
remove_msg_posinfo(Msg,NewMsg,Pos) :- remove_msg_posinfo(Msg,NewMsg,Pos,not_known).
remove_msg_posinfo(Msg,NewMsg,Pos,AlreadyKnown) :-  %print(msg(Msg,AlreadyKnown)),nl,
   atom_codes(Msg,Codes),
   (remove_rowcol(Row,Col,Codes,RestCodes)
    -> Pos=pos(Row,Col,Filename),
       remove_filename_from_codes(RestCodes,Filename,NewCodes)
     ; AlreadyKnown==pos_already_known ->
       Pos = none,
       remove_filename_from_codes(Codes,Filename,NewCodes)
     ; otherwise ->
       NewCodes=Codes, Pos=none
       % do not remove filename: we have not found a position and cannot return the filename
       % see, e.g., public_examples/B/Tickets/Hansen27_NestedMchErr/M1.mch
    ),
   !,
   atom_codes(NewMsg,NewCodes).
remove_msg_posinfo(M,M,none,_).


% ----------------------

get_parser_version(Version) :- java_parser_version(Version),!.
get_parser_version(Version) :-
    get_version_from_parser(Version),
    assert(java_parser_version(Version)),
    !.
get_parser_version(_Vers) :-
    (real_error_occurred -> true % already produced error messages in get_version_from_parser
     ; add_error(get_parser_version,'Could not get parser version.')),
    %missing_parser_diagnostics, % already called if necessary by get_version_from_parser
    fail.


get_version_from_parser(Version) :-
    get_console_parser(In,Out,Err),
    format(In,'version\n',[]),
    read_line(Out,Text),  % if parser cannot be started we receive end_of_file; length will fail
    (Text = end_of_file -> add_error(parsercall,'Could not get parser version: '),
                           read_lines_and_add_as_error(Err),fail
     ; length(Text,Length),
        ( append("Error",_,Text)
                       -> atom_codes(ErrMsg,Text),
                          add_error(parsercall,'Error getting parser version: ',ErrMsg),
                          read_lines_and_add_as_error(Out),fail
        ; Length < 100 -> atom_codes(Version,Text)
        ; otherwise    -> prefix_length(Text,Prefix,100),
                          append(Prefix,"...",Full),
                          safe_name(Msg,Full),
                          add_error_fail(parsercall, 'B parser returned unexpected version string: ', Msg))
    ).

:- use_module(library(sockets),[socket_select/7]).
% check if a stream is ready for reading
stream_ready_to_read(Out) :- var(Out),!,
    add_internal_error('Illegal stream: ',stream_ready_to_read(Out)),fail.
stream_ready_to_read(Out) :-
    socket_select([],_, [Out],RReady, [],_, 1), % wait 1 sec at most %print(ready(Out,RReady)),nl,
    RReady == [Out].


read_lines_and_add_as_error(Out) :- read_lines_until_eof(Out,false,ErrMsgCodes,[]),
    (ErrMsgCodes=[] -> true
      ; ErrMsgCodes=[10] -> true % just a single newline
      ; ErrMsgCodes=[13] -> true % ditto
      ; safe_name(ErrMsg,ErrMsgCodes),
        add_error(parsercall,'Additional information: ',ErrMsg)).

read_line_if_ready(Stream,Line) :- stream_ready_to_read(Stream), !, read_line(Stream,Line).
read_line_if_ready(_,end_of_file). % pretend we are at the eof

read_lines_until_eof(Out,NewlineRequired) --> {read_line_if_ready(Out,Text)},
    !,
    ({Text = end_of_file} -> []
      ; ({NewlineRequired==true} -> "\n" ; []),
         Text, %{format("read: ~s~n",[Text])},
         read_lines_until_eof(Out,true)).
read_lines_until_eof(_Out,_) --> [].

get_java_command(JavaCmdW,ClasspathW) :-
    FileList = ['probcliparser.jar'],
    (parser_location(Classpath) -> true
        ; otherwise -> libraries_as_classpath(FileList,Classpath)),
    get_java_command_path(JavaCmdW),
    replace_windows_path_backslashes(Classpath,ClasspathW).


% a predicate to check whether we have correct java version number and whether java seems to work ok
% ResultStatus=compatible if everything is ok
check_java_version(VersionCheckResult,ResultStatus) :- get_java_fullversion(Version),!,
   check_java_version_aux(Version,VersionCheckResult,ResultStatus).
check_java_version(VersionCheckResult,error) :- get_java_command(JavaCmdW,ClasspathW),!,
   ajoin(['*** Unable to launch java command: ',JavaCmdW,' (classpath: ',ClasspathW,')!'],VersionCheckResult).
check_java_version('*** Unable to launch Java or get path to java command!',error). % should not happen

check_java_version_aux(VersionCodes,VersionCheckResult,ResultStatus) :-
     get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom),!,
     check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,ResultStatus).
check_java_version_aux(VersionCodes,VersionCheckResult,error) :-
     atom_codes(VersionA,VersionCodes),
     ajoin(['*** Unable to identify Java version number: ',VersionA, ' !'],VersionCheckResult).

check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,compatible) :- java_version_ok_for_parser(V1,V2,V3Atom),
     get_java_version(_),!, % if get_java_version fails, this is an indication that Java not fully installed with admin rights
     ajoin(['Java is correctly installed and version ',V1,'.',V2,'.',V3Atom,
            ' is compatible with ProB requirements (>= 1.7).'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,error) :- java_version_ok_for_parser(V1,V2,V3Atom),!,
     ajoin(['*** Java version ',V1,'.',V2,'.',V3Atom,
            ' is compatible with ProB requirements (>= 1.7) ',
            'but does not seem to be correctly installed: reinstall Java with admin rights!'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,incompatible) :-
     get_java_version(_),!,
     ajoin(['*** Java is correctly installed but version ',V1,'.',V2,'.',V3Atom,
            ' is *not* compatible with ProB requirements (>= 1.7)!'],VersionCheckResult).
check_java_version_aux2(V1,V2,V3Atom,VersionCheckResult,error) :-
     ajoin(['*** Java is not correctly installed and version ',V1,'.',V2,'.',V3Atom,
            ' is *not* compatible with ProB requirements (>= 1.7)!'],VersionCheckResult).


% we need Java 1.7 or higher
java_version_ok_for_parser(V1,_V2,_) :- V1>1.
java_version_ok_for_parser(1,V2,_) :- V2>=7.

get_java_fullversion(V1,V2,V3Atom) :- get_java_fullversion(VersionCodes),
     get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom).

:- use_module(tools,[split_chars/3]).
get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom) :-
    append("java full version """,Tail,VersionCodes),
    get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom).
get_java_fullversion_numbers(VersionCodes,V1,V2,V3Atom) :-
    append("openjdk full version """,Tail,VersionCodes),
    get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom).
get_java_fullversion_numbers_aux(Tail,V1,V2,V3Atom) :-
    split_chars(Tail,".",[V1C,V2C|FurtherC]),
    append(FurtherC,V3C),
    number_codes(V1,V1C), number_codes(V2,V2C),
    (append(V3C2,[_,_],V3C) % try remove trailing " and newline
     -> atom_codes(V3Atom,V3C2) ; atom_codes(V3Atom,V3C)).

% get java fullversion as code list; format "java full version "...."\n"
get_java_fullversion(Version) :-
   get_java_command_path(JavaCmdW),
   (my_system_call(JavaCmdW, ['-fullversion'],java_version,Text) -> Text=Version
     ; format(user_error,'Could not execute ~w -fullversion~n',[JavaCmdW]),
       fail).

% when java -fullversion works but not java -version it seems that Java was improperly
% installed without admin rights (http://stackoverflow.com/questions/11808829/jre-1-7-returns-java-lang-noclassdeffounderror-java-lang-object)
get_java_version(Version) :-
   get_java_command_path(JavaCmdW),
   (my_system_call(JavaCmdW, ['-version'],java_version,Text) -> Text=Version
     ; format(user_error,'Could not execute ~w -version~n',[JavaCmdW]),
       fail).
% should we check for JAVA_PATH first ? would allow user to override java; but default pref would have to be set to ''
get_java_command_path(JavaCmdW) :-
    absolute_file_name(path(java),
                       JavaCmd,
                       [access(exist),extensions(['.exe','']),solutions(all),file_errors(fail)]),
    debug_println(8,obtained_java_cmd(JavaCmd)),
    replace_windows_path_backslashes(JavaCmd,JavaCmdW),
    !.
get_java_command_path(JavaCmdW) :- get_preference(path_to_java,JavaCmd),
    JavaCmd \= '',
    debug_println(8,path_to_java(JavaCmd)),
    (file_exists(JavaCmd) -> true
     ; directory_exists(JavaCmd) -> add_warning(get_java_command_path,'No Java Runtime (7 or newer) found; please correct the JAVA_PATH advanced preference (it has to point to the java tool, *not* the directory enclosing it): ',JavaCmd)
     ; add_warning(get_java_command_path,'No Java Runtime (7 or newer) found; please correct the JAVA_PATH advanced preference: ',JavaCmd)),
     replace_windows_path_backslashes(JavaCmd,JavaCmdW),
     !.
get_java_command_path(_JavaCmd) :-
    add_error(get_java_command_path,'Could not get path to the java tool. Make sure a Java Runtime (7 or newer) is installed',''),
    fail.

libraries_as_classpath(Libs,Classpath) :-
   %findall(PLD,user:file_search_path(prob_lib,PLD),LL2), print(prob_lib_directory(LL2)),nl,
    maplist(library_abs_name,Libs,ALibs),
    ( host_platform(windows) -> Sep = (';')
    ; otherwise              -> Sep = (':')),
    ajoin_with_sep(ALibs,Sep,Classpath).
library_abs_name(Lib,Abs) :- absolute_file_name(prob_lib(Lib),Abs).


call_ltl_parser(Formulas, CtlOrLtl, Result) :-
    get_ltl_lang_spec(LangSpec),
    maplist(parse_temporal_formula(CtlOrLtl,LangSpec),Formulas,Result).

:- use_module(specfile,[b_or_z_mode/0,csp_with_bz_mode/0]).
get_ltl_lang_spec('B,none') :- csp_with_bz_mode,!.
get_ltl_lang_spec('B') :- b_or_z_mode,!.
get_ltl_lang_spec(none).

/* unused code :
generate_formula_codes([F]) -->
    !,generate_formula_codes2(F),"\n".
generate_formula_codes([F|Rest]) -->
    generate_formula_codes2(F), "###\n",
    generate_formula_codes(Rest).
generate_formula_codes2(F) -->
    write_to_codes(F).
*/


% ---------------------------

% call the TLA2B parser to translate a TLA file into a B machine
call_tla2b_parser(TLAFile) :-
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(TLAFile,TLAFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('TLA2B.jar'),TLA2BJAR),
   append(XOpts,['-jar',TLA2BJAR,TLAFileW],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   (my_system_call(JavaCmdW, FullOpts,tla2b_parser) -> true
     ; print_error('Be sure to download TLA2B.zip from http://nightly.cobra.cs.uni-duesseldorf.de/tla,'),
       print_error('then unzip the archive and put TLA2B.jar into the ProB lib/ directory.'),
       fail).

call_cspmj_parser(CSPFile,PrologFile) :-
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(CSPFile,CSPFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('cspmj.jar'),CSPMJAR),
   get_writable_compiled_filename(CSPFile,'.pl',PrologFile),
   string_concatenate('--prologOut=', PrologFile, PrologFileOutArg),
   append(XOpts,['-jar',CSPMJAR,'-parse',CSPFileW,PrologFileOutArg],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   (my_system_call(JavaCmdW, FullOpts,cspmj_parser) -> true
     ; print_error('Be sure that cspmj.jar parser is installed.'),
       fail).


tla2prob_filename(TLAFile,GeneratedProbFile) :-
    split_filename(TLAFile,Basename,_Extension),
    atom_chars(Basename,BasenameC),
    append(BasenameC,['.','p','r','o','b'],TLABC),
    atom_chars(GeneratedProbFile,TLABC),!.
tla2prob_filename(Filename, GeneratedProbFile) :-
    add_failed_call_error(tla2b_filename(Filename,GeneratedProbFile)),fail.

tla2b_filename(TLAFile,BFile) :-
    split_filename(TLAFile,Basename,_Extension),
    atom_chars(Basename,BasenameC),
    append(BasenameC,['_',t,l,a,'.','m','c','h'],TLABC),
    atom_chars(BFile,TLABC),!.
tla2b_filename(Filename, BFile) :-
    add_failed_call_error(tla2b_filename(Filename,BFile)),fail.


my_system_call(Command,Options,Origin) :- my_system_call(Command,Options,Origin,_Text).
my_system_call(Command,Options,Origin,Text) :-
   system_call(Command,Options,Text,Exit), %nl,print(exit(Exit)),nl,nl,
   treat_exit_code(Exit,Command,Options,Text,Origin).
treat_exit_code(exit(0),_,_,_,_) :- !.
treat_exit_code(Exit,Command,Options,Text,Origin) :-
   debug_println(9,treat_exit_code(Exit,Command,Options,Text,Origin)),
   catch( my_read_from_codes(Text,Term),
          _,
          (safe_name(T,Text),Term = T)), !,
    print_error(Term),
    add_error(Origin,'Error occurred during sytem_call: ',Term),
    fail.


% replace backslashes by forward slashes in atoms, only under windows
% TODO(DP,14.8.2008): Check if still needed with SICStus 4.0.4 and if
% everybody uses that version
replace_windows_path_backslashes(Old,New) :-
    ( host_platform(windows) ->
        safe_name(Old,OldStr),
        replace_string_backslashes(OldStr,NewStr),
        safe_name(New,NewStr)
    ; otherwise ->
        Old=New).

replace_string_backslashes([],[]) :- !.
replace_string_backslashes([C|OldRest],[N|NewRest]) :- !,
    ( C=92 /* backslash */ ->
        N=47 /* forward slash */
    ; otherwise ->
        N=C ),
    replace_string_backslashes(OldRest,NewRest).
replace_string_backslashes(X,Y) :-
  add_error(parsercall,'Illegal call: ',replace_string_backslashes(X,Y)),
  fail.

% ---------------------------
:- use_module(pathes,[runtime_application_path/1]).
% call the Promela parser to translate a .pml file into a Prolog AST
call_promela_parser(PMLFile) :-
   promela_prolog_filename(PMLFile,PrologFile),
   (file_exists(PrologFile) -> delete_file(PrologFile) ; true),
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(PMLFile,PMLFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('Promela.jar'),PromelaJAR),
   runtime_application_path(AppPath),
   append(XOpts,['-jar',PromelaJAR,PMLFileW,AppPath],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   my_system_call(JavaCmdW, FullOpts,promela_parser).

promela_prolog_filename(PMLFile,PrologFile) :-
    atom_chars(PMLFile,BasenameC),
    append(BasenameC,['.','p','l'],NewC),
    atom_chars(PrologFile,NewC),!.
promela_prolog_filename(Filename, PrologFile) :-
    add_failed_call_error(promela_prolog_filename(Filename,PrologFile)),fail.
