% (c) 2014-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(enabling_predicates, [compute_enabling_predicate/5]).

/* Modules and Infos for the code coverage analysis */
:- use_module(probsrc(module_information)).
:- module_info(group,model_checker).
:- module_info(description,'This module provides predicates computing weakest preconditions without taking the preconditions of the substitutions into account').

% Standard SICSTUS prolog libraries
:- use_module(library(lists)).

% Classical B prolog modules
:- use_module(probsrc(bmachine),[b_get_machine_operation/4, b_get_invariant_from_machine/1]).
:- use_module(probsrc(bsyntaxtree),[conjunct_predicates/2, conjunction_to_list/2,
                                    get_texpr_expr/2, create_texpr/4, %remove_all_infos/2, 
                                    create_exists/3, safe_create_texpr/3, get_texpr_type/2, create_typed_id/3,
                                    predicate_components_in_scope/3,get_texpr_ids/2, some_id_occurs_in_expr/2, find_identifier_uses/3]).
:- use_module(probsrc(bmachine_eventb), [prime_variables/2, is_primed_id/1]).

% Predicates for calculating the weakest-precondition
:- use_module(probsrc(weakest_preconditions), [weakest_precondition/3]).

% POR modules
:- use_module(probporsrc(b_simplifier),[simplify_b_predicate/2]).

% Other ProB modules
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(translate),[translate_bexpression/2, translate_substitution/2]).
:- use_module(probsrc(debug), [debug_println/2, debug_mode/1]).
:- use_module(probsrc(error_manager),[add_error/3,add_error_fail/3]).


%enumeration_warning(enumerating(n),'INTEGER',1:sup,1:3,critical)
% TODO: After computing the weakest-precondition try to simplify the predicate
compute_enabling_predicate(OpName,OmitOpNameGuard,Pred,FindInvViolations,Res) :-
  b_get_machine_operation(OpName,_Res,Par,Body),
  (FindInvViolations==1 ->
    b_get_invariant_from_machine(Inv),
    conjunct_predicates([Inv,Pred],Pred1)
  ;otherwise ->
    Pred1 = Pred
  ),
  % TODO: add a case for Event-B 
  % compute wp when the source operation is deterministic
  get_preference(enable_graph_depth,Depth),
  (Depth==0 ->
    WP = b(truth,pred,[])
    %add_warning(weakest_precondition,'Weakest precondition cannot be computed: ',(body(Body),pred(Pred))),
    %Res = invariant
  ; otherwise ->
    create_event_body(OpName,Par,Body,NewBody),
    compute_ep(NewBody,OmitOpNameGuard,Pred1,WP1),
    %remove_all_infos(WP1,WP2), % consider to use remove_all_infos/2 (problem: all nodeid's will be replaced by a variable,
                                % this causes instatiations errors when using predicate simplifier.)
    remove_nodeid_infos(WP1,WP)
  ),
  print_weakest_precondition(OpName,Body,Pred,WP),
  %% print(weakest_precondition(WP)),nl,
  simplify_b_predicate(WP,Simplified),
  print_b_expression(Simplified),
  Res=predicate(Simplified).

create_event_body(OpName,Par,Body,Res) :-
  create_event_body1(OpName,Par,Body,Res).

create_event_body1(_OpName,[],Body,Res) :- !,
  Res = Body.
create_event_body1(OpName,Par,Body,Res) :-
  is_list(Par),!,
  get_texpr_expr(Body,Expr),
  get_precondition_body(Expr,Par,OpName,InterimRes),
  create_texpr(InterimRes,subst,[],Res).
create_event_body1(OpName,Par,Body,_Res) :-
  add_error(create_event_body,'The Parameters should be wrapped in a prolog list: ', OpName:Par/Body),fail.

get_precondition_body(precondition(Condition,Action),Par,_OpName,Res) :- !,
  Res = any(Par,Condition,Action).
get_precondition_body(Body,_Par,_OpName,Res) :-
  functor(Body,rlevent,11),!,
  Res = Body.
get_precondition_body(OpName,Stmt,_Pre,_InnerBody) :-
  add_error(get_precondition_body,'Statement not covered yet: ', OpName:Stmt).

compute_ep(Body,OmitBodyGuard,Pred,WP) :-
  (OmitBodyGuard ->
    get_texpr_expr(Body,Expr),
    compute_ep1(Expr,Pred,WP)
  ; otherwise ->
    weakest_precondition(Body,Pred,WP)
  ).

compute_ep1(precondition(_Condition,Action),Pred,WP) :- !,
  % here we do not add the Condition as a conjunct
  weakest_precondition(Action,Pred,WP).
compute_ep1(select([When]),Pred,WP) :- !,
  get_texpr_expr(When,select_when(_PreCond, Body)),
  weakest_precondition(Body,Pred,WP).
%% compute_ep1(if([If]),Pred,WP) :- !,
%%   get_texpr_expr(If,if_elsif(_Test,Body)),
%%   weakest_precondition(Body,Pred,WP).
compute_ep1(any(Parameters,Guard,Action),Pred,WP) :- !,
  remove_free_predicates(Parameters,Guard,NewGuard),
  weakest_precondition(Action,Pred,InnerWP),
  safe_create_texpr(conjunct(NewGuard,InnerWP),pred,UnquantifiedWP),
  create_exists(Parameters,UnquantifiedWP,WP).
  %% weakest_preconditions: weakest_precondition2(any(Parameters,NewGuard,Action),Pred,WP). % using weakest precondition as enabling predicate is not correct
compute_ep1(rlevent(_Name,_Section,_Stat,Parameters,Guard,_Thm,Act,_VWit,_PWit,_Unmod,_AbsEvents),Pred,EP) :- !,
  %% debug_println(9,rlevent(_Name,_Section,_Stat,Parameters,Guard,_Thm,Act,_VWit,_PWit,_Unmod,_AbsEvents)),
  % we have "ANY t WHERE Q THEN T END" statement (here we need to add the guard to the WP)
  % we can sort out the free predicates from Q (at the point of testing the WP we know that there is a representation of the non-deterministic event)
  remove_free_predicates(Parameters,Guard,RestrictedGuard),
  % we normalise each event for the case that there are non-deterministic assignments
  (is_list(Act) ->
    normalise_to_deterministic_actions_event(Act,NewBody,Parameters,NewParameters,RestrictedGuard,NewGuard)
  ; otherwise ->
    normalise_to_deterministic_actions_event([Act],NewBody,Parameters,NewParameters,RestrictedGuard,NewGuard)
  ),
  weakest_precondition(NewBody,Pred,WP),
  (NewParameters = [] ->
    EP = WP
  ; otherwise ->
    safe_create_texpr(conjunct(NewGuard,WP),pred,UnquantifiedEP),
    create_exists(NewParameters,UnquantifiedEP,EP)
  ).
compute_ep1(Body,Pred,EP) :-
  create_texpr(Body,subst,[],WrappedBody),
  weakest_precondition(WrappedBody,Pred,EP).
  
normalise_to_deterministic_actions_event(Acts,NewBody,Parameters,NewParameters,Guard,NewGuard) :-
  normalise_nondeterministic_actions(Acts,PrimeParams,NewConjuncts,NewActs),
  append(Parameters,PrimeParams,NewParameters),
  conjunct_predicates([Guard|NewConjuncts],NewGuard),
  create_texpr(parallel(NewActs),subst,[],NewBody).
  %% safe_create_texpr(conjunct(Guard,NewConjuncts),pred,NewGuard).

normalise_nondeterministic_actions([],[],[],[]).
normalise_nondeterministic_actions([Act|Acts],Parameters,NewConjuncts,NewActs) :-
  get_texpr_expr(Act,Expr),
  (Expr = becomes_element_of(IDs,Set) ->
    prime_variables(IDs,Params),
    weakest_preconditions: create_members(Set,Params,SetCond),
    create_texpr(assign(IDs,Parameters),subst,[],PrimeAssign),
    append(Params,Parameters1,Parameters),
    NewConjuncts=[SetCond|NewConjuncts1],
    NewActs = [PrimeAssign|NewActs1]
  ; Expr = becomes_such(Vars,Condition) -> %trace,
    find_identifier_uses(Condition,[],IDs),
    include(is_primed_id,IDs,PrimeIDs),
    NewConjuncts=[Condition|NewConjuncts1],
    (length(Vars,N),length(PrimeIDs,N) ->
       sort(Vars,Vars1),sort(PrimeIDs,PrimeIDs1),
       maplist(get_texpr_type,Vars1,Types),
       maplist(create_typed_id,PrimeIDs1,Types,PrimedVars),
       create_texpr(assign(Vars1,PrimedVars),subst,[],PrimeAssign),
       append(PrimedVars,Parameters1,Parameters),
       NewActs = [PrimeAssign|NewActs1]
    ; otherwise ->
       add_error_fail(normalise_nondeterministic_actions,'Number of varibales does not agree with number of primed variables:  ',becomes_such(Vars,Condition))
    )
  ; otherwise ->
    NewConjuncts = NewConjuncts1,
    NewActs = [Act|NewActs1],
    Parameters = Parameters1
  ),
  normalise_nondeterministic_actions(Acts,Parameters1,NewConjuncts1,NewActs1).

remove_free_predicates(Parameters,Guard,NewGuard) :-
  conjunction_to_list(Guard,Conjuncts),
  get_texpr_ids(Parameters,Ids),
  filter_out_global_guards(Conjuncts,Ids,RestConjuncts),
  conjunct_predicates(RestConjuncts,NewGuard),
  %% predicate_components_in_scope(Guard,Parameters,Res),
  (debug_mode(on) ->
    print('####### Remove global guards #######'),nl,
    print('Identifiers: '),print(Ids),nl,
    remove_nodeid_infos(Guard,GuardNoInfos),
    print('Guard: '),translate:print_bexpr(GuardNoInfos),nl,
    print('NewGuard: '),translate:print_bexpr(NewGuard),nl,
    print('####################################'),nl
  ; otherwise -> true).

filter_out_global_guards([],_Ids,[]).
filter_out_global_guards([TExpr|TExprs],Ids,Res) :-
  (some_id_occurs_in_expr(Ids,TExpr) ->
    Res=[TExpr|Res1]
  ; otherwise ->
    Res = Res1
  ), filter_out_global_guards(TExprs,Ids,Res1).

/* For debugging only */
print_weakest_precondition(Opname,Body,Guard,WP) :-
  (debug_mode(on) ->
   print('######### Weakest Precondition #########'),nl,
   remove_nodeid_infos(Guard,GuardNoInfos),
   translate_bexpression(GuardNoInfos,Guard_PP),
   translate_bexpression(WP,WP_PP),
   translate_substitution(Body,RealBody),
   print('Operation/Event '),print(Opname),print(' :: '),nl,
   print('['),print(RealBody),print(']'),
   print('('),print(Guard_PP),print(')'),
   print(' = '),print(WP_PP),nl,
   print('########################################'),nl
  ;otherwise -> true).

print_b_expression(BExpr) :-
  (debug_mode(on) ->
   translate_bexpression(BExpr,String),
   conjunction_to_list(BExpr,List),
   print(String),nl,
   print('Conjunctions list: [ '),print_conjunct_l(List)
  ; true).

print_conjunct_l([]) :-
  print(' ]'),nl.
print_conjunct_l([Expr]) :- !,
  translate_bexpression(Expr,String),
  print(String),print(' ]'),nl.
print_conjunct_l([Expr|Exprs]) :-
  translate_bexpression(Expr,String),
  print(String),print(', '),
  print_conjunct_l(Exprs).

remove_nodeid_infos(Expr,Res) :-
  var(Expr),!,
  Res=Expr.
remove_nodeid_infos(Expr,Res) :-
  is_list(Expr),!,
  maplist(remove_nodeid_infos,Expr,Res).
remove_nodeid_infos(Expr,Res) :-
  Expr =.. [F|Args],
  remove_nodeid_infos1(F,Args,Res).

remove_nodeid_infos1(nodeid,_Args,Res) :-
  % ignore additional node infos
  !, Res =.. [nodeid,none].
remove_nodeid_infos1(F,Args,Res) :-
  maplist(remove_nodeid_infos,Args,ArgsRes),
  Res =.. [F|ArgsRes].
