% (c) 2009-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(preferences,
    [init_preferences/0,
     tcltk_get_preference/2, get_computed_preference/2,
     get_preference/2, get_preferences_list/1, get_preferences_list/2,
     set_preference/2,
     call_with_preference/3,
     temporary_set_preference/2, temporary_set_preference/3,
     reset_temporary_preference/1, reset_temporary_preference/2,
     reset_all_temporary_preferences/0,
     preference_category/2, preference_description/2, preference_val_type/2, preference_default_value/2,
     preference/2,
     is_of_type/2,
     get_ith_preference_type/2, get_ith_preference_value/2, set_ith_preference_value/2,
     get_ith_preference_type/3, get_ith_preference_value/3, set_ith_preference_value/3,
     load_preferences/1,save_preferences/1,
     backup_preferences/0, revert_preferences/0, reset_to_defaults/1, reset_for_benchmarking/0,
     set_preference_group/2,
     activate_recent_documents/0,deactivate_recent_documents/0,

     add_recent_document/1,  /* use to add a file to recent documents list */
     get_recent_documents/1, clear_recent_documents/0,
     add_element_to_history/3, get_history_elements_list/3,

     eclipse_preference/2, print_eclipse_prefs/0,
     advanced_eclipse_preference/2, obsolete_eclipse_preference/1,
     get_non_default_preferences/1,

     get_time_out_preference_with_factor/2,
     add_time_outs/3,
     time_out_preference_disabled/0
    ]).

:- use_module(tools, [safe_on_exception/3, print_message/1,  translate_term_into_atom/2,
     print_dynamic_pred/3, host_platform/1]).
:- use_module(tools_strings, [string_concatenate/3, ajoin/2]).
:- use_module(tools_printing, [print_error/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module manages the various preferences of ProB. Preferences can be grouped, saved and restored and validated from file.').

%:- use_module(debug). % no longer required
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(library(lists)).

:- volatile change_flag/0.
:- dynamic change_flag/0.
assert_change_flag :- assert(change_flag).

init_preferences :- retractall(change_flag),
   reset_to_defaults, check_preferences.

check_preferences :-
   %print('% '), preference_description(P,_), print(P), print(' '),
   preferences:preference_default_value(P,_),
   ((preferences:preference(P,Val),
     preferences:preference_val_type(P,T),
     preferences:preference_category(P,_))
     -> preferences:is_of_type(Val,T)
     ;  add_error(check_preferences,'Internal error for preference: ',P)),
   fail.
check_preferences. % :- print_message('Preferences Checked').


load_preferences(File) :- print_message(load_preferences(File)),
   safe_on_exception(Exception1,see(File),
       (print_message('### cannot open preference file'),print_message(File),
        print_message(Exception1),fail)),
   safe_on_exception(_Exc2,read_preferences,
       (print_message('### corrupted preference file'),print_message(File))),
    !.
    %print(loaded_preferences(File)),nl.
load_preferences(File) :- print_message('### error occured reading preference file'),
   print_message(File).

read_preferences :-  clear_dynamic_preference_predicates,
   call_cleanup(read_preferences2,seen).
read_preferences2 :-
   read(Term), Term \= end_of_file, !, %print(pref(Term)),nl,
   ((Term = preference(Pref,NVal))
      -> (volatile_preference(Pref) -> true
          ; retract(preference(Pref,_)) -> assert(preference(Pref,NVal))
          ; obsolete_preference(Pref) -> print_message(obsolete(Pref))
          ; add_message(load_preferences,'Unknown preference in preference file: ',Pref)
         )
      ;  Term = ':-'(dynamic(_)) -> true
      ;  dynamic_preference_predicate(Term) -> assert(Term)
      ;  (Term = (Head:-fail),dynamic_preference_predicate(Head)) -> true
      ;  add_error(load_preferences,'Unrecognised:',Term)
    ),
    read_preferences2.
read_preferences2 :- retractall(change_flag), recompute_nr_of_recent_documents,
                     recompute_nr_of_searched_and_replaced_patterns,recompute_nr_of_eval_history_elements.

clear_dynamic_preference_predicates :- dynamic_preference_predicate(X), retractall(X),fail.
clear_dynamic_preference_predicates.

% declare which dynamic predicates are stored in Preference file
dynamic_preference_predicate(recent_documents(_)).
dynamic_preference_predicate(searched_patterns(_)).
dynamic_preference_predicate(replaced_patterns(_)).
dynamic_preference_predicate(eval_history_elements(_)).
dynamic_preference_predicate(eval_csp_history_elements(_)).
dynamic_preference_predicate(checked_ltl_formulas(_)).

save_preferences(File) :- change_flag,!,
   % catch exception in case we cannot write preferences file
   on_exception(Exception,save_preferences2(File),
       (string_concatenate('Cannot write preferences file ',File,Msg),
        add_error(save_preferences,Msg,Exception))).
save_preferences(File) :- print_message(preferences_unchanged(File)).

save_preferences2(File) :-
   print_message(save_preferences(File)),
   tell(File),
   (print_preferences_for_saving -> true ; true),
   told,
   retractall(change_flag).

% if we have temporarily set a preference, then print the old value
print_preferences_for_saving :-
   preference_default_value(Pref,_),
   preference(Pref,CurVal),
   (temp_pref(Pref,OldVal,CurVal) -> V=OldVal ; V=CurVal),
   write_term(preference(Pref,V),[quoted(true)]),print('.'),nl,
   fail.
print_preferences_for_saving :- print_dynamic_preferences.

:- assert_must_succeed(print_preferences).
print_preferences :- print_dynamic_pred(preferences,preference,2),
   print_dynamic_preferences.

print_dynamic_preferences :-
   dynamic_preference_predicate(P), functor(P,F,N),
   print_dynamic_pred(preferences,F,N),fail.
print_dynamic_preferences.

:- assert_must_succeed(get_non_default_preferences(_)).
  
get_non_default_preferences(list(L)) :-
  findall(PS,non_default_pref(PS),L).

non_default_pref(PrefStr) :- preference(Pref,Val), preference_default_value(Pref,DVal),
   DVal \= Val,
   (eclipse_preference(ECLPREF,Pref) ->
      ajoin([Pref,' (',ECLPREF,') == ',Val,' (default = ',DVal, ')'],PrefStr)
      ; 
      ajoin([Pref,' == ',Val,' (default = ',DVal, ')'],PrefStr)
    ).

:- volatile change_flag_backup/0, pref_backup/2.
:- dynamic change_flag_backup/0.
:- dynamic pref_backup/2.

backup_preferences :- retractall(pref_backup(_,_)),retractall(change_flag_backup),
                     preference(P,Val), assert(pref_backup(P,Val)),fail.
backup_preferences :- (change_flag -> assert(change_flag_backup) ; true).

revert_preferences :- \+(pref_backup(_,_)),!, print_error('No previously backed-up preferences in revert_preferences!').
revert_preferences :- retractall(preference(_,_)),pref_backup(P,Val), assert(preference(P,Val)),fail.
revert_preferences :- retractall(change_flag), (change_flag_backup -> assert_change_flag ; true).

reset_to_defaults :- reset_to_defaults(_).

reset_to_defaults(Category) :- preference_category(Pref,Category),
   preference_default_value(Pref,Val),
   retractall(preference(Pref,_)),
   assert(preference(Pref,Val)),
   fail.
reset_to_defaults(_) :- assert_change_flag.


reset_for_benchmarking :-
   set_preference(maxNrOfInitialisations,3),
   set_preference(maxNrOfEnablingsPerOperation,10),
   set_preference(maxint,4),
   set_preference(minint,-1),
   set_preference(time_out,3500),
   set_preference(globalsets_fdrange,3),
   set_preference(show_initialisation_arguments,true),
   set_preference(show_eventb_any_arguments,false),
   %set_preference(animate_skip_operations,true),
   set_preference(cspm_animate_all_processes_without_arguments,false),
   set_preference(cspm_animate_all_processes,false).
  % set_preference(convert_closures_into_explicit_form_for_store,true).

:- meta_predicate call_with_preference(0,-,-).

call_with_preference(Call,Pref,Val) :-
    temporary_set_preference(Pref,Val,CHNG1),
    call_cleanup(Call,
                 reset_temporary_preference(Pref,CHNG1)).


% ------------------------


:- assert_must_succeed( \+((eclipse_preference(_EP,P), \+preference_default_value(P,_Val)))).
eclipse_preference('MAXINT',maxint).
eclipse_preference('MININT',minint).
eclipse_preference('DEFAULT_SETSIZE',globalsets_fdrange).
eclipse_preference('MAX_INITIALISATIONS',maxNrOfInitialisations).
eclipse_preference('MAX_OPERATIONS',maxNrOfEnablingsPerOperation).
%eclipse_preference('ANIMATE_SKIP_OPERATIONS',animate_skip_operations).
%eclipse_preference('EXPAND_CLOSURES_FOR_STATE',convert_closures_into_explicit_form_for_store).
eclipse_preference('SYMBOLIC',convert_comprehension_sets_into_closures).
eclipse_preference('CLPFD',use_clpfd_solver).
eclipse_preference('CHR',use_chr_solver).
eclipse_preference('SMT',use_smt_mode).
eclipse_preference('CSE',use_common_subexpression_elimination).
eclipse_preference('STATIC_ORDERING',use_static_ordering).
eclipse_preference('COMPRESSION',use_state_packing).
%eclipse_preference('SYMMETRY_MARKERS',use_symmetry_marker_hash).
%eclipse_preference('PERMUTATION_FLOODING',enable_permutation_reducation).
eclipse_preference('SYMMETRY_MODE',symmetry_mode).
eclipse_preference('STATIC_SYMMETRY_DETECTION',use_static_symmetry_detection).
eclipse_preference('TIME_OUT',time_out).
eclipse_preference('PROOF_INFO',use_po).
eclipse_preference('TRY_FIND_ABORT',find_abort_values). % maybe rename into TRY_FIND_WD_ERROR ?
eclipse_preference('NUMBER_OF_ANIMATED_ABSTRACTIONS',number_animated_abstractions).
eclipse_preference('USE_RECORD_CONSTRUCTION',use_record_construction).
eclipse_preference('KODKOD',try_kodkod_on_load).
eclipse_preference('MEMO',use_closure_expansion_memoization).
eclipse_preference('RANDOMISE_ENUMERATION_ORDER',randomise_enumeration_order).
eclipse_preference('PGE',pge).
eclipse_preference('OPTIMIZE_AST',optimize_ast).

eclipse_preference(A,B) :- advanced_eclipse_preference(A,B).

% these will not be shown in ProB for Rodin for example:
advanced_eclipse_preference('OPERATION_REUSE',try_operation_reuse).  % not available for Event-B this doesn't work ; see eg. prob_examples/examples/EventBPrologPackages/Advance_WP2/v6_Sep2014/ex_mch.eventb -animate 300 which generates deadlock
advanced_eclipse_preference('NORMALIZE_AST',normalize_ast).
advanced_eclipse_preference('KODKOD_ONLY_FULL',kodkod_for_components).
advanced_eclipse_preference('KODKOD_RAISE_WARNINGS',kodkod_raise_warnings).
advanced_eclipse_preference('KODKOD_MAX_NR_SOLS',kodkod_max_nr_solutions).
advanced_eclipse_preference('KODKOD_SYMMETRY',kodkod_symmetry_level).
advanced_eclipse_preference('KODKOD_SAT_SOLVER',kodkod_sat_solver).
advanced_eclipse_preference('CSE_PRED',use_common_subexpression_also_for_predicates).
advanced_eclipse_preference('CSE_WD_ONLY',use_common_subexpression_wd_only).
advanced_eclipse_preference('ALLOW_INCOMPLETE_SETUP_CONSTANTS',allow_incomplete_partial_setup_constants).
advanced_eclipse_preference('PARTITION_PROPERTIES',partition_predicates).
advanced_eclipse_preference('PARTITION_PROPERTIES_INLINE',partition_predicates_inline_equalities).
advanced_eclipse_preference('SHOW_EVENTB_ANY_VALUES',show_eventb_any_arguments).
advanced_eclipse_preference('RANDOMISE_OPERATION_ORDER',randomise_operation_order).
advanced_eclipse_preference('EXPAND_FORALL_UPTO',expand_forall_upto).
advanced_eclipse_preference('MAX_DISPLAY_SET',expand_avl_upto).
advanced_eclipse_preference('WARN_WHEN_EXPANDING_INFINITE_CLOSURES',warn_when_expanding_infinite_closures).
advanced_eclipse_preference('ENUMERATE_INFINITE_TYPES',allow_enumeration_of_infinite_types).
advanced_eclipse_preference('WARN_IF_DEFINITION_HIDES_VARIABLE',warn_if_definition_hides_variable).
advanced_eclipse_preference('TRACE_INFO',provide_trace_information).
advanced_eclipse_preference('DOUBLE_EVALUATION',double_evaluation_when_analysing).
%advanced_eclipse_preference('RECURSIVE',allow_recursive_closures).
advanced_eclipse_preference('IGNORE_HASH_COLLISIONS',ignore_hash_collisions).
advanced_eclipse_preference('FORGET_STATE_SPACE',forget_state_space).
%advanced_eclipse_preference('LARGE_PARSER_JVM',use_large_jvm_for_parser).
advanced_eclipse_preference('CSP_STRIP_SOURCE_LOC',cspm_strip_source_location).
advanced_eclipse_preference('INVARIANT_CHECKING',do_invariant_checking).
%advanced_eclipse_preference('NEGATED_INVARIANT_CHECKING',do_neginvariant_checking). % use double_evaluation_when_analysing
advanced_eclipse_preference('ALLOW_SIMULTANEOUS_ASSIGNMENTS',allow_simultaneous_assignments).
advanced_eclipse_preference('BUGLY',bugly_pp_scrambling). % B Ugly
advanced_eclipse_preference('ATELIERB_KRT_PATH',path_to_atb_krt).
advanced_eclipse_preference('JAVA_PATH',path_to_java).
advanced_eclipse_preference('TRY_ATELIERB_PROVERS',try_atb_provers).
advanced_eclipse_preference('SYMBOLIC_MC_TRY_OTHER_SOLVERS',symbolic_mc_try_other_solvers).
advanced_eclipse_preference('CSE_SUBST',use_common_subexpression_also_for_substitutions).
advanced_eclipse_preference('SMT_SUPPORTED_INTERPRETER',smt_supported_interpreter).
advanced_eclipse_preference('PP_FROZEN_INFOS',translate_print_frozen_infos).


advanced_eclipse_preference('DOT_PROP',dot_print_node_properties).
advanced_eclipse_preference('DOT_IDS',dot_print_node_ids).
advanced_eclipse_preference('DOT_INFO',dot_print_node_info).
advanced_eclipse_preference('DOT_LOOPS',dot_print_self_loops).
advanced_eclipse_preference('DOT_FUNCTIONS',dot_print_functions).
advanced_eclipse_preference('DOT_ROOT',dot_print_root).
%advanced_eclipse_preference('DOT_LEAVES',dot_print_leaves).
advanced_eclipse_preference('DOT_EDGE_LABELS',dot_print_edge_labels).
advanced_eclipse_preference('DOT_ARGUMENTS',dot_print_action_arguments).
advanced_eclipse_preference('DOT_COLOR_ARC',dot_normal_arc_colour).
advanced_eclipse_preference('DOT_COLOR_NODE',dot_normal_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_ERROR',dot_invariant_violated_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_OPEN',dot_open_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_GOAL',dot_goal_node_colour).
advanced_eclipse_preference('DOT_PEN_WIDTH',dot_node_penwidth).
advanced_eclipse_preference('DOT_SHOW_OP_READ_WRITES',dot_enabling_show_readwrites).
advanced_eclipse_preference('DOT',path_to_dot).
advanced_eclipse_preference('SFDP',path_to_sfdp).

advanced_eclipse_preference('TK_CUSTOM_STATE_VIEW_PADDING',tk_custom_state_viewer_padding).

advanced_eclipse_preference('BOOL_AS_PREDICATE',bool_expression_as_predicate).
advanced_eclipse_preference('ALLOW_UNTYPED_IDENTIFIERS',allow_untyped_identifiers).
advanced_eclipse_preference('REPL_UNICODE',repl_unicode).
advanced_eclipse_preference('REPL_CACHE_PARSING',repl_cache_parsing).
advanced_eclipse_preference('LATEX_ENCODING',latex_encoding).
advanced_eclipse_preference('XML_ENCODING',xml_encoding).
advanced_eclipse_preference('BBRESULTS',view_probcli_errors_using_bbresults).
advanced_eclipse_preference('IGNORE_PRJ_TYPES',ignore_prj_types).
advanced_eclipse_preference('FILTER_UNUSED',filter_unused_constants).
advanced_eclipse_preference('DISPROVER_MODE',disprover_mode).
advanced_eclipse_preference('TRACE_UPON_ERROR',trace_upon_error).
advanced_eclipse_preference('STRICT_RAISE_WARNINGS',strict_raise_warnings).
advanced_eclipse_preference('STRICT_RAISE_ENUM_WARNINGS',strict_raise_enum_warnings).
advanced_eclipse_preference('EDITOR',path_to_text_editor_launch).
advanced_eclipse_preference('TLC_WORKERS',tlc_number_of_workers). % will be useful once we can call TLC from probcli

advanced_eclipse_preference('INTERNAL_ARGUMENT_PREFIX',prefix_internal_operation_arguments).
advanced_eclipse_preference('MINIMAL_TEST_SUITES',generate_minimal_nr_of_testcases).


obsolete_eclipse_preference('EXPAND_CLOSURES_FOR_STATE').
obsolete_eclipse_preference('BPARSER').
obsolete_eclipse_preference('DOT_LEAVES').

print_eclipse_prefs :- eclipse_preference(X,P),
   print('     '),print(X), print(' : '),
   preference_val_type(P,VT),
   print(VT), print(' ==> '),
   preference_description(P,Desc),
   print(Desc),nl,
   fail.
print_eclipse_prefs.


:- dynamic preference/2.

% add clause in case init_preferences forgotten
preference(A,B) :- format(user_error,'Preferences not initalised (~w,~w)',[A,B]),trace,
   init_preferences,!, preference(A,B).
   %add_internal_error('Preferences not initialised: ',preference(A,B)),fail.


% ------------------

preference_default_value(maxint,3). % 2147483647
preference_default_value(minint,-1). %-2147483648
preference_default_value(number_animated_abstractions, 20).
preference_default_value(globalsets_fdrange,2).
preference_default_value(do_invariant_checking,true).
%preference_default_value(do_full_invariant_checking,true).
%preference_default_value(use_new_kernel,false).
preference_default_value(find_abort_values,false).
preference_default_value(treat_outermost_pre_as_select,true).
preference_default_value(require_operations_to_assign_all_outputs,false).
preference_default_value(allow_simultaneous_assignments,false).
preference_default_value(cspm_animate_all_processes_without_arguments,false).
preference_default_value(cspm_animate_all_processes,false).
%preference_default_value(cspm_display_inout,false).
preference_default_value(cspm_detailed_animation,false).
preference_default_value(cspm_strip_source_location,false).
%preference_default_value(csp_event_visible_to_user,false).
%preference_default_value(cspm_allow_incomplete_records,false).
%preference_default_value(findOnlyOneSol,false).
preference_default_value(maxNrOfInitialisations,4).
preference_default_value(maxNrOfEnablingsPerOperation,10).
preference_default_value(time_out,2500).
preference_default_value(debug_time_out_factor,5).
%preference_default_value(use_large_jvm_for_parser,false).
preference_default_value(show_initialisation_arguments,true).
%preference_default_value(animate_skip_operations,true).
preference_default_value(show_eventb_any_arguments,false).
preference_default_value(convert_comprehension_sets_into_closures,false).  /* <------------  was true */
preference_default_value(use_clpfd_solver,true). % Default was changed from false to true on November 28 2013; commit d26b1bf1edeb29d630cf798c0773bfa1ba71219b
preference_default_value(use_chr_solver,false).
preference_default_value(use_smt_mode,false).
preference_default_value(use_common_subexpression_elimination,false). % CSE
preference_default_value(use_common_subexpression_also_for_predicates,true). % CSE_PRED
preference_default_value(use_common_subexpression_also_for_substitutions,false). % CSE_SUBST
preference_default_value(use_common_subexpression_wd_only,false). % CSE_WD_ONLY
preference_default_value(normalize_ast,false). % NORMALIZE_AST
preference_default_value(optimize_ast,true). % OPTIMIZE_AST
preference_default_value(ltsmin_guard_splitting,false).
preference_default_value(ltsmin_do_not_evaluate_top_level_guards,false).
preference_default_value(fail_if_clpfd_timeout,false).
%preference_default_value(allow_recursive_closures,false).
preference_default_value(use_static_ordering,false).
preference_default_value(use_state_packing,false).
preference_default_value(provide_trace_information,false).
preference_default_value(partition_predicates,true).
preference_default_value(partition_predicates_inline_equalities,true).
preference_default_value(use_closure_expansion_memoization,false).
preference_default_value(allow_incomplete_partial_setup_constants,false).
preference_default_value(re_execute_operations_with_side_effects,false).
preference_default_value(warn_if_definition_hides_variable,true).
preference_default_value(warn_when_expanding_infinite_closures,5).
preference_default_value(allow_enumeration_of_infinite_types,true).
preference_default_value(strict_raise_warnings,false). % Default false as add_closure_warning otherwise will e.g. then be raised as errors; this can confuse predicate_evaluator into returning a not-well-defined result or everybody else relying ErrOcc arg of exit_error_scope
preference_default_value(strict_raise_enum_warnings,false).
preference_default_value(perform_stricter_static_checks,false).
preference_default_value(ignore_prj_types,false).
%preference_default_value(never_convert_closures_into_explicit_form,false).
%preference_default_value(convert_types_into_closures,true).
%preference_default_value(convert_closures_into_explicit_form_for_store,true). % <--- maybe remove preference
%preference_default_value(allow_empty_global_sets,false).
%preference_default_value(test_case_gen_stop_at_first_error,false).
preference_default_value(prefix_internal_operation_arguments,'$none'). %all operations arguments are external and shown in testcases
preference_default_value(generate_minimal_nr_of_testcases,true).
preference_default_value(use_tk_custom_state_viewer,true).
preference_default_value(tk_custom_state_viewer_padding,0).
preference_default_value(use_small_window,false).
preference_default_value(tk_show_source_pane,true).
preference_default_value(ignore_hash_collisions,false).
preference_default_value(forget_state_space,false).
%preference_default_value(enable_permutation_reducation,false).
%preference_default_value(use_symmetry_marker_hash,false).
%preference_default_value(enable_canonical_symmetry,false).
%preference_default_value(use_nauty_symmetry_reduction,false).
preference_default_value(symmetry_mode,off).
preference_default_value(use_static_symmetry_detection,true).
preference_default_value(bool_expression_as_predicate,false).
preference_default_value(allow_untyped_identifiers,false).
preference_default_value(repl_unicode,false).
preference_default_value(repl_cache_parsing,false).
preference_default_value(latex_encoding,auto).
preference_default_value(xml_encoding,auto).
preference_default_value(view_probcli_errors_using_bbresults,false).
preference_default_value(eventtrace,false).
preference_default_value(store_event_transinfo,false).
preference_default_value(try_kodkod_on_load,false).
preference_default_value(kodkod_for_components,true).
preference_default_value(kodkod_raise_warnings,false).
preference_default_value(kodkod_max_nr_solutions,22).
preference_default_value(kodkod_symmetry_level,0).
preference_default_value(kodkod_sat_solver,sat4j).
preference_default_value(smt_supported_interpreter,false).
preference_default_value(use_record_construction,true). % was false
preference_default_value(seen_machines_included,false).

preference_default_value(dot_print_edge_labels,true).
preference_default_value(dot_print_action_arguments,true).
preference_default_value(dot_print_functions,false).
preference_default_value(dot_print_self_loops,true).
preference_default_value(dot_print_arc_colors,true).
preference_default_value(dot_print_root,true).
%preference_default_value(dot_print_leaves,true).
preference_default_value(dot_print_node_ids,false).
preference_default_value(dot_print_use_constants,true).
preference_default_value(dot_print_node_info,true).
preference_default_value(dot_use_ps_viewer,false).
preference_default_value(dot_use_alterate_dot_viewer,false).
preference_default_value(path_to_dot,'dot').
preference_default_value(path_to_dotty,VIEWER) :-
  (possible_program(dot_viewer,E) -> VIEWER=E ; VIEWER ='dotty').
preference_default_value(path_to_dotty2,'dotty').
%preference_default_value(path_to_dotty,'/Applications/OmniGraffle/OmniGraffle.app/Contents/MacOS/OmniGraffle').
preference_default_value(path_to_sfdp,'sfdp'). % Tool useful to view large DOT files
preference_default_value(path_to_ltsmin,'prob2lts-sym'). % Tool useful to view large DOT files
preference_default_value(path_to_ps_viewer,VIEWER) :-
  (possible_program(ps_viewer,E) -> VIEWER=E ; VIEWER ='gv').
%preference_default_value(path_to_ps_viewer,'open -a /Applications/Preview.app') :- host_platform(darwin).
%preference_default_value(path_to_ps_viewer,'C:/Program Files/Ghostgum/gsview/gsview32') :- host_platform(windows).
%preference_default_value(path_to_ps_viewer,'gv') :- host_platform(linux) ; host_platform(unknown).
%preference_default_value(path_to_ps_viewer,'open -a /Applications/MacGhostViewX_Folder/MacGhostViewX.app').
preference_default_value(dot_root_shape,invtriangle).
preference_default_value(dot_node_penwidth,2).
preference_default_value(dot_normal_node_shape,box). % egg,ellipse
preference_default_value(dot_current_node_shape,doubleoctagon).
preference_default_value(dot_open_node_colour,'#F4E3C1'). %'#E88C59' '#C96D37'
preference_default_value(dot_scope_limit_node_colour,gray).
preference_default_value(dot_normal_node_colour,'#99BF38').
preference_default_value(dot_invariant_violated_node_colour,'#FF3800'). %'#FF4409'
preference_default_value(dot_fill_normal_nodes,false).
preference_default_value(dot_normal_arc_colour,'#006391').
preference_default_value(dot_node_font_size,12).
preference_default_value(dot_edge_font_size,12).
preference_default_value(dot_colour_goal_nodes,true).
preference_default_value(dot_goal_node_colour,orange).
preference_default_value(dot_print_node_properties,false).
preference_default_value(dot_counterexample_node_colour,brown).
preference_default_value(dot_counterexample_op_colour,brown).
preference_default_value(dot_hierarchy_max_ids,0).

preference_default_value(dot_event_hierarchy_horizontal,true).
preference_default_value(dot_event_hierarchy_machine_colour,gray95).
preference_default_value(dot_event_hierarchy_refines_colour,'#C0A060'). % Dark Khaki; see http://www.gpeters.com/color/color-schemes.php?search_term=Art+Nouveau
preference_default_value(dot_event_hierarchy_new_event_colour,'#C06040'). % Coral '#C08040').  % Peru
preference_default_value(dot_event_hierarchy_rename_event_colour,'#E0E0A0'). % Pale Goldenrod
preference_default_value(dot_event_hierarchy_rename_unchanged_event_colour,'#E0F0E0'). % Honeydew
preference_default_value(dot_event_hierarchy_unchanged_event_colour,gray92).
preference_default_value(dot_event_hierarchy_grd_strengthening_event_colour,gray83).
preference_default_value(dot_event_hierarchy_grd_keeping_event_colour,'#E0C080'). % Burlywood
preference_default_value(dot_event_hierarchy_edge_colour,'#806040'). % Sienna
preference_default_value(dot_event_hierarchy_extends_colour,gray50).

preference_default_value(dot_projection_non_det_edge_colour,'#806040'). % Sienna; alternative :#A06040 Light Salmon
preference_default_value(dot_projection_det_edge_colour,black).
preference_default_value(dot_projection_non_definite_edge_style,dashed).
preference_default_value(dot_projection_definite_edge_style,solid).
preference_default_value(dot_projection_label_limit,75).

preference_default_value(dot_enabling_show_readwrites,true).

preference_default_value(dot_definitions_use_sub_graph,true).
preference_default_value(dot_definitions_show_all,false).

preference_default_value(font_size,'Monaco 12') :- host_platform(darwin).
% Monaco looks better than Courier on Mac, Consolas
% Note: ctext seems to use a non-monospaced font for TkFixedFont
preference_default_value(font_size,'TkFixedFont 12') :- host_platform(X), X\==darwin. % alternative: Courier, Monaco, Consolas
preference_default_value(font_name,'Monaco') :- host_platform(darwin).
preference_default_value(font_name,'TkFixedFont') :- host_platform(X), X\==darwin. % alternative: Courier, Monaco
preference_default_value(font_size_only,12).
preference_default_value(use_font_size_for_columns,false).
preference_default_value(allow_source_code_editing,true).
preference_default_value(show_line_numbers,false).
preference_default_value(show_full_error_span,false).
preference_default_value(highlight_brackets,true).
preference_default_value(highlight_current_line,true).
preference_default_value(ask_when_content_changed,false).
preference_default_value(user_not_beginner,false).
preference_default_value(user_is_an_expert_with_accessto_source_distribution,false).
%preference_default_value(default_to_runtime_type_checking_on_startup_for_expert,false).
preference_default_value(path_to_probe,'C:/probe/probe-1.30-windows/probe.exe') :- host_platform(windows).
preference_default_value(path_to_probe,'more') :- host_platform(X), X\==windows.
preference_default_value(path_to_spin,'xspin').
preference_default_value(path_to_cspm,'/usr/bin/cspm').
preference_default_value(path_to_text_editor,EDITOR) :-
  (possible_program(editor,E) -> EDITOR=E ; EDITOR ='emacs').
preference_default_value(path_to_text_editor_launch,EDITOR) :-
  (possible_program(editor_launch,E) -> EDITOR=E ; EDITOR ='emacs').
preference_default_value(path_to_fuzz,'fuzz').
preference_default_value(path_to_fdr,'fdr2').
preference_default_value(path_to_csp_typechecker,FDR3) :-
  (possible_program(fdr3,E) -> FDR3=E ; FDR3 ='refines').
preference_default_value(path_to_bcomp,'bcomp').
preference_default_value(path_to_atb_krt,KRT) :-
  (possible_program(krt,E) -> KRT=E ; KRT ='krt').
preference_default_value(path_to_java,JJ) :-
  (possible_program(java,E) -> JJ=E ; JJ ='java').
preference_default_value(try_atb_provers,false).
preference_default_value(symbolic_mc_try_other_solvers,false).
preference_default_value(ltl_to_ba_tool,'lib/ltl2ba').

preference_default_value(number_of_recent_documents,15).
preference_default_value(number_of_searched_patterns,9).
preference_default_value(number_of_replaced_patterns,9).
preference_default_value(number_of_eval_history_elements,50).
preference_default_value(number_of_eval_csp_history_elements,50).
preference_default_value(number_of_checked_ltl_formulas,20).

preference_default_value(do_syntax_highlighting,true).
preference_default_value(sh_type_colour,darkred).
preference_default_value(sh_logical_colour,tomato).
preference_default_value(sh_assignments_colour,darkslateblue).
preference_default_value(sh_operators,'#5883be').
preference_default_value(sh_top_level_keywords,blue).
preference_default_value(sh_control_keywords,darkviolet).
%preference_default_value(sh_includes_keywords,darkslateblue).
preference_default_value(sh_comments,'#784e50').
preference_default_value(sh_pragmas,'#5b5b5b').
preference_default_value(sh_unsupported_background,yellow).

preference_default_value(expand_forall_upto,100).
preference_default_value(double_evaluation_when_analysing,true).
preference_default_value(expand_avl_upto,100).
preference_default_value(bugly_pp_scrambling,false).

preference_default_value(allow_new_ops_in_refinement,false).
preference_default_value(allow_local_operation_calls,false).
preference_default_value(use_po,true).
preference_default_value(try_operation_reuse,false).
preference_default_value(randomise_operation_order,false).
preference_default_value(randomise_enumeration_order,false).
preference_default_value(set_rand,false).
preference_default_value(rand_seed,2342452).

preference_default_value(use_cbc_analysis,true).
preference_default_value(timeout_cbc_analysis,300).

preference_default_value(pge,off).
preference_default_value(por,off).
preference_default_value(por_heur,random).
preference_default_value(use_optimistic_por,false).
preference_default_value(use_por_for_ltl,false).
preference_default_value(use_safety_ltl_model_checker,false).

preference_default_value(enable_graph,false).
preference_default_value(enable_graph_depth,1).

preference_default_value(dependency_enable_predicates,false).

preference_default_value(filter_unused_constants,false).
preference_default_value(translate_force_all_typing_infos,false).
preference_default_value(translate_print_typing_infos,false).
preference_default_value(translate_print_frozen_infos,false).
preference_default_value(translate_ids_to_parseable_format,false).
preference_default_value(translate_suppress_rodin_positions_flag,false).

/* Disprover mode - deactivates some optimizations only valid for animation */
preference_default_value(disprover_mode,false).
preference_default_value(trace_upon_error,false).
preference_default_value(tlc_number_of_workers,2).
preference_default_value(unsat_core_algorithm,divide_and_conquer).


preference_default_value(machines_path,Dir) :-
   catch(absolute_file_name(examples('.'),Dir),_,Dir='.').

% the following causes problems when loading ProB:
%:- use_module(plugins,[plugin_preference_default_value/2]).
preference_default_value(X,Y) :- plugins:plugin_preference_default_value(X,Y).


:- use_module(library(file_systems),[file_exists/1,directory_exists/1]).
possible_program(Type,E) :- host_platform(P),possible_program_aux(Type,P,E),
   (file_exists(E) -> true ; P=darwin, directory_exists(E)).

:-use_module(library(system),[environ/2]).
possible_program_aux(editor_launch,darwin,'/usr/local/bin/bbedit').
possible_program_aux(editor_launch,_,PATH) :- environ('EDITOR',PATH).
possible_program_aux(editor_launch,darwin,'/usr/local/bin/atom').
possible_program_aux(editor_launch,linux,'/usr/local/bin/atom').
possible_program_aux(editor_launch,linux,'/usr/bin/vim').
possible_program_aux(editor,darwin,'/Applications/BBEdit.app').
possible_program_aux(editor,darwin,'/Applications/Sublime Text 2.app').
possible_program_aux(editor,darwin,'/Applications/TextWrangler.app').
possible_program_aux(editor,darwin,'/Applications/TextMate.app').
possible_program_aux(editor,darwin,'/Applications/TextEdit.app').
possible_program_aux(editor,_,PATH) :- environ('EDITOR',PATH). % this is the Path in the Terminal; maybe the user wants other programs for the Tcl/Tk GUI; hence this is the last lookup
possible_program_aux(ps_viewer,darwin,'/Applications/Preview.app').
possible_program_aux(ps_viewer,windows,'C:/Program Files/Ghostgum/gsview/gsview32').
possible_program_aux(dot_viewer,darwin,'/Applications/Graphviz.app').
possible_program_aux(krt,darwin,'/Applications/AtelierB.app/AB/bbin/macosx/krt').
possible_program_aux(java,windows,'C:/Program Files/Java/jre7/bin/java.exe').
possible_program_aux(java,darwin,'/usr/bin/java').
possible_program_aux(java,linux,'/usr/bin/java').
possible_program_aux(fdr3,darwin,'/Applications/FDR3.app/Contents/MacOS/refines').

/* ------------------------- */


:- assert_must_succeed( \+((preference_description(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_description(P,_Val)))).
preference_description(maxNrOfInitialisations,'Max Number of Initialisations Computed').
preference_description(maxNrOfEnablingsPerOperation,'Max Number of Enablings per Operation Computed').
preference_description(time_out,'Time out for computing enabled transitions (in ms, is multiplied by a factor for other computations, value 2147483646 turns time outs off)').
preference_description(globalsets_fdrange,'Size of unspecified deferred sets in SETS section').
preference_description(maxint,'value of MAXINT, used for expressions such as xx::NAT (2147483647 for 4 byte ints)').
preference_description(minint,'value of MININT, used for expressions such as xx::INT (-2147483648 for 4 byte ints)'). % Atelier-B uses 2147483647
preference_description(number_animated_abstractions,'How many levels of refined models are animated by default').
%preference_description(findOnlyOneSol,'Find only one solution per Operation or Initialisation').
preference_description(debug_time_out_factor,'Debug Properties: Multiplying Factor for Timeout').
%preference_description(use_large_jvm_for_parser,'Use large Heap and Stack size for Java Parser').
preference_description(do_invariant_checking,'Check invariant while animating and model checking').
%preference_description(do_neginvariant_checking,'Check negation of invariant').
%preference_description(do_full_invariant_checking,'Check types when checking invariant').
%preference_description(use_new_kernel,'Use new Kernel (only some basic operations supported, only fore PRE and INITs)').
preference_description(find_abort_values,'Try more aggressively to detect not well-defined expressions (e.g. applying function outside of domain), may slow down animator').
preference_description(cspm_animate_all_processes,'Animate all CSP-M processes (also processes with arguments)').
preference_description(cspm_animate_all_processes_without_arguments,'Animate all CSP-M processes without arguments (not just MAIN)').
%preference_description(cspm_display_inout,'Display unsynchronised inputs/outputs as ?/!').
preference_description(cspm_detailed_animation,'Provide more fine-grained source-location animation for CSP (can slow model checking down)').
preference_description(cspm_strip_source_location,'Strip CSP source-location information (for RHS of processes, except for internal choice)').
%preference_description(csp_event_visible_to_user,'Return the CSP event as visible event to the user (option only relevant for models in \'CSP || B\' mode).').
%preference_description(cspm_allow_incomplete_records,'Allow construction of incomplete records (e.g., for Casper files)').
preference_description(use_clpfd_solver,'Use CLP(FD) solver for B integers (restricts range to -2^28..2^28-1 on 32 bit machines)').
preference_description(use_chr_solver,'Use CHR based solver. Should speed up detection of unsolvable predicates, but might slow down solving in general.').
preference_description(convert_comprehension_sets_into_closures,'Lazy expansion of lambdas and set comprehensions (SYMBOLIC)').
preference_description(use_smt_mode,'Enable SMT-Mode (aggressive treatment of : and /: inside predicates)'). % will be set automatically in CBC mode; should normally not be set by user
preference_description(use_common_subexpression_elimination,'Enable CSE (Common Subexpression Elimination)').
preference_description(use_common_subexpression_also_for_predicates,'Apply CSE (if enabled) also to predicates').
preference_description(use_common_subexpression_also_for_substitutions,'Apply CSE (if enabled) also to substitutions (operation/event bodies) [EXPERIMENTAL]').
preference_description(use_common_subexpression_wd_only,'Apply CSE (if enabled) only to well-defined sub-formulas').
preference_description(normalize_ast,'Normalize AST before passing to ProB kernel (rewrites many B predicates into simpler form)').
preference_description(optimize_ast,'Optimize AST (Abstract Syntax Tree) before passing to ProB kernel (rewrites many B predicates into more efficient form)').
preference_description(ltsmin_guard_splitting,'Detect GUARDS of operations which do not depend on parameters (for LTSMin); these guards will not be evaluated by the interpreter when ltsmin_do_not_evaluate_top_level_guards is true').
preference_description(ltsmin_do_not_evaluate_top_level_guards,'Do not evaluate top-level GUARDS detected when ltsmin_guard_splitting is true').
preference_description(fail_if_clpfd_timeout,'Assume failure if posting a CLP(FD) constraint times-out').
%preference_description(allow_recursive_closures,'Lazy expansion of *Recursive* set Comprehensions and lambdas').
preference_description(use_static_ordering,'Use static ordering to enumerate constants which occur in most PROPERTIES first').
preference_description(use_state_packing,'Use more aggressive COMPRESSION when storing states').
preference_description(provide_trace_information,'Provide various tracing information on the terminal/console.').
preference_description(partition_predicates,'Partition predicates (PROPERTIES) into components').
preference_description(partition_predicates_inline_equalities,'Inline equalities when partitioning into components').
preference_description(use_closure_expansion_memoization,'MEMOize expansions of set comprehensions and lambda abstractions.').
preference_description(allow_incomplete_partial_setup_constants,'Allow ProB to proceed even if only part of the CONSTANTS have been found.').
preference_description(re_execute_operations_with_side_effects,'Re-execute operations with side-effects in the animator (only works for the forward button at the moment).').
preference_description(warn_if_definition_hides_variable,'Warn if a DEFINITION hides a variable with the same name').
preference_description(warn_when_expanding_infinite_closures,'Warn when expanding infinite closures if MAXINT larger than: ').
preference_description(allow_enumeration_of_infinite_types,'Allow enumeration of infinite types').
preference_description(strict_raise_warnings,'Treat warnings as errors (Warning: can generate spurious NOT-WELL-DEFINED messages).').
preference_description(strict_raise_enum_warnings,'Treat ENUMERATION warnings as errors (Warning: can generate spurious NOT-WELL-DEFINED messages).').
preference_description(perform_stricter_static_checks,'Perform stricter additional static checks on formulas (automatically set by ProB).').
preference_description(ignore_prj_types,'Ignore types of prj1 and prj2 (i.e., prj1({1},{1})(2,3) is considered well-defined').

%preference_description(never_convert_closures_into_explicit_form,'Never convert closures back into explicit form (only use for debugging !)').
%preference_description(convert_types_into_closures,'Lazy expansion of type expressions (POW,-->,...)').
%preference_description(convert_closures_into_explicit_form_for_store,'Convert lazy form back into explicit form for Variables, Constants, Operation Arguments').
%preference_description(allow_empty_global_sets,'Allow empty global SETs').
%preference_description(test_case_gen_stop_at_first_error,'Stop at first error in test case generation').
preference_description(prefix_internal_operation_arguments,'Prefix of internal operation arguments (these are not included in test cases)').
preference_description(generate_minimal_nr_of_testcases,'Generate minimal suite of mcm testcases (and not one per event)').
preference_description(use_tk_custom_state_viewer,'Use custom Tk viewer to represent state of B Machine').
preference_description(tk_custom_state_viewer_padding,'Padding (in pixels) between images of custom TK state viewer').
preference_description(use_small_window,'Use a small main window for ProB (useful on small screens)').
preference_description(tk_show_source_pane,'Show the B Source Code').
preference_description(ignore_hash_collisions,'Ignore Hash Collisions (if true not all states may be computed, visited states are not memorised !)').
preference_description(forget_state_space,'Do not remember state space (mainly useful in conjunction with Ignore Hash Collisions)').
%preference_description(enable_permutation_reducation,'Enable Permutation Flooding').
%preference_description(use_symmetry_marker_hash,'Use Symmetry Hash Marker').
%preference_description(enable_canonical_symmetry,'Enable Symmetry Reduction by Canonical Form Computation').
%preference_description(use_nauty_symmetry_reduction,'Use Nauty ((c) McKay) C Library for Symmetry Reduction with Canonical Form Computation').
preference_description(symmetry_mode,'Symmetry Mode: off,flood,canon,nauty,hash').
preference_description(use_static_symmetry_detection,'Detect symmetries in forall/exists and in PROPERTIES for deferred sets').
preference_description(bool_expression_as_predicate,'Extended Predicate Syntax (allow p&q as syntactic sugar for (p=TRUE) & (q=TRUE), only possible in REPL and DEPRECATED)' ).
preference_description(allow_untyped_identifiers,'Allow UNTYPED identifiers (they will be assigned new deferred sets as type)' ).
preference_description(repl_unicode,'Use UNICODE in REPL and editing window').
preference_description(repl_cache_parsing,'Cache parsing in REPL (useful if same formula gets parsed multiple times, e.g., for Latex processing)').
preference_description(latex_encoding,'Encoding used for Latex processing with -latex (auto,ISO-8859-1,ISO-8859-2,UTF-8,UTF-16,...)').
preference_description(xml_encoding,'Encoding used for XML processing, e.g., with -logxml (auto,ISO-8859-1,ISO-8859-2,UTF-8,UTF-16,...)').
preference_description(view_probcli_errors_using_bbresults,'Show probcli errors using the BBEdit (11.6 or newer) bbresults command').

preference_description(eventtrace,'Store trace for each event (may cause large memory requirement)').
preference_description(store_event_transinfo,'Store Event-B Event-Info for each event (may cause larger memory requirement)').
preference_description(try_kodkod_on_load,'Apply translation to KODKOD on PROPERTIES when loading machines').
preference_description(kodkod_for_components,'Translate only complete components to KODKOD').
preference_description(kodkod_raise_warnings,'Raise WARNING when component or predicate cannot be translated to KODKOD').
preference_description(kodkod_max_nr_solutions,'Maximum number of solutions computed and sent by KODKOD per batch.').
preference_description(kodkod_symmetry_level,'Symmetry level for Kodkod (0=off, 20=on),  the higher this value, the more symmetries will be broken, you should probably only turn this on when MAX_INITIALISATIONS is set to 1').
%  from Kodkod API: If a non-symmetric solver is chosen for this.solver, this value controls the maximum length of the generated lex-leader symmetry breaking predicate. In general, the higher this value, the more symmetries will be broken. But setting the value too high may have the opposite effect and slow down the solving. The default value for this property is 20.
preference_description(kodkod_sat_solver,'SAT solver used for Kodkod (sat4j, minisat, lingeling, glucose)').
preference_description(smt_supported_interpreter,'Use Z3 / CVC4 for predicates inside the B interpreter').
preference_description(use_record_construction,'Records: Check if axioms/properties describe a record pattern').
preference_description(seen_machines_included,'Seen/used machines are included (no operations available) instead of extended').
preference_description(show_eventb_any_arguments,'Show top-level ANY variable values of B Operations without parameters as parameters').
preference_description(show_initialisation_arguments,'Show effect of initialisation and setup constants in operation name').
%preference_description(animate_skip_operations,'Animate operations which are skip or PRE C THEN skip').
preference_description(treat_outermost_pre_as_select,'Treat outermost PRE as SELECT').
preference_description(require_operations_to_assign_all_outputs,'Require operations to assign to all outputs').
preference_description(allow_simultaneous_assignments,'Allow B operations to assign multiple times to same variable (ASM style)').

%preference_description(check_prolog_b_ast,'Check the abstract syntax tree for B machines').
preference_description(dot_print_edge_labels,'Show Edge Labels (Operation Names)').
preference_description(dot_print_action_arguments,'Print Operation Arguments (when Edge Labels shown)').
preference_description(dot_print_functions,'Print Functions (operations that return values w/o changing state)').
preference_description(dot_print_self_loops,'Print Self-Loops').
%preference_description(dot_print_leaves,'Print Leaves').
preference_description(dot_print_arc_colors,'Colour the arcs (disable for import in OmniGraffle)').
preference_description(dot_print_root,'Print Root Node').
preference_description(dot_print_node_ids,'Print Node Identifiers').
preference_description(dot_print_node_info,'Print Node information (values of variables)').
preference_description(dot_print_use_constants,'Use constants for deferred set values if possible').
preference_description(dot_use_ps_viewer,'Use Postscript Viewer').
preference_description(dot_use_alterate_dot_viewer,'Use alternate Dot viewer').
preference_description(path_to_dot,'Path/Command for dot program (using "/" even on Windows. Try avoiding spaces in path)').
preference_description(path_to_dotty,'Path/Command for dot viewer (e.g., dotty)').
preference_description(path_to_dotty2,'Path/Command for alternate dot viewer').
preference_description(path_to_sfdp,'Path/Command for sfdp program for dot files (using "/" even on Windows. Try avoiding spaces in path)').
preference_description(path_to_ltsmin,'Path/Command to LTSmin for symbolic model checking (not available for Windows.)').
preference_description(path_to_ps_viewer,'Path/Command for Postscript viewer').
preference_description(dot_root_shape,'Shape for root node (invtriangle,triangle, ellipse, box, diamond,...)').
preference_description(dot_normal_node_shape,'Shape for normal node').
preference_description(dot_current_node_shape,'Shape for current node').
preference_description(dot_open_node_colour,'Colour for open node (red,green,blue,gray,... or rgb hex #ff00ff) or adjacent to an open node if print leaves is off').
preference_description(dot_scope_limit_node_colour,'Colour for open, but uninteresting nodes (red,green,blue,gray,... or rgb hex #ff00ff)').
preference_description(dot_normal_node_colour,'Colour for normal node').
preference_description(dot_node_penwidth,'Pen width for nodes').
preference_description(dot_invariant_violated_node_colour,'Colour for invariant violation').
preference_description(dot_fill_normal_nodes,'Fill normal nodes with color').
preference_description(dot_normal_arc_colour,'Colour for normal arcs').
preference_description(dot_node_font_size,'Font Size for Node labels').
preference_description(dot_edge_font_size,'Font Size for Edge labels').
preference_description(dot_colour_goal_nodes,'Colour the nodes that satisfy the GOAL predicate').
preference_description(dot_goal_node_colour,'Colour for nodes satisfying the GOAL predicate').
preference_description(dot_print_node_properties,'Show properties of statespace node (useful for CSP models)').
preference_description(dot_counterexample_node_colour,'Colour for counterexample nodes').
preference_description(dot_counterexample_op_colour,'Colour for counterexample transitions').
preference_description(dot_hierarchy_max_ids,'Max. number of identifiers shown in hierarchy view (-1 to show all)').

preference_description(dot_event_hierarchy_horizontal,'Organize refinements horizontally').
preference_description(dot_event_hierarchy_machine_colour,'Colour for machine event cluster').
preference_description(dot_event_hierarchy_refines_colour,'Colour for events refining themselves (i.e., not renamed and modifying guard or action)').
preference_description(dot_event_hierarchy_new_event_colour,'Colour for new events (not refining any event)').
preference_description(dot_event_hierarchy_rename_event_colour,'Colour for renamed events (with change in guard or action)').
preference_description(dot_event_hierarchy_rename_unchanged_event_colour,'Colour for renamed unchanged events').
preference_description(dot_event_hierarchy_unchanged_event_colour,'Colour for unchanged events (not renamed, no guards or actions added)').
preference_description(dot_event_hierarchy_grd_strengthening_event_colour,'Colour for events just adding a guard (not renamed, no actions added)').
preference_description(dot_event_hierarchy_grd_keeping_event_colour,'Colour for events just changing an action (not renamed, keeping guard unchanged)').
preference_description(dot_event_hierarchy_edge_colour,'Edge colour for refine relation').
preference_description(dot_event_hierarchy_extends_colour,'Edge colour for extends relation').

preference_description(dot_projection_non_det_edge_colour,'Edge colour for deterministic edges').
preference_description(dot_projection_det_edge_colour,'Edge colour for non-deterministic edges').
preference_description(dot_projection_non_definite_edge_style,'Style for non-definite edges (solid,dashed,dotted,bold,invis)').
preference_description(dot_projection_definite_edge_style,'Style for definite edges (solid,dashed,dotted,bold,invis)').
preference_description(dot_projection_label_limit,'Max. nr of characters for node labels').

preference_description(dot_enabling_show_readwrites,'Show read/write info for enabling diagrams').

preference_description(dot_definitions_use_sub_graph,'Show expression-, predicate-, substitution-DEFINITIONS as sub-graphs').
preference_description(dot_definitions_show_all,'Show all DEFINITIONS').

preference_description(font_size,'Font size in main window').
preference_description(font_name,'Font name in text editor').
preference_description(font_size_only,'Font size in text editor').
preference_description(use_font_size_for_columns,'Use font size also for columns (you need to reselect font size + Courier will be used)').
preference_description(allow_source_code_editing,'Allow editing and saving of main B Machine file').
preference_description(show_line_numbers,'Show line numbers in main window').
preference_description(show_full_error_span,'Show also the end line and column in error messages').
preference_description(highlight_brackets,'Highlights brackets').
preference_description(highlight_current_line,'Highlights current line').
preference_description(ask_when_content_changed,'Asking when file content changed from external program').
preference_description(user_not_beginner,'User is not a beginner').
preference_description(user_is_an_expert_with_accessto_source_distribution,'User is expert with access to source distribution of ProB').
%preference_description(default_to_runtime_type_checking_on_startup_for_expert,'When starting up (in expert mode) turn runtime type checking on').
preference_description(path_to_probe,'Path to Probe or other tool to view .csp files').
preference_description(path_to_spin,'Path to Spin or other tool to view .prom files').
preference_description(path_to_cspm,'Path to cspm to view and analyze .csp files.').
preference_description(path_to_text_editor,'Path to External Text Editor').
preference_description(path_to_text_editor_launch,'Path to External Text Editor for probcli (uses EDITOR environment variable by default)'). % different to path_to_text_editor, as from SICStus we cannot launch .app ??!
preference_description(path_to_fuzz,'Path to the fuzz tool for ProZ').
preference_description(path_to_fdr,'Path to the FDR2 tool').
preference_description(path_to_csp_typechecker,'Path to a CSP type checking tool, such as refines from FDR3 (taking --typecheck as argument)').
preference_description(path_to_bcomp,'Path to a bcomp (Atelier B Parser and Type Checker )').
preference_description(path_to_java,'Path to the "java" command (in case java is not found on the PATH)').
preference_description(path_to_atb_krt,'Path to Atelier B Provers krt tool').
preference_description(try_atb_provers,'Try Atelier B Provers for unknown SMTLIB specifications').
preference_description(symbolic_mc_try_other_solvers,'Symbolic Model Checking: Try different solver / settings combinations').
preference_description(number_of_recent_documents,'Number of recently opened file names memorised').
preference_description(number_of_searched_patterns,'Number of recently searched patterns in text').
preference_description(number_of_replaced_patterns,'Number of recently replaced patterns in text').
preference_description(number_of_eval_history_elements,'Number of recently runs of commands in the eval window').
preference_description(number_of_eval_csp_history_elements,'Number of recently runs of commands in the eval window').
preference_description(number_of_checked_ltl_formulas,'Number of recently checked ltl formulas').
preference_description(ltl_to_ba_tool,'Tool in lib/ directory for generatin a Buechi automaton from LTL.').

preference_description(do_syntax_highlighting,'Enable Syntax Colour Highlighting').
preference_description(sh_type_colour,'Colour of typing constructs: NAT,POW,... (red,green,... or rgb hex #ff00ff)').
preference_description(sh_logical_colour,'Colour of logical operators: &,or,...').
preference_description(sh_assignments_colour,'Colour of assignments: :=, <-- and strings').
preference_description(sh_operators,'Colour of normal operators ran,dom,...').
preference_description(sh_top_level_keywords,'Colour of top level keywords: MACHINE, INVARIANT,...').
preference_description(sh_control_keywords,'Colour of control keywords: IF, CASE,...').
%preference_description(sh_includes_keywords,'Colour of inclusion related keywords: USES,...').
preference_description(sh_comments,'Colour of comments /* ... */').
preference_description(sh_pragmas,'Colour of pragmas /*@ ... */').
preference_description(sh_unsupported_background,'Background colour for unsupported features (closure,...)').

preference_description(expand_forall_upto,'When analysing predicates: max. domain size for expansion of forall:').
preference_description(double_evaluation_when_analysing,'Evaluate PREDICATES positively and negatively when analysing:').
preference_description(expand_avl_upto,'Max size for pretty-printing sets (-1 means no limit) :').
preference_description(bugly_pp_scrambling,'Scramble identifiers when pretty-printing (BUGLY):').


preference_description(allow_new_ops_in_refinement,'Allow introduction of new operations in refinements').
preference_description(allow_local_operation_calls,'Allow to call (local) operations in same machine').


preference_description(use_po,'Use PROOF information to restrict invariant checking to affected clauses.').
%preference_description(use_po,'Restrict invariant checking to affected clauses. Also remove clauses that are proven (EventB)').
preference_description(try_operation_reuse,'Try and reuse previously computed operation effects in B/Event-B (true,false,full)').
preference_description(randomise_operation_order,'Randomise order of operations when computing successor states.').
preference_description(randomise_enumeration_order,'Randomise enumeration of variables.').
preference_description(set_rand,'Use a fix seed for random number generator. Allows to reproduce results.').
preference_description(rand_seed,'Seed for Random Number Generator').

preference_description(machines_path,'Last open directory').

% Options for enabling/disabling the POR and PGE optimisations for Model Checking of B and Event-B models
preference_description(use_cbc_analysis,'Use the constraint solver for more refined analysis of the action relations (in case of enabling/dependency determination for POR).').
preference_description(timeout_cbc_analysis,'Time out for computing independence and enable relations between operations.').

preference_description(pge,
  'Use the Partial Guard Evaluation optimisation for the state space exploration of B models.').

preference_description(use_optimistic_por,'Ignoring the timeouted action relations in the Enable graph.').
preference_description(por,'Type of Partial Order Reduction Technique (ample_sets sleep_sets).').
preference_description(por_heur,'Type of heuritic for choosing a valid ample set.').
preference_description(use_por_for_ltl,'Use Partial Order Reduction for LTL.').
preference_description(use_safety_ltl_model_checker,'Rocognise safety properties and check these using the safety property model checker').

preference_description(enable_graph,'Option for using the enable graph module for partial order reduction.').
preference_description(enable_graph_depth,'The depth of evaluating the weakest-preconditions along a path in an enable graph.').

preference_description(dependency_enable_predicates,'Compute in each state whether two events can disable each other.').

preference_description(filter_unused_constants,'Filter out unused constants').
preference_description(translate_force_all_typing_infos,'Print all typing infos when pretty-printing').
preference_description(translate_print_typing_infos,'Print needed typing infos when pretty-printing').
preference_description(translate_print_frozen_infos,'Print Prolog frozen goal infos for unbound variables').
preference_description(translate_ids_to_parseable_format,'Print Identifiers in a format that can be parsed again').
preference_description(translate_suppress_rodin_positions_flag,'Suppres Rodin position information when pretty-printing').

preference_description(disprover_mode,'Skip certain optimizations that effect soundness of proofs (for Rodin DISPROVER Plugin)').
preference_description(trace_upon_error,'Trace upon error (requires running from source)').
preference_description(tlc_number_of_workers,'Number of parallel workers for external tool TLC (for -workers option)').
preference_description(unsat_core_algorithm,'Algorithm to compute the unsat core of predicates. (divide_and_conquer or linear)').



%:- use_module(plugins,[plugin_preference_description/2]).
preference_description(X,Y) :- plugins:plugin_preference_description(X,Y).


/* ------------------------- */


%preference_update_action(do_full_invariant_checking,_) :- !, b_types_precompile.
%preference_change_requires_reload(enable_canonical_symmetry,true) :- !, .
%preference_update_action(use_nauty_symmetry_reduction,true)  :- !, graph_iso_nauty:nauty_init.
%:- use_module(probsrc(bmachine),[bmachine_is_precompiled/0]).
%:- use_module(probsrc(graph_iso_nauty),[nauty_init/0]).
preference_update_action(symmetry_mode,nauty)  :- !,
       (bmachine:bmachine_is_precompiled -> graph_iso_nauty:nauty_init ; true).
preference_update_action(_,_).

/* ------------------------- */

:- assert_must_succeed( \+((preference_val_type(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_val_type(P,_Val)))).
preference_val_type(maxint,nat1).
preference_val_type(minint,neg).
preference_val_type(number_animated_abstractions,nat).
preference_val_type(globalsets_fdrange,nat).
preference_val_type(do_invariant_checking,bool).
preference_val_type(find_abort_values,bool).
preference_val_type(treat_outermost_pre_as_select,bool).
preference_val_type(require_operations_to_assign_all_outputs,bool).
preference_val_type(allow_simultaneous_assignments,bool).
preference_val_type(cspm_animate_all_processes_without_arguments,bool).
preference_val_type(cspm_animate_all_processes,bool).
preference_val_type(cspm_detailed_animation,bool).
preference_val_type(cspm_strip_source_location,bool).
%preference_val_type(csp_event_visible_to_user,bool).
preference_val_type(show_initialisation_arguments,bool).
preference_val_type(show_eventb_any_arguments,bool).
preference_val_type(maxNrOfInitialisations,nat).
preference_val_type(maxNrOfEnablingsPerOperation,nat).
preference_val_type(time_out,nat1).
preference_val_type(debug_time_out_factor,nat1).
%preference_val_type(use_large_jvm_for_parser,bool).
preference_val_type(convert_comprehension_sets_into_closures,bool).
preference_val_type(use_clpfd_solver,bool).
preference_val_type(use_chr_solver,bool).
preference_val_type(use_smt_mode,bool).
preference_val_type(use_common_subexpression_elimination,bool).
preference_val_type(use_common_subexpression_also_for_predicates,bool).
preference_val_type(use_common_subexpression_also_for_substitutions,bool).
preference_val_type(use_common_subexpression_wd_only,bool).
preference_val_type(normalize_ast,bool).
preference_val_type(optimize_ast,bool).
preference_val_type(ltsmin_guard_splitting,bool).
preference_val_type(ltsmin_do_not_evaluate_top_level_guards,bool).
preference_val_type(fail_if_clpfd_timeout,bool).
%preference_val_type(allow_recursive_closures,bool).
preference_val_type(use_static_ordering,bool).
preference_val_type(use_state_packing,bool).
preference_val_type(provide_trace_information,bool).
preference_val_type(partition_predicates,bool).
preference_val_type(partition_predicates_inline_equalities,bool).
preference_val_type(use_closure_expansion_memoization,bool).
preference_val_type(allow_incomplete_partial_setup_constants,bool).
preference_val_type(re_execute_operations_with_side_effects,bool).
preference_val_type(warn_if_definition_hides_variable,bool).
preference_val_type(warn_when_expanding_infinite_closures,int).
preference_val_type(allow_enumeration_of_infinite_types,bool).
preference_val_type(strict_raise_warnings,bool).
preference_val_type(strict_raise_enum_warnings,bool).
preference_val_type(perform_stricter_static_checks,bool).
preference_val_type(ignore_prj_types,bool).
preference_val_type(prefix_internal_operation_arguments,string).
preference_val_type(generate_minimal_nr_of_testcases,bool).
preference_val_type(use_tk_custom_state_viewer,bool).
preference_val_type(tk_custom_state_viewer_padding,nat).
preference_val_type(use_small_window,bool).
preference_val_type(tk_show_source_pane,bool).
preference_val_type(ignore_hash_collisions,bool).
preference_val_type(forget_state_space,bool).
preference_val_type(symmetry_mode,[off,flood,nauty,hash]).
preference_val_type(use_static_symmetry_detection,bool).
preference_val_type(bool_expression_as_predicate,bool).
preference_val_type(allow_untyped_identifiers,bool).
preference_val_type(repl_unicode,bool).
preference_val_type(repl_cache_parsing,bool).
preference_val_type(latex_encoding,text_encoding).
preference_val_type(xml_encoding,text_encoding).
preference_val_type(view_probcli_errors_using_bbresults,bool).
preference_val_type(eventtrace,bool).
preference_val_type(store_event_transinfo,bool).
preference_val_type(try_kodkod_on_load,bool).
preference_val_type(kodkod_for_components,bool).
preference_val_type(kodkod_raise_warnings,bool).
preference_val_type(kodkod_max_nr_solutions,nat1).
preference_val_type(kodkod_symmetry_level,nat).
preference_val_type(kodkod_sat_solver,[sat4j,minisat,lingeling,glucose]).
preference_val_type(smt_supported_interpreter,bool).
preference_val_type(use_record_construction,bool).
preference_val_type(seen_machines_included,bool).

preference_val_type(dot_print_node_ids,bool).
preference_val_type(dot_print_use_constants,bool).
preference_val_type(dot_print_node_info,bool).
preference_val_type(dot_print_edge_labels,bool).
preference_val_type(dot_print_action_arguments,bool).
preference_val_type(dot_print_functions,bool).
preference_val_type(dot_print_self_loops,bool).
preference_val_type(dot_print_root,bool).
%preference_val_type(dot_print_leaves,bool).
preference_val_type(dot_print_arc_colors,bool).
preference_val_type(dot_use_ps_viewer,bool).
preference_val_type(dot_use_alterate_dot_viewer,bool).
preference_val_type(path_to_dot,path).
preference_val_type(path_to_dotty,path).
preference_val_type(path_to_dotty2,path).
preference_val_type(path_to_sfdp,path).
preference_val_type(path_to_ltsmin,path).
preference_val_type(path_to_ps_viewer,path).
preference_val_type(dot_root_shape,dot_shape).
preference_val_type(dot_node_penwidth,nat).
preference_val_type(dot_normal_node_shape,dot_shape).
preference_val_type(dot_current_node_shape,dot_shape).
preference_val_type(dot_open_node_colour,rgb_color).
preference_val_type(dot_scope_limit_node_colour,rgb_color).
preference_val_type(dot_normal_node_colour,rgb_color).
preference_val_type(dot_invariant_violated_node_colour,rgb_color).
preference_val_type(dot_fill_normal_nodes,bool).
preference_val_type(dot_normal_arc_colour,rgb_color).
preference_val_type(dot_node_font_size,nat1).
preference_val_type(dot_edge_font_size,nat1).
preference_val_type(dot_colour_goal_nodes,bool).
preference_val_type(dot_goal_node_colour,rgb_color).
preference_val_type(dot_print_node_properties,bool).
preference_val_type(dot_counterexample_node_colour,rgb_color).
preference_val_type(dot_counterexample_op_colour,rgb_color).
preference_val_type(dot_hierarchy_max_ids,int).

preference_val_type(dot_event_hierarchy_horizontal,bool).
preference_val_type(dot_event_hierarchy_machine_colour,rgb_color).
preference_val_type(dot_event_hierarchy_refines_colour,rgb_color).
preference_val_type(dot_event_hierarchy_new_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_rename_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_rename_unchanged_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_unchanged_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_grd_strengthening_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_grd_keeping_event_colour,rgb_color).
preference_val_type(dot_event_hierarchy_edge_colour,rgb_color).
preference_val_type(dot_event_hierarchy_extends_colour,rgb_color).

preference_val_type(dot_projection_non_det_edge_colour,rgb_color).
preference_val_type(dot_projection_det_edge_colour,rgb_color).
preference_val_type(dot_projection_non_definite_edge_style,dot_line_style).
preference_val_type(dot_projection_definite_edge_style,dot_line_style).
preference_val_type(dot_projection_label_limit,nat).

preference_val_type(dot_enabling_show_readwrites,bool).

preference_val_type(dot_definitions_use_sub_graph,bool).
preference_val_type(dot_definitions_show_all,bool).

preference_val_type(font_name,string).
preference_val_type(font_size_only,nat1).
preference_val_type(font_size,string).
preference_val_type(use_font_size_for_columns,bool).
preference_val_type(allow_source_code_editing,bool).
preference_val_type(show_line_numbers,bool).
preference_val_type(show_full_error_span,bool).
preference_val_type(highlight_brackets,bool).
preference_val_type(highlight_current_line,bool).
preference_val_type(ask_when_content_changed,bool).

preference_val_type(user_not_beginner,bool).
preference_val_type(user_is_an_expert_with_accessto_source_distribution,bool).
%preference_val_type(default_to_runtime_type_checking_on_startup_for_expert,bool).
preference_val_type(path_to_probe,path).
preference_val_type(path_to_fuzz,path).
preference_val_type(path_to_fdr,path).
preference_val_type(path_to_csp_typechecker,path).
preference_val_type(path_to_bcomp,path). % Atelier B parser & type checker
preference_val_type(path_to_java,path). % Path to Java
preference_val_type(path_to_atb_krt,path).
preference_val_type(try_atb_provers,bool).
preference_val_type(symbolic_mc_try_other_solvers,bool).
preference_val_type(path_to_spin,path).
preference_val_type(path_to_cspm,path).
preference_val_type(path_to_text_editor,path).
preference_val_type(path_to_text_editor_launch,path).
preference_val_type(number_of_recent_documents,nat).
preference_val_type(number_of_searched_patterns,nat).
preference_val_type(number_of_replaced_patterns,nat).
preference_val_type(number_of_eval_history_elements,nat).
preference_val_type(number_of_eval_csp_history_elements,nat).
preference_val_type(number_of_checked_ltl_formulas,nat).
preference_val_type(ltl_to_ba_tool,path).

preference_val_type(do_syntax_highlighting,bool).
preference_val_type(sh_type_colour,rgb_color).
preference_val_type(sh_logical_colour,rgb_color).
preference_val_type(sh_assignments_colour,rgb_color).
preference_val_type(sh_operators,rgb_color).
preference_val_type(sh_top_level_keywords,rgb_color).
preference_val_type(sh_control_keywords,rgb_color).
%preference_val_type(sh_includes_keywords,rgb_color).
preference_val_type(sh_comments,rgb_color).
preference_val_type(sh_pragmas,rgb_color).
preference_val_type(sh_unsupported_background,rgb_color).

preference_val_type(expand_forall_upto,nat).
preference_val_type(double_evaluation_when_analysing,bool).
preference_val_type(expand_avl_upto,int).
preference_val_type(bugly_pp_scrambling,bool).

preference_val_type(allow_new_ops_in_refinement,bool).
preference_val_type(allow_local_operation_calls,bool).
preference_val_type(use_po,bool).
preference_val_type(try_operation_reuse,[true,false,full]).
preference_val_type(randomise_operation_order,bool).
preference_val_type(randomise_enumeration_order,bool).

preference_val_type(set_rand,bool).
preference_val_type(rand_seed,nat).

%% Options' meaning
% disabled: skipping guard evaluation of actions known to be disabled at state
% enabled: skipping guard evaluation of actions known to be enabled at state
% full: skipping guard evaluations of actions known to be enabled or disabled at state
% disabled2, enabled2, full2: same for these, but another (more precise) way for computing the sets of enabled and disabled actions is used 
preference_val_type(pge,[off,disabled,enabled,full,disabled2,enabled2,full2]).

preference_val_type(use_cbc_analysis,bool).
preference_val_type(timeout_cbc_analysis,nat1).

preference_val_type(use_optimistic_por,bool).
preference_val_type(por,[off,ample_sets,sleep_sets]).
preference_val_type(por_heur,[first,random,minimal]).
preference_val_type(use_por_for_ltl,bool).
preference_val_type(use_safety_ltl_model_checker,bool).

preference_val_type(enable_graph,bool).
preference_val_type(enable_graph_depth,int).

preference_val_type(dependency_enable_predicates,bool).

preference_val_type(filter_unused_constants,bool).
preference_val_type(translate_force_all_typing_infos,bool).
preference_val_type(translate_print_typing_infos,bool).
preference_val_type(translate_print_frozen_infos,bool).
preference_val_type(translate_ids_to_parseable_format,bool).
preference_val_type(translate_suppress_rodin_positions_flag,bool).

preference_val_type(disprover_mode,bool).
preference_val_type(trace_upon_error,bool).
preference_val_type(tlc_number_of_workers,nat1).

preference_val_type(unsat_core_algorithm,[divide_and_conquer,linear]).


preference_val_type(machines_path,path). %directory_path). % preference not so critical; default value may not exist

%:- use_module(plugins,[plugin_preference_val_type/2]).
preference_val_type(X,Y) :- plugins:plugin_preference_val_type(X,Y).


/* ------------------------- */


:- assert_must_succeed( \+((preference_category(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_category(P,_Val)))).
preference_category(unsat_core_algorithm,advanced).
preference_category(maxint,animation).
preference_category(minint,animation).
preference_category(globalsets_fdrange,animation).
preference_category(do_invariant_checking,advanced).
%preference_category(do_full_invariant_checking,animation).
%preference_category(use_new_kernel,advanced).
preference_category(find_abort_values,animation).
preference_category(treat_outermost_pre_as_select,animation).
preference_category(require_operations_to_assign_all_outputs,animation).
preference_category(allow_simultaneous_assignments,hidden).
preference_category(cspm_animate_all_processes_without_arguments,csp_prefs).
preference_category(cspm_animate_all_processes,csp_prefs).
%preference_category(cspm_display_inout,csp_prefs).
preference_category(cspm_detailed_animation,csp_prefs).
preference_category(cspm_strip_source_location,csp_prefs).
%preference_category(csp_event_visible_to_user,csp_prefs).
%preference_category(cspm_allow_incomplete_records,csp_prefs).
%preference_category(findOnlyOneSol,advanced).
preference_category(show_eventb_any_arguments,animation).
preference_category(maxNrOfInitialisations,animation).
preference_category(maxNrOfEnablingsPerOperation,animation).
preference_category(time_out,animation).
preference_category(number_animated_abstractions,animation).
preference_category(show_initialisation_arguments,animation).
%preference_category(animate_skip_operations,animation).
preference_category(debug_time_out_factor,advanced).
%preference_category(use_large_jvm_for_parser,advanced).
preference_category(convert_comprehension_sets_into_closures,animation).
preference_category(use_clpfd_solver,animation).
preference_category(use_chr_solver,animation).
preference_category(use_smt_mode,advanced).
preference_category(use_common_subexpression_elimination,advanced).
preference_category(use_common_subexpression_also_for_predicates,advanced).
preference_category(use_common_subexpression_also_for_substitutions,advanced).
preference_category(use_common_subexpression_wd_only,advanced).
preference_category(normalize_ast,hidden).
preference_category(optimize_ast,advanced).
preference_category(ltsmin_guard_splitting,hidden).
preference_category(ltsmin_do_not_evaluate_top_level_guards,hidden).
preference_category(fail_if_clpfd_timeout,advanced).
%preference_category(allow_recursive_closures,hidden).
preference_category(use_static_ordering,advanced).
preference_category(use_state_packing,advanced).
preference_category(provide_trace_information,advanced).
preference_category(partition_predicates,advanced).
preference_category(partition_predicates_inline_equalities,advanced).
preference_category(use_closure_expansion_memoization,advanced).
preference_category(allow_incomplete_partial_setup_constants,advanced).
preference_category(re_execute_operations_with_side_effects,advanced).
preference_category(warn_if_definition_hides_variable,advanced).
preference_category(warn_when_expanding_infinite_closures,advanced).
preference_category(allow_enumeration_of_infinite_types,advanced).
preference_category(strict_raise_warnings,advanced).
preference_category(strict_raise_enum_warnings,hidden). % only for testing
preference_category(perform_stricter_static_checks,hidden). % set by prob
preference_category(ignore_prj_types,advanced).
%preference_category(never_convert_closures_into_explicit_form,hidden).
%preference_category(convert_types_into_closures,animation).
%preference_category(convert_closures_into_explicit_form_for_store,animation).
%preference_category(allow_empty_global_sets,advanced).
preference_category(ignore_hash_collisions,advanced).
preference_category(forget_state_space,advanced).
%preference_category(enable_permutation_reducation,advanced).
%preference_category(use_symmetry_marker_hash,advanced).
%preference_category(enable_canonical_symmetry,advanced).
%preference_category(use_nauty_symmetry_reduction,advanced).
preference_category(symmetry_mode,hidden).
preference_category(use_static_symmetry_detection,advanced).
preference_category(bool_expression_as_predicate,hidden).
preference_category(allow_untyped_identifiers,hidden).
preference_category(repl_unicode,syntax_highlighting).
preference_category(repl_cache_parsing,hidden).
preference_category(latex_encoding,hidden).
preference_category(xml_encoding,hidden).
preference_category(view_probcli_errors_using_bbresults,hidden).
preference_category(eventtrace,animation).
preference_category(store_event_transinfo,hidden). % automatically set by probcli
preference_category(try_kodkod_on_load,hidden).
preference_category(kodkod_for_components,advanced).
preference_category(kodkod_raise_warnings,advanced).
preference_category(kodkod_max_nr_solutions,advanced).
preference_category(kodkod_symmetry_level,advanced).
preference_category(kodkod_sat_solver,advanced).
preference_category(smt_supported_interpreter,advanced).
preference_category(use_record_construction,advanced).
preference_category(seen_machines_included,animation).

%preference_category(test_case_gen_stop_at_first_error,advanced).
preference_category(prefix_internal_operation_arguments,advanced).
preference_category(generate_minimal_nr_of_testcases,advanced).
preference_category(use_tk_custom_state_viewer,hidden). % actually : if user wants to turn it off; simply do not define ANIMATION_FUNCTION
preference_category(tk_custom_state_viewer_padding,animation).
preference_category(use_small_window,animation).
preference_category(tk_show_source_pane,animation).

preference_category(path_to_ltsmin,advanced).


preference_category(dot_print_node_ids,dot).
preference_category(dot_print_use_constants,bool).
preference_category(dot_print_node_info,dot).
preference_category(dot_print_edge_labels,dot).
preference_category(dot_print_action_arguments,dot).
preference_category(dot_print_functions,dot).
preference_category(dot_print_self_loops,dot).
preference_category(dot_print_root,dot).
%preference_category(dot_print_leaves,dot).
preference_category(dot_print_arc_colors,dot).
preference_category(dot_use_ps_viewer,hidden).
preference_category(dot_use_alterate_dot_viewer,hidden).
preference_category(path_to_dotty,dot).
preference_category(path_to_dotty2,dot).
preference_category(path_to_dot,dot).
preference_category(path_to_sfdp,dot).
preference_category(path_to_ps_viewer,dot).
preference_category(dot_root_shape,dot).
preference_category(dot_node_penwidth,dot).
preference_category(dot_normal_node_shape,dot).
preference_category(dot_current_node_shape,dot).
preference_category(dot_open_node_colour,dot).
preference_category(dot_scope_limit_node_colour,dot).
preference_category(dot_normal_node_colour,dot).
preference_category(dot_invariant_violated_node_colour,dot).
preference_category(dot_fill_normal_nodes,dot).
preference_category(dot_normal_arc_colour,dot).

%preference_category(dot_font_size,hidden).
preference_category(dot_node_font_size,dot).
preference_category(dot_edge_font_size,dot).
preference_category(dot_colour_goal_nodes,dot).
preference_category(dot_goal_node_colour,dot).
preference_category(dot_print_node_properties,dot).
preference_category(dot_counterexample_node_colour,dot).
preference_category(dot_counterexample_op_colour,dot).
preference_category(dot_hierarchy_max_ids,dot).

preference_category(dot_event_hierarchy_horizontal,dot_event_hierarchy).
preference_category(dot_event_hierarchy_machine_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_refines_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_new_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_rename_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_unchanged_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_rename_unchanged_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_grd_strengthening_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_grd_keeping_event_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_edge_colour,dot_event_hierarchy).
preference_category(dot_event_hierarchy_extends_colour,dot_event_hierarchy).

preference_category(dot_projection_non_det_edge_colour,dot_projection).
preference_category(dot_projection_det_edge_colour,dot_projection).
preference_category(dot_projection_non_definite_edge_style,dot_projection).
preference_category(dot_projection_definite_edge_style,dot_projection).
preference_category(dot_projection_label_limit,dot_projection).

preference_category(dot_enabling_show_readwrites,dot).

preference_category(dot_definitions_use_sub_graph,dot_definitions).
preference_category(dot_definitions_show_all,dot_definitions).

preference_category(font_size,gui_prefs).
preference_category(font_name,gui_prefs).
preference_category(font_size_only,gui_prefs).
preference_category(use_font_size_for_columns,advanced).
preference_category(allow_source_code_editing,gui_prefs).
preference_category(show_line_numbers,gui_prefs).
preference_category(highlight_brackets,gui_prefs).
preference_category(highlight_current_line,gui_prefs).
preference_category(ask_when_content_changed,gui_prefs).
preference_category(user_not_beginner,gui_prefs).

preference_category(user_is_an_expert_with_accessto_source_distribution,hidden).
%preference_category(default_to_runtime_type_checking_on_startup_for_expert,hidden).
preference_category(path_to_probe,csp_prefs).
preference_category(path_to_fuzz,advanced).
preference_category(path_to_fdr,csp_prefs).
preference_category(path_to_csp_typechecker,csp_prefs).
preference_category(path_to_bcomp,advanced).
preference_category(path_to_java,advanced).
preference_category(path_to_atb_krt,advanced). % not yet used from within ProB Tcl/Tk, ...
preference_category(try_atb_provers,advanced).
preference_category(symbolic_mc_try_other_solvers,advanced).
preference_category(path_to_spin,advanced).
preference_category(path_to_cspm,advanced).
preference_category(path_to_text_editor,advanced).
preference_category(path_to_text_editor_launch,hidden).
preference_category(number_of_recent_documents,advanced).
preference_category(number_of_searched_patterns,advanced).
preference_category(number_of_replaced_patterns,advanced).
preference_category(number_of_eval_history_elements,advanced).
preference_category(number_of_eval_csp_history_elements,advanced).
preference_category(number_of_checked_ltl_formulas,advanced).
preference_category(allow_new_ops_in_refinement,advanced).
preference_category(allow_local_operation_calls,advanced).
preference_category(use_po,advanced).
preference_category(try_operation_reuse,advanced).
preference_category(randomise_operation_order,advanced).
preference_category(randomise_enumeration_order,advanced).
preference_category(set_rand,advanced).
preference_category(rand_seed,advanced).
preference_category(ltl_to_ba_tool,advanced).

preference_category(do_syntax_highlighting,syntax_highlighting).
preference_category(sh_type_colour,syntax_highlighting).
preference_category(sh_logical_colour,syntax_highlighting).
preference_category(sh_assignments_colour,syntax_highlighting).
preference_category(sh_operators,syntax_highlighting).
preference_category(sh_top_level_keywords,syntax_highlighting).
preference_category(sh_control_keywords,syntax_highlighting).
%preference_category(sh_includes_keywords,syntax_highlighting).
preference_category(sh_comments,syntax_highlighting).
preference_category(sh_pragmas,syntax_highlighting).
preference_category(sh_unsupported_background,syntax_highlighting).

preference_category(expand_forall_upto,advanced).
preference_category(double_evaluation_when_analysing,advanced).
preference_category(expand_avl_upto,advanced).
preference_category(bugly_pp_scrambling,advanced).
preference_category(show_full_error_span,hidden).

preference_category(machines_path,hidden).

/* Partial techniques for optimising the Model Checker */

preference_category(pge,model_checker).

preference_category(use_cbc_analysis,advanced).
preference_category(timeout_cbc_analysis,advanced).
preference_category(use_optimistic_por,model_checker).
preference_category(por,model_checker).
preference_category(por_heur,model_checker).
preference_category(use_por_for_ltl,model_checker).
preference_category(use_safety_ltl_model_checker,model_checker).
preference_category(enable_graph,model_checker).
preference_category(enable_graph_depth,model_checker).
preference_category(dependency_enable_predicates,model_checker).

%------------------------------------

preference_category(filter_unused_constants,advanced).
preference_category(translate_force_all_typing_infos,hidden).
preference_category(translate_print_typing_infos,hidden).
preference_category(translate_print_frozen_infos,hidden).
preference_category(translate_ids_to_parseable_format,hidden).
preference_category(translate_suppress_rodin_positions_flag,hidden).

preference_category(disprover_mode,advanced).
preference_category(trace_upon_error,hidden).
preference_category(tlc_number_of_workers,hidden).


%:- use_module(plugins,[plugin_preference_category/2]).
preference_category(X,Y) :- plugins:plugin_preference_category(X,Y).



/* ------------------------- */
:- assert_must_succeed( \+((preferences:obsolete_preference(P),preference_default_value(P,_Val)))).
obsolete_preference(do_neginvariant_checking). % use double_evaluation_when_analysing
obsolete_preference(check_prolog_b_ast).
obsolete_preference(use_large_jvm_for_parser).
obsolete_preference(default_to_runtime_type_checking_on_startup_for_expert).
obsolete_preference(animate_skip_operations).
obsolete_preference(debug_try_minimise).
obsolete_preference(convert_types_into_closures).
obsolete_preference(use_avl_trees_for_sets).
obsolete_preference(dot_font_size).
obsolete_preference(only_label_base_types).
obsolete_preference(use_new_kernel).
obsolete_preference(findOnlyOneSol).
obsolete_preference(enable_permutation_reducation).
obsolete_preference(use_symmetry_marker_hash).
obsolete_preference(enable_canonical_symmetry).
obsolete_preference(use_nauty_symmetry_reduction).
obsolete_preference(do_full_invariant_checking).
obsolete_preference(test_case_gen_stop_at_first_error).
obsolete_preference(use_failure_refinement).
obsolete_preference(do_on_the_fly_refinement).
obsolete_preference(never_convert_closures_into_explicit_form).
obsolete_preference(allow_empty_global_sets).
obsolete_preference(multilevel_animation).
obsolete_preference(remove_multilevel_prefix).
obsolete_preference(cspm_allow_incomplete_records).
obsolete_preference(cspm_display_inout).
obsolete_preference(allow_recursive_closures).
obsolete_preference(absint_abstract_domain_module).
obsolete_preference(absint_use_widening).
obsolete_preference(absint_soft_widening).
obsolete_preference(absint_extrapolation_threshold).
obsolete_preference(convert_closures_into_explicit_form_for_store).
obsolete_preference(use_por).
obsolete_preference(use_pge).
obsolete_preference(pge_dis).
obsolete_preference(pge_ind).
obsolete_preference(pge_keep).
obsolete_preference(use_pge_keep).
obsolete_preference(use_pge_dis).
obsolete_preference(use_pge_ind).
obsolete_preference(selected_bparser).
obsolete_preference(dot_print_leaves).
obsolete_preference(csp_event_visible_to_user).

% might be re-included later
obsolete_preference(plugin(absint,use_widening)).
obsolete_preference(plugin(absint,soft_widening)).
obsolete_preference(plugin(absint,extrapolation_threshold)).
obsolete_preference(plugin(absint,retract_widening_related_states)).

% replaced by api calls
obsolete_preference(path_to_smt_solver).

% has been renamed
obsolete_preference(try_atb_provers_for_smtlib).

volatile_preference(store_event_transinfo). % not necessary to store in file; value will not be changed when loading pref file
% dangerous preferences: user should explicitly reset it every time or put it explicitly into file
volatile_preference(fail_if_clpfd_timeout).
volatile_preference(ignore_hash_collisions).
volatile_preference(forget_state_space).
volatile_preference(filter_unused_constants).
volatile_preference(disprover_mode).
volatile_preference(trace_upon_error).
volatile_preference(translate_force_all_typing_infos).
volatile_preference(translate_print_typing_infos).
volatile_preference(translate_ids_to_parseable_format).
volatile_preference(translate_suppress_rodin_positions_flag).

/* ------------------------- */
is_of_type(El,Type) :- type_element(Type,El).

type_element(bool,true).
type_element(bool,false).
type_element(directory_path,X) :- atomic(X),check_directory_exists(X).
type_element(int,X) :- number(X).
type_element(neg,X) :- number(X), X=<0.
type_element(nat,X) :- number(X), X>=0.
type_element(nat1,X) :- number(X), X>0.
type_element(range(Low,Up),X) :- number(X), X>=Low, X=<Up.
type_element(string,X) :- atomic(X).
type_element(path,X) :- atomic(X). % some paths are files, some directorys (.app) and some aliases
type_element(file_path,X) :- atomic(X),check_file_exists(X).
type_element(term,X) :- nonvar(X).
type_element(ListOfCsts,X) :- ListOfCsts=[_|_], member(X,ListOfCsts).
type_element(por,X) :-
   member(X,[off,ample_sets,sleep_sets]).
type_element(dot_shape,X) :- is_of_type(X,[triangle,ellipse,box,diamond,hexagon,octagon,house,
                                         invtriangle,invhouse,invtrapez,doubleoctagon,egg]).
type_element(dot_line_style,X) :- is_of_type(X,[solid, dashed, dotted, bold, invis]).
type_element(rgb_color,X) :- valid_rgb_color(X).
type_element(text_encoding,X) :-
      is_of_type(X,[auto,'ISO-8859-1','ISO-8859-2','ISO-8859-15',
                    'UTF-8','UTF-16','UTF-16LE','UTF-16BE','UTF-32','UTF-32LE','UTF-32BE',
                    'ANSI_X3.4-1968', 'windows 1252']).

valid_rgb_color(X) :- tk_color(X).
valid_rgb_color(X) :- valid_gray(X).
valid_rgb_color(X) :-  atomic(X),name(X,[35,H1,H2,H3,H4,H5,H6]),
      hex(H1),hex(H2),hex(H3),hex(H4),hex(H5),hex(H6). /* should be of the form #ff00ff */

tk_color(red).
tk_color(green).
tk_color(blue).
tk_color(yellow).
tk_color(orange).
tk_color(black).
tk_color(white).
tk_color(gray).
tk_color(brown).
tk_color(violet).
tk_color(darkred).
tk_color(tomato).
tk_color(darkblue).
tk_color(darkgray). % this is not a valid X11 color; but it is recognised by Tk (but not by sfdp)
tk_color(darkviolet).
tk_color(darkslateblue).
tk_color(lightblue).
tk_color(lightgray).
tk_color(maroon2).
tk_color(olivedrab2).
tk_color(steelblue).
tk_color(chartreuse3).
tk_color(chartreuse4).
tk_color('LightGray').
tk_color('DarkSlateGray1').
tk_color('DarkSlateGray2').
tk_color('DarkSlateGray3').
tk_color('DarkSlateGray4').
tk_color('LightSteelBlue1').
tk_color('LightSteelBlue2').
tk_color('LightSteelBlue3').
tk_color('LightSteelBlue4').
tk_color(sienna).
tk_color(firebrick).

valid_gray(G) :- atom(G), atom_codes(G,GC), append("gray",[N1|T],GC), digit(N1),
   (T=[] -> true ; T=[N2],digit(N2)).
digit(N) :- N >47, N<58.


hex(X) :- atomic(X), number(X), ((X<103,X>96) ; (X<71,X>64) ; (X>47,X<58)).

:- use_module(library(file_systems)).
check_file_exists(X) :- file_exists(X) -> true ; add_message(preferences,'Warning file for preference does not exist: ',X).
check_directory_exists(X) :-
  directory_exists(X) -> true ; add_message(preferences,'Warning directory for preference does not exist: ',X).
:- public check_path_exists/1.
check_path_exists(X) :- (file_exists(X) -> true
 ; directory_exists(X) -> true
 ; add_message(preferences,'Warning path for preference does not exist: ',X)).

:- dynamic temp_pref/3.
% use to temporarily set preferences;
% changes can be undone with reset_temporary_preference
% change will not be saved if reset_temporary_preference not called (e.g., due to bug/timeout/Ctrl-C)
temporary_set_preference(Pref,Val) :- temporary_set_preference(Pref,Val,_).
temporary_set_preference(Pref,Val,ChangeOccured) :- get_preference(Pref,OldVal),!,
   (OldVal=Val -> ChangeOccured=false
     ;  set_preference(Pref,Val), ChangeOccured = true,
        (retract(temp_pref(Pref,RealOldVal,_))  %% as we retract we cannot really nest temporary preference calls yet; TO DO fix this
           -> assert(temp_pref(Pref,RealOldVal,Val)) % store information that new value is Val
           ;  assert(temp_pref(Pref,OldVal,Val))
        )
         %,debug:debug_println(9,temp_set_pref(Pref,Val))
    ).
reset_temporary_preference(P) :- reset_temporary_preference(P,_).
reset_temporary_preference(_,ChangeOccured) :- ChangeOccured==false,!. % nothing to do
reset_temporary_preference(Pref,_) :-
    (retract(temp_pref(Pref,OldVal,NewVal))
      -> (preference(Pref,NewVal)
            -> set_preference(Pref,OldVal) %, debug:debug_println(9,reset_temp_pref(Pref,OldVal))
            ;  true /* user has explicitly changed the preference; then do not revert */
         )
     ; true).

reset_all_temporary_preferences :- temp_pref(Pref,_,_),reset_temporary_preference(Pref),fail.
reset_all_temporary_preferences.

set_preference(Pref,Val) :- var(Pref),!,
   print_error('Variable as preference not allowed'),
   print_error(set_preference(Pref,Val)).
set_preference(Pref,Val) :-
   (preference_val_type(Pref,Type)
     -> (is_of_type(Val,Type)
         -> true
         ;  print_error('New Preference value of incorrect type: '),
            print_error(Pref), print_error(Val),
             %Val =.. L, print_error(val(L)),
             %print_error(expected(Type)),
            print_error('No changes performed!'),
            fail
         )
      ; print_error('Do not know type of preference'), print_error(Pref)
   ),
   retractall(temp_pref(Pref,_,_)), % user has overriden temporary choice
   (retract(preference(Pref,OldVal)) -> New=false ; New=true),
   assert(preference(Pref,Val)),
   ((New=false,OldVal=Val) -> true ;
                   %debug:debug_println(9,update_pref(Pref,OldVal,Val)),
                   assert_change_flag,
                   preference_update_action(Pref,Val)).


tcltk_get_preference(Pref,Val) :- /* same as get_preference, but convert non-atomic into atoms */
   get_preference(Pref,V),
   (atomic(V) -> Val=V
              ;  translate_term_into_atom(V,Val)
   ).

get_preference(Pref,Val) :-
   (preference(Pref,V)
      -> Val=V
      ; (preference_default_value(Pref,V)
         -> Val=V
         ;  print_error('Invalid Preference in get_preference:'),
            print_error(Pref),
            Val = 0
         )
   ).

% rules for computed/derived preferences:
get_computed_preference(debug_time_out,Res) :-
          preference(debug_time_out_factor,TFactor), !,
          get_time_out_preference_with_factor(TFactor,Res).
get_computed_preference(convert_closures_into_explicit_form_for_store,Value) :-
          preference(convert_comprehension_sets_into_closures,SymbolicV),!,
          (SymbolicV=true -> Value=false ; Value=true).
get_computed_preference(X,_) :- print_error('Invalid Preference in get_computed_preference:'),
            print_error(X).

get_time_out_preference_with_factor(TFactor,Res) :-
          preference(time_out,CurTO),
          (disabled_time_out(CurTO)
           -> disabled_time_out(Res) % keep disabled_time_out value
           ; DTimeOut is integer(CurTO * TFactor),
             prune_time_out_preference(DTimeOut,Res)).
add_time_outs(TO1,TO2,ResTO) :-
   (disabled_time_out(TO1) -> ResTO=TO1
    ; disabled_time_out(TO2) -> ResTO=TO2
    ; ResTO is TO1+TO2
   ).

% value of 2147483646 means time-out was disabled
disabled_time_out(2147483646).
prune_time_out_preference(X,Res) :- (X>= 2147483646 -> disabled_time_out(Res) ; Res=X).

time_out_preference_disabled :- preference(time_out,2147483646).

cur_pref_string(Pref,String) :-
   preference_description(Pref,Desc),
   (eclipse_preference(ECLPREF,Pref) -> ajoin([ECLPREF,': ',Desc],String)
    ; String=Desc). %string_concatenate(Desc,';',String)).
 /*  preference(Pref,Val),
   string_concatenate(Val,';',V1),
   string_concatenate(' = ',V1,V2),
   string_concatenate(Desc,V2,String). */

   /* useful predicates for Tcl/Tk listbox interaction: */
get_preferences_list(List) :-
   findall(S,cur_pref_string(_,S),List).
get_preferences_list(Category,list(List)) :-
   findall(S,(virtual_preference_category(Pref,Category),cur_pref_string(Pref,S)),List).

% add some virtual categories like eclipse
virtual_preference_category(Pref,Category) :- Category==eclipse,!,
   findall(e(E,P),(eclipse_preference(E,P), \+ preference_category(P,hidden)),List),
   sort(List,SList),
   member(e(_,Pref),SList).
virtual_preference_category(Pref,Category) :- preference_description(Pref,_Desc), preference_category(Pref,Category).

get_ith_preference_type(Pos,Type) :- get_ith_preference_type(_,Pos,Type).
get_ith_preference_type(Category,Pos,Type) :-
    findall(S, (virtual_preference_category(Pref,Category),
                preference_val_type(Pref,S)), List),
    (nth1(Pos,List,FType) -> Type=FType
                              ; (add_error(get_ith_preference_type,'Could not find preference: ',nth1(Pos,List)),fail)).

get_ith_preference_value(Pos,Val) :- get_ith_preference_value(_,Pos,Val).
get_ith_preference_value(Category,Pos,Val) :-
    findall(S, (virtual_preference_category(Pref,Category),
                tcltk_get_preference(Pref,S)), List),
    (nth1(Pos,List,FVal) -> Val=FVal
                             ; (add_error(get_ith_preference_value,'Could not find preference: ',nth1(Pos,List)),fail)).

set_ith_preference_value(Pos,Val) :- set_ith_preference_value(_,Pos,Val).
set_ith_preference_value(Category,Pos,Val) :-
    findall(S, virtual_preference_category(S,Category), List),
    (nth1(Pos,List,Pref) -> set_preference(Pref,Val)
                             ;  add_error(set_ith_preference_value,'Could not find preference: ',nth1(Pos,List)),fail).


/* ------------------------- */

:- dynamic nr_of_recent_documents/1.
:- dynamic recent_documents/1.
nr_of_recent_documents(0).

recompute_nr_of_recent_documents :-
  retractall(nr_of_recent_documents(_)),
  findall(X,recent_documents(X),List),
  length(List,Nr),
  assert(nr_of_recent_documents(Nr)).


:- dynamic add_recent_documents_active/0.
add_recent_documents_active.

activate_recent_documents :- (add_recent_documents_active -> true ; assert(add_recent_documents_active)).

deactivate_recent_documents :- retractall(add_recent_documents_active).

/* use to add a file to recent documents list */
add_recent_document(_) :- \+ add_recent_documents_active, !.
add_recent_document(Document) :-
   (retract(nr_of_recent_documents(Nr)) -> true ; Nr=0),
   (retract(recent_documents(Document))
     -> N1 =Nr ; N1 is Nr+1
   ),
   (preference(number_of_recent_documents,Max) -> true ; Max=10),
   (N1>Max -> (retract(recent_documents(_)),assert(nr_of_recent_documents(Nr)))
           ;   assert(nr_of_recent_documents(N1))),
   assertz(recent_documents(Document)),
   assert_change_flag.

get_recent_documents(list(Res)) :-
   findall(X,recent_documents(X),List),
   reverse(List,Res).

clear_recent_documents :-
  retractall(recent_documents(_)),
  retractall(nr_of_recent_documents(_)),
  assert(nr_of_recent_documents(0)).


/* use to add a string to the searched patterns list */
:- dynamic number_of_searched_patterns/1.
:- dynamic number_of_replaced_patterns/1.
:- dynamic number_of_eval_history_elements/1.
:- dynamic number_of_eval_csp_history_elements/1.
:- dynamic searched_patterns/1.
:- dynamic replaced_patterns/1.
:- dynamic eval_history_elements/1.
:- dynamic eval_csp_history_elements/1.
:- dynamic checked_ltl_formulas/1.

number_of_searched_patterns(0).
number_of_replaced_patterns(0).
number_of_eval_history_elements(0).
number_of_eval_csp_history_elements(0).

add_element_to_history(Predicate,Pattern,MaxNumb) :-
   merge_prefix_with_suffix(number_of_,Predicate,Name),
   get_predicate_term_with_single_argument(Name,Nr,TermNum),
   (retract(TermNum) -> true ; Nr=0),
   get_predicate_term_with_single_argument(Predicate,Pattern,Term),
   (retract(Term) -> N1=Nr ; N1 is Nr+1),
   (preference(TermNum,Max) -> true ; Max=MaxNumb),
   get_predicate_term_with_single_argument(Predicate,_,Term1),
   (N1>Max -> (retract(Term1),assert(TermNum))
            ;    get_predicate_term_with_single_argument(Name,N1,TermNum1),
                 assert(TermNum1)),assertz(Term),
   assert_change_flag.

get_history_elements_list(Predicate,Reversed,list(Res)) :-
   get_predicate_term_with_single_argument(Predicate,X,Term),
   findall(X,Term,List),
   (Reversed=true -> reverse(List,Res); Res=List).


recompute_nr_of_searched_and_replaced_patterns :-
  retractall(number_of_searched_patterns(_)),retractall(number_of_replaced_patterns(_)),
  findall(X,searched_patterns(X),List),findall(Y,replaced_patterns(Y),List2),
  length(List,Nr),length(List2,Nr2),
  assert(number_of_searched_patterns(Nr)),assert(number_of_replaced_patterns(Nr2)).

recompute_nr_of_eval_history_elements :-
  retractall(number_of_eval_history_elements(_)),retractall(number_of_eval_csp_history_elements(_)),
  findall(B,eval_history_elements(B),BList),findall(C,eval_csp_history_elements(C),CList),
  length(BList,Nr1),length(CList,Nr2),assert(number_of_eval_history_elements(Nr1)),
  assert(number_of_eval_csp_history_elements(Nr2)).

/* not used
:- dynamic number_of_checked_ltl_formulas/1.
number_of_checked_ltl_formulas(0).
recompute_nr_of_checked_ltl_formulas :-
  retractall(number_of_checked_ltl_formulas(_)),
  findall(F,checked_ltl_formulas(F),List),
  length(List,Nr),assert(number_of_checked_ltl_formulas(Nr)). */

merge_prefix_with_suffix(Prefix,Ending,Name) :-
   name(Prefix,Pr),name(Ending,End),
   append(Pr,End,CodeName),name(Name,CodeName).

get_predicate_term_with_single_argument(Name,Arg,Term) :-
   functor(Term,Name,1),arg(1,Term,Arg).



/* ------------------------- */

set_preference_group(Group,_) :- \+ preference_group_description(Group,_),!,
   add_error(set_preference_mode,'Preference mode does not exist: ',Group).
set_preference_group(Group,default) :- preference_group_val(Group,Scheme,_,_),!,
   ( preference_group_val(Group,Scheme,Pref,_),preference_default_value(Pref,Val),
     set_preference(Pref,Val),fail
     ; true).
set_preference_group(Group,Scheme) :- preference_group_val(Group,Scheme,Pref,Val),
   set_preference(Pref,Val),fail.
set_preference_group(_,_).


preference_group_description(integer,'Values for MAXINT and MININT').
preference_group_description(time_out,'No TIME_OUT').
preference_group_description(dot_colors,'Colours for Dot graphs').

preference_group_val(integer,int32,maxint,2147483647). % 2147483647
preference_group_val(integer,int32,minint,-2147483648). %-2147483648

preference_group_val(time_out,disable_time_out,time_out,TO) :- disabled_time_out(TO). % this value turns time_out off, e.g., in safe_time_out

preference_group_val(dot_colors,classic,dot_normal_node_colour, olivedrab2).
preference_group_val(dot_colors,classic,dot_normal_arc_colour,steelblue).
preference_group_val(dot_colors,classic,dot_open_node_colour,red).
preference_group_val(dot_colors,classic,dot_goal_node_colour,orange).
preference_group_val(dot_colors,classic,dot_invariant_violated_node_colour,maroon2).
preference_group_val(dot_colors,classic,dot_counterexample_node_colour,brown).
preference_group_val(dot_colors,classic,dot_counterexample_op_colour,brown).

preference_group_val(dot_colors,winter,dot_normal_node_colour, '#425955').
preference_group_val(dot_colors,winter,dot_normal_arc_colour,'#1B1D26').
preference_group_val(dot_colors,winter,dot_open_node_colour,'#BFBD9F').
preference_group_val(dot_colors,winter,dot_goal_node_colour,'#CC8604').
preference_group_val(dot_colors,winter,dot_invariant_violated_node_colour,'#9C080C').

preference_group_val(dot_colors,dreams,dot_normal_node_colour, '#5D7359').
preference_group_val(dot_colors,dreams,dot_invariant_violated_node_colour,'#661C0E').
preference_group_val(dot_colors,dreams,dot_normal_arc_colour,'#8C5430').
preference_group_val(dot_colors,dreams,dot_open_node_colour,'#D6AA5C').
preference_group_val(dot_colors,dreams,dot_goal_node_colour,'#FFC800').


:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,init_preferences,
                           'Initialise Preferences').
:- register_event_listener(clear_specification,reset_all_temporary_preferences,
                           'Reset Temporary Preferences.').
:- register_event_listener(start_unit_tests,(backup_preferences,
                                            set_preference(minint,-1),
                                            set_preference(maxint,3)),
                           'Setting up Preferences for Unit Tests').
:- register_event_listener(stop_unit_tests, revert_preferences, 'Reverting Preferences').
