% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(self_check,
    [assert_pre/2, assert_post/2,pp_mnf/1, pp_cll/1, mnf/1, mnf/2,
     mnf_det/1, det_call/1, force_det_call/1,
     assert_must_succeed/1,assert_must_fail/1,
     assert_must_succeed_multiple/1, assert_must_succeed_any/1,
     must_fail/1, must_succeed/1, must_succeed_without_residue/1,
     
     run_time_type_check_mode/1,
     turn_on_run_time_type_checks/0,turn_off_run_time_type_checks/0,
     
     get_module_list/1,
   
     perform_self_check/0, perform_self_check/1,
     perform_verbose_self_check/0, perform_verbose_self_check/1,
     display_self_checks/0,

     disable_interaction_on_errors/0,
     
     % time_out_call/2, time_out_with_factor_call/3, time_out_call/1, % now in tools_timeout
     
     check_deterministic/1, check_det/1,
     
     list_skeleton/1,

     self_checks_exist/0
    
     ]).

:- use_module(debugging_calls).
%:- disable_debugging_calls.

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides predicates to define and run unit tests.').


% first code which asserts pre-, post-conditions and self-checks

% first a copy of print_error; so that this bit of self-check does not rely on other modules
my_print_error(Error) :-
	current_output(X),
	set_output(user_error),
	(var(Error) 
	  -> print_message(error,'_')
	  ;  (write('! '),write_term(Error,[max_depth(50),numbervars(true)]),nl)),
	set_output(X).

:- dynamic pre_condition/2, post_condition/2.
:- meta_predicate assert_pre(0,0).
assert_pre(X,Pre) :-  % print(adding_pre(X)),nl,
    ((nonvar(X),nonvar(Pre))
       -> (retractall(pre_condition(X,_)),assert(pre_condition(X,Pre)))
       ;  (my_print_error('### illegal variable(s) in: '),
           my_print_error(assert_pre(X,Pre)),fail)
    ).
:- meta_predicate assert_post(0,0).
assert_post(X,Post) :- 
    ((nonvar(X),nonvar(Post))
       -> (retractall(post_condition(X,_)),assert(post_condition(X,Post)))
       ;  (my_print_error('### illegal variable(s) in: '),
           my_print_error(assert_post(X,Post)),fail)
    ).

:- dynamic self_check/2.
:- dynamic self_check_module/1.
:- volatile self_check/2, self_check_module/1. % this means that self-check cannot be run in the compiled versions !

% used to check if there are self checks. if not, tcl/tk interface does not show the button
self_checks_exist :- self_check(_,_).

% has a strange behaviour; investigate :- meta_predicate self_check(:,*). 

%add_self_check(_,_) :- !. % comment in to not include self-checks in distribution
add_self_check(Module,X) :-
    (self_check_module(Module) -> true; 
                     otherwise -> assert(self_check_module(Module))),
    (nonvar(X)
                   -> (self_check(X,Module) -> true ;
                                  otherwise -> assert(self_check(X,Module))); 
         otherwise -> my_print_error('### trying to assert variable as self_check: '),my_print_error(X),fail
    ).


:- meta_predicate assert_must_succeed_any(0).
:- meta_predicate assert_must_succeed(0).
:- meta_predicate assert_must_succeed_multiple(0).
:- meta_predicate assert_must_fail(0).
%:- meta_predicate kernel_waitflags:assert_must_abort_wf(0,*).

assert_must_succeed_any(M:X) :-
    add_self_check(M,must_succeed(M:X)).
assert_must_succeed(M:X) :-
    add_self_check(M,must_succeed_without_residue(M:X)).
assert_must_succeed_multiple(M:X) :-
    add_self_check(M,must_succeed_multiple_without_residue(M:X)).
assert_must_fail(M:X) :-
    add_self_check(M,must_fail(M:X)).


% ------------------------

:- use_module(gensym). % so that self-checks can make use of this module
:- use_module(debug).
:- use_module(tools_printing).   
:- use_module(typechecker). % so that self-checks can make use of this module
:- use_module(error_manager).
:- use_module(junit_tests).
:- use_module(library(lists)).
:- use_module(tools_meta,[call_residue/2]).
:- use_module(tools,[safe_numbervars/3]).

/* Example use:

:- assert_pre(user:b_fd_type(_G,_L,_U),true).
:- assert_post(user:b_fd_type(G,L,U),(atomic(G),(integer(L),integer(U)))).

*/

% avoid dependency on tools
print_message(X) :- print_message(informational,X).


   
/* some auxilary predicates that can be used for pre,post conditions: */

% used in some pre- and post-conditions
list_skeleton(X) :- 
  (nonvar(X),list_skel2(X) -> true
              ; (my_print_error('### not list skeleton: '), my_print_error(X),fail)).
list_skel2([]).
list_skel2([_H|T]) :- nonvar(T), list_skel2(T).



:- dynamic run_time_type_check_mode/1.
run_time_type_check_mode(on).

turn_on_run_time_type_checks :- 
      retract(run_time_type_check_mode(_)),
      assert(run_time_type_check_mode(on)).
      
turn_off_run_time_type_checks :- 
      retract(run_time_type_check_mode(_)),
      assert(run_time_type_check_mode(off)).


/* ===================================================== */

:- dynamic prepost_no_error_so_far/0.

prepost_no_error_so_far. % :- fail. 
/* as errors are now displayed by error_manager,  no need for interaction ?? */

disable_interaction_on_errors :-
	retractall(prepost_no_error_so_far).


/* ===================================================== */

:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
:- meta_predicate pp_mnf(0), pp_cll(0), mnf(0), mnf(-,0), det_call(0).
:- meta_predicate mnf_call(0), mnf_call_with_pp(-,0), prepost_mnf_call(0).
:- meta_predicate prepost_call(0).

pp_mnf(X) :- prepost_mnf_call(X).
pp_cll(X) :- prepost_call(X).
mnf(X) :- mnf_call(X).
mnf(ProgramPoint,X) :- mnf_call_with_pp(ProgramPoint,X).


/* ===================================================== */


:- use_module(library(timeout)).
:- meta_predicate check_exception_call(0).
:- meta_predicate rt_timeout_call(0).

rt_timeout_call(Call) :- run_time_type_check_mode(off),!,check_exception_call(Call).
rt_timeout_call(Call) :- time_out(check_exception_call(Call),5000,TimeOutRes), 
  (TimeOutRes = success -> true ; inc_error_count, add_error(self_check,'### TIMEOUT: ',Call),fail).

% just catch exceptions and print them nicely
check_exception_call(X) :- 
   on_exception(Error, call(X),
                ( write('! EXCEPTION: '),nl,
                  write('! '),print_numbervars(X),nl,
                  write('! '),write_term(Error,[max_depth(20),numbervars(true)]),nl,nl,
                  throw(Error)) ).

print_numbervars(X) :- safe_numbervars(X,0,_), print(X),fail.
print_numbervars(_).
   
prepost_call(X) :-
	/* print(pre(X)),nl, */
	(verify_pre(X) -> true ; my_print_error(verify_pre_failed(X))),
	rt_timeout_call(X),
	/* print(post(X)),nl, */
	(verify_post(X) -> true ; my_print_error(verify_post_failed(X))).

verify_pre(Call) :-
  (run_time_type_check_mode(off) -> true
   ;
	(pre_condition(Call,Pre)
	->( (\+(rt_timeout_call(Pre))
		-> (get_predicate_arity(Call,Pred,Arity),
		    add_error(verify_pre,'### PRE-CONDITION ERROR OCCURRED: ',Pred/Arity),
	   	    add_error(verify_pre,'### CALL: ',Call),nl,
            prepost_user_interaction
		   )
		;  (true)
	     )
	  )
	; (get_predicate_arity(Call,Pred,Arity),
	   add_error(verify_pre,'### No PRE-CONDITION for ',Pred/Arity),
	   print_term_summary(Call),
	   prepost_user_interaction )
	)).

verify_post(Call) :-
  (run_time_type_check_mode(off) -> true
   ;
	(post_condition(Call,Post)
	->( (\+(rt_timeout_call(Post))
		-> (get_predicate_arity(Call,Pred,Arity),
		    add_error(verify_post,'### POST-CONDITION ERROR OCCURRED: ',Pred/Arity),
	   	    add_error(verify_post,'### CALL: ',Call),nl,
                    prepost_user_interaction
		   )
		;  (true)
	     )
	  )
	; (get_predicate_arity(Call,Pred,Arity),
	   add_error(verify_post,'### No POST-CONDITION for ',Pred/Arity),
	   prepost_user_interaction )
	)).


prepost_user_interaction :- prepost_no_error_so_far,!,
       my_print_error('### => Stop at next error (y/n/halt/trace) => '),
       read(Answer),
       (Answer='y' -> true 
        ; Answer='halt' -> halt
        ; Answer='trace' -> trace
        ; retract(prepost_no_error_so_far)).
prepost_user_interaction.


get_predicate_arity(':'(Module,Call),Res,Arity) :- !,Res=Module:Pred,
   functor(Call,Pred,Arity).
get_predicate_arity(Call,Pred,Arity) :-
   functor(Call,Pred,Arity).
   
/* ===================================================== */

prepost_mnf_call(X) :- % print(prepost_mnf_call(X)),nl,
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(prepost_call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',X),
	     print_call(X),
         prepost_user_interaction,
         fail)
       )
  ).
	
mnf_call(X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',X),
	     print_call(X),
         prepost_user_interaction,
         fail)
       )
  ).
	
print_call(X) :-
	     print('### Call: '),print_quoted_with_max_depth(X,10),nl,
	     print('### Summary: '),print_term_summary(X),nl,
	     print('### Predicate: '), 
	     (X = ':'(M,C) -> (print(M),print(':')) ; C=X),
	     functor(C,F,N),
	     print(F), print('/'), print(N),nl.
 
	
mnf_call_with_pp(ProgramPoint,X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(rt_timeout_call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',(X,ProgramPoint)),
	     print_call(X),
	     print('### at program point:'),print(ProgramPoint),nl,
         prepost_user_interaction,
         fail)
       )
  ).

:- volatile found_det_sol/4.
:- dynamic found_det_sol/4.
:- meta_predicate mnf_det(0).

mnf_det(X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  (get_functor_module(X,Module,F,Arity),
       retractall(found_det_sol(F,Arity,Module,_)),  /* cannot be nested for same predicate !*/
       copy_term(X,CopyX),
       mnf_call(X),
       (found_det_sol(F,Arity,Module,Previous) ->
          (add_error(mnf,'### WARNING CALL HAS MULTIPLE SOLUTIONS: ',CopyX),nl,
	       print('### Solution 1: '),print_quoted_with_max_depth(Previous,60),nl,nl,
	       print('### Solution 2: '),print_quoted_with_max_depth(X,60),nl,nl,
           prepost_user_interaction)
         ; assert(found_det_sol(F,Arity,Module,X))
       )
      )
  ).
  
  % det_call(';'(X=a,X=b)).
get_functor_module(Module:Call,Module,F,Arity) :- !, functor(Call,F,Arity).
get_functor_module(Call,unknown,F,Arity) :- functor(Call,F,Arity).

:-  meta_predicate force_det_call(0).
det_call(X) :- 
	  (run_time_type_check_mode(off)
	   -> call(X)
	   ;  force_det_call(X)
	  ).
  
:- meta_predicate residue_check_call(0).
residue_check_call(X) :-
        copy_term(X,CX),
		call_residue(X,CallResidue),
		(CallResidue=[] -> true
		  ;   add_error(must_succeed_without_residue,'### Call has residue: ',X),
			  add_error(must_succeed_without_residue,'### Residue: ',CallResidue),
			  safe_numbervars(CX,0,_),
			  format('### Original call: ~w~n',[CX])
		).

force_det_call(X) :- 
       get_functor_module(X,Module,F,Arity),
       retractall(found_det_sol(F,Arity,Module,_)),  /* cannot be nested for same predicate !*/
       copy_term(X,CopyX),
       %print(det_call(X)),nl,
       residue_check_call(X), %  was rt_timeout_call(X)
       (found_det_sol(F,Arity,Module,Previous) ->
          (add_error(mnf,'### WARNING CALL HAS MULTIPLE SOLUTIONS: ',CopyX),nl,
	       print('### Solution 1: '),print_quoted_with_max_depth(Previous,60),nl,nl,
	       print('### Solution 2: '),print_quoted_with_max_depth(X,60),nl,nl,
           prepost_user_interaction)
         ; assert(found_det_sol(F,Arity,Module,X))
       ).

/* ===================================================== */

:- meta_predicate must_fail(0).
must_fail(X) :-
	copy_term(X,Y),
	rt_timeout_call(X),!,
	(safe_numbervars(Y,0,_) -> true ; true),
	inc_error_count,
	add_error(must_fail,'### Unit Test Failed !!! Call succeeded: ',Y),
	my_print_error('### The call: '),
	my_print_error(Y),
	my_print_error('### should have failed but succeeded with:'),
	my_print_error(X).
must_fail(_X).


:- meta_predicate must_succeed(0).

must_succeed(X) :-  %print(must_suceed(X)),nl,
	\+ rt_timeout_call(X),!,
	(safe_numbervars(X,0,_) -> true ; true),
	inc_error_count,
	add_error(must_succeed,'### Unit Test Failed !!! Call failed: ',X).
%	my_print_error('### The call: '),
%	my_print_error(X),
%	my_print_error('### should have succeeded but failed !').
must_succeed(_X). %  :- print(ok_must_succeed(_X)),nl.

:- volatile found_must_succeed_sol/1, found_id/1.
:- dynamic found_must_succeed_sol/1, found_id/1.
found_id(0).
get_found_id(Nr) :- retract(found_id(Nr)), N1 is Nr+1, assert(found_id(N1)).
reset_found_id :- retractall(found_must_succeed_sol(_)),
 retractall(found_id(_)), assert(found_id(0)).

:- meta_predicate must_succeed_without_residue(0).
must_succeed_without_residue(X) :- get_found_id(ID),
	must_succeed(residue_check_call(X)),
	retractall(found_must_succeed_sol(ID)),
	%residue_check_call(X),
	(found_must_succeed_sol(ID)
	  -> (add_error(must_succeed_without_residue,'### Self-Check has multiple solutions: ',X),
	      inc_error_count,
	      !)
	  ;  assert(found_must_succeed_sol(ID)),fail
	 ).
must_succeed_without_residue(_).

:- meta_predicate must_succeed_multiple_without_residue(0).
:- public must_succeed_multiple_without_residue/1. % used by assert_must_succeed_multiple
must_succeed_multiple_without_residue(X) :- get_found_id(ID),
	must_succeed(X),
	retractall(found_must_succeed_sol(ID)),
	call_residue(X,CallResidue),
	(CallResidue=[] -> true
	  ;   add_error(must_succeed_multiple_without_residue,'### Self-Check has residue: ',X),
	      add_error(must_succeed_multiple_without_residue,'### Residue: ',CallResidue),
	      inc_error_count
	),
	(found_must_succeed_sol(ID)
	  -> (true,!)
	  ;  assert(found_must_succeed_sol(ID)),fail
	 ).
must_succeed_multiple_without_residue(X) :- 
	inc_error_count,
    add_error(must_succeed_multiple_without_residue,
              '### Self-Check did not succeed multiple times: ',X).


:- meta_predicate safe_call(0).

safe_call(X) :- on_exception(Exception,
                call(X),
                (print(exception(X,Exception)),nl,nl,
	             inc_error_count,
                 add_error(safe_call,'### Exception occurred during self-check: ',X:Exception))).

:- dynamic starttime/1.

get_module_list(ML) :- findall(Module,self_check_module(Module),Modules), sort(Modules,ML).

perform_self_check :- perform_self_check(_).



:- dynamic errors_in_module/2, current_module_under_test/1.
set_current_module(M) :- retractall(current_module_under_test(_)),
   assert(current_module_under_test(M)).
inc_error_count :- current_module_under_test(Module),!,
   (retract(errors_in_module(Module,Nr)) -> true ; Nr=0),
   N1 is Nr+1,
   assert(errors_in_module(Module,N1)).
inc_error_count :- print('No current module.'),nl.
print_error_summary :- errors_in_module(Module,Nr), format('Module: ~w errors: ~w.~n',[Module,Nr]), fail.
print_error_summary.

check_module(Module,Results) :- 
   findall(testcase(Module,X), self_check(X,Module), Calls),
   maplist(check_test_case, Calls, Results).

check_test_case(testcase(Module,Call), Verdict) :-
   set_error_context(self_check(Module,Call)),
   statistics(runtime,_),
   print('.'),
   safe_call(Call),
   flush_output,
   statistics(runtime,[_,T]),
   (T>20 -> print('*'),print(T),print('*') ; true),
   (T>2000 -> nl, print(Module), print(' : '), print(Call),nl ; true),
   (junit_mode(_) -> (get_all_errors_and_reset(Errors) -> V=error(Errors)
                                           ; otherwise -> V=pass)
   ; otherwise -> true),
   create_junit_result(Call,'Selfcheck', Module, T, V, Verdict).

:- use_module(eventhandling,[announce_event/1]).
perform_self_check(Module) :- nl, reset_found_id,
    retractall(errors_in_module(_,_)),
    ( nonvar(Module) -> true;
           otherwise -> self_check_module(Module)),
   set_current_module(Module),
   nl,
   print_message('---------------------'),
   print_message(('Performing Self-Check for Module ':Module)),
   print_message('---------------------'),
   announce_event(start_unit_tests),
   flush_output,
   print('% '),
   retractall(starttime(_)),
   statistics(runtime,[Start,_]), assert(starttime(Start)),
   check_module(Module,TestResults),
   print_junit(TestResults,Module),
   fail.
perform_self_check(_) :- 
   statistics(runtime,[End,_]),
   starttime(Start), Tot is End-Start,
   nl, print('Time: '), print(Tot), print(' ms'),
   nl,
   announce_event(stop_unit_tests),
   clear_error_context,
   (error_manager:test_error_occurred(Source)
     -> print_message('Self-Check FAILED !!!'),
        format('% Source ~w ~n',[Source]),
        print_error_summary,
        fail
     ;  true),
   print_message('Self-Check Successful'),
   print_message('-------------------'), 
   flush_output.
   
   
perform_verbose_self_check :- perform_verbose_self_check(_).

perform_verbose_self_check(Module) :- nl, reset_found_id,
( nonvar(Module) -> true;
          otherwise -> self_check_module(Module)),
  nl,
  print_message('---------------------'),
  print_message(('Performing Self-Check for Module ',Module)),
  flush_output,
   self_check(X,Module),
   print(X), flush_output,
   set_error_context(self_check(Module,X)),
   statistics(runtime,_),
   safe_call(X),
   statistics(runtime,[_,T]),
   print(' ---> '), print(T), print(' ms'),nl,
   ((T>20) -> (print_message(warning,'### LONG !'),nl) ; true),
   fail.

perform_verbose_self_check(_) :- nl,
   (error_manager:test_error_occurred(_)
     -> print_message('Self-Check FAILED !!!'),fail
     ;  true),
   print_message('Self-Check Successful'),
   print_message('---------------------'),
   flush_output.

display_self_checks :- display_self_checks(_).

/* self_check:display_self_checks */
display_self_checks(Module) :- 
( nonvar(Module) -> true;
          otherwise -> self_check_module(Module)),
  nl,print_message('---------------------'),
  print_message(('Performing Self-Check for Module ',Module)),
flush_output,
   self_check(X,Module),
   print_message(X),
   fail.
display_self_checks(_) :- nl,
   print_message('---------------------'), flush_output.

% ------------------------
:- meta_predicate check_deterministic(0).
:- meta_predicate check_det(0).
:- meta_predicate check_det2(0,-).

:- dynamic det_counter/1.
det_counter(0).
check_deterministic(Call) :-
   retract(det_counter(X)), X1 is X+1, assert(det_counter(X1)),
   check_det2(Call,X).
check_det(Call) :- check_deterministic(Call).

:- volatile calling/1.
:- dynamic calling/1.
check_det2(Call,X) :- assert(calling(X)),
  call(Call),
  (retract(calling(X)) -> true
    ; nl,
      print('### Call has multiple solutions: '),nl,
      print('### '), print(Call),nl,
      nl
   ).
check_det2(Call,X) :- (retract(calling(X)) -> print(fails(Call)),nl ; fail).
