% (c) 2015-2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(smt_solvers_interface, [smt_solve_predicate/4,
                                  smt_is_contradiction/1,smt_is_contradiction/2,
                                  smt_add_predicate/5,
                                  init/0]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,smt_solvers).
:- module_info(description,'This module provides an interface to SMT-solvers like CVC4 or Z3.').

:- use_module(library(lists), [maplist/2,maplist/3,maplist/4,
                               append/2,exclude/3,
                               nth0/3,some/2]).
:- use_module(library(sets), [list_to_set/2]).

:- use_module(probsrc(bsyntaxtree), [map_over_typed_bexpr/3,
                            find_typed_identifier_uses/2,
                            find_typed_identifier_uses/3,
                            get_texpr_type/2,
                            get_texpr_info/2,
                            get_texpr_expr/2,
                            get_texpr_id/2,
                            conjunction_to_list/2,
                            conjunct_predicates/2,
                            disjunction_to_list/2,
                            syntaxtraversion/6]).
:- use_module(probsrc(error_manager), [add_error_fail/3]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(preferences), [get_preference/2]).
:- use_module(probsrc(b_global_sets), [b_global_deferred_set/1,
                                       unfixed_deferred_set/1,
                                       b_get_global_enumerated_sets/1,
                                       is_b_global_constant/3,
                                       b_global_set_cardinality/2]).
:- use_module(probsrc(custom_explicit_sets), [expand_custom_set_to_list_now/2]).
:- use_module(probsrc(eventhandling),[register_event_listener/3]).
:- use_module(probsrc(external_functions), [is_external_function_name/1]).
:- use_module(probsrc(debug), [debug_format/3]).

:- use_module(model_translation, [translate_smt_model_to_b_values/4]).
:- use_module(ast_cleanup_for_smt, [clean_up/2]).
:- use_module(prob_state_predicates, [create_state_pred/4]).
:- use_module(solver_dispatcher).
:- use_module(smt_common_predicates).

init :-
    preferences:get_preference(smt_supported_interpreter,true), !,
    register_event_listener(start_solving,retractall(backjumping_symbols(_)),'clean smt backjumping information'),
    register_event_listener(start_solving,init_interface(z3),'Init Z3 Interface for a new solving'),
    register_event_listener(end_solving,reset_interface(z3),'Reset Z3 Interface after solving').
init.

:- block smt_add_predicate(-,?,?,?,?).
smt_add_predicate(_EnumWf,Pred,ProBLocalState,ProBState,Symbol) :-
    backjump_on_symbol(Symbol),
    smt_add_predicate_to_solver(z3,Pred,ProBLocalState,ProBState,Symbol).

smt_add_predicate_to_solver(Solver,Pred,ProBLocalState,ProBState,Symbol) :-
    init_interface(Solver),
    smt_solver_interface_call(Solver,push_frame),
    clean_up(Pred,CPred),
    find_typed_identifier_uses(CPred,[],Identifiers),
    (contains_introduced_identifiers(Identifiers,CPred)
     -> debug_format(9,'z3: skip: ~w~n because there were introduced identifiers', [Symbol])
     ;  create_state_pred(Identifiers,ProBLocalState,ProBState,StatePred),
        conjunction_to_list(StatePred,StatePredConjuncts),
        exclude(contains_unsupported_type_or_expression(Solver),StatePredConjuncts,FilteredStatePredConjuncts),
        conjunct_predicates(FilteredStatePredConjuncts,FilteredStatePred),
        conjunct_predicates([FilteredStatePred,CPred],FullPred),
        setup_sets(Solver,GlobalIdentifiers),
        create_smt_state(Solver,Identifiers,IdentifierState,_),
        append(IdentifierState,GlobalIdentifiers,GlobalState),
        mk_expr_skip(Solver,FullPred,[],GlobalState,Expr),
        translate_bexpression(FullPred,PPExpr),
        debug_format(9,'z3: add expr: ~w using symbol: ~w~n', [PPExpr,Symbol]),
        !, smt_add_predicate_aux(Solver,Expr,Symbol)).
smt_add_predicate_to_solver(Solver,_,_,_,_) :- % in case the predicate contains unsupported types, etc.
    smt_solver_interface_call(Solver,pop_frame).

contains_introduced_identifiers(FreeIDs,TExpr) :-
    syntaxtraversion(TExpr,_,Type,Infos,Subs,_),
    (member(introduced_by(_),Infos) % identifier that has been introduced
     -> % check if it is not bound by some quantifiers
        get_texpr_id(TExpr,Identifier),
        member(b(identifier(Identifier),Type,_),FreeIDs)
     ;  some(contains_introduced_identifiers(FreeIDs),Subs)).

smt_add_predicate_aux(Solver,Expr,Symbol) :-
    get_preference(time_out,TO),
    QuickTO is TO // 100,
    % smt solver is called multiple times anyway
    % and keeps its state inbetween
    % thus we can use only a fraction of the time_out
    % after e.g. 100 predicates, the whole timeout will be spent.
    smt_solver_interface_call(Solver,add_interpreter_constraint(Expr,Symbol,QuickTO,SMTResult)),
    (SMTResult = unsat(Core) -> % inconsistency detected
        debug_format(9,'z3: false with core: ~w~n',[Core]),
        store_core_for_backjumping(Core),
         fail ;
     otherwise            -> true).
smt_add_predicate_aux(Solver,_,_) :-
    % remove frame upon backtracking
    smt_solver_interface_call(Solver,pop_frame), fail.

:- dynamic backjumping_symbols/1.
store_core_for_backjumping(Core) :-
    retractall(backjumping_symbols(_)),
    cleanup_core_for_backjumping(Core,CleanCore),
    assert(backjumping_symbols(CleanCore)).
cleanup_core_for_backjumping([],[]).
cleanup_core_for_backjumping([E|Es],[CE|CEs]) :-
    tools:split_atom(E,['|'],[CE]),
    cleanup_core_for_backjumping(Es,CEs).

backjump_on_symbol(Symbol) :-
    backjumping_symbols(Symbols),
    member(Symbol,Symbols),
    debug_format(9,'smt check for backjump:~ncurrent symbol: ~w~nbackjumping symbols: ~w~n',[Symbol,Symbols]),
    !, debug_format(9,'backjumping!',[]), fail.
backjump_on_symbol(_) :-
    retractall(backjumping_symbols(_)).

smt_solve_predicate(Solver,Pred,State,Result) :-
    clean_up(Pred,CPred), !,
    smt_solve_predicate_aux(Solver,CPred,State,Result).

smt_solve_predicate_aux(Solver,Pred,_,Result) :-
    contains_unsupported_type_or_expression(Solver,Pred,Un), !,
    Result = no_solution_found(unsupported_type_or_expression(Un)).
smt_solve_predicate_aux(Solver,Pred,ProBState,Result) :-
    reset_interface(Solver) ->
        find_typed_identifier_uses(Pred,Identifiers),

        create_state_pred(Identifiers,[],ProBState,StatePred),
        conjunction_to_list(StatePred,StatePredConjuncts),
        exclude(contains_unsupported_type_or_expression(Solver),StatePredConjuncts,FilteredStatePredConjuncts),
        conjunct_predicates(FilteredStatePredConjuncts,FilteredStatePred),
        conjunct_predicates([FilteredStatePred,Pred],FullPred),

        setup_sets(Solver,GlobalIdentifiers),
        create_smt_state(Solver,Identifiers,IdentifierState,_),
        append(IdentifierState,GlobalIdentifiers,GlobalState),

        % trace,
        mk_expr(FullPred,Solver,[],GlobalState,PExpr),

        translate_bexpression(FullPred,PPPred),
        debug_format(9,'Sending to ~w:~w~n',[Solver,PPPred]),
        flush_output,

        !, % all preparations done, cut before calling solver
           % to avoid backtracking into setup and finding solutions twice
        catch(check_sat(Solver,PExpr,ProBState,GlobalState,Identifiers,Result),
              model_translation_failed, Result = no_solution_found(model_translation_failed))
    ; Result = no_solution_found(solver_not_available).
smt_solve_predicate_aux(_,_Pred,_,no_solution_found(translation_or_setup_failed)).

smt_is_contradiction(Pred) :-
    clean_up(Pred,CPred), !,
    (smt_is_contradiction(cvc4,CPred) ;
     smt_is_contradiction(z3,CPred)).
smt_is_contradiction(Solver,CPred) :-
    reset_interface(Solver), !,
    find_typed_identifier_uses(CPred,Identifiers),
    setup_sets(Solver,GlobalIdentifiers),
    create_smt_state(Solver,Identifiers,IdentifierState,_),
    append(IdentifierState,GlobalIdentifiers,GlobalState),
    mk_expr_skip(Solver,CPred,[],GlobalState,Expr),
    !, % all preparations done, cut before calling cvc4
       % to avoid backtracking into setup and finding solutions twice
    catch(smt_solver_interface_call(Solver,query(Expr,unsat)),_, fail).

mk_expr_skip(Solver,Pred,[],GlobalState,Expr) :-
    clean_up(Pred,CPred), !,
    conjunction_to_list(CPred,List),
    exclude(contains_unsupported_type_or_expression(Solver),List,NList),
    conjunct_predicates(NList,NewCPred),
    mk_expr(NewCPred,Solver,[],GlobalState,Expr).
mk_expr_skip(_,Pred,_,_,_) :- !,
    add_error_fail(smt_solvers_interface,'mk_expr_skip failed',Pred).

check_sat(Solver,Expr,ProBState,State,Identifiers,Result) :-
    catch(check_sat_aux(Solver,Expr,ProBState,State,Identifiers,Result),
          SolverException,
          treat_exception(Solver,SolverException,Result)).
treat_exception(Solver,Exc,Result) :- known_solver_exception(Exc,Reason),!,
    reset_interface(Solver),Result = no_solution_found(Reason).
treat_exception(_,Exc,_) :- throw(Exc).
known_solver_exception(z3_exception(Reason),Reason).
known_solver_exception(logic_exception(Reason),Reason).

check_sat_aux(Solver,Expr,ProBState,State,Identifiers,Result) :-
    get_preference(time_out,TO),
    smt_solver_interface_call(Solver,query(Expr,TO,SMTResult)),
    get_model_or_return_result(SMTResult,Solver,Expr,ProBState,State,Identifiers,Result).

get_model_or_return_result(unsat,_,_,_,_,_,contradiction_found).
get_model_or_return_result(unknown,_,_,_,_,_,no_solution_found(solver_answered_unknown)).
get_model_or_return_result(sat,Solver,_,ProBState,State,Identifiers,solution(Bindings)) :-
    exclude(is_smt_temp_var,Identifiers,FilteredIDs),
    (translate_smt_model_to_b_values(Solver,State,FilteredIDs,Bindings)
     -> maplist(bind_in_prob_state(Bindings),ProBState)
     ; throw(model_translation_failed)).
get_model_or_return_result(sat,Solver,Expr,ProBState,State,Identifiers,Result) :-
    next_result(Solver,Expr,ProBState,State,Identifiers,Result).

bind_in_prob_state(Bindings,bind(Id,Val)) :-
    member(binding(Id,Val,_),Bindings).

% no global identifiers -> no other solution
next_result(_,_,_,_,[],Result) :- !,
    Result = contradiction_found.
next_result(Solver,Expr,ProBState,State,Identifiers,Result) :-
    findall(Id,member(id(_,Id),State),ExprIds),
    smt_solver_interface_call(Solver,conjoin_negated_state(ExprIds,Expr,NewExpr)),
    check_sat(Solver,NewExpr,ProBState,State,Identifiers,Result).

unary_operator(negation).
unary_operator(unary_minus).
binary_operator(equivalence).
binary_operator(implication).
binary_operator(less).
binary_operator(less_equal).
binary_operator(greater).
binary_operator(greater_equal).
binary_operator(add).
binary_operator(modulo).
binary_operator(minus).
binary_operator(multiplication).
binary_operator(div).
binary_operator(power_of).
binary_operator(member).
binary_operator(subset).
binary_operator(union).
binary_operator(intersection).
binary_operator(set_subtraction).

mk_expr_maplist(Solver,LS,GS,Op,Expr) :-
    mk_expr(Op,Solver,LS,GS,Expr).

% unary operators
mk_expr(b(Op,_Type,_Infos),Solver,LS,GS,Expr) :-
    Op =.. [Functor,A],
    unary_operator(Functor), !,
    mk_expr(A,Solver,LS,GS,AE),
    smt_solver_interface_call(Solver,mk_op_arglist(Functor,[AE],Expr)).
% binary operators
mk_expr(b(Op,_Type,_Infos),Solver,LS,GS,Expr) :-
    Op =.. [Functor,A,B],
    binary_operator(Functor), !,
    mk_expr(A,Solver,LS,GS,AE), mk_expr(B,Solver,LS,GS,BE),
    smt_solver_interface_call(Solver,mk_op(Functor,AE,BE,Expr)).
% conjunct and disjunct are deconstructed to lists
mk_expr(b(conjunct(A,B),pred,Infos),Solver,LS,GS,Expr) :- !,
    conjunction_to_list(b(conjunct(A,B),pred,Infos),List),
    maplist(mk_expr_maplist(Solver,LS,GS),List,ExprList),
    % truth/0 are removed from the conjunction during list / from list
    % the list might be empty now!
    (ExprList = []  -> smt_solver_interface_call(Solver,mk_bool_const(boolean_true,Expr)) ;
     ExprList = [X] -> Expr = X ;
     otherwise      -> smt_solver_interface_call(Solver,mk_op_arglist(conjunct,ExprList,Expr))).
mk_expr(b(disjunct(A,B),pred,Infos),Solver,LS,GS,Expr) :- !,
    disjunction_to_list(b(disjunct(A,B),pred,Infos),List),
    maplist(mk_expr_maplist(Solver,LS,GS),List,ExprList),
    smt_solver_interface_call(Solver,mk_op_arglist(disjunct,ExprList,Expr)).
% special case for if then else
mk_expr(b(if_then_else(Pred,A,B),_Type,_Infos),Solver,LS,GS,Expr) :- !,
    maplist(mk_expr_maplist(Solver,LS,GS),[Pred,A,B],ArgList),
    smt_solver_interface_call(Solver,mk_op_arglist(if_then_else,ArgList,Expr)).
% special case for partition
mk_expr(b(partition(S,Sets),_Type,_Infos),Solver,LS,GS,Expr) :- !,
    % definition: sets are disjoint and S is the union
    maplist(mk_expr_maplist(Solver,LS,GS),Sets,SetExprs),
    pairwise_union(SetExprs,Solver,UnionOfSets),
    mk_expr(S,Solver,LS,GS,SExpr),
    smt_solver_interface_call(Solver,mk_op(equal,SExpr,UnionOfSets,EqualToUnion)), % S = Union(Sets)
    % enforce sets pairwise disjoint (= intersection is empty)
    get_texpr_type(S,Type),
    smt_solver_interface_call(Solver,mk_empty_set(Type,EmptySet)),
    pairwise_disjoint(SetExprs,Solver,EmptySet,PWConstraint),
    smt_solver_interface_call(Solver,mk_op_arglist(conjunct,[EqualToUnion,PWConstraint],Expr)).
% quantifiers
mk_expr(b(forall(IDs,b(truth,pred,_),B),pred,_Infos),Solver,LS,GS,Expr) :- !,
    create_bounded_smt_state(Solver,IDs,State,SMTExprs),
    append(State,LS,NLS),
    mk_expr(B,Solver,NLS,GS,BE),
    smt_solver_interface_call(Solver,mk_quantifier(forall,SMTExprs,BE,Expr)).
mk_expr(b(forall(IDs,A,B),pred,_Infos),Solver,LS,GS,Expr) :- !,
    create_bounded_smt_state(Solver,IDs,State,SMTExprs),
    append(State,LS,NLS),
    mk_expr(A,Solver,NLS,GS,AE), mk_expr(B,Solver,NLS,GS,BE),
    smt_solver_interface_call(Solver,mk_op(implication,AE,BE,InnerExpr)),
    smt_solver_interface_call(Solver,mk_quantifier(forall,SMTExprs,InnerExpr,Expr)).
mk_expr(b(exists(IDs,B),pred,_Infos),Solver,LS,GS,Expr) :- !,
    create_bounded_smt_state(Solver,IDs,State,SMTExprs),
    append(State,LS,NLS),
    mk_expr(B,Solver,NLS,GS,InnerExpr),
    smt_solver_interface_call(Solver,mk_quantifier(exists,SMTExprs,InnerExpr,Expr)).
% equality is a special case as it needs to distinguish
% booleans and other types
mk_expr(b(equal(A,B),pred,_Infos),Solver,LS,GS,Expr) :- !,
    mk_expr(A,Solver,LS,GS,AE), mk_expr(B,Solver,LS,GS,BE),
    get_texpr_type(A,Type),
    (Type = boolean -> smt_solver_interface_call(Solver,mk_op(equivalence,AE,BE,Expr)) ;
     otherwise      -> smt_solver_interface_call(Solver,mk_op(equal,AE,BE,Expr))).
% records
mk_expr(b(rec(ListOfFields),Type,_Infos),Solver,LS,GS,Expr) :- !,
    findall(FieldExpr, member(field(_,FieldExpr),ListOfFields), FieldExprs),
    maplist(mk_expr_maplist(Solver,LS,GS),FieldExprs,SMTFieldExprs),
    smt_solver_interface_call(Solver,mk_record_const(Type,SMTFieldExprs,Expr)).
mk_expr(b(record_field(Record,FieldName),Type,_Infos),Solver,LS,GS,Expr) :- !,
    get_texpr_type(Record,record(RFields)),
    nth0(FieldNr,RFields,field(FieldName,Type)),
    % debug_format(9,'Field: ~w Type: ~w~n-> FieldNr: ~w~n',[FieldName,RFields,FieldNr]),
    mk_expr(Record,Solver,LS,GS,SMTRecordExpr),
    smt_solver_interface_call(Solver,mk_record_field(SMTRecordExpr,FieldNr,FieldName,Expr)).
% couple construction
mk_expr(b(couple(A,B),Type,_Infos),Solver,LS,GS,Expr) :- !,
    mk_expr(A,Solver,LS,GS,AE), mk_expr(B,Solver,LS,GS,BE),
    smt_solver_interface_call(Solver,mk_couple(Type,AE,BE,Expr)).
% constants and identifiers
mk_expr(b(identifier(Id),_Type,_Infos),_Solver,LS,GS,Expr) :- !,
    lookup(Id,LS,GS,Expr).
mk_expr(b(integer(Int),integer,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_int_const(Int,Expr)).
mk_expr(b(min_int,integer,_Infos),Solver,_LS,_GS,Expr) :- !,
    get_preference(minint,MinInt),
    smt_solver_interface_call(Solver,mk_int_const(MinInt,Expr)).
mk_expr(b(max_int,integer,_Infos),Solver,_LS,_GS,Expr) :- !,
    get_preference(maxint,MaxInt),
    smt_solver_interface_call(Solver,mk_int_const(MaxInt,Expr)).
mk_expr(b(boolean_true,boolean,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_bool_const(boolean_true,Expr)).
mk_expr(b(boolean_false,boolean,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_bool_const(boolean_false,Expr)).
mk_expr(b(string(S),string,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_string_const(S,Expr)).
mk_expr(b(truth,pred,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_bool_const(boolean_true,Expr)).
mk_expr(b(falsity,pred,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_bool_const(boolean_false,Expr)).
mk_expr(b(bool_set,set(boolean),_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_bool_const(boolean_true,True)),
    smt_solver_interface_call(Solver,mk_bool_const(boolean_false,False)),
    smt_solver_interface_call(Solver,mk_set([True,False],Expr)).
mk_expr(b(set_extension(List),set(T),_Infos),Solver,LS,GS,Expr) :- !,
    maplist(mk_expr_maplist(Solver,LS,GS),List,SubExprs),
    (SubExprs = [] -> smt_solver_interface_call(Solver,mk_empty_set(set(T),Expr)) ;
     otherwise     -> smt_solver_interface_call(Solver,mk_set(SubExprs,Expr))).
mk_expr(b(empty_set,Type,_Infos),Solver,_LS,_GS,Expr) :- !,
    smt_solver_interface_call(Solver,mk_empty_set(Type,Expr)).
% given values
mk_expr(b(value(V),Type,_Infos),Solver,LS,GS,Expr) :- !,
    mk_value(Type,V,Solver,LS,GS,Expr).
% ast nodes that can be ignored
mk_expr(b(convert_bool(P),boolean,_Infos),Solver,LS,GS,Expr) :- !,
    mk_expr(P,Solver,LS,GS,Expr).
% lets
mk_expr(b(let_expression(ListOfIDs,ListOfDefinitions,LetExpression),_Type,_I),Solver,LS,GS,Expr) :-
    create_let_state(ListOfIDs,ListOfDefinitions,Solver,LS,GS,NewLS),
    mk_expr(LetExpression,Solver,NewLS,GS,Expr).
mk_expr(b(let_predicate(ListOfIDs,ListOfDefinitions,LetExpression),pred,_I),Solver,LS,GS,Expr) :-
    create_let_state(ListOfIDs,ListOfDefinitions,Solver,LS,GS,NewLS),
    mk_expr(LetExpression,Solver,NewLS,GS,Expr).
% error case
mk_expr(Expr,_,_,_,_) :- !,
    add_error_fail(smt_solvers_interface,'mk_expr failed',Expr).

create_let_state([],[],_,S,_,S).
create_let_state([ID|IDs],[Def|Defs],Solver,StateSoFar,GS,State) :-
    get_texpr_id(ID,UnwrappedId),
    mk_expr(Def,Solver,StateSoFar,GS,Expr),
    create_let_state(IDs,Defs,Solver,[id(UnwrappedId,Expr)|StateSoFar],GS,State).

mk_value_maplist(Type,Solver,LS,GS,Val,Expr) :-
    mk_value(Type,Val,Solver,LS,GS,Expr).
mk_value(integer,int(Int),Solver,_LS,_GS,Expr) :-
    smt_solver_interface_call(Solver,mk_int_const(Int,Expr)).
mk_value(boolean,pred_false,Solver,_LS,_GS,Expr) :-
    smt_solver_interface_call(Solver,mk_bool_const(boolean_false,Expr)).
mk_value(boolean,pred_true,Solver,_LS,_GS,Expr) :-
    smt_solver_interface_call(Solver,mk_bool_const(boolean_true,Expr)).
% only supported thanks to z3 full sets
mk_value(set(integer),global_set('INTEGER'),z3,_LS,_GS,Expr) :-
    smt_solver_interface_call(z3,mk_full_set(set(integer),Expr)).
mk_value(set(global(X)),global_set(X),_Solver,LS,GS,Expr) :-
    lookup(X,LS,GS,Expr).
mk_value(set(X),avl_set(Val),Solver,LS,GS,Expr) :-
    expand_custom_set_to_list_now(avl_set(Val),Values),
    mk_value(set(X),Values,Solver,LS,GS,Expr).
mk_value(set(X),List,Solver,LS,GS,Expr) :-
    maplist(mk_value_maplist(X,Solver,LS,GS),List,SubExprs),
    (SubExprs = [] -> smt_solver_interface_call(Solver,mk_empty_set(set(X),Expr)) ;
     otherwise     -> smt_solver_interface_call(Solver,mk_set(SubExprs,Expr))).
mk_value(couple(AT,BT),(A,B),Solver,LS,GS,Expr) :-
    mk_value(AT,A,Solver,LS,GS,AE),
    mk_value(BT,B,Solver,LS,GS,BE),
    smt_solver_interface_call(Solver,mk_couple(couple(AT,BT),AE,BE,Expr)).
mk_value(global(Name),fd(Num,Name),_Solver,LS,GS,Expr) :-
    is_b_global_constant(Name,Num,Constant),
    lookup(Constant,LS,GS,Expr).
% error case
mk_value(Type,Value,_,_,_,_) :- !,
    add_error_fail(smt_solvers_interface,'mk_value failed',[Type,Value]).

pairwise_union([S],_Solver,S) :- !.
pairwise_union([S|Sets],Solver,Expr) :-
    pairwise_union(Sets,Solver,SS),
    smt_solver_interface_call(Solver,mk_op(union,S,SS,Expr)).

pairwise_disjoint([H,T],Solver,EmptySet,Constraint) :- !,
    pairwise_disjoint_mk_expr(Solver,H,T,EmptySet,Constraint).
pairwise_disjoint([H|T],Solver,EmptySet,Constraint) :-
    pairwise_disjoint(H,T,Solver,EmptySet,SubConstraints),
    pairwise_disjoint(T,Solver,EmptySet,SubConstraint),
    smt_solver_interface_call(Solver,mk_op_arglist(conjunct,[SubConstraint|SubConstraints],Constraint)).

pairwise_disjoint(E,[Last],Solver,EmptySet,[C]) :- !,
    pairwise_disjoint_mk_expr(Solver,E,Last,EmptySet,C).
pairwise_disjoint(E,[H|T],Solver,EmptySet,[C|Cs]) :-
    pairwise_disjoint_mk_expr(Solver,E,H,EmptySet,C),
    pairwise_disjoint(E,T,Solver,EmptySet,Cs).

pairwise_disjoint_mk_expr(Solver,A,B,EmptySet,PWConstraint) :-
    smt_solver_interface_call(Solver,mk_op(intersection,A,B,Inter)),
    smt_solver_interface_call(Solver,mk_op(equal,Inter,EmptySet,PWConstraint)).

mk_sort(Solver,Set) :-
    smt_solver_interface_call(Solver,mk_sort(Set)).

setup_sets(Solver,GlobalIdentifiers) :-
    % deferred sets -> only create a sort
    findall(X,b_global_deferred_set(X),DeferredSets),
    maplist(mk_sort(Solver),DeferredSets),
    mk_deferred_set_identifiers(Solver,DeferredSets,DefSetIdentifiers),
    % enumerated sets -> create a sort
    % & set up distinct identifiers
    b_get_global_enumerated_sets(EnumeratedSets),
    % can not use setof / bagof for some reason
    list_to_set(EnumeratedSets,EnumeratedSetsNoDuplicates),
    maplist(setup_enumerated_set(Solver),EnumeratedSetsNoDuplicates,IdentifierLists),
    append([DefSetIdentifiers|IdentifierLists],GlobalIdentifiers).

% currently only functional for z3
mk_deferred_set_identifiers(cvc4,_,[]).
mk_deferred_set_identifiers(z3,DeferredSets,DefSetIdentifiers) :-
    maplist(mk_deferred_set_identifiers2,DeferredSets,DefSetIdentifiers).
mk_deferred_set_identifiers2(SetName,id(SetName,SetExpression)) :-
    smt_solver_interface_call(z3,mk_full_set(set(global(SetName)),SetExpression)).
setup_enumerated_set(Solver,SetName,[SetIdentifier|Identifiers]) :-
    % find all global constants belonging to the given set
    findall(b(identifier(Cst),global(SetName),[]),
            is_b_global_constant(SetName,_,Cst),
            SetElements),
    b_global_set_cardinality(SetName,Cardinality),
    % make a new sort and create identifiers
    smt_solver_interface_call(Solver,mk_sort_with_cardinality(SetName,Cardinality)),
    create_smt_state(Solver,SetElements,Identifiers,Exprs),
    smt_solver_interface_call(Solver,setup_enumerated_set(Exprs)),
    SetIdentifier = id(SetName,SetExpression),
    smt_solver_interface_call(Solver,mk_set(Exprs,SetExpression)).

create_smt_state(Solver,Identifiers,State,Exprs) :-
    exclude(is_external_function_identifier,Identifiers,LessIdentifiers),
    maplist(create_smt_identifier(Solver),LessIdentifiers,State,Exprs).
create_bounded_smt_state(Solver,Identifiers,State,Exprs) :-
    maplist(create_bounded_smt_identifier(Solver),Identifiers,State,Exprs).
create_smt_identifier(Solver,b(identifier(Id),Type,_Infos),id(Id,Expr),Expr) :-
    smt_solver_interface_call(Solver,mk_var(Type,Id,Expr)).
create_bounded_smt_identifier(Solver,b(identifier(Id),Type,_Infos),id(Id,Expr),Expr) :-
    smt_solver_interface_call(Solver,mk_bounded_var(Type,Id,Expr)).

is_external_function_identifier(Id) :-
    get_texpr_id(Id,Name),
    is_external_function_name(Name).

lookup(Id,[],[id(Id,Expr)|_],Expr) :- !.
lookup(Id,[],[_|T],Expr) :- lookup(Id,[],T,Expr).
lookup(Id,[id(Id,Expr)|_],_,Expr) :- !.
lookup(Id,[_|T],GS,Expr) :- lookup(Id,T,GS,Expr).

contains_unsupported_type_or_expression(Solver,TExpr) :-
    contains_unsupported_type_or_expression(Solver,TExpr,_).
contains_unsupported_type_or_expression(Solver,TExpr,Unsupported) :-
    % TODO: shouldnt this expansion by done automatically,
    % because map_over_typed_bexpr is a meta-predicate?
    map_over_typed_bexpr(smt_solvers_interface:contains_unsupported_aux(Solver),TExpr,Unsupported).
contains_unsupported_aux(Solver,TExpr,Expr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    unsupported_expr_or_type(Expr,Type,Solver).

% these expressions are not supported at all
unsupported_expr_or_type(general_union(_),_,_).
unsupported_expr_or_type(general_intersection(_),_,_).
unsupported_expr_or_type(general_product(_,_,_),_,_).
unsupported_expr_or_type(general_sum(_,_,_),_,_).
unsupported_expr_or_type(composition(_,_),_,_).
unsupported_expr_or_type(first_of_pair(_),_,_).
unsupported_expr_or_type(second_of_pair(_),_,_).
unsupported_expr_or_type(iteration(_,_),_,_).
unsupported_expr_or_type(closure(_),_,_).
unsupported_expr_or_type(pow_subset(_),_,_).
unsupported_expr_or_type(pow1_subset(_),_,_).
unsupported_expr_or_type(fin_subset(_),_,_).
unsupported_expr_or_type(fin1_subset(_),_,_).
unsupported_expr_or_type(external_pred_call(_,_),_,_).
unsupported_expr_or_type(external_function_call(_,_),_,_).
unsupported_expr_or_type(function(Name,_),_,_) :-
    is_external_function_name(Name).
% the following exprs should have been removed by a rewrite rule
unsupported_expr_or_type(min(_),_,_).
unsupported_expr_or_type(max(_),_,_).
unsupported_expr_or_type(card(_),_,_).
unsupported_expr_or_type(finite(_),_,_).
unsupported_expr_or_type(comprehension_set(_,_),_,_).
unsupported_expr_or_type(interval(_,_),_,_).
unsupported_expr_or_type(integer_set(_),_,_).
unsupported_expr_or_type(parallel_product(_,_),_,_).
unsupported_expr_or_type(direct_product(_,_),_,_).
unsupported_expr_or_type(domain_restriction(_,_),_,_).
unsupported_expr_or_type(range_restriction(_,_),_,_).
unsupported_expr_or_type(domain_subtraction(_,_),_,_).
unsupported_expr_or_type(range_subtraction(_,_),_,_).
unsupported_expr_or_type(overwrite(_,_),_,_).
unsupported_expr_or_type(cartesian_product(_,_),_,_).
unsupported_expr_or_type(image(_,_),_,_).
unsupported_expr_or_type(range(_),_,_).
unsupported_expr_or_type(domain(_),_,_).
unsupported_expr_or_type(function(_,_),_,_).
unsupported_expr_or_type(not_equal(_,_),_,_).
unsupported_expr_or_type(not_member(_,_),_,_).
unsupported_expr_or_type(relations(_,_),_,_).
unsupported_expr_or_type(partial_function(_,_),_,_).
unsupported_expr_or_type(total_function(_,_),_,_).
unsupported_expr_or_type(partial_injection(_,_),_,_).
unsupported_expr_or_type(total_injection(_,_),_,_).
unsupported_expr_or_type(partial_surjection(_,_),_,_).
unsupported_expr_or_type(total_surjection(_,_),_,_).
unsupported_expr_or_type(total_relation(_,_),_,_).
unsupported_expr_or_type(surjection_relation(_,_),_,_).
unsupported_expr_or_type(total_surjection_relation(_,_),_,_).
% certain values are not supported yet
unsupported_expr_or_type(value(X),_,_) :- \+ ground(X). % can not translate unknown values
unsupported_expr_or_type(value(couple(A,B)),couple(AT,BT),Solver) :-
    unsupported_expr_or_type(value(A),AT,Solver) ;
    unsupported_expr_or_type(value(B),BT,Solver).
unsupported_expr_or_type(value(closure(_,_,_)),_,_).
unsupported_expr_or_type(value(global_set('INTEGER')),_,cvc4).
unsupported_expr_or_type(value(_),Type,_) :-
    unsupported_value_type(Type).
unsupported_expr_or_type(value(fd(Num,SetName)),global(SetName),_) :-
    \+ is_b_global_constant(SetName,Num,_). % has been invented by prob but is unnamed
% cvc4 does not support generating a full set of a sort, thus there
% is currently no way of expressing a complete deferred set
unsupported_expr_or_type(identifier(SetName),set(global(SetName)),cvc4).
% some types are not supported / can not be translated at the moment
unsupported_expr_or_type(_,Type,Solver) :- unsupported_type(Type,Solver).
unsupported_type(string,z3).
unsupported_type(any,_).
unsupported_type(seq(_),_).
unsupported_type(set(X),S) :-
    unsupported_type(X,S).
unsupported_type(couple(A,B),S) :-
    unsupported_type(A,S) ; unsupported_type(B,S).
% currently unsupported values
unsupported_value_type(set(T)) :-
    unsupported_value_type(T).
unsupported_value_type(couple(A,B)) :-
    unsupported_value_type(A) ; unsupported_value_type(B).
unsupported_value_type(global(S)) :-
    unfixed_deferred_set(S).
