% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bmc,[bmc_symbolic_model_check/1,
               get_constraint/2,
               constraint_transitions/2,
               replay_counter_example/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'A finite-state symbolic model checker based on BMC.').

:- use_module(library(lists), [append/2,maplist/3]).

:- use_module(extension('counter/counter'),[counter_init/0,
                                            new_counter/1]).

:- use_module(probsrc(before_after_predicates), [before_after_predicate_list_disjunct_with_equalities/3,
                                                 before_after_predicate_with_equalities/3,
                                                 before_after_predicate_for_operation/2]).
:- use_module(probsrc(translate), [translate_bexpression/2]).
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,
                                 b_get_properties_from_machine/1,
                                 b_get_machine_operation/4,
                                 b_get_machine_constants/1,
                                 b_get_machine_variables/1,
                                 b_get_initialisation_from_machine/2,
                                 b_specialized_invariant_for_op/2,
                                 get_proven_invariant/2]).
:- use_module(probsrc(state_space),[state_space_reset/0]).
:- use_module(probsrc(preferences),[get_preference/2]).
:- use_module(probsrc(bsyntaxtree), [conjunct_predicates/2,
                                     disjunct_predicates/2,
                                     disjunction_to_list/2,
                                     transform_bexpr/3,
                                     syntaxtransformation/5,
                                     create_negation/2,
                                     get_texpr_expr/2,
                                     get_texpr_type/2,
                                     unique_typed_id/3,
                                     find_typed_identifier_uses/2]).
:- use_module(symbolic_model_checker(predicate_handling), [get_initial_state_predicate/1,
                                                           prime_predicate/2,
                                                           create_state_predicate/2,
                                                           create_constants_state_predicate/2,
                                                           deprime_state/2,
                                                           prime_n_times/3]).
:- use_module(symbolic_model_checker(solver_handling), [solve/2]).

bmc_symbolic_model_check(Res) :-
    counter_init, new_counter(ic3solvercalls),
    bmc_symbolic_model_check(0,Res).

bmc_symbolic_model_check(25,limit_reached) :- !.
bmc_symbolic_model_check(K,Res) :-
    format('K = ~w~n', [K]),
    get_constraint(K,Constraint),
    % translate_bexpression(Constraint,PPConstraint),
    % format('BMC Constraint: ~w~n',[PPConstraint]),
    solve(Constraint,Result),
    bmc_symbolic_model_check_aux(Result,K,Res).

bmc_symbolic_model_check_aux(solution(Sol),_K,Res) :- !,
    Res = counterexample_found,
    replay_counter_example(Sol).
bmc_symbolic_model_check_aux(contradiction_found(_),K,Res) :- !,
    K2 is K + 1,
    bmc_symbolic_model_check(K2,Res).
bmc_symbolic_model_check_aux(_,_K,solver_and_provers_too_weak).

get_constraint(K,Base) :-
    get_preference(use_po,false), !,
    get_initial_state_predicate(I),
    b_get_invariant_from_machine(P),
    K2 is K - 1,
    constraint_transitions(K2,Transitions),
    constraint_property(K,P,Property),
    conjunct_predicates([I,Transitions,Property],Base).
get_constraint(K,Base) :-
    get_preference(use_po,true), !,
    proof_supported_transition_predicate(K,Base).

proof_supported_transition_predicate(0,Constraint) :- !,
    get_initial_state_predicate(I),
    b_get_invariant_from_machine(P),
    create_negation(P,NegP),
    conjunct_predicates([I,NegP],Constraint).
proof_supported_transition_predicate(N,Constraint) :-
    N2 is N - 2, N1 is N - 1,
    get_initial_state_predicate(I),
    constraint_transitions(N2,TWoProp),
    proof_supported_transition_predicate_aux(N1,PSPred),
    conjunct_predicates([I,TWoProp,PSPred],Constraint).

proof_supported_transition_predicate_aux(N,TPred) :-
    findall(op(OpName,Pred),(before_after_predicate_for_operation(OpName,Pred)),Ops),
    maplist(op_to_step_constraint,Ops,SingleSteps),
    disjunct_predicates(SingleSteps,Step),
    prime_n_times(N,Step,TPred).

op_to_step_constraint(op(OpName,BaPred),StepConstraint) :-
    (b_specialized_invariant_for_op(OpName,UncheckedInvariant)
     -> true ; b_get_invariant_from_machine(UncheckedInvariant)),
    create_negation(UncheckedInvariant,NegUncheckedInvariant),
    prime_predicate(NegUncheckedInvariant,PrimedNegUncheckedInvariant),
    conjunct_predicates([BaPred,PrimedNegUncheckedInvariant],StepConstraint).

replay_counter_example(Sol) :-
    state_space_reset,replay_aux(Sol,_).
replay_aux(Sol,CurId) :-
    create_constants_state_predicate(Sol,ConstantsStatePred), !,
    b_trace_checking:find_successor_state(CurId,ConstantsStatePred,SuccID,_),
    replay_aux2(Sol,SuccID).
replay_aux(Sol,CurId) :-
    replay_aux2(Sol,CurId).
replay_aux2([],_).
replay_aux2(Sol,CurId) :-
    create_state_predicate(Sol,StatePred),
    b_trace_checking:find_successor_state(CurId,StatePred,SuccID,_),
    deprime_state(Sol,DPSol),
    replay_aux2(DPSol,SuccID).

constraint_transitions(K,Transitions) :-
    get_preference(use_po,false), !,
    transition_predicate(T),
    constraint_transitions(K,T,Transitions).
constraint_transitions(K,Transitions) :-
    get_preference(use_po,true), !,
    findall(op(OpName,Pred),(before_after_predicate_for_operation(OpName,Pred)),Ops),
    maplist(op_to_transition,Ops,SingleSteps),
    disjunct_predicates(SingleSteps,Step),
    constraint_transitions(K,Step,Transitions).

op_to_transition(op(OpName,BaPred),StepConstraint) :-
    (get_proven_invariant(OpName,ProvenInvariant)
     -> true ; ProvenInvariant = b(truth,pred,[])),
    prime_predicate(ProvenInvariant,PrimedProvenInvariant),
    conjunct_predicates([BaPred,PrimedProvenInvariant],StepConstraint).

constraint_transitions(K,T,Transitions) :-
    constraint_transitions_aux(K,T,Conjuncts),
    conjunct_predicates(Conjuncts,Transitions).
constraint_transitions_aux(X,_,[]) :- X < 0, !.
constraint_transitions_aux(0,T,[T]) :- !.
constraint_transitions_aux(N,T,[T|Rest]) :-
    prime_predicate(T,PT),
    N2 is N - 1,
    constraint_transitions_aux(N2,PT,Rest).

constraint_property(K,P,Property) :-
    create_negation(P,NegP),
    constraint_property_aux(K,NegP,Disjuncts),
    disjunct_predicates(Disjuncts,Property).
constraint_property_aux(0,P,[P]) :- !.
constraint_property_aux(N,P,[P|Rest]) :-
    prime_predicate(P,PP),
    N2 is N - 1,
    constraint_property_aux(N2,PP,Rest).

transition_predicate(TransitionPredicate) :-
    findall(Body,b_get_machine_operation(_Name,_Results,_Parameters,Body),ListOfBodies),
    b_get_machine_variables(Variables),
    b_get_machine_constants(Constants),
    append(Variables,Constants,VarsAndConsts),
    before_after_predicate_list_disjunct_with_equalities(ListOfBodies,VarsAndConsts,TransitionPredicate).
