% (c) 2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(predicate_abstraction, [get_abstract_domain/1,
                                  concrete_state_to_abstract_state/3]).

:- use_module(probsrc(module_information)).
:- module_info(group,symbolic_model_checker).
:- module_info(description,'An abstract domain using predicate abstraction.').

:- use_module(library(lists), [append/2]).
:- use_module(library(ordsets), [list_to_ord_set/2]).

:- use_module(probsrc(bsyntaxtree), [create_implication/3,
                                     create_negation/2,
                                     safe_create_texpr/3,
                                     conjunct_predicates/2,
                                     conjunction_to_list/2]).
:- use_module(probsrc(bmachine),[b_get_invariant_from_machine/1,
                                 b_get_machine_variables/1]).

:- use_module(symbolic_model_checker(solver_handling)).
:- use_module(symbolic_model_checker(predicate_handling)).

% the abstract domain consists of:
% 1. the conjuncts of the invariant
% 2. the conjuncts of the predicate describing the initial state
% 3. pairwise inequalities x < y for all variables
get_abstract_domain(Set) :-
    get_conjuncts_of_invariant(Invariant),
    get_conjuncts_of_initial_state(Init),
    pairwise_inequalities_of_variables(PW),
    append([Invariant,Init,PW],L),
    list_to_ord_set(L, Set).

get_conjuncts_of_invariant(C) :-
    b_get_invariant_from_machine(I),
    conjunction_to_list(I,C).

get_conjuncts_of_initial_state(C) :-
    get_initial_state_predicate(I),
    conjunction_to_list(I,C).

pairwise_inequalities_of_variables(PW) :-
    b_get_machine_variables(Vars),
    pairwise_inequalities_of_variables(Vars,PW).

pairwise_inequalities_of_variables([],[]).
pairwise_inequalities_of_variables([V1|T],PW) :- !,
    pairwise_inequalities_of_variables(T,V1,PWTemp),
    pairwise_inequalities_of_variables(T,PWTemp2),
    append(PWTemp,PWTemp2,PW).
pairwise_inequalities_of_variables([],_,[]).
pairwise_inequalities_of_variables([V1|Vs],V2,[PW|PWs]) :-
    safe_create_texpr(less(V1,V2),predicate,PW),
    pairwise_inequalities_of_variables(Vs,V2,PWs).

concrete_state_to_abstract_state(Dom,CState,AbsState) :-
    concrete_state_to_abstract_state_aux(Dom,CState,AbsStateConjuncts),
    conjunct_predicates(AbsStateConjuncts,AbsState).
concrete_state_to_abstract_state_aux([],_ConcreteState,[]).
concrete_state_to_abstract_state_aux([ADomElement|Elements],ConcreteState,[ADomElement|T]) :-
    create_implication(ConcreteState,ADomElement,Implies),
    solve_negated(Implies,contradiction_found(_)), !,
    concrete_state_to_abstract_state_aux(Elements,ConcreteState,T).
concrete_state_to_abstract_state_aux([ADomElement|Elements],ConcreteState,[NegDomElement|T]) :-
    create_negation(ADomElement,NegDomElement),
    create_implication(ConcreteState,NegDomElement,Implies),
    solve_negated(Implies,contradiction_found(_)), !,
    concrete_state_to_abstract_state_aux(Elements,ConcreteState,T).
concrete_state_to_abstract_state_aux([_|Elements],ConcreteState,T) :-
    concrete_state_to_abstract_state_aux(Elements,ConcreteState,T).
