% (c) 2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(xml2b,[convert_xml_to_b/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,external_functions).
:- module_info(description,'This module transforms XML into a B representation.').

:- use_module(library(xml)).
:- use_module(library(lists)).
:- use_module(custom_explicit_sets,[convert_to_avl/2]).
:- use_module(error_manager).

% type of generated XML elements:
/*      struct(
        recId: NATURAL1,
        pId:NATURAL,
        element:STRING,
        attributes: STRING +-> STRING,
        meta: STRING +-> STRING
        )
*/

% convert a code list to a B sequence/set of records
convert_xml_to_b(Codes,BContent) :-
     xml_parse(Codes,xml(_Atts,Content),[format(false)]), % format(false) means we also see all newlines
     !,
     reset_counter(xml2b_current_line), set_counter(xml2b_current_line,1),
     convert_xml_content_to_b(Content,0,1,_,BContent,[]).
convert_xml_to_b(_,_) :- add_error(convert_xml_to_b,'XML Parsing failed'),fail.


convert_xml_content_to_b([],_ParentId,NextId,OutId) --> {OutId=NextId}.
convert_xml_content_to_b([H|T],ParentId,NextId,OutId) -->
   convert_xml_element(H,ParentId,NextId,Id2),
   convert_xml_content_to_b(T,ParentId,Id2,OutId).

% TO DO: do we need to parse namespaces...
% TO DO: store meta-information

convert_xml_element(Element, ParentId,NextId,OutId) --> %{nl,print(Element),nl},
    {xml_element(Element,Tag,Attributes,Content,LineNr)},
    !,
    [(int(NextId),
       rec([ % note: fields have to be in order
          field(attributes,BAttributes), % partial function of Attributes
          field(element,string(Tag)),
          field(meta,MetaAttributes),  % such as xmlLineNumber
          field(pId,int(ParentId)), % id of the parent XML record
          field(recId,int(NextId)) % id of the XML element/record
           ])
     )],
    {Id2 is NextId+1,
     number_codes(LineNr,LCC), atom_codes(LineNrAtom,LCC),
     convert_to_avl([(string(xmlLineNumber),string(LineNrAtom))],MetaAttributes),
     maplist(convert_xml_attributes,Attributes,BAttributesList),
     convert_to_avl(BAttributesList,BAttributes)
    },
     convert_xml_content_to_b(Content,NextId,Id2,OutId).
convert_xml_element(pcdata(_Codes),_,Id,Id) --> !. % should only happen when Codes consists solely of newlines
convert_xml_element(_,_,Id,Id) --> [].


xml_element(element(Tag,Attributes,Content),Tag,Attributes,Content,LineNr) :- get_counter(xml2b_current_line,LineNr).
xml_element(pcdata(Codes),'CText',['='(text,NewCodes)],[],LineNr) :-
    peel_and_count_newlines(Codes,C1),
    C1 \= [],
    get_counter(xml2b_current_line,LineNr),
    reverse(C1,RC),
    peel_and_count_newlines(RC,RC2), % remove newlines at end; try avoid B strings with newlines in them
    reverse(RC2,NewCodes).

convert_xml_attributes('='(Attr,Codes),(string(Attr),string(Atom))) :- !,atom_codes(Atom,Codes),
   count_newlines(Codes).
convert_xml_attributes(A,(string(error),string(A))) :- format('**** UNKNOWN ATTRIBUTE: ~w~n',[A]).


:- use_module(extension('counter/counter')).

count_newlines([]).
count_newlines([H|T]) :- (newline(H) -> inc_counter(xml2b_current_line) ; true), count_newlines(T).

peel_and_count_newlines([],[]).
peel_and_count_newlines([H|T],Res) :- newline(H),!, inc_counter(xml2b_current_line), peel_and_count_newlines(T,Res).
peel_and_count_newlines([H|T],Res) :- whitespace(H),!, peel_and_count_newlines(T,Res).
peel_and_count_newlines(R,R).


newline(10).
newline(13).

whitespace(9). % tab
whitespace(32).

xml2b_startup :- % call once at startup to ensure all counters exist
      counter_init,
      new_counter(xml2b_current_line).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(startup_prob,xml2b_startup,
                           'Initialise xml2b Counters.').
